<?php

/**
 * HCE project node request API.
 * Basic request API for network transport.
 *
 * @author bgv <bgv.hce@gmail.com>
 * @link http://hierarchical-cluster-engine.com/
 * @copyright Copyright &copy; 2013-2014 IOIX Ukraine
 * @license http://hierarchical-cluster-engine.com/license/
 * @package HCE project node API
 * @since 0.1
 */
namespace HCE\transport {

	require_once 'zmsg.php';
	require_once 'Constants.inc.php';
	require_once 'Response.inc.php';
	require_once 'Connection.inc.php';

	/**
	 * HCE Request class implements basic functionality of client-side of HCE Cluster Node application interaction
	 */
	class Request {
		/**
		 * Connection object
		 */
		protected $connection = NULL;

		/**
		 * Request message Id
		 */
		protected $messageId = NULL;

		/**
		 * Request body
		 */
		protected $messageBody = NULL;

		/**
		 * Request timeout ms
		 */
		protected $timeout = PROTOCOL_MSG_RESPONSE_TIMEOUT_DEFAULT;

		/**
		 * Request error code of last operation result
		 */
		protected $error = 0;

		/**
		 * Set connection
		 *
		 * @param $connection -
		 *        	connection object, if omitted or null, new admin localhost connection object created
		 *
		 */
		public function setConnection($connection = null) {
			if ($connection == NULL) {
				$this->connection = new \HCE\transport\Connection ();
			} else {
				$this->connection = $connection;
			}
		}

		/**
		 * Returns connection object
		 *
		 * @return connection object
		 */
		public function getConnection() {
			return $this->connection;
		}

		/**
		 * Set message Id
		 *
		 * @param $messageId -
		 *        	message unique Id
		 *
		 * @return boolean : true if Id is not empty and false in opposite
		 */
		public function setMessageId($messageId) {
			if ($messageId == '') {
				return false;
			} else {
				$this->messageId = $messageId;
				return true;
			}
		}

		/**
		 * Returns message Id
		 *
		 * @return message Id
		 */
		public function getMessageId() {
			return $this->messageId;
		}

		/**
		 * Set message body
		 *
		 * @param $messageBody -
		 *        	message body
		 *
		 */
		public function setMessageBody($messageBody) {
			$this->messageBody = $messageBody;
		}

		/**
		 * Returns message body
		 *
		 * @return message body
		 */
		public function getMessageBody() {
			return $this->messageBody;
		}

		/**
		 * Set error code
		 *
		 * @param $errorCode -
		 *        	error code
		 *
		 */
		public function setErrorCode($errorCode) {
			$this->error = $errorCode;
		}

		/**
		 * Returns erro code of last executed request
		 *
		 * @return error code
		 */
		public function getErrorCode() {
			return $this->error;
		}

		/**
		 * create instance and set the HCE connection
		 *
		 * @param $connection -
		 *        	connection object
		 *
		 */
		public function __construct($connection = null) {
			// Set connection
			$this->setConnection ( $connection );

			// Set message Id
			$this->setMessageId ( \HCE\transport\IdGenerator::getMessageId () );

			// Set message Body
			$this->setMessageBody ( '' );
		}

		/**
		 * delete HCE connection (disconnect zmq socket, free zmq socket and context)
		 */
		public function __destruct() {
			$this->connection = NULL;
		}

		/**
		 * Execute request and returns response object instance
		 *
		 * @param $msgBody -
		 *        	command message body to send
		 *
		 * @param $timeout -
		 *        	timeout value of response
		 *
		 * @return Response object
		 */
		public function execute($msgBody = NULL, $timeout = null) {
			// Reset error state
			$this->setErrorCode ( 0 );

			// New response object
			$response = new \HCE\transport\Response ();

			$message = array (
					'id' => $this->getMessageId (),
					'body' => ($msgBody === NULL ? $this->getMessageBody () : $msgBody)
			);
			// Send message
			$ret = $this->messageSend ( $message );

			// Check state
			if ($ret ['error'] == PROTOCOL_ERROR_OK) {
				// Receive response
				$responseItem = $this->messageReceive ( $timeout === null ? $this->timeout : $timeout );
				// Check state
				if ($responseItem ['error'] == PROTOCOL_ERROR_OK) {
					if ($responseItem ['messages'] !== NULL) {
						// Check response items corespondence for requested message and filter items set
						$responseItems = array ();
						$wrongItems = 0;
						foreach ( $responseItem ['messages'] as $responseItem ) {
							if ($responseItem ['id'] == $message ['id']) {
								$responseItems [] = $responseItem;
							} else {
								$wrongItems ++;
							}
						}
						if ($wrongItems > 0) {
							$response->setErrorCode ( PROTOCOL_ERROR_WRONG_RESPONSE_MSG_ID );
						}
						$response->setResponses ( $responseItems );
					}
				} else {
					// Set own error code
					$this->setErrorCode ( $responseItem ['error'] );
					// Set error code for response object
					$response->setErrorCode ( $responseItem ['error'] );
				}
			} else {
				// Set error code
				$this->setErrorCode ( $ret ['error'] );
				// Set error code for response object
				$response->setErrorCode ( $ret ['error'] );
			}

			return $response;
		}

		/**
		 * send HCE connection message
		 *
		 * @param $fields_array -
		 *        	array of message fields 'id' and 'body'
		 *
		 * @return HCE message fields array or error PROTOCOL_ERROR_FIELDS_NOT_SET - fields array not set proper way
		 */
		public function messageSend($fields_array) {
			// Init return HCE message array
			$ret = array (
					'error' => PROTOCOL_ERROR_OK,
					'message' => null
			);

			if (! isset ( $fields_array ['body'] ) || ! isset ( $fields_array ['id'] )) {
				$ret ['error'] = PROTOCOL_ERROR_FIELDS_NOT_SET;
			} else {
				// Create message
				$ret ['message'] = new \Zmsg ( $this->connection->socket );

				// Set fields
				$ret ['message']->wrap ( $fields_array ['body'], NULL );
				$ret ['message']->wrap ( $fields_array ['id'], NULL );

				// Send message
				$ret ['message']->send ();
			}

			return $ret;
		}

		/**
		 * receive HCE connection message
		 *
		 * @param $timeout -
		 *        	receive timeout, msec
		 *
		 * @return HCE responses array or PROTOCOL_ERROR_TIMEOUT - if timeout reached
		 */
		protected function messageReceive($timeout = PROTOCOL_MSG_RESPONSE_TIMEOUT_DEFAULT) {
			// Init return HCE response item array
			$ret = array (
					'error' => PROTOCOL_ERROR_OK,
					'messages' => array ()
			);

			// Poll socket for a reply, with timeout
			$read = $write = array ();
			$poll = new \ZMQPoll ();
			$poll->add ( $this->connection->socket, \ZMQ::POLL_IN );
			$events = $poll->poll ( $read, $write, $timeout );
			// Handle events
			if ($events) {
				foreach ( $read as $socket ) {
					$zmsg_r = new \Zmsg ( $socket );
					$zmsg_r->recv ();
					$ret ['messages'] [] = array (
							'error' => PROTOCOL_ERROR_OK,
							'message' => $zmsg_r,
							'id' => $zmsg_r->unwrap (),
							'body' => $zmsg_r->unwrap ()
					);
					// var_dump ( $ret );
				}
			} else {
				$ret ['error'] = PROTOCOL_ERROR_TIMEOUT;
			}
			$poll->remove ( $this->connection->socket );

			return $ret;
		}
	}
}

?>