"""
HCE project, Python bindings, Distributed Tasks Manager application.
Event objects definitions.

@package: dtm
@file dbi.py
@author Oleksii <developers.hce@gmail.com>
@link: http://hierarchical-cluster-engine.com/
@copyright: Copyright &copy; 2013-2014 IOIX Ukraine
@license: http://hierarchical-cluster-engine.com/license/
@since: 0.1
"""


import Constants as CONSTANTS
from flask import Flask
from flask.ext.sqlalchemy import SQLAlchemy  # pylint: disable-all
from sqlalchemy.sql import select
from sqlalchemy import delete
import sqlalchemy
from sqlalchemy.engine.reflection import Inspector
from sqlalchemy import create_engine, MetaData, Table
from sqlalchemy import schema, types
from sqlalchemy.pool import StaticPool
from sqlalchemy.pool import SingletonThreadPool
import app.Consts as APP_CONSTS
import logging
from threading import Lock


# Logger initialization
logger = logging.getLogger(APP_CONSTS.LOGGER_NAME)

#instantiate global database object
app = Flask(__name__)
db = SQLAlchemy(app)
lock = Lock()


##Common use wrapper class to interact with the ORM databases
#Provide CRUD interface (create, read, update, delete) entries
#
class DBI(object):


  ##constructor
  #initialize fields
  #
  #@param config_dic dict object instance from config section
  #
  def __init__(self, config_dic, bindName=None):
    db_name = config_dic["db_name"]
    """
    db_name - database name
    Examples:
    db_name=sqlite://               # in memory
    db_name=sqlite:///:memory:      # in memory
    db_name=sqlite:///del.db        # relative path to file
    db_name=sqlite:////tmp/del.db   # absolute path to file
    """
    #app.config['SQLALCHEMY_DATABASE_URI'] = "sqlite://%s" % db_name
    #app.config['SQLALCHEMY_DATABASE_URI'] = db_name + "?timeout=60&isolation_level=immediate"
    app.config['SQLALCHEMY_DATABASE_URI'] = db_name
    if bindName is not None:
      newBindItem = {bindName: db_name}
      if ('SQLALCHEMY_BINDS' not in app.config) or (app.config['SQLALCHEMY_BINDS'] is None):
        app.config['SQLALCHEMY_BINDS'] = newBindItem
      else:
        app.config['SQLALCHEMY_BINDS'].update(newBindItem)

    db.create_all()
    db.create_all(bind=bindName)

    """
    db = create_engine('sqlite://',
                    connect_args={'check_same_thread':False},
                    poolclass=StaticPool)
    """
    """
    db = create_engine('sqlite://',
                    poolclass=SingletonThreadPool)
    """

    """
    db = create_engine('sqlite:///:memory:', echo=True)
    metadata = schema.MetaData()
    metadata.bind = db
    metadata.create_all(checkfirst=True)
    """
    #db = create_engine('mysql://scott:tiger@localhost/foo')

    #print(vars(db))

    #db = create_engine(db_name, strategy="threadlocal")
    #db.begin()
    #self.conn = db.connect()

    # Workaround - disable sqlite use journal files on disk
    from sqlalchemy.engine import Engine
    from sqlalchemy import event
    @event.listens_for(Engine, "connect")
    def set_sqlite_pragma(dbapi_connection, connection_record):
      cursor = dbapi_connection.cursor()
      cursor.execute("PRAGMA journal_mode=MEMORY")
      cursor.execute("PRAGMA synchronous=OFF")
      cursor.close()

    self.errorCode = CONSTANTS.DBI_SUCCESS_CODE
    self.errorMessage = CONSTANTS.DBI_SUCCESS_MSG


  ##get error code
  #return error code last performed action (e.g. insert, update, etc.)
  #
  #@return error code last performed action (e.g. insert, update, etc.)
  #
  def getErrorCode(self):
    return self.errorCode


  ##get error message
  #return error message last performed action (e.g. insert, update, etc.)
  #
  #@return error message last performed action (e.g. insert, update, etc.)
  #
  def getErrorMsg(self):
    return self.errorMessage


  ##insert
  #insert objects into the database
  #
  #@param tasks - list of objects (not mandatory the same type) for insertion
  #@return the same list of objects
  #@error can be checked by the getError() method
  #
  def insert(self, tasks):
    with lock:
      self.errorCode = CONSTANTS.DBI_SUCCESS_CODE
      if hasattr(tasks, "__len__"):
        for task in tasks:
          db.session.add(task)
      else:
        db.session.add(tasks)
      try:
        db.session.commit()
      # Firstly, we check if exception was thrown by the DBI itself
      # If so we rollback DB
      except sqlalchemy.exc.SQLAlchemyError as err:
        db.session.rollback()
        self.errorCode = CONSTANTS.DBI_INSERT_ERROR_CODE
        self.errorMessage = CONSTANTS.DBI_INSERT_ERROR_MSG + ": " + err.message
        raise DBIErr(self.dbi.getErrorCode(), self.dbi.getErrorMsg())
      except Exception as err:
        self.errorCode = CONSTANTS.DBI_INSERT_ERROR_CODE
        self.errorMessage = CONSTANTS.DBI_INSERT_ERROR_MSG + ": " + err.message
        raise DBIErr(self.dbi.getErrorCode(), self.dbi.getErrorMsg())
      return tasks


  ##fetch
  #fetch objects from the database
  #
  #@param obj - list of objects (not mandatory the same type) for fetching
  #@param clause - sql clause for fetching the object from table
  #@return list of fetched objects
  #@error can be checked by the getError() method
  #
  def fetch(self, obj, clause):
    with lock:
      self.errorCode = CONSTANTS.DBI_SUCCESS_CODE
      fetched = [None]
      try:
        rows = db.session.query(type(obj)).filter(clause).all()
        if len(rows):
          fetched = rows
      # Firstly, we check if exception was thrown by the DBI itself
      # If so we rollback DB
      except sqlalchemy.exc.SQLAlchemyError as err:
        db.session.rollback()
        self.errorCode = CONSTANTS.DBI_FETCH_ERROR_CODE
        self.errorMessage = CONSTANTS.DBI_FETCH_ERROR_MSG + ": " + err.message
        raise DBIErr(self.dbi.getErrorCode(), self.dbi.getErrorMsg())
      except Exception as err:
        self.errorCode = CONSTANTS.DBI_FETCH_ERROR_CODE
        self.errorMessage = CONSTANTS.DBI_FETCH_ERROR_MSG + ": " + err.message
        raise DBIErr(self.dbi.getErrorCode(), self.dbi.getErrorMsg())
      return fetched


  ##sql
  #fetch objects from the database
  #
  #@param obj - list of objects (not mandatory the same type) for fetching
  #@param clause - sql clause for fetching the object from table
  #@return list of fetched objects
  #@error can be checked by the getError() method
  #
  def sql(self, obj, clause):
    with lock:
      self.errorCode = CONSTANTS.DBI_SUCCESS_CODE
      fetched = []
      try:
        rows = db.session.query(type(obj)).from_statement(clause).all()
        if len(rows):
          fetched = rows
      # Firstly, we check if exception was thrown by the DBI itself
      # If so we rollback DB
      except sqlalchemy.exc.SQLAlchemyError as err:
        db.session.rollback()
        self.errorCode = CONSTANTS.DBI_SQL_ERROR_CODE
        self.errorMessage = CONSTANTS.DBI_SQL_ERROR_MSG + ": " + err.message + ", clause:\n" + str(clause)
        raise DBIErr(self.dbi.getErrorCode(), self.dbi.getErrorMsg())
      except Exception as err:
        self.errorCode = CONSTANTS.DBI_SQL_ERROR_CODE
        self.errorMessage = CONSTANTS.DBI_SQL_ERROR_MSG + ": " + err.message + ", clause:\n" + str(clause)
        raise DBIErr(self.dbi.getErrorCode(), self.dbi.getErrorMsg())
      return fetched


  ##fetchAll
  #fetch all objects from the database
  #
  #@param tasks - object type for fetching
  #@return list of fetched objects
  #@error can be checked by the getError() method
  #
  def fetchAll(self, tasks):
    with lock:
      self.errorCode = CONSTANTS.DBI_SUCCESS_CODE
      fetched = []
      try:
        conn = db.engine.connect()
        if hasattr(tasks, "__len__"):
          for task in tasks:
            fetched.append(conn.execute(select([type(task)])).fetchall())
        else:
          fetched.append(conn.execute(select([type(tasks)])).fetchall())
      # Firstly, we check if exception was thrown by the DBI itself
      # If so we rollback DB
      except sqlalchemy.exc.SQLAlchemyError as err:
        db.session.rollback()
        self.errorCode = CONSTANTS.DBI_FETCH_ERROR_CODE
        self.errorMessage = CONSTANTS.DBI_FETCH_ERROR_MSG + ": " + err.message
        raise DBIErr(self.dbi.getErrorCode(), self.dbi.getErrorMsg())
      except Exception as err:
        self.errorCode = CONSTANTS.DBI_FETCH_ERROR_CODE
        self.errorMessage = CONSTANTS.DBI_FETCH_ERROR_MSG + ": " + err.message
        raise DBIErr(self.dbi.getErrorCode(), self.dbi.getErrorMsg())
      return fetched


  ##update
  #update objects in the database
  #
  #@param tasks - list of objects (not mandatory the same type) for updating
  #@param clause - sql clause for updating the object in the table
  #@return list of updated objects
  #@error can be checked by the getError() method
  #
  def update(self, obj, clause):
    with lock:
      self.errorCode = CONSTANTS.DBI_SUCCESS_CODE
      updated = []
      try:
        updated_task = db.session.query(type(obj)).filter(clause).first()
        if updated_task:
          attributes = [attr for attr in dir(obj) if not attr.startswith('__') and not attr.startswith('_') and getattr(obj, attr) is not None]
          #attributes = [attr for attr in dir(obj) if not attr.startswith('__') and not attr.startswith('_')]
          for attr in attributes:
            setattr(updated_task, attr, getattr(obj, attr))
          updated.append(updated_task)
        db.session.commit()
      # Firstly, we check if exception was thrown by the DBI itself
      # If so we rollback DB
      except sqlalchemy.exc.SQLAlchemyError as err:
        db.session.rollback()
        self.errorCode = CONSTANTS.DBI_UPDATE_ERROR_CODE
        self.errorMessage = CONSTANTS.DBI_UPDATE_ERROR_MSG + ": " + err.message
        raise DBIErr(self.dbi.getErrorCode(), self.dbi.getErrorMsg())
      except Exception as err:
        self.errorCode = CONSTANTS.DBI_UPDATE_ERROR_CODE
        self.errorMessage = CONSTANTS.DBI_UPDATE_ERROR_MSG + ": " + err.message
        raise DBIErr(self.dbi.getErrorCode(), self.dbi.getErrorMsg())
      return updated


  ##insertOnUpdate
  #insert new object or update object if exists in the table
  #@param tasks - list of objects (not mandatory the same type) for updating
  #@param clause - (deprecated) sql clause for updating the object in the table
  #@return list of updated objects
  #@error can be checked by the getError() method
  #
  def insertOnUpdate(self, tasks, clause):
    with lock:
      self.errorCode = CONSTANTS.DBI_SUCCESS_CODE
      try:
        if hasattr(tasks, "__len__"):
          for task in tasks:
            db.session.merge(task)
        else:
          db.session.merge(tasks)
        db.session.commit()
      # Firstly, we check if exception was thrown by the DBI itself
      # If so we rollback DB
      except sqlalchemy.exc.SQLAlchemyError as err:
        db.session.rollback()
        self.errorCode = CONSTANTS.DBI_INSERT_ERROR_CODE
        self.errorMessage = CONSTANTS.DBI_INSERT_ERROR_MSG + ": " + err.message
        raise DBIErr(self.dbi.getErrorCode(), self.dbi.getErrorMsg())
      except Exception as err:
        self.errorCode = CONSTANTS.DBI_INSERT_ERROR_CODE
        self.errorMessage = CONSTANTS.DBI_INSERT_ERROR_MSG + ": " + err.message
        raise DBIErr(self.dbi.getErrorCode(), self.dbi.getErrorMsg())
      return tasks


  ##delete
  #delete objects from the database
  #@param obj- list of objects (not mandatory the same type) for deleting
  #@param clause - sql clause for deleting the object from the table
  #@return list of deleted objects
  #@error can be checked by the getError() method
  #
  def delete(self, obj, clause):
    with lock:
      self.errorCode = CONSTANTS.DBI_SUCCESS_CODE
      deleted = []
      try:
        deleted_task = db.session.query(type(obj)).filter(clause).first()
        if deleted_task:
          deleted.append(deleted_task)
          db.session.delete(deleted_task)
        db.session.commit()
      # Firstly, we check if exception was thrown by the DBI itself
      # If so we rollback DB
      except sqlalchemy.exc.SQLAlchemyError as err:
        db.session.rollback()
        self.errorCode = CONSTANTS.DBI_DELETE_ERROR_CODE
        self.errorMessage = CONSTANTS.DBI_DELETE_ERROR_MSG + ": " + err.message
        raise DBIErr(self.dbi.getErrorCode(), self.dbi.getErrorMsg())
      except Exception as err:
        self.errorCode = CONSTANTS.DBI_DELETE_ERROR_CODE
        self.errorMessage = CONSTANTS.DBI_DELETE_ERROR_MSG + ": " + err.message
        raise DBIErr(self.dbi.getErrorCode(), self.dbi.getErrorMsg())
      return deleted


  ##deleteAll
  #delete all objects from the database
  #@param tasks - object type for deleting
  #@return list of deleted objects
  #@error can be checked by the getError() method
  #
  def deleteAll(self, tasks):
    with lock:
      deleted = []
      self.errorCode = CONSTANTS.DBI_SUCCESS_CODE
      try:
        conn = db.engine.connect()
        if hasattr(tasks, "__len__"):
          for task in tasks:
            conn.execute(type(task).delete())
        else:
          #conn.execute(type(tasks).delete())
          #db.session.query(delete(tasks))
          #db.session.delete(tasks)
          ress = db.session.query(tasks).filter(tasks.UThreads == "").first()
          #ress = conn.execute(select([type(tasks)])).fetchall()
          for res in ress:
            db.session.delete(res)
        db.session.commit()
      # Firstly, we check if exception was thrown by the DBI itself
      # If so we rollback DB
      except sqlalchemy.exc.SQLAlchemyError as err:
        db.session.rollback()
        self.errorCode = CONSTANTS.DBI_DELETE_ERROR_CODE
        self.errorMessage = CONSTANTS.DBI_DELETE_ERROR_MSG + ": " + err.message
        raise DBIErr(self.dbi.getErrorCode(), self.dbi.getErrorMsg())
      except Exception as err:
        self.errorCode = CONSTANTS.DBI_DELETE_ERROR_CODE
        self.errorMessage = CONSTANTS.DBI_DELETE_ERROR_MSG + ": " + err.message
        raise DBIErr(self.dbi.getErrorCode(), self.dbi.getErrorMsg())
      return deleted


# #Class is used to inform about error of DBI object
#
#
class DBIErr(Exception):
  
  def __init__(self, errCode, message):
    Exception.__init__(self, message)
    self.errCode = errCode



