'''
Created on Apr 8, 2014

@package: dc
@author: scorp, bgv
@link: http://hierarchical-cluster-engine.com/
@copyright: Copyright &copy; 2013-2014 IOIX Ukraine
@license: http://hierarchical-cluster-engine.com/license/
@since: 0.1
'''

import logging
from app.BaseServerManager import BaseServerManager
from Constants import EVENT_TYPES, LOGGER_NAME
import transport.Consts as consts
import app.Utils as Utils  # pylint: disable=F0401
from dc import EventObjects
import dtm.EventObjects
import dc.Constants as DC_CONSTANTS

# Logger initialization
logger = logging.getLogger(LOGGER_NAME)

class ClientInterfaceService(BaseServerManager):

  CONFIG_SECTION = "ClientInterfaceService"
  CONFIG_SERVER_HOST = "serverHost"
  CONFIG_SERVER_PORT = "serverPort"
  CONFIG_SITES_MANAGER = "clientSitesManager"
  SERVER_CONNECTION_NAME = "server"
  CONNECTION_PREFIX = "Connection"
  CONFIG_BATCH_TASKS_MANAGER_REALTIME = "clientBatchTasksManagerRealTime"

  # #constructor
  # initialise all connections and event handlers
  # @param configParser config parser object
  # @param connectBuilderLight instance of ConnectBuilderLight object
  #
  def __init__(self, configParser, connectBuilderLight):
    '''
    Constructor
    '''
    BaseServerManager.__init__(self)

    serverHost = configParser.get(self.CONFIG_SECTION, self.CONFIG_SERVER_HOST)
    serverPort = configParser.get(self.CONFIG_SECTION, self.CONFIG_SERVER_PORT)
    server = serverHost + ":" + str(serverPort)
    self.sitesManager = configParser.get(self.CONFIG_SECTION, self.CONFIG_SITES_MANAGER)
    self.batchTasksManagerRealTime = configParser.get(self.CONFIG_SECTION, self.CONFIG_BATCH_TASKS_MANAGER_REALTIME)

    serverConnection = connectBuilderLight.build(consts.SERVER_CONNECT, server, consts.TCP_TYPE)
    sitesManagerConnection = connectBuilderLight.build(consts.CLIENT_CONNECT, self.sitesManager)
    batchTasksManagerRealTimeConnection = connectBuilderLight.build(consts.CLIENT_CONNECT,
                                                                    self.batchTasksManagerRealTime)

    self.addConnection(self.SERVER_CONNECTION_NAME + self.CONNECTION_PREFIX, serverConnection)
    self.addConnection(str(self.sitesManager) + self.CONNECTION_PREFIX, sitesManagerConnection)
    self.addConnection(str(self.batchTasksManagerRealTime) + self.CONNECTION_PREFIX,
                       batchTasksManagerRealTimeConnection)

    self.setEventHandler(EVENT_TYPES.SITE_NEW, self.onSitesManagerRoute)
    self.setEventHandler(EVENT_TYPES.SITE_UPDATE, self.onSitesManagerRoute)
    self.setEventHandler(EVENT_TYPES.SITE_STATUS, self.onSitesManagerRoute)
    self.setEventHandler(EVENT_TYPES.SITE_DELETE, self.onSitesManagerRoute)
    self.setEventHandler(EVENT_TYPES.SITE_CLEANUP, self.onSitesManagerRoute)
    self.setEventHandler(EVENT_TYPES.SITE_FIND, self.onSitesManagerRoute)

    self.setEventHandler(EVENT_TYPES.URL_NEW, self.onSitesManagerRoute)
    self.setEventHandler(EVENT_TYPES.URL_STATUS, self.onSitesManagerRoute)
    self.setEventHandler(EVENT_TYPES.URL_UPDATE, self.onSitesManagerRoute)
    self.setEventHandler(EVENT_TYPES.URL_FETCH, self.onSitesManagerRoute)
    self.setEventHandler(EVENT_TYPES.URL_DELETE, self.onSitesManagerRoute)
    self.setEventHandler(EVENT_TYPES.URL_CLEANUP, self.onSitesManagerRoute)
    self.setEventHandler(EVENT_TYPES.URL_CONTENT, self.onSitesManagerRoute)
    self.setEventHandler(EVENT_TYPES.SQL_CUSTOM, self.onSitesManagerRoute)

    self.setEventHandler(EVENT_TYPES.SITE_NEW_RESPONSE, self.onDCClientRoute)
    self.setEventHandler(EVENT_TYPES.SITE_UPDATE_RESPONSE, self.onDCClientRoute)
    self.setEventHandler(EVENT_TYPES.SITE_STATUS_RESPONSE, self.onDCClientRoute)
    self.setEventHandler(EVENT_TYPES.SITE_DELETE_RESPONSE, self.onDCClientRoute)
    self.setEventHandler(EVENT_TYPES.SITE_CLEANUP_RESPONSE, self.onDCClientRoute)
    self.setEventHandler(EVENT_TYPES.SITE_FIND_RESPONSE, self.onDCClientRoute)

    self.setEventHandler(EVENT_TYPES.URL_NEW_RESPONSE, self.onDCClientRoute)
    self.setEventHandler(EVENT_TYPES.URL_STATUS_RESPONSE, self.onDCClientRoute)
    self.setEventHandler(EVENT_TYPES.URL_UPDATE_RESPONSE, self.onDCClientRoute)
    self.setEventHandler(EVENT_TYPES.URL_FETCH_RESPONSE, self.onDCClientRoute)
    self.setEventHandler(EVENT_TYPES.URL_DELETE_RESPONSE, self.onDCClientRoute)
    self.setEventHandler(EVENT_TYPES.URL_CLEANUP_RESPONSE, self.onDCClientRoute)
    self.setEventHandler(EVENT_TYPES.URL_CONTENT_RESPONSE, self.onDCClientRoute)
    self.setEventHandler(EVENT_TYPES.SQL_CUSTOM_RESPONSE, self.onDCClientRoute)

    self.setEventHandler(EVENT_TYPES.BATCH, self.onBatchTasksManagerRealTimeRoute)
    self.setEventHandler(EVENT_TYPES.BATCH_RESPONSE, self.onDCClientRoute)

    # map of incoming event, which are in processing
    # event.uid => event without eventObj field
    self.processEvents = dict()



  # #handler to route all event to SitesManager
  #
  # @param even instance of Event object
  def onSitesManagerRoute(self, event):
    try:
      logger.debug("Received event: " + Utils.varDump(event))
      self.send(str(self.sitesManager) + self.CONNECTION_PREFIX, event)
      self.registerEvent(event)
    except KeyError as err:
      logger.error(err.message)



  # #handler to route all event to BatchTasksManagerRealTime
  #
  # @param even instance of Event object
  def onBatchTasksManagerRealTimeRoute(self, event):
    try:
      logger.debug("Received event: " + Utils.varDump(event))
      self.send(str(self.batchTasksManagerRealTime) + self.CONNECTION_PREFIX, event)
      self.registerEvent(event)
    except KeyError as err:
      logger.error(err.message)



  # #handler to route all response event to DCClient
  #
  # @param even instance of Event object
  def onDCClientRoute(self, event):
    try:
      request_event = self.getRequestEvent(event)
      if event.cookie is not None and isinstance(event.cookie, dict) and\
         DC_CONSTANTS.MERGE_PARAM_NAME in event.cookie and bool(event.cookie[DC_CONSTANTS.MERGE_PARAM_NAME]) == False:
        logger.debug("No merge results specified in Event.cookie!")
      else:
        logger.debug("Results merge try by default.")
        event.eventObj = self.mergeResultsData(event.eventType, event.eventObj)
      self.reply(request_event, event)
      logger.debug("Results sent to DCC.")
      self.unregisterEvent(request_event)
    except KeyError as err:
      logger.error(err.message)



  # #add event in map of processing events
  #
  # @param even instance of Event object
  def registerEvent(self, event):
    event.eventObj = None
    self.processEvents[event.uid] = event



  # #get request event from processEvents map
  # @param event instance of Event  object
  # @return event instance of Event object
  def getRequestEvent(self, event):
    return self.processEvents[event.uid]



  # #delete event in map of processing events
  #
  # @param even instance of Event object
  def unregisterEvent(self, event):
    del self.processEvents[event.uid]



  # #Merge results of operation request to represent results from several hosts as single merged data
  #
  # @param eventType type of event, see constants definition
  # @param eventObj response object for requested operation
  # @return eventObj object changed or not
  def mergeResultsData(self, eventType, eventObj):
    if isinstance(eventObj, EventObjects.ClientResponse):
      #If number of client response items grater than one perform the merging, else leave untouched
      if len(eventObj.itemsList) > 1:
        if isinstance(eventObj.itemsList[0], EventObjects.ClientResponseItem):
          #Create new list of ClientResponseItem items and fill it with first item from response
          newItemObject = None
          newHost = ""
          newPort = ""
          newNode = ""
          newTime = ""
          newErrorMessage = ""
          newErrorCode = ""
          logger.debug("Response items number is " + str(len(eventObj.itemsList)))
          mergedCounter = 0
          #Cycle response items
          for clientResponseItem in eventObj.itemsList:
            #If response item exists
            if clientResponseItem.itemObject is not None:
              mergedCounter = mergedCounter + 1
              logger.debug("clientResponseItem:\n" + Utils.varDump(clientResponseItem))
              newHost += clientResponseItem.host + ";"
              newPort += str(clientResponseItem.port) + ";"
              newNode += clientResponseItem.node + ";"
              newTime += str(clientResponseItem.time) + ";"
              newErrorMessage += clientResponseItem.errorMessage
              newErrorCode += str(clientResponseItem.errorCode) + ";"
              if eventType == EVENT_TYPES.SITE_FIND_RESPONSE:
                #Merge SITE_FIND operation results response
                newItemObject = self.mergeResultsSiteFind(newItemObject, clientResponseItem.itemObject)
              elif eventType == EVENT_TYPES.URL_FETCH_RESPONSE or eventType == EVENT_TYPES.URL_STATUS_RESPONSE:
                #Merge URL_FETCH operation results response
                newItemObject = self.mergeResultsURLFetch(newItemObject, clientResponseItem.itemObject)
              elif eventType == EVENT_TYPES.URL_CONTENT_RESPONSE:
                #Merge URL_CONTENT operation results response
                newItemObject = self.mergeResultsURLContent(newItemObject, clientResponseItem.itemObject)
              elif eventType == EVENT_TYPES.SITE_STATUS_RESPONSE:
                #Merge SITE_STATUS operation results response
                newItemObject = self.mergeResultsSiteStatus(newItemObject, clientResponseItem.itemObject)
              #elif eventType == EVENT_TYPES.SITE_NEW_RESPONSE:
              #  #Merge SITE_NEW operation results response
              #  newItemObject = self.mergeResultsGeneralResponse(newItemObject, clientResponseItem.itemObject)
              else:
                l = {EVENT_TYPES.SITE_NEW_RESPONSE, EVENT_TYPES.SITE_UPDATE_RESPONSE, EVENT_TYPES.SITE_DELETE_RESPONSE,
                     EVENT_TYPES.SITE_CLEANUP_RESPONSE, EVENT_TYPES.URL_NEW_RESPONSE, EVENT_TYPES.URL_UPDATE_RESPONSE,
                     EVENT_TYPES.URL_DELETE_RESPONSE, EVENT_TYPES.URL_CLEANUP_RESPONSE, EVENT_TYPES.URL_UPDATE_RESPONSE,
                     EVENT_TYPES.URL_DELETE_RESPONSE, EVENT_TYPES.URL_CLEANUP_RESPONSE}  # pylint: disable=unused-import
                if eventType in l:
                  #Merge another result types that no need special processing of fields but check and leave just one
                  newItemObject = self.mergeResultsGeneralResponse(newItemObject, clientResponseItem.itemObject)
            else:
              #Object is null or wrong type
              logger.error("The clientResponseItem.itemObject is None!")
          #Replace items list with newly created and merged if itemObject is set
          if newItemObject is not None:
            eventObj.itemsList = [eventObj.itemsList[0]]
            eventObj.itemsList[0].errorCode = newErrorCode
            eventObj.itemsList[0].errorMessage = newErrorMessage
            eventObj.itemsList[0].itemObject = newItemObject
            eventObj.itemsList[0].host = newHost
            eventObj.itemsList[0].port = newPort
            eventObj.itemsList[0].node = newNode
            eventObj.itemsList[0].time = newTime
            logger.debug("Merge done with " + str(mergedCounter) + " items.")
          else:
            logger.debug("No items collected while merge procedure, merge skipped.")
        else:
          #Wrong type
          logger.error("Wrong eventObj.itemsList[0] type " + str(type(eventObj.itemsList[0])) + \
                         " expected EventObjects.ClientResponseItem\n" + Utils.varDump(eventObj.itemsList[0]))
    else:
      logger.error("Wrong eventObj type " + str(type(eventObj)) + " expected EventObjects.ClientResponse\n" + \
                   Utils.varDump(eventObj))

    return eventObj



  # #Merge results of operation URL_CONTENT
  #
  # @param newItemsList new itemsList for new merged itemObject
  # @param itemsList current items list from current itemObject of node response results
  # @return new itemsList for merged itemObject
  def mergeResultsURLContent(self, newItemsList, itemsList):
    #If this is first item in results list, init itemObject with empty list
    if newItemsList is None:
      newItemsList = []

    if isinstance(itemsList, list):
      #For each URLContentResponse object in the clientResponseItem.itemObject
      for urlContentResponse in itemsList:
        #If raw or processed content exists
        if len(urlContentResponse.rawContents) > 0 or len(urlContentResponse.processedContents) > 0:
          #Check is candidate valid
          for itemObject in newItemsList:
            #Is urlMd5 already exists in accumulated list
            if itemObject.siteId == urlContentResponse.siteId and\
              (itemObject.urlMd5 == urlContentResponse.urlMd5 or\
               (urlContentResponse.rawContentMd5 != '' and urlContentResponse.rawContentMd5 != '""' and\
                itemObject.rawContentMd5 == urlContentResponse.rawContentMd5) or\
               (urlContentResponse.contentURLMd5 != '' and urlContentResponse.contentURLMd5 != '""' and\
                itemObject.contentURLMd5 == urlContentResponse.contentURLMd5)):
              #Item already exists
              logger.debug("Already exists urlContentResponse:\n" + Utils.varDump(urlContentResponse))
              #TODO: check timestamps, but now just reject the same item by urlMd5
              urlContentResponse = None
              break
          #Add content to list
          if urlContentResponse is not None:
            newItemsList.append(urlContentResponse)
            logger.debug("Added urlContentResponse:\n" + Utils.varDump(urlContentResponse))
          else:
            logger.debug("Rejected urlContentResponse")
        else:
          logger.debug("Empty contents lists in urlContentResponse:\n" + Utils.varDump(urlContentResponse))
    else:
      #Object is null or wrong type
      logger.error("Wrong type of clientResponseItem.itemObject\n" + Utils.varDump(itemsList))

    return newItemsList



  # #Merge results of operation SITE_STATUS
  #
  # @param mergedSite new merged Site object
  # @param currentSite current Site object from itemObject item of response results
  # @return new merged Site object
  def mergeResultsSiteStatus(self, mergedSite, currentSite):
    if isinstance(currentSite, EventObjects.Site):
      #If this is first time call for results list init. with this Site object
      if mergedSite is None:
        mergedSite = currentSite
      else:
        logger.debug("Merge with Site object:\n" + Utils.varDump(currentSite))
        #Merge site's data
        mergedSite = self.mergeResultsSiteFields(mergedSite, currentSite)
    else:
      #Object is null or wrong type
      logger.error("Wrong type of currentSite object: " + str(type(currentSite)) + ", Site expected\n" + \
                   Utils.varDump(currentSite))

    return mergedSite



  # #Merge Site object fields
  #
  # @param siteToMerge Site object that is used as base
  # @param siteMergeWith Site object that is merged with base
  # @return merged siteToMerge object
  def mergeResultsSiteFields(self, siteToMerge, siteMergeWith):
    #Set merged values for correspondent fields of the Site objects
    siteToMerge.resources += siteMergeWith.resources
    siteToMerge.contents += siteMergeWith.contents
    siteToMerge.iterations = max([siteToMerge.iterations, siteMergeWith.iterations])
    siteToMerge.errors += siteMergeWith.errors
    siteToMerge.size += siteMergeWith.size
    siteToMerge.avgSpeed = min([siteToMerge.avgSpeed, siteMergeWith.avgSpeed])
    siteToMerge.avgSpeedCounter = min([siteToMerge.avgSpeedCounter, siteMergeWith.avgSpeedCounter])

    return siteToMerge



  # #Merge results of operation SITE_FIND
  #
  # @param newItemsList new itemsList for new merged itemObject
  # @param itemsList current items list from current itemObject of node response results
  # @return new itemsList for merged itemObject
  def mergeResultsSiteFind(self, newItemsList, itemsList):
    #If this is first item in results list, init itemObject with empty list
    if newItemsList is None:
      newItemsList = []

    if isinstance(itemsList, list):
      #For each Site object in the clientResponseItem.itemObject
      for site in itemsList:
        #Check presence of this site in the newItemsList
        present = False
        for addedSite in newItemsList:
          if addedSite.id == site.id:
            #Merge site's data
            addedSite = self.mergeResultsSiteFields(addedSite, site)
            present = True
            break
        if not present:
          #Add content to list
          newItemsList.append(site)
          logger.debug("Added Site:\n" + Utils.varDump(site))
    else:
      #Object is null or wrong type
      logger.error("Wrong type of clientResponseItem.itemObject\n" + Utils.varDump(itemsList))

    return newItemsList



  # #Merge results of operation URL_FETCH
  #
  # @param newItemsList new itemsList for new merged itemObject
  # @param itemsList current items list from current itemObject of node response results
  # @return new itemsList for merged itemObject
  def mergeResultsURLFetch(self, newItemsList, itemsList):
    #If this is first item in results list, init itemObject with empty list
    if newItemsList is None:
      newItemsList = []

    if isinstance(itemsList, list):
      #For each URL object in the clientResponseItem.itemObject
      for url in itemsList:
        #Check presence of this url in the newItemsList
        present = False
        for addedURL in newItemsList:
          if addedURL.urlMd5 == url.urlMd5:
            #Chose URL object with higher status value
            if addedURL.status <= url.status and addedURL.crawled < url.crawled:
              addedURL = url
            present = True
            break
        if not present:
          #Add content to list
          newItemsList.append(url)
          logger.debug("Added Site:\n" + Utils.varDump(url))
    else:
      #Object is null or wrong type
      logger.error("Wrong type of clientResponseItem.itemObject\n" + Utils.varDump(itemsList))

    return newItemsList



  # #Merge results of operation for GeneralResponse item object
  #
  # @param mergedGeneralResponse new merged GeneralResponse object
  # @param currentGeneralResponse current GeneralResponse object from itemObject item of response results
  # @return new merged GeneralResponse object
  def mergeResultsGeneralResponse(self, mergedGeneralResponse, currentGeneralResponse):
    if isinstance(currentGeneralResponse, dtm.EventObjects.GeneralResponse):
      if mergedGeneralResponse is None:
        #If this is first time call for results list init. with this Site object
        mergedGeneralResponse = currentGeneralResponse
        logger.debug("Merge init GeneralResponse object:\n" + Utils.varDump(currentGeneralResponse))
      else:
        logger.debug("Merge with GeneralResponse object:\n" + Utils.varDump(currentGeneralResponse))
        #Merge fields values
        mergedGeneralResponse.errorCode = str(mergedGeneralResponse.errorCode) + ";" + \
                                             str(currentGeneralResponse.errorCode)
        mergedGeneralResponse.errorMessage = str(mergedGeneralResponse.errorMessage) + ";" + \
                                             str(currentGeneralResponse.errorMessage)
        mergedGeneralResponse.statuses.extend(currentGeneralResponse.statuses)
    else:
      #Object is null or wrong type
      logger.error("Wrong type of currentGeneralResponse object: " + str(type(currentGeneralResponse)) + \
                   ", dtm.GeneralResponse expected\n" + Utils.varDump(currentGeneralResponse))

    return mergedGeneralResponse

