"""
@package: dc
@file CollectURLs.py
@author Scorp <developers.hce@gmail.com>
@link: http://hierarchical-cluster-engine.com/
@copyright: Copyright &copy; 2013-2014 IOIX Ukraine
@license: http://hierarchical-cluster-engine.com/license/
@since: 0.1
"""
import logging
import lxml.html
import hashlib
import re
import app.Consts as APP_CONSTS
import dc_processor.Constants as PCONSTS
import app.Utils as Utils
import dc.EventObjects
import CrawlerTask
from app.Filters import Filters
try:
  import feedparser
except ImportError:
  feedparser = None


logger = logging.getLogger(APP_CONSTS.LOGGER_NAME)


class CollectURLs(object):

  BINARY_CONTENT_TYPE_PATTERN = re.compile('(text)|(xml)', re.I)
  WITH_URL_XPATH_LIST = ["//a/@href", "//applet/@codebase", "//area/@href", "//base/@href", \
  "//blockquote/@cite", "//body/@background", "//del/@cite", \
  "//frame/@src", "//head/@profile", "//iframe/@src", "//input/@src", "//ins/@cite", \
  "//object/@codebase", "//q/@cite", "//audio/@src", \
  "//button/@formaction", "//command/@icon", "//embed/@src", "//html/@manifest", \
  "//input/@formaction", "//source/@src", "//video/@src", "//video/@poster", "//category/@domain",
  "//link", "//guid"]
  COLLECT_POST_DATA_NAME = "COLLECT_POST_DATA"
  COLLECT_POST_DATA = "1"
  DC_URLS_TABLE_PREFIX = "urls_"
  PATTERN_WITH_PROTOCOL = re.compile('[a-zA-Z]+:(//)?')
  DETECT_MIME_COLLECTED_URL = "2"
  DETECT_MIME_TIMEOUT = 1


  def __init__(self):
    self.crawledResource = None
    self.url = None
    self.dom = None
    self.realUrl = None
    self.processorName = None
    self.batchItem = None
    self.urlXpathList = None
    self.feedItems = None
    self.feed = None
    self.siteProperties = None
    self.site = None
    self.dbWrapper = None
    self.autoRemoveProps = None
    self.autoDetectMime = None
    self.processContentTypes = None
    self.postForms = None
    self.urlProcess = None


  # #checkClassFields method checks all class's mandatory fields
  #
  def checkClassFields(self):
    for field in self.__dict__:
      if field not in ["feedItems", "feed", "processorName", "autoDetectMime", "processContentTypes", "postForms"]:
        if not hasattr(self, field) or getattr(self, field) is None:
          logger.debug(">>> [CollectURLs] Mandatory field must be initialized, field Name = " + field)
          raise Exception(">>> [CollectURLs] Mandatory field must be initialized, field Name = " + field)


  # #process method
  #
  def process(self):
    self.checkClassFields()
    localSiteId = self.batchItem.siteId if self.batchItem.siteId else "0"
    nextStep = True
    internalLinks, externalLinks = [], []

    if self.url.type == dc.EventObjects.URL.TYPE_SINGLE:
      logger.debug("URL type: single")
      nextStep = False

    if nextStep and not self.BINARY_CONTENT_TYPE_PATTERN.search(self.crawledResource.content_type):
      logger.debug("Binary content: <<%s>>", self.crawledResource.content_type)
      nextStep = False

    # don't parse url for 4XX or 5XX response
    if nextStep:
      code_type = int(self.crawledResource.http_code) / 100
      if code_type == 4 or code_type == 5:
        nextStep = False

    if nextStep and not self.crawledResource.html_content:
      nextStep = False

    if nextStep and self.dom is None:
      logger.debug("DOM is None")
      nextStep = False

    if nextStep:

      urlSet = set()
      self.processProcessor(urlSet, self.dom, self.urlXpathList, self.batchItem.urlObj)

      formUrls, formMethods, formFields = self.extractFormURL(self.dom, self.siteProperties)
      urlSet.update(formUrls)
      #(3) END
      urlTable = self.DC_URLS_TABLE_PREFIX + localSiteId

      numUrlsFromPage = 0
      self.urlProcess.urlTable = urlTable
      currentCnt = self.urlProcess.readCurrentCnt(self.site.maxURLs)

      params = []
      localFilters = None
      for elemUrl in urlSet:
        localUrl = elemUrl
        self.urlProcess.urlObj = self.url
        self.urlProcess.url = elemUrl
        retUrl, retContinue = self.urlProcess.processURL(self.realUrl, internalLinks, externalLinks)
        if retUrl is not None:
          elemUrl = retUrl
        if retContinue:
          continue

        if not self.filtersApply(localFilters, elemUrl, self.url.depth, self.dbWrapper, localSiteId):
          continue

        urlMd5 = hashlib.md5(elemUrl).hexdigest()

        self.urlProcess.url = elemUrl
        self.urlProcess.siteId = localSiteId
        self.urlProcess.urlTable = urlTable
        if self.urlProcess.isUrlExist(self.site.recrawlPeriod, self.siteProperties, urlMd5):
          continue

        if self.site.maxURLs:
          if currentCnt < self.site.maxURLs:
            logger.debug("Site MaxURLs limit need to be checked.")
            currentCnt += 1
          elif self.urlProcess.autoRemoveURL(self.autoRemoveProps, self.site.recrawlPeriod, urlTable,
                                            self.dbWrapper) == 0:
            # should not use break, because need add to internal/external links although reach the MaxURL limit
            continue

        # support maxURLsFromPage
        if  self.site.maxURLsFromPage and numUrlsFromPage >= self.site.maxURLsFromPage:
          logger.debug("Site maxURLsFromPage limit reached! Site.MaxURLsFromPage: %s, current_page_num: %s",
                          self.site.maxURLsFromPage, numUrlsFromPage)
          continue

        #detect collected url mime type and ignore non-match URL
        #(7) Detect collected url mime type and ignore non-match URL
        detectedMime = ''
        if self.autoDetectMime == self.DETECT_MIME_COLLECTED_URL and self.processContentTypes is not None:
          self.urlProcess.url = elemUrl
          detectedMime = self.urlProcess.detectUrlMime()
          if detectedMime not in self.processContentTypes:
            continue
        numUrlsFromPage += 1
        #(7) END

        self.urlProcess.siteId = localSiteId
        depth = self.urlProcess.getDepthFromUrl(self.batchItem.urlId)

        self.urlProcess.url = elemUrl
        self.urlProcess.siteId = localSiteId
        self.urlProcess.urlObj = self.url
        localUrlObj = self.urlProcess.createUrlObjForCollectURLs(urlMd5, formMethods, self.batchItem.urlId, depth,
                                                                 detectedMime, self.site.maxURLsFromPage)
        if self.processorName == PCONSTS.PROCESSOR_FEED_PARSER or self.processorName == PCONSTS.PROCESSOR_RSS:
          self.urlProcess.url = elemUrl
          self.urlProcess.siteId = localSiteId
          self.urlProcess.urlObj = localUrlObj
          localRet = self.urlProcess.fillRssFieldInUrlObj(localUrl, self.url.url, self.batchItem, self.processorName,
                                                          self.feed)
          self.urlProcess.urlObj = None
          if localRet is not None:
            self.feedItems[urlMd5] = localRet
        params.append(localUrlObj)

      result = self.dbWrapper.urlNew(params)
      logger.debug(">>> Url New result: "+str(result))
      self.dbWrapper.fieldRecalculating(localSiteId)

      if formFields is not None and self.postForms is not None:
        fieldParams = self.getFieldParams(formFields, self.postForms, localSiteId)
        self.insertNewSiteProperties(fieldParams, self.dbWrapper, localSiteId)

    return nextStep, internalLinks, externalLinks


  # #processProcessor method
  #
  def processProcessor(self, urlSet, dom, urlXpathList, urlObj):
    if (self.processorName == PCONSTS.PROCESSOR_FEED_PARSER or self.processorName == PCONSTS.PROCESSOR_RSS) \
      and urlObj.type != dc.EventObjects.URL.TYPE_FETCHED:
      if feedparser is not None:
        try:
          self.feedItems = {}
          self.feed = feedparser.parse(self.crawledResource.html_content)
          urlSet.update(entry.link for entry in self.feed.entries)
        except TypeError, err:
          logger.debug("WRONG CONTENT FOR URL <"+str(urlObj.url)+"> not rss feed. "+str(err.message))
      else:
        logger.debug("feedparser module not found")
    # won't collect urls from rss feed resources
    elif self.processorName != PCONSTS.PROCESSOR_RSS:
      # Added support of urlXpathList as site's properties
      if len(urlXpathList):
        logger.debug("Site has URL_XPATH_LIST property: %s", str(urlXpathList))
      else:
        urlXpathList = self.WITH_URL_XPATH_LIST
        logger.debug("Site hasn't URL_XPATH_LIST property. Use default xpath's' list.")
      for xpath in urlXpathList:
        elem = dom.xpath(xpath)
        elem_type = type(elem)
        if elem_type == list and len(elem) > 0 and hasattr(elem[0], "tail"):
          urlSet.update([el.tail for el in elem])
        elif elem_type == list and len(elem) > 0 and isinstance(elem[0], lxml.html.HtmlElement):
          urlSet.update([el.text for el in elem])
        else:
          urlSet.update(elem)
    logger.debug("urlSet: %s", str(urlSet))


  # # extrace URL, form action, and fields from html dom
  #
  # @param dom the dom tree
  # @return form_urls    sequence of urls extracted
  # @return form_methods  dict of form methods, {form_action: form_method}
  # @return form_fields  dict of fields {field_name: field_value}
  def extractFormURL(self, dom, siteProperties):
    formUrls, formMethods, formFields = [], {}, {}
    if self.COLLECT_POST_DATA_NAME in siteProperties and \
    siteProperties['COLLECT_POST_DATA'] == self.COLLECT_POST_DATA:
      for form in dom.xpath("//form"):
        formAction = None
        formMethod = 'get'
        for attr in form.keys():
          if attr.lower() == "action":
            formAction = form.get(attr)
            formUrls.append(formAction)
          elif attr.lower() == "method":
            formMethod = form.get(attr)
        if not formAction:
          continue
        formMethods[formAction] = formMethod
        for field in form.getchildren():
          tagName, tagValue = None, ""
          for fieldTag in field.keys():
            if fieldTag.lower() == "name":
              tagName = field.get(fieldTag)
            elif fieldTag.lower() == "value":
              tagValue = field.get(fieldTag)
          if tagName:
            formFields[tagName] = tagValue
      logger.info("extracted form data, formUrls:%s, formMethods:%s, formFields:%s", \
      formUrls, formMethods, formFields)
    return formUrls, formMethods, formFields


  # #filtersApply applies filters and returns bool result
  #
  def filtersApply(self, localFilters, localUrl, depth, wrapper, siteId):
    ret = False
    fValue = Utils.generateReplacementDict()
    fValue.update({"MAX_DEPTH": str(depth)})
    if localFilters is None:
      localFilters = Filters(None, wrapper, siteId)
    
    logger.debug(">>> url before filter = " + localUrl)
    fResult = localFilters.filterAll(Filters.STAGE_AFTER_DOM, fValue, Filters.LOGIC_OR, localUrl)
    logger.debug(">>> filter result - " + str(fResult))
    '''
    if len(fResult) == 0:
      ret = True
    else:
      for elem in fResult:
        if elem < 0:
          ret = True
          break
    '''
    '''
    for elem in fResult:
      if elem > 0:
        ret = ret or (elem > 0)
    '''
    for elem in fResult:
      if elem > 0:
        ret = True
        break
    return ret


  # #getFieldParams method fill and returns post forms
  #
  #formFields - post fields dict
  #postForms - post form list
  #siteId - site's id
  def getFieldParams(self, formFields, postForms, siteId):
    ret = []
    for fieldName, fieldValue in formFields.iteritems():
      if fieldName in postForms:
        continue
      logger.debug("field_name: %s", fieldName)
      logger.debug("field_value: %s", fieldValue)
      ret.append((siteId, "HTTP_POST_FORM_" + fieldName, fieldValue))
    return ret


  # #insertNewSiteProperties update Site (add new properties in correspond table)
  #
  #params - params list
  #wrapper - db-task wrapper
  #siteId - site's id
  def insertNewSiteProperties(self, params, wrapper, siteId):
    if siteId is not None and hasattr(params, '__iter__') and len(params) > 0:
      localSiteUpdate = dc.EventObjects.SiteUpdate(siteId)
      for attr in localSiteUpdate.__dict__:
        if hasattr(localSiteUpdate, attr):
          setattr(localSiteUpdate, attr, None)
      localSiteUpdate.updateType = dc.EventObjects.SiteUpdate.UPDATE_TYPE_APPEND
      localSiteUpdate.id = siteId
      localSiteUpdate.properties = []
      for param in params:
        newPropElem = {}
        newPropElem["siteId"] = param[0]
        newPropElem["name"] = param[1]
        newPropElem["value"] = param[2]
        localSiteUpdate.properties.append(newPropElem)
      wrapper.siteNewOrUpdate(localSiteUpdate, stype=dc.EventObjects.SiteUpdate)
