"""
HCE project, Python bindings, Distributed Tasks Manager application.
web page fetchers.

@package: dc
@file Fetcher.py
@author madk <developers.hce@gmail.com>
@link: http://hierarchical-cluster-engine.com/
@copyright: Copyright &copy; 2013-2014 IOIX Ukraine
@license: http://hierarchical-cluster-engine.com/license/
@since: 0.1
"""

# #The BaseFetcher class, defines the interface of fetchers
# #this class don't implement any fetcher
# concrete fetcher class should extends from this class
#

from requests.auth import HTTPBasicAuth
import requests
import requests.exceptions
import re
from app.Utils import varDump
import dc_crawler.Constants as CONSTS

#max content size for use chardet to detect charset
MAX_CONTENT_SIZE_FOR_CHARDET = 5000000
class BaseFetcher(object):

  fetchers = None

  TYP_NORMAL = 1
  TYP_DYNAMIC = 2
  TYP_EXTERNAL = 3
  TYP_URLLIB = 4

  @staticmethod
  def init():
    #enumerate content_types we don't want to fetch
    BaseFetcher.restircted_conten_types = ["audio/mpeg", "application/pdf"]

    BaseFetcher.fetchers = {
      BaseFetcher.TYP_NORMAL : RequestsFetcher(),
      BaseFetcher.TYP_DYNAMIC: GhostpyFetcher(),
      BaseFetcher.TYP_URLLIB: URLLibFetcher(),
      BaseFetcher.TYP_EXTERNAL: ExternalFetcher()
    }

  # #fetch a url, and return the response
  #
  # @param url, the url to fetch
  # @param method, fetch HTTP method
  # @param headers, request headers dict
  # @param timeout, request timeout(seconds)
  # @param allow_redirects, should follow redirect
  # @param proxies, proxy setting, tuple of proxy_type, host, port, username, password
  # @param auth, basic auth setting, tuple of username and password
  # @param data, post data, used only when method is post
  # @return Response object
  def open(self,
    url,
    external_url=None,
    method='get',
    headers={},
    timeout=100,
    allow_redirects=True,
    proxies=None,
    auth=None,
    data=None,
    logger=None,
    allowed_content_types=None,
    max_resource_size=None,
    max_redirects=CONSTS.MAX_HTTP_REDIRECTS_LIMIT):
    pass


  # #get fetched by fetch type
  #
  # @param typ, the fetch type
  # @return fetcher
  @staticmethod
  def get_fetcher(typ):
    if not BaseFetcher.fetchers:
      BaseFetcher.init()
    if typ in BaseFetcher.fetchers:
      return BaseFetcher.fetchers[typ]
    else:
      raise BaseException("unsuported fetch type:%s" % (typ,))


  # #check whether the fetcher have meta resource
  #
  # @return whether the fetcher have meta resource
  def should_have_meta_res(self):
    return False



# # Fetcher base on the requests module, cann't execute javascript
#
#
class RequestsFetcher(BaseFetcher):

  # #fetch a url, and return the response
  #
  # @param url, the url to fetch
  # @param method, fetch HTTP method
  # @param headers, request headers dict
  # @param timeout, request timeout(seconds)
  # @param allow_redirects, should follow redirect
  # @param proxies, proxy setting
  # @param auth, basic auth setting, tuple of name and password
  # @param data, post data, used only when method is post
  # @return Response object
  def open(self,
    url,
    external_url=None,
    method='get',
    headers=None,
    timeout=100,
    allow_redirects=True,
    proxies=None,
    auth=None,
    data=None,
    logger=None,
    allowed_content_types=None,
    max_resource_size=None,
    max_redirects=CONSTS.MAX_HTTP_REDIRECTS_LIMIT):

    if auth:
      auth = HTTPBasicAuth(auth[0], auth[1])

    proxy_setting = None
    if proxies:
      proxy_type, proxy_host, proxy_port, proxy_user, proxy_passwd = proxies
      if proxy_type.lower() == 'http':

        if proxy_user:
          proxies = "http://%s:%s@%s:%s" % (proxy_user, proxy_passwd, proxy_host, proxy_port)
        else:
          proxies = "http://%s:%s" % (proxy_host, proxy_port)

        proxy_setting = {"http" : proxies}
    rSession = requests.Session()
    rSession.max_redirects = int(max_redirects)
    method_func = rSession.__getattribute__(method)
    logger.debug("max_resource_size: %s" % str(max_resource_size))
    impl_res = None
    res = Response()
    try:
      impl_res = method_func(url,
        timeout=timeout,
        headers=headers,
        allow_redirects=allow_redirects,
        proxies=proxy_setting,
        auth=auth,
        data=data,
        stream=True)
      #logger.debug("impl_res: " + varDump(impl_res))
      #try to prevent huge content fetching
      if "content-length" in impl_res.headers and \
        max_resource_size!=CONSTS.MAX_HTTP_SIZE_UNLIMIT and \
        int(impl_res.headers['content-length']) > max_resource_size:
        logger.debug("Content size overshooted. content-length: %s, max_resource_size: %s" % str(impl_res.headers['content-length'], str(max_resource_size)))
        res.content_size = int(impl_res.headers['content-length'])
        res.headers = impl_res.headers
        res.redirects = impl_res.history
        res.status_code = impl_res.status_code
        res.url = impl_res.url
      else:
        #don't detect charset for binary content type or BIG response
        if impl_res.headers.get('content-type', '').startswith('application') or  \
          impl_res.headers.get('content-type', '').startswith('audio') or  \
          len(impl_res.content) >= MAX_CONTENT_SIZE_FOR_CHARDET:
          detected_encoding = impl_res.encoding
        else:
          # use chardet to improve encoding detect
          detected_encoding = impl_res.apparent_encoding
          if detected_encoding != 'ascii' and detected_encoding != 'ISO-8859-2':
            if impl_res.encoding != detected_encoding:
              impl_res.encoding = detected_encoding
        res.encoding = impl_res.encoding
        res.url = impl_res.url
        res.status_code = impl_res.status_code
        res.redirects = impl_res.history
        #Fix for pages that has xml document tag but no html structure inside
        text_buffer = self.fixWrongXMLHeader(impl_res.text)
        if impl_res.headers.get('content-type', '').startswith('application'):
          res.unicode_content = impl_res.content
        else:
          res.unicode_content = text_buffer
        res.str_content = impl_res.content
        if impl_res.headers.get('content-type', '').startswith('application'):
          res.rendered_unicode_content = impl_res.content
        else:
          res.rendered_unicode_content = text_buffer

        res.content_size = impl_res.raw.tell()
        res.headers = impl_res.headers
        res.cookies = requests.utils.dict_from_cookiejar(impl_res.cookies)
    except requests.exceptions.RequestException, err:
      res.error_mask = CONSTS.ERROR_FETCH_AMBIGUOUS_REQUEST
    except requests.exceptions.ConnectionError, err:
      res.error_mask = CONSTS.ERROR_FETCH_CONNECTION_ERROR
    except requests.exceptions.HTTPError, err:
      res.error_mask = CONSTS.ERROR_FETCH_HTTP_ERROR
    except requests.exceptions.URLRequired, err:
      res.error_mask = CONSTS.ERROR_FETCH_INVALID_URL
    except requests.exceptions.TooManyRedirects, err:
      res.error_mask = CONSTS.ERROR_FETCH_TOO_MANY_REDIRECTS
    #except requests.exceptions.ConnectTimeout, err:
    #  res.error_mask = CONSTS.ERROR_FETCH_CONNECTION_TIMEOUT
    #except requests.exceptions.ReadTimeout, err:
    #  res.error_mask = CONSTS.ERROR_FETCH_READ_TIMEOUT
    except requests.exceptions.Timeout, err:
      res.error_mask = CONSTS.ERROR_FETCH_TIMEOUT
    except Exception, err:
      res.error_mask = CONSTS.ERROR_FETCH_UNDEFINED
      #logger.error("Requests fetcher has thrown exception: %s" % str(err.message))
    return res


  # #Fix the wrong html document structure in case of XML header is defined first
  #
  # @param content input content
  # @return fixed contentStr
  def fixWrongXMLHeader(self, contentStr):
    #text_file = open("/tmp/fetcher_log.txt", "w")
    #text_file.write("Fetcher: start\n")

    if contentStr.startswith('<?xml ') and '<html' in contentStr and '<head' in contentStr:
      #text_file.write("Fetcher: xml detected!\n")
      p = re.compile(r'<\?xml .*\?>')
      contentStr = p.sub('', contentStr, count=1)
      #text_file.write(contentStr)

    #text_file.close()

    return contentStr



# # Fetcher base on the ghost.py module, which can execute javascript
#
#
class GhostpyFetcher(BaseFetcher):
  # #fetch a url, and return the response
  #
  # @param url, the url to fetch
  # @param method, fetch HTTP method
  # @param headers, request headers dict
  # @param timeout, request timeout(seconds)
  # @param allow_redirects, should follow redirect
  # @param proxies, proxy setting
  # @param auth, basic auth setting
  # @param data, post data, used only when method is post
  # @return Response object
  def open(self,
    url,
    external_url=None,
    method='get',
    headers=None,
    timeout=100,
    allow_redirects=True,
    proxies=None,
    auth=None,
    data=None,
    logger=None,
    allowed_content_types=None,
    max_resource_size=None,
    max_redirects=1):


    if 'Accept-Encoding' in headers:
      if "gzip" in headers['Accept-Encoding']:
        if logger:
          logger.warn("GhostpyFetcher don't support GZIP, Accept-Encoding request header was ignored")
        del headers['Accept-Encoding']
    from ghost import Ghost  # pylint: disable-all
    agent = headers.get('User-Agent', '')
    browser = Ghost(user_agent=agent)
    if proxies:
      browser.set_proxy(proxies[0], proxies[1], int(proxies[2]), proxies[3], proxies[4])

    try:
      page_res, _ = browser.open(url, method,
        headers=headers,
        auth=auth)
    except ghost.TimeoutError:  # pylint: disable-all
      raise requests.exceptions.Timeout()
    res = Response()
    res.url = page_res.url
    res.status_code = page_res.http_status
    #TODO get all redirect histories
    res.redirects = []
    res.unicode_content = browser.content
    charset, _ = browser.evaluate("document.characterSet")
    if charset:
      charset = str(charset)
    if logger:
      logger.info("got page charset %s by execute javascript", charset)
    if not charset:
      charset = "utf-8"
    res.encoding = charset
    res.str_content = browser.content.encode(res.encoding, res.encoding)

    res.rendered_unicode_content = browser.content
    res.content_size = len(res.str_content)
    res.headers = page_res.headers
    res.cookies = {}
    for cookie in browser.cookies:
      res.cookies[str(cookie.name())] = str(cookie.value())
    try:
      if isinstance(page_res.content, unicode):
        res.meta_res = page_res.content
      else:
        res.meta_res = unicode(page_res.content, charset)
    except BaseException:
      if logger:
        logger.warn("error when get meta resource", exc_info=True)
    if res.meta_res == u'None':
      if logger:
        logger.warn("got None from page response, use rendered content")
      res.meta_res = res.rendered_unicode_content
    res.dynamic_fetcher_type = "ghost"
    return res


  # #check whether the fetcher have meta resource
  #
  # @return whether the fetcher have meta resource
  def should_have_meta_res(self):
    return True


# # external Fetcher
#
#
class ExternalFetcher(BaseFetcher):
  # #fetch a url, and return the response
  #
  # @param url, the url to fetch
  # @param external_url, the external url template
  # @param method, fetch HTTP method
  # @param headers, request headers dict
  # @param timeout, request timeout(seconds)
  # @param allow_redirects, should follow redirect
  # @param proxies, proxy setting
  # @param auth, basic auth setting
  # @param data, post data, used only when method is post
  # @return Response object
  def open(self,
           url,
           external_url=None, **kwargs):
    if external_url:
      url = external_url.replace('%url%', url)
    
    kwargs['method'] = 'get'
    return BaseFetcher.get_fetcher(BaseFetcher.TYP_NORMAL).open(url, **kwargs)



# # urllib Fetcher
#
#
class URLLibFetcher(BaseFetcher):
  # #fetch a url, and return the response
  #
  # @param url, the url to fetch
  # @param external_url, the external url template
  # @param method, fetch HTTP method
  # @param headers, request headers dict
  # @param timeout, request timeout(seconds)
  # @param allow_redirects, should follow redirect
  # @param proxies, proxy setting
  # @param auth, basic auth setting
  # @param data, post data, used only when method is post
  # @return Response object
  def open(self,
    url,
    external_url=None, **kwargs):
    if external_url:
      url = external_url.replace('%url%', url)

    import urllib2
    import os

    logger = kwargs['logger']
    allowed_content_types = kwargs['allowed_content_types']
    max_resource_size = kwargs["max_resource_size"]

    res = Response()
    logger.debug("url: <%s>" % url)
    response = None
    try:
      response = urllib2.urlopen(url)
      headers_info = response.info()
      if headers_info is not None:
        if headers_info.type in allowed_content_types:
          if response is not None:
            #res.encoding = impl_res.encoding
            #res.cookies = requests.utils.dict_from_cookiejar(impl_res.cookies)
            res.url = response.geturl()
            res.status_code = response.getcode()
            content_response = response.read()
            res.unicode_content = content_response
            res.str_content = content_response
            res.rendered_unicode_content = content_response
            res.content_size = len(content_response)
            headers = {}
            headers["content-length"] = res.content_size
            headers["content-type"] = headers_info.type
            res.headers = headers
            history = []
            res.redirects = history
          else:
            logger.debug("URLLib return empty response.")
        else:
          logger.debug("Content-Type not allowed. headers_info.type: %s" % str(headers_info.type))
      else:
        logger.debug("URLLib info is empty.")
    except urllib2.HTTPError, err:
    #except Exception, err:
      logger.debug("Exception <%s>", str(err.code))
    return res



# #The Response class
# represents an web page response
class Response(object):
  def __init__(self):
    #final url
    self.url = None
    # http status code
    self.status_code = None
    # redirect lists
    self.redirects = None
    # unicode content
    self.unicode_content = None
    # str content
    self.str_content = None
    # rendered(by command line browser) unicode content
    self.rendered_unicode_content = None
    # http response content size
    self.content_size = None
    # content encoding
    self.encoding = None
    # headers
    self.headers = None
    #meta resource
    self.meta_res = None
    #cookies
    self.cookies = None
    #dynamic fetcher type
    self.dynamic_fetcher_type = None
    #error mask from fetcher
    self.error_mask = CONSTS.ERROR_OK
