"""
@package: dc
@file ResourceProcess.py
@author Scorp <developers.hce@gmail.com>
@link: http://hierarchical-cluster-engine.com/
@copyright: Copyright &copy; 2013-2014 IOIX Ukraine
@license: http://hierarchical-cluster-engine.com/license/
@since: 0.1
"""
import logging
import app.Consts as APP_CONSTS
import dc.EventObjects as dc_event
import dc_crawler.Constants as CONSTS
import lxml.html
import lxml.etree
import tidylib
import time
from dateutil.parser import *
from dc_crawler.CrawledResource import CrawledResource
from app.Utils import SQLExpression

logger = logging.getLogger(APP_CONSTS.LOGGER_NAME)


class ResourceProcess(object):
  
  RECOVER_IF_FAILED = "2"
  
  def __init__(self):
    self.dbWrapper = None
    self.batchItem = None
    self.resource = None
    self.urlObj = None


  # #checkClassFields method checks all class's mandatory fields
  #
  def checkClassFields(self, checkList):
    for field in self.__dict__:
      if field in checkList and (not hasattr(self, field) or getattr(self, field) is None):
        raise Exception(">>> [ResourceProcess] Mandatory field must be initialized, field Name = " + field)


  # #calcLastModified - generates and returns lastModified string
  #
  # @param startTime - resource start time
  # @param res - resource object fetched from request lib
  # @param headers - resource http headers
  # @param crawledTime - resource crawled time
  # @param defaultIcrCrawlTime - default increment time
  # @returns valid or not valid checking bool value
  def generateResource(self, startTime, res, headers, crawledTime, defaultIcrCrawlTime):
    # use charset to improve encoding detect
    resource = CrawledResource()
    resource.meta_content = res.meta_res
    resource.crawling_time = int((crawledTime - startTime) * 1000)
    resource.bps = res.content_size / resource.crawling_time * 1000
    logger.warn("crawling_time: %s, bps: %s", resource.crawling_time, resource.bps)
    resource.http_code = res.status_code
    logger.warn("headers is :%s", res.headers)
    resource.content_type = res.headers.get('Content-Type', 'text/html').split(';')[0]
    if res.encoding:
      resource.charset = res.encoding
    else:
      resource.charset = ""
    resource.html_request = headers
    resource.response_header = '\r\n'.join(['%s: %s' % (k, v) for k, v in res.headers.iteritems()])
    resource.last_modified = self.calcLastModified(resource, res, defaultIcrCrawlTime)
    logger.debug("request is: %s", resource.html_request)
    logger.debug("response is: %s", resource.response_header)
    return resource


  # #calcLastModified - generates and returns lastModified string
  #
  # @param resource - own resource object
  # @param res - resource object fetched from request lib
  # @param defaultIcrCrawlTime - incrementing time values
  # @returns lastModified string
  def calcLastModified(self, resource, res, defaultIcrCrawlTime):
    self.checkClassFields(["urlObj"])
    if resource.http_code == 304:
      resource.last_modified = self.urlObj.tcDate
      #resource.last_modified = self.url["TcDate"]
    elif 'Last-Modified' in res.headers:
      resource.last_modified = res.headers['Last-Modified']
      resource.last_modified = parse(resource.last_modified).strftime('%Y-%m-%d %H:%M:%S')
    elif 'Date' in res.headers:
      resource.last_modified = res.headers['Date']
      resource.last_modified = parse(resource.last_modified).strftime('%Y-%m-%d %H:%M:%S')
    else:
      resource.last_modified = time.strftime('%Y-%m-%d %H:%M:%S', time.gmtime(time.time() - defaultIcrCrawlTime))
    logger.debug("LastModified date:" + str(resource.last_modified))
    return str(resource.last_modified)


  # #addSiteSize update sites table to increase size
  #
  # @param size content size of this crawler
  def addSiteSize(self, size):
    self.checkClassFields(["dbWrapper", "batchItem"])
    localSiteUpdate = dc_event.SiteUpdate(self.batchItem.siteId)
    for attr in localSiteUpdate.__dict__:
      if hasattr(localSiteUpdate, attr):
        setattr(localSiteUpdate, attr, None)
    localSiteUpdate.id = self.batchItem.siteId
    localSiteUpdate.tcDate = SQLExpression("NOW()")
    localSiteUpdate.size = SQLExpression(("`Size` + %s" % str(size)))
    self.dbWrapper.siteNewOrUpdate(localSiteUpdate)


  # #checkResourcesResponse checks resource's fields on valid
  #
  # @param res - current resource object, just fetched from request library
  # @param maxResourceSize - max resources size
  # @param updateSiteCallback - update site callback
  # @returns valid or not valid checking bool value
  def checkResourcesResponse(self, res, maxResourceSize, updateSiteCallback):
    ret = True
    self.checkClassFields(["resource"])
    resourceSize = res.content_size
    logger.debug("MaxResourceSize: " + str(maxResourceSize) + " ResourceSize: " + str(resourceSize))
    if resourceSize == 0 and self.resource.http_code / 100 != 3:
      self.resource.error_mask = CONSTS.ERROR_EMPTY_RESPONSE
      updateSiteCallback(CONSTS.ERROR_EMPTY_RESPONSE)
      ret = False
    elif maxResourceSize and resourceSize > maxResourceSize:
      self.resource.error_mask = CONSTS.ERROR_RESPONSE_SIZE_ERROR
      updateSiteCallback(CONSTS.ERROR_RESPONSE_SIZE_ERROR)
      logger.debug("Site MaxResourceSize limit overshooted.")
      ret = False
    else:
      self.resource.html_content = res.rendered_unicode_content
      self.resource.binary_content = res.str_content
    if ret and (res.status_code / 100 == 4 or res.status_code / 100 == 5):
      self.resource.error_mask = CONSTS.ERROR_HTTP_ERROR
      updateSiteCallback(CONSTS.ERROR_HTTP_ERROR)
      ret = False
    if ret:
      self.addSiteSize(resourceSize)
    return ret


  # #domParser generates DOM object and returns it
  #
  # @param html_recover - html recover type
  # @param res - incoming resource
  # @returns just generated dom object
  def domParser(self, htmlRecover, rendered_unicode_content, http_code):
    ret = None
    parser = lxml.etree.HTMLParser(encoding='utf-8')
    if http_code == 304:
      ret = lxml.html.fromstring("<html></html>", parser=parser)
    else:
      try:
        ret = lxml.html.fromstring(rendered_unicode_content.encode("utf-8"), parser=parser)
      except Exception:
        logger.debug("Wrong DOM model structure.")
        if htmlRecover is not None and htmlRecover == self.RECOVER_IF_FAILED:
          logger.debug("Try to fix DOM by tidylib.")
          tidy_content, errors = tidylib.tidy_document(rendered_unicode_content)
          logger.debug("tidylib errors: %s", str(errors))
          ret = lxml.html.fromstring(tidy_content, parser=parser)
    return ret
