"""@package docstring
 @file base_extractor.py
 @author Alexey <developers.hce@gmail.com>
 @link http://hierarchical-cluster-engine.com/
 @copyright Copyright &copy; 2013 IOIX Ukraine
 @license http://hierarchical-cluster-engine.com/license/
 @package HCE project node API
 @since 0.1
"""


import logging
import copy
import re
import json
import MySQLdb as mdb
import MySQLdb.cursors
import types
from contextlib import closing
import dc_processor.Constants as CONSTS
import app.Consts as APP_CONSTS
from app.Utils import varDump


# Logger initialization
logger = logging.getLogger(APP_CONSTS.LOGGER_NAME)


# time execution limit
def signal_handler(signum, frame):
  logger.debug("Time execution limit was reached: %s seconds." % str(CONSTS.TIME_EXECUTION_LIMIT))
  raise Exception("Timed out!")


# Local class constants
ERR_MSG_ADJUST_PUB_DATE = "Error in adjustPubDate: "
ERR_MSG_ADJUST_MEDIA = "Error in adjustMedia: "
ERR_MSG_ADJUST_CONTENT_UTF8_ENCODED = "Error in adjustContentUTF8Encoded: "

ERR_MSG_OK = ""

EMPTY_DATE = ""

# Adjust publication date
# If incoming publication date is array just return one of them (first)
# @param dates publication date, extracted from content
def adjustPubDate(dates):
  #logger.debug("dates: %s", dates)
  pub_date = EMPTY_DATE
  try:
    # TODO: improve to return most appropriate
    # if dates and any(i.isdigit() for i in dates):
    if isinstance(dates, list) and len(dates):
      # pub_date = dates[0]
      pub_date = " ".join(dates)
    else:
      pub_date = dates
    if pub_date and len(dates) and not re.search(r'\d+', pub_date):
      pub_date = EMPTY_DATE
  except Exception as err:
    logger.error(ERR_MSG_ADJUST_PUB_DATE + err.message)
  return pub_date


# Adjust data in media tag
# If media are PR (partial reference) adjust path
# @param medias media extracted from content
def adjustMedia(medias):
  return medias
  # valid_http_url = HttpUrl()
  # res = []
  # try:
  #  if isinstance(medias, list):
  #    for media in medias:
  #      if valid_http_url(media):
  #        res.append(media)
  # except Exception as err:
  #  logger.error(ERR_MSG_ADJUST_MEDIA + err.message)
  # return res


# Adjust data in content_encoded tag
# If content are non-meaningfull adjust it
# @param data content extracted from content
def adjustContentUTF8Encoded(data):
  return data


# Adjust data in content_encoded tag
# If content are non-meaningfull adjust it
# @param data content extracted from content
def adjustLink(data):
  if type(data) == list and len(data) > 1:
    data = data[0]
  return data



def adjustNone(data):
  return data


# #The BaseExtractor class
# This is the base class for custom extractors
# Provide basic functionality such as add tag, etc.
class BaseExtractor(object):

  properties = None

  tag = {
    CONSTS.TAG_MEDIA: adjustMedia,
    CONSTS.TAG_CONTENT_UTF8_ENCODED: adjustContentUTF8Encoded,
    CONSTS.TAG_PUB_DATE: adjustPubDate,
    CONSTS.TAG_TITLE: adjustNone,
    CONSTS.TAG_LINK: adjustLink,
    CONSTS.TAG_DESCRIPTION: adjustNone,
    CONSTS.TAG_DC_DATE: adjustNone,
    CONSTS.TAG_AUTHOR: adjustNone,
    CONSTS.TAG_GUID: adjustNone,
    CONSTS.TAG_KEYWORDS: adjustNone,
    CONSTS.TAG_MEDIA_THUMBNAIL: adjustNone,
    CONSTS.TAG_ENCLOSURE: adjustNone,
    CONSTS.TAG_MEDIA_CONTENT: adjustNone,
    CONSTS.TAG_GOOGLE: adjustNone,
    CONSTS.TAG_GOOGLE_TOTAL: adjustNone,
    CONSTS.HTML_LANG: adjustNone
  }


  tagsMask = {
    CONSTS.TAG_MEDIA: 1,
    CONSTS.TAG_CONTENT_UTF8_ENCODED: 1 << 1, CONSTS.CONTENT: 1 << 1,
    CONSTS.TAG_PUB_DATE: 1 << 2, CONSTS.PUBLISHED: 1 << 2,
    CONSTS.TAG_TITLE: 1 << 3,

    CONSTS.TAG_LINK: 1 << 4,
    CONSTS.TAG_DESCRIPTION: 1 << 5,
    CONSTS.UPDATED_PARSED: 1 << 6,
    CONSTS.TAG_DC_DATE: 1 << 7,
    
    CONSTS.TAG_AUTHOR: 1 << 8,
    CONSTS.TAG_GUID: 1 << 9,
    CONSTS.TAG_KEYWORDS: 1 << 10,
    CONSTS.TAG_MEDIA_THUMBNAIL: 1 << 11,
    
    CONSTS.TAG_ENCLOSURE: 1 << 12,
    CONSTS.TAG_MEDIA_CONTENT: 1 << 13,
    CONSTS.TAG_GOOGLE: 1 << 14,
    CONSTS.TAG_GOOGLE_TOTAL: 1 << 15,
    
    CONSTS.HTML_LANG: 1 << 16,
    CONSTS.PARENT_RSS_FEED: 1 << 17,
    CONSTS.PARENT_RSS_FEED_URLMD5: 1 << 18,
    CONSTS.SUMMARY_DETAIL: 1 << 19,
    
    CONSTS.SUMMARY: 1 << 20,
    CONSTS.COMMENTNS: 1 << 21,
    CONSTS.TAGS: 1 << 22,
    CONSTS.UPDATED: 1 << 23
}


  def __init__(self, config):
    self.db_dc_scraper_db_connect = None
    self.db_dc_scraper = None
    self.config = config
    
    #load MySQL db data
    self.loadDBBackend()

    self.name = "Base extractor"
    self.rank = CONSTS.SCRAPER_RANK_INIT

    #support processing modes
    self.process_mode = CONSTS.PROCESS_ALGORITHM_REGULAR
    self.modules= {}

    self.data = {"extractor":"Base extractor", "data":"", "name":""}
    self.loadPropertiesFromDB()

  def __str__(self):
    return "%s" % (self.name)


  def __repr__(self):
    return repr((self.name, self.rank))


  def isTagNotFilled(self, result, tag_name):
    ret = True
    if tag_name in result.tags:
      if type(result.tags[tag_name]) in types.StringTypes:
	ret = (result.tags[tag_name].strip() == "")
      elif type(result.tags[tag_name]) is types.ListType:
	if len(result.tags[tag_name]) > 0:
	  ret = False
      elif type(result.tags[tag_name]) is types.DictType:
	if "data" in result.tags[tag_name] and type(result.tags[tag_name]["data"]) in types.StringTypes:
	  ret = (result.tags[tag_name]["data"].strip() == "")
    return ret
    '''
    if (isinstance(result.tags.get(tag_name),list) and len(result.tags.get(tag_name))) \
      or result.tags.get(tag_name) != "":
      pass
    else:
      pass
    '''
    #return result.tags.get(tag_name) == ""


  def isTagValueNotEmpty(self, tag_value):
    full = None
    if type(tag_value) == list:
      # FIX IT: if list check if list items not empty
      #logger.debug("Tag data value: <%s>", tag_value)
      # Check if data in tag is empty list
      # return None if data list hold empty values
      # FIX IT: rewrite it in more convenient way
      if len(tag_value) == 0:
        full = None
      else:
        full = tag_value
      '''
      if tag_value == ['']:
        #logger.debug("Tag data value is empty: <%s>", tag_value)
        full = None
      else:
        full = tag_value
      '''
    else:
      full = tag_value
    logger.debug("Returned full value: <%s>", full)
    return full


  def addTag(self, result, tag_name, tag_value, xpath=""):
    ret = False
    try:
      if tag_value and not isinstance(tag_value, list):
        pass
      if tag_value and isinstance(tag_value, list):
        pass
      tag_value = self.tag[tag_name](tag_value)
    except Exception as err:
      logger.debug("Missing adjustment for section: " + err.message)
    result.errorCode = 0
    result.errorMessage = ERR_MSG_OK

    if (tag_name not in result.tags.keys() and self.isTagValueNotEmpty(tag_value) is not None) or \
    (self.isTagNotFilled(result, tag_name) and self.isTagValueNotEmpty(tag_value) is not None):
      logger.info("PASS")
      data = {"extractor":"Base extractor", "data":"", "name":""}
      data["data"] = tag_value
      data["name"] = tag_name
      data["xpath"] = xpath
      data["extractor"] = self.__class__.__name__
      result.tags[tag_name] = data
      ret = True
    return ret


  def loadDBBackend(self):
    try:
      className = self.__class__.__name__
      logger.debug("className: %s" % className)
      dbHost = self.config.get(CONSTS.DB_SECTION, CONSTS.DB_HOST)
      dbPort = int(self.config.get(CONSTS.DB_SECTION, CONSTS.DB_PORT))
      dbUser = self.config.get(CONSTS.DB_SECTION, CONSTS.DB_USER)
      dbPWD = self.config.get(CONSTS.DB_SECTION, CONSTS.DB_PWD)
      self.db_dc_scraper_db = self.config.get(CONSTS.DB_SECTION, CONSTS.DB_SCRAPERS)
      self.db_dc_scraper = self.config.get(className, CONSTS.DB_SCRAPERS)
      self.db_dc_scraper_db_connect = mdb.connect(dbHost, dbUser, dbPWD, self.db_dc_scraper_db, dbPort)
    except Exception, err:
      logger.debug(CONSTS.MSG_ERROR_LOAD_DB_BACKEND + str(err.message))


  ##
  #
  #
  def loadPropertiesFromDB(self):
    logger.debug("Start load Scraper's properties from DB")
    self.DBConnector = self.db_dc_scraper_db_connect
    try:
      query = "SELECT `Value` FROM %s WHERE `Name`='%s'" % (self.db_dc_scraper, self.__class__.__name__)
      logger.debug("query: %s" % query)
      res = self.executeQuery(query)
      if len(res)>0:
        #self.properties = json.loads(res)
        self.properties = json.loads(res[0]["Value"])[CONSTS.PROPERTIES_KEY]
      else:
        logger.debug("properties wasn't found in %s table" % self.db_dc_scraper)
    except Exception, err:
      logger.debug("Error load Scraper's properties from DB: %s" % str(err.message))
    logger.debug("End load Scraper's properties from DB")


  # #helper function for correct query execution
  #
  # @param query query for execution
  def executeQuery(self, query):
    logger.info("Connector: " + str(self.DBConnector))
    try:
      with closing(self.DBConnector.cursor(MySQLdb.cursors.DictCursor)) as cursor:
        cursor.execute(query)
        self.DBConnector.commit()
        return cursor.fetchall()
    except mdb.Error as err:  # @todo logging in db_task
      self.DBConnector.rollback()
      logger.error("%s %s" % (err.args[0], err.args[1]))
      raise err


  ## calculateMetrics
  #
  def calculateMetrics(self, response):
    logger.info("calculateMetrics")
    try:
      for metric in response.metrics:
        logger.info("metric's name: %s" % metric.name)
        logger.info("response's tags: %s" % varDump(response.tags))
        logger.info("metric instance: %s" % varDump(metric))
        metric.calculateMetricValue(response.tags)
    except Exception, err:
      logger.error(CONSTS.MSG_ERROR_CALC_METRICS + str(err.message))
      raise err
