'''
Created on Mar 05, 2014

@package: dtm
@author: scorp
@link: http://hierarchical-cluster-engine.com/
@copyright: Copyright &copy; 2013-2014 IOIX Ukraine
@license: http://hierarchical-cluster-engine.com/license/
@since: 0.1
'''

from app.BaseServerManager import BaseServerManager
from app.LogFormatter import LogFormatterEvent
from dbi.dbi import DBI
from dbi.dbi import DBIErr
from transport.ConnectionBuilderLight import ConnectionBuilderLight
from transport.Event import EventBuilder
from Constants import EVENT_TYPES as EVENT
from EEResponsesTable import EEResponsesTable
from EventObjects import GeneralResponse
from TasksDataTable import TasksDataTable
import transport.Consts
import ConfigParser
import logging
#import pickle
try:
  import cPickle as pickle
except ImportError:
  import pickle
import Constants as CONSTANTS
import dbi.Constants as DBI_CONSTANTS


logger = logging.getLogger(CONSTANTS.LOGGER_NAME)


##Class contents TasksDataManager module implementation
#
class TasksDataManager(BaseServerManager):


  TASK_DATA_MANAGER_SERV_NAME = "Server"
  TASK_DATA_MANAGER_SERV_CONFIG_NAME = "ServerName"


  ##constructor
  #initialise all class variable and recieve config as param
  def __init__(self, config, connectionBuilder=None):
    BaseServerManager.__init__(self)
    self.moduleName = self.__class__.__name__
    self.servIndex = 1
    self.eventBuilder = EventBuilder()
    self.setEventHandler(EVENT.NEW_TASK, self.onNewTask)
    self.setEventHandler(EVENT.FETCH_TASK_DATA, self.onFetchTask)
    self.setEventHandler(EVENT.UPDATE_TASK, self.onUpdateTask)
    self.setEventHandler(EVENT.DELETE_TASK_DATA, self.onDeleteTask)
    self.setEventHandler(EVENT.INSERT_EE_DATA, self.onInsertEEResponse)
    self.setEventHandler(EVENT.FETCH_EE_DATA, self.onFetchEEResponse)
    self.setEventHandler(EVENT.DELETE_EE_DATA, self.onDeleteEEResponse)
    self.connectionBuilder = connectionBuilder
    self.config = config
    self.dbi = None
    self.connectionInit()
    self.dbInit()


  ##connectionInit method
  #initializes internal inproc connection
  def connectionInit(self):
    if self.connectionBuilder == None:
      self.connectionBuilder = ConnectionBuilderLight()
    try:
      localConnection = self.connectionBuilder.build(transport.Consts.SERVER_CONNECT, \
                        self.config.get(self.moduleName, self.TASK_DATA_MANAGER_SERV_CONFIG_NAME))
      self.addConnection(self.TASK_DATA_MANAGER_SERV_NAME + str(self.servIndex), localConnection)
      self.servIndex = self.servIndex + 1
    except ConfigParser.NoSectionError:
      logger.error(">>> TasksDataManager can't read config - Section Error")
    except ConfigParser.NoOptionError:
      logger.error(">>> TasksDataManager can't read config - Option Error")


  ##dbInit method
  #initializes internal database API
  def dbInit(self):
    if self.config != None:
      dic = None
      dic = dict(self.config.items(self.moduleName))
      if dic != None:
        self.dbi = DBI(dic, "dtm_tdm")


  ##dbiProcessing method
  #Method contains main processing with database API
  def dbiProcessing(self, data, eventType):
    ret = None
    retEventType = None
    if eventType == EVENT.NEW_TASK:
      self.dbi.insert(data)
      retEventType = EVENT.NEW_TASK_RESPONSE
    elif eventType == EVENT.FETCH_TASK_DATA:
      ret = self.dbi.fetch(data, "id=%s" % data.id)
      retEventType = EVENT.FETCH_TASK_DATA_RESPONSE
    elif eventType == EVENT.UPDATE_TASK:
      self.dbi.update(data, "id=%s" % data.id)
      retEventType = EVENT.UPDATE_TASK_RESPONSE
    elif eventType == EVENT.DELETE_TASK_DATA:
      self.dbi.delete(data, "id=%s" % data.id)
      retEventType = EVENT.DELETE_TASK_DATA_RESPONSE
    elif eventType == EVENT.INSERT_EE_DATA:
      self.dbi.insert(data)
      retEventType = EVENT.INSERT_EE_DATA_RESPONSE
    elif eventType == EVENT.FETCH_EE_DATA:
      ret = self.dbi.fetch(data, "id=%s" % data.id)
      retEventType = EVENT.FETCH_EE_DATA_RESPONSE
    elif eventType == EVENT.DELETE_EE_DATA:
      self.dbi.delete(data, "id=%s" % data.id)
      retEventType = EVENT.DELETE_EE_DATA_RESPONSE
    return (ret, retEventType)


  ##eventProcessing method
  #Method contains main processing with incoming events
  def eventProcessing(self, event):
    serializeStr = pickle.dumps(event.eventObj)
    if event.eventType == EVENT.NEW_TASK or \
       event.eventType == EVENT.FETCH_TASK_DATA or \
       event.eventType == EVENT.UPDATE_TASK or \
       event.eventType == EVENT.DELETE_TASK_DATA:
      data = TasksDataTable()
    else:
      data = EEResponsesTable()
    data.id = event.eventObj.id
    data.data = serializeStr
    dbiRet = None
    retEventType = None
    try:
      dbiRet, retEventType = self.dbiProcessing(data, event.eventType)
    except DBIErr as err:
      dbiRet = GeneralResponse()
      dbiRet.errorCode = err.errCode
      dbiRet.errorMessage = ">>> Some DBI error in TasksDataManager.eventProcessing [" + str(err.message) + "]"
    else:
      if dbiRet is not None and len(dbiRet) > 0 and dbiRet[0] != None and dbiRet[0].data != None:
        dbiRet = pickle.loads(str(dbiRet[0].data))
      else:
        dbiRet = GeneralResponse()
    retEvent = self.eventBuilder.build(retEventType, dbiRet)
    return retEvent


  ##eventProcessing method
  #Method contains main error processing for incoming events
  def badEventType(self, msg, event):
    errorStr = msg + str(event.eventType)
    logger.error(LogFormatterEvent(event, [], errorStr))
#    raise Exception(errorStr)


  ##Callbacks methods
  #Event callbacks
  def onNewTask(self, event):
    logger.info(LogFormatterEvent(event, [], ">>> TasksDataManager [NEW_TASK_REQUEST] Handler start"))
    if event.eventType != EVENT.NEW_TASK:
      self.badEventType(">>> Wrong Event type [NEW_TASK_REQUEST] != ", event)
    self.reply(event, self.eventProcessing(event))
    logger.info(LogFormatterEvent(event, [], ">>> TasksDataManager [NEW_TASK_REQUEST] Handler finish"))


  def onFetchTask(self, event):
    logger.info(LogFormatterEvent(event, [], ">>> TasksDataManager [FETCH_TASK_REQUEST] Handler start"))
    if event.eventType != EVENT.FETCH_TASK_DATA:
      self.badEventType(">>> Wrong Event type [FETCH_TASK_REQUEST] != ", event)
    self.reply(event, self.eventProcessing(event))
    logger.info(LogFormatterEvent(event, [], ">>> TasksDataManager [FETCH_TASK_REQUEST] Handler finish"))


  def onUpdateTask(self, event):
    logger.info(LogFormatterEvent(event, [], ">>> TasksDataManager [UPDATE_TASK_REQUEST] Handler start"))
    if event.eventType != EVENT.UPDATE_TASK:
      self.badEventType(">>> Wrong Event type [UPDATE_TASK_REQUEST] != ", event)
    self.reply(event, self.eventProcessing(event))
    logger.info(LogFormatterEvent(event, [], ">>> TasksDataManager [UPDATE_TASK_REQUEST] Handler finish"))


  def onDeleteTask(self, event):
    logger.info(LogFormatterEvent(event, [], ">>> TasksDataManager [DELETE_TASK_DATA_REQUEST] Handler start"))
    if event.eventType != EVENT.DELETE_TASK_DATA:
      self.badEventType(">>> Wrong Event type [DELETE_TASK_DATA_REQUEST] != ", event)
    self.reply(event, self.eventProcessing(event))
    logger.info(LogFormatterEvent(event, [], ">>> TasksDataManager [DELETE_TASK_DATA_REQUEST] Handler finish"))


  def onInsertEEResponse(self, event):
    logger.info(LogFormatterEvent(event, [], ">>> TasksDataManager [INSERT_EE_REQUEST] Handler start"))
    if event.eventType != EVENT.INSERT_EE_DATA:
      self.badEventType(">>> Wrong Event type [INSERT_EE_REQUEST] != ", event)
    self.reply(event, self.eventProcessing(event))
    logger.info(LogFormatterEvent(event, [], ">>> TasksDataManager [INSERT_EE_REQUEST] Handler finish"))


  def onFetchEEResponse(self, event):
    logger.info(LogFormatterEvent(event, [], ">>> TasksDataManager [FETCH_EE_REQUEST] Handler start"))
    if event.eventType != EVENT.FETCH_EE_DATA:
      self.badEventType(">>> Wrong Event type [FETCH_EE_REQUEST] != ", event)
    self.reply(event, self.eventProcessing(event))
    logger.info(LogFormatterEvent(event, [], ">>> TasksDataManager [FETCH_EE_REQUEST] Handler finish"))


  def onDeleteEEResponse(self, event):
    logger.info(LogFormatterEvent(event, [], ">>> TasksDataManager [DELETE_EE_REQUEST] Handler start"))
    if event.eventType != EVENT.DELETE_EE_DATA:
      self.badEventType(">>> Wrong Event type [DELETE_EE_REQUEST] != ", event)
    self.reply(event, self.eventProcessing(event))
    logger.info(LogFormatterEvent(event, [], ">>> TasksDataManager [DELETE_EE_REQUEST] Handler finish"))
