#include "Admin.hpp"
#include "AdminCommandParameters.hpp"

namespace HCE {
  namespace handlers {

    const unsigned char ADMIN_REQUEST_TYPE_UNDEFINED = 0;
    const unsigned char ADMIN_REQUEST_TYPE_EXTERNAL = 1;
    const unsigned char ADMIN_REQUEST_TYPE_INTERNAL = 2;

    Admin::Admin(std::string name, zmq::context_t& ctx, const std::string& clientIdentity, std::string connectionString, int64_t startedAt, unsigned char logPriority) :
        Handler(name, ctx, connectionString, startedAt, logPriority), _adminServerSock(NULL), _clientIdentity(clientIdentity), _clientsManagerInproc(NULL), _statRequestsTotal(0){

      //!<Init all the rest
      initialize();
    }

    Admin::~Admin(){
      deinitialize();
    }

    int Admin::initialize(void){
      logger.log(NODE_LOG_MODE_INFORMATION) << "Initialize client [" << _clientIdentity << "] with connection [" << _connectionString << "]" << flush;

      //!<Rebuild inproc_proxy connection
      rebuildServerConnection(_inprocAdminSock, NODE_CS_INPROC_ADMIN, ZMQ_ROUTER);

      //!<Rebuild adminserver connection
      if(!rebuildServerConnection(_adminServerSock, _connectionString, ZMQ_ROUTER)){
        //Fatal error, can't bind admin port, abort
        ::exit(NODE_EXIT_CODE_GENERAL_CRITICAL_ERROR);
      }

      //!<Init clients queue manager
      _clientsManagerInproc = new ClientsQueueManager(&_queueInproc, 0);

      //!<Init stat vars
      _statRequestsTotal = 0;

      return true;
    }

    void Admin::deinitialize(void){
      if(_adminServerSock){
        if(_adminServerSock->connected()){
          _adminServerSock->close();
        }
        delete _adminServerSock;
        _adminServerSock = NULL;
      }

      _queueInproc.clear();

      if(_clientsManagerInproc){
        delete _clientsManagerInproc;
      }
    }

    void Admin::handleInternalMessage(void){
      //!<Fetch message from dataserver
      zmsg msg(*_inprocAdminSock);

      logger.log(NODE_LOG_MODE_TRACE) << "MSG FROM inproc_admin\n" << msg.dump() << flush;

      //!<Process dataserver's message
      if(msg.parts() == NODE_MSG_DATA_MIN_FIELDS){
        //!<Control message
        processControlMessageInproc(msg);
      }else if(msg.parts() > NODE_MSG_DATA_MIN_FIELDS){
        //!<Data message
        processDataMessageInproc(msg);
      }else{
        logger.log(NODE_LOG_MODE_TRACE) << "ERROR of message structure inproc_admin, wrong fields number " << msg.parts()
            << endl << msg.dump() << flush;
      }
    }

    void Admin::registerClientInproc(const std::string& identity){
      _clientsManagerInproc->remove(identity);
      _clientsManagerInproc->append(identity);
      logger.log(NODE_LOG_MODE_DEBUG) << "READY message inproc from client [" << identity << "] received" << flush;
    }

    void Admin::processControlMessageInproc(zmsg& msg){
      //!<Fetch the clientIdentity field
      std::string clientIdentity((const char*)msg.pop_front().c_str());
      std::string addr((msg.address())?msg.address():"");

      if(msg.address() == NODE_MSG_READY){
        registerClientInproc(clientIdentity);
      }else if(addr.compare(0, NODE_MSG_PROPERTIES.length(), NODE_MSG_PROPERTIES)==0){
        //!<Update clients property
        HCE::handlers::PropertyMessage propertyMessage;
        propertyMessage.extract(addr);
        if (!propertyMessage.isError()){
          _clientsManagerInproc->refresh(clientIdentity, propertyMessage.getProperty());
          logger.log(NODE_LOG_MODE_TRACE) << "Received property message from client [" << clientIdentity << "] " << flush;
        }else{
          //!<Invalid property message
          logger.log(NODE_LOG_MODE_ERROR) << "Operation extract property message from client [" << clientIdentity << "] has error: " << propertyMessage.getErrorMsg() << flush;
          logger.log(NODE_LOG_MODE_TRACE) << msg.dump() << flush;
        }
      }else{
        //!<Invalid message type
        logger.log(NODE_LOG_MODE_ERROR) << "INVALID inproc control message from client [" << clientIdentity << "]" << flush;
        logger.log(NODE_LOG_MODE_TRACE) << msg.dump() << flush;
      }
    }

    std::vector<std::string> Admin::parseAdminRequest(const std::string& command){
      //!<Parse command string
      std::vector<std::string> parts;
      //!<Split on parts
      std::istringstream partsStream(command);
      std::string tmpString;
      while(std::getline(partsStream, tmpString, NODE_ADMIN_COMMAND_DELIMITER)){
        parts.push_back(tmpString);
      }

      return parts;
    }

    unsigned char Admin::processAdminRequest(zmsg& msg){
      unsigned char ret = ADMIN_REQUEST_TYPE_UNDEFINED;

      //!<Fetch the clientIdentity field
      std::string clientIdentity((const char*)msg.pop_front().c_str());

      //!<Fetch message Id
      std::string messageId((const char*)msg.pop_front().c_str());
      //!<Fetch message body
      std::string messageBody((const char*)msg.pop_front().c_str());
      logger.log(NODE_LOG_MODE_DEBUG)<< "RECEIVED REQUEST from adminserver client [" << clientIdentity << "], message [" << messageId << "] body " << messageBody << flush;

      //!<Parse message body, supposes minimum 3 items: 0 - DESTINATION HANDLER IDENTITY, 1 - COMMAND_NAME, 2 - PARAMETERS
      std::vector<std::string> requestItems = parseAdminRequest(messageBody);
      if (requestItems.size() < 3){
        logger.log(NODE_LOG_MODE_ERROR) << "Received admin request has wrong items count = " << requestItems.size() << flush;
      }

      //!<Check items number and command destination type
      if(requestItems.size() > 2){
        logger.log(NODE_LOG_MODE_TRACE) << "Admin::processAdminRequest  requestItems[0]: " << requestItems[0] << flush;
        logger.log(NODE_LOG_MODE_TRACE) << "Admin::processAdminRequest  requestItems[1]: " << requestItems[1] << flush;
        logger.log(NODE_LOG_MODE_TRACE) << "Admin::processAdminRequest  requestItems[2]: " << requestItems[2] << flush;

        //!<Parse command parameters
        AdminCommandParameters adminCommandParameters(requestItems[2]);
        if (adminCommandParameters.isError()){
          logger.log(NODE_LOG_MODE_ERROR) << "Operation extract of command parameters has error: " << adminCommandParameters.getErrorMsg() << flush;
        }else{
          logger.log(NODE_LOG_MODE_TRACE) << "Admin::processAdminRequest  adminCommandParameters.getParameters(): " << adminCommandParameters.getParameters() << flush;
          logger.log(NODE_LOG_MODE_TRACE) << "Admin::processAdminRequest  adminCommandParameters.getRealtime(): " << adminCommandParameters.getRealtime() << flush;
          logger.log(NODE_LOG_MODE_TRACE) << "Admin::processAdminRequest  adminCommandParameters.isRealtime(): " << std::boolalpha << adminCommandParameters.isRealtime() << flush;
        }

        if (adminCommandParameters.isJsonString() && requestItems[1]==NODE_MSG_PROPERTIES)
          requestItems[2] = adminCommandParameters.getParameters();

        if(((_clientsManagerInproc->exists(std::string(requestItems[0].c_str())) && requestItems[1]!=NODE_MSG_PROPERTIES) ||
            (_clientsManagerInproc->exists(std::string(requestItems[0].c_str())) && requestItems[1]==NODE_MSG_PROPERTIES && adminCommandParameters.isRealtime()))){

          //!<External admin command for inproc handler
          //!<Insert inproc destination identity
          msg.append(requestItems[0].c_str());
          //!<Insert source message Id
          msg.append(messageId.c_str());
          //!<Insert command name
          msg.append(requestItems[1].c_str());
          //!<Insert command parameters
          msg.append(requestItems[2].c_str());
          //!<Insert source client identity
          msg.append(clientIdentity.c_str());

          logger.log(NODE_LOG_MODE_DEBUG) << "External admin command message prepared for client [" << requestItems[0].c_str() << "], command ["
              << requestItems[1].c_str() << "], params [" << requestItems[2].c_str() << "] " << " bytes" << flush;

          ret = ADMIN_REQUEST_TYPE_EXTERNAL;

        }else if (_clientsManagerInproc->exists(std::string(requestItems[0].c_str())) && !adminCommandParameters.isRealtime()){

          //!<Response message value
          std::string responseMsg = _clientsManagerInproc->getProperty(requestItems[0]);
          //!<Insert source client identity
          msg.append(clientIdentity.c_str());
          //!<Insert source message Id
          msg.append(messageId.c_str());
          //!<Insert response message
          msg.append(responseMsg.c_str());

          ret = ADMIN_REQUEST_TYPE_INTERNAL;

        }else if(requestItems[0] == "" || requestItems[0] == NODE_ADMIN_HANDLER){

          //!<Response message value
          std::string responseMsg = processAdminCommands(requestItems[1], requestItems[2]);

          if(responseMsg == ""){
            //!<Wrong or not supported command
            responseMsg = NODE_ADMIN_ERROR_ERROR + NODE_ADMIN_COMMAND_DELIMITER + NODE_ADMIN_ERROR_UNSUPPORTED_COMMAND;
            logger.log(NODE_LOG_MODE_DEBUG) << "Internal admin command [" << requestItems[1].c_str() << "], params [" << requestItems[2].c_str() << "] " << " not supported" << flush;
          }else{
            ret = ADMIN_REQUEST_TYPE_INTERNAL;
            logger.log(NODE_LOG_MODE_DEBUG) << "Internal admin command processed successfully" << flush;
          }
          //!<Insert source client identity
          msg.append(clientIdentity.c_str());
          //!<Insert source message Id
          msg.append(messageId.c_str());
          //!<Insert response message
          msg.append(responseMsg.c_str());

        }else{
          //!<Wrong or not supported command
          std::string responseMsg = NODE_ADMIN_ERROR_ERROR + NODE_ADMIN_COMMAND_DELIMITER + NODE_ADMIN_ERROR_HANDLER_NOT_FOUND + " [" + requestItems[0] + "]";

          logger.log(NODE_LOG_MODE_DEBUG) << "Handler [" << requestItems[0] << "] not registered or not connected to admin server." << flush;

          //!<Insert source client identity
          msg.append(clientIdentity.c_str());
          //!<Insert source message Id
          msg.append(messageId.c_str());
          //!<Insert response message
          msg.append(responseMsg.c_str());
        }
      }else{
        //!<Wrong incoming command message body format from manager or destination admin handler (class instance by name) not connected
        //!<Insert source client identity
        msg.append(clientIdentity.c_str());
        //!<Insert source message Id
        msg.append(messageId.c_str());
        //!<Insert error message
        msg.append(NODE_ADMIN_ERROR_WRONG_FORMAT.c_str());

        logger.log(NODE_LOG_MODE_DEBUG) << "ADMIN REQUEST MESSAGE ERROR - items " << requestItems.size() << " " << NODE_ADMIN_ERROR_WRONG_FORMAT << flush;
      }

      logger.log(NODE_LOG_MODE_TRACE) << "MSG FROM processAdminRequest():" << endl << msg.dump() << flush;

      //Register total requests number
      _statRequestsTotal++;

      return ret;
    }

    void Admin::stopClients(void){
      //!<Send NODE_MSG_STOP message for all clients connected to internal admin socket
      for(std::vector<ClientWorkerItem>::iterator it = _queueInproc.begin(); it < _queueInproc.end(); ++it){
        //!<Create response message and fill with fields
        zmsg clientMsg(_clientIdentity.c_str());      //Source handler identity
        clientMsg.wrap(it->identity.c_str(), NULL);   //Destination handler identity
        clientMsg.append(NODE_MSG_STOP.c_str());      //Message Id
        clientMsg.append(NODE_MSG_STOP.c_str());      //Message body
        clientMsg.append(NODE_ADMIN_HANDLER.c_str()); //Message source identity

        //!<Send response
        clientMsg.send(*_inprocAdminSock);

        logger.log(NODE_LOG_MODE_DEBUG) << "STOP REQUEST SENT to _adminServerSock for client [" << it->identity << "]" << flush;
      }
    }

    std::string Admin::processAdminCommands(std::string& command, std::string& parameters){
      std::string ret = "";

      logger.log(NODE_LOG_MODE_DEBUG) << "Admin internal command [" << command << "], parameters [" << parameters << "]" << flush;

      //!<Process commands
      if(command == NODE_MSG_SHUTDOWN){
        //!<Get timeout of shutdown action (ms), 0 - default
        unsigned int timeout = 0;
        std::istringstream(parameters) >> timeout;
        if(timeout > 0){
          //!<Wait timeout
          s_sleep(timeout);
        }

        //Stop all clients handlers connected by inproc ZMQ socket
        stopClients();

        //!<Stop admin handler main loop
        setInProgress(false);

        ret = NODE_ADMIN_ERROR_OK;

        logger.log(NODE_LOG_MODE_DEBUG) << "Done with timeout " << timeout << " ms" << flush;
      }else if(command == NODE_MSG_ECHO){
        std::stringstream result;
        result << NODE_ADMIN_ERROR_OK << NODE_ADMIN_COMMAND_DELIMITER << "command=" << NODE_MSG_ECHO;
        ret = result.str();

        logger.log(NODE_LOG_MODE_DEBUG) << "Done" << flush;
      }else if(command == NODE_MSG_TIME){
        std::stringstream timeStream;
        timeStream << (unsigned int)(s_clock() - _startedAt);

        std::stringstream result;
        result << NODE_ADMIN_ERROR_OK << NODE_ADMIN_COMMAND_DELIMITER << "time=" << timeStream.str();
        ret = result.str();

        logger.log(NODE_LOG_MODE_DEBUG) << "Done" << flush;
      }else if(command == NODE_MSG_PROPERTIES){
        //!<Make property information
        ret = getPropertyInformation();
      }else if(command == NODE_MSG_LOG_LEVEL_GET){
        ret = NODE_ADMIN_ERROR_OK + NODE_ADMIN_COMMAND_DELIMITER + NODE_ADMIN_PROPERTY_NAME_LOG + "=" + std::to_string(_logPriority) + NODE_ADMIN_COMMAND_DELIMITER + _name;
      }else if(command == NODE_MSG_LOG_LEVEL_SET){
        //!<Parse command parameters and fetch log level value
        int logp = _logPriority;
        //Parse log priority from parameters
        std::istringstream(parameters) >> logp;
        //Make response
        ret = NODE_ADMIN_ERROR_OK + NODE_ADMIN_COMMAND_DELIMITER + NODE_ADMIN_PROPERTY_NAME_LOG + "=" + std::to_string(_logPriority) + NODE_ADMIN_COMMAND_DELIMITER + _name;
        //Set new log priority
        _logPriority = logp;
        Poco::Logger::root().setLevel(_logPriority);
      }else if(command == NODE_MSG_POLL_TIMEOUT_GET){
        ret = NODE_ADMIN_ERROR_OK + NODE_ADMIN_COMMAND_DELIMITER + NODE_ADMIN_PROPERTY_POLL_TIMEOUT + "=" + std::to_string(_pollTimeout/1000);
      }else if(command == NODE_MSG_POLL_TIMEOUT_SET){
        //!<Parse command parameters and fetch log level value
        unsigned int pollTimeout = _pollTimeout;
        //Parse log priority from parameters
        std::istringstream(parameters) >> pollTimeout;
        //Make response
        ret = NODE_ADMIN_ERROR_OK + NODE_ADMIN_COMMAND_DELIMITER + NODE_ADMIN_PROPERTY_POLL_TIMEOUT + "=" + std::to_string(_pollTimeout/1000);
        //Set new poll timeout value
        _pollTimeout = (pollTimeout*1000);
      }

      return ret;
    }

    //!<Get property information
    std::string Admin::getPropertyInformation(void){
      std::stringstream result;
      result << NODE_ADMIN_ERROR_OK
          << NODE_ADMIN_COMMAND_DELIMITER << getProperties() << "clientIdentity=" << _clientIdentity
          << NODE_ADMIN_COMMAND_DELIMITER << "clients=" << _clientsManagerInproc->getAdminClients()
          << NODE_ADMIN_COMMAND_DELIMITER << "requestsTotal=" << _statRequestsTotal;
      return result.str();
    }

    void Admin::processDataMessage(zmsg& msg){
      unsigned char destinationType = processAdminRequest(msg);

      if(destinationType == ADMIN_REQUEST_TYPE_EXTERNAL){
        //!<Send admin request message to inproc client to handle admin message by destination handler
        msg.send(*_inprocAdminSock);

        logger.log(NODE_LOG_MODE_DEBUG) << "REQUEST SENT to inproc_admin socket to execute externally" << flush;
      }else if(destinationType == ADMIN_REQUEST_TYPE_INTERNAL){
        //!<Send okay response message, command was executed internally
        msg.send(*_adminServerSock);

        logger.log(NODE_LOG_MODE_DEBUG) << "REQUEST was executed internally" << flush;
      }else{
        //!<Send error response message - wrong format of requested command or another parse error, send error response message to manager
        msg.send(*_adminServerSock);

        logger.log(NODE_LOG_MODE_DEBUG) << "REQUEST execution error : wrong format of requested command or another parse error" << flush;
      }
    }

    void Admin::processDataMessageInproc(zmsg& msg){
      //!<Fetch the clientIdentity field
      std::string clientIdentity((const char*)msg.pop_front().c_str());
      //!<Fetch message Id
      std::string messageId((const char*)msg.pop_front().c_str());
      //!<Fetch message body
      std::string messageBody((const char*)msg.pop_front().c_str());
      //!<Fetch the source clientIdentity field
      std::string sourceClientIdentity((const char*)msg.pop_front().c_str());

      logger.log(NODE_LOG_MODE_DEBUG) << "RECEIVED RESPONSE from inproc client [" << clientIdentity << "], message [" << messageId << "] body [" << messageBody << "]" << flush;

      //!<Create new response message
      zmsg responseMsg;
      //!<Insert client identity
      responseMsg.append(sourceClientIdentity.c_str());
      //!<Insert source message Id
      responseMsg.append(messageId.c_str());
      //!<Insert message body
      responseMsg.append(messageBody.c_str());

      //!<Send response to manager
      responseMsg.send(*_adminServerSock);

      logger.log(NODE_LOG_MODE_DEBUG) << "RESPONSE SENT to '_adminServerSock'" << flush;
    }

    void Admin::handleExternalMessage(void){
      //!<Fetch message from dataserver
      zmsg msg(*_adminServerSock);

      logger.log(NODE_LOG_MODE_TRACE) << "MSG FROM adminserver" << endl << msg.dump() << flush;

      //!<Process dataserver's message
      if(msg.parts() > NODE_MSG_DATA_MIN_FIELDS){
        //!<Data message
        processDataMessage(msg);
      }else{
        logger.log(NODE_LOG_MODE_TRACE) << "ERROR of message structure adminserver, wrong fields number " << msg.parts()
            << endl << msg.dump() << flush;
      }
    }

    void* Admin::process(void){
      if(_adminServerSock && _inprocAdminSock){
          logger.log(NODE_LOG_MODE_INFORMATION) << NODE_ADMIN_ERROR_MAIN_MSG_LOOP_STARTED << flush;

        //!<Enter main loop of processing of message
        while(getInProgress()){
          try {
            //!<Init polling pool
            zmq::pollitem_t items[] = {{*_adminServerSock, 0, ZMQ_POLLIN, 0}, {*_inprocAdminSock, 0, ZMQ_POLLIN, 0}};

            //!<Poll dataserver
            zmq::poll(items, 2, getPollTimeout());

            //!<Handle adminserver socket's message
            if(items[0].revents & ZMQ_POLLIN){
              handleExternalMessage();
            }

            //!<Handle inproc_admin socket message
            if(items[1].revents & ZMQ_POLLIN){
              handleInternalMessage();
            }
          }catch(zmq::error_t& error){
            std::cout << "Admin E: " << error.what() << std::endl;
          }catch(std::exception& e){
            std::cout << "Admin E: " << e.what() << std::endl;
          }catch(...){
            std::cout << "Admin E: Unknown exception" << std::endl;
          }

        }

        logger.log(NODE_LOG_MODE_INFORMATION) << NODE_ADMIN_ERROR_MAIN_MSG_LOOP_STOPPED << flush;
      }else{
        logger.log(NODE_LOG_MODE_FATAL) << "Admin::process() aborted because not all required sockets created : " << NODE_CRITICAL_ERROR_MESSAGE << flush;

        //TODO: possible more intelligent termination implementation
        ::exit(NODE_EXIT_CODE_GENERAL_CRITICAL_ERROR);
      }

      return NULL;
    }

  }
}
