#include "DataClientData.hpp"
#include "DRCEResourceExtractor.hpp"
#include "ServiceMessages.hpp"

namespace HCE {
  namespace handlers {

    DataClientData::DataClientData(std::string name, zmq::context_t& ctx, const std::string& clientIdentity, std::string connectionString, unsigned char dataClientMode, int64_t startedAt,
        unsigned char logPriority, const std::string& iniFile) :
        Handler(name, ctx, connectionString, startedAt, logPriority, iniFile), DataNodeOptions(),
        _clientIdentity(clientIdentity), _dataClientSockMode(dataClientMode), _iniFile(iniFile), _shiftCounter(){

      //!<Init all the rest
      initialize();
    }

    DataClientData::~DataClientData(){
      deinitialize();
    }

    int DataClientData::initialize(void){
      logger.log(NODE_LOG_MODE_INFORMATION) << "Initialize client [" << _clientIdentity << "] with connection [" << _connectionString << "]" << flush;

      //!<Init sockets
      _inprocDataSock = NULL;
      _dataClientSock = NULL;
      _dataClientSockConnected = false;

      //!<Node init from file
      initNode(_iniFile, logger);

      //!<Heartbeat at time mark
      _heartbeatAt = s_clock() + getHeartbeatDelay();
      //!<Heartbeated time mark
      _heartbeatedAt = s_clock();

      //!<Send out periodical stats at regular intervals
      _pstatsAt = s_clock() + STAT_INTERVAL;

      //!<Stat vars
      _statRequests = 0;
      _statResponses = 0;
      _statRequestsSize = 0;
      _statResponsesSize = 0;
      _statRequestsProcessed = 0;
      _statResponsesProcessed = 0;
      _statRequestsProcessedSize = 0;
      _statResponsesProcessedSize = 0;
      _statQPS = 0;
      _statRPS = 0;
      _statRequestsBandwidth = 0;
      _statResponsesBandwidth = 0;

      //!<Rebuild admin client socket
      rebuildAdminClientSocket();

      //!<Rebuild inproc_proxy connection
      rebuildClientConnection(_inprocDataSock, NODE_CS_INPROC_DATA, ZMQ_DEALER, _clientIdentity);

      //!<Create dataclient socket, connect to upper node and send READY message
      rebuildDataClientSocket();

      return true;
    }

    void DataClientData::deinitialize(void){
      if(_inprocDataSock){
        delete _inprocDataSock;
      }
      _inprocDataSock = NULL;

      if(_dataClientSock){
        sendByeMessage(_dataClientSock, _clientIdentity);
        _dataClientSockConnected = false;
        delete _dataClientSock;
      }
      _dataClientSock = NULL;

      HCE::drce::ResourceUsageExtractor::shutdown();
    }

    void DataClientData::rebuildDataClientSocket(void){
      if(_connectionString != "" && _connectionString != "0"){
        _dataClientSockConnected = rebuildClientConnection(_dataClientSock, _connectionString, ZMQ_DEALER, _clientIdentity, 0, NODE_MSG_READY);
      }else{
        logger.log(NODE_LOG_MODE_INFORMATION) << "Client connection was not rebuilt due connection string empty value [" << _connectionString << "]" << flush;
      }
    }

    void DataClientData::handleExternalMessage(void){
      //!<Fetch message from dataclient
      zmsg clientMsg(*_dataClientSock);

      std::string addr((clientMsg.address())?clientMsg.address():"");

      logger.log(NODE_LOG_MODE_TRACE) << "DataClientData::handleExternalMessage  addr: " << addr << flush;

      //!<Process message
      if(clientMsg.parts() == NODE_MSG_CMD_MIN_FIELDS){
        //!<Control message
        if(addr.compare(0, NODE_MSG_HEARTBEAT.length(), NODE_MSG_HEARTBEAT)==0){
          logger.log(NODE_LOG_MODE_DEBUG) << "HEARTBEAT message from server" << flush;

          //!<Register heartbit time mark
          _heartbeatedAt = s_clock();

          if (getHeartbeatMode() == HEARBEAT_MODE_IMMEDIATE_AFTER_HB ||
              getHeartbeatMode() == HEARBEAT_MODE_MIXED){
            //!<Execute sending heartbeat
            sendHeartbeat();

            logger.log(NODE_LOG_MODE_TRACE) << "HEARBEAT_MODE_IMMEDIATE_AFTER_HB || HEARBEAT_MODE_MIXED" << flush;
          }

        }else{
          logger.log(NODE_LOG_MODE_TRACE) << "ERROR control message" << endl << clientMsg.dump() << flush;
        }
      }else{
        //!<Data message
        logger.log(NODE_LOG_MODE_TRACE) << "MSG FROM dataclient" << endl << clientMsg.dump() << flush;

        //!<Create identity of dataclient's message
        std::string clientIdentity((const char*)clientMsg.pop_front().c_str());
        std::string messageId((const char*)clientMsg.pop_front().c_str());
        std::string messageBody((const char*)clientMsg.pop_front().c_str());
        std::string messageRoute((clientMsg.parts())?(const char*)clientMsg.pop_front().c_str():"");

        logger.log(NODE_LOG_MODE_TRACE) << "DataClientData::handleExternalMessage  Recieved body: \"" << messageBody << "\"" << flush;
        logger.log(NODE_LOG_MODE_TRACE) << "DataClientData::handleExternalMessage  Recieved route: \"" << messageRoute << "\"" << flush;

        _statRequests++;
        _statRequestsSize += messageBody.size();

        logger.log(NODE_LOG_MODE_DEBUG) << "RECEIVED REQUEST from upper server [" << clientIdentity << "], message [" << messageId << "] body " << messageBody.size()
            << " bytes n=" << _statRequests << flush;

        //!<Create response message and set message Id
        zmsg msg(messageId.c_str());
        //!<Set message body
        msg.append(messageBody.c_str());
        //!<Set message route
        msg.append(messageRoute.c_str());

        if(_dataClientSockMode == 0){
          //!<Send response for dataclient from end-point node
          msg.send(*_dataClientSock);
          _statResponses++;
          _statResponsesSize += _statRequestsSize;

          logger.log(NODE_LOG_MODE_TRACE) << " MSG TO dataclient" << endl << msg.dump() << flush;
          logger.log(NODE_LOG_MODE_DEBUG) << " RESPONSE to dataclient socket from end-point node to upper server [" << clientIdentity << "] sent, n=" << _statResponses << flush;

        }else{
          //!<Send response for data handler of end-point node
          msg.send(*_inprocDataSock);
          logger.log(NODE_LOG_MODE_DEBUG) << "REQUEST SENT to inproc_data socket" << flush;
        }

        //!<Register heartbeated time mark
        _heartbeatedAt = s_clock();

        if (getHeartbeatMode() == HEARBEAT_MODE_ADAPTIVE){

          //!<Reset timestamp for shift counter
          _shiftCounter.reset();
          //!<Move heartbeat time forward
          _heartbeatAt = s_clock() + getHeartbeatDelay();
        }
      }
    }

    void DataClientData::handleInternalMessage(void){
      //!<Get message from inproc_data
      zmsg msg(*_inprocDataSock);

      logger.log(NODE_LOG_MODE_TRACE) << "MSG FROM inproc_data" << endl << msg.dump() << flush;

      //!<Re-send response message directly, not using the responses map
      msg.send(*_dataClientSock);

      _statResponses++;
      _statResponsesSize += msg.size();

      logger.log(NODE_LOG_MODE_DEBUG) << "RESPONSE to dataclient socket from data node to upper server sent, n=" << _statResponses << flush;

      //!<Move heartbit time forward
      _heartbeatAt = s_clock() + getHeartbeatDelay();
    }

    void DataClientData::heartbit(void){
      if (getHeartbeatMode() == HEARBEAT_MODE_FIXED_TIME ||
          getHeartbeatMode() == HEARBEAT_MODE_MIXED){
        if((s_clock() > _heartbeatAt) && _dataClientSockConnected){

          //!<Execute sending heartbeat
          sendHeartbeat();

          logger.log(NODE_LOG_MODE_TRACE) << "HEARBEAT_MODE_FIXED_TIME || HEARBEAT_MODE_MIXED" << flush;
          logger.log(NODE_LOG_MODE_DEBUG) << "SENT HEARTBIT from client [" << _clientIdentity << "] to server" << flush;
        }
      }else if(getHeartbeatMode() == HEARBEAT_MODE_ADAPTIVE){
        _shiftCounter.responceTimeStamp(_heartbeatedAt);
        _shiftCounter.recount(getHeartbeatTimeout());

        logger.log(NODE_LOG_MODE_TRACE) << "DataClientData::heartbit  HEARBEAT_MODE_ADAPTIVE  delta = " << (s_clock()- _heartbeatAt+getHeartbeatDelay())
            << " _shiftCounter.getShift()= " << _shiftCounter.getShift() << flush;

        size_t delta = s_clock()-_heartbeatAt+getHeartbeatDelay();
        if (delta > (getHeartbeatTimeout()+_shiftCounter.getShift()) && _dataClientSockConnected){
          //!<Execute sending heartbeat
          sendHeartbeat();
          //!<Reset shift counter
          _shiftCounter.reset();

          logger.log(NODE_LOG_MODE_DEBUG) << "SENT HEARTBIT from client [" << _clientIdentity << "] to server _heartbeatAt = '"
              << Poco::DateTimeFormatter::format(Poco::LocalDateTime(Poco::Timestamp(Poco::Timestamp::fromEpochTime(_heartbeatAt/1000))), "%Y-%m-%d %H:%M:%S") << "'" << flush;
        }
        //!<Register heartbeated time mark
        _heartbeatedAt = s_clock();
      }
    }

    void DataClientData::sendHeartbeat(void){
      //!<Send heartbeat with resource usage message
//      zmsg msg(NODE_MSG_HEARTBEAT.c_str());
//      msg.append(getResourceUsageMessage().c_str());
//      msg.send(*_dataClientSock);

      std::vector<ClientWorkerItem> queue;
      HCE::handlers::HeartbeatMessage heartbeatMessage(queue, getResourceUsageMessage());
      std::string message = heartbeatMessage.build(NODE_MSG_HEARTBEAT);
      s_send(*_dataClientSock, message);

      logger.log(NODE_LOG_MODE_TRACE) << "SENT HEARTBIT MESSAGE: " << message << flush;

      //!<Reset heartbit time
      _heartbeatAt = s_clock() + getHeartbeatDelay();
    }

    std::string DataClientData::getResourceUsageMessage(void) const{
      //!<Get resource usage data
      HCE::drce::ResourceUsageExtractor::getInstance().update();
      HCE::drce::ResourceUsage resourceUsage = HCE::drce::ResourceUsageExtractor::getInstance().extract();
      HCE::drce::ResourceUsageConverter resourceUsageConverter(resourceUsage);
      return resourceUsageConverter.toString();
    }

    void DataClientData::logPeriodicStatistics(void){
      if(s_clock() > _pstatsAt){
        _statQPS = (_statRequests - _statRequestsProcessed) / (float)((STAT_INTERVAL) / 1000.0);
        _statRPS = ((_statResponses - _statResponsesProcessed) / (float)((STAT_INTERVAL) / 1000.0));
        _statRequestsBandwidth = (_statRequestsSize - _statRequestsProcessedSize) / (float)((STAT_INTERVAL) / 1000.0);
        _statResponsesBandwidth = (_statResponsesSize - _statResponsesProcessedSize) / (float)((STAT_INTERVAL) / 1000.0);

        _statRequestsProcessed = _statRequests;
        _statResponsesProcessed = _statResponses;
        _statRequestsProcessedSize = _statRequestsSize;
        _statResponsesProcessedSize = _statResponsesSize;

        logger.log(NODE_LOG_MODE_DEBUG) << "requests=" << _statRequests << ", responses=" << _statResponses << ", qps=" << _statQPS << ", rps=" << _statRPS
            << ", requestsSize=" << _statRequestsSize << ", responsesSize=" << _statResponsesSize << ", requestsBandwidth="
            << _statRequestsBandwidth << ", responsesBandwidth=" << _statResponsesBandwidth << flush;

        //!<Reset pstat time
        _pstatsAt = s_clock() + STAT_INTERVAL;
      }
    }

    std::string DataClientData::getPropertyInformation(void){
      std::stringstream result;
      result << NODE_ADMIN_ERROR_OK
          << NODE_ADMIN_COMMAND_DELIMITER << getProperties() << "requests=" << _statRequests
          << NODE_ADMIN_COMMAND_DELIMITER << "responses=" << _statResponses
          << NODE_ADMIN_COMMAND_DELIMITER << "requestsSize=" << _statRequestsSize
          << NODE_ADMIN_COMMAND_DELIMITER << "responsesSize=" << _statResponsesSize
          << NODE_ADMIN_COMMAND_DELIMITER << "clientIdentity=" << _clientIdentity
          << NODE_ADMIN_COMMAND_DELIMITER << "dataClientMode=" << static_cast<int>(_dataClientSockMode)
          << NODE_ADMIN_COMMAND_DELIMITER << "pstatsAt=" << static_cast<uint64_t>(_pstatsAt)
          << NODE_ADMIN_COMMAND_DELIMITER << "heartbeatAt=" << static_cast<uint64_t>(_heartbeatAt)
          << NODE_ADMIN_COMMAND_DELIMITER << "heartbeatedAt=" << static_cast<uint64_t>(_heartbeatedAt)
          << NODE_ADMIN_COMMAND_DELIMITER << "statQPS=" << _statQPS
          << NODE_ADMIN_COMMAND_DELIMITER << "statRPS=" << _statRPS
          << NODE_ADMIN_COMMAND_DELIMITER << "statRequestsBandwidth=" << _statRequestsBandwidth
          << NODE_ADMIN_COMMAND_DELIMITER << "statResponsesBandwidth=" << _statResponsesBandwidth;
      return result.str();
    }

    void DataClientData::externalConnectionRefresh(void){

      if (getHeartbeatMode() != HEARBEAT_MODE_ADAPTIVE){

        if(((s_clock() - _heartbeatedAt) > getHeartbeatTimeout()) && getInProgress() && _dataClientSockConnected){
          logger.log(NODE_LOG_MODE_INFORMATION) << "RECONNECT client [" << _clientIdentity << "] because HEARTBIT not received during time of "
              << (getHeartbeatTimeout()+_shiftCounter.getShift()) << " msec" << flush;

          //!<Rebuild data socket
          rebuildDataClientSocket();

          //!<Reset time mark
          _heartbeatedAt = s_clock();

          //!<Move heartbit time forward
          _heartbeatAt = s_clock() + getHeartbeatDelay();
        }
      }else{ // HEARBEAT_MODE_ADAPTIVE

        _shiftCounter.recount(getHeartbeatTimeout());

        logger.log(NODE_LOG_MODE_TRACE) << "DataClientData::externalConnectionRefresh  (s_clock()-_heartbeatedAt)= " << (s_clock()-_heartbeatedAt)
            << " getHeartbeatTimeout()= " << getHeartbeatTimeout() << " _shiftCounter.getShift()= " << _shiftCounter.getShift() << flush;

        if(((s_clock() - _heartbeatedAt) > (getHeartbeatTimeout()+_shiftCounter.getShift())) && getInProgress() && _dataClientSockConnected){
          logger.log(NODE_LOG_MODE_INFORMATION) << "RECONNECT client [" << _clientIdentity << "] because HEARTBIT not received during time of "
              << (getHeartbeatTimeout()+_shiftCounter.getShift()) << " msec" << flush;

          //!<Rebuild data socket
          rebuildDataClientSocket();

          //!<Reset time mark
          _heartbeatedAt = s_clock();

          _shiftCounter.reset();

          //!<Move heartbit time forward
          _heartbeatAt = s_clock() + getHeartbeatDelay();
        }
      }
    }

    std::string DataClientData::processAdminCommands(std::string& command, std::string& parameters){
      std::string ret = NODE_ADMIN_ERROR_OK;

      logger.log(NODE_LOG_MODE_DEBUG) << "Admin command requested [" << command << "], parameters [" << parameters << "]" << flush;

      //!<Process commands
      if(command == NODE_MSG_STAT){
        std::stringstream result;
        result << NODE_ADMIN_ERROR_OK << NODE_ADMIN_COMMAND_DELIMITER << "requests=" << _statRequests << ",responses=" << _statResponses << ",requests_size=" << _statRequestsSize
            << ",responses_size=" << _statResponsesSize;
        ret = result.str();
      }else if(command == NODE_MSG_REBUILD_CLIENT_CONNECTION){
        //!<Rebuild client connection
        rebuildDataClientSocket();
      }else if(command == NODE_MSG_DISCONNECT_CLIENT_CONNECTION){
        //!<Disconnect client connection
        _dataClientSock->close();
        _dataClientSockConnected = false;
      }else if(command == NODE_MSG_UPDATE_SCHEMA){
        setConnectionString(parameters);
      }else if(command == NODE_MSG_PROPERTIES){
        //!<Make property information
        ret = getPropertyInformation();
      }else if(command == NODE_MSG_HEARTBEAT_DELAY_GET){
        ret = NODE_ADMIN_ERROR_OK + NODE_ADMIN_COMMAND_DELIMITER + NODE_ADMIN_PROPERTY_HEARTBEAT_DELAY + "=" + std::to_string(getHeartbeatDelay()/1000);
      }else if(command == NODE_MSG_HEARTBEAT_DELAY_SET){
        //!<Parse command parameters
        unsigned int heartbeatDelay = HEARTBEAT_DELAY;
        //!<Parse heartbeat delay from parameters
        std::istringstream(parameters) >> heartbeatDelay;
        ret = NODE_ADMIN_ERROR_OK + NODE_ADMIN_COMMAND_DELIMITER + NODE_ADMIN_PROPERTY_HEARTBEAT_DELAY + "=" + std::to_string(getHeartbeatDelay()/1000);
        //!<Set new heartbeat delay value
        setHeartbeatDelay(heartbeatDelay*1000);
      }else if(command == NODE_MSG_HEARTBEAT_TIMEOUT_GET){
        ret = NODE_ADMIN_ERROR_OK + NODE_ADMIN_COMMAND_DELIMITER + NODE_ADMIN_PROPERTY_HEARTBEAT_TIMEOUT + "=" + std::to_string(getHeartbeatTimeout()/1000);
      }else if(command == NODE_MSG_HEARTBEAT_TIMEOUT_SET){
        //!<Parse command parameters
        unsigned int heartbeatTimeout = HEARTBEAT_TIMEOUT;
        //!<Parse heartbeat timeout from parameters
        std::istringstream(parameters) >> heartbeatTimeout;
        ret = NODE_ADMIN_ERROR_OK + NODE_ADMIN_COMMAND_DELIMITER + NODE_ADMIN_PROPERTY_HEARTBEAT_TIMEOUT + "=" + std::to_string(getHeartbeatTimeout()/1000);
        //!<Set new heartbeat timeout value
        setHeartbeatTimeout(heartbeatTimeout*1000);
      }else if(command == NODE_MSG_HEARTBEAT_MODE_GET){
        ret = NODE_ADMIN_ERROR_OK + NODE_ADMIN_COMMAND_DELIMITER + NODE_ADMIN_PROPERTY_HEARTBEAT_MODE + "=" + std::to_string(getHeartbeatMode());
      }else if(command == NODE_MSG_HEARTBEAT_MODE_SET){
        //!<Parse command parameters
        int hbMode = HEARBEAT_MODE_DEFAULT;
        //!<Parse heartbeat mode from parameters
        std::istringstream(parameters) >> hbMode;
        ret = NODE_ADMIN_ERROR_OK + NODE_ADMIN_COMMAND_DELIMITER + NODE_ADMIN_PROPERTY_HEARTBEAT_MODE + "=" + std::to_string(getHeartbeatMode());
        //!<Set new heartbeat mode value
        setHeartbeatMode(static_cast<unsigned int>(hbMode));
      }else{
        //!<Unsupported command
        ret = NODE_ADMIN_ERROR_ERROR + NODE_ADMIN_COMMAND_DELIMITER + NODE_ADMIN_ERROR_UNSUPPORTED_COMMAND;
      }

      return ret;
    }

    void* DataClientData::process(void){
      logger.log(NODE_LOG_MODE_INFORMATION) << NODE_ADMIN_ERROR_MAIN_MSG_LOOP_STARTED << flush;

      long delay = getPollTimeout();

      while(getInProgress()){
        try{
          //!<Init polling pool
          std::vector<zmq::pollitem_t> items;
          if(_dataClientSockConnected){
            items ={{*_dataClientSock, 0, ZMQ_POLLIN, 0},{*_inprocDataSock, 0, ZMQ_POLLIN, 0},{*_inprocAdminSock, 0, ZMQ_POLLIN, 0}};
          } else{
            items ={{*_inprocDataSock, 0, ZMQ_POLLIN, 0},{*_inprocAdminSock, 0, ZMQ_POLLIN, 0}};
          }

          //!<Reinit delay if necessary
          if (delay){
            delay = getPollTimeout();
          }

          //!<Poll dataclient
          zmq::poll(&items[0], items.size(), delay);

          //!<Handle dataclient socket's message
          if(_dataClientSockConnected){
            if(items[0].revents & ZMQ_POLLIN){
              handleExternalMessage();
            }
          }

          //!<Handle inproc_data socket message
          if(items[items.size() - 2].revents & ZMQ_POLLIN){
            handleInternalMessage();
          }

          //!<Handle inproc_admin's socket message
          if(items[items.size() - 1].revents & ZMQ_POLLIN){
            handleAdminMessage();
            if(getInProgress() == false){
              delay = 0;
              setInProgress(true);
              continue;
            }
          }

          if(getInProgress() == true && delay == 0){
            setInProgress(false);
            continue;
          }

          //!<Send heartbeat to server if it's time
          heartbit();

          //!<Out periodical stats
          logPeriodicStatistics();

          //!<Send periodical property values
          sendPeriodicProperties();

          //!<Heartbited interval check and connection refresh
          externalConnectionRefresh();

        }catch(zmq::error_t& error){
          std::cout << "DataClientData E: " << error.what() << std::endl;
        }catch(std::exception& e){
          std::cout << "DataClientData E: " << e.what() << std::endl;
        }catch(...){
          std::cout << "DataClientData E: Unknown exception" << std::endl;
        }
      }

      logger.log(NODE_LOG_MODE_INFORMATION) << NODE_ADMIN_ERROR_MAIN_MSG_LOOP_STOPPED << flush;

      return NULL;
    }
  }
}
