#include "DataClientProxy.hpp"
#include "ServiceMessages.hpp"
#include "DRCEResourceUsage.hpp"
#include "DRCEResourceExtractor.hpp"

namespace HCE {
  namespace handlers {

    DataClientProxy::DataClientProxy(std::string name, zmq::context_t& ctx, const std::string& clientIdentity, std::string connectionString, int64_t startedAt,
        unsigned char logPriority, const std::string& iniFile) :
        Handler(name, ctx, connectionString, startedAt, logPriority, iniFile), DataNodeOptions(),
        _clientIdentity(clientIdentity), _iniFile(iniFile), _shiftCounter(), _clientsQueue(){

      //!<Init all the rest
      initialize();
    }

    DataClientProxy::~DataClientProxy(){
      deinitialize();
    }

    int DataClientProxy::initialize(void){
      logger.log(NODE_LOG_MODE_INFORMATION) << "Initialize client [" << _clientIdentity << "] with connection [" << _connectionString << "]" << flush;

      //!<Init sockets
      _inprocProxySock = NULL;
      _inprocReducerOutSock = NULL;
      _dataClientSock = NULL;
      _dataClientSockConnected = false;

      //!<Node init from file
      initNode(_iniFile, logger);

      //!<Heartbeat at time mark
      _heartbeatAt = s_clock() + getHeartbeatDelay();
      //!<Heartbeated time mark
      _heartbeatedAt = s_clock();

      //!<Stat vars
      _statRequests = 0;
      _statResponses = 0;

      //!<Rebuild admin client socket
      rebuildAdminClientSocket();

      //!<Rebuild inproc_proxy connection
      rebuildClientConnection(_inprocProxySock, NODE_CS_INPROC_PROXY, ZMQ_DEALER, _clientIdentity);

      //!<Rebuild inproc_reducer_out connection
      rebuildClientConnection(_inprocReducerOutSock, NODE_CS_INPROC_REDUCER_OUT, ZMQ_DEALER);

      //!<Create dataclient socket, connect to upper node and send READY message
      rebuildDataClientSocket();

      return true;
    }

    void DataClientProxy::deinitialize(void){
      if(_inprocProxySock){
        if(_inprocProxySock->connected()){
          _inprocProxySock->close();
        }
        delete _inprocProxySock;
        _inprocProxySock = NULL;
      }

      if(_inprocReducerOutSock){
        if(_inprocReducerOutSock->connected()){
          _inprocReducerOutSock->close();
        }
        delete _inprocReducerOutSock;
        _inprocReducerOutSock = NULL;
      }

      if(_dataClientSock){
        if(_dataClientSock->connected()){
          sendByeMessage(_dataClientSock, _clientIdentity);
          _dataClientSockConnected = false;
          _dataClientSock->close();
        }
        delete _dataClientSock;
        _dataClientSock = NULL;
      }

      HCE::drce::ResourceUsageExtractor::shutdown();
    }

    void DataClientProxy::rebuildDataClientSocket(void){
      _dataClientSockConnected = rebuildClientConnection(_dataClientSock, _connectionString, ZMQ_DEALER, _clientIdentity, 0, NODE_MSG_READY);
    }

    void DataClientProxy::processControlMessage(zmsg& msg){

      std::string addr((msg.address())?msg.address():"");
      if(addr.compare(0, NODE_MSG_HEARTBEAT.length(), NODE_MSG_HEARTBEAT)==0){
        logger.log(NODE_LOG_MODE_DEBUG) << "RECEIVED HEARTBEAT from server" << flush;

        //!<Register heartbit time mark
        _heartbeatedAt = s_clock();

        if (getHeartbeatMode() == HEARBEAT_MODE_IMMEDIATE_AFTER_HB ||
            getHeartbeatMode() == HEARBEAT_MODE_MIXED){
          //!<Execute sending heartbeat
          sendHeartbeat();
          logger.log(NODE_LOG_MODE_TRACE) << "HEARBEAT_MODE_IMMEDIATE_AFTER_HB || HEARBEAT_MODE_MIXED" << flush;
        }
      }else{
        logger.log(NODE_LOG_MODE_TRACE) << "ERROR: wrong control message received" << endl << msg.dump() << flush;
      }
    }

    void DataClientProxy::processDataMessage(zmsg& msg){
      _statRequests++;

      logger.log(NODE_LOG_MODE_TRACE) << "MSG FROM dataclient" << endl << msg.dump() << flush;
      logger.log(NODE_LOG_MODE_DEBUG) << "RECEIVED REQUEST from upper server, n=" << _statRequests << flush;

      //!<Proxy message to inproc handler
      msg.send(*_inprocProxySock);

      logger.log(NODE_LOG_MODE_DEBUG) << "REQUEST SENT to inproc_proxy socket" << flush;

      if (getHeartbeatMode() == HEARBEAT_MODE_ADAPTIVE){
        //!<Register heartbeated time mark
        _heartbeatedAt = s_clock();
        //!<Execute sending heartbeat
        _shiftCounter.reset();
        //!<Move heartbeat time forward
        _heartbeatAt = s_clock() + getHeartbeatDelay();
      }
    }

    void DataClientProxy::handleExternalMessage(void){
      //!<Fetch message from dataclient
      zmsg clientMsg(*_dataClientSock);

      if(clientMsg.parts() == NODE_MSG_CMD_MIN_FIELDS){
        //!<Control message
        processControlMessage(clientMsg);
      }else{
        //!<Data message
        processDataMessage(clientMsg);
      }

      //!<Register heartbit time mark
      _heartbeatedAt = s_clock();
    }

    void DataClientProxy::handleInternalMessage(void){
      //!<Fetch message from inproc_reducer_out
      zmsg msg(*_inprocReducerOutSock);

      if (msg.parts() == NODE_MSG_CMD_MIN_FIELDS)
      {
        char* addrp = msg.address();
        std::string addr ((addrp)?addrp:"");
        if (addr.compare(0, NODE_MSG_ROUTES.length(), NODE_MSG_ROUTES)==0)
        {
          s_send(*_dataClientSock, addr);  // resend message to server
        }
      }
      else
      {
        _statResponses++;

        logger.log(NODE_LOG_MODE_TRACE) << "MSG FROM inproc_reducer_out" << endl << msg.dump() << flush;
        logger.log(NODE_LOG_MODE_DEBUG) << "RESPONSE SENT to dataclient socket from proxy node to upper server, n=" << _statResponses << flush;

        //!<Send response message to server
        msg.send(*_dataClientSock);
      }
    }

    void DataClientProxy::handleRouteMessage(void){

      //!<Fetch message from inproc_proxy
      zmsg msg(*_inprocProxySock);
      logger.log(NODE_LOG_MODE_TRACE) << " MSG FROM inproc_proxy" << endl << msg.dump() << flush;

      std::string addr((msg.address())?msg.address():"");

      if (addr.compare(0, NODE_MSG_ROUTES.length(), NODE_MSG_ROUTES)==0)
      {
        RouteMessage routeMessage(_clientsQueue);
        routeMessage.extract(NODE_MSG_ROUTES, addr);
        if (routeMessage.isError())
          logger.log(NODE_LOG_MODE_ERROR) << "Operation extract ROUTE MESSAGE has error: " << routeMessage.getErrorMsg() << flush;
        else
        {
          logger.log(NODE_LOG_MODE_TRACE) << "Operation extract ROUTE MESSAGE was made SUCCESSFULLY" << flush;
        }
      }
    }

    void DataClientProxy::heartbit(void){
      if (getHeartbeatMode() == HEARBEAT_MODE_FIXED_TIME ||
          getHeartbeatMode() == HEARBEAT_MODE_MIXED){
        if(s_clock() > _heartbeatAt && _dataClientSockConnected){
          sendHeartbeat();

          logger.log(NODE_LOG_MODE_DEBUG) << "SENT HEARTBIT from client [" << _clientIdentity << "] to server" << flush;
        }
      }else if (getHeartbeatMode() == HEARBEAT_MODE_ADAPTIVE){
        _shiftCounter.responceTimeStamp(_heartbeatedAt);
        _shiftCounter.recount(getHeartbeatTimeout());

        logger.log(NODE_LOG_MODE_TRACE) << "DataClientProxy::heartbit  HEARBEAT_MODE_ADAPTIVE  delta = " << (s_clock()-_heartbeatedAt+getHeartbeatDelay())
            << " _shiftCounter.getShift()= " << _shiftCounter.getShift() << flush;

        size_t delta = s_clock()-_heartbeatAt+getHeartbeatDelay();
        if (delta > (getHeartbeatTimeout()+_shiftCounter.getShift()) && _dataClientSockConnected){
          
          sendHeartbeat();
          _shiftCounter.reset();

          logger.log(NODE_LOG_MODE_DEBUG) << "SENT HEARTBIT from client [" << _clientIdentity << "] to server _heartbeatAt = '"
              << Poco::DateTimeFormatter::format(Poco::LocalDateTime(Poco::Timestamp(Poco::Timestamp::fromEpochTime(_heartbeatAt/1000))), "%Y-%m-%d %H:%M:%S") << "'" << flush;
        }
        //!<Register heartbeated time mark
        _heartbeatedAt = s_clock();
      }
    }

    void DataClientProxy::sendHeartbeat(void){
      //!<Get resource usage data
      HCE::drce::ResourceUsageExtractor::getInstance().update();
      HCE::drce::ResourceUsage resourceUsage = HCE::drce::ResourceUsageExtractor::getInstance().extract();
      HCE::drce::ResourceUsageConverter resourceUsageConverter(resourceUsage);
      //!<Build heartbeat message
      HCE::handlers::HeartbeatMessage heartbeatMessage(_clientsQueue, resourceUsageConverter.toString());
      std::string message = heartbeatMessage.build(NODE_MSG_HEARTBEAT);

      logger.log(NODE_LOG_MODE_TRACE) << "SENT HEARTBIT MESSAGE: " << message << flush;
      //!<Send heartbeat message
      s_send(*_dataClientSock, message);

      //!<Reset heartbit time
      _heartbeatAt = s_clock() + getHeartbeatDelay();
    }

    void DataClientProxy::externalConnectionRefresh(void){

      if (getHeartbeatMode() != HEARBEAT_MODE_ADAPTIVE){

        if(((s_clock() - _heartbeatedAt) > getHeartbeatTimeout()) && getInProgress() && _dataClientSockConnected){
          logger.log(NODE_LOG_MODE_INFORMATION) << "RECONNECT client [" << _clientIdentity << "] because HEARTBIT not received during time of "
              << (getHeartbeatTimeout()+_shiftCounter.getShift()) << " msec" << flush;

          //!<Rebuild data socket
          rebuildDataClientSocket();

          //Reset time mark
          _heartbeatedAt = s_clock();

          //!<Move heartbit time forward
          _heartbeatAt = s_clock() + getHeartbeatDelay();
        }
      }else{ // HEARBEAT_MODE_ADAPTIVE

        _shiftCounter.recount(getHeartbeatTimeout());

        logger.log(NODE_LOG_MODE_TRACE) << "DataClientProxy::externalConnectionRefresh  (s_clock()-_heartbeatedAt)= " << (s_clock()-_heartbeatedAt) <<
            " getHeartbeatTimeout()= " << getHeartbeatTimeout() << " _shiftCounter.getShift()= " << _shiftCounter.getShift() << flush;

        if(((s_clock() - _heartbeatedAt) > (getHeartbeatTimeout()+_shiftCounter.getShift())) && getInProgress() && _dataClientSockConnected){
          logger.log(NODE_LOG_MODE_INFORMATION) << "RECONNECT client [" << _clientIdentity << "] because HEARTBIT not received during time of "
              << (getHeartbeatTimeout()+_shiftCounter.getShift()) << " msec" << flush;

          //!<Rebuild data socket
          rebuildDataClientSocket();

          //Reset time mark
          _heartbeatedAt = s_clock();

          //!<Move heartbit time forward
          _heartbeatAt = s_clock() + getHeartbeatDelay();
        }
      }
    }

    //!<Get property information
    std::string DataClientProxy::getPropertyInformation(void){
      std::stringstream result;
      result << NODE_ADMIN_ERROR_OK
          << NODE_ADMIN_COMMAND_DELIMITER << getProperties() << "requests=" << _statRequests
          << NODE_ADMIN_COMMAND_DELIMITER << "responses=" << _statResponses
          << NODE_ADMIN_COMMAND_DELIMITER << "clientIdentity=" << _clientIdentity
          << NODE_ADMIN_COMMAND_DELIMITER << "heartbeatAt=" << static_cast<uint64_t>(_heartbeatAt)
          << NODE_ADMIN_COMMAND_DELIMITER << "heartbeatedAt=" << static_cast<uint64_t>(_heartbeatedAt);
      return result.str();
    }

    std::string DataClientProxy::processAdminCommands(std::string& command, std::string& parameters){
      std::string ret = NODE_ADMIN_ERROR_OK;

      logger.log(NODE_LOG_MODE_DEBUG) << "Admin command requested [" << command << "], parameters [" << parameters << "]" << flush;

      //!<Process commands
      if(command == NODE_MSG_STAT){
        std::stringstream result;
        result << NODE_ADMIN_ERROR_OK << NODE_ADMIN_COMMAND_DELIMITER << "requests=" << _statRequests << ",responses=" << _statResponses;
        ret = result.str();
      }else if(command == NODE_MSG_REBUILD_CLIENT_CONNECTION){
        //!<Rebuild client connection
        rebuildDataClientSocket();
      }else if(command == NODE_MSG_DISCONNECT_CLIENT_CONNECTION){
        //!<Disconnect client connection
        _dataClientSock->close();
        _dataClientSockConnected = false;
      }else if(command == NODE_MSG_UPDATE_SCHEMA){
        setConnectionString(parameters);
      }else if(command == NODE_MSG_PROPERTIES){
        //!<Make property information
        ret = getPropertyInformation();
      }else if(command == NODE_MSG_HEARTBEAT_DELAY_GET){
        ret = NODE_ADMIN_ERROR_OK + NODE_ADMIN_COMMAND_DELIMITER + NODE_ADMIN_PROPERTY_HEARTBEAT_DELAY + "=" + std::to_string(getHeartbeatDelay()/1000);
      }else if(command == NODE_MSG_HEARTBEAT_DELAY_SET){
        //!<Parse command parameters
        unsigned int heartbeatDelay = HEARTBEAT_DELAY;
        //!<Parse log priority from parameters
        std::istringstream(parameters) >> heartbeatDelay;
        ret = NODE_ADMIN_ERROR_OK + NODE_ADMIN_COMMAND_DELIMITER + NODE_ADMIN_PROPERTY_HEARTBEAT_DELAY + "=" + std::to_string(getHeartbeatDelay()/1000);
        //!<Set new heartbeat delay value
        setHeartbeatDelay(heartbeatDelay*1000);
      }else if(command == NODE_MSG_HEARTBEAT_TIMEOUT_GET){
        ret = NODE_ADMIN_ERROR_OK + NODE_ADMIN_COMMAND_DELIMITER + NODE_ADMIN_PROPERTY_HEARTBEAT_TIMEOUT + "=" + std::to_string(getHeartbeatTimeout()/1000);
      }else if(command == NODE_MSG_HEARTBEAT_TIMEOUT_SET){
        //!<Parse command parameters
        int hbTimeout = HEARTBEAT_TIMEOUT;
        //!<Parse heartbeat timeout from parameters
        std::istringstream(parameters) >> hbTimeout;
        ret = NODE_ADMIN_ERROR_OK + NODE_ADMIN_COMMAND_DELIMITER + NODE_ADMIN_PROPERTY_HEARTBEAT_TIMEOUT + "=" + std::to_string(getHeartbeatTimeout()/1000);
        //!<Set new heartbeat timeout value
        setHeartbeatTimeout(static_cast<unsigned int>(hbTimeout*1000));
      }else if(command == NODE_MSG_HEARTBEAT_MODE_GET){
        ret = NODE_ADMIN_ERROR_OK + NODE_ADMIN_COMMAND_DELIMITER + NODE_ADMIN_PROPERTY_HEARTBEAT_MODE + "=" + std::to_string(getHeartbeatMode());
      }else if(command == NODE_MSG_HEARTBEAT_MODE_SET){
        //!<Parse command parameters
        int hbMode = HEARBEAT_MODE_DEFAULT;
        //!<Parse heartbeat mode from parameters
        std::istringstream(parameters) >> hbMode;
        ret = NODE_ADMIN_ERROR_OK + NODE_ADMIN_COMMAND_DELIMITER + NODE_ADMIN_PROPERTY_HEARTBEAT_MODE + "=" + std::to_string(getHeartbeatMode());
        //!<Set new heartbeat mode value
        setHeartbeatMode(static_cast<unsigned int>(hbMode));
      }else{
        //!<Unsupported command
        ret = NODE_ADMIN_ERROR_ERROR + NODE_ADMIN_COMMAND_DELIMITER + NODE_ADMIN_ERROR_UNSUPPORTED_COMMAND;
      }

      return ret;
    }

    void* DataClientProxy::process(void){
      logger.log(NODE_LOG_MODE_DEBUG) << NODE_ADMIN_ERROR_MAIN_MSG_LOOP_STARTED << flush;

      while(getInProgress()){
        try {
          //!<Init polling pool
          std::vector<zmq::pollitem_t> items;
          if(_dataClientSockConnected){
            items ={{*_dataClientSock, 0, ZMQ_POLLIN, 0},{*_inprocReducerOutSock, 0, ZMQ_POLLIN, 0},{*_inprocAdminSock, 0, ZMQ_POLLIN, 0},{*_inprocProxySock, 0, ZMQ_POLLIN, 0}}; //add new socket
          } else{
            items ={{*_inprocReducerOutSock, 0, ZMQ_POLLIN, 0},{*_inprocAdminSock, 0, ZMQ_POLLIN, 0},{*_inprocProxySock, 0, ZMQ_POLLIN, 0}}; //add new socket inproc_proxy
          }

          //!<Poll dataclient sockets
          zmq::poll(&items[0], items.size(), getPollTimeout());

          if(_dataClientSockConnected){
            //!<Handle dataclient socket's message
            if(items[0].revents & ZMQ_POLLIN){
              handleExternalMessage();
            }
          }

          //!<Handle inproc_reducer_out socket message
          if(items[items.size() - 3].revents & ZMQ_POLLIN){
            handleInternalMessage();
          }

          //!<Handle inproc_admin's socket message
          if(items[items.size() - 2].revents & ZMQ_POLLIN){
            handleAdminMessage();
          }

          //!<Handle inproc_proxy's socket message
          if(items[items.size() - 1].revents & ZMQ_POLLIN){
            handleRouteMessage();
          }

          //!<Send heartbeat to server
          heartbit();

          //!<Send periodical property values
          sendPeriodicProperties();

          //!<Heartbited interval check and connection refresh
          externalConnectionRefresh();
        }catch(zmq::error_t& error){
          std::cout << "DataClientProxy E: " << error.what() << std::endl;
        }catch(std::exception& e){
          std::cout << "DataClientProxy E: " << e.what() << std::endl;
        }catch(...){
          std::cout << "DataClientProxy E: Unknown exception" << std::endl;
        }
      }
      logger.log(NODE_LOG_MODE_INFORMATION) << NODE_ADMIN_ERROR_MAIN_MSG_LOOP_STOPPED << flush;

      return NULL;
    }

  }
}
