#include "DataProcessorData.hpp"

namespace HCE {
  namespace handlers {
    DataProcessorData::DataProcessorData(const std::string& nodeName, std::string name, zmq::context_t& ctx, int64_t startedAt, unsigned char logPriority, const std::string& iniFile) :
        Handler(name, ctx, std::string(""), startedAt, logPriority, iniFile), _nodeName(nodeName), _inprocDataSock(NULL), _tasksStateNotificationSock(NULL), _statClientResponses(0), _statSphinxRequestsSearch(
            0), _statSphinxRequestsAdmin(0), _statSphinxTimeTotalSearch(0), _statSphinxTimeTotalAdmin(0), _iniFile(iniFile), _sphinxFO(NULL), _dataProcessingMode(
            NODE_DATA_PROCESSING_MODE_NORMAL), _dataProcessingFakeResults(0), _dataProcessingFakeResultsMax(0), _nodeNumber(0),
            _state_notification_host(""), _state_notification_port(""), _tasksStateNotificationFunctor(NULL) {

      //!<Init all the rest
      initialize();
    }

    DataProcessorData::~DataProcessorData(){
      deinitialize();
    }

    int DataProcessorData::initialize(void){
      logger.log(NODE_LOG_MODE_INFORMATION) << "Initialize..." << flush;

      //!<Init sockets
      _inprocDataSock = NULL;
      _tasksStateNotificationSock = NULL;

      _statClientResponses = 0;
      _statSphinxTimeTotalSearch = 0;
      _statSphinxTimeTotalAdmin = 0;
      _statDRCERequests = 0;
      _statDRCETimeTotal = 0;

      _nodeNumber = 0;

      _state_notification_host = "";
      _state_notification_port = "";

      _tasksStateNotificationFunctor = NULL;

      //!<Node init
      nodeInit(_iniFile);

      //!<Sphinx FO init
      dataStorageInitSphinx(_iniFile);

      //!<DRCE FO init
      dataStorageInitDRCE(_iniFile);

      //!<Rebuild admin client socket
      rebuildAdminClientSocket();

      //!<Rebuild inproc_proxy connection
      rebuildServerConnection(_inprocDataSock, NODE_CS_INPROC_DATA, ZMQ_DEALER);

      return true;
    }

    void DataProcessorData::deinitialize(void){
      if(_inprocDataSock){
        if(_inprocDataSock->connected()){
          _inprocDataSock->close();
        }
        delete _inprocDataSock;
        _inprocDataSock = NULL;
      }

      if(_sphinxFO){
        delete _sphinxFO;
        _sphinxFO = NULL;
      }

      if(_tasksStateNotificationSock){
        delete _tasksStateNotificationSock;
        _tasksStateNotificationSock = NULL;
      }

      if(_tasksStateNotificationFunctor){
        delete _tasksStateNotificationFunctor;
        _tasksStateNotificationFunctor = NULL;
      }

      if(_drceFO){
        delete _drceFO;
        _drceFO = NULL;
      }
    }

    std::string DataProcessorData::getNodeName(void) const{
      return _nodeName;
    }

    bool DataProcessorData::rebuildTasksStateNotificationConnection(){
      bool ret = true;

      if(_state_notification_host != "" && _state_notification_port != ""){
        //Init the tasks state notification update connection
        ret = rebuildClientConnection(_tasksStateNotificationSock, ZMQ_PROTOCOL_PREFIX_TCP + _state_notification_host + ":" + _state_notification_port, ZMQ_DEALER, _nodeName, -1,
            std::string(""), std::string(""));
      }else{
        ret = false;
      }

      return ret;
    }

    void DataProcessorData::dataStorageInitSphinx(const std::string& iniFile){
      //!<Sphinx FO init parameters
      std::string sphinx_home_dir = "";
      std::string sphinx_default_index;
      bool sphinx_start_searchd = NODE_FUNCTIONAL_SPHINX_NOT_START_SEARCHD;
      bool sphinx_stop_searchd = NODE_FUNCTIONAL_SPHINX_NOT_STOP_SEARCHD;
      unsigned int sphinx_ranker = NODE_FUNCTIONAL_SPHINX_RANKER_0;
      std::string sphinx_ranker_expression = "";

      logger.log(NODE_LOG_MODE_INFORMATION) << "load Sphinx FO settings from " << iniFile << " file" << flush;

      //!<Process FO's ini file
      try{
        Poco::AutoPtr<Poco::Util::IniFileConfiguration> pConf(new Poco::Util::IniFileConfiguration(iniFile));
        try{
          sphinx_home_dir = pConf->getString("sphinx.home_dir");
          sphinx_default_index = pConf->getString("sphinx.default_index");
          sphinx_start_searchd = pConf->getBool("sphinx.start_searchd");
          sphinx_stop_searchd = pConf->getBool("sphinx.stop_searchd");
          sphinx_ranker = pConf->getInt("sphinx.ranker");
          sphinx_ranker_expression = pConf->getString("sphinx.ranker_expression");
        } catch(Poco::SyntaxException& e){
          //!<Log some error of config file processing
          logger.log(NODE_LOG_MODE_ERROR) << NODE_FUNCTIONAL_ERROR_SPHINX << " data node ini file processing error: " << e.displayText() << flush;
        }
      } catch(Poco::Exception& e){
        //!<Log some error of config file processing
        logger.log(NODE_LOG_MODE_ERROR) << " data node ini file access error: " << e.displayText() << flush;
      }

      //If Sphinx home dir not set, define it as current user home dir
      if(sphinx_home_dir == ""){
        sphinx_home_dir = getUserHomeDir() + NODE_INI_FILE_PATH_DELIMITER_CHAR + NODE_FUNCTIONAL_PROJECT_ROOT_DIR;
      }else{
        if(sphinx_home_dir[0] != NODE_INI_FILE_PATH_START_CHAR){
          sphinx_home_dir = getUserHomeDir() + NODE_INI_FILE_PATH_DELIMITER_CHAR + sphinx_home_dir;
        }
      }

      logger.log(NODE_LOG_MODE_INFORMATION) << "Sphinx FO initialization:" << flush;
      logger.log(NODE_LOG_MODE_INFORMATION) << "nodeName=" << getNodeName() << flush;
      logger.log(NODE_LOG_MODE_INFORMATION) << "home_dir=" << sphinx_home_dir << flush;
      logger.log(NODE_LOG_MODE_INFORMATION) << "default_index=" << sphinx_default_index << flush;
      logger.log(NODE_LOG_MODE_INFORMATION) << "start_searchd=" << sphinx_start_searchd << flush;
      logger.log(NODE_LOG_MODE_INFORMATION) << "stop_searchd=" << sphinx_stop_searchd << flush;
      logger.log(NODE_LOG_MODE_INFORMATION) << "_nodeNumber=" << _nodeNumber << flush;
      logger.log(NODE_LOG_MODE_INFORMATION) << "ranker=" << sphinx_ranker << flush;
      logger.log(NODE_LOG_MODE_INFORMATION) << "ranker_expression=" << sphinx_ranker_expression << flush;

      //!<Create Sphinx Sphinx FO and init
      _sphinxFO = new HCE::sphinx::SphinxFunctionalObject(getNodeName(), _nodeNumber, sphinx_home_dir, sphinx_default_index, sphinx_start_searchd, sphinx_stop_searchd);
      if(_sphinxFO){
        if(_sphinxFO->getErrorCode()){
          //!<Log some error of Sphinx FO
          logger.log(NODE_LOG_MODE_ERROR) << NODE_FUNCTIONAL_ERROR_SPHINX << " Sphinx FO initialization error " << _sphinxFO->getErrorCode() << " : " << _sphinxFO->getErrorMsg()
              << endl << NODE_FUNCTIONAL_ERROR_SPHINX_INTERNAL_LOG_BEGIN << _sphinxFO->logMsg() << NODE_FUNCTIONAL_ERROR_SPHINX_INTERNAL_LOG_END << flush;
        }
      }else{
        logger.log(NODE_LOG_MODE_FATAL) << NODE_FUNCTIONAL_ERROR_SPHINX << " FO instantiate : " << NODE_CRITICAL_ERROR_MESSAGE << flush;
        ::exit(NODE_EXIT_CODE_SPHINX_CRITICAL_ERROR);
      }

      //!<Turn checking the incoming data ON
      _sphinxFO->setCheckIncomingDocuments(true);

      //!<Set ranker
      _sphinxFO->setRankingMode(sphinx_ranker, sphinx_ranker_expression);

      logger.log(NODE_LOG_MODE_DEBUG) << "SPHO_INSTANTIATE [" << _sphinxFO->logMsg() << "]" << flush;
    }

    void DataProcessorData::dataStorageInitDRCE(const std::string& iniFile){
      //!<DRCE FO init parameters
      std::string drce_home_dir = "";
      std::string drce_node_host = "";
      std::string drce_node_port = "";
      std::string drce_tasks_data_dir = "";
      std::string drce_tasks_status_dir = "";
      unsigned int drce_max_threads = 32;
      unsigned int drce_tasks_queue_dump_period = 10000;  // time interval period in msec.
      unsigned int drce_resources_monitor_period = 10000; // time interval period in msec.
      bool drce_state_notification_reset_error_state = false;
      std::string drce_env_vars = "";
      size_t drce_resources_extractor_max_threads = 1024;
      size_t drce_resources_extractor_max_processes = 1024;
      size_t drce_resources_extractor_max_disk_size = 0;
      size_t drce_resources_extractor_max_vram_size = 0;
      size_t drce_resources_extractor_max_rram_size = 0;

      logger.log(NODE_LOG_MODE_INFORMATION) << "load DRCE FO settings from " << iniFile << " file" << flush;

      //!<Process FO's ini file
      try{
        Poco::AutoPtr<Poco::Util::IniFileConfiguration> pConf(new Poco::Util::IniFileConfiguration(iniFile));
        try{
          drce_home_dir = pConf->getString("drce.home_dir");
          drce_tasks_data_dir = pConf->getString("drce.tasks_data_dir");
          drce_tasks_status_dir = pConf->getString("drce.tasks_status_dir");
          drce_node_host = pConf->getString("drce.node_host");
          drce_node_port = pConf->getString("drce.node_port");
          drce_max_threads = pConf->getInt("drce.max_threads");
          drce_tasks_queue_dump_period = pConf->getInt("drce.tasks_queue_dump_period");
          drce_resources_monitor_period = pConf->getInt("drce.resources_monitor_period");
          _state_notification_host = pConf->getString("drce.state_notification_host");
          _state_notification_port = pConf->getString("drce.state_notification_port");
          drce_state_notification_reset_error_state = pConf->getBool("drce.state_notification_reset_error_code");
          drce_env_vars = pConf->getString("drce.env_vars");
          drce_resources_extractor_max_threads = pConf->getInt("drce.drce_resources_extractor_max_threads");
          drce_resources_extractor_max_processes = pConf->getInt("drce.drce_resources_extractor_max_processes");
          drce_resources_extractor_max_disk_size = pConf->getInt("drce.drce_resources_extractor_max_disk_size");
          drce_resources_extractor_max_vram_size = pConf->getInt("drce.drce_resources_extractor_max_vram_size");
          drce_resources_extractor_max_rram_size = pConf->getInt("drce.drce_resources_extractor_max_rram_size");
        } catch(Poco::SyntaxException& e){
          //!<Log some error of config file processing
          logger.log(NODE_LOG_MODE_ERROR) << NODE_FUNCTIONAL_ERROR_DRCE << " data node ini file processing error: " << e.displayText() << flush;
        }
      } catch(Poco::Exception& e){
        //!<Log some error of config file processing
        logger.log(NODE_LOG_MODE_ERROR) << " data node ini file access error: " << e.displayText() << flush;
      }

      //If DRCE home dir not set, define it as current user home dir
      if(drce_home_dir == ""){
        drce_home_dir = getUserHomeDir() + NODE_INI_FILE_PATH_DELIMITER_CHAR + NODE_FUNCTIONAL_PROJECT_ROOT_DIR;
      }else{
        //If first path character is not slash, this means that the path specified is relative to the home dir
        if(drce_home_dir[0] != NODE_INI_FILE_PATH_START_CHAR){
          drce_home_dir = getUserHomeDir() + NODE_INI_FILE_PATH_DELIMITER_CHAR + drce_home_dir;
        }
      }

      //If drce_tasks_data_dir not set
      if(drce_tasks_data_dir == ""){
        drce_tasks_data_dir = drce_home_dir + NODE_INI_FILE_PATH_DELIMITER_CHAR + NODE_FUNCTIONAL_DRCE_TASKS_DATA_DIR;
      }else{
        //If first path character is not slash, this means that the path specified is relative to the home dir
        if(drce_tasks_data_dir[0] != NODE_INI_FILE_PATH_START_CHAR){
          drce_tasks_data_dir = drce_home_dir + NODE_INI_FILE_PATH_DELIMITER_CHAR + NODE_FUNCTIONAL_DRCE_TASKS_DATA_DIR;
        }
      }

      //If drce_tasks_status_dir not set
      if(drce_tasks_status_dir == ""){
        drce_tasks_status_dir = drce_home_dir + NODE_INI_FILE_PATH_DELIMITER_CHAR + NODE_FUNCTIONAL_DRCE_TASKS_STATUS_DIR;
      }else{
        //If first path character is not slash, this means that the path specified is relative to the home dir
        if(drce_tasks_status_dir[0] != NODE_INI_FILE_PATH_START_CHAR){
          drce_tasks_status_dir = drce_home_dir + NODE_INI_FILE_PATH_DELIMITER_CHAR + NODE_FUNCTIONAL_DRCE_TASKS_STATUS_DIR;
        }
      }

      logger.log(NODE_LOG_MODE_INFORMATION) << "DRCE FO initialization:" << flush;
      logger.log(NODE_LOG_MODE_INFORMATION) << "nodeName=" << getNodeName() << flush;
      logger.log(NODE_LOG_MODE_INFORMATION) << "home_dir=" << drce_home_dir << flush;
      logger.log(NODE_LOG_MODE_INFORMATION) << "drce_tasks_data_dir=" << drce_tasks_data_dir << flush;
      logger.log(NODE_LOG_MODE_INFORMATION) << "drce_tasks_status_dir=" << drce_tasks_status_dir << flush;
      logger.log(NODE_LOG_MODE_INFORMATION) << "drce_node_host=" << drce_node_host << flush;
      logger.log(NODE_LOG_MODE_INFORMATION) << "drce_node_port=" << drce_node_port << flush;
      logger.log(NODE_LOG_MODE_INFORMATION) << "drce_max_threads=" << drce_max_threads << flush;
      logger.log(NODE_LOG_MODE_INFORMATION) << "drce_tasks_queue_dump_period=" << drce_tasks_queue_dump_period << flush;
      logger.log(NODE_LOG_MODE_INFORMATION) << "drce_resources_monitor_period=" << drce_resources_monitor_period << flush;
      logger.log(NODE_LOG_MODE_INFORMATION) << "_state_notification_host=" << _state_notification_host << flush;
      logger.log(NODE_LOG_MODE_INFORMATION) << "_state_notification_port=" << _state_notification_port << flush;

      //!<Create DRCE FO and init
      _drceFO = new HCE::drce::DRCEFunctionalObject(getNodeName(), drce_home_dir, drce_tasks_data_dir, drce_tasks_status_dir, drce_node_host, drce_node_port);

      if(_drceFO){
        if(_drceFO->getErrorCode()){
          //!<Log some error of DRCE FO
          logger.log(NODE_LOG_MODE_FATAL) << NODE_FUNCTIONAL_ERROR_DRCE << " DRCE FO initialization error " << _drceFO->getErrorCode() << " : " << _drceFO->getErrorMsg()
              << endl << NODE_FUNCTIONAL_ERROR_DRCE_INTERNAL_LOG_BEGIN << _drceFO->logMsg() << NODE_FUNCTIONAL_ERROR_DRCE_INTERNAL_LOG_END << flush;
          ::exit(NODE_EXIT_CODE_DRCE_CRITICAL_ERROR);
        }
      }else{
        logger.log(NODE_LOG_MODE_FATAL) << NODE_FUNCTIONAL_ERROR_DRCE << " FO instantiate : " << NODE_CRITICAL_ERROR_MESSAGE << flush;
        ::exit(NODE_EXIT_CODE_DRCE_CRITICAL_ERROR);
      }

      //!<Set max threads count for DRCE workers threads pool
      _drceFO->setMaxThreadCount(drce_max_threads);
      //!<Set tasks queue dump period interval (in msec.)
      _drceFO->setTasksQueueDumpPeriod(drce_tasks_queue_dump_period);
      //!<Set resource monitor time period (in msec.)
      _drceFO->setResourceMonitorTimePeriod(drce_resources_monitor_period);
      //!<Set mandatory environment variables
      _drceFO->setEnvironments(drce_env_vars);
      //!<Set max threads count for resource extractor
      _drceFO->setResourceExtractorMaxThreadsCount(drce_resources_extractor_max_threads);
      //!<Set max processes count for resource extractor
      _drceFO->setResourceExtractorMaxProcessesCount(drce_resources_extractor_max_processes);
      //!<Set max disk size for resource extractor
      if (drce_resources_extractor_max_disk_size)
        _drceFO->setResourceExtractorMaxDiskSize(drce_resources_extractor_max_disk_size);
      //!<Set max virtual memory size for resource extractor
      if (drce_resources_extractor_max_vram_size)
        _drceFO->setResourceExtractorMaxVramSize(drce_resources_extractor_max_vram_size);
      //!<Set max rss memory size for resource extractor
      if (drce_resources_extractor_max_rram_size)
        _drceFO->setResourceExtractorMaxRramSize(drce_resources_extractor_max_rram_size);

      //!<Create connection socket and connect to tasks state notifications service
      if(rebuildTasksStateNotificationConnection()){
        logger.log(NODE_LOG_MODE_DEBUG) << "Tasks state notification connection created!" << flush;
        //!<Create instance or the notification functor
        _tasksStateNotificationFunctor = new TasksStateNotificationFunctor(_tasksStateNotificationSock);
        //!<Set notification function to DRCE FO instance
        _drceFO->setNotificationFunctor(*_tasksStateNotificationFunctor);
        //!<Set flag reset error state for notification
        _drceFO->setResetErrorCodeStateNotification(drce_state_notification_reset_error_state);
        logger.log(NODE_LOG_MODE_DEBUG) << "Tasks state notification functor is set!" << flush;
      }

      logger.log(NODE_LOG_MODE_DEBUG) << "DRCE_FO_INSTANTIATE [" << _drceFO->logMsg() << "]" << flush;
    }

    std::string DataProcessorData::getUserHomeDir(void){
      std::string homeDir;

      struct passwd *pw = getpwuid(getuid());
      const char *homedir = pw->pw_dir;
      if(homedir){
        homeDir = homedir;
      }

      return homeDir;
    }

    void DataProcessorData::nodeInit(const std::string& iniFile){
      //!<Node init parameters
      unsigned int processingMode = NODE_DATA_PROCESSING_MODE_NORMAL;

      logger.log(NODE_LOG_MODE_INFORMATION) << "load node settings from " << iniFile << " file" << flush;

      //!<Process Node's ini file
      try{
        Poco::AutoPtr<Poco::Util::IniFileConfiguration> pConf(new Poco::Util::IniFileConfiguration(iniFile));
        try{
          //Set node number from ini
          _nodeNumber = pConf->getInt("node.number");
          //Set data processing mode from ini
          processingMode = pConf->getInt("node.data_processing_mode");
          if(processingMode > 0){
            _dataProcessingFakeResults = pConf->getInt("node.data_processing_fake_results");
            _dataProcessingFakeResultsMax = pConf->getInt("node.data_processing_fake_max_results");
          }
        } catch(Poco::SyntaxException& e){
          //!<Log some error of config file processing
          logger.log(NODE_LOG_MODE_ERROR) << "nodeInit() - data node ini file processing error: " << e.displayText() << flush;
        }
      } catch(Poco::Exception& e){
        //!<Log some error of config file processing
        logger.log(NODE_LOG_MODE_ERROR) << "nodeInit() - data node ini file access error: " << e.displayText() << flush;
      }

      _dataProcessingMode = processingMode;

      logger.log(NODE_LOG_MODE_INFORMATION) << "node.number: " << _nodeNumber << flush;
      logger.log(NODE_LOG_MODE_INFORMATION) << "node.data_processing_mode: " << _dataProcessingMode << flush;
      logger.log(NODE_LOG_MODE_INFORMATION) << "node.data_processing_fake_results: " << _dataProcessingFakeResults << flush;
      logger.log(NODE_LOG_MODE_INFORMATION) << "node.data_processing_fake_results_max: " << _dataProcessingFakeResultsMax << flush;
    }

    std::string DataProcessorData::processDataRequest(const std::string& messageBody){
      std::string ret;

      if(messageBody.size() > 0 && messageBody[0] == NODE_JSON_BEGIN_CHAR){
        logger.log(NODE_LOG_MODE_TRACE) << "JSON processing..." << flush;

        //!<Possible json message
        ret = processDataRequestJson(messageBody);
      }else{
        logger.log(NODE_LOG_MODE_TRACE) << "PLAIN_TEXT processing..." << flush;

        //!<Plain text message
        ret = processDataRequestPlainText(messageBody);
      }

      return ret;
    }

    std::string DataProcessorData::processDataRequestJson(const std::string& messageBody){
      //Response json string
      std::string ret;

      //Create default empty cover object for results
      HCE::JsonMessageCover msgCoverOutEmpty;

      //Serialize default empty cover object
      if(!msgCoverOutEmpty.serialize(ret)){
        logger.log(NODE_LOG_MODE_DEBUG) << "Default empty cover object serialize error " << msgCoverOutEmpty.getErrorCode() << " : " << msgCoverOutEmpty.getErrorMsg() << flush;
      }

      logger.log(NODE_LOG_MODE_DEBUG) << "Request json : " << messageBody << flush;

      //Unserialize request json
      HCE::JsonMessageCover msgCoverIn(messageBody);

      if(msgCoverIn.getErrorCode()){
        logger.log(NODE_LOG_MODE_ERROR) << "Data request cover json deserialization error : " << msgCoverIn.getErrorMsg() << flush;
      }else{
        logger.log(NODE_LOG_MODE_DEBUG) << "Json request object from cover: [" << msgCoverIn.getData() << "]" << flush;

        if(msgCoverIn.getType() == HCE::types::MessageType::mtSphinx){
          //Process Sphinx request type
          std::string sphinxResultDataJson = processDataRequestSphinx(msgCoverIn.getData());

          //Create cover object for results
          HCE::JsonMessageCover msgCoverOut(msgCoverIn.getType(), sphinxResultDataJson, msgCoverIn.getTTL());

          //Serialize output cover object
          if(!msgCoverOut.serialize(ret)){
            //!<If serialization error return default result to reducer
            logger.log(NODE_LOG_MODE_ERROR) << "Cover object for Sphinx request serialize error " << msgCoverOut.getErrorCode() << " : " << msgCoverOut.getErrorMsg() << flush;
          }
        }else{
          if(msgCoverIn.getType() == HCE::types::MessageType::mtDrce){
            //Process DRCE request type
            std::string drceResultDataJson = processDataRequestDRCE(msgCoverIn.getData());

            //Create cover object for results
            HCE::JsonMessageCover msgCoverOut(msgCoverIn.getType(), drceResultDataJson, msgCoverIn.getTTL());

            //Serialize output cover object
            if(!msgCoverOut.serialize(ret)){
              //!<If serialization error return default result to reducer
              logger.log(NODE_LOG_MODE_ERROR) << "Cover object for DRCE request serialize error " << msgCoverOut.getErrorCode() << " : " << msgCoverOut.getErrorMsg() << flush;
            }
          }else{
            logger.log(NODE_LOG_MODE_ERROR) << "Data request type error : type " << static_cast<int>(msgCoverIn.getType()) << " not supported" << flush;
          }
        }
      }

      logger.log(NODE_LOG_MODE_DEBUG) << "Result json : " << ret << flush;

      return ret;
    }

    std::string DataProcessorData::processDataRequestSphinx(const std::string& messageBody){
      //Create empty Sphinx result object to return it for farther processing by default
      HCE::sphinx::SphinxDefaultJSON defaultJson;
      std::string ret = defaultJson.getJSON();

      //Create message cover object and fill it with default results json to return it to upper level processing (reducing) in case of some error happened
      HCE::JsonMessageCover msgCover(HCE::types::MessageType::mtSphinx, ret);
      if(!msgCover.serialize(ret)){
        //!<If serialization error return default result to reducer
        logger.log(NODE_LOG_MODE_ERROR) << NODE_FUNCTIONAL_ERROR_SPHINX << " cover object from default results - serialization error "
            << msgCover.getErrorCode() << " : " << msgCover.getErrorMsg() << flush;
      }

      logger.log(NODE_LOG_MODE_DEBUG) << "Sphinx data handler request, json=[" << messageBody << "]" << flush;

      //If node data processing mode is normal - enter FO usage, else in simulation mode - return default results
      if(_dataProcessingMode == NODE_DATA_PROCESSING_MODE_NORMAL){
        //!<Create input object for Sphinx FO
        HCE::sphinx::SphinxInputJsonMessage inputMsg(messageBody);
        if(!inputMsg.isError()){
          if(_sphinxFO->getErrorCode()){
            //!<Return empty string to reducer in case of searcher state error
            logger.log(NODE_LOG_MODE_ERROR) << NODE_FUNCTIONAL_ERROR_SPHINX << " searcher state error " << _sphinxFO->getErrorCode() << " : " << _sphinxFO->getErrorMsg()
                << endl << NODE_FUNCTIONAL_ERROR_SPHINX_INTERNAL_LOG_BEGIN << _sphinxFO->logMsg() << NODE_FUNCTIONAL_ERROR_SPHINX_INTERNAL_LOG_END << flush;

            //!<Reset searcher error state to get next message to start from scratch
            _sphinxFO->resetError();
          }else{
            //!<Execute search request
            HCE::sphinx::SphinxOutputJsonMessage outputMessage = _sphinxFO->Process(inputMsg);

            //TODO temporary for filters debug reason, need to be removed in release
            logger.log(NODE_LOG_MODE_DEBUG) << NODE_FUNCTIONAL_ERROR_SPHINX_INTERNAL_LOG_BEGIN << _sphinxFO->logMsg() << NODE_FUNCTIONAL_ERROR_SPHINX_INTERNAL_LOG_END << flush;
            //TODO

            if(_sphinxFO->getErrorCode()){
              //!<Log some error of Sphinx FO
              logger.log(NODE_LOG_MODE_ERROR) << NODE_FUNCTIONAL_ERROR_SPHINX << " search process error " << _sphinxFO->getErrorCode() << " : " << _sphinxFO->getErrorMsg()
                  << endl << NODE_FUNCTIONAL_ERROR_SPHINX_INTERNAL_LOG_BEGIN << _sphinxFO->logMsg() << NODE_FUNCTIONAL_ERROR_SPHINX_INTERNAL_LOG_END << flush;
              //!<Reset searcher error state to get next message to start from scratch
              _sphinxFO->resetError();
            }

            if(outputMessage.getErrorCode()){
              //!<Log some error of search request execution
              logger.log(NODE_LOG_MODE_ERROR) << NODE_FUNCTIONAL_ERROR_SPHINX << " output message error " << outputMessage.getErrorCode() << " : " << outputMessage.getErrorMessage() << flush;
            }

            logger.log(NODE_LOG_MODE_DEBUG) << "Sphinx time:" << outputMessage.getTime() << flush;

            //Return Sphinx results data json string of data from Sphinx FO
            ret = outputMessage.getData();

            //Register Sphinx request time
            _statSphinxTimeTotalSearch += outputMessage.getTime();
          }
        }else{
          //!<Return empty json string to reducer in case of de-serialisation error to direct the processing on json type
          logger.log(NODE_LOG_MODE_DEBUG) << NODE_FUNCTIONAL_ERROR_SPHINX << " deserialization error " << inputMsg.getErrorCode() << " : " << inputMsg.getErrorMsg() << flush;
        }
      }else{
        //Fake processing mode, random generator of results set
        //Create message cover object and fill it with default results json to return it to upper level processing (reducing) in case of some error happened
        if(defaultJson.makeJSON(_dataProcessingFakeResults, 1, _dataProcessingFakeResultsMax)){
          //Return Sphinx results data json string of data from fake results generator
          ret = defaultJson.getJSON();
        }else{
          //TODO
          logger.log(NODE_LOG_MODE_ERROR) << "Error create fake results " << _dataProcessingFakeResults << ", zero results returned for Sphinx" << flush;
        }
      }

      logger.log(NODE_LOG_MODE_DEBUG) << "Sphinx results json : " << ret << flush;

      //Register Sphinx requests number stat
      _statSphinxRequestsSearch++;

      return ret;
    }

    std::string DataProcessorData::processDataRequestDRCE(const std::string& messageBody){
      //Create empty DRCE result object to return it for farther processing by default
      HCE::drce::DRCEDefaultJSON defaultJson;
      std::string ret = defaultJson.getJSON();

      logger.log(NODE_LOG_MODE_DEBUG) << "DRCE data handler request, json=[" << messageBody << "]" << flush;

      //If node data processing mode is normal - enter FO usage, else in simulation mode - return default results
      if(_dataProcessingMode == NODE_DATA_PROCESSING_MODE_NORMAL){
        //!<Create input object for DRCE FO
        HCE::drce::DRCEInputJsonMessage inputMsg(messageBody);
        if(!inputMsg.isError()){
          if(_drceFO->getErrorCode()){
            //!<Return empty string to reducer in case of DRCE state error
            logger.log(NODE_LOG_MODE_ERROR) << NODE_FUNCTIONAL_ERROR_DRCE << " DRCE state error " << _drceFO->getErrorCode() << " : " << _drceFO->getErrorMsg()
                << endl << NODE_FUNCTIONAL_ERROR_DRCE_INTERNAL_LOG_BEGIN << _drceFO->logMsg() << NODE_FUNCTIONAL_ERROR_DRCE_INTERNAL_LOG_END << flush;

            //!<Reset DRCE error state to get next message to start from scratch
            _drceFO->resetError();
          }else{
            //!<Execute DRCE request
            HCE::drce::DRCEOutputJsonMessage outputMessage = _drceFO->Process(inputMsg);

            //TODO temporary for filters debug reason, need to be removed in release
            logger.log(NODE_LOG_MODE_DEBUG) << NODE_FUNCTIONAL_ERROR_DRCE_INTERNAL_LOG_BEGIN << _drceFO->logMsg() << NODE_FUNCTIONAL_ERROR_DRCE_INTERNAL_LOG_END << flush;
            //TODO

            if(_drceFO->getErrorCode()){
              //!<Log some error of DRCE FO
              logger.log(NODE_LOG_MODE_ERROR) << NODE_FUNCTIONAL_ERROR_DRCE << " execution error " << _drceFO->getErrorCode() << " : " << _drceFO->getErrorMsg()
                  << endl << NODE_FUNCTIONAL_ERROR_DRCE_INTERNAL_LOG_BEGIN << _drceFO->logMsg() << NODE_FUNCTIONAL_ERROR_DRCE_INTERNAL_LOG_END << flush;

              //!<Reset DRCE error state to get next message to start from scratch
              _drceFO->resetError();
            }

            if(outputMessage.getErrorCode()){
              //!<Log some error of DRCE request execution
              logger.log(NODE_LOG_MODE_ERROR) << NODE_FUNCTIONAL_ERROR_DRCE << " output message error " << outputMessage.getErrorCode() << " : " << outputMessage.getErrorMsg() << flush;
            }
            logger.log(NODE_LOG_MODE_DEBUG) << "DRCE time:" << _drceFO->getRequestTime() << flush;

            //Return DRCE results data json string of data from the DRCEOutputJsonMessage object
            outputMessage.serialize(ret);

            //Register DRCE request time
            _statDRCETimeTotal += _drceFO->getRequestTime();
          }
        }else{
          //!<Return empty json string to reducer in case of de-serialisation error to direct the processing on json type
          logger.log(NODE_LOG_MODE_ERROR) << NODE_FUNCTIONAL_ERROR_DRCE << " deserialization error " << inputMsg.getErrorCode() << " : " << inputMsg.getErrorMsg() << flush;
          //!<Fill error code and error message for default json
          defaultJson.setJsonErrorCode(inputMsg.getErrorCode());
          defaultJson.setJsonErrorMessage(inputMsg.getErrorMsg());
          defaultJson.setJsonRequestType(inputMsg.getRequestType());
          //defaultJson.setJsonRequestId(inputMsg.getRequestId());
          defaultJson.setJsonNodeHost(_drceFO->getNodeHost());
          defaultJson.setJsonNodePort(_drceFO->getNodePort());
          defaultJson.setJsonNodeName(_drceFO->getNodeName());
          defaultJson.setJsonState(DRCE_REQUEST_STATE_ERROR_SET_AS_NEW_TASK);
          defaultJson.makeJSON();
          ret = defaultJson.getJSON();
        }
      }else{
        //Fake processing mode, random generator of results set
        //Create message cover object and fill it with default results json to return it to upper level processing (reducing) in case of some error happened
        if(defaultJson.makeJSON(_dataProcessingFakeResults)){
          //Return DRCE results data json string of data from fake results generator
          ret = defaultJson.getJSON();
        }else{
          logger.log(NODE_LOG_MODE_ERROR) << "Error create fake results " << _dataProcessingFakeResults << ", zero results returned for DRCE" << flush;
        }
      }

      //TODO: test, need to be removed
      //ret = DRCE_FAKE_RESPONSE;

      logger.log(NODE_LOG_MODE_DEBUG) << "DRCE results json : " << ret << flush;

      //Register DRCE requests number stat
      _statDRCERequests++;

      return ret;
    }

    std::string DataProcessorData::processDataRequestPlainText(const std::string& messageBody){
      //!<Parse request and fetch items number
      unsigned int items_number;
      std::istringstream(messageBody) >> items_number;
      //!<Generate items
      std::vector<unsigned long long> items(items_number);
      //!<Randomize items values
      if(items_number > 0){
        for(unsigned long long i = 0; i < items.size(); ++i){
          items[i] = (rand() % RAND_MAX) + 1;
        }
      }else{
        items.resize(1);
      }

      logger.log(NODE_LOG_MODE_DEBUG) << "DATA items requested:" << items_number << ", items generated:" << items.size() << flush;

      //!<Sord items
      sort(items.begin(), items.end(), std::less<unsigned long long>());

      logger.log(NODE_LOG_MODE_DEBUG) << "DATA sorted" << flush;
      //!<Serialize items
      std::ostringstream oss;
      std::copy(items.begin(), items.end() - 1, std::ostream_iterator<int>(oss, " "));
      oss << items.back();
      std::string body(oss.str());

      return body;
    }

    void DataProcessorData::handleInternalMessage(void){
      //!<Fetch message
      zmsg msg(*_inprocDataSock);

      logger.log(NODE_LOG_MODE_TRACE) << "MSG FROM inproc_data" << endl << msg.dump() << flush;
      //!<Get identity and body of request message message
      std::string messageId((const char*)msg.pop_front().c_str());
      std::string messageBody((const char*)msg.pop_front().c_str());
      std::string messageRoute((msg.parts())?(const char*)msg.pop_front().c_str():"");

      _statClientResponses++;

      logger.log(NODE_LOG_MODE_DEBUG) << "DATA PROCESSOR RECEIVED message [" << messageId << "] body " << messageBody.size() << " bytes n=" << _statClientResponses << flush;

      //!<Process data request
      std::string body = processDataRequest(messageBody);

      //!<Create response message
      msg.append(messageId.c_str());
      msg.append(body.c_str());
      msg.append(messageRoute.c_str());

      logger.log(NODE_LOG_MODE_TRACE) << "DATA PROCESSOR RESPONSE message [" << messageId << "] body [" << body << "]" << flush;

      //!<Send response
      msg.send(*_inprocDataSock);

      logger.log(NODE_LOG_MODE_DEBUG) << "DATA PROCESSOR RESPONSE message [" << messageId << "] body " << body.size() << " bytes sent" << flush;
    }

    std::string DataProcessorData::processAdminCommands(std::string & command, std::string & parameters){
      std::string ret = NODE_ADMIN_ERROR_OK;

      logger.log(NODE_LOG_MODE_DEBUG) << "Admin command requested [" << command << "], parameters [" << parameters << "]" << flush;

      //!<Process commands
      if(command == NODE_MSG_STAT){
        std::stringstream result;
        result << NODE_ADMIN_ERROR_OK << NODE_ADMIN_COMMAND_DELIMITER << "responses=" << _statClientResponses << NODE_ADMIN_COMMAND_DELIMITER << "avgSphinxSearchTime="
            << (_statSphinxTimeTotalSearch / (_statSphinxRequestsSearch + 1)) << NODE_ADMIN_COMMAND_DELIMITER << "avgSphinxAdminTime="
            << (_statSphinxTimeTotalAdmin / (_statSphinxRequestsAdmin + 1)) << NODE_ADMIN_COMMAND_DELIMITER << "sphinxSearchRequests=" << _statSphinxRequestsSearch
            << NODE_ADMIN_COMMAND_DELIMITER << "sphinxAdminRequests=" << _statSphinxRequestsAdmin << NODE_ADMIN_COMMAND_DELIMITER << "DRCERequests=" << _statDRCERequests
            << NODE_ADMIN_COMMAND_DELIMITER << "avgDRCETime=" << (_statDRCETimeTotal / (_statDRCERequests + 1));
        ret = result.str();
      }else if(command == NODE_MSG_PROPERTIES){
        //!<Make property information
        ret = getPropertyInformation();
      }else if(command == NODE_MSG_SET_DATA_PROCESSING_MODE){
        std::stringstream result;

        //!<Parse command parameters and fetch processing mode value
        unsigned int processingMode;
        std::istringstream(parameters) >> processingMode;

        result << NODE_ADMIN_ERROR_OK << NODE_ADMIN_COMMAND_DELIMITER << std::to_string(_dataProcessingMode) << NODE_ADMIN_COMMAND_DELIMITER << _name;

        //Change data processing mode
        _dataProcessingMode = processingMode;

        ret = result.str();
      }else if(command == NODE_MSG_SPHINX){
        //!<Sphinx functional object command
        ret = processAdminCommandsSphinx(parameters);
      }else if(command == NODE_MSG_DRCE){
        //!<DRCE functional object command
        ret = NODE_ADMIN_ERROR_OK + NODE_ADMIN_COMMAND_DELIMITER + processDataRequestDRCE(parameters);
        //ret = NODE_ADMIN_ERROR_OK + NODE_ADMIN_COMMAND_DELIMITER + DRCE_FAKE_RESPONSE;
      }else if((command == NODE_MSG_DRCE_SET_HOST) || (command == NODE_MSG_DRCE_GET_HOST)){
        //!<Prepare response message
        ret = NODE_ADMIN_ERROR_OK + NODE_ADMIN_COMMAND_DELIMITER + _drceFO->getNodeHost() + NODE_ADMIN_COMMAND_DELIMITER;
        if(command == NODE_MSG_DRCE_SET_HOST){
          //Change DRCE FO host
          _drceFO->setNodeHost(parameters);
        }
      }else if((command == NODE_MSG_DRCE_SET_PORT) || (command == NODE_MSG_DRCE_GET_PORT)){
        //!<Prepare response message
        ret = NODE_ADMIN_ERROR_OK + NODE_ADMIN_COMMAND_DELIMITER + _drceFO->getNodePort() + NODE_ADMIN_COMMAND_DELIMITER;
        if(command == NODE_MSG_DRCE_SET_PORT){
          //Change DRCE FO port
          _drceFO->setNodePort(parameters);
        }
      }else if(command == NODE_MSG_DRCE_GET_TASKS){
        ret = _drceFO->getCurrentTasksQueue();
      }else if(command == NODE_MSG_DRCE_GET_TASKS_INFO){
        ret = _drceFO->getListInfoAllTasks();
      }else if(command == NODE_MSG_RESOURCE_USAGE){
        ret = _drceFO->getHostResourceUsage();
      }else{
        //!<Unsupported command
        ret = NODE_ADMIN_ERROR_ERROR + NODE_ADMIN_COMMAND_DELIMITER + NODE_ADMIN_ERROR_UNSUPPORTED_COMMAND;
      }

      return ret;
    }

    std::string DataProcessorData::getPropertyInformation(void){
      std::stringstream result;
      result  << NODE_ADMIN_ERROR_OK
          << NODE_ADMIN_COMMAND_DELIMITER << getProperties() << "responses=" << _statClientResponses
          << NODE_ADMIN_COMMAND_DELIMITER << "iniFile=" << _iniFile
          << NODE_ADMIN_COMMAND_DELIMITER << "nodeName=" << _nodeName
          << NODE_ADMIN_COMMAND_DELIMITER << "dataProcessingMode=" << _dataProcessingMode
          << NODE_ADMIN_COMMAND_DELIMITER << "avgSphinxSearchTime=" << (_statSphinxTimeTotalSearch / (_statSphinxRequestsSearch + 1))
          << NODE_ADMIN_COMMAND_DELIMITER << "avgSphinxAdminTime=" << (_statSphinxTimeTotalAdmin / (_statSphinxRequestsAdmin + 1))
          << NODE_ADMIN_COMMAND_DELIMITER << "sphinxSearchRequests=" << _statSphinxRequestsSearch
          << NODE_ADMIN_COMMAND_DELIMITER << "sphinxAdminRequests=" << _statSphinxRequestsAdmin
          << NODE_ADMIN_COMMAND_DELIMITER << "sphinxDataStorageProperties=" << getDataStorageSphinxProperties()
          << NODE_ADMIN_COMMAND_DELIMITER << "drceDataStorageProperties=" << getDataStorageDRCEProperties()
          << NODE_ADMIN_COMMAND_DELIMITER << "DRCERequests=" << _statDRCERequests
          << NODE_ADMIN_COMMAND_DELIMITER << "avgDRCETime=" << (_statDRCETimeTotal / (_statDRCERequests + 1))
          << NODE_ADMIN_COMMAND_DELIMITER << "nodeNumber=" << _nodeNumber;
      return result.str();
    }

    std::string DataProcessorData::getDataStorageSphinxProperties(void){
      std::stringstream ret;
      ret << "isConnected" << NODE_ADMIN_COMMAND_PROPERTY_ITEM_PAIR_DELIMITER << _sphinxFO->isConnected() << NODE_ADMIN_COMMAND_PROPERTY_ITEMS_DELIMITER << "indexName"
          << NODE_ADMIN_COMMAND_PROPERTY_ITEM_PAIR_DELIMITER << _sphinxFO->getIndexName() << NODE_ADMIN_COMMAND_PROPERTY_ITEMS_DELIMITER << "dataDir"
          << NODE_ADMIN_COMMAND_PROPERTY_ITEM_PAIR_DELIMITER << _sphinxFO->getDataDir() << NODE_ADMIN_COMMAND_PROPERTY_ITEMS_DELIMITER << "etcDir"
          << NODE_ADMIN_COMMAND_PROPERTY_ITEM_PAIR_DELIMITER << _sphinxFO->getEtcDir() << NODE_ADMIN_COMMAND_PROPERTY_ITEMS_DELIMITER << "homeDir"
          << NODE_ADMIN_COMMAND_PROPERTY_ITEM_PAIR_DELIMITER << _sphinxFO->getHomeDir() << NODE_ADMIN_COMMAND_PROPERTY_ITEMS_DELIMITER << "logDir"
          << NODE_ADMIN_COMMAND_PROPERTY_ITEM_PAIR_DELIMITER << _sphinxFO->getLogDir() << NODE_ADMIN_COMMAND_PROPERTY_ITEMS_DELIMITER << "runDir"
          << NODE_ADMIN_COMMAND_PROPERTY_ITEM_PAIR_DELIMITER << _sphinxFO->getRunDir() << NODE_ADMIN_COMMAND_PROPERTY_ITEMS_DELIMITER << "searchdHost"
          << NODE_ADMIN_COMMAND_PROPERTY_ITEM_PAIR_DELIMITER << _sphinxFO->getServerHost() << NODE_ADMIN_COMMAND_PROPERTY_ITEMS_DELIMITER << "searchdPort"
          << NODE_ADMIN_COMMAND_PROPERTY_ITEM_PAIR_DELIMITER << _sphinxFO->getServerPort() << NODE_ADMIN_COMMAND_PROPERTY_ITEMS_DELIMITER << "timeoutedRequests"
          << NODE_ADMIN_COMMAND_PROPERTY_ITEM_PAIR_DELIMITER << _sphinxFO->getTimeoutedRequests() << NODE_ADMIN_COMMAND_PROPERTY_ITEMS_DELIMITER << "ranker"
          << NODE_ADMIN_COMMAND_PROPERTY_ITEM_PAIR_DELIMITER << _sphinxFO->getRankingMode() << NODE_ADMIN_COMMAND_PROPERTY_ITEMS_DELIMITER << "ranker_expression"
          << NODE_ADMIN_COMMAND_PROPERTY_ITEM_PAIR_DELIMITER << Poco::replace(_sphinxFO->getRankerExpression(), "=", "%3D") << NODE_ADMIN_COMMAND_PROPERTY_ITEMS_DELIMITER << "nodeNumber"
          << NODE_ADMIN_COMMAND_PROPERTY_ITEM_PAIR_DELIMITER << _sphinxFO->getNodeNumber() << NODE_ADMIN_COMMAND_PROPERTY_ITEMS_DELIMITER << "nodeName"
          << NODE_ADMIN_COMMAND_PROPERTY_ITEM_PAIR_DELIMITER << _sphinxFO->getNodeName();

      return ret.str();
    }

    std::string DataProcessorData::getDataStorageDRCEProperties(void){
      std::stringstream ret;
      HCE::drce::StatVariables statVariables = _drceFO->getStatVariables();
      ret << "nodeName" << NODE_ADMIN_COMMAND_PROPERTY_ITEM_PAIR_DELIMITER << _drceFO->getNodeName() << NODE_ADMIN_COMMAND_PROPERTY_ITEMS_DELIMITER
          << "nodeHost" << NODE_ADMIN_COMMAND_PROPERTY_ITEM_PAIR_DELIMITER << _drceFO->getNodeHost() << NODE_ADMIN_COMMAND_PROPERTY_ITEMS_DELIMITER
          << "nodePort" << NODE_ADMIN_COMMAND_PROPERTY_ITEM_PAIR_DELIMITER << _drceFO->getNodePort() << NODE_ADMIN_COMMAND_PROPERTY_ITEMS_DELIMITER
          << "homeDir" << NODE_ADMIN_COMMAND_PROPERTY_ITEM_PAIR_DELIMITER << _drceFO->getHomeDir() << NODE_ADMIN_COMMAND_PROPERTY_ITEMS_DELIMITER
          << "tasksDataDir" << NODE_ADMIN_COMMAND_PROPERTY_ITEM_PAIR_DELIMITER << _drceFO->getTasksDataDir() << NODE_ADMIN_COMMAND_PROPERTY_ITEMS_DELIMITER
          << "tasksStatusDir" << NODE_ADMIN_COMMAND_PROPERTY_ITEM_PAIR_DELIMITER << _drceFO->getTasksStatusDir() << NODE_ADMIN_COMMAND_PROPERTY_ITEMS_DELIMITER
          << "maxThreadCount" << NODE_ADMIN_COMMAND_PROPERTY_ITEM_PAIR_DELIMITER << _drceFO->getMaxThreadCount() << NODE_ADMIN_COMMAND_PROPERTY_ITEMS_DELIMITER
          << "tasksQueueDumpPeriod" << NODE_ADMIN_COMMAND_PROPERTY_ITEM_PAIR_DELIMITER << _drceFO->getTasksQueueDumpPeriod() << NODE_ADMIN_COMMAND_PROPERTY_ITEMS_DELIMITER
          << "resourceMonitorTimePeriod" << NODE_ADMIN_COMMAND_PROPERTY_ITEM_PAIR_DELIMITER << _drceFO->getResourceMonitorTimePeriod() << NODE_ADMIN_COMMAND_PROPERTY_ITEMS_DELIMITER
          << "asyncTasksCount" << NODE_ADMIN_COMMAND_PROPERTY_ITEM_PAIR_DELIMITER << _drceFO->getAsyncTasksCount() << NODE_ADMIN_COMMAND_PROPERTY_ITEMS_DELIMITER
          << "requestTime" << NODE_ADMIN_COMMAND_PROPERTY_ITEM_PAIR_DELIMITER << _drceFO->getRequestTime() << NODE_ADMIN_COMMAND_PROPERTY_ITEMS_DELIMITER

          << "timeAsyncTask.min" << NODE_ADMIN_COMMAND_PROPERTY_ITEM_PAIR_DELIMITER << statVariables.timeAsyncTask.min << NODE_ADMIN_COMMAND_PROPERTY_ITEMS_DELIMITER
          << "timeAsyncTask.avg" << NODE_ADMIN_COMMAND_PROPERTY_ITEM_PAIR_DELIMITER << statVariables.timeAsyncTask.avg << NODE_ADMIN_COMMAND_PROPERTY_ITEMS_DELIMITER
          << "timeAsyncTask.max" << NODE_ADMIN_COMMAND_PROPERTY_ITEM_PAIR_DELIMITER << statVariables.timeAsyncTask.max << NODE_ADMIN_COMMAND_PROPERTY_ITEMS_DELIMITER

          << "timeSyncTask.min" << NODE_ADMIN_COMMAND_PROPERTY_ITEM_PAIR_DELIMITER << statVariables.timeSyncTask.min << NODE_ADMIN_COMMAND_PROPERTY_ITEMS_DELIMITER
          << "timeSyncTask.avg" << NODE_ADMIN_COMMAND_PROPERTY_ITEM_PAIR_DELIMITER << statVariables.timeSyncTask.avg << NODE_ADMIN_COMMAND_PROPERTY_ITEMS_DELIMITER
          << "timeSyncTask.max" << NODE_ADMIN_COMMAND_PROPERTY_ITEM_PAIR_DELIMITER << statVariables.timeSyncTask.max << NODE_ADMIN_COMMAND_PROPERTY_ITEMS_DELIMITER

          << "sizeInputBufferSyncTasks.min" << NODE_ADMIN_COMMAND_PROPERTY_ITEM_PAIR_DELIMITER << statVariables.sizeInputBufferSyncTasks.min << NODE_ADMIN_COMMAND_PROPERTY_ITEMS_DELIMITER
          << "sizeInputBufferSyncTasks.avg" << NODE_ADMIN_COMMAND_PROPERTY_ITEM_PAIR_DELIMITER << statVariables.sizeInputBufferSyncTasks.avg << NODE_ADMIN_COMMAND_PROPERTY_ITEMS_DELIMITER
          << "sizeInputBufferSyncTasks.max" << NODE_ADMIN_COMMAND_PROPERTY_ITEM_PAIR_DELIMITER << statVariables.sizeInputBufferSyncTasks.max << NODE_ADMIN_COMMAND_PROPERTY_ITEMS_DELIMITER

          << "sizeOutputBufferSyncTasks.min" << NODE_ADMIN_COMMAND_PROPERTY_ITEM_PAIR_DELIMITER << statVariables.sizeOutputBufferSyncTasks.min << NODE_ADMIN_COMMAND_PROPERTY_ITEMS_DELIMITER
          << "sizeOutputBufferSyncTasks.avg" << NODE_ADMIN_COMMAND_PROPERTY_ITEM_PAIR_DELIMITER << statVariables.sizeOutputBufferSyncTasks.avg << NODE_ADMIN_COMMAND_PROPERTY_ITEMS_DELIMITER
          << "sizeOutputBufferSyncTasks.max" << NODE_ADMIN_COMMAND_PROPERTY_ITEM_PAIR_DELIMITER << statVariables.sizeOutputBufferSyncTasks.max << NODE_ADMIN_COMMAND_PROPERTY_ITEMS_DELIMITER

          << "sizeInputBufferAsyncTasks.min" << NODE_ADMIN_COMMAND_PROPERTY_ITEM_PAIR_DELIMITER << statVariables.sizeInputBufferAsyncTasks.min << NODE_ADMIN_COMMAND_PROPERTY_ITEMS_DELIMITER
          << "sizeInputBufferAsyncTasks.avg" << NODE_ADMIN_COMMAND_PROPERTY_ITEM_PAIR_DELIMITER << statVariables.sizeInputBufferAsyncTasks.avg << NODE_ADMIN_COMMAND_PROPERTY_ITEMS_DELIMITER
          << "sizeInputBufferAsyncTasks.max" << NODE_ADMIN_COMMAND_PROPERTY_ITEM_PAIR_DELIMITER << statVariables.sizeInputBufferAsyncTasks.max << NODE_ADMIN_COMMAND_PROPERTY_ITEMS_DELIMITER

          << "sizeOutputBufferAsyncTasks.min" << NODE_ADMIN_COMMAND_PROPERTY_ITEM_PAIR_DELIMITER << statVariables.sizeOutputBufferAsyncTasks.min << NODE_ADMIN_COMMAND_PROPERTY_ITEMS_DELIMITER
          << "sizeOutputBufferAsyncTasks.avg" << NODE_ADMIN_COMMAND_PROPERTY_ITEM_PAIR_DELIMITER << statVariables.sizeOutputBufferAsyncTasks.avg << NODE_ADMIN_COMMAND_PROPERTY_ITEMS_DELIMITER
          << "sizeOutputBufferAsyncTasks.max" << NODE_ADMIN_COMMAND_PROPERTY_ITEM_PAIR_DELIMITER << statVariables.sizeOutputBufferAsyncTasks.max << NODE_ADMIN_COMMAND_PROPERTY_ITEMS_DELIMITER

          << "cpuUsageAsyncTasks.min" << NODE_ADMIN_COMMAND_PROPERTY_ITEM_PAIR_DELIMITER << statVariables.cpuUsageAsyncTasks.min << NODE_ADMIN_COMMAND_PROPERTY_ITEMS_DELIMITER
          << "cpuUsageAsyncTasks.avg" << NODE_ADMIN_COMMAND_PROPERTY_ITEM_PAIR_DELIMITER << statVariables.cpuUsageAsyncTasks.avg << NODE_ADMIN_COMMAND_PROPERTY_ITEMS_DELIMITER
          << "cpuUsageAsyncTasks.max" << NODE_ADMIN_COMMAND_PROPERTY_ITEM_PAIR_DELIMITER << statVariables.cpuUsageAsyncTasks.max << NODE_ADMIN_COMMAND_PROPERTY_ITEMS_DELIMITER

          << "memoryUsageAsyncTasks.min" << NODE_ADMIN_COMMAND_PROPERTY_ITEM_PAIR_DELIMITER << statVariables.memoryUsageAsyncTasks.min << NODE_ADMIN_COMMAND_PROPERTY_ITEMS_DELIMITER
          << "memoryUsageAsyncTasks.avg" << NODE_ADMIN_COMMAND_PROPERTY_ITEM_PAIR_DELIMITER << statVariables.memoryUsageAsyncTasks.avg << NODE_ADMIN_COMMAND_PROPERTY_ITEMS_DELIMITER
          << "memoryUsageAsyncTasks.max" << NODE_ADMIN_COMMAND_PROPERTY_ITEM_PAIR_DELIMITER << statVariables.memoryUsageAsyncTasks.max << NODE_ADMIN_COMMAND_PROPERTY_ITEMS_DELIMITER

          << "countSyncTasksForMinute.min" << NODE_ADMIN_COMMAND_PROPERTY_ITEM_PAIR_DELIMITER << statVariables.countSyncTasksForMinute.min << NODE_ADMIN_COMMAND_PROPERTY_ITEMS_DELIMITER
          << "countSyncTasksForMinute.avg" << NODE_ADMIN_COMMAND_PROPERTY_ITEM_PAIR_DELIMITER << statVariables.countSyncTasksForMinute.avg << NODE_ADMIN_COMMAND_PROPERTY_ITEMS_DELIMITER
          << "countSyncTasksForMinute.max" << NODE_ADMIN_COMMAND_PROPERTY_ITEM_PAIR_DELIMITER << statVariables.countSyncTasksForMinute.max << NODE_ADMIN_COMMAND_PROPERTY_ITEMS_DELIMITER

          << "countAsyncTasksForMinute.min" << NODE_ADMIN_COMMAND_PROPERTY_ITEM_PAIR_DELIMITER << statVariables.countAsyncTasksForMinute.min << NODE_ADMIN_COMMAND_PROPERTY_ITEMS_DELIMITER
          << "countAsyncTasksForMinute.avg" << NODE_ADMIN_COMMAND_PROPERTY_ITEM_PAIR_DELIMITER << statVariables.countAsyncTasksForMinute.avg << NODE_ADMIN_COMMAND_PROPERTY_ITEMS_DELIMITER
          << "countAsyncTasksForMinute.max" << NODE_ADMIN_COMMAND_PROPERTY_ITEM_PAIR_DELIMITER << statVariables.countAsyncTasksForMinute.max << NODE_ADMIN_COMMAND_PROPERTY_ITEMS_DELIMITER

          << "countSyncTasks" << NODE_ADMIN_COMMAND_PROPERTY_ITEM_PAIR_DELIMITER << statVariables.countSyncTasks << NODE_ADMIN_COMMAND_PROPERTY_ITEMS_DELIMITER
          << "countAsyncTasks" << NODE_ADMIN_COMMAND_PROPERTY_ITEM_PAIR_DELIMITER << statVariables.countAsyncTasks << NODE_ADMIN_COMMAND_PROPERTY_ITEMS_DELIMITER
          << "countSyncTasksFail" << NODE_ADMIN_COMMAND_PROPERTY_ITEM_PAIR_DELIMITER << statVariables.countSyncTasksFail << NODE_ADMIN_COMMAND_PROPERTY_ITEMS_DELIMITER
          << "countAsyncTasksFail" << NODE_ADMIN_COMMAND_PROPERTY_ITEM_PAIR_DELIMITER << statVariables.countAsyncTasksFail;

      return ret.str();
    }

    std::string DataProcessorData::processAdminCommandsSphinx(std::string & command){
      std::string ret = NODE_FUNCTIONAL_ERROR_SPHINX_ADMIN_JSON_MSG;

      logger.log(NODE_LOG_MODE_DEBUG) << NODE_MSG_SPHINX << " admin command requested [" << command << "]" << flush;

      //!<Create input object for Sphinx FO
      HCE::sphinx::SphinxInputJsonMessage inputMsg(command);
      if(!inputMsg.isError()){
        if(_sphinxFO->getErrorCode()){
          logger.log(NODE_LOG_MODE_ERROR) << NODE_FUNCTIONAL_ERROR_SPHINX_ADMIN << " before process indexer state error "
              << _sphinxFO->getErrorCode() << " : " << _sphinxFO->getErrorMsg() << flush;
          //!<Reset searcher error state to get next message to start from scratch
          _sphinxFO->resetError();
        }else{
          //!<Execute admin request
          HCE::sphinx::SphinxOutputJsonMessage outputMessage = _sphinxFO->Process(inputMsg);
          //Sphinx FO error
          if(_sphinxFO->getErrorCode()){
            //!<Log some error of Sphinx FO
            logger.log(NODE_LOG_MODE_ERROR) << NODE_FUNCTIONAL_ERROR_SPHINX_ADMIN << " after process indexer state error " << _sphinxFO->getErrorCode() << " : " << _sphinxFO->getErrorMsg()
                << endl << NODE_FUNCTIONAL_ERROR_SPHINX_INTERNAL_LOG_BEGIN << _sphinxFO->logMsg() << NODE_FUNCTIONAL_ERROR_SPHINX_INTERNAL_LOG_END << flush;
            //!<Reset indexer error state to get next message to start from scratch
            _sphinxFO->resetError();
          }
          //Output message error
          if(outputMessage.getErrorCode()){
            //!<Log some error of indexer request execution
            logger.log(NODE_LOG_MODE_ERROR) << NODE_FUNCTIONAL_ERROR_SPHINX_ADMIN << " output message error " << outputMessage.getErrorCode() << " : " << outputMessage.getErrorMessage() << flush;
          }
          //Get json message from output object
          if(!outputMessage.serialize(ret)){
            //!<Log some error of indexer request execution
            logger.log(NODE_LOG_MODE_ERROR) << NODE_FUNCTIONAL_ERROR_SPHINX_ADMIN << " output message serialization error " << outputMessage.getErrorCode() << " : "
                << outputMessage.getErrorMessage() << flush;

            //!<Return default response with general error
            ret = NODE_FUNCTIONAL_ERROR_SPHINX_ADMIN_JSON_MSG;
          }
          //Register Sphinx request time
          _statSphinxTimeTotalAdmin += outputMessage.getTime();
        }
      }else{
        logger.log(NODE_LOG_MODE_ERROR) << NODE_FUNCTIONAL_ERROR_SPHINX_ADMIN << " deserialization error " << inputMsg.getErrorCode() << " : " << inputMsg.getErrorMsg() << flush;
      }

      //Register Sphinx admin requests number stat
      _statSphinxRequestsAdmin++;

      return ret;
    }

    void* DataProcessorData::process(void){
      if(_inprocDataSock && _inprocAdminSock){
        logger.log(NODE_LOG_MODE_INFORMATION) << NODE_ADMIN_ERROR_MAIN_MSG_LOOP_STARTED << flush;

        //!<Enter main loop of processing of message
        while(getInProgress()){
          try{
            //!<Init polling pool
            zmq::pollitem_t items[] = {{*_inprocDataSock, 0, ZMQ_POLLIN, 0}, {*_inprocAdminSock, 0, ZMQ_POLLIN, 0}};

            //!<Poll inproc_data server
            zmq::poll(items, 2, getPollTimeout());

            //!<Handle income data message
            if(items[0].revents & ZMQ_POLLIN){
              handleInternalMessage();
            }

            //!<Handle inproc_admin's socket message
            if(items[1].revents & ZMQ_POLLIN){
              handleAdminMessage();
            }

            //!<Send periodical property values
            sendPeriodicProperties();

          }catch(zmq::error_t& error){
            std::cout << "DataProcessorData E: " << error.what() << std::endl;
          }catch(std::exception& e){
            std::cout << "DataProcessorData E: " << e.what() << std::endl;
          }catch(...){
            std::cout << "DataProcessorData E: Unknown exception" << std::endl;
          }
        }

        logger.log(NODE_LOG_MODE_INFORMATION) << NODE_ADMIN_ERROR_MAIN_MSG_LOOP_STOPPED << flush;
      }else{
        logger.log(NODE_LOG_MODE_FATAL) << "DataProcessorData::process() aborted because not all required sockets created : " << NODE_CRITICAL_ERROR_MESSAGE << flush;
        ::exit(NODE_EXIT_CODE_GENERAL_CRITICAL_ERROR);
      }

      return NULL;
    }

  }
}
