#include "DataReducerProxy.hpp"

namespace HCE {
  namespace handlers {

    DataReducerProxy::DataReducerProxy(std::string name, zmq::context_t& ctx, unsigned char nodeMode, int64_t startedAt,
        unsigned char logPriority, const std::string& iniFile) :
        Handler(name, ctx, std::string(""), startedAt, logPriority, iniFile),
        _inprocReducerInSock(NULL), _inprocReducerOutSock(NULL), _statRequests(0), _statResponses(0),
        _clientsNumber(0), _nodeMode(nodeMode), _reducerTtlTasksCleanupAt(TTL_TASKS_CLEANUP_INTERVAL_DEFAULT), _extendedReducerFO(NULL){

      initialize();
    }

    DataReducerProxy::~DataReducerProxy(){
      deinitialize();
    }

    int DataReducerProxy::initialize(void){
      logger.log(NODE_LOG_MODE_INFORMATION) << "Initialize..." << flush;

      //!<%Reducer FO initialization
      reducerInit();

      _statRequests = 0;
      _statResponses = 0;
      _clientsNumber = 0;

      //!<!<Rebuild admin client socket
      rebuildAdminClientSocket();

      //TODO: must be replaced with parent methods for client and server sockets rebuild
      //!<Create inproc_reducer_in socket
      _inprocReducerInSock = new zmq::socket_t(_context, ZMQ_DEALER);
      //!<Bind to inproc_reducer_in socket
      _inprocReducerInSock->bind(NODE_CS_INPROC_REDUCER_IN.c_str());
      logger.log(NODE_LOG_MODE_DEBUG) << "Bound to inproc_reducer_in" << flush;

      //!<Create inproc_reducer_out socket
      _inprocReducerOutSock = new zmq::socket_t(_context, ZMQ_DEALER);
      //!<Bind to inproc_reducer_out socket
      _inprocReducerOutSock->bind(NODE_CS_INPROC_REDUCER_OUT.c_str());
      logger.log(NODE_LOG_MODE_DEBUG) << "Bound to inproc_reducer_out" << flush;

      return true;
    }

    void DataReducerProxy::deinitialize(void){
      if(_inprocReducerInSock){
        if(_inprocReducerInSock->connected()){
          _inprocReducerInSock->close();
        }
        delete _inprocReducerInSock;
        _inprocReducerInSock = NULL;
      }
      if(_inprocReducerOutSock){
        if(_inprocReducerOutSock->connected()){
          _inprocReducerOutSock->close();
        }
        delete _inprocReducerOutSock;
        _inprocReducerOutSock = NULL;
      }

      _responses.clear();

      if(_extendedReducerFO){
        delete _extendedReducerFO;
        _extendedReducerFO = NULL;
      }
    }

    void DataReducerProxy::reducerInit(){
      //!<Create _extendedReducerFO instance
      _extendedReducerFO = new HCE::reduce::ReducerFunctionalObject();

      if(_extendedReducerFO){
        logger.log(NODE_LOG_MODE_INFORMATION) << " Sphinx reducer FO instance created." << flush;
      }else{
        logger.log(NODE_LOG_MODE_FATAL) << NODE_FUNCTIONAL_ERROR_SPHINX_REDUCER << " FO instantiate : " << NODE_CRITICAL_ERROR_MESSAGE << flush;
        ::exit(NODE_EXIT_CODE_SPHINX_CRITICAL_ERROR);
      }
    }

    void DataReducerProxy::defaultReducerResponseItemCreate(std::string& messageId, std::string& messageBody){
      DefaultReducerResponseItem* ri = new DefaultReducerResponseItem;
      if(ri){
        //!<Accumulate message body
        ri->bodies = messageBody;
        //!<Set messages counter
        ri->count = 1;
        //!<Insert new item in to the responses map
        _responses[messageId] = ri;
      }else{
        logger.log(NODE_LOG_MODE_ERROR) << "ERROR create new response item" << flush;
      }
    }

    void DataReducerProxy::defaultReducerResponseItemAccumulate(std::map<std::string, DefaultReducerResponseItem*>::iterator& DefaultReducerResponseItemIterator, std::string& messageBody){
      //!<Request found, accumulate next message, inc counter
      DefaultReducerResponseItemIterator->second->bodies += " " + messageBody;
      DefaultReducerResponseItemIterator->second->count++;
    }

    std::string DataReducerProxy::defaultReducerResponseItemReduce(std::map<std::string, DefaultReducerResponseItem*>::iterator& DefaultReducerResponseItemIterator){
      //!<Convert string buffer to vector of "unsigned long long"
      std::stringstream ss(DefaultReducerResponseItemIterator->second->bodies);
      std::vector<unsigned long long> items((std::istream_iterator<unsigned long long>(ss)), std::istream_iterator<unsigned long long>());
      //!<Sort vector
      sort(items.begin(), items.end(), std::less<unsigned long long>());
      //!<Remove duplicated items
      items.erase(std::unique(items.begin(), items.end()), items.end());
      //!<Convert from vector to string
      std::ostringstream oss;
      std::copy(items.begin(), items.end() - 1, std::ostream_iterator<unsigned long long>(oss, " "));
      oss << items.back();
      std::string body(oss.str());

      return body;
    }

    void DataReducerProxy::defaultReducerReducePlainText(std::string& messageId, std::string& messageBody){
      //!<Init output message with Id
      zmsg msg(messageId.c_str());

      //!<Check in queries messages accumulator map
      std::map<std::string, DefaultReducerResponseItem*>::iterator DefaultReducerResponseItemIterator = _responses.find(messageId);
      if(DefaultReducerResponseItemIterator == _responses.end()){
        //!<Request not found, create new responses item
        defaultReducerResponseItemCreate(messageId, messageBody);
      }else{
        //!<Accumulate message
        defaultReducerResponseItemAccumulate(DefaultReducerResponseItemIterator, messageBody);
        //!<If accumulated messages number is equal to clients number - reduce task
        if(DefaultReducerResponseItemIterator->second->count == _clientsNumber){
          //!<Reduce all accumulated messages and get resulted context
          std::string body = defaultReducerResponseItemReduce(DefaultReducerResponseItemIterator);

          //!<Add message body
          msg.append(body.c_str());

          logger.log(NODE_LOG_MODE_TRACE) << "MSG TO inproc_reducer_out" << endl << msg.dump() << flush;

          //!<Send response message to server
          msg.send(*_inprocReducerOutSock);

          _statResponses++;
          logger.log(NODE_LOG_MODE_DEBUG) << "RESPONSE from plain text reducer to inproc_reducer_out socket for upper server sent, message [" << messageId << "], size:" << body.size()
              << " bytes, accumulated:" << DefaultReducerResponseItemIterator->second->count << ", n=" << _statResponses << flush;

          //!<Delete responses item
          delete DefaultReducerResponseItemIterator->second;

          //!<Delete responses map item
          _responses.erase(DefaultReducerResponseItemIterator);
        }
      }
    }

    void DataReducerProxy::reduceJson(std::string& messageId, std::string& messageBody){
      //Generate itemId on the basis of messageId string
      Poco::Checksum checksum(Poco::Checksum::TYPE_CRC32);
      checksum.update(messageId);
      unsigned long long itemId = checksum.checksum();

      logger.log(NODE_LOG_MODE_DEBUG) << "MERGE_JSON2: [" << messageBody << "], messageId=" << messageId << ", clietns: " << _clientsNumber << ", itemId: " << itemId << flush;

      _extendedReducerFO->accumulate(messageBody, itemId);
      if(_extendedReducerFO->getErrCode() != HCE::reduce::OK){
        logger.log(NODE_LOG_MODE_ERROR) << NODE_FUNCTIONAL_ERROR_SPHINX_REDUCER << " accumulate message error " << _extendedReducerFO->getErrCode() << " : " << _extendedReducerFO->getErrMsg() << flush;
      }

      if(_extendedReducerFO->isCompleteTask(itemId, (int)_clientsNumber) == true){
        Poco::SharedPtr<HCE::reduce_types::ReducingOutputMessage> reducingOutputMessage = _extendedReducerFO->reduce(itemId);

        logger.log(NODE_LOG_MODE_DEBUG) << "Reducer output json [" << reducingOutputMessage->getBody() << "], type: " << static_cast<int>(reducingOutputMessage->getProcessingType()) << flush;

        //Response message body json string
        std::string reducerResultJson;
        HCE::sphinx::SphinxDefaultJSON defaultJson;

        if(reducingOutputMessage->getProcessingErrorCode()){
          //Error of reducer processing, return default empty result
          logger.log(NODE_LOG_MODE_DEBUG) << NODE_FUNCTIONAL_ERROR_SPHINX_REDUCER << " reduce task error " << reducingOutputMessage->getProcessingErrorCode() << " : "
              << reducingOutputMessage->getProcessingErrorMsg() << flush;
          //Create empty Sphinx result object to return it for farther processing
          reducerResultJson = defaultJson.getJSON();
        }else{
          //Get body to return to upper level reducer
          reducerResultJson = reducingOutputMessage->getBody();
        }

        //Create message cover object to return it to upper level processing (reducing)
        HCE::JsonMessageCover msgCover(reducingOutputMessage->getProcessingType(), reducerResultJson, reducingOutputMessage->getTTL());
        if(!msgCover.serialize(reducerResultJson)){
          //!<If serialization error return default result to reducer
          logger.log(NODE_LOG_MODE_ERROR) << NODE_FUNCTIONAL_ERROR_SPHINX << " cover searialization error " << msgCover.getErrorCode() << " : " << msgCover.getErrorMsg() << flush;

          //Create empty Sphinx result object to return it for farther processing
          reducerResultJson = defaultJson.getJSON();
        }

        logger.log(NODE_LOG_MODE_DEBUG) << "Reducer covered json : " << reducerResultJson << flush;

        //!<Init output message with Id
        zmsg msg(messageId.c_str());
        //!<Add message body
        msg.append(reducerResultJson.c_str());

        logger.log(NODE_LOG_MODE_TRACE) << "MSG TO inproc_reducer_out" << endl << msg.dump() << flush;

        //!<Send response message to server
        msg.send(*_inprocReducerOutSock);

        _statResponses++;

        logger.log(NODE_LOG_MODE_DEBUG) << "RESPONSE from json reducer to inproc_reducer_out socket for upper server sent, message [" << messageId << "], size:" << reducerResultJson.size()
              << " bytes, items accumulated:" << /*DefaultReducerResponseItemIterator->second->count <<*/", time:" << reducingOutputMessage->getProcessingElapsedTime()
              << ", n=" << _statResponses << flush;
      }else{
        //TODO: possible call for garbage collector or log some timing stats when not all messages are accumulated for the reduce task
        logger.log(NODE_LOG_MODE_DEBUG) << "Reducer task is not complete : messageId=" << messageId << ", clietns: " << _clientsNumber << ", itemId: " << itemId << "\n"
            << endl << NODE_FUNCTIONAL_ERROR_SPHINX_REDUCER << " error " << _extendedReducerFO->getErrCode() << " : " << _extendedReducerFO->getErrMsg() << flush;
      }
    }

    void DataReducerProxy::handleInternalMessage(void){
      //!<Fetch message from inproc_reducer_in
      zmsg msg(*_inprocReducerInSock);

      logger.log(NODE_LOG_MODE_TRACE) << "MSG FROM inproc_reducer_in" << endl << msg.dump() << flush;

      _statRequests++;

      //!<Fetch fields from message
      std::string messageId((const char*)msg.pop_front().c_str());
      std::string messageBody((const char*)msg.pop_front().c_str());
      std::string messageRoute((const char*)msg.pop_front().c_str());
      std::istringstream((const char*)msg.pop_front().c_str()) >> _clientsNumber;

      logger.log(NODE_LOG_MODE_DEBUG) << "REQUEST from inproc_reducer_in, nodeMode=" << static_cast<unsigned int>(_nodeMode) << ", message [" << messageId
          << "], size:" << messageBody.size() << ", clients:" << _clientsNumber << " , n=" << _statRequests << flush;

      //!<If node in SHARD mode (send request to all data nodes) and if clients number > 1 - use Reducer to collect messages and reduce results
      if((_clientsNumber > 1) && (_nodeMode == NODE_MODE_PROXY || _nodeMode == NODE_MODE_ROUTER)){
        if(messageBody.size() > 0 && messageBody[0] == NODE_JSON_BEGIN_CHAR){
          reduceJson(messageId, messageBody);
        }else{
          defaultReducerReducePlainText(messageId, messageBody);
        }
      }else{
        //!<If node "Single client" or REPLICA mode (send request to one data node roud-robin way) - just re-send message from one data node to upper level w/o Reducer usage
        //!<Add fields back to message
        msg.append(messageId.c_str());
        msg.append(messageBody.c_str());
        msg.append(messageRoute.c_str());//??
        _statResponses++;

        logger.log(NODE_LOG_MODE_DEBUG) << "RESPONSE to inproc_reducer_out socket from proxy node redirector to upper server sent, message [" << messageId << "], size:" << messageBody.size()
            << ", n=" << _statResponses << flush;
        logger.log(NODE_LOG_MODE_TRACE) << "MSG TO inproc_reducer_out" << endl << msg.dump() << flush;

        //!<Send response message to server
        msg.send(*_inprocReducerOutSock);
      }
    }

    std::string DataReducerProxy::processAdminCommands(std::string & command, std::string & parameters){
      std::string ret = NODE_ADMIN_ERROR_OK;

      logger.log(NODE_LOG_MODE_DEBUG) << "Admin command requested [" << command << "], parameters [" << parameters << "]" << flush;

      //!<Process commands
      if(command == NODE_MSG_STAT){
        std::stringstream result;
        result << NODE_ADMIN_ERROR_OK << NODE_ADMIN_COMMAND_DELIMITER << "requests=" << _statRequests << NODE_ADMIN_COMMAND_DELIMITER << "responses=" << _statResponses
            << NODE_ADMIN_COMMAND_DELIMITER << "clientsNumber=" << _clientsNumber;
        ret = result.str();
      }else if(command == NODE_MSG_PROPERTIES){
        //!<Make property information
        ret = getPropertyInformation();
/*
      }else if(command == NODE_MSG_GET_DATA_PROCESSING_MODE){
        std::stringstream result;

        result << NODE_ADMIN_ERROR_OK << NODE_ADMIN_COMMAND_DELIMITER << std::to_string(static_cast<unsigned int>(_nodeMode));

        ret = result.str();
      }else if(command == NODE_MSG_SET_DATA_PROCESSING_MODE){
        std::stringstream result;

        //!<Parse command parameters and fetch processing mode value
        unsigned int processingMode;
        std::istringstream(parameters) >> processingMode;

        result << NODE_ADMIN_ERROR_OK << NODE_ADMIN_COMMAND_DELIMITER << std::to_string(static_cast<unsigned int>(_nodeMode));

        //Change data processing mode
        _nodeMode = static_cast<unsigned char>(processingMode);

        ret = result.str();
*/
      }else if(command == NODE_MSG_MANAGER_MODE_GET){
        ret = NODE_ADMIN_ERROR_OK + NODE_ADMIN_COMMAND_DELIMITER + NODE_ADMIN_PROPERTY_NAME_MMODE + "=" + std::to_string(static_cast<unsigned int>(_nodeMode));
      }else if(command == NODE_MSG_MANAGER_MODE_SET){
        //!<Parse command parameters and fetch log level value
        int nodeMode = _nodeMode;
        //Parse log priority from parameters
        std::istringstream(parameters) >> nodeMode;
        //Make response
        ret = NODE_ADMIN_ERROR_OK + NODE_ADMIN_COMMAND_DELIMITER + NODE_ADMIN_PROPERTY_NAME_MMODE + "=" + std::to_string(static_cast<unsigned int>(_nodeMode));
        //Set new log priority
        _nodeMode = nodeMode;
      }else{
        //!<Unsupported command
        ret = NODE_ADMIN_ERROR_ERROR + NODE_ADMIN_COMMAND_DELIMITER + NODE_ADMIN_ERROR_UNSUPPORTED_COMMAND;
      }

      return ret;
    }

    void DataReducerProxy::reducerTasksCleanupByTtl(void){
      if(s_clock() > _reducerTtlTasksCleanupAt){

        unsigned int tasks = _extendedReducerFO->cleanupExpiredTasksByTTL(TTL_TASKS_CLEANUP_NUMBER_DEFAULT);
        unsigned int tasksItems = _extendedReducerFO->cleanupExpiredTasksByTTLQueue(TTL_TASKS_QUEUE_CLEANUP_NUMBER_DEFAULT, TTL_TASKS_QUEUE_CLEANUP_TIME);

        if(tasks > 0 || tasksItems){
          logger.log(NODE_LOG_MODE_TRACE) << "TTL terminated tasks and queue cleaned, " << tasks << " tasks, " << tasksItems << " queue items removed" << flush;
        }

        //!<Reset cleanup time
        _reducerTtlTasksCleanupAt = s_clock() + TTL_TASKS_CLEANUP_INTERVAL_DEFAULT;
      }
    }

    //!<Get property information
    std::string DataReducerProxy::getPropertyInformation(void){
      std::stringstream result;
      result << NODE_ADMIN_ERROR_OK
          << NODE_ADMIN_COMMAND_DELIMITER << getProperties() << "requests=" << _statRequests
          << NODE_ADMIN_COMMAND_DELIMITER << "responses=" << _statResponses
          << NODE_ADMIN_COMMAND_DELIMITER << "clientsNumber=" << _clientsNumber
          << NODE_ADMIN_COMMAND_DELIMITER << "nodeMode=" << static_cast<unsigned int>(_nodeMode)
          << NODE_ADMIN_COMMAND_DELIMITER << "reducerTasks=" << _extendedReducerFO->getTasksNumber()
          << NODE_ADMIN_COMMAND_DELIMITER << "reducerRejectedMessages=" << _extendedReducerFO->getRejectedMessages();
      return result.str();
    }

    void* DataReducerProxy::process(void){
      if(_inprocReducerInSock){
        logger.log(NODE_LOG_MODE_DEBUG) << NODE_ADMIN_ERROR_MAIN_MSG_LOOP_STARTED << flush;

        //!<Enter main loop of processing of message
        while(getInProgress()){
          try {
            //!<Init polling pool
            zmq::pollitem_t items[] = {{*_inprocReducerInSock, 0, ZMQ_POLLIN, 0}, {*_inprocAdminSock, 0, ZMQ_POLLIN, 0}};

            //!<Poll dataclient
            zmq::poll(items, 2, getPollTimeout());

            //!<Handle message
            if(items[0].revents & ZMQ_POLLIN){
              handleInternalMessage();
            }

            //!<Handle inproc_admin's socket message
            if(items[1].revents & ZMQ_POLLIN){
              handleAdminMessage();
            }

            //!<Cleanup TTL terminated tasks and task's queue
            reducerTasksCleanupByTtl();

            //!<Send periodical property values
            sendPeriodicProperties();

          }catch(zmq::error_t& error){
            std::cout << "DataReducerProxy E: " << error.what() << std::endl;
          }catch(std::exception& e){
            std::cout << "DataReducerProxy E: " << e.what() << std::endl;
          }catch(...){
            std::cout << "DataReducerProxy E: Unknown exception" << std::endl;
          }
        }

        logger.log(NODE_LOG_MODE_INFORMATION) << NODE_ADMIN_ERROR_MAIN_MSG_LOOP_STOPPED << flush;
      }else{
        logger.log(NODE_LOG_MODE_FATAL) << "DataReducerProxy::process() aborted because not all required sockets created : " << NODE_CRITICAL_ERROR_MESSAGE << flush;
        ::exit(NODE_EXIT_CODE_GENERAL_CRITICAL_ERROR);
      }

      return NULL;
    }

  }
}
