/**
 * @file DataReducerProxy.hpp
 *
 * HCE project DataReducerProxy handler.\n
 * Implements functionality for data reducer handler.\n
 * Implements data reducing functionality of responses of proxy nodes.\n
 * Container of data reducer functional object that reduces upstream data flows according with end-user functionality and supported data storage engines like Sphinx.
 *
 * @author bgv <bgv.hce@gmail.com>
 * @link http://hierarchical-cluster-engine.com/
 * @copyright Copyright &copy; 2013 IOIX Ukraine
 * @license http://hierarchical-cluster-engine.com/license/
 * @package HCE project node API
 * @since 0.1
 */

#ifndef __HCE_DATAREDUCER_PROXY_INCLUDED__
#define __HCE_DATAREDUCER_PROXY_INCLUDED__

#include <sstream>
#include <map>
#include <iterator>
#include <Poco/Checksum.h>

#include "Handler.hpp"

//Sphinx functional object dependencies
#include "SphinxInputJsonMessage.hpp"
#include "SphinxOutputJsonMessage.hpp"
#include "SphinxSearcher.hpp"
#include "SphinxError.hpp"
//Reducer functional object dependencies
#include "ReducerFunctionalObject.hpp"
//Message cover object
#include "JsonMessageCover.hpp"

namespace HCE {
  namespace handlers {

    class DataReducerProxy: public HCE::handlers::Handler {
      public:
        const int TTL_TASKS_CLEANUP_INTERVAL_DEFAULT = 10000;
        const int TTL_TASKS_CLEANUP_NUMBER_DEFAULT = 100;
        const int TTL_TASKS_QUEUE_CLEANUP_NUMBER_DEFAULT = 1000;
        const int TTL_TASKS_QUEUE_CLEANUP_TIME = 200;

        DataReducerProxy(std::string name, zmq::context_t& ctx, unsigned char nodeMode, int64_t startedAt, unsigned char logPriority, const std::string& iniFile);
        ~DataReducerProxy();

      protected:
        int initialize(void);
        void deinitialize(void);
        void* process(void);

        void handleInternalMessage(void);
        std::string processAdminCommands(std::string& command, std::string& parameters);

        //!<%Reducer FO initialization
        void reducerInit(void);

        //!<Plain text reducer part. Creates response item to accumulate all responses from registered replica nodes.
        void defaultReducerResponseItemCreate(std::string& messageId, std::string& messageBody);
        //!<Plain text reducer part. Accumulates response messages from all registered replica nodes in corresponded response item according with message Id.
        static void defaultReducerResponseItemAccumulate(std::map<std::string, DefaultReducerResponseItem*>::iterator& DefaultReducerResponseItemIterator, std::string& messageBody);
        //!<Plain text reducer part. Reduces data accumulated. Removes duplicates and sorts.
        std::string defaultReducerResponseItemReduce(std::map<std::string, DefaultReducerResponseItem*>::iterator& DefaultReducerResponseItemIterator);
        //!<Plain text messages reducer. Process plain text messages for reduce task
        void defaultReducerReducePlainText(std::string& messageId, std::string& messageBody);
        //!<Structured messages reducer. Process Json-based messages like Sphinx search results for reduce task
        void defaultReducerReduceJson(std::string& messageId, std::string& messageBody);

        //!<Structured messages reducer involved extended reducer FO. Process Json-based messages like Sphinx search results for reduce task
        void reduceJson(std::string& messageId, std::string& messageBody);
        //!<TTL tasks queue cleanup for reduceJson
        void reducerTasksCleanupByTtl(void);

        //!<Get property information
        std::string getPropertyInformation(void);

        //!<ZMQ sockets
        zmq::socket_t* _inprocReducerInSock;
        zmq::socket_t* _inprocReducerOutSock;

        //!<Statistical variables
        unsigned int _statRequests;
        unsigned int _statResponses;

        //!<Number of connected and registered clients
        unsigned int _clientsNumber;

        //!<Node mode: router, shard manager, replica manager or replica.
        unsigned char _nodeMode;

        //!<Time when ttl terminated tasks queue will be cleaned partially
        int64_t _reducerTtlTasksCleanupAt;

        //Sphinx reducer FO
        HCE::reduce::ReducerFunctionalObject* _extendedReducerFO;

        //TODO: only for model purposes, need to be removed in final version
        //!<Accumulator map for plain text messages
        std::map<std::string, DefaultReducerResponseItem*> _responses;
    };

  }
}

#endif /* __HCE_DATAREDUCER_PROXY_INCLUDED__ */
