#include <stdlib.h>
#include <time.h>
#include <random>

#include "DataServerProxy.hpp"
#include "ServiceMessages.hpp"
//DRCE FO
#include "DRCEFunctionalObject.hpp"
#include "DRCEMessageConst.hpp"
#include "DRCEError.hpp"
//Sphinx FO
#include "SphinxFunctionalObject.hpp"
#include "SphinxMessageConst.hpp"
#include "SphinxError.hpp"
//Message cover object
#include "JsonMessageCover.hpp"
//Resource extractor
#include "DRCEResourceExtractor.hpp"
//Resource usage checker
#include "ResourceUsageLimitsChecker.hpp"

namespace HCE {
  namespace handlers {

    DataServerProxy::DataServerProxy(std::string name, zmq::context_t& ctx, const std::string& clientIdentity,
        std::string connectionString, unsigned char nodeMode, int64_t startedAt,
        unsigned char logPriority, const std::string& iniFile) :
        Handler(name, ctx, connectionString, startedAt, logPriority, iniFile), DataNodeOptions(),
        _clientIdentity(clientIdentity), _nodeMode(nodeMode), _previousClientNumber(0), _iniFile(iniFile),
        calculatorAlgorithmData(), resourceUsageLimitsData(), typeCalculatorAlgorithm(CalculatorAlgorithm::Types::caDefault),
        _resourcesUsageCollectedSize(HCE::handlers::RESOURCES_USAGE_COLLECTED_SIZE),
        _resourcesUsageMaxAllowedRange(HCE::handlers::RESOURCES_USAGE_MAX_ALLOWED_RANGE),
        _resourceUsageManager(logger){

      //!<Init all the rest
      initialize();
    }

    DataServerProxy::~DataServerProxy(){
      deinitialize();
    }

    int DataServerProxy::initialize(void){
      logger.log(NODE_LOG_MODE_INFORMATION) << "Initialize client [" << _clientIdentity << "] with connection [" << _connectionString << "]" << flush;

      //!<Init sockets
      _inprocProxySock = NULL;
      _inprocReducerInSock = NULL;
      _dataServerSock = NULL;

      //!<Node init from file
      initNode(_iniFile, logger);

      //!<Heartbeat at time mark
      _heartbeatAt = s_clock() + getHeartbeatDelay();

      //!<Send out periodical stats at regular intervals
      _pstatsAt = s_clock() + STAT_INTERVAL;

      //!<Stat fields
      _statClientRequests = 0;
      _statClientResponses = 0;
      _statRequests = 0;
      _statClientRequestsSize = 0;
      _statClientResponsesSize = 0;
      _statRequestsSize = 0;
      _periodicStatClientRequests = 0;
      _periodicStatClientResponses = 0;
      _statRequestsProcessed = 0;
      _statRequestsProcessedSize = 0;
      _periodicStatClientRequestsSize = 0;
      _periodicStatClientResponsesSize = 0;
      _statQPSClient = 0;
      _statQPSServer = 0;
      _statRPSClient = 0;
      _statRPSServer = 0;
      _statRequestsBandwidthClient = 0;
      _statRequestsBandwidthServer = 0;
      _statResponsesBandwidthClient = 0;
      _statResponsesBandwidthServer = 0;

      //!<Node init
      nodeInit(_iniFile);

      _clientsManager = new ClientsQueueManager(&_clientsQueue, 0, &_resourceUsageManager);
      //!<Apply heartbeat timeout value to client manager
      _clientsManager->setHeartbeatTimeout(getHeartbeatTimeout());
      //!<Apply max resources usage collected size value to client manager
      _clientsManager->setMaxResourcesUsageCollectedSize(_resourcesUsageCollectedSize);

      //!<Rebuild admin client socket
      rebuildAdminClientSocket();

      //!<Rebuild inproc_proxy connection
      rebuildServerConnection(_inprocProxySock, NODE_CS_INPROC_PROXY, ZMQ_DEALER);
      //!<Rebuild inproc_reducer_in connection
      rebuildClientConnection(_inprocReducerInSock, NODE_CS_INPROC_REDUCER_IN, ZMQ_DEALER);
      //!<Rebuild external server connection
      rebuildServerConnection(_dataServerSock, _connectionString, ZMQ_ROUTER);

      return true;
    }

    void DataServerProxy::deinitialize(void){
      if(_inprocProxySock){
        delete _inprocProxySock;
      }
      if(_inprocReducerInSock){
        delete _inprocReducerInSock;
      }
      if(_dataServerSock){
        delete _dataServerSock;
      }
      _inprocProxySock = NULL;
      _inprocReducerInSock = NULL;
      _dataServerSock = NULL;
      _clientsQueue.clear();
      if(_clientsManager){
        delete _clientsManager;
      }
      HCE::drce::ResourceUsageExtractor::shutdown();
    }

    void DataServerProxy::handleInternalMessage(void){
      //!<Fetch message from inprocontrol
      zmsg msg(*_inprocProxySock);
      logger.log(NODE_LOG_MODE_TRACE) << "MSG FROM inproc_proxy" << endl << msg.dump() << flush;

      if (msg.parts() == NODE_MSG_CMD_MIN_FIELDS){

        char* addrp = msg.address();
        std::string addr ((addrp)?addrp:"");
        if (addr.compare(0, NODE_MSG_ROUTES.length(), NODE_MSG_ROUTES)==0){
          std::vector<ClientWorkerItem> clientQueue;
          RouteMessage routeMessage(clientQueue);
          routeMessage.extract(NODE_MSG_ROUTES, addr);

          _clientsManager->append(clientQueue);

          RouteMessage routeMsg(_clientsQueue);

          s_send(*_inprocReducerInSock, routeMsg.build(NODE_MSG_ROUTES));
        }
      }else{
        //!<Create identity, messageId and body of dataclient's message
        std::string clientIdentity((const char*)msg.pop_front().c_str());
        std::string messageId((const char*)msg.pop_front().c_str());
        std::string messageBody((const char*)msg.pop_front().c_str());
        std::string messageRoute((msg.parts())?(const char*)msg.pop_front().c_str():"");

        logger.log(NODE_LOG_MODE_DEBUG) << "RECEIVED REQUEST from upper server [" << clientIdentity << "], message [" << messageId << "] body " << messageBody.size()
            << " bytes, n=" << ++_statRequests << flush;

        if(_clientsQueue.empty()){
          //!<No clients connected, make response message and send response for reducer
          sendEmptyResponseMessage(msg, messageBody, EmptyResponseType::ertNoClients, messageId, messageRoute);
          logger.log(NODE_LOG_MODE_DEBUG) << "No client`s connection registered!" << flush;
          logger.log(NODE_LOG_MODE_TRACE) << msg.dump() << flush;
        }else{
          //!<Get allowed clients names
          std::vector<std::string> routeNames = std::forward<std::vector<std::string> >(RouteMessage::extractRouteNames(_clientsQueue));

          unsigned int nodeMode(_nodeMode);
          CalculatorAlgorithm::Types resourcesUsageAlgorithm = CalculatorAlgorithm::Types::caDefault;
          CalculatorAlgorithmData algorithmData(calculatorAlgorithmData);
          ResourceUsageLimitsData limitsData(resourceUsageLimitsData);

          if (!messageRoute.empty() && nodeMode != NODE_MODE_ROUTER){
            logger.log(NODE_LOG_MODE_TRACE) << "!!! DataServerProxy::handleInternalMessage  >>> (!messageRoute.empty() && nodeMode != NODE_MODE_ROUTER) <<<" << flush;

            //!<Extract route JSON from routeMessage
            std::string json(messageRoute);
            size_t pos = json.find_last_of('}');

            logger.log(NODE_LOG_MODE_TRACE) << "!!! DataServerProxy::handleInternalMessage  pos = " << pos << flush;

            if (pos!=std::string::npos){
              json = json.substr(0, pos+1);

              logger.log(NODE_LOG_MODE_TRACE) << "!!! DataServerProxy::handleInternalMessage  json(1): '" << json << "'" << flush;
            }else{
              json.clear();
            }
            //!<If JSON was successful extracted, update route names
            if (!json.empty()){
              logger.log(NODE_LOG_MODE_TRACE) << "!!! DataServerProxy::handleInternalMessage  json(2): '" << json << "'" << flush;

              //!<Get nodes list from route message
              RequestRouteMessage requestRouteMessage(json); //deserialize route json
              if (!requestRouteMessage.isError()){
                unsigned int managerRole = requestRouteMessage.getManagerRole();
                std::vector<std::string> identityNames = std::forward<std::vector<std::string> >(requestRouteMessage.getNodesNames());
                //!<Set current resource usage algorithm
                resourcesUsageAlgorithm = (static_cast<CalculatorAlgorithm::Types>(requestRouteMessage.getResourcesUsageAlgorithm()) != typeCalculatorAlgorithm)?
                    static_cast<CalculatorAlgorithm::Types>(requestRouteMessage.getResourcesUsageAlgorithm()):typeCalculatorAlgorithm;
                //!<Set current resource usage algorithm weights
                if (!requestRouteMessage.getResourcesUsageAlgorithmWeights().empty()){
                  HCE::handlers::DictionaryConverter<std::string, double> converter(requestRouteMessage.getResourcesUsageAlgorithmWeights());
                  algorithmData = converter.toObject().cast<HCE::handlers::CalculatorAlgorithmData>();
                }
                //!<Set current resource usage limits
                if (!requestRouteMessage.getResourcesUsageLimits().empty()){
                  HCE::handlers::DictionaryConverter<std::string, double> converter(requestRouteMessage.getResourcesUsageLimits());
                  limitsData = converter.toObject().cast<HCE::handlers::ResourceUsageLimitsData>();
                }
                //!<Check allowed data node mode
                if (managerRole == NODE_MODE_PROXY || managerRole == NODE_MODE_SHARD || managerRole == NODE_MODE_SHARD_RND || managerRole == NODE_MODE_SHARD_RU)
                  nodeMode = managerRole;

                logger.log(NODE_LOG_MODE_TRACE) << "!!! DataServerProxy::handleInternalMessage  routeNames.size() = " << routeNames.size() << flush;
                logger.log(NODE_LOG_MODE_TRACE) << "!!! DataServerProxy::handleInternalMessage  identityNames.size() = " << identityNames.size() << flush;

                if (!identityNames.empty()){
                  std::vector<std::string> routes;
                  for (const std::string& routeName:routeNames){
                    for (const std::string& identityName:identityNames){
                      logger.log(NODE_LOG_MODE_TRACE) << "!!! DataServerProxy::handleInternalMessage  routeName: " << routeName
                          << " identityName: " << identityName << " isEqual: " << std::boolalpha << (routeName==identityName) << flush;
                      if (routeName==identityName){
                        routes.push_back(identityName);
                      }
                    }
                  }
                  routeNames = routes;
                }
              }else{
                logger.log(NODE_LOG_MODE_DEBUG) << "!!! DataServerProxy::handleInternalMessage  ErrorMsg: " << requestRouteMessage.getErrorMsg() << flush;
                logger.log(NODE_LOG_MODE_DEBUG) << "!!! DataServerProxy::handleInternalMessage  messageRoute: " << messageRoute << flush;
              }
            }
          }

          logger.log(NODE_LOG_MODE_TRACE) << "!!! DataServerProxy::handleInternalMessage  _nodeMode: " << _nodeMode << flush;
          logger.log(NODE_LOG_MODE_TRACE) << "!!! DataServerProxy::handleInternalMessage   nodeMode: " << nodeMode << flush;
          logger.log(NODE_LOG_MODE_TRACE) << "!!! DataServerProxy::handleInternalMessage  routeNames.size() = " << routeNames.size() << flush;

          for (size_t i=0;i<routeNames.size();++i){
            logger.log(NODE_LOG_MODE_TRACE) << "!!! DataServerProxy::handleInternalMessage  routeNames[" << i << "] = " << routeNames[i] << flush;
          }

          if(routeNames.empty()){
          //!<No clients connected, make response message and send response for reducer
          sendEmptyResponseMessage(msg, messageBody, EmptyResponseType::ertNoClients, messageId, messageRoute);
          logger.log(NODE_LOG_MODE_DEBUG) << "No valid route names!" << flush;
          logger.log(NODE_LOG_MODE_TRACE) << msg.dump() << flush;

          //!<Some client(s) registered
          }else if(nodeMode == NODE_MODE_PROXY || nodeMode == NODE_MODE_ROUTER){

            std::stringstream routes;
            routes << messageRoute << ROUTE_LIST_COUNT_DELIMITER << _clientsQueue.size();
            //!<Re-send data request for all clients connected to dataserver
            for (size_t i=0;i<_clientsQueue.size();++i){
              //!<Create and send response message
              sendResponseMessage(_clientsQueue[i].identity, messageId, messageBody, routes.str());

              _statClientRequests++;
              _statClientRequestsSize += messageBody.size();

              logger.log(NODE_LOG_MODE_DEBUG) << "REQUEST SENT_ALL to dataserver socket for client [" << _clientsQueue[i].identity
                  << "], message [" << messageId << "] body " << messageBody.size() << ", n=" << _statClientRequests << flush;
            }
          }else{

            if(nodeMode == NODE_MODE_SHARD || nodeMode == NODE_MODE_SHARD_RND || nodeMode == NODE_MODE_SHARD_RU) {
              std::string identity;
              //!<Make local copy
              std::vector<ClientWorkerItem> clientsQueue;
              for (size_t i=0;i<routeNames.size();++i){
                for (size_t k=0;k<_clientsQueue.size();++k){
                  if (routeNames[i] == _clientsQueue[k].identity){
                    clientsQueue.push_back(ClientWorkerItem(_clientsQueue[k]));
                    break;
                  }
                }
              }

              if (clientsQueue.empty()){
                //!<No clients connected, make response message and send response for reducer
                sendEmptyResponseMessage(msg, messageBody, EmptyResponseType::ertNoClients, messageId, messageRoute);
                logger.log(NODE_LOG_MODE_DEBUG) << "No valid route names!" << flush;
                logger.log(NODE_LOG_MODE_TRACE) << msg.dump() << flush;
              }else{

                if (nodeMode == NODE_MODE_SHARD_RND){
                  //!<Get identity use random algorithm
                  identity = getIdentityRandom(_clientsQueue, _previousClientNumber);
                }else if(nodeMode == NODE_MODE_SHARD_RU){
                  //!<Get identity use resource usage algorithm
                  identity = getIdentityResourceUsage(clientsQueue, _previousClientNumber, resourcesUsageAlgorithm, algorithmData);
                  if (identity.empty()){
                    identity = getIdentityRandom(_clientsQueue, _previousClientNumber);
                  }
                  _clientsManager->incrementFrequency(identity);
                  logger.log(NODE_LOG_MODE_DEBUG) << "Choosen identity: '" << identity << "'" << flush;
                }else{
                  //!<Sort clients by frequency of queries done
                  std::sort(clientsQueue.begin(), clientsQueue.end(), ClientWorkerItem::sortByFrequencyPredicate);
                  //!<Inc freq of best (with minimal freq)
                  _clientsManager->incrementFrequency(clientsQueue[0].identity);
                  //!<Save identity for next use
                  if (!clientsQueue.empty())
                    identity = clientsQueue[0].identity;
                }


                if (identity.empty()){
                  //!<No clients connected, make response message and send response for reducer
                  sendEmptyResponseMessage(msg, messageBody, EmptyResponseType::ertNoClients, messageId, messageRoute);
                  logger.log(NODE_LOG_MODE_DEBUG) << "Empty client name cannot be use for routing balance!" << flush;
                  logger.log(NODE_LOG_MODE_TRACE) << msg.dump() << flush;
                }else{

                  logger.log(NODE_LOG_MODE_DEBUG) << "resourcesUsageAlgorithm = " << (unsigned int)resourcesUsageAlgorithm << flush;
                  logger.log(NODE_LOG_MODE_DEBUG) << "algorithmData = " << HCE::handlers::DictionaryConverter<std::string, double>(algorithmData).toString() << flush;
                  logger.log(NODE_LOG_MODE_DEBUG) << "limitsData = " << HCE::handlers::DictionaryConverter<std::string, double>(limitsData).toString() << flush;

                  ResourceUsageCollection resourceUsageCollection = _clientsManager->getResources(identity);

                  HCE::handlers::DictionaryConverter<std::string, double> converter(_resourceUsageManager.calculateAverage(identity, resourceUsageCollection));
                  HCE::handlers::CalculatorNodeResourceData resourceData = converter.toObject().cast<HCE::handlers::CalculatorNodeResourceData>(identity);

                  std::string errorMsg;
                  HCE::handlers::ResourceUsageLimitsChecker resourceUsageLimitsChecker(limitsData);
                  if (!resourceUsageLimitsChecker.safeCheck(resourceData, errorMsg)) {
                    //!<Node has not allowed resource usage limits
                    sendEmptyResponseMessage(msg, messageBody, EmptyResponseType::ertNotEnoughResourceLimits, messageId, messageRoute, errorMsg);
                    logger.log(NODE_LOG_MODE_TRACE) << "Node '" << identity << "' has not allowed resources usage limits, message skipped" << endl << msg.dump() << flush;
                  }else{
                    std::stringstream routes;
                    routes << messageRoute << ROUTE_LIST_COUNT_DELIMITER << 1;
                    //!<Create and send response message
                    sendResponseMessage(identity, messageId, messageBody, routes.str());

                    _statClientRequests++;
                    _statClientRequestsSize += messageBody.size();

                    logger.log(NODE_LOG_MODE_DEBUG) << "REQUEST SENT_RR to inproc_proxy socket from [" << identity << "], message [" << messageId << "] body "
                        << messageBody.size() << ", n=" << _statClientRequests << flush;
                  }
                }
              }
            }else{
              //!<Node mode not supported, make response message and send response for reducer
              sendEmptyResponseMessage(msg, messageBody, EmptyResponseType::ertBadNodeMode, messageId, messageRoute);
              logger.log(NODE_LOG_MODE_TRACE) << "Node mode value " << _nodeMode << " not supported, message skipped" << endl << msg.dump() << flush;
            }
          }
        }
      }
    }

    size_t DataServerProxy::getRandomNodeNumber(size_t maxNumber){
      //!<Initialize random seed
      std::random_device rd;
      std::mt19937_64 gen(rd());
      std::uniform_real_distribution<double> dis(0, maxNumber);
      //!<Generate index number limited clients queue size
      return static_cast<size_t>(dis(gen));
    }

    std::string DataServerProxy::getIdentityRandom(std::vector<ClientWorkerItem>& clients, size_t& previousClientNumber){
      logger.log(NODE_LOG_MODE_TRACE) << "NODE_MODE_SHARD_RND  clients count = " << clients.size() << " previousClientNumber = " << _previousClientNumber << flush;
      std::string identity;
      size_t queueIndex = 0;
      do{
        queueIndex = getRandomNodeNumber(clients.size());

        logger.log(NODE_LOG_MODE_TRACE) << "NODE_MODE_SHARD_RND  getRandomNodeNumber() return queueIndex = " << queueIndex << flush;

        if (clients.empty() || clients.size()==1){
          logger.log(NODE_LOG_MODE_TRACE) << "NODE_MODE_SHARD_RND  do-while broken because: clients count==1" << flush;
          break;
        }

      }while(queueIndex==_previousClientNumber);

      if (!clients.empty()){

        std::vector<ClientWorkerItem> clientsQueue(clients);
        if (clientsQueue.size() > 1){
          //!<Erase previous client
          clientsQueue.erase(clientsQueue.begin()+_previousClientNumber);
        }

        if (!clientsQueue.empty()){
          //!<Sort clients by frequency of queries done
          std::sort(clientsQueue.begin(), clientsQueue.end(), ClientWorkerItem::sortByFrequencyPredicate);
          //!<Save more greater node's identity name
          identity = clientsQueue[0].identity;

          for (queueIndex=0;queueIndex<clients.size();++queueIndex){
            if (clients[queueIndex].identity == identity){
              //!<Save previous client number
              previousClientNumber = queueIndex;
              break;
            }
          }
        }
      }
      //!<Return identity name
      return identity;
    }

    std::string DataServerProxy::getIdentityResourceUsage(std::vector<ClientWorkerItem>& clients,
        size_t& previousClientNumber,
        CalculatorAlgorithm::Types typeAlgorithm,
        const CalculatorAlgorithmData& algorithmData){

      logger.log(NODE_LOG_MODE_TRACE) << "NODE_MODE_SHARD_RU  client count = " << clients.size() << flush;

      HCE::handlers::NodeResourceDataList nodeResourceDataList;
      for (size_t i=0;i<clients.size();++i){
        if (!clients[i].resources.empty()){
          //!<Calculate average value of resources usage
          HCE::handlers::DictionaryConverter<std::string, double> converter(_resourceUsageManager.calculateAverage(clients[i].identity, clients[i].resources));
          HCE::handlers::CalculatorNodeResourceData resourceData = converter.toObject().cast<HCE::handlers::CalculatorNodeResourceData, std::string>(clients[i].identity);
          nodeResourceDataList.addItem(std::forward<HCE::handlers::CalculatorNodeResourceData>(resourceData));

          logger.log(NODE_LOG_MODE_TRACE) << i << ") [ " << clients[i].identity << " ] avg: '" << converter.toString() << "'" << flush;
        }
      }

      HCE::handlers::CalculatorNodeExecution calculatorNodeExecution(typeAlgorithm);
      std::string identity = calculatorNodeExecution.calculate(algorithmData, nodeResourceDataList, _resourceUsageManager);
      if (!identity.empty()){
        ResourceUsageCollection resourceUsageCollection = _clientsManager->getResources(identity);
        std::vector<double> listWeights = calculatorNodeExecution.getListWeights(algorithmData, resourceUsageCollection);
        _resourceUsageManager.setListWeights(identity, listWeights);

        if (_resourceUsageManager.isAllowedRange(identity, _resourcesUsageMaxAllowedRange)){
          identity.clear();
        }
      }

      for (size_t i=0;i<clients.size();++i){
        if (clients[i].identity==identity){
          previousClientNumber = i;
          break;
        }
      }
      return identity;
    }

    void DataServerProxy::sendEmptyResponseMessage(zmsg& msg,
        const std::string& coverMessageBody,
        EmptyResponseType responseType,
        const std::string& messageId,
        const std::string& messageRoute,
        const std::string& errorMsg){
      //!<Make response message and send response for reducer
      std::string messageBody = makeEmptyResponseMessage(coverMessageBody, responseType, messageRoute, errorMsg);
      //!<Insert message Id back
      msg.append(messageId.c_str());
      //!<Insert body back
      msg.append(messageBody.c_str());
      //!<Insert route back
      msg.append(messageRoute.c_str());
      //!<Create current clients number
      std::stringstream clientsNumberStream;
      clientsNumberStream << _clientsQueue.size();
      //!<Insert clients number field
      msg.append(clientsNumberStream.str().c_str());
      //!<Send message
      msg.send(*_inprocReducerInSock);
    }

    void DataServerProxy::sendResponseMessage(const std::string& identity, const std::string& messageId, const std::string& messageBody, const std::string& messageRoute){
      //!<Create response message and fill with fields
      zmsg clientMsg(_clientIdentity.c_str());
      clientMsg.wrap(identity.c_str(), NULL);
      clientMsg.append(messageId.c_str());
      clientMsg.append(messageBody.c_str());
      clientMsg.append(messageRoute.c_str());
      //!<Send response
      clientMsg.send(*_dataServerSock);
    }

    void DataServerProxy::nodeInit(const std::string& iniFile){

      if (!iniFile.empty()){
        logger.log(NODE_LOG_MODE_INFORMATION) << "load node settings from '" << iniFile << "' file" << flush;

        size_t drce_resources_extractor_max_threads = 1024;
        size_t drce_resources_extractor_max_processes = 1024;
        size_t drce_resources_extractor_max_disk_size = 0;
        size_t drce_resources_extractor_max_vram_size = 0;
        size_t drce_resources_extractor_max_rram_size = 0;
        std::string resourcesUsageAlgorithmWeights;
        std::string resourcesUsageLimits;
        unsigned int resourcesUsageCollectedSize = HCE::handlers::RESOURCES_USAGE_COLLECTED_SIZE;
        double resourcesUsageMaxAllowedRange = HCE::handlers::RESOURCES_USAGE_MAX_ALLOWED_RANGE;

        //!<Process Node's ini file
        try{
          Poco::AutoPtr<Poco::Util::IniFileConfiguration> pConf(new Poco::Util::IniFileConfiguration(iniFile));
          try{
            //!<Set type of calcuator algorithm
            typeCalculatorAlgorithm = static_cast<CalculatorAlgorithm::Types>(pConf->getInt("node.resources_usage_alg"));
            //!<Set weights for default algorithm
            resourcesUsageAlgorithmWeights = pConf->getString("node.resources_usage_alg0_weights");
            if (!resourcesUsageAlgorithmWeights.empty()) {
              HCE::handlers::DictionaryConverter<std::string, double> converter(resourcesUsageAlgorithmWeights);
              calculatorAlgorithmData = converter.toObject().cast<HCE::handlers::CalculatorAlgorithmData>();
            }
            //!<Set order for calculator algorithm data
            calculatorAlgorithmData.setOrder(pConf->getInt("node.resources_usage_order"));

            //!<Set resources usage limits
            resourcesUsageLimits = pConf->getString("node.resources_usage_limits");
            if (!resourcesUsageLimits.empty()) {
              HCE::handlers::DictionaryConverter<std::string, double> converter(resourcesUsageLimits);
              resourceUsageLimitsData = converter.toObject().cast<HCE::handlers::ResourceUsageLimitsData>();
            }

            //!<Set resources usage collected size
            resourcesUsageCollectedSize = pConf->getInt("node.resources_usage_collected_size");
            _resourcesUsageCollectedSize = resourcesUsageCollectedSize;

            //!<Set resource usage max allowed range (in percent)
            resourcesUsageMaxAllowedRange = pConf->getInt("node.resources_usage_max_allowed_range");
            _resourcesUsageMaxAllowedRange = resourcesUsageMaxAllowedRange;

            //!<Create instance of resource usage extractor
            HCE::drce::ResourceUsageExtractor::getInstance();

            //!<Extract threads and processes counts from configuration file
            drce_resources_extractor_max_threads = pConf->getInt("drce.drce_resources_extractor_max_threads");
            drce_resources_extractor_max_processes = pConf->getInt("drce.drce_resources_extractor_max_processes");
            drce_resources_extractor_max_disk_size = pConf->getInt("drce.drce_resources_extractor_max_disk_size");
            drce_resources_extractor_max_vram_size = pConf->getInt("drce.drce_resources_extractor_max_vram_size");
            drce_resources_extractor_max_rram_size = pConf->getInt("drce.drce_resources_extractor_max_rram_size");

          } catch(Poco::SyntaxException& e){
            //!<Log some error of config file processing
            logger.log(NODE_LOG_MODE_ERROR) << "nodeInit() - data node ini file processing error: " << e.displayText() << flush;
          }
        } catch(Poco::Exception& e){
          //!<Log some error of config file processing
          logger.log(NODE_LOG_MODE_ERROR) << "nodeInit() - data node ini file access error: " << e.displayText() << flush;
        }
        //!<Create instance of resource usage extractor and initialization parameters
        HCE::drce::ResourceUsageExtractor::getInstance().setMaxThreadsCount(drce_resources_extractor_max_threads);
        HCE::drce::ResourceUsageExtractor::getInstance().setMaxProcessesCount(drce_resources_extractor_max_processes);
        HCE::drce::ResourceUsageExtractor::getInstance().setMaxDiskSize(drce_resources_extractor_max_disk_size);
        HCE::drce::ResourceUsageExtractor::getInstance().setMaxVramSize(drce_resources_extractor_max_vram_size);
        HCE::drce::ResourceUsageExtractor::getInstance().setMaxRramSize(drce_resources_extractor_max_rram_size);

        logger.log(NODE_LOG_MODE_INFORMATION) << "node.resources_usage_order: " << calculatorAlgorithmData.getOrder() << flush;
        logger.log(NODE_LOG_MODE_INFORMATION) << "node.resources_usage_alg: " << static_cast<unsigned int>(typeCalculatorAlgorithm) << flush;
        logger.log(NODE_LOG_MODE_INFORMATION) << "node.resources_usage_alg0_weights: " << resourcesUsageAlgorithmWeights << flush;
        logger.log(NODE_LOG_MODE_INFORMATION) << "node.resources_usage_limits: " << resourcesUsageLimits << flush;
        logger.log(NODE_LOG_MODE_INFORMATION) << "node.resources_usage_collected_size: " << resourcesUsageCollectedSize << flush;
        logger.log(NODE_LOG_MODE_INFORMATION) << "node.resources_usage_max_allowed_range: " << resourcesUsageMaxAllowedRange << flush;
        logger.log(NODE_LOG_MODE_INFORMATION) << "drce.drce_resources_extractor_max_threads: " << drce_resources_extractor_max_threads << flush;
        logger.log(NODE_LOG_MODE_INFORMATION) << "drce.drce_resources_extractor_max_processes: " << drce_resources_extractor_max_processes << flush;
        logger.log(NODE_LOG_MODE_INFORMATION) << "drce.drce_resources_extractor_max_disk_size: " << drce_resources_extractor_max_disk_size << flush;
        logger.log(NODE_LOG_MODE_INFORMATION) << "drce.drce_resources_extractor_max_vram_size: " << drce_resources_extractor_max_vram_size << flush;
        logger.log(NODE_LOG_MODE_INFORMATION) << "drce.drce_resources_extractor_max_rram_size: " << drce_resources_extractor_max_rram_size << flush;
      }
    }

    void DataServerProxy::registerClient(const std::string& identity){
      _clientsManager->remove(identity);
      _clientsManager->append(identity, false);

      logger.log(NODE_LOG_MODE_DEBUG) << "READY message from client [" << identity << "] received" << flush;
    }

    void DataServerProxy::processControlMessage(zmsg& msg){

      //!<Fetch the clientIdentity field
      std::string clientIdentity((const char*)msg.pop_front().c_str());

      std::string addr((msg.address())?msg.address():"");
      std::string body((msg.body())?msg.body():"");

      logger.log(NODE_LOG_MODE_TRACE) << "DataServerProxy::processControlMessage  clientIdentity: " << clientIdentity << flush;
      logger.log(NODE_LOG_MODE_TRACE) << "DataServerProxy::processControlMessage            addr: " << addr << flush;
      logger.log(NODE_LOG_MODE_TRACE) << "DataServerProxy::processControlMessage            body: " << body << flush;

      if(addr == NODE_MSG_READY){
        registerClient(clientIdentity);
      }else if(addr.compare(0, NODE_MSG_HEARTBEAT.length(), NODE_MSG_HEARTBEAT)==0){

        std::vector<ClientWorkerItem> clientQueue;
        HCE::handlers::HeartbeatMessage heartbeatMessage(clientQueue);
        heartbeatMessage.extract(addr, NODE_MSG_HEARTBEAT);
        if (heartbeatMessage.isError())
          logger.log(NODE_LOG_MODE_ERROR) << "Operation extract heartbeat message has error: " << heartbeatMessage.getErrorMsg() << flush;
        else
        {
          _clientsManager->refresh(clientIdentity, clientQueue, heartbeatMessage.getResources());

          logger.log(NODE_LOG_MODE_TRACE) << "DataServerProxy::processControlMessage  _clientsManager->refresh(" << clientIdentity << ")" << flush;

          RouteMessage routeMsg(_clientsQueue);
          std::string route = routeMsg.build(NODE_MSG_ROUTES);
          if (routeMsg.isError())
            logger.log(NODE_LOG_MODE_ERROR) << "Operation build route message has error: " << routeMsg.getErrorMsg() << flush;
          else
          {
            logger.log(NODE_LOG_MODE_TRACE) << "DataServerProxy::processControlMessage  route: " << route << flush;
            s_send(*_inprocProxySock, route);
          }
        }
        logger.log(NODE_LOG_MODE_DEBUG) << "HEARTBEAT message from client [" << clientIdentity << "] received" << flush;

      }else if(addr == NODE_MSG_BYE){
        //!<Delete client from queue
        _clientsManager->remove(clientIdentity);

        logger.log(NODE_LOG_MODE_DEBUG) << "BYE message from client [" << clientIdentity << "] received" << flush;
      }else{
        //!<Invalid message type
        logger.log(NODE_LOG_MODE_DEBUG) << "INVALID command message from client [" << clientIdentity << "]" << flush;
        logger.log(NODE_LOG_MODE_TRACE) << msg.dump() << flush;
      }
    }

    void DataServerProxy::processDataMessage(zmsg& msg){
      //!<Fetch the clientIdentity field
      std::string clientIdentity((const char*)msg.pop_front().c_str());

      if(!_clientsManager->refresh(clientIdentity)){
        logger.log(NODE_LOG_MODE_DEBUG) << "MSG STRUCTURE ERROR wrong client identity [" << clientIdentity << "]"
            << endl << "Clients:[" << _clientsManager->getDump() << "]" << flush;
        logger.log(NODE_LOG_MODE_TRACE) << msg.dump() << flush;
      }

      //!<Get message Id
      std::string messageId((const char*)msg.pop_front().c_str());
      //!<Get message body
      std::string messageBody((const char*)msg.pop_front().c_str());
      _statClientResponsesSize += messageBody.size();
      logger.log(NODE_LOG_MODE_DEBUG) << "RECEIVED RESPONSE from down client [" << clientIdentity << "], message [" << messageId << "] body " << messageBody.size()
          << " bytes n=" << _statClientResponses << flush;

      std::string messageRoute((msg.parts())?(const char*)msg.pop_front().c_str():"");

      logger.log(NODE_LOG_MODE_TRACE) << "DataServerProxy::processDataMessage  messageRoute: " << messageRoute << flush;

      size_t pos = messageRoute.find_last_of(ROUTE_LIST_COUNT_DELIMITER);
      //!<Create current clients number
      std::stringstream clientsNumberStream;

      if (pos!=std::string::npos && (pos+1)!=std::string::npos) {
        unsigned int clientsNumber = 0;
        std::istringstream(messageRoute.substr(pos+1)) >> clientsNumber;
        if (clientsNumber && clientsNumber < _clientsQueue.size()) {
          clientsNumberStream << clientsNumber;
        }
        else
          clientsNumberStream << _clientsQueue.size();

        messageRoute = messageRoute.substr(0, pos);
      }else{
        clientsNumberStream << _clientsQueue.size();
      }

      logger.log(NODE_LOG_MODE_TRACE) << "DataServerProxy::processDataMessage  messageRoute: '" << messageRoute << "', clientsNumberStream: '" << clientsNumberStream.str() << "'" << flush;

      //!<Insert message Id back
      msg.append(messageId.c_str());
      //!<Insert body back
      msg.append(messageBody.c_str());

      //!<Insert route back
      msg.append(messageRoute.c_str());

      //!<Insert clients number field
      msg.append(clientsNumberStream.str().c_str());

      //!<Send message
      msg.send(*_inprocReducerInSock);

      logger.log(NODE_LOG_MODE_DEBUG) << "RESPONSE SENT to inproc_reducer_in socket, message [" << messageId << "] body " << messageBody.size() << " bytes sent" << flush;
    }

    std::string DataServerProxy::makeEmptyResponseMessage(const std::string& messageBody,
        EmptyResponseType responseType,
        const std::string& messageRoute,
        const std::string& errorMsg){

      HCE::JsonMessageCover msgCover(messageBody);
      std::string resultMessageBody;
      std::stringstream errorMessage;

      switch(msgCover.getType()){
        case HCE::types::MessageType::mtSphinx: {

            unsigned int errorCode = HCE::sphinx::NO_ERROR;

            if (responseType==EmptyResponseType::ertNoClients){
              //!<No target node clients available
              errorCode = HCE::sphinx::ERROR_TARGET_CLIENTS_NOT_AVAILABLE;
              errorMessage << HCE::sphinx::message_const::targetClientsNotAvailable;
            }else if(responseType==EmptyResponseType::ertBadNodeMode){
              //!<Node mode not supported
              errorCode = HCE::sphinx::ERROR_NODE_MODE_NOT_SUPPORTED;
              errorMessage << HCE::sphinx::message_const::nodeNotSupported;
            }else if(responseType==EmptyResponseType::ertNotEnoughResourceLimits){
              //!<Chosen node has not allowed resource usage limits
              errorCode = HCE::sphinx::ERROR_NODE_HAS_NOT_ENOUGH_RESOURCE_LIMITS;
              errorMessage << HCE::sphinx::message_const::nodeHasNotEnoughResources;
            }

            if (!errorMsg.empty()){
              errorMessage << ". " << errorMsg;
            }
            resultMessageBody = HCE::sphinx::SphinxFunctionalObject::makeErrorMessage(msgCover.getData(),
                errorCode,
                errorMessage.str());
          }
          break;
        case HCE::types::MessageType::mtDrce: {

            unsigned int errorCode = HCE::drce::NO_ERROR;

            if (responseType==EmptyResponseType::ertNoClients){
              //!<No target node clients available
              errorCode = HCE::drce::ERROR_TARGET_CLIENTS_NOT_AVAILABLE;
              errorMessage  << HCE::drce::message_const::targetClientsNotAvailable;
            }
            else if(responseType==EmptyResponseType::ertBadNodeMode){
              //!<Node mode not supported
              errorCode = HCE::drce::ERROR_NODE_MODE_NOT_SUPPORTED;
              errorMessage << HCE::drce::message_const::nodeNotSupported;
            }else if(responseType==EmptyResponseType::ertNotEnoughResourceLimits){
              //!<Chosen node has not allowed resource usage limits
              errorCode = HCE::drce::ERROR_NODE_HAS_NOT_ENOUGH_RESOURCE_LIMITS;
              errorMessage << HCE::drce::message_const::nodeHasNotEnoughResources;
            }

            if (!errorMsg.empty()){
              errorMessage << ". " << errorMsg;
            }
            resultMessageBody = HCE::drce::DRCEFunctionalObject::makeErrorMessage(msgCover.getData(),
                HCE::drce::DRCETaskRequest::TaskState::NOT_SET_AS_NEW,
                errorCode,
                errorMessage.str());
          }
          break;
        default:;
      }

      std::string jsonMsg;
      msgCover.setData(resultMessageBody);
      if(!msgCover.serialize(jsonMsg)){
        logger.log(NODE_LOG_MODE_ERROR) << "Serialization error of cover object from empty results " << msgCover.getErrorCode() << " : " << msgCover.getErrorMsg() << flush;
      }

      return jsonMsg;
    }

    void DataServerProxy::handleExternalMessage(void){
      //!<Fetch message from dataserver
      zmsg msg(*_dataServerSock);

      logger.log(NODE_LOG_MODE_TRACE) << "MSG FROM dataserver" << endl << msg.dump() << flush;

      //!<Process dataserver's message
      if(msg.parts() == NODE_MSG_DATA_MIN_FIELDS){
        //!<Control message
        processControlMessage(msg);
      }else if(msg.parts() > NODE_MSG_DATA_MIN_FIELDS){
        //!<Inc data responses counter
        _statClientResponses++;

        //!<Data message
        processDataMessage(msg);
      }else{
        logger.log(NODE_LOG_MODE_TRACE) << "ERROR of message structure, wrong fields number " << msg.parts() << endl << msg.dump() << flush;
      }
    }

    std::string DataServerProxy::processAdminCommands(std::string& command, std::string& parameters){
      std::string ret = NODE_ADMIN_ERROR_OK;

      logger.log(NODE_LOG_MODE_DEBUG) << "Admin command requested [" << command << "], parameters [" << parameters << "]" << flush;

      //!<Process commands
      if(command == NODE_MSG_STAT){
        std::stringstream result;
        result << NODE_ADMIN_ERROR_OK << NODE_ADMIN_COMMAND_DELIMITER << "clientRequests=" << _statClientRequests << ",clientResponses=" << _statClientResponses << ",clientRequestsSize="
            << _statClientRequestsSize << ",clientResponsesSize=" << _statClientResponsesSize << ",Requests=" << _statRequests << ",requestsSize=" << _statRequestsSize;
        ret = result.str();
      }else if(command == NODE_MSG_rebuildServerConnection){
        //!<Rebuild server connection
        rebuildServerConnection(_dataServerSock, _connectionString, ZMQ_ROUTER);
      }else if(command == NODE_MSG_DISCONNECT_SERVER_CONNECTION){
        //!<Disconnect server connection
        _dataServerSock->close();
      }else if(command == NODE_MSG_UPDATE_SCHEMA){
        setConnectionString(parameters);
      }else if(command == NODE_MSG_PROPERTIES){
        //!<Make property information
        ret = getPropertyInformation();
      }else if(command == NODE_MSG_MANAGER_MODE_GET){
        ret = NODE_ADMIN_ERROR_OK + NODE_ADMIN_COMMAND_DELIMITER + NODE_ADMIN_PROPERTY_NAME_MMODE + "=" + std::to_string(static_cast<unsigned int>(_nodeMode));
      }else if(command == NODE_MSG_MANAGER_MODE_SET){
        //!<Parse command parameters and fetch manager mode value
        unsigned int nodeMode = static_cast<unsigned int>(_nodeMode);
        //Parse manager mode from parameters
        std::istringstream(parameters) >> nodeMode;
        //Make response
        ret = NODE_ADMIN_ERROR_OK + NODE_ADMIN_COMMAND_DELIMITER + NODE_ADMIN_PROPERTY_NAME_MMODE + "=" + std::to_string(static_cast<unsigned int>(_nodeMode));
        //Set new manager mode
        _nodeMode = nodeMode;
      }else if(command == NODE_MSG_MANAGER_RESOURCES_COLLECTED_SIZE_GET){
        ret = NODE_ADMIN_ERROR_OK + NODE_ADMIN_COMMAND_DELIMITER + NODE_ADMIN_PROPERTY_RESOURCES_COLLECTED_SIZE + "=" + std::to_string(_resourceUsageManager.getMaxSize());
      }else if(command == NODE_MSG_MANAGER_RESOURCES_COLLECTED_SIZE_SET){
        //!<Parse command parameters and fetch resources usage collected size
        unsigned int resourcesUsageCollectedSize = _resourcesUsageCollectedSize;
        //Parse resources usage collected size from parameters
        std::istringstream(parameters) >> resourcesUsageCollectedSize;
        //Make response
        ret = NODE_ADMIN_ERROR_OK + NODE_ADMIN_COMMAND_DELIMITER + NODE_ADMIN_PROPERTY_RESOURCES_COLLECTED_SIZE + "=" + std::to_string(_resourceUsageManager.getMaxSize());
        //Set new resources usage collected size
        _resourceUsageManager.setMaxSize(resourcesUsageCollectedSize);
        _clientsManager->setMaxResourcesUsageCollectedSize(resourcesUsageCollectedSize);
        _resourcesUsageCollectedSize = resourcesUsageCollectedSize;
      }else if(command == NODE_MSG_MANAGER_PURGE_MODE_GET){
        ret = NODE_ADMIN_ERROR_OK + NODE_ADMIN_COMMAND_DELIMITER + NODE_ADMIN_PROPERTY_NAME_MMODE + "=" + std::to_string(static_cast<unsigned int>(_clientsManager->getPurgeMode()));
      }else if(command == NODE_MSG_MANAGER_PURGE_MODE_SET){
        //!<Parse command parameters
        int purgeMode = 0;
        //!<Parse purge mode from parameters
        std::istringstream(parameters) >> purgeMode;
        //!<Make response
        ret = NODE_ADMIN_ERROR_OK + NODE_ADMIN_COMMAND_DELIMITER + NODE_ADMIN_PROPERTY_NAME_MMODE + "=" + std::to_string(static_cast<unsigned int>(_clientsManager->getPurgeMode()));
        //!<Set new purge mode
        _clientsManager->setPurgeMode(purgeMode);
      }else if(command == NODE_MSG_HEARTBEAT_DELAY_GET){
        ret = NODE_ADMIN_ERROR_OK + NODE_ADMIN_COMMAND_DELIMITER + NODE_ADMIN_PROPERTY_HEARTBEAT_DELAY + "=" + std::to_string(getHeartbeatDelay()/1000);
      }else if(command == NODE_MSG_HEARTBEAT_DELAY_SET){
        //!<Parse command parameters
        int hbDelay = HEARTBEAT_DELAY;
        //!<Parse heartbeat delay from parameters
        std::istringstream(parameters) >> hbDelay;
        ret = NODE_ADMIN_ERROR_OK + NODE_ADMIN_COMMAND_DELIMITER + NODE_ADMIN_PROPERTY_HEARTBEAT_DELAY + "=" + std::to_string(getHeartbeatDelay()/1000);
        //!<Set new heartbeat delay value
        setHeartbeatDelay(static_cast<unsigned int>(hbDelay*1000));
      }else if(command == NODE_MSG_HEARTBEAT_TIMEOUT_GET){
        ret = NODE_ADMIN_ERROR_OK + NODE_ADMIN_COMMAND_DELIMITER + NODE_ADMIN_PROPERTY_HEARTBEAT_TIMEOUT + "=" + std::to_string(_clientsManager->getHeartbeatTimeout()/1000);
      }else if(command == NODE_MSG_HEARTBEAT_TIMEOUT_SET){
        //!<Parse command parameters
        int hbTimeout = HEARTBEAT_TIMEOUT;
        //!<Parse heartbeat timeout from parameters
        std::istringstream(parameters) >> hbTimeout;
        ret = NODE_ADMIN_ERROR_OK + NODE_ADMIN_COMMAND_DELIMITER + NODE_ADMIN_PROPERTY_HEARTBEAT_TIMEOUT + "=" + std::to_string(_clientsManager->getHeartbeatTimeout()/1000);
        //!<Set new heartbeat timeout value
        setHeartbeatTimeout(static_cast<unsigned int>(hbTimeout*1000));
        _clientsManager->setHeartbeatTimeout(getHeartbeatTimeout());
      }else if(command == NODE_MSG_HEARTBEAT_MODE_GET){
        ret = NODE_ADMIN_ERROR_OK + NODE_ADMIN_COMMAND_DELIMITER + NODE_ADMIN_PROPERTY_HEARTBEAT_MODE + "=" + std::to_string(getHeartbeatMode());
      }else if(command == NODE_MSG_HEARTBEAT_MODE_SET){
        //!<Parse command parameters
        int hbMode = HEARBEAT_MODE_DEFAULT;
        //!<Parse heartbeat mode from parameters
        std::istringstream(parameters) >> hbMode;
        ret = NODE_ADMIN_ERROR_OK + NODE_ADMIN_COMMAND_DELIMITER + NODE_ADMIN_PROPERTY_HEARTBEAT_MODE + "=" + std::to_string(getHeartbeatMode());
        //!<Set new heartbeat mode value
        setHeartbeatMode(static_cast<unsigned int>(hbMode));
      }else if(command == NODE_MSG_ROUTES){
        //!<Return cluster node routes list
        std::vector<ClientWorkerItem> clientsQueue;
        RouteMessage routeMessage(clientsQueue);
        ClientsQueueManager clientsManager(&clientsQueue, 0);
        clientsManager.refresh(_clientIdentity, _clientsQueue);
        ret = routeMessage.build();
      }else if(command == NODE_MSG_RESOURCE_USAGE){
        //!<Return cluster node resource usage
        std::string json = "";
        HCE::drce::ResourceUsageExtractor::getInstance().update();
        HCE::drce::ResourceUsage resourceUsage = HCE::drce::ResourceUsageExtractor::getInstance().extract();
        HCE::drce::ResourceUsageSerializator resourceUsageSerializator(resourceUsage);
        resourceUsageSerializator.serialize(json);
        if (resourceUsageSerializator.isError())
          logger.log(NODE_LOG_MODE_ERROR) << "Get resource usage error: "<< resourceUsageSerializator.getErrorMsg() << HCE::flush;
        ret = json;
      }else{
        //!<Unsupported command
        ret = NODE_ADMIN_ERROR_ERROR + NODE_ADMIN_COMMAND_DELIMITER + NODE_ADMIN_ERROR_UNSUPPORTED_COMMAND;
      }
      return ret;
    }

    //!<Get property information
    std::string DataServerProxy::getPropertyInformation(void){
      std::stringstream result;
      result << NODE_ADMIN_ERROR_OK
          << NODE_ADMIN_COMMAND_DELIMITER << getProperties() << "clientRequests=" << _statClientRequests
          << NODE_ADMIN_COMMAND_DELIMITER << "clientResponses=" << _statClientResponses
          << NODE_ADMIN_COMMAND_DELIMITER << "clientRequestsSize=" << _statClientRequestsSize
          << NODE_ADMIN_COMMAND_DELIMITER << "clientResponsesSize=" << _statClientResponsesSize
          << NODE_ADMIN_COMMAND_DELIMITER << "requests=" << _statRequests
          << NODE_ADMIN_COMMAND_DELIMITER << "requestsSize=" << _statRequestsSize
          << NODE_ADMIN_COMMAND_DELIMITER << "nodeMode=" << static_cast<int>(_nodeMode)
          << NODE_ADMIN_COMMAND_DELIMITER << "clients=" << _clientsManager->getDump()
          << NODE_ADMIN_COMMAND_DELIMITER << "statQPSClient=" << _statQPSClient
          << NODE_ADMIN_COMMAND_DELIMITER << "statQPSServer=" << _statQPSServer
          << NODE_ADMIN_COMMAND_DELIMITER << "statRPSClient=" << _statRPSClient
          << NODE_ADMIN_COMMAND_DELIMITER << "statRequestsBandwidthClient=" << _statRequestsBandwidthClient
          << NODE_ADMIN_COMMAND_DELIMITER << "statRequestsBandwidthServer=" << _statRequestsBandwidthServer
          << NODE_ADMIN_COMMAND_DELIMITER << "statResponsesBandwidthClient=" << _statResponsesBandwidthClient;
      return result.str();
    }

    void DataServerProxy::logPeriodicStatistics(void){
      //!<Out periodical stats
      if(s_clock() > _pstatsAt){

        _statQPSClient = (_statClientRequests - _periodicStatClientRequests) / (float)((STAT_INTERVAL) / 1000.0);
        _statQPSServer = (_statRequests - _statRequestsProcessed) / (float)((STAT_INTERVAL) / 1000.0);
        _statRPSClient = (_statClientResponses - _periodicStatClientResponses) / (float)((STAT_INTERVAL) / 1000.0);
        _statRPSServer = 0;
        _statRequestsBandwidthClient = (_statClientRequestsSize - _periodicStatClientRequestsSize) / (float)((STAT_INTERVAL) / 1000.0);
        _statRequestsBandwidthServer = (_statRequestsSize - _statRequestsProcessedSize) / (float)((STAT_INTERVAL) / 1000.0);
        _statResponsesBandwidthClient = (_statClientResponsesSize - _periodicStatClientResponsesSize) / (float)((STAT_INTERVAL) / 1000.0);
        _statResponsesBandwidthServer = 0;

        _statRequestsProcessed = _statRequests;
        _statRequestsProcessedSize = _statRequestsSize;
        _periodicStatClientRequests = _statClientRequests;
        _periodicStatClientResponses = _statClientResponses;
        _periodicStatClientRequestsSize = _statClientRequestsSize;
        _periodicStatClientResponsesSize = _statClientResponsesSize;

        logger.log(NODE_LOG_MODE_DEBUG) << "SRequests=" << _statRequests << ", CRequests=" << _statClientRequests << ", CResponses=" << _statClientResponses
            << ", SQPS=" << _statQPSServer << ", CQPS=" << _statQPSClient << ", CRPS=" << _statRPSClient << ", SRequestsSize=" << _statRequestsSize
            << ", CRequestsSize=" << _statClientRequestsSize << ", CResponsesSize=" << _statClientResponsesSize
            << ", SRequestsBandwidth=" << _statRequestsBandwidthServer << ", CRequestsBandwidth=" << _statRequestsBandwidthClient
            << ", CResponsesBandwidth=" << _statResponsesBandwidthClient << flush;

        //Shift next log stat action time
        _pstatsAt = s_clock() + STAT_INTERVAL;
      }
    }

    void DataServerProxy::heartbit(void){

      logger.log(NODE_LOG_MODE_TRACE) << "DataServerProxy::heartbit  HeartbeatMode: " << getHeartbeatMode() << " _heartbeatAt = '"
          << Poco::DateTimeFormatter::format(Poco::LocalDateTime(Poco::Timestamp(Poco::Timestamp::fromEpochTime(_heartbeatAt/1000))), "%Y-%m-%d %H:%M:%S") << "'" << flush;

      //!<Send heartbeats to idle clients if it's time
      if(s_clock() > _heartbeatAt){
        if (getHeartbeatMode() == HEARBEAT_MODE_ADAPTIVE){

          for(std::vector<ClientWorkerItem>::iterator it = _clientsQueue.begin(); it < _clientsQueue.end(); ++it){
            it->shiftCounter.responceTimeStamp(it->rtimestamp);
            it->shiftCounter.recount(getHeartbeatTimeout());

            logger.log(NODE_LOG_MODE_TRACE) << "[ " << it->identity << " ] delta = " << abs(s_clock()-_heartbeatAt+getHeartbeatDelay())
                << " _shiftCounter.getShift()= " << it->shiftCounter.getShift() << flush;

            size_t delta = s_clock()-_heartbeatAt+getHeartbeatDelay();
            if (delta > (getHeartbeatTimeout()+it->shiftCounter.getShift())){

              //!<Execute sending heartbeat
              sendHeartbeat(it->identity);
              //!<Reset response shift counter
              it->shiftCounter.reset();

              logger.log(NODE_LOG_MODE_DEBUG) << "SENT HEARTBIT from client [" << _clientIdentity << "] to server _heartbeatAt = '"
                  << Poco::DateTimeFormatter::format(Poco::LocalDateTime(Poco::Timestamp(Poco::Timestamp::fromEpochTime(_heartbeatAt/1000))), "%Y-%m-%d %H:%M:%S") << "'" << flush;
            }
          }
         //!<Send heartbeats to idle clients if it's time
        }else{

          for(std::vector<ClientWorkerItem>::iterator it = _clientsQueue.begin(); it != _clientsQueue.end(); ++it){
            //!<Execute sending heartbeat
            sendHeartbeat(it->identity);
            logger.log(NODE_LOG_MODE_DEBUG) << "HEARTBEAT message for client [" << it->identity.c_str() << "] sent by timer reason" << flush;
          }
        }
        //!<Reset heartbit time
        _heartbeatAt = s_clock() + getHeartbeatDelay();
      }
    }

    void DataServerProxy::sendHeartbeat(const std::string& identity){
      //!<Execute sending heartbeat
      zmsg msg(NODE_MSG_HEARTBEAT.c_str());
      msg.wrap(identity.c_str(), NULL);
      msg.send(*_dataServerSock);
    }

    void* DataServerProxy::process(void){
      logger.log(NODE_LOG_MODE_DEBUG) << NODE_ADMIN_ERROR_MAIN_MSG_LOOP_STARTED << flush;

      while(getInProgress()){
        try {
          std::vector<zmq::pollitem_t> items;
          if(_dataServerSock->connected()){
            items ={{*_dataServerSock, 0, ZMQ_POLLIN, 0},{*_inprocProxySock, 0, ZMQ_POLLIN, 0},{*_inprocAdminSock, 0, ZMQ_POLLIN, 0}};
          } else{
            items ={{*_inprocProxySock, 0, ZMQ_POLLIN, 0},{*_inprocAdminSock, 0, ZMQ_POLLIN, 0}};
          }

          //!<Poll dataserver
          zmq::poll(&items[0], items.size(), getPollTimeout());

          //!<Handle dataserver socket's message if socket is connected
          if(_dataServerSock->connected()){
            if(items[0].revents & ZMQ_POLLIN){
              handleExternalMessage();
            }
          }

          //!<Handle inproc_proxy socket message
          if(items[items.size() - 2].revents & ZMQ_POLLIN){
            handleInternalMessage();
          }

          //!<Handle inproc_admin's socket message
          if(items[items.size() - 1].revents & ZMQ_POLLIN){
            handleAdminMessage();
          }

          //!<Heartbeats for clients
          heartbit();

          //!<Log stats
          logPeriodicStatistics();

          //!<Send periodical property values
          sendPeriodicProperties();

          //!<Remove expired clients from queue
          std::string r = _clientsManager->purge();
          if(r != ""){
            logger.log(NODE_LOG_MODE_INFORMATION) << "PURGED_CLIENTS: " << r << flush;
          }
        }catch(zmq::error_t& error){
          std::cout << "DataServerProxy E: " << error.what() << std::endl;
        }catch(std::exception& e){
          std::cout << "DataServerProxy E: " << e.what() << std::endl;
        }catch(...){
          std::cout << "DataServerProxy E: Unknown exception" << std::endl;
        }
      }

      logger.log(NODE_LOG_MODE_INFORMATION) << NODE_ADMIN_ERROR_MAIN_MSG_LOOP_STOPPED << flush;

      return NULL;
    }

  }
}
