/**
 * @file DataServerProxy.hpp
 *
 * HCE project DataServerProxy handler.
 * Implements functionality for data server proxy handler.
 * Contains client data requests server used in data and proxy nodes.
 *
 * @author bgv <bgv.hce@gmail.com>
 * @link http://hierarchical-cluster-engine.com/
 * @copyright Copyright &copy; 2013 IOIX Ukraine
 * @license http://hierarchical-cluster-engine.com/license/
 * @package HCE project node API
 * @since 0.1
 */

#ifndef __HCE_DATASERVER_PROXY_INCLUDED__
#define __HCE_DATASERVER_PROXY_INCLUDED__

#include <sstream>
#include <map>

#include "Handler.hpp"
#include "DataNodeOptions.hpp"

//Node ini
#include <Poco/Util/IniFileConfiguration.h>
#include "Poco/String.h"
//Message cover object
#include "JsonMessageCover.hpp"
//Calculator balancing nodes for execution
#include "CalculatorNodeExecution.hpp"
//Checker of resource usage limits
#include "ResourceUsageLimitsChecker.hpp"

namespace HCE {
  namespace handlers {

    class DataServerProxy: public HCE::handlers::Handler, public HCE::handlers::DataNodeOptions {
      public:
        DataServerProxy(std::string name, zmq::context_t& ctx, const std::string& clientIdentity, std::string connectionString,
            unsigned char nodeMode, int64_t startedAt, unsigned char logPriority, const std::string& iniFile="");
        ~DataServerProxy();

      protected:
        int initialize(void);
        void deinitialize(void);
        void* process(void);

        void handleInternalMessage(void);
        void handleExternalMessage(void);

        std::string processAdminCommands(std::string& command, std::string& parameters);

        void logPeriodicStatistics(void);
        void heartbit(void);
        void sendHeartbeat(const std::string& identity);
        //void heartbitResponse(const std::string& identity); //TODO: experimental_heartbitResponse
        void registerClient(const std::string& identity);

        void processControlMessage(zmsg& msg);
        void processDataMessage(zmsg& msg);

        //!<Get property information
        std::string getPropertyInformation(void);

        enum class EmptyResponseType : std::int8_t
        {
          ertNoClients=0,
          ertBadNodeMode=1,
          ertNotEnoughResourceLimits=2
        };
        //!<Make empty response message according to response type
        std::string makeEmptyResponseMessage(const std::string& messageBody,
            EmptyResponseType responseType,
            const std::string& messageRoute,
            const std::string& errorMsg="");

        //!<Send empty response message to reducer
        void sendEmptyResponseMessage(zmsg& msg,
            const std::string& coverMessageBody,
            EmptyResponseType responseType,
            const std::string& messageId,
            const std::string& messageRoute,
            const std::string& errorMsg="");

        //!<Make and send response message
        void sendResponseMessage(const std::string& identity,
            const std::string& messageId,
            const std::string& messageBody,
            const std::string& messageRoute);

        //!<Node initialization from ini file
        void nodeInit(const std::string& iniFile);

        //!<ZMQ Sockets
        zmq::socket_t* _inprocProxySock;
        zmq::socket_t* _inprocReducerInSock;
        zmq::socket_t* _dataServerSock;

        //!<Client ID
        std::string _clientIdentity;

        //!<Data node mode
        unsigned char _nodeMode;
        //!<Previous client number
        size_t _previousClientNumber;
        //!<Random client number genertor
        size_t getRandomNodeNumber(size_t maxNumber);
        //!<Get identity use random algorithm
        std::string getIdentityRandom(std::vector<ClientWorkerItem>& clients,
            size_t& previousClientNumber);
        //!<Get identity use resource usage algorithm
        std::string getIdentityResourceUsage(std::vector<ClientWorkerItem>& clients,
            size_t& previousClientNumber,
            CalculatorAlgorithm::Types typeAlgorithm,
            const CalculatorAlgorithmData& algorithmData);

        //!<Cluster node ini file name.
        std::string _iniFile;
        //!<Algorithm data for balancing calculator
        CalculatorAlgorithmData calculatorAlgorithmData;
        //!<Resource usage limits data
        ResourceUsageLimitsData resourceUsageLimitsData;
        //!<Type of algorithm for balancing calculator
        CalculatorAlgorithm::Types typeCalculatorAlgorithm;
        //!<Resource usage collected size
        unsigned int _resourcesUsageCollectedSize;
        //!<Resource usage max allowed range for calculation of weight in percent
        double _resourcesUsageMaxAllowedRange;
        //!<Resource usage manager
        ResourceUsageManager _resourceUsageManager;

        //!<Queue of connected clients
        std::vector<ClientWorkerItem> _clientsQueue;
        ClientsQueueManager* _clientsManager;

        //!<Heartbeat at time mark
        int64_t _heartbeatAt;

        //!<Stat vars
        unsigned int _statClientRequests;
        unsigned int _statClientResponses;
        unsigned int _statRequests;
        unsigned int _statClientRequestsSize;
        unsigned int _statClientResponsesSize;
        unsigned int _statRequestsSize;
        unsigned int _periodicStatClientRequests;
        unsigned int _periodicStatClientResponses;
        unsigned int _statRequestsProcessed;
        unsigned int _statRequestsProcessedSize;
        unsigned int _periodicStatClientRequestsSize;
        unsigned int _periodicStatClientResponsesSize;
        float _statQPSClient;
        float _statQPSServer;
        float _statRPSClient;
        float _statRPSServer;
        float _statRequestsBandwidthClient;
        float _statRequestsBandwidthServer;
        float _statResponsesBandwidthClient;
        float _statResponsesBandwidthServer;

        //!<Send out periodical stats at regular intervals
        int64_t _pstatsAt = s_clock() + HEARTBEAT_DELAY;
    };
  }
}

#endif /* __HCE_DATASERVER_PROXY_INCLUDED__ */
