/**
 * @Handler.hpp
 *
 * HCE project node handler's parent.
 * Implements basic functionality common for any kind of node handler.
 *
 * @author bgv <bgv.hce@gmail.com>
 * @link http://hierarchical-cluster-engine.com/
 * @copyright Copyright &copy; 2013 IOIX Ukraine
 * @license http://hierarchical-cluster-engine.com/license/
 * @package HCE project node API
 * @since 0.1
 */

#ifndef __HCE_HANDLER_INCLUDED__
#define __HCE_HANDLER_INCLUDED__

#include <bitset>

#include "zmsg.hpp"
#include "HandlerTypes.hpp"
#include "ClientsQueueManager.hpp"
#include "AdminCommandParameters.hpp"

//POCO date time formatter dep.
#include <Poco/Timestamp.h>
#include <Poco/DateTimeFormatter.h>
#include <Poco/Logger.h>
#include <Poco/AutoPtr.h>
#include <Poco/Configurable.h>
#include <Poco/Util/IniFileConfiguration.h>
#include <Poco/String.h>

#include "LoggerStream.hpp"
#include "ServiceMessages.hpp"

namespace HCE {
  namespace handlers {

    class Handler {
      public:
        Handler(const std::string& name, zmq::context_t& context, const std::string& connectionString,
            int64_t startedAt, unsigned char logPriority, const std::string& iniFile="") :
            _inprocAdminSock(NULL), _name(name), _context(context), _startedAt(startedAt), _logPriority(logPriority), _logMask(NODE_LOG_DEAFULT_MASK),
            _logTimeFormat(NODE_LOG_DEAFULT_TIME_FORMAT), log(), logger(name), _connectionRebuildTriesMax(NODE_CONNECTION_REBUILD_MAX_TRIES),
            _connectionRebuildTriesDelay(NODE_CONNECTION_REBUILD_DELAY_DEFAULT), _connectionString(connectionString),
            _inProgress(true), _pollTimeout(POLL_TIMEOUT), _propertyInterval(PROPERTY_INTERVAL),
            _propertyIntervalAt(s_clock()+PROPERTY_INTERVAL){
          //!<Read initialization data from file
          initHandler(iniFile, logger);
        }

        virtual ~Handler(){
          if(_inprocAdminSock){
            if(_inprocAdminSock->connected()){
              _inprocAdminSock->close();
            }
            delete _inprocAdminSock;
            _inprocAdminSock = NULL;
          }
        }

        static void* main(void *that){
          return (static_cast<Handler*>(that))->process();
        }

        std::string getName(void) const{
          return _name;
        }

        void setName(const std::string& name){
          _name = name;
        }

        void setBindTriesMaxNumber(const unsigned int& bindTriesMaxNumber){
          _connectionRebuildTriesMax = bindTriesMaxNumber;
        }

        unsigned int getBindTriesMaxNumber(void) const{
          return _connectionRebuildTriesMax;
        }

        void setBindTriesDelay(const unsigned int& bindTriesDelay){
          _connectionRebuildTriesDelay = bindTriesDelay;
        }

        unsigned int getBindTriesDelay(void) const{
          return _connectionRebuildTriesDelay;
        }

        void setConnectionString(const std::string& connectionString){
          _connectionString = connectionString;
        }

        std::string getConnectionString(void) const{
          return _connectionString;
        }

        void setInProgress(bool in_progress){
          _inProgress = in_progress;
        }

        bool getInProgress(void) const{
          return _inProgress;
        }

        void setPollTimeout(unsigned int pollTimeout){
          if (pollTimeout>=MIN_ALLOWED_POLL_TIMEOUT){
            _pollTimeout = pollTimeout;
          }
        }

        unsigned int getPollTimeout(void) const{
          return _pollTimeout;
        }

        void setPropertyInterval(unsigned int propertyInterval){
          if (propertyInterval >= MIN_ALLOWED_PROPERTY_INTERVAL){
            _propertyInterval=propertyInterval;
          }
        }

        unsigned int getPropertyInterval(void) const{
          return _propertyInterval;
        }

        void shutdown(void){
          _inProgress = false;
        }

      protected:

        void logFlush(int priority = NODE_LOG_MODE_DEFAULT){
          if(priority == _logPriority || priority == NODE_LOG_MODE_DEFAULT){
            //Fill time mark
            std::string timeMark = Poco::DateTimeFormatter::format(Poco::Timestamp(static_cast<Poco::Timestamp::TimeVal>(s_clock() * 1000)), _logTimeFormat);

            //Backup log message
            std::string message = log.str();

            //Reset log accumulator
            log.str(std::string());
            log.clear();

            //Create log context
            if(_logMask[0]){
              //Out time diff
              log << std::setfill('0') << std::setw(8) << ((unsigned int)(s_clock() - _startedAt)) << " ";
            }
            //Out three more log components
            log << ((_logMask[1] ? timeMark : "") + (_logMask[2] ? "[" + _name + "] " : "") + (_logMask[3] ? message : "")) << std::endl;

            //Out final log context
            std::cout << log.str();
            std::cout.flush();
          }

          //Reset log accumulator
          log.str(std::string());
          log.clear();
        }

        //!<Node init from file
        void initHandler(const std::string& iniFile, LoggerStream& logger){
          if (!iniFile.empty()){
            logger.log(NODE_LOG_MODE_INFORMATION) << "load node settings from '" << iniFile << "' file" << flush;

            unsigned int propertyInterval = PROPERTY_INTERVAL/1000;
             //!<Process Node's ini file
            try
            {
              Poco::AutoPtr<Poco::Util::IniFileConfiguration> pConf(new Poco::Util::IniFileConfiguration(iniFile));
              try
              {
                propertyInterval = pConf->getInt("node.property_interval");
              }
              catch(Poco::SyntaxException& e)
              {
                //!<Log some error of config file processing
                logger.log(NODE_LOG_MODE_ERROR) << "nodeInit() - data node options ini file processing error: '" << e.displayText() << "'" << flush;
              }
            }
            catch(Poco::Exception& e)
            {
              //!<Log some error of config file processing
              logger.log(NODE_LOG_MODE_ERROR) << "nodeInit() - data node options ini file access error: '" << e.displayText() << "'" << flush;
            }
            //!<Apply parameters
            setPropertyInterval(propertyInterval*1000);

            logger.log(NODE_LOG_MODE_INFORMATION) << "property_interval = " << propertyInterval << flush;
          }
        }

        void handleAdminMessage(void){
          //!<Fetch request message from inproc_admin socket
          zmsg msg(*_inprocAdminSock);

          logger.log(NODE_LOG_MODE_TRACE) << " MSG FROM inproc_admin\n" << msg.dump() << flush;

          //!<Process message
          processAdminMessage(msg);

          //!<Send response message
          msg.send(*_inprocAdminSock);
        }

        void processAdminMessage(zmsg& msg){
          //!<Fetch message Id
          std::string messageId((const char*)msg.pop_front().c_str());
          //!<Fetch message command
          std::string messageCommand((const char*)msg.pop_front().c_str());
          //!<Fetch message command parameters
          std::string messageParams((const char*)msg.pop_front().c_str());
          //!<Fetch message source Identity
          std::string messageSrcIdentity((const char*)msg.pop_front().c_str());

          logger.log(NODE_LOG_MODE_DEBUG) << "RECEIVED admin message from inproc_admin id [" << messageId << "], body [" << messageCommand << "], params ["
              << messageParams << "], source identity [" << messageSrcIdentity << "]" << flush;

          //!<Create response
          std::string response = NODE_ADMIN_ERROR_OK + NODE_ADMIN_COMMAND_DELIMITER;

          if(messageCommand == NODE_MSG_ECHO){
            response += NODE_MSG_ECHO + NODE_ADMIN_COMMAND_DELIMITER + _name;
          }else if(messageCommand == NODE_MSG_LOG_LEVEL_GET){
            response += NODE_ADMIN_PROPERTY_NAME_LOG + "=" + std::to_string(_logPriority) + NODE_ADMIN_COMMAND_DELIMITER + _name;
          }else if(messageCommand == NODE_MSG_LOG_LEVEL_SET){
            //!<Parse command parameters and fetch log level value
            int logp = _logPriority;

            std::istringstream(messageParams) >> logp;
            response += NODE_ADMIN_PROPERTY_NAME_LOG + "=" + std::to_string(_logPriority) + NODE_ADMIN_COMMAND_DELIMITER + _name;
            _logPriority = logp;
            Poco::Logger::root().setLevel(_logPriority);
          }else if(messageCommand == NODE_MSG_POLL_TIMEOUT_GET){
            response += NODE_ADMIN_PROPERTY_POLL_TIMEOUT + "=" + std::to_string(_pollTimeout/1000);
          }else if(messageCommand == NODE_MSG_POLL_TIMEOUT_SET){
            //!<Parse command parameters and fetch log level value
            unsigned int pollTimeout = _pollTimeout;
            //Parse log priority from parameters
            std::istringstream(messageParams) >> pollTimeout;
            //Make response
            response += NODE_ADMIN_PROPERTY_POLL_TIMEOUT + "=" + std::to_string(_pollTimeout/1000);
            //Set new poll timeout value
            setPollTimeout(pollTimeout*1000);
          }else if(messageCommand == NODE_MSG_PROPERTY_INTERVAL_GET){
            response += NODE_ADMIN_PROPERTY_PROPERTY_INTERVAL + "=" + std::to_string(_propertyInterval/1000);
          }else if(messageCommand == NODE_MSG_PROPERTY_INTERVAL_SET){
            //!<Parse command parameters and fetch log level value
            unsigned int propertyInterval = _propertyInterval;
            //Parse log priority from parameters
            std::istringstream(messageParams) >> propertyInterval;
            //Make response
            response += NODE_ADMIN_PROPERTY_PROPERTY_INTERVAL + "=" + std::to_string(_propertyInterval/1000);
            //Set new poll timeout value
            setPropertyInterval(propertyInterval*1000);
          }else if(messageCommand == NODE_MSG_STOP){
            //!<Stop handler main loop as a result of processing of the "Shutdown" command for "Admin" handler
            setInProgress(false);
          }else{
            //!<Call admin commands processing redefined in the derivative class instance
            response = processAdminCommands(messageCommand, messageParams);
          }

          //!<Insert message Id
          msg.append(messageId.c_str());
          //!<Insert response
          msg.append(response.c_str());
          //!<Insert messageSrcIdentity
          msg.append(messageSrcIdentity.c_str());
        }

        virtual std::string processAdminCommands(std::string& command, std::string& parameters){
          std::string ret("DEFAULT stub for command processor!");

          return ret;
        }

        void rebuildAdminClientSocket(void){
          if(_inprocAdminSock){
            if(_inprocAdminSock->connected()){
              _inprocAdminSock->close();
            }
            delete _inprocAdminSock;
            _inprocAdminSock = NULL;
          }

          //!<Create inproc_admin socket
          _inprocAdminSock = new zmq::socket_t(_context, ZMQ_DEALER);
          if(_inprocAdminSock){
            //!<Set identity for client socket
            _inprocAdminSock->setsockopt(ZMQ_IDENTITY, _name.c_str(), _name.size());
            //!<Connect to inproc_admin socket
            _inprocAdminSock->connect(NODE_CS_INPROC_ADMIN.c_str());
            logger.log(NODE_LOG_MODE_DEBUG) << "Connected inproc_admin as [" << _name << "]" << flush;
            //!<Send ready message
            logger.log(NODE_LOG_MODE_DEBUG) << "Sending ready message to admin..." << flush;
            s_send(*_inprocAdminSock, NODE_MSG_READY);
            logger.log(NODE_LOG_MODE_DEBUG) << "Sent ready message to admin" << flush;
          }else{
            logger.log(NODE_LOG_MODE_FATAL) << "Admin client socket instantiate : " << NODE_CRITICAL_ERROR_MESSAGE << flush;
            ::abort();
          }
        }

        bool rebuildServerConnection(zmq::socket_t*& sock_p, const std::string& connectionString, const int type){
          bool bound = false;

          //!<Delete socket if already exists
          if(sock_p != NULL){
            //TODO: experimental
            if(sock_p->connected()){
              sock_p->close();
              s_sleep(100);
            }
            delete sock_p;
            sock_p = NULL;
          }

          //!<Create server's socket
          sock_p = new zmq::socket_t(_context, type);
          if(sock_p != NULL){
            //!<Try to bind
            unsigned int counter = 0;
            while(counter < _connectionRebuildTriesMax){
              try{
                sock_p->bind(connectionString.c_str());
                bound = true;
                break;
              } catch(zmq::error_t &e){
                logger.log(NODE_LOG_MODE_ERROR) << "Bind exeption : " << e.what() << flush;
                counter++;
                s_sleep(_connectionRebuildTriesDelay);
              }
            }
          }else{
            logger.log(NODE_LOG_MODE_FATAL) << "Server connection socket type " << type << " instantiate : " << NODE_CRITICAL_ERROR_MESSAGE << flush;
            ::exit(NODE_EXIT_CODE_GENERAL_CRITICAL_ERROR);
          }

          logger.log(NODE_LOG_MODE_DEBUG) << "Bind to [" << connectionString << "] ";
          if(bound){
            logger.log(NODE_LOG_MODE_DEBUG) << "success";
          }else{
            logger.log(NODE_LOG_MODE_DEBUG) << "error";
          }
          logger.log(NODE_LOG_MODE_DEBUG) << flush;

          return bound;
        }

        void rebuildClientConnection(zmq::socket_t*& sock_p, const std::string& connectionString, const int type, const std::string& identity=""){
          rebuildClientConnection(sock_p, connectionString, type, identity, -1);
        }

        bool rebuildClientConnection(zmq::socket_t*& sock_p, const std::string& connectionString, const int type, const std::string& identity, int linger, const std::string& readyMessage =
            "", const std::string& byeMessage = NODE_MSG_BYE){
          bool connected = false;

          //!<Delete socket if already exists
          if(sock_p != NULL){
            if(sock_p->connected()){
              if(byeMessage != ""){
                //!<Send BYE message to the server (especially for the Paranoid Pirate Protocol implementations)
                sendByeMessage(sock_p, identity, byeMessage);
              }

              logger.log(NODE_LOG_MODE_DEBUG) << "rebuildServerConnection()::close();" << flush;
              sock_p->close();
            }
            delete sock_p;
            sock_p = NULL;
          }

          //!<Create socket
          sock_p = new zmq::socket_t(_context, type);
          if(sock_p != NULL){
            if(identity != ""){
              //!<Set identity to route responses from router
              sock_p->setsockopt(ZMQ_IDENTITY, identity.c_str(), identity.size());
            }

            //!<Try to connect
            unsigned int counter = 0;
            while(counter < _connectionRebuildTriesMax){
              try{
                sock_p->connect(connectionString.c_str());
                connected = true;
                break;
              } catch(zmq::error_t &e){
                logger.log(NODE_LOG_MODE_ERROR) << "Connect [" << connectionString << "] warning : " << e.what() << flush;
                counter++;
                s_sleep(_connectionRebuildTriesDelay);
              }
            }
          }else{
            logger.log(NODE_LOG_MODE_FATAL) << "Client connection socket type " << type << " instantiate : " << NODE_CRITICAL_ERROR_MESSAGE << flush;
            ::exit(NODE_EXIT_CODE_GENERAL_CRITICAL_ERROR);
          }

          logger.log(NODE_LOG_MODE_INFORMATION) << "Connect to [" << connectionString << "] ";
          if(connected){
            logger.log(NODE_LOG_MODE_INFORMATION) << "success" << flush;

            if(linger != -1){
              //!<Set linger for the socket (0 - to not wait at close time)
              sock_p->setsockopt(ZMQ_LINGER, &linger, sizeof(linger));
              logger.log(NODE_LOG_MODE_DEBUG) << "Linger value " << linger << " for [" << connectionString << "] set" << flush;
            }

            if(readyMessage != ""){
              //!<Send READY message to the server (especially for the Paranoid Pirate Protocol implementations)
              s_send(*sock_p, readyMessage);
              logger.log(NODE_LOG_MODE_INFORMATION) << "READY message sent from client [" << identity << "] to [" << connectionString << "]" << flush;
            }
          }else{
            logger.log(NODE_LOG_MODE_DEBUG) << "error" << flush;
          }

          return connected;
        }

        void sendByeMessage(zmq::socket_t*& sock_p, const std::string& identity="", const std::string& byeMessage = NODE_MSG_BYE){
          //!<Send BYE message to the server (especially for the Paranoid Pirate Protocol implementations)
          s_send(*sock_p, byeMessage);
          logger.log(NODE_LOG_MODE_INFORMATION) << "BYE message sent from client [" << ((identity.empty())?_name:identity) << "]" << flush;
        }

        std::string getProperties(void) const{
          std::stringstream result;
          result << "name=" << _name << NODE_ADMIN_COMMAND_DELIMITER << "startedAt=" << _startedAt << NODE_ADMIN_COMMAND_DELIMITER << "logPriority=" << static_cast<int>(_logPriority)
              << NODE_ADMIN_COMMAND_DELIMITER << "connectionRebuildTriesMax=" << _connectionRebuildTriesMax << NODE_ADMIN_COMMAND_DELIMITER << "connectionRebuildTriesDelay="
              << _connectionRebuildTriesDelay << NODE_ADMIN_COMMAND_DELIMITER << "connectionString=" << _connectionString << NODE_ADMIN_COMMAND_DELIMITER;
          return result.str();
        }

        //!<Get property information
        virtual std::string getPropertyInformation(void){
          std::string ret("DEFAULT stub for property informaion!");
          return ret;
        }

        //!<Send periodical property values
        void sendPeriodicProperties(void){
          if(s_clock() > _propertyIntervalAt){

            HCE::handlers::PropertyMessage propertyMessage(_name, getPropertyInformation());
            std::string message = propertyMessage.build();
            if (!propertyMessage.isError()){
              s_send(*_inprocAdminSock, message);

              if (_name == "DataServerProxy"){
                //!<Log periodic properties sending
                logger.log(NODE_LOG_MODE_TRACE) << "Send 'Periodic Properties' from DataServerProxy to _inprocAdminSock" << flush;
              }
            }else{
              logger.log(NODE_LOG_MODE_ERROR) << "Operation build property message has error: " << propertyMessage.getErrorMsg() << flush;
            }
            _propertyIntervalAt = s_clock() + getPropertyInterval();
          }
        }

        virtual int initialize(void) = 0;
        virtual void deinitialize(void) = 0;
        virtual void* process(void) = 0;

        void handleInternalMessage(void);
        void handleExternalMessage(void);

        void processControlMessage(zmsg& msg);
        void processDataMessage(zmsg& msg);

        void logPeriodicStatistics(void);
        void heartbit(void);
//        void heartbitResponse(const std::string& identity);
        void registerClient(const std::string& identity);

        //Admin socket
        zmq::socket_t* _inprocAdminSock;

        //Handler instance name
        std::string _name;

        //ZMQ context
        zmq::context_t& _context;

        //Start time mark
        int64_t _startedAt;

        //Log priority level
        int _logPriority;
        //Log components mask
        std::bitset<4> _logMask;
        //Log time format POCO specification
        std::string _logTimeFormat;
        //Log stream buffer (use only in method logFlush)
        std::stringstream log;
        //Logger stream use now as log buffer
        LoggerStream logger;

        //ZMQ connections rebuild tries settings
        unsigned int _connectionRebuildTriesMax;
        unsigned int _connectionRebuildTriesDelay;

        //!<ZMQ connection string
        std::string _connectionString;

        //!<Main loop of sockets polling condition
        bool _inProgress;

        //!<Poll timeout value
        unsigned int _pollTimeout;

        //!<Property interval between resend Property values
        unsigned int _propertyInterval;
        //!<Send out periodical property values to Admin class at regular intervals
        int64_t _propertyIntervalAt;
    };

  }
}

#endif /* __HCE_HANDLER_INCLUDED__ */

