/**
 * @file HandlerTypes.hpp
 *
 * HCE project handler_types definitions.
 * Contains general common definitions of types and constants for any kind of handlers.
 *
 * @author bgv <bgv.hce@gmail.com>
 * @link http://hierarchical-cluster-engine.com/
 * @copyright Copyright &copy; 2013 IOIX Ukraine
 * @license http://hierarchical-cluster-engine.com/license/
 * @package HCE project node API
 * @since 0.1
 */

#ifndef __HCE_HANDLER_TYPES_INCLUDED__
#define __HCE_HANDLER_TYPES_INCLUDED__

#include "zmsg.hpp"
#include "ClientWorkerItem.hpp"

namespace HCE {
  namespace handlers {
    //!<Delay intervals
    const unsigned int HEARTBEAT_DELAY = 15000;                  //!<  Delay receiving of heartbeat messages from client, msec
    const unsigned int HEARTBEAT_TIMEOUT = 60000;                //!<  Heartbeat messages sent period interval, msec
    const unsigned int LOG_STATISTIC_PERIOD = 30000;             //!<  Statistics log interval, msec
    const unsigned int RECONNECT_INTERVAL_INIT = 10000;          //!<  Initial reconnect period, msec
    const unsigned int RECONNECR_INTERVAL_MAX = 30000;           //!<  Max reconnect period that can be reached during an exponential backoff
    const unsigned int POLL_TIMEOUT = HEARTBEAT_DELAY/2;         //!<  POLL timeout of main sockets polling in handlers, msec
    const unsigned int STAT_INTERVAL = 60000;                    //!<  Stat logging action period interval, msec
    const unsigned int PROPERTY_INTERVAL = 10000;                //!<  Interval between resend PROPERTY values from Handles to Admin class

    const unsigned int MIN_ALLOWED_HEARTBEAT_DELAY = 5000;       //!<  Min allowed delay receiving of heartbeat messages from client, msec
    const unsigned int MIN_ALLOWED_HEARTBEAT_TIMEOUT = 5000;     //!<  Min allowed heartbeat messages sent period interval, msec
    const unsigned int MIN_ALLOWED_POLL_TIMEOUT = 5000;          //!<  Min allowed POLL timeout of main sockets polling in handlers, msec
    const unsigned int MIN_ALLOWED_PROPERTY_INTERVAL = 5000;     //!<  Min allowed interval between resend PROPERTY values from Handles to Admin class

    //!<Heartbeat modes
    const unsigned int HEARBEAT_MODE_FIXED_TIME = 0;
    const unsigned int HEARBEAT_MODE_IMMEDIATE_AFTER_HB = 1;
    const unsigned int HEARBEAT_MODE_MIXED = 2;
    const unsigned int HEARBEAT_MODE_ADAPTIVE = 3;

    const unsigned int HEARBEAT_MODE_DEFAULT = HEARBEAT_MODE_FIXED_TIME;
    const unsigned int HEARBEAT_MODE_MAX_NUMBER = HEARBEAT_MODE_ADAPTIVE;

    //!<Data processing and shard model modes
    const unsigned char NODE_MODE_PROXY = 0;            //!< Multicast requests balancing (supposes proportional sharding of documents as sharding model)
    const unsigned char NODE_MODE_SHARD = 1;            //!< Round robin requests balancing (supposes mirroring of data documents as sharding model)
    const unsigned char NODE_MODE_DATA = 2;             //!< Node in data mode, i.e. - end of chain in cluster tree
    const unsigned char NODE_MODE_ROUTER = 3;           //!< Multicast requests balancing
    const unsigned char NODE_MODE_SHARD_RND = 4;        //!< Random single-cast balancing
    const unsigned char NODE_MODE_SHARD_RU = 5;         //!< Resources usage balancing (supposes choose of data node based resources usage)

    //!<Messages fields number structure
    const unsigned int NODE_MSG_DATA_MIN_FIELDS = 2; //!<Minimal fields number to treat message as valid data message
    const unsigned int NODE_MSG_CMD_MIN_FIELDS = 1;  //!<Minimal fields number to treat message as command (heartbit, ready and another node states)

    //!<Log mode priorities
    const int NODE_LOG_MODE_NONE        = 0; //!< A none error. In the application turn off any log.
    const int NODE_LOG_MODE_FATAL       = 1; //!< A fatal error. The application will most likely terminate. This is the highest priority.
    const int NODE_LOG_MODE_CRITICAL    = 2; //!< A critical error. The application might not be able to continue running successfully.
    const int NODE_LOG_MODE_ERROR       = 3; //!< An error. An operation did not complete successfully, but the application as a whole is not affected.
    const int NODE_LOG_MODE_WARNING     = 4; //!< A warning. An operation completed with an unexpected result.
    const int NODE_LOG_MODE_NOTICE      = 5; //!< A notice, which is an information with just a higher priority.
    const int NODE_LOG_MODE_INFORMATION = 6; //!< An informational message, usually denoting the successful completion of an operation.
    const int NODE_LOG_MODE_DEBUG       = 7; //!< A debugging message.
    const int NODE_LOG_MODE_TRACE       = 8; //!< A tracing message. This is the lowest priority.

    const int NODE_LOG_MODE_DEFAULT = NODE_LOG_MODE_INFORMATION; //!< Log mode default to enable logging all modes

    //!<Exit codes
    const unsigned int NODE_EXIT_CODE_GENERAL_CRITICAL_ERROR = 1;
    const unsigned int NODE_EXIT_CODE_SPHINX_CRITICAL_ERROR = 2;
    const unsigned int NODE_EXIT_CODE_DRCE_CRITICAL_ERROR = 3;

    //!<Connection delays and intervals
    const unsigned int NODE_CONNECTION_REBUILD_DELAY_DEFAULT = 60000;  //Connection rebuild delay, ms
    const unsigned int NODE_CONNECTION_REBUILD_MAX_TRIES = 100;        //Connection rebuild max tries number

    //!<Message structure signatures
    const char NODE_ROUTER_MSG_DELIMITER = '@';
    const char NODE_ADMIN_COMMAND_DELIMITER = 0x09;  //!<':';
    const char NODE_JSON_BEGIN_CHAR = '{';
    const std::string NODE_JSON_EMPTY_DEFAULT = "{}";
    const char NODE_ADMIN_COMMAND_PROPERTY_ITEMS_DELIMITER = ',';
    const char NODE_ADMIN_COMMAND_PROPERTY_ITEM_PAIR_DELIMITER = ':';
    const char ROUTE_TABLE_MSG_TITLE_DELIMITER = ':';
    const char ROUTE_LIST_NAMES_DELIMITER = ',';
    const char ROUTE_LIST_COUNT_DELIMITER = ':';
    const char HEARTBEAT_MSG_DELIMITER = ':';
    const char PROPERTY_MSG_DELIMITER = ':';
    const std::string HEARTBEAT_JSON_RESOURCES = "resources";
    const std::string HEARTBEAT_JSON_ROUTES = "routes";

    //Ini file params values related
    const char NODE_INI_FILE_PATH_START_CHAR = '/';
    const char NODE_INI_FILE_PATH_DELIMITER_CHAR = '/';

    //!<Data modes
    const unsigned int NODE_DATA_PROCESSING_MODE_NORMAL = 0;         //!< Regular data processing supposes to involve FO call(s)
    const unsigned int NODE_DATA_PROCESSING_MODE_SIMULATE = 1;       //!< Simulation that avoid natural FO call(s) and to use auto-generated results to return

    //!<ZMQ connection protocols names prefix
    const std::string ZMQ_PROTOCOL_PREFIX_TCP = "tcp://";
    const std::string ZMQ_PROTOCOL_PREFIX_INPROC = "inproc://";

    //!<Internal connection strings
    const std::string NODE_CS_INPROC_ADMIN = ZMQ_PROTOCOL_PREFIX_INPROC + "inproc_admin";
    const std::string NODE_CS_INPROC_PROXY = ZMQ_PROTOCOL_PREFIX_INPROC + "inproc_proxy";
    const std::string NODE_CS_INPROC_DATA = ZMQ_PROTOCOL_PREFIX_INPROC + "inproc_data";
    const std::string NODE_CS_INPROC_REDUCER_OUT = ZMQ_PROTOCOL_PREFIX_INPROC + "inproc_reducer_out";
    const std::string NODE_CS_INPROC_REDUCER_IN = ZMQ_PROTOCOL_PREFIX_INPROC + "inproc_reducer_in";

    //!<General handlers command messages
    const std::string NODE_MSG_READY = "READY";
    const std::string NODE_MSG_HEARTBEAT = "HEARTBEAT";
    const std::string NODE_MSG_BYE = "BYE";

    //!<Admin handlers command messages supported by all handlers via parent
    const std::string NODE_MSG_ECHO = "ECHO";
    const std::string NODE_MSG_LOG_LEVEL_SET = "LLSET";
    const std::string NODE_MSG_LOG_LEVEL_GET = "LLGET";
    const std::string NODE_MSG_HEARTBEAT_DELAY_SET = "HEARTBEAT_DELAY_SET";
    const std::string NODE_MSG_HEARTBEAT_DELAY_GET = "HEARTBEAT_DELAY_GET";
    const std::string NODE_MSG_HEARTBEAT_TIMEOUT_SET = "HEARTBEAT_TIMEOUT_SET";
    const std::string NODE_MSG_HEARTBEAT_TIMEOUT_GET = "HEARTBEAT_TIMEOUT_GET";
    const std::string NODE_MSG_HEARTBEAT_MODE_SET = "HEARTBEAT_MODE_SET";
    const std::string NODE_MSG_HEARTBEAT_MODE_GET = "HEARTBEAT_MODE_GET";
    const std::string NODE_MSG_POLL_TIMEOUT_SET = "POLL_TIMEOUT_SET";
    const std::string NODE_MSG_POLL_TIMEOUT_GET = "POLL_TIMEOUT_GET";
    const std::string NODE_MSG_PROPERTY_INTERVAL_SET = "PROPERTY_INTERVAL_SET";
    const std::string NODE_MSG_PROPERTY_INTERVAL_GET = "PROPERTY_INTERVAL_GET";

    //!<Admin handlers command messages supported by specific handler depends on command
    const std::string NODE_MSG_STAT = "STAT";
    const std::string NODE_MSG_TIME = "TIME";
    const std::string NODE_MSG_PROPERTIES = "PROPERTIES";
    const std::string NODE_MSG_rebuildServerConnection = "rebuildServerConnection";
    const std::string NODE_MSG_REBUILD_CLIENT_CONNECTION = "REBUILD_CLIENT_CONNECTION";
    const std::string NODE_MSG_UPDATE_SCHEMA = "UPDATE_SCHEMA";
    const std::string NODE_MSG_DISCONNECT_SERVER_CONNECTION = "DISCONNECT_SERVER_CONNECTION";
    const std::string NODE_MSG_DISCONNECT_CLIENT_CONNECTION = "DISCONNECT_CLIENT_CONNECTION";
    const std::string NODE_MSG_SHUTDOWN = "SHUTDOWN";
    const std::string NODE_MSG_SPHINX = "SPHINX";
    const std::string NODE_MSG_DRCE = "DRCE";
    const std::string NODE_MSG_DRCE_SET_HOST = "DRCE_SET_HOST";
    const std::string NODE_MSG_DRCE_GET_HOST = "DRCE_GET_HOST";
    const std::string NODE_MSG_DRCE_SET_PORT = "DRCE_SET_PORT";
    const std::string NODE_MSG_DRCE_GET_PORT = "DRCE_GET_PORT";
    const std::string NODE_MSG_DRCE_GET_TASKS = "DRCE_GET_TASKS"; // for return async task queue
    const std::string NODE_MSG_DRCE_GET_TASKS_INFO = "DRCE_GET_TASKS_INFO"; // for return async tasks information

    const std::string NODE_MSG_ROUTES = "NODE_ROUTES"; // for return routes of cluster nodes
    const std::string NODE_MSG_RESOURCE_USAGE = "NODE_RESOURCE_USAGE"; // for return resource usage array
    const std::string NODE_MSG_SET_DATA_PROCESSING_MODE = "SET_DATA_PROCESSING_MODE";
    const std::string NODE_MSG_GET_DATA_PROCESSING_MODE = "GET_DATA_PROCESSING_MODE";
    const std::string NODE_MSG_MANAGER_MODE_SET = "MMSET";
    const std::string NODE_MSG_MANAGER_MODE_GET = "MMGET";
    const std::string NODE_MSG_MANAGER_PURGE_MODE_SET = "MPMSET";
    const std::string NODE_MSG_MANAGER_PURGE_MODE_GET = "MPMGET";
    const std::string NODE_MSG_MANAGER_RESOURCES_COLLECTED_SIZE_SET = "MRCSSET";
    const std::string NODE_MSG_MANAGER_RESOURCES_COLLECTED_SIZE_GET = "MRCSGET";
    const std::string NODE_MSG_STOP = "STOP"; //Supported by all handlers except of "Admin"

    //!<Error messages
    const std::string NODE_ADMIN_ERROR_WRONG_FORMAT = "Wrong admin message format, expected string - DESTINATION_IDENTITY:COMMAND_NAME:PARAMETERS, or destination admin handler not found";
    const std::string NODE_ADMIN_ERROR_UNSUPPORTED_COMMAND = "REDEFINED admin command handler - unsupported admin command!";
    const std::string NODE_ADMIN_ERROR_HANDLER_NOT_FOUND = "Handler not found or not connected to inproc admin server";
    const std::string NODE_ADMIN_ERROR_UNSUPPORTED_CC_REBUILD = "Client connection rebuild not supported for handler ";
    const std::string NODE_ADMIN_ERROR_OK = "OK";
    const std::string NODE_ADMIN_ERROR_ERROR = "ERROR";
    const std::string NODE_ADMIN_ERROR_MAIN_MSG_LOOP_STARTED = "Main messages loop started";
    const std::string NODE_ADMIN_ERROR_MAIN_MSG_LOOP_STOPPED = "Main messages loop stopped";

    //!<Internal handler names
    const std::string NODE_ADMIN_HANDLER = "Admin";

    //!<Functional object Sphinx errors
    const std::string NODE_FUNCTIONAL_ERROR_SPHINX = "SPHINX_FO_ERR";
    const std::string NODE_FUNCTIONAL_ERROR_SPHINX_ADMIN = "SPHINX_FO_ADMIN_ERR";
    const std::string NODE_FUNCTIONAL_ERROR_SPHINX_ADMIN_JSON_MSG = "{\"error_code\":\"1\", \"error_message\": \"Sphinx FO error\", \"data\":\"ERROR\", \"time\":\"-1\"}";
    const std::string NODE_FUNCTIONAL_ERROR_SPHINX_INTERNAL_LOG_BEGIN = "\n!!!SPHINXFO!!![\n";
    const std::string NODE_FUNCTIONAL_ERROR_SPHINX_INTERNAL_LOG_END = "\n]!!!SPHINXFO!!!\n";
    const std::string NODE_CRITICAL_ERROR_MESSAGE = "Critical error, can't continue, execution aborted!";
    const std::string NODE_FUNCTIONAL_ERROR_SPHINX_REDUCER = "SPHINX_REDUCER_FO_ERR";

    //!<Functional object Sphinx defaults
    const std::string NODE_FUNCTIONAL_PROJECT_ROOT_DIR = "hce-node-bundle";
    const bool NODE_FUNCTIONAL_SPHINX_NOT_START_SEARCHD = false;
    const bool NODE_FUNCTIONAL_SPHINX_NOT_STOP_SEARCHD = false;
    const unsigned int NODE_FUNCTIONAL_SPHINX_RANKER_0 = 0;
    const std::string NODE_FUNCTIONAL_DRCE_TASKS_DATA_DIR = "data";
    const std::string NODE_FUNCTIONAL_DRCE_TASKS_STATUS_DIR = "log";

    //!<Functional object DRCE errors
    const std::string NODE_FUNCTIONAL_ERROR_DRCE = "DRCE_FO_ERR";
    const std::string NODE_FUNCTIONAL_ERROR_DRCE_INTERNAL_LOG_BEGIN = "\n!!!DRCEFO!!![\n";
    const std::string NODE_FUNCTIONAL_ERROR_DRCE_INTERNAL_LOG_END = "\n]!!!DRCEFO!!!\n";

    //!<Admin command properties names
    const std::string NODE_ADMIN_PROPERTY_NAME_LOG = "log";
    const std::string NODE_ADMIN_PROPERTY_NAME_MMODE = "mode";
    const std::string NODE_ADMIN_PROPERTY_HEARTBEAT_DELAY = "hb_delay";
    const std::string NODE_ADMIN_PROPERTY_HEARTBEAT_TIMEOUT = "hb_timeout";
    const std::string NODE_ADMIN_PROPERTY_HEARTBEAT_MODE = "hb_mode";
    const std::string NODE_ADMIN_PROPERTY_POLL_TIMEOUT = "poll_timeout";
    const std::string NODE_ADMIN_PROPERTY_PROPERTY_INTERVAL = "property_interval";
    const std::string NODE_ADMIN_PROPERTY_RESOURCES_COLLECTED_SIZE = "collected_size";

    //!<Default resources usage collected size value
    const unsigned int RESOURCES_USAGE_COLLECTED_SIZE = 10;
    //!<Default resources usage max allowed range value in percent
    const double RESOURCES_USAGE_MAX_ALLOWED_RANGE = 10.0;

    //!<Default log time mark format
    const std::string NODE_LOG_DEAFULT_TIME_FORMAT = "[%Y-%m-%d %H:%M:%s] ";
    const int NODE_LOG_DEAFULT_MASK = 8+4+2+1;

    //!<DRCE request error - can't set as new task
    const int DRCE_REQUEST_STATE_ERROR_SET_AS_NEW_TASK = 6;

    //!<Get property information parameters
    const unsigned int PROPERTY_REALTIME_DEFAULT = 1;

    //!<Messages accumulator item
    typedef struct {
        std::string bodies;
        size_t count;
    } DefaultReducerResponseItem;

  }

  const std::string DRCE_FAKE_RESPONSE =
  "[{"
  " \"error_code\": 1,"
  "  \"error_message\": \"!!!FAKE JSON!!!\","
  "  \"id\": 12345,"
  "  \"type\": 1,"
  "  \"host\": \"\","
  "  \"port\": 0,"
  "  \"state\": 0,"
  "  \"pid\": 0,"
  "  \"stdout\": \"\","
  "  \"stderror\": \"\","
  "  \"exit_status\": \"\","
  "  \"files\": [],"
  "  \"node\": \"FAKE NODE\","
  "  \"time\": 0"
  "}]";
}

#endif /* __HCE_HANDLER_TYPES_INCLUDED__ */
