#include "RouterServerProxy.hpp"
#include "ServiceMessages.hpp"

namespace HCE {
  namespace handlers {

    RouterServerProxy::RouterServerProxy(std::string name, zmq::context_t& ctx, const std::string& clientIdentity, std::string connectionString,
        int64_t startedAt, unsigned char logPriority, const std::string& iniFile) :
        Handler(name, ctx, connectionString, startedAt, logPriority, iniFile), _clientIdentity(clientIdentity){

      //!<Init all the rest
      initialize();
    }

    RouterServerProxy::~RouterServerProxy(){
      deinitialize();
    }

    int RouterServerProxy::initialize(void){
      //!<Init sockets
      _inprocProxySock = NULL;
      _inprocReducerOutSock = NULL;
      _routerServerSock = NULL;

      //!<Heartbeat at time mark
      _heartbeatAt = s_clock() + LOG_STATISTIC_PERIOD;

      //!<Stat vars
      _statRequests = 0;
      _statResponses = 0;
      _statRequestsSize = 0;
      _statResponsesSize = 0;
      _statRequestsProcessed = 0;
      _statResponsesProcessed = 0;
      _statRequestsProcessedSize = 0;
      _statResponsesProcessedSize = 0;
      _statQPS = 0;
      _statRPS = 0;
      _statRequestsBandwidth = 0;
      _statResponsesBandwidth = 0;

      //!<Send out periodical stats at regular intervals
      _pstatsAt = s_clock() + LOG_STATISTIC_PERIOD;

      logger.log(NODE_LOG_MODE_INFORMATION) << "Initialize client [" << _clientIdentity << "] with connection [" << _connectionString << "]" << flush;

      //!<Rebuild admin client socket
      rebuildAdminClientSocket();

      //!<Rebuild inproc_proxy connection
      rebuildClientConnection(_inprocProxySock, NODE_CS_INPROC_PROXY, ZMQ_DEALER);
      //!<Rebuild inproc_reducer_out connection
      rebuildClientConnection(_inprocReducerOutSock, NODE_CS_INPROC_REDUCER_OUT, ZMQ_DEALER);
      //!<Rebuild external server connection
      rebuildServerConnection(_routerServerSock, _connectionString, ZMQ_ROUTER);

      return true;
    }

    void RouterServerProxy::deinitialize(void){
      if(_inprocProxySock){
        delete _inprocProxySock;
      }
      if(_inprocReducerOutSock){
        delete _inprocReducerOutSock;
      }
      if(_routerServerSock){
        delete _routerServerSock;
      }
      _inprocProxySock = NULL;
      _inprocReducerOutSock = NULL;
      _routerServerSock = NULL;
      _responses.clear();
    }

    void RouterServerProxy::handleInternalMessage(void){
      //!<Fetch message from inproc_reducer_out
      zmsg msg(*_inprocReducerOutSock);
      logger.log(NODE_LOG_MODE_TRACE) << " MSG FROM inproc_reducer_out" << endl << msg.dump() << flush;

      _statResponses++;
      //!<Parse message, create identity, messageId and body of dataclient's message
      std::string messageId((const char*)msg.pop_front().c_str());
      std::string messageBody((const char*)msg.pop_front().c_str());
      std::string messageRoute((msg.parts())?(const char*)msg.pop_front().c_str():"");

      _statResponsesSize += messageBody.size();
      logger.log(NODE_LOG_MODE_DEBUG) << " RECEIVED RESPONSE from inproc_reducer_out message [" << messageId << "] body " << messageBody.size() << " bytes, n=" << _statResponses << flush;
      //!<Generate unique client_Id and request Id from message_Id and rebuild message to send response to the client via ZMQ router
      std::vector<std::string> parts;
      //!<Split message_Id on two parts
      std::istringstream partsStream(messageId);
      std::string tmpString;

      while(std::getline(partsStream, tmpString, NODE_ROUTER_MSG_DELIMITER)){
        parts.push_back(tmpString);
      }
      if(parts.size() > 1){
        //!<Fill message with router`s client_Id, message_Id and messageBody
        msg.append(parts[0].c_str());
        msg.append(parts[1].c_str());
        msg.append(messageBody.c_str());

        logger.log(NODE_LOG_MODE_TRACE) << "MSG TO routerserver" << endl << msg.dump() << flush;

        //!<Send response message to client router`s client
        msg.send(*_routerServerSock);

        logger.log(NODE_LOG_MODE_DEBUG) << "RESPONSE SENT to routerserver socket from proxy node, n=" << _statResponses << flush;
      }else{
        //!<Wrong message_Id format
        logger.log(NODE_LOG_MODE_ERROR) << "ERROR : Wrong message_Id format [" << messageId << "] received from client" << flush;
      }
    }

    void RouterServerProxy::handleExternalMessage(void){
      //!<Fetch message from dataclient
      zmsg clientMsg(*_routerServerSock);

      //!<Data message
      if(clientMsg.parts() > NODE_MSG_DATA_MIN_FIELDS){
        logger.log(NODE_LOG_MODE_TRACE) << "MSG FROM routerserver" << endl << clientMsg.dump() << flush;

        _statRequests++;
        //!<Parse message, create identity, messageId and body of dataclient's message
        std::string clientIdentity((const char*)clientMsg.pop_front().c_str());
        std::string messageId((const char*)clientMsg.pop_front().c_str());
        std::string messageBody((const char*)clientMsg.pop_front().c_str());
        std::string messageRoute((clientMsg.parts())?(const char*)clientMsg.pop_front().c_str():"");

        logger.log(NODE_LOG_MODE_DEBUG) << "RouterServerProxy::handleExternalMessage  Recieved body: \"" << messageBody << "\"" << flush;
        logger.log(NODE_LOG_MODE_DEBUG) << "RouterServerProxy::handleExternalMessage  Recieved route: \"" << messageRoute << "\"" << flush;

        _statRequestsSize += messageBody.size();
        logger.log(NODE_LOG_MODE_DEBUG) << "RECEIVED REQUEST from router`s client [" << _clientIdentity << "], message [" << messageId << "] body "
            << messageBody.size() << " bytes, n=" << _statRequests << flush;

        logger.log(NODE_LOG_MODE_DEBUG) << "RouterServerProxy::handleExternalMessage  Recieved route: \"" << messageRoute << "\"" << flush;

        //!<Generate unique request Id from client_Id amd message_Id and rebuild message for farther processing
        clientMsg.clear();
        clientMsg.append(clientIdentity.c_str());
        clientMsg.append((clientIdentity + NODE_ROUTER_MSG_DELIMITER + messageId).c_str());
        clientMsg.append(messageBody.c_str());
        clientMsg.append(messageRoute.c_str());

        if(_logPriority >= NODE_LOG_MODE_DEBUG){
          logger.log(NODE_LOG_MODE_DEBUG) << " MSG TO inproc_proxy" << endl << clientMsg.dump() << flush;
        }

        //!<Send message to inproc_proxy socket
        clientMsg.send(*_inprocProxySock);

        logger.log(NODE_LOG_MODE_DEBUG) << " REQUEST SENT to inproc_proxy socket, n=" << _statRequests << flush;
      }else{
        logger.log(NODE_LOG_MODE_ERROR) << "ERROR MESSAGE RECEIVED from routerserver, parts=" << clientMsg.parts() << ", n=" << _statRequests
            << endl << clientMsg.dump() << flush;
      }
    }

    void RouterServerProxy::handleRouteMessage(void){

      //!<Fetch message from inproc_proxy
      zmsg msg(*_inprocProxySock);
      logger.log(NODE_LOG_MODE_DEBUG) << " MSG FROM inproc_proxy" << endl << msg.dump() << flush;

      std::string addr((msg.address())?msg.address():"");
      if (addr.compare(0, NODE_MSG_ROUTES.length(), NODE_MSG_ROUTES)==0)
      {
        std::vector<ClientWorkerItem> clients;
        RouteMessage routeMessage(clients);
        routeMessage.extract(NODE_MSG_ROUTES, addr);
        if (routeMessage.isError())
          logger.log(NODE_LOG_MODE_ERROR) << "Operation extract ROUTE MESSAGE has error: " << routeMessage.getErrorMsg() << flush;
        else
        {
          std::vector<ClientWorkerItem> clientsQueue;
          RouteMessage routesMsg(clientsQueue);
          ClientsQueueManager clientsManager(&clientsQueue, 0);
          clientsManager.refresh(_clientIdentity, clients, "Here must be resources information...");
          std::string route = routesMsg.build();
          if (routesMsg.isError())
            logger.log(NODE_LOG_MODE_ERROR) << "Operation build ROUTE MESSAGE has error: " << routesMsg.getErrorMsg() << flush;
          else
          {
            logger.log(NODE_LOG_MODE_TRACE) << "RouterServerProxy::handleRouteMessage  ROUTE_TABLE: " << route << flush;
            s_send(*_routerServerSock, route);
          }
        }
      }
    }

    std::string RouterServerProxy::processAdminCommands(std::string& command, std::string& parameters){
      std::string ret = NODE_ADMIN_ERROR_OK;

      logger.log(NODE_LOG_MODE_DEBUG) << "Admin command requested [" << command << "], parameters [" << parameters << "]" << flush;

      //!<Process commands
      if(command == NODE_MSG_STAT){
        std::stringstream result;
        result << NODE_ADMIN_ERROR_OK << NODE_ADMIN_COMMAND_DELIMITER << "requests=" << _statRequests << ",responses=" << _statResponses << ",requestsSize=" << _statRequestsSize
            << ",response_size=" << _statResponsesSize;
        ret = result.str();
      }else if(command == NODE_MSG_rebuildServerConnection){
        //!<Rebuild server connection
        rebuildServerConnection(_routerServerSock, _connectionString, ZMQ_ROUTER);
      }else if(command == NODE_MSG_DISCONNECT_SERVER_CONNECTION){
        //!<Disconnect server connection
        _routerServerSock->close();
      }else if(command == NODE_MSG_UPDATE_SCHEMA){
        setConnectionString(parameters);
      }else if(command == NODE_MSG_PROPERTIES){
        //!<Make property information
        ret = getPropertyInformation();
      }else{
        //!<Unsupported command
        ret = NODE_ADMIN_ERROR_ERROR + NODE_ADMIN_COMMAND_DELIMITER + NODE_ADMIN_ERROR_UNSUPPORTED_COMMAND;
      }

      return ret;
    }

    void RouterServerProxy::logPeriodicStatistics(void){
      if(s_clock() > _pstatsAt){
        _statQPS = (_statRequests - _statRequestsProcessed) / (float)((STAT_INTERVAL) / 1000.0);
        _statRPS = ((_statResponses - _statResponsesProcessed) / (float)((STAT_INTERVAL) / 1000.0));
        _statRequestsBandwidth = (_statRequestsSize - _statRequestsProcessedSize) / (float)((STAT_INTERVAL) / 1000.0);
        _statResponsesBandwidth = (_statResponsesSize - _statResponsesProcessedSize) / (float)((STAT_INTERVAL) / 1000.0);

        _statRequestsProcessed = _statRequests;
        _statResponsesProcessed = _statResponses;
        _statRequestsProcessedSize = _statRequestsSize;
        _statResponsesProcessedSize = _statResponsesSize;

        logger.log(NODE_LOG_MODE_DEBUG) << "requests=" << _statRequests << ", responses=" << _statResponses << ", qps=" << _statQPS << ", rps=" << _statRPS
            << ", requestsSize=" << _statRequestsSize << ", responsesSize=" << _statResponsesSize << ", requestsBandwidth=" << _statRequestsBandwidth
            << ", responsesBandwidth=" << _statResponsesBandwidth << flush;

        _pstatsAt = s_clock() + LOG_STATISTIC_PERIOD;
      }
    }

    //!<Get property information
    std::string RouterServerProxy::getPropertyInformation(void){
      std::stringstream result;
      result << NODE_ADMIN_ERROR_OK
          << NODE_ADMIN_COMMAND_DELIMITER << getProperties() << "requests=" << _statRequests << NODE_ADMIN_COMMAND_DELIMITER << "responses=" << _statResponses
          << NODE_ADMIN_COMMAND_DELIMITER << "requestsSize=" << _statRequestsSize << NODE_ADMIN_COMMAND_DELIMITER << "responseSize=" << _statResponsesSize
          << NODE_ADMIN_COMMAND_DELIMITER << "clientIdentity=" << _clientIdentity << NODE_ADMIN_COMMAND_DELIMITER << "heartbeatAt" << static_cast<uint64_t>(_heartbeatAt)
          << NODE_ADMIN_COMMAND_DELIMITER << "pstatsAt=" << static_cast<uint64_t>(_pstatsAt) << NODE_ADMIN_COMMAND_DELIMITER << "statQPS=" << _statQPS
          << NODE_ADMIN_COMMAND_DELIMITER << "statRPS=" << _statRPS << NODE_ADMIN_COMMAND_DELIMITER << "statRequestsBandwidth=" << _statRequestsBandwidth
          << NODE_ADMIN_COMMAND_DELIMITER << "statResponsesBandwidth=" << _statResponsesBandwidth;
      return result.str();
    }

    void* RouterServerProxy::process(void){
      logger.log(NODE_LOG_MODE_DEBUG) << NODE_ADMIN_ERROR_MAIN_MSG_LOOP_STARTED << flush;

      while(getInProgress()){
        try {
          //!<Init polling pool sockets array
          std::vector<zmq::pollitem_t> items;
          if(_routerServerSock->connected()){
            items ={{*_routerServerSock, 0, ZMQ_POLLIN, 0},{*_inprocReducerOutSock, 0, ZMQ_POLLIN, 0},{*_inprocAdminSock, 0, ZMQ_POLLIN, 0},{*_inprocProxySock, 0, ZMQ_POLLIN, 0}};
          } else{
            items ={{*_inprocReducerOutSock, 0, ZMQ_POLLIN, 0},{*_inprocAdminSock, 0, ZMQ_POLLIN, 0},{*_inprocProxySock, 0, ZMQ_POLLIN, 0}};
          }

          //!<Poll sockets
          zmq::poll(&items[0], items.size(), getPollTimeout());

          //!<Handle routerserver's socket message if connected
          if(_routerServerSock->connected()){
            if(items[0].revents & ZMQ_POLLIN){
              handleExternalMessage();
            }
          }

          //!<Handle inproc_reducer_out's socket message
          if(items[items.size() - 3].revents & ZMQ_POLLIN){
            handleInternalMessage();
          }

          //!<Handle inproc_admin's socket message
          if(items[items.size() - 2].revents & ZMQ_POLLIN){
            handleAdminMessage();
          }

          //!<Handle inproc_proxy's socket message
          if(items[items.size() - 1].revents & ZMQ_POLLIN){
            handleRouteMessage();
          }

          //!<Out periodical stats
          logPeriodicStatistics();

          //!<Send periodical property values
          sendPeriodicProperties();
        }catch(zmq::error_t& error){
          std::cout << "RouterServerProxy E: " << error.what() << std::endl;
        }catch(std::exception& e){
          std::cout << "RouterServerProxy E: " << e.what() << std::endl;
        }catch(...){
          std::cout << "RouterServerProxy E: Unknown exception" << std::endl;
        }
      }

      logger.log(NODE_LOG_MODE_DEBUG) << NODE_ADMIN_ERROR_MAIN_MSG_LOOP_STOPPED << flush;

      return NULL;
    }

  }
}
