/**
 * @file RouterServerProxy.hpp
 *
 * HCE project RouterServerProxy handler.
 * Implements functionality for router server handler.
 * Contains client requests router server and cluster nodes (proxy or data) server.
 * This is root of cluster hierarchy and data streams root point for client-side applications connections and upper connection point for
 * data connections of all kind of nodes in cluster. Get request from client application, send request to all connected proxy or data nodes,
 * collects upper data flows in one response and send response back to client-side application.
 *
 * @author bgv <bgv.hce@gmail.com>
 * @link http://hierarchical-cluster-engine.com/
 * @copyright Copyright &copy; 2013 IOIX Ukraine
 * @license http://hierarchical-cluster-engine.com/license/
 * @package HCE project node API
 * @since 0.1
 */

#ifndef __HCE_ROUTERSERVER_PROXY_INCLUDED__
#define __HCE_ROUTERSERVER_PROXY_INCLUDED__

#include <sstream>
#include <map>

#include "Handler.hpp"

namespace HCE {
namespace handlers {

class RouterServerProxy: public HCE::handlers::Handler {
public:
  RouterServerProxy(std::string name, zmq::context_t& ctx, const std::string& clientIdentity, std::string connectionString,
      int64_t startedAt, unsigned char logPriority, const std::string& iniFile);
  virtual ~RouterServerProxy();

protected:
  int initialize(void);
  void deinitialize(void);
  void* process(void);

  void handleInternalMessage(void);
  void handleExternalMessage(void);
  void handleRouteMessage(void);

  std::string processAdminCommands(std::string& command, std::string& parameters);

  void logPeriodicStatistics(void);

  //!<Get property information
  std::string getPropertyInformation(void);

  //!<ZMQ Sockets
  zmq::socket_t* _inprocProxySock;
  zmq::socket_t* _inprocReducerOutSock;
  zmq::socket_t* _routerServerSock;

  //!<Client ID
  std::string _clientIdentity;

  //!<Heartbeat at time mark
  int64_t _heartbeatAt;

  //!<Messages data accumulator map
  std::map<std::string, DefaultReducerResponseItem*> _responses;

  //!<Stat vars
  unsigned int _statRequests;
  unsigned int _statResponses;
  unsigned int _statRequestsSize;
  unsigned int _statResponsesSize;
  unsigned int _statRequestsProcessed;
  unsigned int _statResponsesProcessed;
  unsigned int _statRequestsProcessedSize;
  unsigned int _statResponsesProcessedSize;
  float _statQPS;
  float _statRPS;
  float _statRequestsBandwidth;
  float _statResponsesBandwidth;

  //!<Send out periodical stats at regular intervals
  int64_t _pstatsAt = s_clock() + LOG_STATISTIC_PERIOD;
};

}
}

#endif /* __HCE_ROUTERSERVER_PROXY_INCLUDED__ */
