#include <Poco/JSON/JSON.h>
#include <Poco/JSON/Stringifier.h>
#include <Poco/JSON/Parser.h>
#include <Poco/JSON/JSONException.h>
#include <Poco/Dynamic/Var.h>
#include <sstream>
#include <algorithm>
#include <iterator>

#include "ServiceMessages.hpp"

namespace HCE
{
namespace handlers
{
//-----------------------------------------------------------------------------
RouteMessage::RouteMessage(std::vector<ClientWorkerItem>& queue_)
: inherited(), queue(queue_)
{
}
//-----------------------------------------------------------------------------
void RouteMessage::packToObject(Poco::JSON::Object& obj, const std::vector<ClientWorkerItem>& workerItem)
{
  for (size_t i=0;i<workerItem.size();++i)
  {
    Poco::JSON::Object::Ptr pItemElem = new Poco::JSON::Object();
    if (!pItemElem.isNull())
    {
      packToObject(*pItemElem, workerItem[i].route);
      obj.set(workerItem[i].identity, pItemElem);
    }
  }
}
//-----------------------------------------------------------------------------
bool RouteMessage::serialize(std::string& json)
{
  resetError();
  try
  {
    Poco::JSON::Object::Ptr pObj = new Poco::JSON::Object();
    if (pObj.isNull())
      throw Poco::Exception("Object instance was not created");

    packToObject(*pObj, queue);

    std::stringstream ostr;
    Poco::JSON::Stringifier::stringify(pObj, ostr);
    json = ostr.str();

    _isError = false;
  }
  catch(Poco::Exception& e)
  {
    _isError = true;
    errorMsg = e.displayText();
    errorCode = ERROR_UNSERIALIZE;
  }
  catch(std::exception& e)
  {
    _isError = true;
    errorMsg = e.what();
    errorCode = ERROR_SERIALIZE;
  }
  return !_isError;
}
//-----------------------------------------------------------------------------
void RouteMessage::unpackFromObject(Poco::JSON::Object& obj, std::vector<ClientWorkerItem>& workerItem)
{
  std::vector<std::string> names;
  obj.getNames(names);

  for (size_t i=0;i<names.size();++i)
  {
    Poco::JSON::Object::Ptr pObj = obj.getObject(names[i]);
    if (!pObj.isNull())
    {
      ClientWorkerItem clientWorker(names[i]);
      unpackFromObject(*pObj, clientWorker.route);
      workerItem.push_back(std::forward<ClientWorkerItem>(clientWorker));
    }
  }
}
//-----------------------------------------------------------------------------
bool RouteMessage::unserialize(const std::string& json)
{
  resetError();
  queue.clear();

  try
  {
    Poco::JSON::Parser parser;
    Poco::Dynamic::Var res = parser.parse(json);

    Poco::JSON::Object::Ptr pObj = res.extract<Poco::JSON::Object::Ptr>();

    unpackFromObject(*pObj, queue);

    _isError = false;
  }
  catch(Poco::JSON::JSONException& e)
  {
    _isError = true;
    errorMsg = e.displayText();
    errorCode = ERROR_UNSERIALIZE;
  }
  catch(Poco::Exception& e)
  {
    _isError = true;
    errorMsg = e.displayText();
    errorCode = ERROR_UNSERIALIZE;
  }
  catch(std::exception& e)
  {
    _isError = true;
    errorMsg = e.what();
    errorCode = ERROR_UNSERIALIZE;
  }
  return !_isError;
}
//-----------------------------------------------------------------------------
std::string RouteMessage::build(const std::string& title)
{
  std::string result = "";
  std::string json;
  serialize(json);

  if (title.empty())
    result = json;
  else
    result = (title+ROUTE_TABLE_MSG_TITLE_DELIMITER+json);
  return result;
}
//-----------------------------------------------------------------------------
void RouteMessage::extract(const std::string& title, const std::string& msgBody)
{
  try
  {
    if (msgBody.compare(0, title.length(), title)!=0)
      Poco::Exception("Title of message not found", ERROR_TITLE);

    size_t pos = msgBody.find_first_of(ROUTE_TABLE_MSG_TITLE_DELIMITER);
    if (pos==std::string::npos)
      Poco::Exception("Delimiter of title message not found", ERROR_TITLE);

    std::string json = msgBody.substr(pos+1);
    unserialize(json);
  }
  catch(Poco::Exception& e)
  {
    _isError = true;
    errorMsg = e.displayText();
    errorCode = e.code();
  }
}
//-----------------------------------------------------------------------------
void RouteMessage::clear(void)
{
  queue.clear();
}
//-----------------------------------------------------------------------------
std::vector<std::string> RouteMessage::extractIdentityNames(const std::string& routeMsg)
{
  std::vector<std::string> clients;
  std::string routes(routeMsg);

  size_t pos = routes.find_first_of(ROUTE_LIST_COUNT_DELIMITER);
  if (pos!=std::string::npos)
    routes = routes.substr(0, pos);

  std::replace(routes.begin(), routes.end(), ROUTE_LIST_NAMES_DELIMITER, ' ');

  std::stringstream routesStream(routes);
  std::istream_iterator <std::string> iit(routesStream);
  std::istream_iterator <std::string> eos;

  while (iit != eos)
  {
    clients.push_back((*iit));
    ++iit;
  }
  return clients;
}
//-----------------------------------------------------------------------------
std::string RouteMessage::buildIdentityNames(const std::vector<ClientWorkerItem>& clientQueue)
{
  std::vector<std::string> names;
  for (size_t i=0;i<clientQueue.size();++i)
  {
    names.push_back(clientQueue[i].identity);
  }
  return buildIdentityNames(names);
}
//-----------------------------------------------------------------------------
std::string RouteMessage::buildIdentityNames(const std::vector<std::string>& names)
{
  std::stringstream outMsg;
  for (size_t i=0;i<names.size();++i)
  {
    outMsg << names[i];
    if ((i+1) < names.size())
      outMsg << ROUTE_LIST_NAMES_DELIMITER;
  }
  return outMsg.str();
}
//-----------------------------------------------------------------------------
void RouteMessage::extractRouteNames(std::vector<std::string>& routeNames, const std::vector<ClientWorkerItem>& clientQueue)
{
  for (size_t i=0;i<clientQueue.size();++i)
  {
    routeNames.push_back(clientQueue[i].identity);
    extractRouteNames(routeNames, clientQueue[i].route);
  }
}
//-----------------------------------------------------------------------------
std::vector<std::string> RouteMessage::extractRouteNames(const std::vector<ClientWorkerItem>& clientQueue)
{
  std::vector<std::string> routeNames;
  extractRouteNames(routeNames, clientQueue);
  return routeNames;
}
//-----------------------------------------------------------------------------
std::vector<std::string> RouteMessage::extractRouteNames(void)
{
  return extractRouteNames(queue);
}
//-----------------------------------------------------------------------------
//-----------------------------------------------------------------------------
RequestRouteMessage::RequestRouteMessage(const std::string& json)
: inherited(), managerRole(0), nodesNames(), resourcesUsageAlgorithm(0), resourcesUsageAlgorithmWeights(""), resourcesUsageLimits("")
{
  if (!json.empty())
    unserialize(json);
}
//-----------------------------------------------------------------------------
bool RequestRouteMessage::serialize(std::string& json)
{
  resetError();
  try
  {
    Poco::JSON::Object::Ptr pObj = new Poco::JSON::Object();
    if (pObj.isNull())
      throw Poco::Exception("Object instance was not created");

    pObj->set(MANAGER_ROLE, managerRole);

    Poco::JSON::Array::Ptr pNodesNames = new Poco::JSON::Array();
    if (pNodesNames.isNull())
      throw Poco::Exception("Object instance was not created");

    for (size_t i=0;i<nodesNames.size();++i)
    {
      pNodesNames->add(nodesNames[i]);
    }
    pObj->set(NODES_NAMES, pNodesNames);

    pObj->set(RESOURCES_USAGE_ALGORITHM, resourcesUsageAlgorithm);
    pObj->set(RESOURCES_USAGE_ALGORITHM_WEIGHTS, resourcesUsageAlgorithmWeights);
    pObj->set(RESOURCES_USAGE_LIMITS, resourcesUsageLimits);

    std::stringstream ostr;
    Poco::JSON::Stringifier::stringify(pObj, ostr);
    json = ostr.str();

    _isError = false;
  }
  catch(Poco::Exception& e)
  {
    _isError = true;
    errorMsg = e.displayText();
    errorCode = ERROR_UNSERIALIZE;
  }
  catch(std::exception& e)
  {
    _isError = true;
    errorMsg = e.what();
    errorCode = ERROR_SERIALIZE;
  }
  return !_isError;
}
//-----------------------------------------------------------------------------
bool RequestRouteMessage::unserialize(const std::string& json)
{
  resetError();
  clear();

  try
  {
    Poco::JSON::Parser parser;
    Poco::Dynamic::Var res = parser.parse(json);

    Poco::JSON::Object::Ptr pObj = res.extract<Poco::JSON::Object::Ptr>();

    Poco::Dynamic::Var tmp = pObj->get(MANAGER_ROLE);
    managerRole = convertVarToNumeric<unsigned int>(tmp, 0);

    Poco::JSON::Array::Ptr pNodesNames = pObj->getArray(NODES_NAMES);
    for (size_t i=0;i<pNodesNames->size();++i)
    {
      tmp = pNodesNames->get(i);
      if (tmp.isString())
        nodesNames.push_back(tmp.convert<std::string>());
    }

    tmp = pObj->get(RESOURCES_USAGE_ALGORITHM);
    resourcesUsageAlgorithm = convertVarToNumeric<unsigned int>(tmp, 0);

    tmp = pObj->get(RESOURCES_USAGE_ALGORITHM_WEIGHTS);
    if (tmp.isString())
      resourcesUsageAlgorithmWeights = tmp.convert<std::string>();

    tmp = pObj->get(RESOURCES_USAGE_LIMITS);
    if (tmp.isString())
      resourcesUsageLimits = tmp.convert<std::string>();

    _isError = false;
  }
  catch(Poco::JSON::JSONException& e)
  {
    _isError = true;
    errorMsg = e.displayText();
    errorCode = ERROR_UNSERIALIZE;
  }
  catch(Poco::Exception& e)
  {
    _isError = true;
    errorMsg = e.displayText();
    errorCode = ERROR_UNSERIALIZE;
  }
  return !_isError;
}
//-----------------------------------------------------------------------------
void RequestRouteMessage::clear(void)
{
  managerRole = 0;
  nodesNames.clear();
  resourcesUsageAlgorithm = 0;
  resourcesUsageAlgorithmWeights.clear();
  resourcesUsageLimits.clear();
}
//-----------------------------------------------------------------------------
//-----------------------------------------------------------------------------
HeartbeatMessage::HeartbeatMessage(std::vector<ClientWorkerItem>& queue_, const std::string& resources_)
: inherited(), queue(queue_), resources(resources_)
{
}
//-----------------------------------------------------------------------------
void HeartbeatMessage::packToObject(Poco::JSON::Object& obj, const std::vector<ClientWorkerItem>& workerItem)
{
  for (size_t i=0;i<workerItem.size();++i)
  {
    Poco::JSON::Object::Ptr pItemElem = new Poco::JSON::Object();
    if (!pItemElem.isNull())
    {
      packToObject(*pItemElem, workerItem[i].route);
      obj.set(workerItem[i].identity, pItemElem);
    }
  }
}
//-----------------------------------------------------------------------------
bool HeartbeatMessage::serialize(std::string& json)
{
  resetError();
  try
  {
    Poco::JSON::Object::Ptr pObj = new Poco::JSON::Object();
    if (pObj.isNull())
      throw Poco::Exception("Object instance was not created");

    Poco::JSON::Object::Ptr pRoutesElem = new Poco::JSON::Object();
    if (pRoutesElem.isNull())
      throw Poco::Exception("Object instance was not created");

    packToObject(*pRoutesElem, queue);
    //!<Add routes data to json
    pObj->set(HCE::handlers::HEARTBEAT_JSON_ROUTES, pRoutesElem);
    //!<Add resources data to json
    pObj->set(HCE::handlers::HEARTBEAT_JSON_RESOURCES, resources);

    std::stringstream ostr;
    Poco::JSON::Stringifier::stringify(pObj, ostr);
    json = ostr.str();

    _isError = false;
  }
  catch(Poco::Exception& e)
  {
    _isError = true;
    errorMsg = e.displayText();
    errorCode = ERROR_UNSERIALIZE;
  }
  catch(std::exception& e)
  {
    _isError = true;
    errorMsg = e.what();
    errorCode = ERROR_SERIALIZE;
  }
  return !_isError;
}
//-----------------------------------------------------------------------------
void HeartbeatMessage::unpackFromObject(Poco::JSON::Object& obj, std::vector<ClientWorkerItem>& workerItem)
{
  std::vector<std::string> names;
  obj.getNames(names);

  for (size_t i=0;i<names.size();++i)
  {
    Poco::JSON::Object::Ptr pObj = obj.getObject(names[i]);
    if (!pObj.isNull())
    {
      ClientWorkerItem clientWorker(names[i]);
      unpackFromObject(*pObj, clientWorker.route);
      workerItem.push_back(std::forward<ClientWorkerItem>(clientWorker));
    }
  }
}
//-----------------------------------------------------------------------------
bool HeartbeatMessage::unserialize(const std::string& json)
{
  clear();
  try
  {
    Poco::JSON::Parser parser;
    Poco::Dynamic::Var res = parser.parse(json);

    Poco::JSON::Object::Ptr pObj = res.extract<Poco::JSON::Object::Ptr>();

    Poco::JSON::Object::Ptr pRoutesElem = pObj->getObject(HCE::handlers::HEARTBEAT_JSON_ROUTES);
    if (!pRoutesElem.isNull())
    {
      unpackFromObject(*pRoutesElem, queue);
    }

    Poco::Dynamic::Var tmp = pObj->get(HCE::handlers::HEARTBEAT_JSON_RESOURCES);
    if (tmp.isString())
      resources = tmp.convert<std::string>();

    _isError = false;
  }
  catch(Poco::JSON::JSONException& e)
  {
    _isError = true;
    errorMsg = e.displayText();
    errorCode = ERROR_UNSERIALIZE;
  }
  catch(Poco::Exception& e)
  {
    _isError = true;
    errorMsg = e.displayText();
    errorCode = ERROR_UNSERIALIZE;
  }
  catch(std::exception& e)
  {
    _isError = true;
    errorMsg = e.what();
    errorCode = ERROR_UNSERIALIZE;
  }
  return !_isError;
}
//-----------------------------------------------------------------------------
std::string HeartbeatMessage::build(const std::string& title)
{
  std::string result = "";
  std::string json;
  serialize(json);

  if (title.empty())
    result = json;
  else
    result = (title+HEARTBEAT_MSG_DELIMITER+json);
  return result;
}
//-----------------------------------------------------------------------------
void HeartbeatMessage::extract(const std::string& message, const std::string& title)
{
  try
  {
    if (message.compare(0, title.length(), title)!=0)
      Poco::Exception("Title of message not found", ERROR_TITLE);

    size_t pos = message.find_first_of(HEARTBEAT_MSG_DELIMITER);
    if (pos==std::string::npos)
      Poco::Exception("Delimiter of title message not found", ERROR_TITLE);

    std::string json = message.substr(pos+1);
    unserialize(json);
  }
  catch(Poco::Exception& e)
  {
    _isError = true;
    errorMsg = e.displayText();
    errorCode = e.code();
  }
}
//-----------------------------------------------------------------------------
void HeartbeatMessage::clear(void)
{
  resetError();
  queue.clear();
  resources.clear();
}
//-----------------------------------------------------------------------------
//-----------------------------------------------------------------------------
PropertyMessage::PropertyMessage(const std::string& identity_, const std::string& property_)
: inherited(), identity(identity_), property(property_)
{
}
//-----------------------------------------------------------------------------
bool PropertyMessage::serialize(std::string& json)
{
  resetError();
  try
  {
    Poco::JSON::Object::Ptr pObj = new Poco::JSON::Object();
    if (pObj.isNull())
      throw Poco::Exception("Object instance was not created");

    pObj->set(identity, property);

    std::stringstream ostr;
    Poco::JSON::Stringifier::stringify(pObj, ostr);
    json = ostr.str();

    _isError = false;
  }
  catch(Poco::Exception& e)
  {
    _isError = true;
    errorMsg = e.displayText();
    errorCode = ERROR_UNSERIALIZE;
  }
  catch(std::exception& e)
  {
    _isError = true;
    errorMsg = e.what();
    errorCode = ERROR_SERIALIZE;
  }
  return !_isError;
}
//-----------------------------------------------------------------------------
bool PropertyMessage::unserialize(const std::string& json)
{
  clear();
  try
  {
    Poco::JSON::Parser parser;
    Poco::Dynamic::Var res = parser.parse(json);

    Poco::JSON::Object::Ptr pObj = res.extract<Poco::JSON::Object::Ptr>();

    if (!pObj.isNull())
    {
      std::vector<std::string> names;
      pObj->getNames(names);

      for (size_t i=0;i<names.size();++i)
      {
        identity = names[i];
        Poco::Dynamic::Var tmp = pObj->get(identity);
        if (tmp.isString())
          property = tmp.convert<std::string>();
        break;
      }
    }
    _isError = false;
  }
  catch(Poco::JSON::JSONException& e)
  {
    _isError = true;
    errorMsg = e.displayText();
    errorCode = ERROR_UNSERIALIZE;
  }
  catch(Poco::Exception& e)
  {
    _isError = true;
    errorMsg = e.displayText();
    errorCode = ERROR_UNSERIALIZE;
  }
  catch(std::exception& e)
  {
    _isError = true;
    errorMsg = e.what();
    errorCode = ERROR_UNSERIALIZE;
  }
  return !_isError;
}
//-----------------------------------------------------------------------------
std::string PropertyMessage::build(const std::string& title)
{
  std::string result = "";
  std::string json;
  serialize(json);

  if (title.empty())
    result = json;
  else
    result = (title+PROPERTY_MSG_DELIMITER+json);
  return result;
}
//-----------------------------------------------------------------------------
void PropertyMessage::extract(const std::string& message, const std::string& title)
{
  try
  {
    if (message.compare(0, title.length(), title)!=0)
      Poco::Exception("Title of message not found", ERROR_TITLE);

    size_t pos = message.find_first_of(PROPERTY_MSG_DELIMITER);
    if (pos==std::string::npos)
      Poco::Exception("Delimiter of title message not found", ERROR_TITLE);

    std::string json = message.substr(pos+1);
    unserialize(json);
  }
  catch(Poco::Exception& e)
  {
    _isError = true;
    errorMsg = e.displayText();
    errorCode = e.code();
  }
}
//-----------------------------------------------------------------------------
void PropertyMessage::clear(void)
{
  resetError();
  property.clear();
}
//-----------------------------------------------------------------------------
//-----------------------------------------------------------------------------
} // end namespace handlers
} // end namespace HCE 
