/**
 * @nodemcpp.cpp
 *
 * \mainpage %HCE project node model
 * %HCE project node model application v1.4.0a
 * Implements complete functional model of cluster node application.
 * Supports four basic runtime modes: router, proxy shard, data shard and data node.
 * Uses zmq PPP-based protocol and connection schema, heartbit algorithm, admin management handler and so on.
 * This is an atomic brick for %HCE project architecture structure schema construction. \n\n
 * Sphinx data provider\n
 * First implementation of data storage provider for distributed cluster system that can be built - is Sphinx (c).
 * HCE project node implements complete functionality to: index, search, merge results and storage management of Sphinx search.
 * Future extension will add clusterization of another not distributed DB data storages like sqlite.
 *
 * @author bgv <bgv.hce@gmail.com>
 * @link http://hierarchical-cluster-engine.com/
 * @copyright Copyright &copy; 2013-2015 HCE project, IOIX Ukraine
 * @license http://hierarchical-cluster-engine.com/license/
 * @package HCE project node
 * @since 0.1
 */

#include "main.hpp"

//!<Global pointer to admin handlers
HCE::handlers::Admin* pHandlerAdmin = nullptr;
//-----------------------------------------------------------------------------
void applicationSignalHandler(int sig)
{
  std::stringstream outMsg;
  outMsg << "Application recieved signal number: " << sig;
  Poco::Logger::root().log(Poco::Message("HCE cluster node", outMsg.str(), Poco::Message::Priority::PRIO_DEBUG));

  if (pHandlerAdmin)
  {
    pHandlerAdmin->stopClients();
    pHandlerAdmin->shutdown();
  }
}
//-----------------------------------------------------------------------------
void signalInstall(void)
{
  struct sigaction newAction, oldAction;
  newAction.sa_handler = applicationSignalHandler;
  newAction.sa_flags = 0;
  sigemptyset(&newAction.sa_mask);

  sigaction (SIGHUP, 0, &oldAction);
  if (oldAction.sa_handler != SIG_IGN)
    sigaction (SIGHUP, &newAction, 0);

  sigaction (SIGTERM, 0, &oldAction);
  if (oldAction.sa_handler != SIG_IGN)
    sigaction (SIGTERM, &newAction, 0);

  sigset_t blockMask;
  sigemptyset (&blockMask);
  sigaddset (&blockMask, SIGCHLD);
  sigaddset (&blockMask, SIGTSTP);
  sigaddset (&blockMask, SIGTTOU);
  sigaddset (&blockMask, SIGTTIN);
  sigprocmask (SIG_BLOCK, &blockMask, 0);
}
//-----------------------------------------------------------------------------
enum BgStatus
{
  BACKGROUND_ERROR=-1,
  BACKGRAUND_OK=0,
  IGNORE_PARENT=1
};
//-----------------------------------------------------------------------------
BgStatus gotoBackgraund(void)
{
  BgStatus result = BACKGROUND_ERROR;

  /* Fork off the parent process */
  switch(fork())
  {
    case -1: exit(EXIT_FAILURE);
      break;
    case 0:
      {
        /* Change the file mode mask */
        umask(0);

        /* Create a new SID for the child process */
        if(setsid() != -1)
        {
          /* Change the current working directory */
          if(chdir("/") != -1)
          {
            /* Close out the standard file descriptors */
            if(close(STDIN_FILENO) != -1)
              if(close(STDOUT_FILENO) != -1)
                if(close(STDERR_FILENO) != -1)
                  if(open("/dev/null",O_RDONLY) != -1)
                    if(open("/dev/null",O_WRONLY) != -1)
                      if(dup(1) != -1)
                        result = BACKGRAUND_OK;
          }
        }
      }
      break;
    /* If we got a good PID, then
      we can exit the parent process. */
    default: result = IGNORE_PARENT; // here instead 'exit(EXIT_SUCCESS);'
  }
  return result;
}
//-----------------------------------------------------------------------------
int main(int argc, char* argv[]){
  //!<Init actual cli args map with default values
  std::map<std::string, std::string> actualOptions;
  actualOptions.insert(std::pair<std::string, std::string>(NODE_CLI_NAME, ""));
  actualOptions.insert(std::pair<std::string, std::string>(NODE_CLI_ADMIN, "tcp://*:5546"));
  actualOptions.insert(std::pair<std::string, std::string>(NODE_CLI_CLIENT, "tcp://*:5556"));
  actualOptions.insert(std::pair<std::string, std::string>(NODE_CLI_SERVER, "tcp://*:5557"));
  actualOptions.insert(std::pair<std::string, std::string>(NODE_CLI_DATAMODE, "1"));
  actualOptions.insert(std::pair<std::string, std::string>(NODE_CLI_ROLE, "router"));
  actualOptions.insert(std::pair<std::string, std::string>(NODE_CLI_INI_FILE, ""));
  actualOptions.insert(std::pair<std::string, std::string>(NODE_CLI_LOG_MODE, "hce-node_log.ini"));
  actualOptions.insert(std::pair<std::string, std::string>(NODE_CLI_HELP, " "));
  actualOptions.insert(std::pair<std::string, std::string>(NODE_CLI_VERSION, " "));
  actualOptions.insert(std::pair<std::string, std::string>(NODE_CLI_DEMONIZE, "0"));

  //!<Init node modes map
  std::map<std::string, int> nodeModes;
  nodeModes.insert(std::pair<std::string, int>(NODE_CLI_ROLE_RMANAGER_RND, HCE::handlers::NODE_MODE_SHARD_RND));
  nodeModes.insert(std::pair<std::string, int>(NODE_CLI_ROLE_ROUTER, HCE::handlers::NODE_MODE_ROUTER));
  nodeModes.insert(std::pair<std::string, int>(NODE_CLI_ROLE_REPLICA, HCE::handlers::NODE_MODE_DATA));
  nodeModes.insert(std::pair<std::string, int>(NODE_CLI_ROLE_RMANAGER, HCE::handlers::NODE_MODE_SHARD));
  nodeModes.insert(std::pair<std::string, int>(NODE_CLI_ROLE_RMANAGER_ROUND_ROBIN, HCE::handlers::NODE_MODE_SHARD));
  nodeModes.insert(std::pair<std::string, int>(NODE_CLI_ROLE_RMANAGER_RESOURCES_USAGE, HCE::handlers::NODE_MODE_SHARD_RU));
  nodeModes.insert(std::pair<std::string, int>(NODE_CLI_ROLE_SMANAGER, HCE::handlers::NODE_MODE_PROXY));

  //!<Fill supported cli args
  Poco::Util::OptionSet knownOptions;
  knownOptions.addOption(Poco::Util::Option(NODE_CLI_NAME, "n", "node name", false, NODE_CLI_NAME));
  knownOptions.addOption(Poco::Util::Option(NODE_CLI_ADMIN, "a", "admin listen zmq connection", false, NODE_CLI_ADMIN));
  knownOptions.addOption(Poco::Util::Option(NODE_CLI_CLIENT, "c", "client listen zmq connection", false, NODE_CLI_CLIENT));
  knownOptions.addOption(Poco::Util::Option(NODE_CLI_SERVER, "s", "server listen zmq connection", false, NODE_CLI_SERVER));
  knownOptions.addOption(Poco::Util::Option(NODE_CLI_DATAMODE, "m", "mode 1 - regular, 0 - test only", false, NODE_CLI_DATAMODE));
  knownOptions.addOption(Poco::Util::Option(NODE_CLI_ROLE, "r", "node role: router, smanager, rmanager, replica", false, NODE_CLI_ROLE));
  knownOptions.addOption(Poco::Util::Option(NODE_CLI_INI_FILE, "i", "ini file name", false, NODE_CLI_INI_FILE));
  knownOptions.addOption(Poco::Util::Option(NODE_CLI_LOG_MODE, "l", "log configuration file name", false, NODE_CLI_LOG_MODE));
  knownOptions.addOption(Poco::Util::Option(NODE_CLI_HELP, "h", "display help", false, NODE_CLI_HELP));
  knownOptions.addOption(Poco::Util::Option(NODE_CLI_VERSION, "v", "display version", false, NODE_CLI_VERSION));
  knownOptions.addOption(Poco::Util::Option(NODE_CLI_DEMONIZE, "d", "demonize", false, NODE_CLI_DEMONIZE));

  //!<Parse and check cli args
  Poco::Util::OptionProcessor op(knownOptions);
  op.setUnixStyle(true);
  for(int i = 1; i < argc; i++){
    std::string name;
    std::string value;
    try{
      if(op.process(std::string(argv[i]), name, value)){
        if((name != NODE_CLI_HELP && !value.empty()) || (name == NODE_CLI_HELP && value.empty()) || (name == NODE_CLI_VERSION && value.empty()) ){
          //std::cout << "[" << name << "]" << "=[" << value << "]" << std::endl;
          std::map<std::string, std::string>::iterator actualOption;
          actualOption = actualOptions.find(name);
          if(actualOption != actualOptions.end()){
            //std::cout << "found: " << actualOption->second << std::endl;
            actualOption->second = value;
          }
        }
      }
    } catch(Poco::Exception& exc){
      std::cerr << exc.displayText() << std::endl;
      ::exit(1);
    }
  }

  std::map<std::string, std::string>::iterator actualOption;
  //!<Display help message and terminate
  actualOption = actualOptions.find(NODE_CLI_HELP);
  if(actualOption != actualOptions.end() && actualOption->second == ""){
    std::cout << "Version: " << APP_VERSION << "\nUsage: " << argv[0] << " [--" << NODE_CLI_HELP << "] [--" << NODE_CLI_ADMIN << "=<tcp://<admin_address|*:port>] [--" << NODE_CLI_CLIENT
        << "=[tcp://<client_address:port]] [--" << NODE_CLI_SERVER << "=<tcp://<server_address|*:port>] [--" << NODE_CLI_NAME << "=<nodeName>] [--" << NODE_CLI_DATAMODE
        << "=<0 - randomizer, 1 - real data source>] " << "[--" << NODE_CLI_ROLE << "=<router | smanager | rmanager | replica> [--" << NODE_CLI_INI_FILE << "=<iniFile>] [--"
        << NODE_CLI_LOG_MODE << "=<logger config file name>] [--" << NODE_CLI_VERSION << "] [--" << NODE_CLI_DEMONIZE << "=<0 - without background, 1 - use background>]"
        << std::endl;
    ::exit(0);
  }

  //!<Display version message and terminate
  actualOption = actualOptions.find(NODE_CLI_VERSION);
  if(actualOption != actualOptions.end() && actualOption->second == ""){
    std::cout << APP_VERSION << std::endl;
    ::exit(0);
  }

  //If name not set - init with hostname
  actualOption = actualOptions.find(NODE_CLI_NAME);
  if(actualOption != actualOptions.end() && actualOption->second == ""){
    //Set name from hostname
    struct utsname name;
    if(::uname(&name) > -1){
      actualOption->second = name.nodename;
    }
  }

  /*
   for(actualOption = actualOptions.begin(); actualOption != actualOptions.end(); ++actualOption){
   std::cout << actualOption->first << "=[" << actualOption->second << "]" << std::endl;
   }
   */

  //!<Init list exist loggers
  Poco::Util::IniFileConfiguration::Keys loggers;
  loggers.push_back(HCE_MAIN_LOGGER);
  //Init vars
  //!<Node name
  std::string nodeName = actualOptions.find(NODE_CLI_NAME)->second;
  //!<Connection strings
  std::string adminConnection = actualOptions.find(NODE_CLI_ADMIN)->second;
  unsigned int nodeMode = nodeModes.find(actualOptions.find(NODE_CLI_ROLE)->second)->second;
  //!<Node role-dependent connection strings vars
  std::string dataServerConnection = "0";
  std::string dataClientConnection = "0";
  std::string routerServerConnection = "0";
  if(nodeMode == HCE::handlers::NODE_MODE_ROUTER){
    //Init router
    routerServerConnection = actualOptions.find(NODE_CLI_CLIENT)->second;
    dataServerConnection = actualOptions.find(NODE_CLI_SERVER)->second;
  }else if(nodeMode == HCE::handlers::NODE_MODE_DATA){
    //Init replica
    dataClientConnection = actualOptions.find(NODE_CLI_CLIENT)->second;
    //!<Add to list loggers name task logger
    loggers.push_back(DRCE_TASKS_LOGGER);
  }else{
    //Init shard manager or replica manager
    dataServerConnection = actualOptions.find(NODE_CLI_SERVER)->second;
    dataClientConnection = actualOptions.find(NODE_CLI_CLIENT)->second;
  }
  //!<Init ini file var use absolute path
  Poco::Path iniFilePath(actualOptions.find(NODE_CLI_INI_FILE)->second);
  std::string iniFile = iniFilePath.absolute().toString();
  //!<Init client data mode
  unsigned int clientDataMode = 1;
  std::istringstream(actualOptions.find(NODE_CLI_DATAMODE)->second) >> clientDataMode;
  //!<Read config file name
  std::string configFileName;
  std::istringstream(actualOptions.find(NODE_CLI_LOG_MODE)->second) >> configFileName;
  //!<Init list exist loggers
//  Poco::Util::IniFileConfiguration::Keys loggers;
//  loggers.push_back(HCE_MAIN_LOGGER);
//  loggers.push_back(DRCE_TASKS_LOGGER);
  //!<Load config file
  HCE::handlers::LoggerConfigLoader loggerConfigLoader(nodeName, loggers);
  if (!loggerConfigLoader.loadConfig(configFileName))
  {
    std::cerr << loggerConfigLoader.getErrorMsg() << std::endl;
    std::cerr << "Logger ini file load error, possible file not found, not proper format or another cause..." << std::endl << "Node not started!" << std::endl;
    ::exit(2);
  }

  //!<Init client demonize
  unsigned int clientDemonize = 1;
  std::istringstream(actualOptions.find(NODE_CLI_DEMONIZE)->second) >> clientDemonize;
  //!<Init log/debug level
  unsigned int DEBUG_LEVEL = loggerConfigLoader.getDebugLevel();
  //!<Create logger stream
  HCE::LoggerStream logger(nodeName);

  //!<Check nessesary of demonization
  if (clientDemonize == 1) {
    //!<Demonization of application
    if (daemon(0, 0) == 0) {
      logger.log(HCE::LoggerStream::Priority::PRIO_INFORMATION) << "Demonization success..." << HCE::flush;
    }
    else {
      std::cerr << "Demonization fail: " << strerror(errno) << std::endl;
      ::exit(3);
    }
/*
    BgStatus rc = gotoBackgraund();
    if (rc == BACKGROUND_ERROR){
      std::cerr << "Demonization fail: " << strerror(errno) << std::endl;
      ::exit(3);
    }
    else if (rc == IGNORE_PARENT){
      return 0;
    }
    else{
      logger.log(HCE::LoggerStream::Priority::PRIO_INFORMATION) << "Demonization success..." << HCE::flush;
    }
*/
  }

  //!<Bind signals with handlers
  signalInstall();
  //!<Check zmq library version
  s_version_assert(2, 1);
  //!<Start time mark init
  int64_t startMark = s_clock();
  //!<Init random generator
  srandom((unsigned)time(NULL));
  //!<Create zmq context handler
  zmq::context_t context(1);

  //!<Create thread identifiers
  pthread_t threadDataServer = 0;
  pthread_t threadDataProcessor = 0;
  pthread_t threadDataClient = 0;
  pthread_t threadDataReducer = 0;
  pthread_t threadRouterServer = 0;
  pthread_t threadAdmin = 0;

  //!<General application start header
  logger.log(HCE::LoggerStream::Priority::PRIO_INFORMATION)
      << std::setfill('0') << std::setw(8) << (int)(s_clock() - startMark) << " [] HCE cluster node, version " << APP_VERSION << HCE::flush;

  //!<Create admin handler
  logger.log(HCE::LoggerStream::Priority::PRIO_INFORMATION)
      << std::setfill('0') << std::setw(8) << (int)(s_clock() - startMark) << " [] ADMIN handler creation..." << HCE::flush;

  HCE::handlers::Admin handlerAdmin(std::string("Admin"), context, nodeName, adminConnection, startMark, static_cast<unsigned char>(DEBUG_LEVEL));
  pHandlerAdmin = &handlerAdmin;
  //!<Create admin handler's thread
  if(::pthread_create(&threadAdmin, NULL, &HCE::handlers::Admin::main, &handlerAdmin)){
    logger.log(HCE::LoggerStream::Priority::PRIO_FATAL) << "Error create thread [" << handlerAdmin.getName() << "]!" << HCE::flush;
  }else{
    //!<Check node mode and create correspondent handlers and threads
    switch(nodeMode){
      case 0:
        //!<Proxy mode (SHARD MANAGER)
      case 1:
      case 4:
      case 5: {
        if(nodeMode == HCE::handlers::NODE_MODE_PROXY || nodeMode == HCE::handlers::NODE_MODE_ROUTER){
          //Shard manager mode
          logger.log(HCE::LoggerStream::Priority::PRIO_INFORMATION) << std::setfill('0') << std::setw(8) << (int)(s_clock() - startMark)
              << " [] NODE IN SHARD MANAGER MODE - sends requests to all connected data nodes multicast way" << HCE::flush;
        }else{
          //Replica manager mode
          logger.log(HCE::LoggerStream::Priority::PRIO_INFORMATION) << std::setfill('0') << std::setw(8) << (int)(s_clock() - startMark)
              << " [] NODE IN REPLICA MANAGER MODE - sends requests to only one of connected data nodes round-robin way" << HCE::flush;
        }
        //!<Create data reducer's thread
        HCE::handlers::DataReducerProxy handlerDataReducerProxy(std::string("DataReducerProxy"), context, nodeMode, startMark, (unsigned char)DEBUG_LEVEL, iniFile);
        if(::pthread_create(&threadDataReducer, NULL, &HCE::handlers::DataReducerProxy::main, &handlerDataReducerProxy)){
          logger.log(HCE::LoggerStream::Priority::PRIO_FATAL) << "Error create thread [" << handlerDataReducerProxy.getName() << "]!" << HCE::flush;
        }else{
          //!<Let bind sockets to stabilize...
          s_sleep(100);
          //!<Create dataServer's thread
          HCE::handlers::DataServerProxy handlerDataServerProxy(std::string("DataServerProxy"), context, nodeName, dataServerConnection, nodeMode, startMark, (unsigned char)DEBUG_LEVEL, iniFile);
          if(::pthread_create(&threadDataServer, NULL, &HCE::handlers::DataServerProxy::main, &handlerDataServerProxy)){
            logger.log(HCE::LoggerStream::Priority::PRIO_FATAL) << "Error create thread [" << handlerDataServerProxy.getName() << "]!" << HCE::flush;
          }else{
            //!<Let bind sockets to stabilize...
            s_sleep(100);
            //!<Create and run dataclient's proxy handler
            HCE::handlers::DataClientProxy handlerDataClientProxy(std::string("DataClientProxy"), context, nodeName, dataClientConnection, startMark, (unsigned char)DEBUG_LEVEL, iniFile);
            if(::pthread_create(&threadDataClient, NULL, &HCE::handlers::DataClientProxy::main, &handlerDataClientProxy)){
              logger.log(HCE::LoggerStream::Priority::PRIO_FATAL) << "Error create thread [" << handlerDataClientProxy.getName() << "]!" << HCE::flush;
            }else{
              //!<Wait until admin thread is not finished
              ::pthread_join(threadAdmin, NULL);
              logger.log(HCE::LoggerStream::Priority::PRIO_DEBUG) << "The thread [" << pHandlerAdmin->getName() << "] was joined!" << HCE::flush;

              //!<Shutdown all handlers
              handlerDataClientProxy.shutdown();
              handlerDataServerProxy.shutdown();
              handlerDataReducerProxy.shutdown();
              //!<Wait until dataclient thread is not finished
              ::pthread_join(threadDataClient, NULL);
              logger.log(HCE::LoggerStream::Priority::PRIO_DEBUG) << "The thread [" << handlerDataClientProxy.getName() << "] was joined!" << HCE::flush;
              //!<Wait until dataServer thread is not finished
              ::pthread_join(threadDataServer, NULL);
              logger.log(HCE::LoggerStream::Priority::PRIO_DEBUG) << "The thread [" << handlerDataServerProxy.getName() << "] was joined!" << HCE::flush;
              //!<Wait until datareducer thread is not finished
              ::pthread_join(threadDataReducer, NULL);
              logger.log(HCE::LoggerStream::Priority::PRIO_DEBUG) << "The thread [" << handlerDataReducerProxy.getName() << "] was joined!" << HCE::flush;
            }
          }
        }
        break;
      }
      case 2: {
        //!<Data node mode
        logger.log(HCE::LoggerStream::Priority::PRIO_INFORMATION) << std::setfill('0') << std::setw(8) << (int)(s_clock() - startMark) << " [] NODE IN DATA MODE" << HCE::flush;
        //!<Create data processor thread
        HCE::handlers::DataProcessorData handlerDataProcessorData(nodeName, std::string("DataProcessorData"), context, startMark, (unsigned char)DEBUG_LEVEL, iniFile);
        if(::pthread_create(&threadDataProcessor, NULL, &HCE::handlers::DataProcessorData::main, &handlerDataProcessorData)){
          logger.log(HCE::LoggerStream::Priority::PRIO_FATAL) << "Error create thread [" << handlerDataProcessorData.getName() << "]!" << HCE::flush;
        }else{
          //!<Let sockets to stabilize...
          s_sleep(100);
          //!<Create dataclient's thread data
          HCE::handlers::DataClientData handlerDataClientData(std::string("DataClientData"), context, nodeName, dataClientConnection, (unsigned char)clientDataMode, startMark,
              (unsigned char)DEBUG_LEVEL, iniFile);
          if(::pthread_create(&threadDataClient, NULL, &HCE::handlers::DataClientData::main, &handlerDataClientData)){
            logger.log(HCE::LoggerStream::Priority::PRIO_FATAL) << "Error create thread [" << handlerDataClientData.getName() << "]!" << HCE::flush;
          }else{
            //!<Wait to finish admin thread
            ::pthread_join(threadAdmin, NULL);
            logger.log(HCE::LoggerStream::Priority::PRIO_DEBUG) << "The thread [" << handlerAdmin.getName() << "] was joined!" << HCE::flush;

            //!<Shutdown dataclient thread
            handlerDataClientData.shutdown();
            //!<Wait until dataclient thread is not finished
            ::pthread_join(threadDataClient, NULL);
            logger.log(HCE::LoggerStream::Priority::PRIO_DEBUG) << "The thread [" << handlerDataClientData.getName() << "] was joined!" << HCE::flush;

            //!<Shutdown dataprocessor thread
            handlerDataProcessorData.shutdown();
            //!<Wait until dataprocessor thread is not finished
            ::pthread_join(threadDataProcessor, NULL);
            logger.log(HCE::LoggerStream::Priority::PRIO_DEBUG) << "The thread [" << handlerDataProcessorData.getName() << "] was joined!" << HCE::flush;
          }
        }
        break;
      }
      case 3: {
        //!<Router node mode
        logger.log(HCE::LoggerStream::Priority::PRIO_INFORMATION) << std::setfill('0') << std::setw(8) << (int)(s_clock() - startMark) << " [] NODE IN ROUTER MODE" << HCE::flush;
        //!<Create data reducer's thread
        HCE::handlers::DataReducerProxy handlerDataReducerProxy(std::string("DataReducerProxy"), context, nodeMode, startMark, (unsigned char)DEBUG_LEVEL, iniFile);
        if(::pthread_create(&threadDataReducer, NULL, &HCE::handlers::DataReducerProxy::main, &handlerDataReducerProxy)){
          logger.log(HCE::LoggerStream::Priority::PRIO_FATAL) << "Error create thread [" << handlerDataReducerProxy.getName() << "]!" << HCE::flush;
        }else{
          //!<Let bind sockets to stabilize...
          s_sleep(100);
          //!<Create dataServer's thread
          HCE::handlers::DataServerProxy handlerDataServerProxy(std::string("DataServerProxy"), context, nodeName, dataServerConnection, nodeMode, startMark, (unsigned char)DEBUG_LEVEL, iniFile);
          if(::pthread_create(&threadDataServer, NULL, &HCE::handlers::DataServerProxy::main, &handlerDataServerProxy)){
            logger.log(HCE::LoggerStream::Priority::PRIO_FATAL) << "Error create thread [" << handlerDataServerProxy.getName() << "]!" << HCE::flush;
          }else{
            //!<Let bind sockets to stabilize...
            s_sleep(100);
            //!<Create router's thread
            HCE::handlers::RouterServerProxy handlerRouterServerProxy(std::string("RouterServerProxy"), context, nodeName, routerServerConnection, startMark, (unsigned char)DEBUG_LEVEL, iniFile);
            if(::pthread_create(&threadRouterServer, NULL, &HCE::handlers::RouterServerProxy::main, &handlerRouterServerProxy)){
              logger.log(HCE::LoggerStream::Priority::PRIO_FATAL) << "Error create thread [" << handlerRouterServerProxy.getName() << "]!" << HCE::flush;
            }else{
              //!<Wait to finish data server thread
              ::pthread_join(threadAdmin, NULL);
              logger.log(HCE::LoggerStream::Priority::PRIO_DEBUG) << "The thread [" << handlerAdmin.getName() << "] was joined!" << HCE::flush;
              //!<Shutdown all handlers
              handlerRouterServerProxy.shutdown();
              handlerDataServerProxy.shutdown();
              handlerDataReducerProxy.shutdown();
              //!<Wait until routerserver thread is not finished
              ::pthread_join(threadRouterServer, NULL);
              logger.log(HCE::LoggerStream::Priority::PRIO_DEBUG) << "The thread [" << handlerRouterServerProxy.getName() << "] was joined!" << HCE::flush;
              //!<Wait until dataServer thread is not finished
              ::pthread_join(threadDataServer, NULL);
              logger.log(HCE::LoggerStream::Priority::PRIO_DEBUG) << "The thread [" << handlerDataServerProxy.getName() << "] was joined!" << HCE::flush;
              //!<Wait until datareducer thread is not finished
              ::pthread_join(threadDataReducer, NULL);
              logger.log(HCE::LoggerStream::Priority::PRIO_DEBUG) << "The thread [" << handlerDataReducerProxy.getName() << "] was joined!" << HCE::flush;
            }
          }
        }
        break;
      }
      default: {
        logger.log(HCE::LoggerStream::Priority::PRIO_FATAL) << "Unsupported node mode" << HCE::flush;
        break;
      }
    }
  }

  logger.log(HCE::LoggerStream::Priority::PRIO_INFORMATION) << std::setfill('0') << std::setw(8) << (int)(s_clock() - startMark) << " [] Node was shutdown normally" << HCE::flush;

  return 0;
}
