/**
 * @file DRCECommonTask.hpp
 *
 * Implementation class DRCECommonTask of a Distributed Remote Command Execution functional object in HCE cluster node.
 *
 * @author Alexander Vybornyh <alexander.hce.cluster@gmail.com>
 * @link http://hierarchical-cluster-engine.com/
 * @copyright Copyright &copy; 2013 IOIX Ukraine
 * @license http://hierarchical-cluster-engine.com/license/
 * @package HCE project node API
 * @since 0.1
 */

#ifndef DRCE_COMMON_TASK_HPP
#define DRCE_COMMON_TASK_HPP

#include <Poco/Task.h>
#include <Poco/Timestamp.h>
#include <vector>

#include "DRCENodeOptions.hpp"
#include "CustomMessage.hpp"
#include "CommandExecutor.hpp"
#include "DRCEFilesList.hpp"
#include "DRCEResultData.hpp"
#include "DRCETaskRequestSetExecute.hpp"
#include "DRCEAsyncTasksQueue.hpp"
#include "DRCEResourceMonitor.hpp"
#include "Process.hpp"
#include "DRCEError.hpp"
#include "DRCEInputJsonMessage.hpp"

namespace HCE
{
namespace drce
{
class DRCEAsyncTasksQueue;
//-----------------------------------------------------------------------------
class DRCECommonTask : public Poco::Task
{
public:
  typedef std::vector<std::pair<unsigned int, SessionOptions::CleanupFlag> > SubtasksList;
public:
  class ApplyPidHandler
  {
  public:
    ApplyPidHandler(DRCEAsyncTasksQueue& asyncTasksQueue_,
        AsyncTasks& asyncTask_,
        DRCETaskRequest::TaskState allowedTaskState_,
        SessionOptions::ThreadMode threadMode_);
    ~ApplyPidHandler(void) {}

    void operator()(pid_t pid);
  private:
    DRCEAsyncTasksQueue& asyncTasksQueue;
    AsyncTasks& asyncTask;
    DRCETaskRequest::TaskState allowedTaskState;
    SessionOptions::ThreadMode threadMode;
  };
public:
  DRCECommonTask(const std::string& taskName_, DRCENodeOptions& nodeOptions_, CustomMessage& message_,
      const DRCETaskRequestSetExecute& taskRequestSetExecute_, DRCEAsyncTasksQueue& asyncTasksQueue_,
      DRCEResourceMonitor& resourceMonitor_, const DRCEInputJsonMessage& inputJsonMessage, unsigned int parentTaskId_);

  virtual void runTask(void) = 0;
  virtual DRCEResultDataItem execute(Process::CallbackFunction fn) throw (Poco::Exception) = 0;


  void updateTasks(DRCEResultDataItem& resultDataItem, unsigned int& progressCount);
  static void updateTasks(DRCEResultDataItem& resultDataItem,
      DRCENodeOptions& nodeOptions,
      CustomMessage& message,
      DRCEAsyncTasksQueue& asyncTasksQueue,
      unsigned int& progressCount,
      DRCECommonTask* pCommonTask=nullptr);

  void makeFilesBefore(DRCEFilesList& filesList) throw (Poco::Exception);
  void makeFilesAfter(DRCEFilesList& filesList, DRCEResultDataItem& resultDataItem);

  static size_t getElapsedTimeMsec(const Poco::Timestamp& tsStart);  // return msec

  unsigned int getParentTaskId(void) const {return parentTaskId;}

  void saveRequestData(void);
  void saveResultData(DRCEResultDataItem& resultDataItem);
protected:
  DRCENodeOptions& nodeOptions;
  CustomMessage& message;
  DRCETaskRequestSetExecute taskRequestSetExecute;
  DRCEAsyncTasksQueue& asyncTasksQueue;
  DRCEResourceMonitor& resourceMonitor;
  DRCEInputJsonMessage inputJsonMessage;
  unsigned int parentTaskId;
public:
  void executeSubtasks(DRCEResultDataItem& resultDataItem);
  void cleanup(bool needCleanup=false);
  static void getSubtasksList(DRCEInputJsonMessage& inputJsonMessage, SubtasksList& subtasks);

  static DRCEInputJsonMessage getRequestData(unsigned int taskId,
      DRCENodeOptions& nodeOptions,
      CustomMessage& message);

  static void waitUpdateAllTasks(unsigned int taskId,
      DRCENodeOptions& nodeOptions,
      CustomMessage& message,
      DRCEAsyncTasksQueue& asyncTasksQueue,
      Poco::Task* pTask=nullptr);

  static void waitEndAllTask(DRCEInputJsonMessage& inputJsonMessage,
      DRCEAsyncTasksQueue& asyncTasksQueue,
      Poco::Task* pTask=nullptr);

  static void cleanup(unsigned int taskId,
      DRCENodeOptions& nodeOptions,
      CustomMessage& message,
      bool needCleanup=false);

  static void cleanup(DRCEInputJsonMessage& inputJsonMessage,
      DRCENodeOptions& nodeOptions,
      CustomMessage& message,
      bool needCleanup=false);

  static void saveRequestData(DRCEInputJsonMessage& inputJsonMessage,
      DRCENodeOptions& nodeOptions,
      CustomMessage& message);

  static void saveRequest(unsigned int taskId,
      DRCENodeOptions& nodeOptions,
      CustomMessage& message);

  static void saveRequest(DRCEInputJsonMessage& inputJsonMessage,
      DRCENodeOptions& nodeOptions,
      CustomMessage& message);

  static void saveResultData(DRCEResultDataItem& resultDataItem,
      DRCENodeOptions& nodeOptions,
      CustomMessage& message,
      DRCEAsyncTasksQueue& asyncTasksQueue);

  static DRCEResultDataItem makeResultDataItem(DRCEInputJsonMessage& inputJsonMessage,
      DRCENodeOptions& nodeOptions,
      CustomMessage& message,
      DRCETaskRequest::RequestType requestType,
      DRCETaskRequest::TaskState taskState,
      unsigned int errorCode=NO_ERROR,
      const std::string& errorMessage="",
      unsigned int pid=0,
      size_t timeElapsed=0,
      unsigned int exitStatus=0);

  static unsigned int getHostParentTaskId(unsigned int taskId,
      DRCEAsyncTasksQueue& asyncTasksQueue);

  static DRCEResultDataItem extractStatusResultDataItem(unsigned int taskId,
      DRCENodeOptions& nodeOptions,
      DRCEAsyncTasksQueue& asyncTasksQueue);

  static DRCEResultDataItem extractDataResultDataItem(unsigned int taskId,
      DRCENodeOptions& nodeOptions,
      DRCEAsyncTasksQueue& asyncTasksQueue);

  static void safeCleanup(const std::string& path);
  static DRCEResultDataItem extractResultDataItem(const std::string& fileName);
private:
  typedef Poco::Task inherited;
};
//-----------------------------------------------------------------------------
//-----------------------------------------------------------------------------
} // end namespace drce 
} // end namespace HCE 

#endif // DRCE_COMMON_TASK_HPP
