#include <sstream>
#include <Poco/File.h>
#include <Poco/Logger.h>

#include "DRCEFileStream.hpp"
#include "DRCEMessageConst.hpp"
#include "DRCEError.hpp"
#include "DRCEResultDataSerializator.hpp"

#include "DRCEFileBuilder.hpp"

namespace HCE
{
namespace drce
{
//-----------------------------------------------------------------------------
FileBuilder::FileBuilder(CustomMessage& message_, const std::string& dir_)
: message(message_), dir(dir_)
{
}
//-----------------------------------------------------------------------------
//-----------------------------------------------------------------------------
DataFileBuilder::DataFileBuilder(CustomMessage& message_, const std::string& dir_)
: inherited(message_, dir_)
{
}
//-----------------------------------------------------------------------------
void DataFileBuilder::build(const DRCEResultData& resultData) throw (Poco::Exception)
{
  if (resultData.getItemsCount()!=1)
    throw Poco::Exception(message(message_const::EMPTY_OBJECT_INSTANCE), ERROR_BAD_DATA_CONTENT);

  std::string json = makeJson(resultData);
  if (!json.empty() && resultData.getDataItem(0).getRequestId())
  {
    Poco::File dataFile(FileStream::getDataFileName(dir, resultData.getDataItem(0).getRequestId()));
    if (dataFile.exists())
      dataFile.remove();

    FileStream fileStream(dataFile.path(), FileStream::write);
    fileStream << json;
  }
}
//-----------------------------------------------------------------------------
std::string DataFileBuilder::makeJson(const DRCEResultData& resultData_) throw (Poco::Exception)
{
  std::string json;
  DRCEResultData resultData = resultData_;
  DRCEResultDataSerializator resultSerializator(resultData);
  if (!resultSerializator.serialize(json))
    throw Poco::Exception(resultSerializator.getErrorMsg(), resultSerializator.getErrorCode());

  return json;
}
//-----------------------------------------------------------------------------
//-----------------------------------------------------------------------------
StatusFileBuilder::StatusFileBuilder(CustomMessage& message_, const std::string& dir_)
: inherited(message_, dir_)
{
}
//-----------------------------------------------------------------------------
void StatusFileBuilder::build(const DRCEResultData& resultData) throw (Poco::Exception)
{
  if (resultData.getItemsCount()!=1)
    throw Poco::Exception(message(message_const::EMPTY_OBJECT_INSTANCE), ERROR_BAD_DATA_CONTENT);

  std::string json = getStatusJson(resultData);
  if (!json.empty() && resultData.getDataItem(0).getRequestId())
    FileStream(FileStream::getStatusFileName(dir, resultData.getDataItem(0).getRequestId()), FileStream::write) << json;
}
//-----------------------------------------------------------------------------
std::string StatusFileBuilder::getStatusJson(const DRCEResultData& resultData) throw (Poco::Exception)
{
  DRCEResultData localResultData;
  for (size_t i=0;i<resultData.getItemsCount();++i)
  {
    DRCEResultDataItem resultDataItem(resultData.getDataItem(i));
    cleanup(resultDataItem);
    localResultData.addDataItem(std::forward<DRCEResultDataItem>(resultDataItem));
  }
  return makeJson(localResultData);
}
//-----------------------------------------------------------------------------
void StatusFileBuilder::cleanup(DRCEResultDataItem& resultDataItem)
{
  resultDataItem.setStdoutStream("");
  resultDataItem.setStderrStream("");
  dynamic_cast<DRCEFilesList&>(resultDataItem).clear();
  for (size_t i=0;i<resultDataItem.getSubtasksCount();++i)
  {
    DRCEResultDataItem resultItem(resultDataItem.getSubtaskItem(i));
    cleanup(resultItem);
    resultDataItem.setSubtaskItem(i, std::forward<DRCEResultDataItem>(resultItem));
  }
}
//-----------------------------------------------------------------------------
//-----------------------------------------------------------------------------
RequestFileBuilder::RequestFileBuilder(CustomMessage& message_, const std::string& dir_)
: inherited(message_, dir_)
{
}
//-----------------------------------------------------------------------------
void RequestFileBuilder::build(DRCEInputJsonMessage& inputJsonMessage) throw (Poco::Exception)
{
  std::string resultJson;
  if (!inputJsonMessage.serialize(resultJson))
    throw Poco::Exception(inputJsonMessage.getErrorMsg(), inputJsonMessage.getErrorCode());

  if (!resultJson.empty() && inputJsonMessage.getRequestId())
  {
    Poco::File requestFile(FileStream::getRequestFileName(dir, inputJsonMessage.getRequestId()));
    if (requestFile.exists())
      requestFile.remove();

    FileStream fileStream(requestFile.path(), FileStream::write);
    fileStream << resultJson;
  }
}
//-----------------------------------------------------------------------------
//-----------------------------------------------------------------------------
} // end namespace drce 
} // end namespace HCE 
