#include <Poco/JSON/JSON.h>
#include <Poco/JSON/Stringifier.h>
#include <Poco/JSON/Parser.h>
#include <Poco/JSON/JSONException.h>
#include <Poco/Dynamic/Var.h>
#include <utility>
#include <limits.h>

#include "DRCEError.hpp"
#include "DRCEMessageConst.hpp"
#include "DRCEInputJsonMessage.hpp"
#include "DRCEEncodeDecodeBase64.hpp"
#include "DRCETaskRequestFactory.hpp"

namespace HCE
{
namespace drce
{
//-----------------------------------------------------------------------------
DRCEInputJsonMessage::DRCEInputJsonMessage(const std::string& json)
: IJsonSerializable(), Subtasks(), requestType(DRCETaskRequest::RequestType::rtSetTaskExecute), requestData(""), requestId(0), pTaskRequest(nullptr)
{
  if (!json.empty())
    unserialize(json);
}
//-----------------------------------------------------------------------------
DRCEInputJsonMessage::DRCEInputJsonMessage(const DRCEInputJsonMessage& rhs)
: IJsonSerializable(), Subtasks(), requestType(DRCETaskRequest::RequestType::rtSetTaskExecute), requestData(""), requestId(0), pTaskRequest(nullptr)
{
  (*this) = rhs;
}
//-----------------------------------------------------------------------------
DRCEInputJsonMessage::DRCEInputJsonMessage(DRCEInputJsonMessage&& rhs)
: IJsonSerializable(), Subtasks(), requestType(DRCETaskRequest::RequestType::rtSetTaskExecute), requestData(""), requestId(0), pTaskRequest(nullptr)
{
  (*this) = std::forward<DRCEInputJsonMessage>(rhs);
}
//-----------------------------------------------------------------------------
DRCEInputJsonMessage& DRCEInputJsonMessage::operator=(const DRCEInputJsonMessage& rhs)
{
  if (this != &rhs)
  {
    requestType = rhs.getRequestType();
    requestData = rhs.getRequestData();
    requestId = rhs.getRequestId();
    pTaskRequest = rhs.getTaskRequest();
    (*dynamic_cast<Subtasks*>(this)) = static_cast<Subtasks>(rhs);
  }
  return *this;
}
//-----------------------------------------------------------------------------
DRCEInputJsonMessage& DRCEInputJsonMessage::operator=(DRCEInputJsonMessage&& rhs)
{
  if (this != &rhs)
  {
    requestType = std::move(rhs.getRequestType());
    requestData = std::move(rhs.getRequestData());
    requestId = std::move(rhs.getRequestId());
    pTaskRequest = std::move(rhs.getTaskRequest());
    (*dynamic_cast<Subtasks*>(this)) = std::move(static_cast<Subtasks>(rhs));
  }
  return *this;
}
//-----------------------------------------------------------------------------
void DRCEInputJsonMessage::setRequestData(const std::string& requestData_)
{
  requestData = requestData_;
  if (!requestData.empty())
  {
    DRCETaskRequestFactory taskRequestFactory;
    pTaskRequest = taskRequestFactory.create(requestType);
    if (!pTaskRequest.isNull())
    {
      if (!pTaskRequest->unserialize(requestData))
      {
        _isError = pTaskRequest->isError();
        errorMsg = pTaskRequest->getErrorMsg();
        errorCode = pTaskRequest->getErrorCode();
      }
    }
  }
}
//-----------------------------------------------------------------------------
// cppcheck-suppress unusedFunction
void DRCEInputJsonMessage::setTaskRequest(Poco::SharedPtr<DRCETaskRequest> pTaskRequest_)
{
  pTaskRequest = pTaskRequest_;
  if (!pTaskRequest.isNull())
  {
    requestType = pTaskRequest->getRequestType();
    if (!pTaskRequest->serialize(requestData))
    {
      _isError = pTaskRequest->isError();
      errorMsg = pTaskRequest->getErrorMsg();
      errorCode = pTaskRequest->getErrorCode();
    }
  }
}
//-----------------------------------------------------------------------------
void DRCEInputJsonMessage::clear(void)
{
  requestType = DRCETaskRequest::RequestType::rtSetTaskExecute;
  requestData.clear();
  requestId = 0;
  pTaskRequest = nullptr;
  Subtasks::clear();
}
//-----------------------------------------------------------------------------
void DRCEInputJsonMessage::packToObject(Poco::JSON::Object& obj) throw (Poco::Exception)
{
  obj.set(drce_json_message_const::requestType, static_cast<unsigned int>(requestType));
  obj.set(drce_json_message_const::requestData, DRCEEncodeBase64(requestData));
  obj.set(drce_json_message_const::requestId, requestId);

  Poco::JSON::Array::Ptr pSubtasks = new Poco::JSON::Array();
  if (!pSubtasks.isNull())
  {
    packToArray(*pSubtasks);
    obj.set(drce_json_message_const::subtasksArray, pSubtasks);
  }
}
//-----------------------------------------------------------------------------
void DRCEInputJsonMessage::packToArray(Poco::JSON::Array& arr) throw (Poco::Exception)
{
  for (size_t i=0;i<getSubtasksCount();++i)
  {
    Poco::JSON::Object::Ptr pObj = new Poco::JSON::Object();
    if (!pObj.isNull())
    {
      DRCEInputJsonMessage inputJsonMessage = getSubtaskItem(i);
      inputJsonMessage.packToObject(*pObj);
      arr.add(*pObj);
    }
  }
}
//-----------------------------------------------------------------------------
void DRCEInputJsonMessage::unpackFromObject(Poco::JSON::Object& obj) throw (Poco::Exception)
{
  Poco::Dynamic::Var tmp = obj.get(drce_json_message_const::requestType);
  if (tmp.isEmpty())
   throw Poco::Exception(message_const::badJsonFormat+drce_json_message_const::requestType, ERROR_BAD_JSON_FORMAT);

  requestType = static_cast<DRCETaskRequest::RequestType>(convertVarToNumeric<unsigned int>(tmp, 0));

  tmp = obj.get(drce_json_message_const::requestData);
  if (tmp.isEmpty())
   throw Poco::Exception(message_const::badJsonFormat+drce_json_message_const::requestData, ERROR_BAD_JSON_FORMAT);

  if (tmp.isString())
   setRequestData(DRCEDecodeBase64(tmp.convert<std::string>()));

  if (isError())
   throw Poco::Exception(errorMsg, errorCode);

  if (pTaskRequest.isNull())
   throw Poco::Exception(message_const::badJsonFieldContent+drce_json_message_const::requestData, ERROR_BAD_JSON_FORMAT);

  tmp = obj.get(drce_json_message_const::requestId);
  if (tmp.isEmpty())
   throw Poco::Exception(message_const::badJsonFormat+drce_json_message_const::requestId, ERROR_BAD_JSON_FORMAT);

  long long taskId = convertVarToNumeric<long long>(tmp, 0);
  if (taskId > std::numeric_limits<unsigned int>::max() || taskId < 0)
   throw Poco::Exception(message_const::dataNotRange+" { 0, "+std::to_string(std::numeric_limits<unsigned int>::max())+" }", ERROR_TASK_ID_NOT_RANGE);
  requestId = static_cast<unsigned int>(taskId);

  Poco::JSON::Array::Ptr pSubtasks = obj.getArray(drce_json_message_const::subtasksArray);
  if (!pSubtasks.isNull())
  {
    unpackFromArray(*pSubtasks);
  }
}
//-----------------------------------------------------------------------------
void DRCEInputJsonMessage::unpackFromArray(Poco::JSON::Array& arr) throw (Poco::Exception)
{
  for (size_t i=0;i<arr.size();++i)
  {
    Poco::JSON::Object::Ptr pObj = arr.getObject(i);
    if (!pObj.isNull())
    {
      DRCEInputJsonMessage inputJsonMessage;
      inputJsonMessage.unpackFromObject(*pObj);
      addSubtaskItem(std::forward<DRCEInputJsonMessage>(inputJsonMessage));
    }
  }
}
//-----------------------------------------------------------------------------
bool DRCEInputJsonMessage::serialize(std::string& json)
{
  errorCode = NO_ERROR;
  errorMsg.clear();
  try
  {
    Poco::JSON::Object::Ptr pObj = new Poco::JSON::Object();
    if (pObj.isNull())
      throw Poco::Exception(message_const::objectInstanceWasNotCreated);

    packToObject(*pObj);

    std::stringstream ostr;
    Poco::JSON::Stringifier::stringify(pObj, ostr);
    json = ostr.str();

    _isError = false;
  }
  catch(Poco::Exception& e)
  {
    _isError = true;
    errorMsg = e.message();
    errorCode = e.code();
  }
  catch(std::exception& e)
  {
    _isError = true;
    errorMsg = e.what();
    errorCode = ERROR_SERIALIZE;
  }
  return !_isError;
}
//-----------------------------------------------------------------------------
bool DRCEInputJsonMessage::unserialize(const std::string& json)
{
  clear();
  _isError = false;
  errorCode = NO_ERROR;
  errorMsg.clear();

  try
  {
    Poco::JSON::Parser parser;
    Poco::Dynamic::Var res = parser.parse(json);
    Poco::JSON::Object::Ptr pObj = res.extract<Poco::JSON::Object::Ptr>();
    if (!pObj.isNull())
    {
      unpackFromObject(*pObj);
      _isError = false;
    }
  }
  catch(Poco::JSON::JSONException& e)
  {
    _isError = true;
    errorMsg = e.message();
    errorCode = ERROR_UNSERIALIZE;
  }
  catch(Poco::Exception& e)
  {
    _isError = true;
    errorMsg = e.message();
    errorCode = e.code();
  }
  catch(std::exception& e)
  {
    _isError = true;
    errorMsg = e.what();
    errorCode = ERROR_UNSERIALIZE;
  }
  return !_isError;
}
//-----------------------------------------------------------------------------
std::istream& operator>>(std::istream& is, DRCEInputJsonMessage& inputJsonMessage)
{
  std::string json;
  is.seekg(0, std::ios::end);
  json.resize(is.tellg());
  is.seekg(0, std::ios::beg);
  is.read(const_cast<char*>(json.c_str()), json.size());
  inputJsonMessage.unserialize(json);
  return is;
}
//-----------------------------------------------------------------------------
std::ostream& operator<<(std::ostream& os, const DRCEInputJsonMessage& inputJsonMessage)
{
  std::string json;
  const_cast<DRCEInputJsonMessage&>(inputJsonMessage).serialize(json);
  return os << json;
}
//-----------------------------------------------------------------------------
//-----------------------------------------------------------------------------
} // end namespace drce 
} // end namespace HCE 
