#include <utility>
#include <Poco/Logger.h>
#include <Poco/Timestamp.h>

#include "DRCEResultData.hpp"
#include "DRCEResourceMonitor.hpp"
#include "DRCEMessageConst.hpp"
#include "Process.hpp"
#include "DRCEResourceLimitsChecker.hpp"
#include "DRCEResourceExtractor.hpp"

namespace HCE
{
namespace drce
{
//-----------------------------------------------------------------------------
DRCEResourceMonitor::DRCEResourceMonitor(DRCEAsyncTasksQueue& asyncTasksQueue_, CustomMessage& message_)
: Poco::Runnable(), thread(), mutex(), terminated(false), timePeriod(DEFAULT_TIME_PERIOD),
  asyncTasksQueue(asyncTasksQueue_), message(message_), statVariables(), counterStatVariables(), avgCounts()
{
  HCE::drce::ResourceUsageExtractor::getInstance();
  thread.start(*this);
}
//-----------------------------------------------------------------------------
DRCEResourceMonitor::~DRCEResourceMonitor(void)
{
  terminate();
  thread.join();
  HCE::drce::ResourceUsageExtractor::shutdown();
}
//-----------------------------------------------------------------------------
void DRCEResourceMonitor::terminate(void)
{
  Poco::Mutex::ScopedLock lock(mutex);
  terminated = true;
}
//-----------------------------------------------------------------------------
bool DRCEResourceMonitor::isTerminated(void)
{
  Poco::Mutex::ScopedLock lock(mutex);
  return terminated;
}
//-----------------------------------------------------------------------------
void DRCEResourceMonitor::setTimePeriod(unsigned int timePeriod_)
{
  Poco::Mutex::ScopedLock lock(mutex);
  timePeriod=timePeriod_;
}
//-----------------------------------------------------------------------------
unsigned int DRCEResourceMonitor::getTimePeriod(void)
{
  Poco::Mutex::ScopedLock lock(mutex);
  return timePeriod;
}
//-----------------------------------------------------------------------------
void DRCEResourceMonitor::setStatVariables(StatVariables&& rhs)
{
  Poco::Mutex::ScopedLock lock(mutex);
  statVariables = std::forward<StatVariables>(rhs);
}
//-----------------------------------------------------------------------------
StatVariables DRCEResourceMonitor::getStatVariables(void)
{
  Poco::Mutex::ScopedLock lock(mutex);
  return statVariables;
}
//-----------------------------------------------------------------------------
void DRCEResourceMonitor::dump(StatVariables& statVariables, AvgCounts& avgCounts)
{
  Poco::Mutex::ScopedLock lock(mutex);
  statVariables = this->statVariables;
  avgCounts = this->avgCounts;
}
//-----------------------------------------------------------------------------
void DRCEResourceMonitor::restore(const StatVariables& statVariables, const AvgCounts& avgCounts)
{
  Poco::Mutex::ScopedLock lock(mutex);
  this->statVariables = statVariables;
  this->counterStatVariables = statVariables;
  this->avgCounts = avgCounts;
}
//-----------------------------------------------------------------------------
void DRCEResourceMonitor::fillFieldsArray(DRCEResultData& resultData)
{
  fillFieldsArray(resultData, asyncTasksQueue, message, this);
}
//-----------------------------------------------------------------------------
void DRCEResourceMonitor::fillFieldsArray(DRCEResultData& resultData, DRCEAsyncTasksQueue& asyncTasksQueue_, CustomMessage& message_, DRCEResourceMonitor* pResourceMonitor)
{
  for (size_t i=0;i<resultData.getItemsCount();++i)
  {
    DRCEResultDataItem resultDataItem(resultData.getDataItem(i));
    if (fillFieldsArray(resultDataItem, asyncTasksQueue_, message_, pResourceMonitor))
        resultData.setDataItem(i, resultDataItem);
  }
}
//-----------------------------------------------------------------------------
bool DRCEResourceMonitor::fillFieldsArray(DRCEResultDataItem& resultDataItem, DRCEAsyncTasksQueue& asyncTasksQueue_, CustomMessage& message_, DRCEResourceMonitor* pResourceMonitor)
{
  if (!asyncTasksQueue_.isExistAsyncTask(resultDataItem.getRequestId()))
  {
    std::stringstream outMsg;
    outMsg << "DRCEResourceMonitor::fillFieldsArray  task [ " << resultDataItem.getRequestId() << " ] already not exist...";
    Poco::Logger::root().log(Poco::Message(drce_const::moduleName, outMsg.str(), Poco::Message::Priority::PRIO_TRACE));
    return false;
  }

  AsyncTasks asyncTask = asyncTasksQueue_.getAsyncTask(resultDataItem.getRequestId());

  if (!asyncTask.pid)
  {
    std::stringstream outMsg;
    outMsg << "DRCEResourceMonitor::fillFieldsArray  task [ " << resultDataItem.getRequestId() << " ] pid is empty ...";
    Poco::Logger::root().log(Poco::Message(drce_const::moduleName, outMsg.str(), Poco::Message::Priority::PRIO_TRACE));
    return false;
  }

  if (asyncTask.pid && !HCE::Process::isExistProcess(asyncTask.pid))
  {
    std::stringstream outMsg;
    outMsg << "DRCEResourceMonitor::fillFieldsArray  task [ " << resultDataItem.getRequestId() << " ] has pid [ " << asyncTask.pid << " ] but process already not exist ...";
    Poco::Logger::root().log(Poco::Message(drce_const::moduleName, outMsg.str(), Poco::Message::Priority::PRIO_TRACE));
    return false;
  }

  ProcessInfoExtractor processInfoExtractor;
  ProcessInfoExtractor::FieldsArray fieldsArray = processInfoExtractor.extract(asyncTask);

  std::stringstream outMsg;
  outMsg << "DRCEResourceMonitor::fillFieldsArray  task [ " << resultDataItem.getRequestId() << " ] has state: " << static_cast<int>(asyncTask.state);
  Poco::Logger::root().log(Poco::Message(drce_const::moduleName, outMsg.str(), Poco::Message::Priority::PRIO_TRACE));

  try
  {
    HCE::TrimFormatter <std::string> trim;
    std::map<std::string, std::string> fields;
    fields.insert(std::map<std::string, std::string>::value_type(drce_const::taskCpu, trim(fieldsArray.cpu)));
    fields.insert(std::map<std::string, std::string>::value_type(drce_const::taskVram, std::to_string(fieldsArray.vram)));
    fields.insert(std::map<std::string, std::string>::value_type(drce_const::taskRram, std::to_string(fieldsArray.rram)));
    fields.insert(std::map<std::string, std::string>::value_type(drce_const::taskThreads, std::to_string(fieldsArray.threads)));

    ResourceUsage resourceUsage = HCE::drce::ResourceUsageExtractor::getInstance().extract();
    fields.insert(std::map<std::string, std::string>::value_type(drce_const::hostCpu, trim(resourceUsage.cpu)));
    fields.insert(std::map<std::string, std::string>::value_type(drce_const::hostIowait, trim(resourceUsage.iowait)));
    fields.insert(std::map<std::string, std::string>::value_type(drce_const::hostVram, trim(resourceUsage.vramUsedPercent)));
    fields.insert(std::map<std::string, std::string>::value_type(drce_const::hostRram, trim(resourceUsage.rramUsedPercent)));
    fields.insert(std::map<std::string, std::string>::value_type(drce_const::hostDisk, trim(resourceUsage.diskUsedPercent)));
    fields.insert(std::map<std::string, std::string>::value_type(drce_const::hostProcesses, trim(resourceUsage.processesPercent)));
    fields.insert(std::map<std::string, std::string>::value_type(drce_const::hostThreads, trim(resourceUsage.threadsPercent)));
    fields.insert(std::map<std::string, std::string>::value_type(drce_const::hostLoadAverage, trim(resourceUsage.loadAverage)));

    resultDataItem.setFields(fields);

    // save values Cpu and Memory usage
    if (pResourceMonitor)
    {
      pResourceMonitor->addCpuUsageAsyncTasks(fieldsArray.cpu);
      if (fieldsArray.rram)
        pResourceMonitor->addMemoryUsageAsyncTasks(fieldsArray.rram);
    }
  }
  catch(std::exception& e)
  {
    Poco::Logger::root().log(Poco::Message(drce_const::moduleName, message_(message_const::ERROR, e.what()), Poco::Message::Priority::PRIO_ERROR));
    return false;
  }
  return true;
}
//-----------------------------------------------------------------------------
void DRCEResourceMonitor::run(void)
{
  const unsigned int periodSleep = 1000; // msec
  Poco::Timestamp tsMinuteCounter;
  size_t countSyncTasks(0), countAsyncTasks(0);

  while(!isTerminated())
  {
    unsigned int periodMs = getTimePeriod();
    if (periodMs > periodSleep)
    {
      unsigned int repeat = periodMs/periodSleep; // msec to sec
      for (unsigned int i=0;i<repeat;++i)
      {
        Poco::Thread::sleep(periodSleep);
        if (isTerminated())
          return;
        countTasksForMinute(tsMinuteCounter, countSyncTasks, countAsyncTasks);
      }
    }
    else
    {
      Poco::Thread::sleep(periodMs);
      countTasksForMinute(tsMinuteCounter, countSyncTasks, countAsyncTasks);
    }
    try
    {
      Poco::Logger::root().log(Poco::Message(drce_const::moduleName, "!!! DRCEResourceMonitor::run !!!", Poco::Message::Priority::PRIO_TRACE));

      //!<Extract exist async task queue
      DRCEResultData resultData = asyncTasksQueue.getTasksAsResultData();
      if (resultData.getItemsCount())
      {
        //!<Update resource usage data
        HCE::drce::ResourceUsageExtractor::getInstance().update();
        //!<Fill resource usage values
        fillFieldsArray(resultData);
        asyncTasksQueue.executeNotification(resultData, DRCETaskRequest::RequestType::rtResourceMonitorNotification);
      }
      //!<Update stat variables possible for output
      setStatVariables(std::forward<StatVariables>(recountStatVariables()));
    }
    catch(Poco::Exception& e)
    {
      Poco::Logger::root().log(Poco::Message(drce_const::moduleName, message(message_const::ERROR, e.displayText()), Poco::Message::Priority::PRIO_ERROR));
    }
    catch(std::exception& e)
    {
      Poco::Logger::root().log(Poco::Message(drce_const::moduleName, message(message_const::ERROR, e.what()), Poco::Message::Priority::PRIO_ERROR));
    }
  }
}
//-----------------------------------------------------------------------------
void DRCEResourceMonitor::addTimeAsyncTask(size_t value)
{
  addStatValue<double>(static_cast<double>(value), counterStatVariables.timeAsyncTask, avgCounts.timeAsyncTaskCount);
}
//-----------------------------------------------------------------------------
void DRCEResourceMonitor::addTimeSyncTask(size_t value)
{
  addStatValue<double>(static_cast<double>(value), counterStatVariables.timeSyncTask, avgCounts.timeSyncTaskCount);
}
//-----------------------------------------------------------------------------
void DRCEResourceMonitor::addSizeInputBufferSyncTasks(size_t value)
{
  addStatValue(value, counterStatVariables.sizeInputBufferSyncTasks, avgCounts.sizeInputBufferSyncTasksCount);
}
//-----------------------------------------------------------------------------
void DRCEResourceMonitor::addSizeInputBufferAsyncTasks(size_t value)
{
  addStatValue(value, counterStatVariables.sizeInputBufferAsyncTasks, avgCounts.sizeInputBufferAsyncTasksCount);
}
//-----------------------------------------------------------------------------
void DRCEResourceMonitor::addSizeOutputBufferSyncTasks(size_t value)
{
  addStatValue(value, counterStatVariables.sizeOutputBufferSyncTasks, avgCounts.sizeOutputBufferSyncTasksCount);
}
//-----------------------------------------------------------------------------
void DRCEResourceMonitor::addSizeOutputBufferAsyncTasks(size_t value)
{
  addStatValue(value, counterStatVariables.sizeOutputBufferAsyncTasks, avgCounts.sizeOutputBufferAsyncTasksCount);
}
//-----------------------------------------------------------------------------
void DRCEResourceMonitor::addCpuUsageAsyncTasks(double value)
{
  addStatValue<double>(value, counterStatVariables.cpuUsageAsyncTasks, avgCounts.cpuUsageAsyncTasksCount);
}
//-----------------------------------------------------------------------------
void DRCEResourceMonitor::addMemoryUsageAsyncTasks(size_t value)
{
  addStatValue(value, counterStatVariables.memoryUsageAsyncTasks, avgCounts.memoryUsageAsyncTasksCount);
}
//-----------------------------------------------------------------------------
void DRCEResourceMonitor::addCountAsyncTasksForMinute(double value)
{
  addStatValue<double>(value, counterStatVariables.countAsyncTasksForMinute, avgCounts.countAsyncTasksForMinuteCount);
}
//-----------------------------------------------------------------------------
void DRCEResourceMonitor::addCountSyncTasksForMinute(double value)
{
  addStatValue<double>(value, counterStatVariables.countSyncTasksForMinute, avgCounts.countSyncTasksForMinuteCount);
}
//-----------------------------------------------------------------------------
void DRCEResourceMonitor::incrementCountSyncTasks(void)
{
  ++(counterStatVariables.countSyncTasks);
}
//-----------------------------------------------------------------------------
void DRCEResourceMonitor::incrementCountAsyncTasks(void)
{
  ++(counterStatVariables.countAsyncTasks);
}
//-----------------------------------------------------------------------------
void DRCEResourceMonitor::incrementCountSyncTasksFail(void)
{
  ++(counterStatVariables.countSyncTasksFail);
}
//-----------------------------------------------------------------------------
void DRCEResourceMonitor::incrementCountAsyncTasksFail(void)
{
  ++(counterStatVariables.countAsyncTasksFail);
}
//-----------------------------------------------------------------------------
template <class T> void DRCEResourceMonitor::addStatValue(T value, StatVariableData<T>& statVariableData, size_t& count)
{
  if (value)
  {
    if (statVariableData.min)
    {
      if (value < statVariableData.min)
        statVariableData.min = value;
    }
    else
      statVariableData.min = value;

    if (statVariableData.max)
    {
      if(value > statVariableData.max)
        statVariableData.max = value;
    }
    else
      statVariableData.max = value;

    statVariableData.avg += value;
    ++count;
  }
}
//-----------------------------------------------------------------------------
void DRCEResourceMonitor::countTasksForMinute(Poco::Timestamp& ts, size_t& countSyncTasks, size_t& countAsyncTasks)
{
  const size_t oneMinute = 60000000; // in microseconds
  if (ts.isElapsed(oneMinute) || !countSyncTasks || !countAsyncTasks)
  {
    // update count sync and async tasks if one minute was elapsed
    addCountSyncTasksForMinute(counterStatVariables.countSyncTasks-countSyncTasks);
    countSyncTasks = counterStatVariables.countSyncTasks;
    addCountAsyncTasksForMinute(counterStatVariables.countAsyncTasks-countAsyncTasks);
    countAsyncTasks = counterStatVariables.countAsyncTasks;
  }
}
//-----------------------------------------------------------------------------
StatVariables DRCEResourceMonitor::recountStatVariables(void)
{
  StatVariables localStatVariables(counterStatVariables);
/*
  std::cout << "avgCounts.timeAsyncTaskCount = " << localStatVariables.timeAsyncTask.avg << "/" << avgCounts.timeAsyncTaskCount << std::endl;
  std::cout << "avgCounts.timeSyncTaskCount = " << localStatVariables.timeSyncTask.avg << "/" << avgCounts.timeSyncTaskCount << std::endl;
  std::cout << "avgCounts.sizeInputBufferSyncTasksCount = " << localStatVariables.sizeInputBufferSyncTasks.avg << "/" << avgCounts.sizeInputBufferSyncTasksCount << std::endl;
  std::cout << "avgCounts.sizeOutputBufferSyncTasksCount = " << localStatVariables.sizeOutputBufferSyncTasks.avg << "/" << avgCounts.sizeOutputBufferSyncTasksCount << std::endl;
  std::cout << "avgCounts.sizeInputBufferAsyncTasksCount = " << localStatVariables.sizeInputBufferAsyncTasks.avg << "/" << avgCounts.sizeInputBufferAsyncTasksCount << std::endl;
  std::cout << "avgCounts.sizeOutputBufferAsyncTasksCount = " << localStatVariables.sizeOutputBufferAsyncTasks.avg << "/" << avgCounts.sizeOutputBufferAsyncTasksCount << std::endl;
  std::cout << "avgCounts.cpuUsageAsyncTasksCount = " << localStatVariables.cpuUsageAsyncTasks.avg << "/" << avgCounts.cpuUsageAsyncTasksCount << std::endl;
  std::cout << "avgCounts.memoryUsageAsyncTasksCount = " << localStatVariables.memoryUsageAsyncTasks.avg << "/" << avgCounts.memoryUsageAsyncTasksCount << std::endl;
  std::cout << "avgCounts.countSyncTasksForMinuteCount = " << localStatVariables.countSyncTasksForMinute.avg << "/" << avgCounts.countSyncTasksForMinuteCount << std::endl;
  std::cout << "avgCounts.countAsyncTasksForMinuteCount = " << localStatVariables.countAsyncTasksForMinute.avg << "/" << avgCounts.countAsyncTasksForMinuteCount << std::endl;
*/
  localStatVariables.timeAsyncTask.avg              /= (avgCounts.timeAsyncTaskCount)?avgCounts.timeAsyncTaskCount:1;
  localStatVariables.timeSyncTask.avg               /= (avgCounts.timeSyncTaskCount)?avgCounts.timeSyncTaskCount:1;
  localStatVariables.sizeInputBufferSyncTasks.avg   /= (avgCounts.sizeInputBufferSyncTasksCount)?avgCounts.sizeInputBufferSyncTasksCount:1;
  localStatVariables.sizeOutputBufferSyncTasks.avg  /= (avgCounts.sizeOutputBufferSyncTasksCount)?avgCounts.sizeOutputBufferSyncTasksCount:1;
  localStatVariables.sizeInputBufferAsyncTasks.avg  /= (avgCounts.sizeInputBufferAsyncTasksCount)?avgCounts.sizeInputBufferAsyncTasksCount:1;
  localStatVariables.sizeOutputBufferAsyncTasks.avg /= (avgCounts.sizeOutputBufferAsyncTasksCount)?avgCounts.sizeOutputBufferAsyncTasksCount:1;
  localStatVariables.cpuUsageAsyncTasks.avg         /= (avgCounts.cpuUsageAsyncTasksCount)?avgCounts.cpuUsageAsyncTasksCount:1;
  localStatVariables.memoryUsageAsyncTasks.avg      /= (avgCounts.memoryUsageAsyncTasksCount)?avgCounts.memoryUsageAsyncTasksCount:1;
  localStatVariables.countSyncTasksForMinute.avg    /= (avgCounts.countSyncTasksForMinuteCount)?avgCounts.countSyncTasksForMinuteCount:1;
  localStatVariables.countAsyncTasksForMinute.avg   /= (avgCounts.countAsyncTasksForMinuteCount)?avgCounts.countAsyncTasksForMinuteCount:1;

  return localStatVariables;
}
//-----------------------------------------------------------------------------
//-----------------------------------------------------------------------------
} // end namespace drce
} // end namespace HCE
