#include <Poco/JSON/JSON.h>
#include <Poco/JSON/Stringifier.h>
#include <Poco/JSON/Parser.h>
#include <Poco/JSON/JSONException.h>
#include <Poco/Dynamic/Var.h>

#include "DRCEEncodeDecodeBase64.hpp"
#include "DRCEError.hpp"
#include "DRCEMessageConst.hpp"
#include "DRCEResultDataSerializator.hpp"

namespace HCE
{
namespace drce
{
//-----------------------------------------------------------------------------
DRCEResultDataSerializator::DRCEResultDataSerializator(DRCEResultData& resultData_)
: inherited(), resultData(resultData_)
{
}
//-----------------------------------------------------------------------------
void DRCEResultDataSerializator::packToObject(Poco::JSON::Object& obj, DRCEResultDataItem& resultDataItem) throw (Poco::Exception)
{
  obj.set(drce_json_message_const::errorCode, resultDataItem.getErrorCode());
  obj.set(drce_json_message_const::errorMessage, resultDataItem.getErrorMessage());
  obj.set(drce_json_message_const::requestId, resultDataItem.getRequestId());
  obj.set(drce_json_message_const::requestType, static_cast<unsigned int>(resultDataItem.getRequestType()));
  obj.set(drce_json_message_const::nodeHost, resultDataItem.getNodeHost());
  obj.set(drce_json_message_const::nodePort, resultDataItem.getNodePort());
  obj.set(drce_json_message_const::state, resultDataItem.getState());
  obj.set(drce_json_message_const::pid, resultDataItem.getPid());
  obj.set(drce_json_message_const::stdoutStream, DRCEEncodeBase64(resultDataItem.getStdoutStream()));
  obj.set(drce_json_message_const::stderrStream, DRCEEncodeBase64(resultDataItem.getStderrStream()));
  obj.set(drce_json_message_const::nodeName, resultDataItem.getNodeName());
  obj.set(drce_json_message_const::time, resultDataItem.getTime());
  obj.set(drce_json_message_const::exitStatus, resultDataItem.getExitStatus());

  Poco::JSON::Object::Ptr pFields = new Poco::JSON::Object();
  if (!pFields.isNull())
  {
    for (std::map<std::string, std::string>::const_iterator iter=resultDataItem.getFields().begin();iter!=resultDataItem.getFields().end();++iter)
    {
      pFields->set((*iter).first, (*iter).second);
    }
    obj.set(drce_json_message_const::fields, pFields);
  }

  Poco::JSON::Array::Ptr pFiles = new Poco::JSON::Array();
  if (!pFiles.isNull())
  {
    const size_t filesCount = resultDataItem.getFilesCount();
    for (size_t k=0;k<filesCount;++k)
    {
      Poco::JSON::Object::Ptr pFileElem = new Poco::JSON::Object();
      if (!pFileElem.isNull())
      {
        pFileElem->set(drce_json_message_const::fileName, resultDataItem.getFileItem(k).name);
        pFileElem->set(drce_json_message_const::fileData, resultDataItem.getFileItem(k).data);
        pFileElem->set(drce_json_message_const::actionTypeMask, resultDataItem.getFileItem(k).actionType);
        pFiles->add(pFileElem);
      }
    }
    obj.set(drce_json_message_const::filesArray, pFiles);
  }

  Poco::JSON::Array::Ptr pSubtasks = new Poco::JSON::Array();
  if (!pSubtasks.isNull())
  {
    packToArray(*pSubtasks, resultDataItem);
    obj.set(drce_json_message_const::subtasksArray, pSubtasks);
  }
}
//-----------------------------------------------------------------------------
void DRCEResultDataSerializator::packToArray(Poco::JSON::Array& arr, DRCEResultDataItem& resultDataItem) throw (Poco::Exception)
{
  for (size_t i=0;i<resultDataItem.getSubtasksCount();++i)
  {
    Poco::JSON::Object::Ptr pObj = new Poco::JSON::Object();
    if (!pObj.isNull())
    {
      DRCEResultDataItem resultItem(resultDataItem.getSubtaskItem(i));
      packToObject(*pObj, resultItem);
      arr.add(*pObj);
    }
  }
}
//-----------------------------------------------------------------------------
void DRCEResultDataSerializator::unpackFromObject(Poco::JSON::Object& obj, DRCEResultDataItem& resultDataItem) throw (Poco::Exception)
{
  Poco::Dynamic::Var tmp = obj.get(drce_json_message_const::errorCode);
  resultDataItem.setErrorCode(convertVarToNumeric<unsigned int>(tmp, 0));

  tmp = obj.get(drce_json_message_const::errorMessage);
  if (tmp.isString())
    resultDataItem.setErrorMessage(tmp.convert<std::string>());

  tmp = obj.get(drce_json_message_const::requestId);
  resultDataItem.setRequestId(convertVarToNumeric<unsigned int>(tmp, 0));

  tmp = obj.get(drce_json_message_const::requestType);
  resultDataItem.setRequestType(static_cast<DRCETaskRequest::RequestType>(convertVarToNumeric<unsigned int>(tmp, 0)));

  tmp = obj.get(drce_json_message_const::nodeHost);
  if (tmp.isString())
    resultDataItem.setNodeHost(tmp.convert<std::string>());

  tmp = obj.get(drce_json_message_const::nodePort);
  if (tmp.isString())
    resultDataItem.setNodePort(tmp.convert<std::string>());

  tmp = obj.get(drce_json_message_const::state);
  resultDataItem.setState(convertVarToNumeric<unsigned int>(tmp, 0));

  tmp = obj.get(drce_json_message_const::pid);
  resultDataItem.setPid(convertVarToNumeric<unsigned int>(tmp, 0));

  tmp = obj.get(drce_json_message_const::stdoutStream);
  if (tmp.isString())
    resultDataItem.setStdoutStream(DRCEDecodeBase64(tmp.convert<std::string>()));

  tmp = obj.get(drce_json_message_const::stderrStream);
  if (tmp.isString())
    resultDataItem.setStderrStream(DRCEDecodeBase64(tmp.convert<std::string>()));

  tmp = obj.get(drce_json_message_const::nodeName);
  if (tmp.isString())
    resultDataItem.setNodeName(tmp.convert<std::string>());

  tmp = obj.get(drce_json_message_const::time);
  resultDataItem.setTime(convertVarToNumeric<unsigned int>(tmp, 0));

  tmp = obj.get(drce_json_message_const::exitStatus);
  resultDataItem.setExitStatus(convertVarToNumeric<unsigned int>(tmp, 0));

  Poco::JSON::Object::Ptr pFields = obj.getObject(drce_json_message_const::fields);
  if (!pFields.isNull())
  {
    std::map<std::string, std::string> fields;
    std::vector<std::string> names;
    pFields->getNames(names);
    for (size_t i=0;i<names.size();++i)
    {
      tmp = pFields->get(names[i]);
      if (tmp.isString())
        fields[names[i]] = tmp.convert<std::string>();
    }
    resultDataItem.setFields(fields);
  }

  Poco::JSON::Array::Ptr pFiles = obj.getArray(drce_json_message_const::filesArray);
  if (!pFiles.isNull())
  {
    for (size_t k=0;k<pFiles->size();++k)
    {
      Poco::JSON::Object::Ptr pFileElem = pFiles->getObject(k);
      if (!pFileElem.isNull())
      {
        FileItem fileItem;

        tmp = pFileElem->get(drce_json_message_const::fileName);
        if (tmp.isString())
          fileItem.name = tmp.convert<std::string>();

        tmp = pFileElem->get(drce_json_message_const::fileData);
        if (tmp.isString())
          fileItem.data = tmp.convert<std::string>();

        tmp = pFileElem->get(drce_json_message_const::actionTypeMask);
        fileItem.actionType = convertVarToNumeric<unsigned int>(tmp, 0);

        resultDataItem.addFileItem(std::forward<FileItem>(fileItem));
      }
    }
  }

  Poco::JSON::Array::Ptr pSubtasks = obj.getArray(drce_json_message_const::subtasksArray);
  if (!pSubtasks.isNull())
  {
    unpackFromArray(*pSubtasks, resultDataItem);
  }
}
//-----------------------------------------------------------------------------
void DRCEResultDataSerializator::unpackFromArray(Poco::JSON::Array& arr, DRCEResultDataItem& resultDataItem) throw (Poco::Exception)
{
  for (size_t i=0;i<arr.size();++i)
  {
    Poco::JSON::Object::Ptr pObj = arr.getObject(i);
    if (!pObj.isNull())
    {
      DRCEResultDataItem resultItem;
      unpackFromObject(*pObj, resultItem);
      resultDataItem.addSubtaskItem(resultItem);
    }
  }
}
//-----------------------------------------------------------------------------
bool DRCEResultDataSerializator::serialize(std::string& json)
{
  _isError = false;
  errorCode = NO_ERROR;
  errorMsg.clear();
  try
  {
    Poco::JSON::Array::Ptr pItems = new Poco::JSON::Array();
    if (pItems.isNull())
      throw Poco::Exception(message_const::objectInstanceWasNotCreated);

    for (size_t i=0;i<resultData.getItemsCount();++i)
    {
      Poco::JSON::Object::Ptr pItemElem = new Poco::JSON::Object();

      if (!pItemElem.isNull())
      {
        DRCEResultDataItem resultItem(resultData.getDataItem(i));
        packToObject(*pItemElem, resultItem);
        pItems->add(pItemElem);
      }
    }

    std::stringstream ostr;
    Poco::JSON::Stringifier::stringify(pItems, ostr);
    json = ostr.str();

    _isError = false;
  }
  catch(Poco::Exception& e)
  {
    _isError = true;
    errorMsg = e.displayText();
    errorCode = ERROR_UNSERIALIZE;
  }
  catch(std::exception& e)
  {
    _isError = true;
    errorMsg = e.what();
    errorCode = ERROR_SERIALIZE;
  }
  return !_isError;
}
//-----------------------------------------------------------------------------
bool DRCEResultDataSerializator::unserialize(const std::string& json)
{
  _isError = false;
  resultData.clear();
  errorCode = NO_ERROR;
  errorMsg.clear();

  try
  {
    Poco::JSON::Parser parser;
    Poco::Dynamic::Var res = parser.parse(json);

    Poco::Dynamic::Var tmp;
    Poco::JSON::Array::Ptr pItems = res.extract<Poco::JSON::Array::Ptr>();
    if (!pItems.isNull())
    {
      for (size_t i=0;i<pItems->size();++i)
      {
        Poco::JSON::Object::Ptr pItemElem = pItems->getObject(i);
        if (!pItemElem.isNull())
        {
          DRCEResultDataItem resultDataItem;
          unpackFromObject(*pItemElem, resultDataItem);
          resultData.addDataItem(std::forward<DRCEResultDataItem>(resultDataItem));
        }
      }
    }
    _isError = false;
  }
  catch(Poco::JSON::JSONException& e)
  {
    _isError = true;
    errorMsg = e.displayText();
    errorCode = ERROR_UNSERIALIZE;
  }
  catch(Poco::Exception& e)
  {
    _isError = true;
    errorMsg = e.displayText();
    errorCode = ERROR_UNSERIALIZE;
  }
  return !_isError;
}
//-----------------------------------------------------------------------------
//-----------------------------------------------------------------------------
} // namespace drce
} // namespace HCE
