#include <Poco/JSON/Object.h>
#include <Poco/JSON/Array.h>
#include <Poco/JSON/JSON.h>
#include <Poco/JSON/Stringifier.h>
#include <Poco/JSON/Parser.h>
#include <Poco/JSON/JSONException.h>
#include <Poco/Dynamic/Var.h>
#include <utility>

#include "DRCEError.hpp"
#include "DRCEMessageConst.hpp"
#include "DRCETaskRequestSetExecute.hpp"

namespace HCE
{
namespace drce
{
//-----------------------------------------------------------------------------
DRCETaskRequestSetExecute::DRCETaskRequestSetExecute(const std::string& json)
: DRCETaskRequest(DRCETaskRequest::RequestType::rtSetTaskExecute), DRCEFilesList(), sessionOptions(), commandLine(""), inputStream(""), resourceLimits()
{
  if (!json.empty())
    unserialize(json);
}
//-----------------------------------------------------------------------------
DRCETaskRequestSetExecute::DRCETaskRequestSetExecute(const DRCETaskRequestSetExecute& rhs)
: DRCETaskRequest(DRCETaskRequest::RequestType::rtSetTaskExecute), DRCEFilesList(), sessionOptions(), commandLine(""), inputStream(""), resourceLimits()
{
  (*this) = rhs;
}
//-----------------------------------------------------------------------------
DRCETaskRequestSetExecute::DRCETaskRequestSetExecute(DRCETaskRequestSetExecute&& rhs)
: DRCETaskRequest(DRCETaskRequest::RequestType::rtSetTaskExecute), DRCEFilesList(), sessionOptions(), commandLine(""), inputStream(""), resourceLimits()
{
  (*this) = std::forward<DRCETaskRequestSetExecute>(rhs);
}
//-----------------------------------------------------------------------------
DRCETaskRequestSetExecute& DRCETaskRequestSetExecute::operator=(const DRCETaskRequestSetExecute& rhs)
{
  if (this != &rhs)
  {
    sessionOptions = rhs.getSessionOptions();
    commandLine = rhs.getCommandLine();
    inputStream = rhs.getInputStream();
    resourceLimits = rhs.getResourceLimits();
    (*dynamic_cast<DRCEFilesList*>(this)) = static_cast<DRCEFilesList>(rhs);
    taskId = rhs.getTaskId();
    taskState = rhs.getTaskState();
  }
  return *this;
}
//-----------------------------------------------------------------------------
DRCETaskRequestSetExecute& DRCETaskRequestSetExecute::operator=(DRCETaskRequestSetExecute&& rhs)
{
  if (this != &rhs)
  {
    sessionOptions = std::move(rhs.getSessionOptions());
    commandLine = std::move(rhs.getCommandLine());
    inputStream = std::move(rhs.getInputStream());
    resourceLimits = std::move(rhs.getResourceLimits());
    (*dynamic_cast<DRCEFilesList*>(this)) = std::move(static_cast<DRCEFilesList>(rhs));
    taskId = std::move(rhs.getTaskId());
    taskState = std::move(rhs.getTaskState());
  }
  return *this;
}
//-----------------------------------------------------------------------------
void DRCETaskRequestSetExecute::setSessionOptions(SessionOptions&& sessionOptions_)
{
  sessionOptions = std::move(std::forward<SessionOptions>(sessionOptions_));
}
//-----------------------------------------------------------------------------
// cppcheck-suppress unusedFunction
void DRCETaskRequestSetExecute::setResourceLimits(ResourceLimits&& resourceLimits_)
{
  resourceLimits = std::move(std::forward<ResourceLimits>(resourceLimits_));
}
//-----------------------------------------------------------------------------
void DRCETaskRequestSetExecute::clear(void)
{
  sessionOptions.clear();
  commandLine.clear();
  inputStream.clear();
  resourceLimits.clear();
  files.clear();
}
//-----------------------------------------------------------------------------
bool DRCETaskRequestSetExecute::serialize(std::string& json)
{
  errorCode = NO_ERROR;
  errorMsg.clear();
  try
  {
    Poco::JSON::Object::Ptr obj = new Poco::JSON::Object();

    Poco::JSON::Object::Ptr pSessionOptions = new Poco::JSON::Object();
    pSessionOptions->set(drce_json_message_const::sessionType, static_cast<unsigned int>(sessionOptions.sessionType));
    pSessionOptions->set(drce_json_message_const::port, sessionOptions.port);
    pSessionOptions->set(drce_json_message_const::userName, sessionOptions.userName);
    pSessionOptions->set(drce_json_message_const::userPassword, sessionOptions.userPassword);
    pSessionOptions->set(drce_json_message_const::shellName, sessionOptions.shellName);
    pSessionOptions->set(drce_json_message_const::homeDir, sessionOptions.homeDir);
    pSessionOptions->set(drce_json_message_const::timeout, sessionOptions.timeout);
    pSessionOptions->set(drce_json_message_const::tmode, static_cast<unsigned int>(sessionOptions.tmode));
    pSessionOptions->set(drce_json_message_const::timeMax, sessionOptions.timeMax);
    pSessionOptions->set(drce_json_message_const::cleanup, static_cast<unsigned int>(sessionOptions.cleanup));

    Poco::JSON::Array::Ptr pEnviroment = new Poco::JSON::Array();
    for (size_t i=0;i<sessionOptions.environments.size();++i)
    {
      Poco::JSON::Object::Ptr pEnviromentElem = new Poco::JSON::Object();
      pEnviromentElem->set(sessionOptions.environments[i].first, sessionOptions.environments[i].second);
      pEnviroment->add(pEnviromentElem);
    }
    pSessionOptions->set(drce_json_message_const::environment, pEnviroment);

    obj->set(drce_json_message_const::sessionOptions, pSessionOptions);
    obj->set(drce_json_message_const::commandLine, commandLine);
    obj->set(drce_json_message_const::inputStream, inputStream);

    Poco::JSON::Object::Ptr pResourceLimits = new Poco::JSON::Object();
    pResourceLimits->set(drce_json_message_const::procMax, resourceLimits.procMax);
    pResourceLimits->set(drce_json_message_const::threadsMax, resourceLimits.threadsMax);
    pResourceLimits->set(drce_json_message_const::cpuMax, resourceLimits.cpuMax);
    pResourceLimits->set(drce_json_message_const::iowaitMax, resourceLimits.iowaitMax);
    pResourceLimits->set(drce_json_message_const::vramFree, resourceLimits.vramFree);
    pResourceLimits->set(drce_json_message_const::vramFreeMin, resourceLimits.vramFreeMin);
    pResourceLimits->set(drce_json_message_const::rramFree, resourceLimits.rramFree);
    pResourceLimits->set(drce_json_message_const::rramFreeMin, resourceLimits.rramFreeMin);
    pResourceLimits->set(drce_json_message_const::diskFree, resourceLimits.diskFree);
    pResourceLimits->set(drce_json_message_const::diskFreeMin, resourceLimits.diskFreeMin);
    pResourceLimits->set(drce_json_message_const::iowaitUsageMax, resourceLimits.usageLimits.iowaitMax);
    pResourceLimits->set(drce_json_message_const::vramUsageMax, resourceLimits.usageLimits.vramMax);
    pResourceLimits->set(drce_json_message_const::rramUsageMax, resourceLimits.usageLimits.rramMax);
    pResourceLimits->set(drce_json_message_const::diskUsageMax, resourceLimits.usageLimits.diskMax);
    pResourceLimits->set(drce_json_message_const::diskUsageMaxBytes, resourceLimits.usageLimits.diskMaxBytes);

    obj->set(drce_json_message_const::resourceLimits, pResourceLimits);

    Poco::JSON::Array::Ptr pFiles = new Poco::JSON::Array();

    const size_t filesCount = getFilesCount();
    for (size_t i=0;i<filesCount;++i)
    {
      Poco::JSON::Object::Ptr pFileElem = new Poco::JSON::Object();
      pFileElem->set(drce_json_message_const::actionTypeMask, getFileItem(i).actionType);
      pFileElem->set(drce_json_message_const::fileName, getFileItem(i).name);
      pFileElem->set(drce_json_message_const::fileData, getFileItem(i).data);
      pFiles->add(pFileElem);
    }
    obj->set(drce_json_message_const::filesArray, pFiles);

    std::stringstream ostr;
    Poco::JSON::Stringifier::stringify(obj, ostr);
    json = ostr.str();

    _isError = false;
  }
  catch(std::exception& e)
  {
    _isError = true;
    errorMsg = e.what();
    errorCode = ERROR_SERIALIZE;
  }
  return !_isError;
}
//-----------------------------------------------------------------------------
bool DRCETaskRequestSetExecute::unserialize(const std::string& json)
{
  clear();
  errorCode = NO_ERROR;
  errorMsg.clear();

  try
  {
    Poco::JSON::Parser parser;
    Poco::Dynamic::Var res = parser.parse(json);
    Poco::JSON::Object::Ptr obj = res.extract<Poco::JSON::Object::Ptr>();

    Poco::Dynamic::Var tmp;
    Poco::JSON::Object::Ptr pSessionOptions = obj->getObject(drce_json_message_const::sessionOptions);
    if (!pSessionOptions.isNull())
    {
      tmp = pSessionOptions->get(drce_json_message_const::sessionType);
      if (tmp.isEmpty())
        throw Poco::Exception(message_const::badJsonFormat+drce_json_message_const::sessionType, ERROR_BAD_JSON_FORMAT);
      sessionOptions.sessionType = static_cast<SessionOptions::SessionType>(convertVarToNumeric<unsigned int>(tmp, 0));

      tmp = pSessionOptions->get(drce_json_message_const::port);
      sessionOptions.port = convertVarToNumeric<unsigned int>(tmp, 0);

      tmp = pSessionOptions->get(drce_json_message_const::userName);
      if (tmp.isString())
        sessionOptions.userName = tmp.convert<std::string>();

      tmp = pSessionOptions->get(drce_json_message_const::userPassword);
      if (tmp.isString())
        sessionOptions.userPassword = tmp.convert<std::string>();

      tmp = pSessionOptions->get(drce_json_message_const::shellName);
      if (tmp.isString())
        sessionOptions.shellName = tmp.convert<std::string>();

      tmp = pSessionOptions->get(drce_json_message_const::homeDir);
      if (tmp.isString())
        sessionOptions.homeDir = tmp.convert<std::string>();

      tmp = pSessionOptions->get(drce_json_message_const::timeout);
      sessionOptions.timeout = convertVarToNumeric<unsigned int>(tmp, 0);

      tmp = pSessionOptions->get(drce_json_message_const::tmode);
      if (tmp.isEmpty())
        throw Poco::Exception(message_const::badJsonFormat+drce_json_message_const::tmode, ERROR_BAD_JSON_FORMAT);
      sessionOptions.tmode = static_cast<SessionOptions::ThreadMode>(convertVarToNumeric<unsigned int>(tmp, 0));

      tmp = pSessionOptions->get(drce_json_message_const::timeMax);
      if (tmp.isEmpty())
        throw Poco::Exception(message_const::badJsonFormat+drce_json_message_const::timeMax, ERROR_BAD_JSON_FORMAT);
      sessionOptions.timeMax = convertVarToNumeric<unsigned int>(tmp, 0);

      tmp = pSessionOptions->get(drce_json_message_const::cleanup);
      sessionOptions.cleanup = static_cast<SessionOptions::CleanupFlag>(convertVarToNumeric<unsigned int>(tmp, 0));

      Poco::JSON::Array::Ptr pEnviroment = pSessionOptions->getArray(drce_json_message_const::environment);
      if (!pEnviroment.isNull())
      {
        for (size_t j=0;j<pEnviroment->size();++j)
        {
          Poco::JSON::Object::Ptr pEnviromentElem = pEnviroment->getObject(j);
          if (!pEnviromentElem.isNull())
          {
            std::vector<std::string> names;
            pEnviromentElem->getNames(names);
            for (size_t k=0;k<names.size();++k)
            {
              tmp = pEnviromentElem->get(names[k]);
              if (tmp.isString())
                sessionOptions.environments.push_back(std::make_pair(names[k], tmp.convert<std::string>()));
            }
          }
        }
      }
    }

    tmp = obj->get(drce_json_message_const::commandLine);
    if (tmp.isString())
      commandLine = tmp.convert<std::string>();

    tmp = obj->get(drce_json_message_const::inputStream);
    if (tmp.isString())
      inputStream = tmp.convert<std::string>();

    Poco::JSON::Object::Ptr pResourceLimits = obj->getObject(drce_json_message_const::resourceLimits);
    if (!pResourceLimits.isNull())
    {
      tmp = pResourceLimits->get(drce_json_message_const::procMax);
      resourceLimits.procMax = convertVarToNumeric<unsigned int>(tmp, 0);

      tmp = pResourceLimits->get(drce_json_message_const::threadsMax);
      resourceLimits.threadsMax = convertVarToNumeric<unsigned int>(tmp, 0);

      tmp = pResourceLimits->get(drce_json_message_const::cpuMax);
      resourceLimits.cpuMax = convertVarToNumeric<double>(tmp, 0);

      tmp = pResourceLimits->get(drce_json_message_const::iowaitMax);
      resourceLimits.iowaitMax = convertVarToNumeric<double>(tmp, 0);

      tmp = pResourceLimits->get(drce_json_message_const::vramFree);
      resourceLimits.vramFree = convertVarToNumeric<double>(tmp, 0);

      tmp = pResourceLimits->get(drce_json_message_const::vramFreeMin);
      resourceLimits.vramFreeMin = convertVarToNumeric<size_t>(tmp, 0);

      tmp = pResourceLimits->get(drce_json_message_const::rramFree);
      resourceLimits.rramFree = convertVarToNumeric<double>(tmp, 0);

      tmp = pResourceLimits->get(drce_json_message_const::rramFreeMin);
      resourceLimits.rramFreeMin = convertVarToNumeric<size_t>(tmp, 0);

      tmp = pResourceLimits->get(drce_json_message_const::diskFree);
      resourceLimits.diskFree = convertVarToNumeric<double>(tmp, 0);

      tmp = pResourceLimits->get(drce_json_message_const::diskFreeMin);
      resourceLimits.diskFreeMin = convertVarToNumeric<size_t>(tmp, 0);

      tmp = pResourceLimits->get(drce_json_message_const::iowaitUsageMax);
      resourceLimits.usageLimits.iowaitMax = convertVarToNumeric<double>(tmp, 0);

      tmp = pResourceLimits->get(drce_json_message_const::vramUsageMax);
      resourceLimits.usageLimits.vramMax = convertVarToNumeric<double>(tmp, 0);

      tmp = pResourceLimits->get(drce_json_message_const::rramUsageMax);
      resourceLimits.usageLimits.rramMax = convertVarToNumeric<double>(tmp, 0);

      tmp = pResourceLimits->get(drce_json_message_const::diskUsageMax);
      resourceLimits.usageLimits.diskMax = convertVarToNumeric<double>(tmp, 0);

      tmp = pResourceLimits->get(drce_json_message_const::diskUsageMaxBytes);
      resourceLimits.usageLimits.diskMaxBytes = convertVarToNumeric<size_t>(tmp, 0);
    }

    Poco::JSON::Array::Ptr pFiles = obj->getArray(drce_json_message_const::filesArray);
    if (!pFiles.isNull())
    {
      for (size_t i=0;i<pFiles->size();++i)
      {
        Poco::JSON::Object::Ptr pFileElem = pFiles->getObject(i);
        if (!pFileElem.isNull())
        {
          FileItem fileItem;
          fileItem.actionType = pFileElem->getValue<unsigned int>(drce_json_message_const::actionTypeMask);
          fileItem.name = pFileElem->getValue<std::string>(drce_json_message_const::fileName);
          fileItem.data = pFileElem->getValue<std::string>(drce_json_message_const::fileData);
          addFileItem(std::forward<FileItem>(fileItem));
        }
      }
    }

    _isError = false;
  }
  catch(Poco::JSON::JSONException& e)
  {
    _isError = true;
    errorMsg = e.message();
    errorCode = ERROR_UNSERIALIZE;
  }
  catch(Poco::Exception& e)
  {
    _isError = true;
    errorMsg = e.message();
    errorCode = ERROR_UNSERIALIZE;
  }
  return !_isError;
}
//-----------------------------------------------------------------------------
std::istream& operator>>(std::istream& is, DRCETaskRequestSetExecute& rhs)
{
  std::string json;
  is.seekg(0, std::ios::end);
  json.resize(is.tellg());
  is.seekg(0, std::ios::beg);
  is.read(const_cast<char*>(json.c_str()), json.size());
  rhs.unserialize(json);
  return is;
}
//-----------------------------------------------------------------------------
std::ostream& operator<<(std::ostream& os, const DRCETaskRequestSetExecute& rhs)
{
  std::string json;
  const_cast<DRCETaskRequestSetExecute&>(rhs).serialize(json);
  return os << json;
}
//-----------------------------------------------------------------------------
//-----------------------------------------------------------------------------
} // namespace drce
} // namespace HCE
