#include <signal.h>
#include <sys/wait.h>
#include <Poco/Timestamp.h>
#include <Poco/Logger.h>
#include <Poco/File.h>

#include "Process.hpp"
#include "DRCETaskRequestTerminateExecutor.hpp"
#include "DRCETaskRequestTerminate.hpp"
#include "DRCEFileExtractor.hpp"
#include "DRCEFileBuilder.hpp"
#include "DRCEFileStream.hpp"
#include "DRCECleanupTask.hpp"
#include "DRCECommonTask.hpp"
#include "DRCEReadProcessData.hpp"

namespace HCE
{
namespace drce
{
//-----------------------------------------------------------------------------
DRCETaskRequestTerminateExecutor::DRCETaskRequestTerminateExecutor(DRCEAsyncTasksQueue& asyncTaskQueue_,
    DRCENodeOptions& nodeOptions_, CustomMessage& message_, DRCEResourceMonitor& resourceMonitor_)
: inherited(asyncTaskQueue_, nodeOptions_, message_, resourceMonitor_)
{
}
//-----------------------------------------------------------------------------
DRCEResultDataItem DRCETaskRequestTerminateExecutor::execute(DRCETaskRequest* pTaskRequest, DRCEInputJsonMessage& inputJsonMessage)
{
  return executeTerminate(pTaskRequest, inputJsonMessage, asyncTaskQueue, nodeOptions, message);
}
//-----------------------------------------------------------------------------
DRCEResultDataItem DRCETaskRequestTerminateExecutor::executeTerminate(DRCETaskRequest* pTaskRequest, DRCEInputJsonMessage& inputJsonMessage,
    DRCEAsyncTasksQueue& asyncTaskQueue, DRCENodeOptions& nodeOptions, CustomMessage& message, const std::string& errorMsg, unsigned int errorCode,
    DRCETaskRequest::TaskState taskState)
{
  DRCEResultDataItem resultDataItem = getResultDataItem(*pTaskRequest, nodeOptions, DRCETaskRequest::TaskState::UNDEFINED);

  try
  {
    DRCETaskRequestTerminate* pTaskRequestTerminate = dynamic_cast<DRCETaskRequestTerminate*>(pTaskRequest);
    if (!pTaskRequestTerminate)
      throw Poco::Exception(message(message_const::WRONG_TYPE_TASK_REQUEST), ERROR_BAD_COMMAND_PARAMS);

    if (pTaskRequestTerminate->getDelayValue() == 0 || pTaskRequestTerminate->getRepeatValue() == 0)
      throw Poco::Exception(message(message_const::BAD_COMMAND_PARAMS), ERROR_BAD_COMMAND_PARAMS);

    resultDataItem.setState(static_cast<unsigned int>(DRCETaskRequest::TaskState::NOT_FOUND));

    if (!asyncTaskQueue.isExistAsyncTask(pTaskRequestTerminate->getTaskId()))
    {
      if (!loadStatusFile(asyncTaskQueue,  nodeOptions, message, resultDataItem))
        throw Poco::Exception(message(message_const::TASK_NOT_FOUND), ERROR_TERMINATE_TASK);
    }

    AsyncTasks asyncTask = asyncTaskQueue.getAsyncTask(pTaskRequestTerminate->getTaskId());

    std::stringstream outMsg;
    outMsg << "Request terminate taskId: " << pTaskRequestTerminate->getTaskId();
    Poco::Logger::root().log(Poco::Message(drce_const::moduleName, outMsg.str(), Poco::Message::Priority::PRIO_TRACE));

    DRCECommonTask::SubtasksList subtasks;
    DRCEInputJsonMessage inputMessage = DRCECommonTask::getRequestData(pTaskRequestTerminate->getTaskId(), nodeOptions, message);
    DRCECommonTask::getSubtasksList(inputMessage, subtasks);

    // below block code need to checking
    for (DRCECommonTask::SubtasksList::reverse_iterator iter=subtasks.rbegin();iter!=subtasks.rend();++iter)
    {
      asyncTaskQueue.addSubtask(pTaskRequestTerminate->getTaskId(), (*iter).first, (pTaskRequestTerminate->getCleanupFlag()==DRCETaskRequestTerminate::CleanupFlag::cfDelete));
      asyncTaskQueue.cancelTask((*iter).first);
      terminateTask((*iter).first, *pTaskRequestTerminate, asyncTaskQueue, nodeOptions, message, resultDataItem, errorMsg, errorCode, taskState);
    }

    if (pTaskRequestTerminate->getCleanupFlag()==DRCETaskRequestTerminate::CleanupFlag::cfDelete)
    {
      asyncTaskQueue.setAllSubtaskCleanup(pTaskRequestTerminate->getTaskId());
    }

    Poco::File statusFile(FileStream::getStatusFileName(nodeOptions.getTasksStatusDir(), pTaskRequestTerminate->getTaskId()));
    if (statusFile.exists())
    {
      unsigned int errorCodeMainTask = errorCode;
      std::string errorMessage = errorMsg;

      DRCEResultDataItem resultItem = DRCECommonTask::extractStatusResultDataItem(pTaskRequestTerminate->getTaskId(), nodeOptions, asyncTaskQueue);
      if (resultItem.getErrorCode() == ERROR_TERMINATE_EXPIRED_TASK)
      {
        errorCode = ERROR_TERMINATE_SUBTASK_AFTER_EXPIRED_MAIN_TASK;
        errorCodeMainTask = ERROR_TERMINATE_EXPIRED_TASK;
        errorMessage = message(message_const::TASK_TIME_EXPIRED, asyncTask.timeMax);
        taskState = DRCETaskRequest::TaskState::TERMINATED_BY_TTL;
      }

      resultDataItem = DRCECommonTask::makeResultDataItem(inputMessage, nodeOptions, message,
          DRCETaskRequest::RequestType::rtSetTaskExecute, taskState,
          errorCode, errorMessage, 0, asyncTask.timeElapsed, resultDataItem.getExitStatus());
      resultDataItem.setErrorCode(errorCodeMainTask);

      DRCECommonTask::saveResultData(resultDataItem, nodeOptions, message, asyncTaskQueue);
    }

    //!<Subtasks list
    std::map<unsigned int, bool> sub = asyncTaskQueue.getSubtasks(pTaskRequestTerminate->getTaskId());

    for (std::map<unsigned int, bool>::iterator iter=sub.begin();iter!=sub.end();++iter)
    {
      outMsg.str("");
      outMsg << "DRCETaskRequestTerminateExecutor::executeTerminate  cleanup()  taskId: " << (*iter).first << "  cleanup: " << std::boolalpha << (*iter).second;
      Poco::Logger::root().log(Poco::Message(drce_const::moduleName, outMsg.str(), Poco::Message::Priority::PRIO_TRACE));

      DRCECommonTask::cleanup((*iter).first, nodeOptions, message, (*iter).second);
    }
  }
  catch(Poco::Exception& e)
  {
    resultDataItem.setErrorCode(e.code());
    resultDataItem.setErrorMessage(e.message());
    AsyncTasks asyncTask(asyncTaskQueue.getAsyncTask(pTaskRequest->getTaskId()));
    asyncTask.timeElapsed = DRCECommonTask::getElapsedTimeMsec(Poco::Timestamp::fromEpochTime(asyncTask.timeStart));
    asyncTask.state = DRCETaskRequest::TaskState::CRASHED;
    if (asyncTaskQueue.isExistAsyncTask(pTaskRequest->getTaskId()))
      asyncTaskQueue.safeChangeState(std::forward<AsyncTasks>(asyncTask), &resultDataItem, e.message(), e.code());
  }
  catch(std::exception& e)
  {
    resultDataItem.setErrorCode(ERROR_EXECUTE_TASK);
    resultDataItem.setErrorMessage(e.what());
    AsyncTasks asyncTask(asyncTaskQueue.getAsyncTask(pTaskRequest->getTaskId()));
    asyncTask.timeElapsed = DRCECommonTask::getElapsedTimeMsec(Poco::Timestamp::fromEpochTime(asyncTask.timeStart));
    asyncTask.state = DRCETaskRequest::TaskState::CRASHED;
    if (asyncTaskQueue.isExistAsyncTask(pTaskRequest->getTaskId()))
      asyncTaskQueue.safeChangeState(std::forward<AsyncTasks>(asyncTask), &resultDataItem, std::string(e.what()), ERROR_EXECUTE_TASK);
  }
  return resultDataItem;
}
//-----------------------------------------------------------------------------
void DRCETaskRequestTerminateExecutor::executeDefaultAlgorithm(DRCETaskRequestTerminate& taskRequestTerminate, pid_t pid, DRCEAsyncTasksQueue& asyncTaskQueue,
    DRCENodeOptions& nodeOptions, CustomMessage& message, DRCEResultDataItem& resultDataItem) throw (Poco::Exception)
{
  if (pid)
  {
    kill(pid, SIGTERM);
    resultDataItem.setExitStatus(HCE::drce::drce_const::EXIT_STATUS_BASE+SIGTERM);
    if (HCE::Process::isExistProcess(pid))
    {
      Poco::Thread::sleep(taskRequestTerminate.getDelayValue());

      std::string errorMsg;
      for (unsigned int i=0;i<taskRequestTerminate.getRepeatValue();++i)
      {
        if (kill(pid, SIGKILL) != 0)
          errorMsg = strerror(errno);
        if (HCE::Process::isExistProcess(pid))
          Poco::Thread::sleep(taskRequestTerminate.getDelayValue());
        else
          break;
      }
      if (!errorMsg.empty())
      {
        if (!loadStatusFile(asyncTaskQueue, nodeOptions, message, resultDataItem))
          throw Poco::Exception(message(message_const::PROCESS_WAS_NOT_STOPPED, errorMsg)+
                        message(message_const::PID, pid), ERROR_TERMINATE_TASK);
      }
      resultDataItem.setExitStatus(HCE::drce::drce_const::EXIT_STATUS_BASE+SIGKILL);
    }
  }
}
//-----------------------------------------------------------------------------
void DRCETaskRequestTerminateExecutor::executeCustomAlgorithm(DRCETaskRequestTerminate& taskRequestTerminate, pid_t pid, DRCEAsyncTasksQueue& asyncTaskQueue,
    DRCENodeOptions& nodeOptions, CustomMessage& message, DRCEResultDataItem& resultDataItem) throw (Poco::Exception)
{
  if (pid)
  {
    std::string errorMsg;
    for (unsigned int i=0;i<taskRequestTerminate.getRepeatValue();++i)
    {
      if (kill(pid, taskRequestTerminate.getSignalValue()) != 0)
        errorMsg = strerror(errno);
      if (HCE::Process::isExistProcess(pid))
        Poco::Thread::sleep(taskRequestTerminate.getDelayValue());
      else
        break;
    }
    if (!errorMsg.empty())
    {
      if (!loadStatusFile(asyncTaskQueue, nodeOptions, message, resultDataItem))
        throw Poco::Exception(message(message_const::PROCESS_WAS_NOT_STOPPED, errorMsg)+
                      message(message_const::PID, pid), ERROR_TERMINATE_TASK);
    }
    resultDataItem.setExitStatus(HCE::drce::drce_const::EXIT_STATUS_BASE+taskRequestTerminate.getSignalValue());
  }
}
//-----------------------------------------------------------------------------
bool DRCETaskRequestTerminateExecutor::loadStatusFile(DRCEAsyncTasksQueue& asyncTaskQueue,
    DRCENodeOptions& nodeOptions,
    CustomMessage& message,
    DRCEResultDataItem& resultDataItem)
{
  bool result = false;
  Poco::File statusFile(FileStream::getStatusFileName(nodeOptions.getTasksStatusDir(), resultDataItem.getRequestId()));
  if (statusFile.exists())
  {
    try
    {
      AsyncTaskLocker lock(asyncTaskQueue, resultDataItem.getRequestId());
      DataFileExtractor dataFileExtractor(message);
      resultDataItem = dataFileExtractor.extract(statusFile.path());
      result = true;
    }
    catch(Poco::Exception& e)
    {
      Poco::Logger::root().log(Poco::Message(drce_const::moduleName, e.displayText(), Poco::Message::Priority::PRIO_ERROR));
    }
  }
  return result;
}
//-----------------------------------------------------------------------------
void DRCETaskRequestTerminateExecutor::terminateTask(unsigned int taskId,
    DRCETaskRequestTerminate& taskRequestTerminate,
    DRCEAsyncTasksQueue& asyncTaskQueue,
    DRCENodeOptions& nodeOptions,
    CustomMessage& message,
    DRCEResultDataItem& resultDataItem,
    const std::string& errorMsg,
    unsigned int errorCode,
    DRCETaskRequest::TaskState taskState) throw (Poco::Exception)
{
  if (asyncTaskQueue.isExistAsyncTask(taskId))
  {
    AsyncTasks asyncTask(asyncTaskQueue.getAsyncTask(taskId));
    resultDataItem.setRequestId(taskId);
    resultDataItem.setPid(asyncTask.pid);
    asyncTask.timeElapsed = Poco::Timestamp::fromEpochTime(asyncTask.timeStart).elapsed()/1000;

    std::vector<pid_t> descendants = DRCEReadProcessData::getAllProcessIds(asyncTask.pid);
    switch(taskRequestTerminate.getAlgorithmType())
    {
      case DRCETaskRequestTerminate::AlgorithmType::atDefault:
        {
          executeDefaultAlgorithm(taskRequestTerminate, asyncTask.pid, asyncTaskQueue, nodeOptions, message, resultDataItem);
        }
        break;
      case DRCETaskRequestTerminate::AlgorithmType::atCustom:
        {
          if (taskRequestTerminate.getSignalValue() == 0)
            throw Poco::Exception(message(message_const::BAD_COMMAND_PARAMS), ERROR_BAD_COMMAND_PARAMS);
          executeCustomAlgorithm(taskRequestTerminate, asyncTask.pid, asyncTaskQueue, nodeOptions, message, resultDataItem);
        }
        break;
      default: throw Poco::Exception(message(message_const::WRONG_ALGORITHM_TYPE,
          static_cast<unsigned int>(taskRequestTerminate.getAlgorithmType())), ERROR_BAD_COMMAND_PARAMS);
    }

    std::stringstream outMsg;
    outMsg << "descendants count = " << descendants.size() << " (";
    for (size_t i=0;i<descendants.size();++i)
    {
      if (i!=0)
        outMsg << ", ";
      outMsg << descendants[i];
      kill(descendants[i], SIGKILL);
    }
    outMsg << ") has been killed...";
    Poco::Logger::root().log(Poco::Message(drce_const::moduleName, outMsg.str(), Poco::Message::Priority::PRIO_TRACE));

    resultDataItem.setTime(asyncTask.timeElapsed);
    asyncTask.state = taskState; //((errorCode == ERROR_TERMINATE_EXPIRED_TASK)?DRCETaskRequest::TaskState::TERMINATED_BY_TTL:DRCETaskRequest::TaskState::TERMINATED);
    asyncTaskQueue.safeChangeState(std::forward<AsyncTasks>(asyncTask), &resultDataItem, errorMsg, errorCode);
    // write to tasks log information about finished state of task
    asyncTaskQueue.writeToTasksLog(drce_const::DRCE_TASKS_LOGGER_DELIMITER, drce_const::DRCE_TASKS_LOGGER_DELETED, asyncTask);
  }
  else
  {
    std::stringstream outMsg;
    outMsg << "Task for terminate [ " << taskId << " ] alredy not exist in async queue";
    Poco::Logger::root().log(Poco::Message(drce_const::moduleName, outMsg.str(), Poco::Message::Priority::PRIO_DEBUG));
  }
}
//-----------------------------------------------------------------------------
DRCEResultDataItem DRCETaskRequestTerminateExecutor::makeResultDataItem(DRCEInputJsonMessage& inputJsonMessage, unsigned int exitStatus)
{
  return DRCECommonTask::makeResultDataItem(inputJsonMessage,
      nodeOptions,
      message,
      DRCETaskRequest::RequestType::rtTerminateTask,
      DRCETaskRequest::TaskState::TERMINATED,
      NO_ERROR,
      "",
      0,
      0,
      exitStatus);
}
//-----------------------------------------------------------------------------
void DRCETaskRequestTerminateExecutor::prepareCleanup(unsigned int taskId, DRCEInputJsonMessage& inputJsonMessage)
{
  if (taskId == inputJsonMessage.getRequestId())
  {
    Poco::SharedPtr<DRCETaskRequest> pTaskRequest = inputJsonMessage.getTaskRequest();
    DRCETaskRequestSetExecute* pTaskRequestSetExecute = dynamic_cast<DRCETaskRequestSetExecute*>(pTaskRequest.get());
    if (pTaskRequestSetExecute)
    {
      SessionOptions sessionOptions = pTaskRequestSetExecute->getSessionOptions();
      sessionOptions.cleanup = SessionOptions::CleanupFlag::cfDelete;
      pTaskRequestSetExecute->setSessionOptions(std::forward<SessionOptions>(sessionOptions));
      pTaskRequest = pTaskRequestSetExecute;
      inputJsonMessage.setTaskRequest(pTaskRequest);
    }
  }
  else
  {
    for (size_t i=0;i<inputJsonMessage.getSubtasksCount();++i)
    {
      DRCEInputJsonMessage inputMessage(inputJsonMessage.getSubtaskItem(i));
      prepareCleanup(taskId, inputMessage);
    }
  }
}
//-----------------------------------------------------------------------------
void DRCETaskRequestTerminateExecutor::terminateExpiredTask(const AsyncTasks& asyncTask,
    DRCEAsyncTasksQueue& asyncTaskQueue,
    DRCENodeOptions& nodeOptions,
    CustomMessage& message)
{
  DRCETaskRequestTerminateExecutor::executeTerminate(asyncTask.taskId,
      asyncTaskQueue,
      nodeOptions,
      message,
      message(message_const::TASK_TIME_EXPIRED, asyncTask.timeMax),
      ERROR_TERMINATE_EXPIRED_TASK,
      DRCETaskRequest::TaskState::TERMINATED_BY_TTL);
}
//-----------------------------------------------------------------------------
void DRCETaskRequestTerminateExecutor::executeTerminate(unsigned int taskId, DRCEAsyncTasksQueue& asyncTaskQueue, DRCENodeOptions& nodeOptions, CustomMessage& message,
    const std::string& errorMsg, unsigned int errorCode, DRCETaskRequest::TaskState taskState)
{
  DRCETaskRequestTerminate taskRequestTerminate;
  taskRequestTerminate.setTaskId(taskId);
  taskRequestTerminate.setAlgorithmType(DRCETaskRequestTerminate::AlgorithmType::atCustom);
  taskRequestTerminate.setDelayValue(100);
  taskRequestTerminate.setRepeatValue(3);
  taskRequestTerminate.setSignalValue(9);
  taskRequestTerminate.setCleanupFlag(DRCETaskRequestTerminate::CleanupFlag::cfNotDelete);
  DRCEInputJsonMessage inputJsonMessage;
  DRCETaskRequestTerminateExecutor::executeTerminate(&taskRequestTerminate, inputJsonMessage, asyncTaskQueue, nodeOptions, message, errorMsg, errorCode, taskState);
}
//-----------------------------------------------------------------------------
//-----------------------------------------------------------------------------
} // end namespace drce 
} // end namespace HCE 
