#include <gtest/gtest.h>
#include <fstream>
#include <iostream>
#include <utility>
#include <iomanip>
#include <Poco/File.h>
#include <Poco/Path.h>

#include "DRCEError.hpp"
#include "DRCESessionOptions.hpp"
#include "DRCETaskRequest.hpp"
#include "DRCETaskRequestSetExecute.hpp"
#include "DRCEOutputJsonMessage.hpp"
#include "DRCEFunctionalObject.hpp"
#include "DRCEResultDataSerializator.hpp"
#include "DRCEResultData.hpp"

#include "DRCEPrintStatus.hpp"
#include "DRCEActionTypesTest.hpp"
#include "DRCEEncodeDecodeBase64.hpp"
#include "DRCENotificationExecutor.hpp"
#include "DRCETaskThreadingModeTest.hpp"

namespace HCE
{
namespace drce
{
namespace tests
{
//-----------------------------------------------------------------------------
void DRCEActionTypesTest::testCreateBeforeExec(void)
{
  const std::string fileName = "file1";
  const std::string fileData = "data1";

  DRCENodeOptions nodeOptions = DRCETaskThreadingModeTest::getNodeOptions();
  DRCEFunctionalObject fObj(nodeOptions.getNodeName(), nodeOptions.getHomeDir(), nodeOptions.getTasksDataDir(), nodeOptions.getTasksStatusDir());
  fObj.setNodeHost("NODE_HOST");
  fObj.setNodePort("NODE_PORT");
  fObj.setNodeName("NODE_NAME");

  std::stringstream sourceJson;
  DRCETaskRequestSetExecute sourceMessage;

  sourceMessage.addFileItem(fileName, fileData, FileItem::CREATE_BEFORE_EXEC_AND_WRITE_DATA);

  sourceJson << sourceMessage;

//  std::cout << "sourceJson = " << sourceJson.str() << std::endl;

  ASSERT_FALSE(sourceMessage.isError()) << printError(sourceMessage);
  ASSERT_EQ(sourceMessage.getErrorCode(), NO_ERROR);
  ASSERT_TRUE(sourceMessage.getErrorMsg().empty());
  printSuccess("DRCETaskRequestSetExecute serialize");

  DRCEInputJsonMessage inputJsonMessage;
  inputJsonMessage.setRequestType(DRCETaskRequest::RequestType::rtSetTaskExecute);
  inputJsonMessage.setRequestData(sourceJson.str());
  inputJsonMessage.setRequestId(2345);

  sourceJson.str("");
  sourceJson << inputJsonMessage;

  ASSERT_FALSE(inputJsonMessage.isError()) << printError(inputJsonMessage);
  ASSERT_EQ(inputJsonMessage.getErrorCode(), NO_ERROR);
  ASSERT_TRUE(inputJsonMessage.getErrorMsg().empty());
  printSuccess("DRCETaskRequest serialize");

  std::string resultJson = fObj.Process(sourceJson.str());

//  std::cout << "resultJson = " << resultJson << std::endl;

  ASSERT_FALSE(fObj.isError()) << printError(fObj);
  ASSERT_EQ(fObj.getErrorCode(), NO_ERROR);
  ASSERT_TRUE(fObj.getErrorMsg().empty());
  printSuccess("DRCEFunctionalObject Process");

  DRCEOutputJsonMessage responseJsonMessage(resultJson);

  ASSERT_FALSE(responseJsonMessage.isError()) << printError(responseJsonMessage);
  ASSERT_EQ(responseJsonMessage.getErrorCode(), NO_ERROR);
  ASSERT_TRUE(responseJsonMessage.getErrorMsg().empty());
  printSuccess("DRCEOutputJsonMessage unserialize");

  ASSERT_EQ(responseJsonMessage.getResultData().getItemsCount(), 1);

  ASSERT_EQ(responseJsonMessage.getResultData().getDataItem(0).getNodeName(), "NODE_NAME");
  ASSERT_EQ(responseJsonMessage.getResultData().getDataItem(0).getNodeHost(), "NODE_HOST");
  ASSERT_EQ(responseJsonMessage.getResultData().getDataItem(0).getNodePort(), "NODE_PORT");
  ASSERT_TRUE(responseJsonMessage.getResultData().getDataItem(0).getStdoutStream().empty());
  ASSERT_TRUE(responseJsonMessage.getResultData().getDataItem(0).getStderrStream().empty());
  ASSERT_EQ(responseJsonMessage.getResultData().getDataItem(0).getFilesCount(), 0);

  std::ifstream ifs(fileName.c_str());
  ASSERT_TRUE(ifs.is_open());

  std::string content;
  ifs.seekg(0, std::ios::end);
  content.resize(ifs.tellg());
  ifs.seekg(0, std::ios::beg);
  ifs.read(const_cast<char*>(content.c_str()), content.size());
  ifs.close();
  ASSERT_EQ(content, fileData);

  Poco::File testFile(fileName);
  if (testFile.exists())
    testFile.remove();
  printSuccess("Test actionType CREATE_BEFORE_EXEC_AND_WRITE_DATA");
}
//-----------------------------------------------------------------------------
void DRCEActionTypesTest::testDeleteBeforeExec(void)
{
  const std::string fileName = "file1";
  const std::string fileData = "data1";

  DRCENodeOptions nodeOptions = DRCETaskThreadingModeTest::getNodeOptions();
  DRCEFunctionalObject fObj(nodeOptions.getNodeName(), nodeOptions.getHomeDir(), nodeOptions.getTasksDataDir(), nodeOptions.getTasksStatusDir());
  fObj.setNodeHost("NODE_HOST");
  fObj.setNodePort("NODE_PORT");
  fObj.setNodeName("NODE_NAME");

  std::ofstream ofs(fileName.c_str(), std::fstream::trunc);
  ofs << fileData;
  ofs.close();

  std::stringstream sourceJson;
  DRCETaskRequestSetExecute sourceMessage;

  sourceMessage.addFileItem(fileName, "", FileItem::DELETE_BEFORE_EXEC);

  sourceJson << sourceMessage;

//  std::cout << "sourceJson = " << sourceJson.str() << std::endl;

  ASSERT_FALSE(sourceMessage.isError()) << printError(sourceMessage);
  ASSERT_EQ(sourceMessage.getErrorCode(), NO_ERROR);
  ASSERT_TRUE(sourceMessage.getErrorMsg().empty());
  printSuccess("DRCETaskRequestSetExecute serialize");

  DRCEInputJsonMessage inputJsonMessage;
  inputJsonMessage.setRequestType(DRCETaskRequest::RequestType::rtSetTaskExecute);
  inputJsonMessage.setRequestData(sourceJson.str());
  inputJsonMessage.setRequestId(2345);

  sourceJson.str("");
  sourceJson << inputJsonMessage;

  ASSERT_FALSE(inputJsonMessage.isError()) << printError(inputJsonMessage);
  ASSERT_EQ(inputJsonMessage.getErrorCode(), NO_ERROR);
  ASSERT_TRUE(inputJsonMessage.getErrorMsg().empty());
  printSuccess("DRCETaskRequest serialize");

  std::string resultJson = fObj.Process(sourceJson.str());

//  std::cout << "resultJson = " << resultJson << std::endl;

  ASSERT_FALSE(fObj.isError()) << printError(fObj);
  ASSERT_EQ(fObj.getErrorCode(), NO_ERROR);
  ASSERT_TRUE(fObj.getErrorMsg().empty());
  printSuccess("DRCEFunctionalObject Process");

  DRCEOutputJsonMessage responseJsonMessage(resultJson);

  ASSERT_FALSE(responseJsonMessage.isError()) << printError(responseJsonMessage);
  ASSERT_EQ(responseJsonMessage.getErrorCode(), NO_ERROR);
  ASSERT_TRUE(responseJsonMessage.getErrorMsg().empty());
  printSuccess("DRCEOutputJsonMessage unserialize");

  ASSERT_EQ(responseJsonMessage.getResultData().getItemsCount(), 1);

  ASSERT_EQ(responseJsonMessage.getResultData().getDataItem(0).getNodeName(), "NODE_NAME");
  ASSERT_EQ(responseJsonMessage.getResultData().getDataItem(0).getNodeHost(), "NODE_HOST");
  ASSERT_EQ(responseJsonMessage.getResultData().getDataItem(0).getNodePort(), "NODE_PORT");
  ASSERT_TRUE(responseJsonMessage.getResultData().getDataItem(0).getStdoutStream().empty());
  ASSERT_TRUE(responseJsonMessage.getResultData().getDataItem(0).getStderrStream().empty());
  ASSERT_EQ(responseJsonMessage.getResultData().getDataItem(0).getFilesCount(), 0);

  Poco::File testFile(fileName);
  ASSERT_FALSE(testFile.exists());
  printSuccess("Test actionType DELETE_BEFORE_EXEC");
}
//-----------------------------------------------------------------------------
void DRCEActionTypesTest::testReadAfterExec(void)
{
  const std::string fileName = "file1";
  const std::string fileData = "data1";

  DRCENodeOptions nodeOptions = DRCETaskThreadingModeTest::getNodeOptions();
  DRCEFunctionalObject fObj(nodeOptions.getNodeName(), nodeOptions.getHomeDir(), nodeOptions.getTasksDataDir(), nodeOptions.getTasksStatusDir());
  fObj.setNodeHost("NODE_HOST");
  fObj.setNodePort("NODE_PORT");
  fObj.setNodeName("NODE_NAME");

  Poco::Path testPath(fileName);
  Poco::File testFile(testPath.absolute().toString());

  std::ofstream ofs(testFile.path().c_str(), std::fstream::trunc);
  ofs << fileData << std::flush;
  ofs.close();

  Poco::Thread::sleep(2000);

  std::stringstream sourceJson;
  DRCETaskRequestSetExecute sourceMessage;

  sourceMessage.addFileItem(testFile.path(), "", FileItem::READ_AFTER_EXEC_AND_RETURN_RESULT);

  sourceJson << sourceMessage;

//  std::cout << "sourceJson = " << sourceJson.str() << std::endl;

  ASSERT_FALSE(sourceMessage.isError()) << printError(sourceMessage);
  ASSERT_EQ(sourceMessage.getErrorCode(), NO_ERROR);
  ASSERT_TRUE(sourceMessage.getErrorMsg().empty());
  printSuccess("DRCETaskRequestSetExecute serialize");

  DRCEInputJsonMessage inputJsonMessage;
  inputJsonMessage.setRequestType(DRCETaskRequest::RequestType::rtSetTaskExecute);
  inputJsonMessage.setRequestData(sourceJson.str());
  inputJsonMessage.setRequestId(234567);

  sourceJson.str("");
  sourceJson << inputJsonMessage;

  ASSERT_FALSE(inputJsonMessage.isError()) << printError(inputJsonMessage);
  ASSERT_EQ(inputJsonMessage.getErrorCode(), NO_ERROR);
  ASSERT_TRUE(inputJsonMessage.getErrorMsg().empty());
  printSuccess("DRCETaskRequest serialize");

  std::string resultJson = fObj.Process(sourceJson.str());

  std::cout << "resultJson = " << resultJson << std::endl;

  ASSERT_FALSE(fObj.isError()) << printError(fObj);
  ASSERT_EQ(fObj.getErrorCode(), NO_ERROR);
  ASSERT_TRUE(fObj.getErrorMsg().empty());
  printSuccess("DRCEFunctionalObject Process");

  DRCEOutputJsonMessage responseJsonMessage(resultJson);

  ASSERT_FALSE(responseJsonMessage.isError()) << printError(responseJsonMessage);
  ASSERT_EQ(responseJsonMessage.getErrorCode(), NO_ERROR);
  ASSERT_TRUE(responseJsonMessage.getErrorMsg().empty());
  printSuccess("DRCEOutputJsonMessage unserialize");

  ASSERT_EQ(responseJsonMessage.getResultData().getItemsCount(), 1);

  ASSERT_EQ(responseJsonMessage.getResultData().getDataItem(0).getNodeName(), "NODE_NAME");
  ASSERT_EQ(responseJsonMessage.getResultData().getDataItem(0).getNodeHost(), "NODE_HOST");
  ASSERT_EQ(responseJsonMessage.getResultData().getDataItem(0).getNodePort(), "NODE_PORT");
  ASSERT_TRUE(responseJsonMessage.getResultData().getDataItem(0).getStdoutStream().empty());
  ASSERT_TRUE(responseJsonMessage.getResultData().getDataItem(0).getStderrStream().empty());
  ASSERT_EQ(responseJsonMessage.getResultData().getDataItem(0).getFilesCount(), 1);

  ASSERT_EQ(responseJsonMessage.getResultData().getDataItem(0).getFileItem(0).actionType, FileItem::READ_AFTER_EXEC_AND_RETURN_RESULT);
  ASSERT_EQ(responseJsonMessage.getResultData().getDataItem(0).getFileItem(0).name, testFile.path());
  ASSERT_EQ(responseJsonMessage.getResultData().getDataItem(0).getFileItem(0).data, fileData);

  if (testFile.exists())
    testFile.remove();
  printSuccess("Test actionType READ_AFTER_EXEC_AND_RETURN_RESULT");
}
//-----------------------------------------------------------------------------
void DRCEActionTypesTest::testDeleteAfterExec(void)
{
  const std::string fileName = "file1";
  const std::string fileData = "data1";

  DRCENodeOptions nodeOptions = DRCETaskThreadingModeTest::getNodeOptions();
  DRCEFunctionalObject fObj(nodeOptions.getNodeName(), nodeOptions.getHomeDir(), nodeOptions.getTasksDataDir(), nodeOptions.getTasksStatusDir());
  fObj.setNodeHost("NODE_HOST");
  fObj.setNodePort("NODE_PORT");
  fObj.setNodeName("NODE_NAME");

  std::ofstream ofs(fileName.c_str());
  ofs << fileData;
  ofs.close();

  std::stringstream sourceJson;
  DRCETaskRequestSetExecute sourceMessage;

  sourceMessage.addFileItem(fileName, "", FileItem::DELETE_AFTER_EXEC);

  sourceJson << sourceMessage;

//  std::cout << "sourceJson = " << sourceJson.str() << std::endl;

  ASSERT_FALSE(sourceMessage.isError()) << printError(sourceMessage);
  ASSERT_EQ(sourceMessage.getErrorCode(), NO_ERROR);
  ASSERT_TRUE(sourceMessage.getErrorMsg().empty());
  printSuccess("DRCETaskRequestSetExecute serialize");

  DRCEInputJsonMessage inputJsonMessage;
  inputJsonMessage.setRequestType(DRCETaskRequest::RequestType::rtSetTaskExecute);
  inputJsonMessage.setRequestData(sourceJson.str());
  inputJsonMessage.setRequestId(2345);

  sourceJson.str("");
  sourceJson << inputJsonMessage;

  ASSERT_FALSE(inputJsonMessage.isError()) << printError(inputJsonMessage);
  ASSERT_EQ(inputJsonMessage.getErrorCode(), NO_ERROR);
  ASSERT_TRUE(inputJsonMessage.getErrorMsg().empty());
  printSuccess("DRCETaskRequest serialize");

  std::string resultJson = fObj.Process(sourceJson.str());

//  std::cout << "resultJson = " << resultJson << std::endl;

  ASSERT_FALSE(fObj.isError()) << printError(fObj);
  ASSERT_EQ(fObj.getErrorCode(), NO_ERROR);
  ASSERT_TRUE(fObj.getErrorMsg().empty());
  printSuccess("DRCEFunctionalObject Process");

  DRCEOutputJsonMessage responseJsonMessage(resultJson);

  ASSERT_FALSE(responseJsonMessage.isError()) << printError(responseJsonMessage);
  ASSERT_EQ(responseJsonMessage.getErrorCode(), NO_ERROR);
  ASSERT_TRUE(responseJsonMessage.getErrorMsg().empty());
  printSuccess("DRCEOutputJsonMessage unserialize");

  ASSERT_EQ(responseJsonMessage.getResultData().getItemsCount(), 1);

  ASSERT_EQ(responseJsonMessage.getResultData().getDataItem(0).getNodeName(), "NODE_NAME");
  ASSERT_EQ(responseJsonMessage.getResultData().getDataItem(0).getNodeHost(), "NODE_HOST");
  ASSERT_EQ(responseJsonMessage.getResultData().getDataItem(0).getNodePort(), "NODE_PORT");
  ASSERT_TRUE(responseJsonMessage.getResultData().getDataItem(0).getStdoutStream().empty());
  ASSERT_TRUE(responseJsonMessage.getResultData().getDataItem(0).getStderrStream().empty());
  ASSERT_EQ(responseJsonMessage.getResultData().getDataItem(0).getFilesCount(), 0);

  Poco::File testFile(fileName);
  ASSERT_FALSE(testFile.exists());
  printSuccess("Test actionType DELETE_AFTER_EXEC");
}
//-----------------------------------------------------------------------------
void DRCEActionTypesTest::testEncodeDecodeContent(void)
{
  const std::string fileName = "file1";
  const std::string fileData = "data1";

  DRCENodeOptions nodeOptions = DRCETaskThreadingModeTest::getNodeOptions();
  DRCEFunctionalObject fObj(nodeOptions.getNodeName(), nodeOptions.getHomeDir(), nodeOptions.getTasksDataDir(), nodeOptions.getTasksStatusDir());
  fObj.setNodeHost("NODE_HOST");
  fObj.setNodePort("NODE_PORT");
  fObj.setNodeName("NODE_NAME");

  std::stringstream sourceJson;
  DRCETaskRequestSetExecute sourceMessage;

  sourceMessage.addFileItem(fileName, DRCEEncodeBase64(fileData),
      FileItem::CREATE_BEFORE_EXEC_AND_WRITE_DATA | FileItem::READ_AFTER_EXEC_AND_RETURN_RESULT |
      FileItem::DELETE_AFTER_EXEC | FileItem::ENCODE_DECODE_CONTENT_BASE64);

  sourceJson << sourceMessage;

//  std::cout << "sourceJson = " << sourceJson.str() << std::endl;

  ASSERT_FALSE(sourceMessage.isError()) << printError(sourceMessage);
  ASSERT_EQ(sourceMessage.getErrorCode(), NO_ERROR);
  ASSERT_TRUE(sourceMessage.getErrorMsg().empty());
  printSuccess("DRCETaskRequestSetExecute serialize");

  DRCEInputJsonMessage inputJsonMessage;
  inputJsonMessage.setRequestType(DRCETaskRequest::RequestType::rtSetTaskExecute);
  inputJsonMessage.setRequestData(sourceJson.str());
  inputJsonMessage.setRequestId(2345);

  sourceJson.str("");
  sourceJson << inputJsonMessage;

  ASSERT_FALSE(inputJsonMessage.isError()) << printError(inputJsonMessage);
  ASSERT_EQ(inputJsonMessage.getErrorCode(), NO_ERROR);
  ASSERT_TRUE(inputJsonMessage.getErrorMsg().empty());
  printSuccess("DRCETaskRequest serialize");

  std::string resultJson = fObj.Process(sourceJson.str());

//  std::cout << "resultJson = " << resultJson << std::endl;

  ASSERT_FALSE(fObj.isError()) << printError(fObj);
  ASSERT_EQ(fObj.getErrorCode(), NO_ERROR);
  ASSERT_TRUE(fObj.getErrorMsg().empty());
  printSuccess("DRCEFunctionalObject Process");

  DRCEOutputJsonMessage responseJsonMessage(resultJson);

  ASSERT_FALSE(responseJsonMessage.isError()) << printError(responseJsonMessage);
  ASSERT_EQ(responseJsonMessage.getErrorCode(), NO_ERROR);
  ASSERT_TRUE(responseJsonMessage.getErrorMsg().empty());
  printSuccess("DRCEOutputJsonMessage unserialize");

  ASSERT_EQ(responseJsonMessage.getResultData().getItemsCount(), 1);

  ASSERT_EQ(responseJsonMessage.getResultData().getDataItem(0).getNodeName(), "NODE_NAME");
  ASSERT_EQ(responseJsonMessage.getResultData().getDataItem(0).getNodeHost(), "NODE_HOST");
  ASSERT_EQ(responseJsonMessage.getResultData().getDataItem(0).getNodePort(), "NODE_PORT");
  ASSERT_TRUE(responseJsonMessage.getResultData().getDataItem(0).getStdoutStream().empty());
  ASSERT_TRUE(responseJsonMessage.getResultData().getDataItem(0).getStderrStream().empty());
  ASSERT_EQ(responseJsonMessage.getResultData().getDataItem(0).getFilesCount(), 1);

  ASSERT_EQ(responseJsonMessage.getResultData().getDataItem(0).getFileItem(0).actionType,
      FileItem::CREATE_BEFORE_EXEC_AND_WRITE_DATA |
      FileItem::READ_AFTER_EXEC_AND_RETURN_RESULT |
      FileItem::DELETE_AFTER_EXEC |
      FileItem::ENCODE_DECODE_CONTENT_BASE64);

  ASSERT_EQ(responseJsonMessage.getResultData().getDataItem(0).getFileItem(0).name, fileName);
  ASSERT_EQ(DRCEDecodeBase64(responseJsonMessage.getResultData().getDataItem(0).getFileItem(0).data), fileData);

  Poco::File testFile(fileName);
  ASSERT_FALSE(testFile.exists());
  printSuccess("Test actionType ENCODE_DECODE_CONTENT_BASE64");}
//-----------------------------------------------------------------------------
//-----------------------------------------------------------------------------
} // namespace tests 
} // namespace drce
} // namespace HCE
