#include <gtest/gtest.h>
#include <sstream>
#include <Poco/File.h>

#include "DRCEFileStream.hpp"
#include "DRCENodeOptions.hpp"
#include "DRCEFunctionalObject.hpp"
#include "DRCETaskRequestSetExecute.hpp"
#include "DRCETaskRequestGetData.hpp"
#include "DRCETaskRequestDeleteData.hpp"
#include "DRCETaskRequestTerminate.hpp"

#include "DRCETaskThreadingModeTest.hpp"
#include "DRCEPrintStatus.hpp"
#include "DRCECleanupTest.hpp"
#include "ProcExec.hpp"
#include "DRCEFileExtractor.hpp"
#include "CustomMessage.hpp"
#include "DRCEMessageConst.hpp"

namespace HCE
{
namespace drce
{
namespace tests
{
//-----------------------------------------------------------------------------
DRCEResultDataItem DRCECleanupTest::extractResultData(unsigned int taskId, DRCENodeOptions& nodeOptions)
{
  DRCEResultDataItem resultDataItem;
  try
  {
    MessagesCollection messagesCollection(message_const::messages);
    CustomMessage message(messagesCollection);
    DataFileExtractor dataFileExctractor(message);
    resultDataItem = dataFileExctractor.extract(FileStream::getDataFileName(nodeOptions.getTasksDataDir(), taskId));
  }
  catch(Poco::Exception& e)
  {
    printError("Extract result data ("+e.message()+")");
  }
  return resultDataItem;
}
//-----------------------------------------------------------------------------
std::string DRCECleanupTest::getInputJson(unsigned int startTaskId,
    SessionOptions::ThreadMode threadMode,
    SessionOptions::CleanupFlag cleanupFlag,
    const std::string& commandLine,
    unsigned int timeMax)
{
  DRCEInputJsonMessage inputJsonMessage = getInputJsonMessage(startTaskId, threadMode, cleanupFlag, commandLine, timeMax);

  const unsigned int SZ = 6;
  DRCEInputJsonMessage subtasks[SZ];
  for (unsigned int i=0;i<SZ;++i)
  {
    subtasks[i] = getInputJsonMessage(++startTaskId,
        (!(i%2))?SessionOptions::ThreadMode::tmAsync:SessionOptions::ThreadMode::tmSync,
        SessionOptions::CleanupFlag::cfNotDelete,
        commandLine,
        timeMax);
  }

  subtasks[0].addSubtaskItem(subtasks[2]);
  subtasks[0].addSubtaskItem(subtasks[3]);

  subtasks[1].addSubtaskItem(subtasks[4]);
  subtasks[1].addSubtaskItem(subtasks[5]);

  inputJsonMessage.addSubtaskItem(subtasks[0]);
  inputJsonMessage.addSubtaskItem(subtasks[1]);

  std::stringstream jsonStream;
  jsonStream << inputJsonMessage;

  return jsonStream.str();
}
//-----------------------------------------------------------------------------
DRCEInputJsonMessage DRCECleanupTest::getInputJsonMessage(unsigned int taskId,
    SessionOptions::ThreadMode threadMode,
    SessionOptions::CleanupFlag cleanupFlag,
    const std::string& commandLine,
    unsigned int timeMax)
{
  DRCETaskRequestSetExecute taskRequestSetExecute;
  taskRequestSetExecute.setRequestType(DRCETaskRequest::RequestType::rtSetTaskExecute);
  taskRequestSetExecute.setTaskId(taskId);

  if (commandLine.empty())
  {
    taskRequestSetExecute.setCommandLine("sort");
    taskRequestSetExecute.setInputStream("def\nabc\n123\n");
  }
  else
    taskRequestSetExecute.setCommandLine(commandLine);

  SessionOptions sessionOptions;
  sessionOptions.tmode = threadMode;
  sessionOptions.cleanup = cleanupFlag;
  sessionOptions.timeMax = timeMax;
//  sessionOptions.shellName = "hce-starter-bash.sh";
  taskRequestSetExecute.setSessionOptions(std::forward<SessionOptions>(sessionOptions));

  std::stringstream jsonStream;

  jsonStream << taskRequestSetExecute;

  DRCEInputJsonMessage inputJsonMessage;
  inputJsonMessage.setRequestId(taskId);
  inputJsonMessage.setRequestType(DRCETaskRequest::RequestType::rtSetTaskExecute);
  inputJsonMessage.setRequestData(jsonStream.str());

  return inputJsonMessage;
}
//-----------------------------------------------------------------------------
std::string DRCECleanupTest::getInputJson(unsigned int startTaskId,
    SessionOptions::ThreadMode taskThreadMode,
    SessionOptions::ThreadMode subtasksThreadMode)
{
  DRCEInputJsonMessage inputJsonMessage = getInputJsonMessage(startTaskId, taskThreadMode, SessionOptions::CleanupFlag::cfNotDelete);

  const unsigned int SZ = 6;
  DRCEInputJsonMessage subtasks[SZ];
  for (unsigned int i=0;i<SZ;++i)
  {
    subtasks[i] = getInputJsonMessage(++startTaskId, subtasksThreadMode, SessionOptions::CleanupFlag::cfNotDelete);
  }

  subtasks[0].addSubtaskItem(subtasks[2]);
  subtasks[0].addSubtaskItem(subtasks[3]);

  subtasks[1].addSubtaskItem(subtasks[4]);
  subtasks[1].addSubtaskItem(subtasks[5]);

  inputJsonMessage.addSubtaskItem(subtasks[0]);
  inputJsonMessage.addSubtaskItem(subtasks[1]);

  std::stringstream jsonStream;
  jsonStream << inputJsonMessage;

  return jsonStream.str();
}
//-----------------------------------------------------------------------------
std::string DRCECleanupTest::getInputJson(unsigned int startTaskId, SessionOptions::ThreadMode taskThreadMode)
{
  DRCEInputJsonMessage inputJsonMessage = getInputJsonMessage(startTaskId, taskThreadMode, SessionOptions::CleanupFlag::cfNotDelete);

  const unsigned int SZ = 14;
  DRCEInputJsonMessage subtasks[SZ];
  for (unsigned int i=0;i<SZ;++i)
  {
    subtasks[i] = getInputJsonMessage(++startTaskId,
        (!(i%2))?SessionOptions::ThreadMode::tmAsync:SessionOptions::ThreadMode::tmSync,
            SessionOptions::CleanupFlag::cfNotDelete);
  }

  subtasks[2].addSubtaskItem(subtasks[6]);
  subtasks[2].addSubtaskItem(subtasks[7]);

  subtasks[3].addSubtaskItem(subtasks[8]);
  subtasks[3].addSubtaskItem(subtasks[9]);

  subtasks[4].addSubtaskItem(subtasks[10]);
  subtasks[4].addSubtaskItem(subtasks[11]);

  subtasks[5].addSubtaskItem(subtasks[12]);
  subtasks[5].addSubtaskItem(subtasks[13]);


  subtasks[0].addSubtaskItem(subtasks[2]);
  subtasks[0].addSubtaskItem(subtasks[3]);

  subtasks[1].addSubtaskItem(subtasks[4]);
  subtasks[1].addSubtaskItem(subtasks[5]);

  inputJsonMessage.addSubtaskItem(subtasks[0]);
  inputJsonMessage.addSubtaskItem(subtasks[1]);

  std::stringstream jsonStream;
  jsonStream << inputJsonMessage;

  return jsonStream.str();
}
//-----------------------------------------------------------------------------
std::string DRCECleanupTest::getInputJsonCleanup(unsigned int startTaskId, SessionOptions::ThreadMode taskThreadMode, const std::string& commandLine, unsigned int timeMax)
{
  DRCEInputJsonMessage inputJsonMessage = getInputJsonMessage(startTaskId, taskThreadMode, SessionOptions::CleanupFlag::cfNotDelete, commandLine, timeMax);

  const unsigned int SZ = 14;
  DRCEInputJsonMessage subtasks[SZ];
  for (unsigned int i=0;i<SZ;++i)
  {
    subtasks[i] = getInputJsonMessage(++startTaskId,
        (!(i%2))?SessionOptions::ThreadMode::tmAsync:SessionOptions::ThreadMode::tmSync,
            (!(i%2))?SessionOptions::CleanupFlag::cfDelete:SessionOptions::CleanupFlag::cfNotDelete,
                commandLine,
                timeMax);
  }

  subtasks[2].addSubtaskItem(subtasks[6]);
  subtasks[2].addSubtaskItem(subtasks[7]);

  subtasks[3].addSubtaskItem(subtasks[8]);
  subtasks[3].addSubtaskItem(subtasks[9]);

  subtasks[4].addSubtaskItem(subtasks[10]);
  subtasks[4].addSubtaskItem(subtasks[11]);

  subtasks[5].addSubtaskItem(subtasks[12]);
  subtasks[5].addSubtaskItem(subtasks[13]);


  subtasks[0].addSubtaskItem(subtasks[2]);
  subtasks[0].addSubtaskItem(subtasks[3]);

  subtasks[1].addSubtaskItem(subtasks[4]);
  subtasks[1].addSubtaskItem(subtasks[5]);

  inputJsonMessage.addSubtaskItem(subtasks[0]);
  inputJsonMessage.addSubtaskItem(subtasks[1]);

  std::stringstream jsonStream;
  jsonStream << inputJsonMessage;

  return jsonStream.str();
}
//-----------------------------------------------------------------------------
std::string DRCECleanupTest::getInputJsonCleanup(unsigned int startTaskId, SessionOptions::ThreadMode taskThreadMode,
    SessionOptions::CleanupFlag cleanupFlag, const std::string& commandLine)
{
  DRCEInputJsonMessage inputJsonMessage = getInputJsonMessage(startTaskId, taskThreadMode, SessionOptions::CleanupFlag::cfNotDelete, commandLine);

  const unsigned int SZ = 14;
  DRCEInputJsonMessage subtasks[SZ];
  for (unsigned int i=0;i<SZ;++i)
  {
    subtasks[i] = getInputJsonMessage(++startTaskId,
        (!(i%2))?SessionOptions::ThreadMode::tmAsync:SessionOptions::ThreadMode::tmSync, cleanupFlag, commandLine);
  }

  subtasks[2].addSubtaskItem(subtasks[6]);
  subtasks[2].addSubtaskItem(subtasks[7]);

  subtasks[3].addSubtaskItem(subtasks[8]);
  subtasks[3].addSubtaskItem(subtasks[9]);

  subtasks[4].addSubtaskItem(subtasks[10]);
  subtasks[4].addSubtaskItem(subtasks[11]);

  subtasks[5].addSubtaskItem(subtasks[12]);
  subtasks[5].addSubtaskItem(subtasks[13]);


  subtasks[0].addSubtaskItem(subtasks[2]);
  subtasks[0].addSubtaskItem(subtasks[3]);

  subtasks[1].addSubtaskItem(subtasks[4]);
  subtasks[1].addSubtaskItem(subtasks[5]);

  inputJsonMessage.addSubtaskItem(subtasks[0]);
  inputJsonMessage.addSubtaskItem(subtasks[1]);

  std::stringstream jsonStream;
  jsonStream << inputJsonMessage;

  return jsonStream.str();
}
//-----------------------------------------------------------------------------
bool DRCECleanupTest::isExistTaskData(unsigned int taskId, DRCENodeOptions& nodeOptions)
{
  Poco::File dataFile(FileStream::getDataFileName(nodeOptions.getTasksDataDir(), taskId));
  Poco::File statusFile(FileStream::getStatusFileName(nodeOptions.getTasksStatusDir(), taskId));
  Poco::File requestFile(FileStream::getRequestFileName(nodeOptions.getTasksDataDir(), taskId));

  return (dataFile.exists() && statusFile.exists() && requestFile.exists());
}
//-----------------------------------------------------------------------------
bool DRCECleanupTest::isNotExistTaskData(unsigned int taskId, DRCENodeOptions& nodeOptions)
{
  Poco::File dataFile(FileStream::getDataFileName(nodeOptions.getTasksDataDir(), taskId));
  Poco::File statusFile(FileStream::getStatusFileName(nodeOptions.getTasksStatusDir(), taskId));
  Poco::File requestFile(FileStream::getRequestFileName(nodeOptions.getTasksDataDir(), taskId));

  return (!dataFile.exists() && !statusFile.exists() && !requestFile.exists());
}
//-----------------------------------------------------------------------------
void DRCECleanupTest::cleanupDir(const std::string& dirName)
{
  Poco::File directory(dirName);
  if (directory.exists())
  {
    std::vector<std::string> files;
    directory.list(files);
    for (const std::string fileName : files)
    {
      Poco::File file(fileName);
      if (file.exists())
        file.remove();
    }
  }
}
//-----------------------------------------------------------------------------
void DRCECleanupTest::cleanupTestData(DRCENodeOptions& nodeOptions)
{
#if CLEANUP_TEST == 1
  cleanupDir(nodeOptions.getTasksDataDir());
  cleanupDir(nodeOptions.getTasksStatusDir());
#endif
}
//-----------------------------------------------------------------------------
//-----------------------------------------------------------------------------
void DRCECleanupTest::testAsyncTaskAsyncSubtasks(void)
{
  DRCENodeOptions nodeOptions = DRCETaskThreadingModeTest::getNodeOptions();
  cleanupTestData(nodeOptions);
  DRCEFunctionalObject fObj(nodeOptions.getNodeName(), nodeOptions.getHomeDir(), nodeOptions.getTasksDataDir(), nodeOptions.getTasksStatusDir());
  fObj.setNodeHost("NODE_HOST");
  fObj.setNodePort("NODE_PORT");
  fObj.setNodeName("NODE_NAME");

  unsigned int startTaskId = 600;

  std::string sourceJson = getInputJson(startTaskId, SessionOptions::ThreadMode::tmAsync, SessionOptions::ThreadMode::tmAsync);

//  std::cout << "sourceJson: " << sourceJson << std::endl;

  std::string resultJson = fObj.Process(sourceJson);

  ASSERT_FALSE(fObj.isError()) << printError(fObj);
  ASSERT_EQ(fObj.getErrorCode(), NO_ERROR);
  ASSERT_TRUE(fObj.getErrorMsg().empty());
  printSuccess("DRCEFunctionalObject Process");

//  std::cout << "result JSON: " << resultJson << std::endl;

  // check right result (unserialize and compare two objects)
  DRCEOutputJsonMessage resultJsonMessage(resultJson);

  ASSERT_FALSE(resultJsonMessage.isError()) << printError(resultJsonMessage);
  ASSERT_EQ(resultJsonMessage.getErrorCode(), NO_ERROR);
  ASSERT_TRUE(resultJsonMessage.getErrorMsg().empty());
  printSuccess("DRCEOutputJsonMessage unserialize");

  ASSERT_EQ(resultJsonMessage.getResultData().getItemsCount(), 1);
  ASSERT_TRUE(resultJsonMessage.getResultData().getDataItem(0).getStderrStream().empty());
  ASSERT_EQ(resultJsonMessage.getResultData().getDataItem(0).getErrorCode(), NO_ERROR);
  ASSERT_EQ(resultJsonMessage.getResultData().getDataItem(0).getNodeName(), "NODE_NAME");
  ASSERT_EQ(resultJsonMessage.getResultData().getDataItem(0).getNodeHost(), "NODE_HOST");
  ASSERT_EQ(resultJsonMessage.getResultData().getDataItem(0).getNodePort(), "NODE_PORT");
  ASSERT_TRUE(resultJsonMessage.getResultData().getDataItem(0).getState() == DRCETaskRequest::TaskState::SET_AS_NEW ||
      resultJsonMessage.getResultData().getDataItem(0).getState() == DRCETaskRequest::TaskState::IN_PROGRESS ||
      resultJsonMessage.getResultData().getDataItem(0).getState() == DRCETaskRequest::TaskState::FINISHED);
  ASSERT_EQ(resultJsonMessage.getResultData().getDataItem(0).getSubtasksCount(), 2);

  ASSERT_TRUE(resultJsonMessage.getResultData().getDataItem(0).getSubtaskItem(0).getStderrStream().empty());
  ASSERT_EQ(resultJsonMessage.getResultData().getDataItem(0).getSubtaskItem(0).getErrorCode(), NO_ERROR);
  ASSERT_EQ(resultJsonMessage.getResultData().getDataItem(0).getSubtaskItem(0).getNodeName(), "NODE_NAME");
  ASSERT_EQ(resultJsonMessage.getResultData().getDataItem(0).getSubtaskItem(0).getNodeHost(), "NODE_HOST");
  ASSERT_EQ(resultJsonMessage.getResultData().getDataItem(0).getSubtaskItem(0).getNodePort(), "NODE_PORT");
  ASSERT_TRUE(resultJsonMessage.getResultData().getDataItem(0).getSubtaskItem(0).getState() == DRCETaskRequest::TaskState::SET_AS_NEW ||
      resultJsonMessage.getResultData().getDataItem(0).getSubtaskItem(0).getState() == DRCETaskRequest::TaskState::IN_PROGRESS ||
      resultJsonMessage.getResultData().getDataItem(0).getSubtaskItem(0).getState() == DRCETaskRequest::TaskState::FINISHED ||
      resultJsonMessage.getResultData().getDataItem(0).getSubtaskItem(0).getState() == DRCETaskRequest::TaskState::QUEUED_TO_RUN);
  ASSERT_EQ(resultJsonMessage.getResultData().getDataItem(0).getSubtaskItem(0).getSubtasksCount(), 2);

  ASSERT_TRUE(resultJsonMessage.getResultData().getDataItem(0).getSubtaskItem(1).getStderrStream().empty());
  ASSERT_EQ(resultJsonMessage.getResultData().getDataItem(0).getSubtaskItem(1).getErrorCode(), NO_ERROR);
  ASSERT_EQ(resultJsonMessage.getResultData().getDataItem(0).getSubtaskItem(1).getNodeName(), "NODE_NAME");
  ASSERT_EQ(resultJsonMessage.getResultData().getDataItem(0).getSubtaskItem(1).getNodeHost(), "NODE_HOST");
  ASSERT_EQ(resultJsonMessage.getResultData().getDataItem(0).getSubtaskItem(1).getNodePort(), "NODE_PORT");
  ASSERT_TRUE(resultJsonMessage.getResultData().getDataItem(0).getSubtaskItem(1).getState() == DRCETaskRequest::TaskState::SET_AS_NEW ||
      resultJsonMessage.getResultData().getDataItem(0).getSubtaskItem(1).getState() == DRCETaskRequest::TaskState::IN_PROGRESS ||
      resultJsonMessage.getResultData().getDataItem(0).getSubtaskItem(1).getState() == DRCETaskRequest::TaskState::FINISHED ||
      resultJsonMessage.getResultData().getDataItem(0).getSubtaskItem(1).getState() == DRCETaskRequest::TaskState::QUEUED_TO_RUN);
  ASSERT_EQ(resultJsonMessage.getResultData().getDataItem(0).getSubtaskItem(1).getSubtasksCount(), 2);

  while (fObj.getAsyncTasksCount())
  {
    Poco::Thread::sleep(5000);
    std::cout << "wait end task" << std::endl;
  }
  Poco::Thread::sleep(2000);

  DRCEResultDataItem resultDataItem = DRCECleanupTest::extractResultData(startTaskId, fObj);

  ASSERT_TRUE(resultDataItem.getStderrStream().empty());
  ASSERT_EQ(resultDataItem.getErrorCode(), NO_ERROR);
  ASSERT_EQ(resultDataItem.getNodeName(), fObj.getNodeName());
  ASSERT_EQ(resultDataItem.getNodeHost(), fObj.getNodeHost());
  ASSERT_EQ(resultDataItem.getNodePort(), fObj.getNodePort());
  ASSERT_EQ(resultDataItem.getState(), DRCETaskRequest::TaskState::FINISHED);
  ASSERT_EQ(resultDataItem.getSubtasksCount(), 2);

  ASSERT_TRUE(resultDataItem.getSubtaskItem(0).getStderrStream().empty());
  ASSERT_EQ(resultDataItem.getSubtaskItem(0).getErrorCode(), NO_ERROR);
  ASSERT_EQ(resultDataItem.getSubtaskItem(0).getNodeName(), fObj.getNodeName());
  ASSERT_EQ(resultDataItem.getSubtaskItem(0).getNodeHost(), fObj.getNodeHost());
  ASSERT_EQ(resultDataItem.getSubtaskItem(0).getNodePort(), fObj.getNodePort());
  ASSERT_TRUE(resultDataItem.getSubtaskItem(0).getState() == DRCETaskRequest::TaskState::SET_AS_NEW ||
      resultDataItem.getSubtaskItem(0).getState() == DRCETaskRequest::TaskState::IN_PROGRESS ||
      resultDataItem.getSubtaskItem(0).getState() == DRCETaskRequest::TaskState::FINISHED ||
      resultDataItem.getSubtaskItem(0).getState() == DRCETaskRequest::TaskState::QUEUED_TO_RUN);
  ASSERT_EQ(resultDataItem.getSubtaskItem(0).getSubtasksCount(), 2);

  ASSERT_TRUE(resultDataItem.getSubtaskItem(1).getStderrStream().empty());
  ASSERT_EQ(resultDataItem.getSubtaskItem(1).getErrorCode(), NO_ERROR);
  ASSERT_EQ(resultDataItem.getSubtaskItem(1).getNodeName(), fObj.getNodeName());
  ASSERT_EQ(resultDataItem.getSubtaskItem(1).getNodeHost(), fObj.getNodeHost());
  ASSERT_EQ(resultDataItem.getSubtaskItem(1).getNodePort(), fObj.getNodePort());
  ASSERT_TRUE(resultDataItem.getSubtaskItem(1).getState() == DRCETaskRequest::TaskState::SET_AS_NEW ||
      resultDataItem.getSubtaskItem(1).getState() == DRCETaskRequest::TaskState::IN_PROGRESS ||
      resultDataItem.getSubtaskItem(1).getState() == DRCETaskRequest::TaskState::FINISHED ||
      resultDataItem.getSubtaskItem(1).getState() == DRCETaskRequest::TaskState::QUEUED_TO_RUN);
  ASSERT_EQ(resultDataItem.getSubtaskItem(1).getSubtasksCount(), 2);

  for (unsigned int i=0;i<6;++i)
  {
    ASSERT_TRUE(isExistTaskData(startTaskId+i, nodeOptions));
  }
  cleanupTestData(nodeOptions);
  printSuccess("!!! Test Async task and Async subtasks");
}
//-----------------------------------------------------------------------------
void DRCECleanupTest::testAsyncTaskSyncSubtasks(void)
{
  DRCENodeOptions nodeOptions = DRCETaskThreadingModeTest::getNodeOptions();
  cleanupTestData(nodeOptions);
  DRCEFunctionalObject fObj(nodeOptions.getNodeName(), nodeOptions.getHomeDir(), nodeOptions.getTasksDataDir(), nodeOptions.getTasksStatusDir());
  fObj.setNodeHost("NODE_HOST");
  fObj.setNodePort("NODE_PORT");
  fObj.setNodeName("NODE_NAME");

  unsigned int startTaskId = 600;

  std::string sourceJson = getInputJson(startTaskId, SessionOptions::ThreadMode::tmAsync, SessionOptions::ThreadMode::tmSync);

//  std::cout << "sourceJson: " << sourceJson << std::endl;

  std::string resultJson = fObj.Process(sourceJson);

  ASSERT_FALSE(fObj.isError()) << printError(fObj);
  ASSERT_EQ(fObj.getErrorCode(), NO_ERROR);
  ASSERT_TRUE(fObj.getErrorMsg().empty());
  printSuccess("DRCEFunctionalObject Process");

//  std::cout << "result JSON: " << resultJson << std::endl;

  // check right result (unserialize and compare two objects)
  DRCEOutputJsonMessage resultJsonMessage(resultJson);

  ASSERT_FALSE(resultJsonMessage.isError()) << printError(resultJsonMessage);
  ASSERT_EQ(resultJsonMessage.getErrorCode(), NO_ERROR);
  ASSERT_TRUE(resultJsonMessage.getErrorMsg().empty());
  printSuccess("DRCEOutputJsonMessage unserialize");

  ASSERT_EQ(resultJsonMessage.getResultData().getItemsCount(), 1);
  ASSERT_TRUE(resultJsonMessage.getResultData().getDataItem(0).getStderrStream().empty());
  ASSERT_EQ(resultJsonMessage.getResultData().getDataItem(0).getErrorCode(), NO_ERROR);
  ASSERT_EQ(resultJsonMessage.getResultData().getDataItem(0).getNodeName(), fObj.getNodeName());
  ASSERT_EQ(resultJsonMessage.getResultData().getDataItem(0).getNodeHost(), fObj.getNodeHost());
  ASSERT_EQ(resultJsonMessage.getResultData().getDataItem(0).getNodePort(), fObj.getNodePort());
  ASSERT_TRUE(resultJsonMessage.getResultData().getDataItem(0).getState() == DRCETaskRequest::TaskState::SET_AS_NEW ||
      resultJsonMessage.getResultData().getDataItem(0).getState() == DRCETaskRequest::TaskState::IN_PROGRESS ||
      resultJsonMessage.getResultData().getDataItem(0).getState() == DRCETaskRequest::TaskState::FINISHED)
  << "State: " << resultJsonMessage.getResultData().getDataItem(0).getState();
  ASSERT_EQ(resultJsonMessage.getResultData().getDataItem(0).getSubtasksCount(), 2);

  ASSERT_TRUE(resultJsonMessage.getResultData().getDataItem(0).getSubtaskItem(0).getStderrStream().empty());
  ASSERT_EQ(resultJsonMessage.getResultData().getDataItem(0).getSubtaskItem(0).getErrorCode(), NO_ERROR);
  ASSERT_EQ(resultJsonMessage.getResultData().getDataItem(0).getSubtaskItem(0).getNodeName(), fObj.getNodeName());
  ASSERT_EQ(resultJsonMessage.getResultData().getDataItem(0).getSubtaskItem(0).getNodeHost(), fObj.getNodeHost());
  ASSERT_EQ(resultJsonMessage.getResultData().getDataItem(0).getSubtaskItem(0).getNodePort(), fObj.getNodePort());
  ASSERT_TRUE(resultJsonMessage.getResultData().getDataItem(0).getSubtaskItem(0).getState() == DRCETaskRequest::TaskState::SET_AS_NEW ||
      resultJsonMessage.getResultData().getDataItem(0).getSubtaskItem(0).getState() == DRCETaskRequest::TaskState::IN_PROGRESS ||
      resultJsonMessage.getResultData().getDataItem(0).getSubtaskItem(0).getState() == DRCETaskRequest::TaskState::FINISHED ||
      resultJsonMessage.getResultData().getDataItem(0).getSubtaskItem(0).getState() == DRCETaskRequest::TaskState::QUEUED_TO_RUN);
  ASSERT_EQ(resultJsonMessage.getResultData().getDataItem(0).getSubtaskItem(0).getSubtasksCount(), 2);

  ASSERT_TRUE(resultJsonMessage.getResultData().getDataItem(0).getSubtaskItem(1).getStderrStream().empty());
  ASSERT_EQ(resultJsonMessage.getResultData().getDataItem(0).getSubtaskItem(1).getErrorCode(), NO_ERROR);
  ASSERT_EQ(resultJsonMessage.getResultData().getDataItem(0).getSubtaskItem(1).getNodeName(), fObj.getNodeName());
  ASSERT_EQ(resultJsonMessage.getResultData().getDataItem(0).getSubtaskItem(1).getNodeHost(), fObj.getNodeHost());
  ASSERT_EQ(resultJsonMessage.getResultData().getDataItem(0).getSubtaskItem(1).getNodePort(), fObj.getNodePort());
  ASSERT_TRUE(resultJsonMessage.getResultData().getDataItem(0).getSubtaskItem(1).getState() == DRCETaskRequest::TaskState::SET_AS_NEW ||
      resultJsonMessage.getResultData().getDataItem(0).getSubtaskItem(1).getState() == DRCETaskRequest::TaskState::IN_PROGRESS ||
      resultJsonMessage.getResultData().getDataItem(0).getSubtaskItem(1).getState() == DRCETaskRequest::TaskState::FINISHED ||
      resultJsonMessage.getResultData().getDataItem(0).getSubtaskItem(1).getState() == DRCETaskRequest::TaskState::QUEUED_TO_RUN)
  << "State: " << resultJsonMessage.getResultData().getDataItem(0).getSubtaskItem(1).getState();
  ASSERT_EQ(resultJsonMessage.getResultData().getDataItem(0).getSubtaskItem(1).getSubtasksCount(), 2);

  while (fObj.getAsyncTasksCount())
  {
    Poco::Thread::sleep(5000);
    std::cout << "wait end task" << std::endl;
  }
  Poco::Thread::sleep(2000);

  DRCEResultDataItem resultDataItem = DRCECleanupTest::extractResultData(startTaskId, fObj);

  ASSERT_TRUE(resultDataItem.getStderrStream().empty());
  ASSERT_EQ(resultDataItem.getErrorCode(), NO_ERROR);
  ASSERT_EQ(resultDataItem.getNodeName(), fObj.getNodeName());
  ASSERT_EQ(resultDataItem.getNodeHost(), fObj.getNodeHost());
  ASSERT_EQ(resultDataItem.getNodePort(), fObj.getNodePort());
  ASSERT_EQ(resultDataItem.getState(), DRCETaskRequest::TaskState::FINISHED);
  ASSERT_EQ(resultDataItem.getSubtasksCount(), 2);

  ASSERT_TRUE(resultDataItem.getSubtaskItem(0).getStderrStream().empty());
  ASSERT_EQ(resultDataItem.getSubtaskItem(0).getErrorCode(), NO_ERROR);
  ASSERT_EQ(resultDataItem.getSubtaskItem(0).getNodeName(), fObj.getNodeName());
  ASSERT_EQ(resultDataItem.getSubtaskItem(0).getNodeHost(), fObj.getNodeHost());
  ASSERT_EQ(resultDataItem.getSubtaskItem(0).getNodePort(), fObj.getNodePort());
  ASSERT_EQ(resultDataItem.getSubtaskItem(0).getState(), DRCETaskRequest::TaskState::FINISHED);
  ASSERT_EQ(resultDataItem.getSubtaskItem(0).getSubtasksCount(), 2);

  ASSERT_TRUE(resultDataItem.getSubtaskItem(1).getStderrStream().empty());
  ASSERT_EQ(resultDataItem.getSubtaskItem(1).getErrorCode(), NO_ERROR);
  ASSERT_EQ(resultDataItem.getSubtaskItem(1).getNodeName(), fObj.getNodeName());
  ASSERT_EQ(resultDataItem.getSubtaskItem(1).getNodeHost(), fObj.getNodeHost());
  ASSERT_EQ(resultDataItem.getSubtaskItem(1).getNodePort(), fObj.getNodePort());
  ASSERT_EQ(resultDataItem.getSubtaskItem(1).getState(), DRCETaskRequest::TaskState::FINISHED);
  ASSERT_EQ(resultDataItem.getSubtaskItem(1).getSubtasksCount(), 2);

  for (unsigned int i=0;i<6;++i)
  {
    ASSERT_TRUE(isExistTaskData(startTaskId+i, nodeOptions));
  }
  cleanupTestData(nodeOptions);
  printSuccess("!!! Test Async task and Sync subtasks");
}
//-----------------------------------------------------------------------------
void DRCECleanupTest::testSyncTaskAsyncSubtasks(void)
{
  DRCENodeOptions nodeOptions = DRCETaskThreadingModeTest::getNodeOptions();
  cleanupTestData(nodeOptions);
  DRCEFunctionalObject fObj(nodeOptions.getNodeName(), nodeOptions.getHomeDir(), nodeOptions.getTasksDataDir(), nodeOptions.getTasksStatusDir());
  fObj.setNodeHost("NODE_HOST");
  fObj.setNodePort("NODE_PORT");
  fObj.setNodeName("NODE_NAME");

  unsigned int startTaskId = 600;

  std::string sourceJson = getInputJson(startTaskId, SessionOptions::ThreadMode::tmSync, SessionOptions::ThreadMode::tmAsync);

//  std::cout << "sourceJson: " << sourceJson << std::endl;

  std::string resultJson = fObj.Process(sourceJson);

  ASSERT_FALSE(fObj.isError()) << printError(fObj);
  ASSERT_EQ(fObj.getErrorCode(), NO_ERROR);
  ASSERT_TRUE(fObj.getErrorMsg().empty());
  printSuccess("DRCEFunctionalObject Process");

//  std::cout << "result JSON: " << resultJson << std::endl;

  // check right result (unserialize and compare two objects)
  DRCEOutputJsonMessage resultJsonMessage(resultJson);

  ASSERT_FALSE(resultJsonMessage.isError()) << printError(resultJsonMessage);
  ASSERT_EQ(resultJsonMessage.getErrorCode(), NO_ERROR);
  ASSERT_TRUE(resultJsonMessage.getErrorMsg().empty());
  printSuccess("DRCEOutputJsonMessage unserialize");

  while (fObj.getAsyncTasksCount())
  {
    Poco::Thread::sleep(5000);
    std::cout << "wait end task" << std::endl;
  }
  Poco::Thread::sleep(2000);

  ASSERT_EQ(resultJsonMessage.getResultData().getItemsCount(), 1);
  ASSERT_TRUE(resultJsonMessage.getResultData().getDataItem(0).getStderrStream().empty());
  ASSERT_EQ(resultJsonMessage.getResultData().getDataItem(0).getErrorCode(), NO_ERROR);
  ASSERT_EQ(resultJsonMessage.getResultData().getDataItem(0).getNodeName(), fObj.getNodeName());
  ASSERT_EQ(resultJsonMessage.getResultData().getDataItem(0).getNodeHost(), fObj.getNodeHost());
  ASSERT_EQ(resultJsonMessage.getResultData().getDataItem(0).getNodePort(), fObj.getNodePort());
  ASSERT_EQ(resultJsonMessage.getResultData().getDataItem(0).getState(), DRCETaskRequest::TaskState::FINISHED);
  ASSERT_EQ(resultJsonMessage.getResultData().getDataItem(0).getSubtasksCount(), 2);

  ASSERT_TRUE(resultJsonMessage.getResultData().getDataItem(0).getSubtaskItem(0).getStderrStream().empty());
  ASSERT_EQ(resultJsonMessage.getResultData().getDataItem(0).getSubtaskItem(0).getErrorCode(), NO_ERROR);
  ASSERT_EQ(resultJsonMessage.getResultData().getDataItem(0).getSubtaskItem(0).getNodeName(), fObj.getNodeName());
  ASSERT_EQ(resultJsonMessage.getResultData().getDataItem(0).getSubtaskItem(0).getNodeHost(), fObj.getNodeHost());
  ASSERT_EQ(resultJsonMessage.getResultData().getDataItem(0).getSubtaskItem(0).getNodePort(), fObj.getNodePort());
  ASSERT_TRUE(resultJsonMessage.getResultData().getDataItem(0).getSubtaskItem(0).getState() == DRCETaskRequest::TaskState::SET_AS_NEW ||
      resultJsonMessage.getResultData().getDataItem(0).getSubtaskItem(0).getState() == DRCETaskRequest::TaskState::IN_PROGRESS ||
      resultJsonMessage.getResultData().getDataItem(0).getSubtaskItem(0).getState() == DRCETaskRequest::TaskState::FINISHED);
  ASSERT_EQ(resultJsonMessage.getResultData().getDataItem(0).getSubtaskItem(0).getSubtasksCount(), 2);

  ASSERT_TRUE(resultJsonMessage.getResultData().getDataItem(0).getSubtaskItem(1).getStderrStream().empty());
  ASSERT_EQ(resultJsonMessage.getResultData().getDataItem(0).getSubtaskItem(1).getErrorCode(), NO_ERROR);
  ASSERT_EQ(resultJsonMessage.getResultData().getDataItem(0).getSubtaskItem(1).getNodeName(), fObj.getNodeName());
  ASSERT_EQ(resultJsonMessage.getResultData().getDataItem(0).getSubtaskItem(1).getNodeHost(), fObj.getNodeHost());
  ASSERT_EQ(resultJsonMessage.getResultData().getDataItem(0).getSubtaskItem(1).getNodePort(), fObj.getNodePort());
  ASSERT_TRUE(resultJsonMessage.getResultData().getDataItem(0).getSubtaskItem(1).getState() == DRCETaskRequest::TaskState::SET_AS_NEW ||
      resultJsonMessage.getResultData().getDataItem(0).getSubtaskItem(1).getState() == DRCETaskRequest::TaskState::IN_PROGRESS ||
      resultJsonMessage.getResultData().getDataItem(0).getSubtaskItem(1).getState() == DRCETaskRequest::TaskState::FINISHED)
  << "State: " << resultJsonMessage.getResultData().getDataItem(0).getSubtaskItem(1).getState();
  ASSERT_EQ(resultJsonMessage.getResultData().getDataItem(0).getSubtaskItem(1).getSubtasksCount(), 2);

  for (unsigned int i=0;i<6;++i)
  {
    ASSERT_TRUE(isExistTaskData(startTaskId+i, nodeOptions));
  }
  cleanupTestData(nodeOptions);
  printSuccess("!!! Test Sync task and Async subtasks");
}
//-----------------------------------------------------------------------------
void DRCECleanupTest::testSyncTaskSyncSubtasks(void)
{
  DRCENodeOptions nodeOptions = DRCETaskThreadingModeTest::getNodeOptions();
  cleanupTestData(nodeOptions);
  DRCEFunctionalObject fObj(nodeOptions.getNodeName(), nodeOptions.getHomeDir(), nodeOptions.getTasksDataDir(), nodeOptions.getTasksStatusDir());
  fObj.setNodeHost("NODE_HOST");
  fObj.setNodePort("NODE_PORT");
  fObj.setNodeName("NODE_NAME");

  unsigned int startTaskId = 600;

  std::string sourceJson = getInputJson(startTaskId, SessionOptions::ThreadMode::tmSync, SessionOptions::ThreadMode::tmSync);

//  std::cout << "sourceJson: " << sourceJson << std::endl;

  std::string resultJson = fObj.Process(sourceJson);

  ASSERT_FALSE(fObj.isError()) << printError(fObj);
  ASSERT_EQ(fObj.getErrorCode(), NO_ERROR);
  ASSERT_TRUE(fObj.getErrorMsg().empty());
  printSuccess("DRCEFunctionalObject Process");

//  std::cout << "result JSON: " << resultJson << std::endl;

  // check right result (unserialize and compare two objects)
  DRCEOutputJsonMessage resultJsonMessage(resultJson);

  ASSERT_FALSE(resultJsonMessage.isError()) << printError(resultJsonMessage);
  ASSERT_EQ(resultJsonMessage.getErrorCode(), NO_ERROR);
  ASSERT_TRUE(resultJsonMessage.getErrorMsg().empty());
  printSuccess("DRCEOutputJsonMessage unserialize");

  while (fObj.getAsyncTasksCount() || fObj.hasTask(startTaskId))
  {
    Poco::Thread::sleep(5000);
    std::cout << "wait end task" << std::endl;
  }
  Poco::Thread::sleep(5000);

  ASSERT_EQ(resultJsonMessage.getResultData().getItemsCount(), 1);
  ASSERT_TRUE(resultJsonMessage.getResultData().getDataItem(0).getStderrStream().empty());
  ASSERT_EQ(resultJsonMessage.getResultData().getDataItem(0).getErrorCode(), NO_ERROR);
  ASSERT_EQ(resultJsonMessage.getResultData().getDataItem(0).getNodeName(), fObj.getNodeName());
  ASSERT_EQ(resultJsonMessage.getResultData().getDataItem(0).getNodeHost(), fObj.getNodeHost());
  ASSERT_EQ(resultJsonMessage.getResultData().getDataItem(0).getNodePort(), fObj.getNodePort());
  ASSERT_EQ(resultJsonMessage.getResultData().getDataItem(0).getState(), DRCETaskRequest::TaskState::FINISHED);
  ASSERT_EQ(resultJsonMessage.getResultData().getDataItem(0).getSubtasksCount(), 2);

  ASSERT_TRUE(resultJsonMessage.getResultData().getDataItem(0).getSubtaskItem(0).getStderrStream().empty());
  ASSERT_EQ(resultJsonMessage.getResultData().getDataItem(0).getSubtaskItem(0).getErrorCode(), NO_ERROR);
  ASSERT_EQ(resultJsonMessage.getResultData().getDataItem(0).getSubtaskItem(0).getNodeName(), fObj.getNodeName());
  ASSERT_EQ(resultJsonMessage.getResultData().getDataItem(0).getSubtaskItem(0).getNodeHost(), fObj.getNodeHost());
  ASSERT_EQ(resultJsonMessage.getResultData().getDataItem(0).getSubtaskItem(0).getNodePort(), fObj.getNodePort());
  ASSERT_EQ(resultJsonMessage.getResultData().getDataItem(0).getSubtaskItem(0).getState(), DRCETaskRequest::TaskState::FINISHED);
  ASSERT_EQ(resultJsonMessage.getResultData().getDataItem(0).getSubtaskItem(0).getSubtasksCount(), 2);

  ASSERT_TRUE(resultJsonMessage.getResultData().getDataItem(0).getSubtaskItem(1).getStderrStream().empty());
  ASSERT_EQ(resultJsonMessage.getResultData().getDataItem(0).getSubtaskItem(1).getErrorCode(), NO_ERROR);
  ASSERT_EQ(resultJsonMessage.getResultData().getDataItem(0).getSubtaskItem(1).getNodeName(), fObj.getNodeName());
  ASSERT_EQ(resultJsonMessage.getResultData().getDataItem(0).getSubtaskItem(1).getNodeHost(), fObj.getNodeHost());
  ASSERT_EQ(resultJsonMessage.getResultData().getDataItem(0).getSubtaskItem(1).getNodePort(), fObj.getNodePort());
  ASSERT_EQ(resultJsonMessage.getResultData().getDataItem(0).getSubtaskItem(1).getState(), DRCETaskRequest::TaskState::FINISHED);
  ASSERT_EQ(resultJsonMessage.getResultData().getDataItem(0).getSubtaskItem(1).getSubtasksCount(), 2);

  for (unsigned int i=0;i<6;++i)
  {
    ASSERT_TRUE(isExistTaskData((startTaskId+i), nodeOptions)) << " Task ID: " << (startTaskId+i);
  }
  cleanupTestData(nodeOptions);
  printSuccess("!!! Test Sync task and Sync subtasks");
}
//-----------------------------------------------------------------------------
void DRCECleanupTest::testAsyncTaskMixSubtasks(void)
{
  DRCENodeOptions nodeOptions = DRCETaskThreadingModeTest::getNodeOptions();
  cleanupTestData(nodeOptions);
  DRCEFunctionalObject fObj(nodeOptions.getNodeName(), nodeOptions.getHomeDir(), nodeOptions.getTasksDataDir(), nodeOptions.getTasksStatusDir());
  fObj.setNodeHost("NODE_HOST");
  fObj.setNodePort("NODE_PORT");
  fObj.setNodeName("NODE_NAME");

  unsigned int startTaskId = 600;

  std::string sourceJson = getInputJson(startTaskId, SessionOptions::ThreadMode::tmAsync);

//  std::cout << "sourceJson: " << sourceJson << std::endl;

  std::string resultJson = fObj.Process(sourceJson);

  ASSERT_FALSE(fObj.isError()) << printError(fObj);
  ASSERT_EQ(fObj.getErrorCode(), NO_ERROR);
  ASSERT_TRUE(fObj.getErrorMsg().empty());
  printSuccess("DRCEFunctionalObject Process");

//  std::cout << "result JSON: " << resultJson << std::endl;

  // check right result (unserialize and compare two objects)
  DRCEOutputJsonMessage resultJsonMessage(resultJson);

  ASSERT_FALSE(resultJsonMessage.isError()) << printError(resultJsonMessage);
  ASSERT_EQ(resultJsonMessage.getErrorCode(), NO_ERROR);
  ASSERT_TRUE(resultJsonMessage.getErrorMsg().empty());
  printSuccess("DRCEOutputJsonMessage unserialize");

  ASSERT_EQ(resultJsonMessage.getResultData().getItemsCount(), 1);
  ASSERT_TRUE(resultJsonMessage.getResultData().getDataItem(0).getStderrStream().empty());
  ASSERT_EQ(resultJsonMessage.getResultData().getDataItem(0).getErrorCode(), NO_ERROR);
  ASSERT_EQ(resultJsonMessage.getResultData().getDataItem(0).getNodeName(), fObj.getNodeName());
  ASSERT_EQ(resultJsonMessage.getResultData().getDataItem(0).getNodeHost(), fObj.getNodeHost());
  ASSERT_EQ(resultJsonMessage.getResultData().getDataItem(0).getNodePort(), fObj.getNodePort());
  ASSERT_TRUE(resultJsonMessage.getResultData().getDataItem(0).getState() == DRCETaskRequest::TaskState::SET_AS_NEW ||
      resultJsonMessage.getResultData().getDataItem(0).getState() == DRCETaskRequest::TaskState::IN_PROGRESS ||
      resultJsonMessage.getResultData().getDataItem(0).getState() == DRCETaskRequest::TaskState::FINISHED);
  ASSERT_EQ(resultJsonMessage.getResultData().getDataItem(0).getSubtasksCount(), 2);

  ASSERT_TRUE(resultJsonMessage.getResultData().getDataItem(0).getSubtaskItem(0).getStderrStream().empty());
  ASSERT_EQ(resultJsonMessage.getResultData().getDataItem(0).getSubtaskItem(0).getErrorCode(), NO_ERROR);
  ASSERT_EQ(resultJsonMessage.getResultData().getDataItem(0).getSubtaskItem(0).getNodeName(), fObj.getNodeName());
  ASSERT_EQ(resultJsonMessage.getResultData().getDataItem(0).getSubtaskItem(0).getNodeHost(), fObj.getNodeHost());
  ASSERT_EQ(resultJsonMessage.getResultData().getDataItem(0).getSubtaskItem(0).getNodePort(), fObj.getNodePort());
  ASSERT_TRUE(resultJsonMessage.getResultData().getDataItem(0).getSubtaskItem(0).getState() == DRCETaskRequest::TaskState::SET_AS_NEW ||
      resultJsonMessage.getResultData().getDataItem(0).getSubtaskItem(0).getState() == DRCETaskRequest::TaskState::IN_PROGRESS ||
      resultJsonMessage.getResultData().getDataItem(0).getSubtaskItem(0).getState() == DRCETaskRequest::TaskState::FINISHED ||
      resultJsonMessage.getResultData().getDataItem(0).getSubtaskItem(0).getState() == DRCETaskRequest::TaskState::QUEUED_TO_RUN);
  ASSERT_EQ(resultJsonMessage.getResultData().getDataItem(0).getSubtaskItem(0).getSubtasksCount(), 2);

  ASSERT_TRUE(resultJsonMessage.getResultData().getDataItem(0).getSubtaskItem(1).getStderrStream().empty());
  ASSERT_EQ(resultJsonMessage.getResultData().getDataItem(0).getSubtaskItem(1).getErrorCode(), NO_ERROR);
  ASSERT_EQ(resultJsonMessage.getResultData().getDataItem(0).getSubtaskItem(1).getNodeName(), fObj.getNodeName());
  ASSERT_EQ(resultJsonMessage.getResultData().getDataItem(0).getSubtaskItem(1).getNodeHost(), fObj.getNodeHost());
  ASSERT_EQ(resultJsonMessage.getResultData().getDataItem(0).getSubtaskItem(1).getNodePort(), fObj.getNodePort());
  ASSERT_TRUE(resultJsonMessage.getResultData().getDataItem(0).getSubtaskItem(1).getState() == DRCETaskRequest::TaskState::SET_AS_NEW ||
      resultJsonMessage.getResultData().getDataItem(0).getSubtaskItem(1).getState() == DRCETaskRequest::TaskState::IN_PROGRESS ||
      resultJsonMessage.getResultData().getDataItem(0).getSubtaskItem(1).getState() == DRCETaskRequest::TaskState::FINISHED ||
      resultJsonMessage.getResultData().getDataItem(0).getSubtaskItem(1).getState() == DRCETaskRequest::TaskState::QUEUED_TO_RUN);
  ASSERT_EQ(resultJsonMessage.getResultData().getDataItem(0).getSubtaskItem(1).getSubtasksCount(), 2);

  while (fObj.getAsyncTasksCount())
  {
    Poco::Thread::sleep(5000);
    std::cout << "wait end task" << std::endl;
  }
  Poco::Thread::sleep(2000);

  DRCEResultDataItem resultDataItem = DRCECleanupTest::extractResultData(startTaskId, fObj);

  ASSERT_TRUE(resultDataItem.getStderrStream().empty());
  ASSERT_EQ(resultDataItem.getErrorCode(), NO_ERROR);
  ASSERT_EQ(resultDataItem.getNodeName(), fObj.getNodeName());
  ASSERT_EQ(resultDataItem.getNodeHost(), fObj.getNodeHost());
  ASSERT_EQ(resultDataItem.getNodePort(), fObj.getNodePort());
  ASSERT_EQ(resultDataItem.getState(), DRCETaskRequest::TaskState::FINISHED);
  ASSERT_EQ(resultDataItem.getSubtasksCount(), 2);

  ASSERT_TRUE(resultDataItem.getSubtaskItem(0).getStderrStream().empty());
  ASSERT_EQ(resultDataItem.getSubtaskItem(0).getErrorCode(), NO_ERROR);
  ASSERT_EQ(resultDataItem.getSubtaskItem(0).getNodeName(), fObj.getNodeName());
  ASSERT_EQ(resultDataItem.getSubtaskItem(0).getNodeHost(), fObj.getNodeHost());
  ASSERT_EQ(resultDataItem.getSubtaskItem(0).getNodePort(), fObj.getNodePort());
  ASSERT_EQ(resultDataItem.getSubtaskItem(0).getState(), DRCETaskRequest::TaskState::FINISHED);
  ASSERT_EQ(resultDataItem.getSubtaskItem(0).getSubtasksCount(), 2);

  ASSERT_TRUE(resultDataItem.getSubtaskItem(1).getStderrStream().empty());
  ASSERT_EQ(resultDataItem.getSubtaskItem(1).getErrorCode(), NO_ERROR);
  ASSERT_EQ(resultDataItem.getSubtaskItem(1).getNodeName(), fObj.getNodeName());
  ASSERT_EQ(resultDataItem.getSubtaskItem(1).getNodeHost(), fObj.getNodeHost());
  ASSERT_EQ(resultDataItem.getSubtaskItem(1).getNodePort(), fObj.getNodePort());
  ASSERT_EQ(resultDataItem.getSubtaskItem(1).getState(), DRCETaskRequest::TaskState::FINISHED);
  ASSERT_EQ(resultDataItem.getSubtaskItem(1).getSubtasksCount(), 2);

  for (unsigned int i=0;i<14;++i)
  {
    ASSERT_TRUE(isExistTaskData(startTaskId+i, nodeOptions));
  }
  cleanupTestData(nodeOptions);
  printSuccess("!!! Test Async task and Mix subtasks");
}
//-----------------------------------------------------------------------------
void DRCECleanupTest::testSyncTaskMixSubtasks(void)
{
  DRCENodeOptions nodeOptions = DRCETaskThreadingModeTest::getNodeOptions();
  cleanupTestData(nodeOptions);
  DRCEFunctionalObject fObj(nodeOptions.getNodeName(), nodeOptions.getHomeDir(), nodeOptions.getTasksDataDir(), nodeOptions.getTasksStatusDir());
  fObj.setNodeHost("NODE_HOST");
  fObj.setNodePort("NODE_PORT");
  fObj.setNodeName("NODE_NAME");

  unsigned int startTaskId = 600;

  std::string sourceJson = getInputJson(startTaskId, SessionOptions::ThreadMode::tmSync);

//  std::cout << "sourceJson: " << sourceJson << std::endl;

  std::string resultJson = fObj.Process(sourceJson);

  ASSERT_FALSE(fObj.isError()) << printError(fObj);
  ASSERT_EQ(fObj.getErrorCode(), NO_ERROR);
  ASSERT_TRUE(fObj.getErrorMsg().empty());
  printSuccess("DRCEFunctionalObject Process");

//  std::cout << "result JSON: " << resultJson << std::endl;

  // check right result (unserialize and compare two objects)
  DRCEOutputJsonMessage resultJsonMessage(resultJson);

  ASSERT_FALSE(resultJsonMessage.isError()) << printError(resultJsonMessage);
  ASSERT_EQ(resultJsonMessage.getErrorCode(), NO_ERROR);
  ASSERT_TRUE(resultJsonMessage.getErrorMsg().empty());
  printSuccess("DRCEOutputJsonMessage unserialize");

  while (fObj.getAsyncTasksCount())
  {
    Poco::Thread::sleep(5000);
    std::cout << "wait end task" << std::endl;
  }
  Poco::Thread::sleep(5000);

  ASSERT_EQ(resultJsonMessage.getResultData().getItemsCount(), 1);
  ASSERT_TRUE(resultJsonMessage.getResultData().getDataItem(0).getStderrStream().empty());
  ASSERT_EQ(resultJsonMessage.getResultData().getDataItem(0).getErrorCode(), NO_ERROR);
  ASSERT_EQ(resultJsonMessage.getResultData().getDataItem(0).getNodeName(), fObj.getNodeName());
  ASSERT_EQ(resultJsonMessage.getResultData().getDataItem(0).getNodeHost(), fObj.getNodeHost());
  ASSERT_EQ(resultJsonMessage.getResultData().getDataItem(0).getNodePort(), fObj.getNodePort());
  ASSERT_EQ(resultJsonMessage.getResultData().getDataItem(0).getState(), DRCETaskRequest::TaskState::FINISHED);
  ASSERT_EQ(resultJsonMessage.getResultData().getDataItem(0).getSubtasksCount(), 2);

  ASSERT_TRUE(resultJsonMessage.getResultData().getDataItem(0).getSubtaskItem(0).getStderrStream().empty());
  ASSERT_EQ(resultJsonMessage.getResultData().getDataItem(0).getSubtaskItem(0).getErrorCode(), NO_ERROR);
  ASSERT_EQ(resultJsonMessage.getResultData().getDataItem(0).getSubtaskItem(0).getNodeName(), fObj.getNodeName());
  ASSERT_EQ(resultJsonMessage.getResultData().getDataItem(0).getSubtaskItem(0).getNodeHost(), fObj.getNodeHost());
  ASSERT_EQ(resultJsonMessage.getResultData().getDataItem(0).getSubtaskItem(0).getNodePort(), fObj.getNodePort());
  ASSERT_TRUE(resultJsonMessage.getResultData().getDataItem(0).getSubtaskItem(0).getState() == DRCETaskRequest::TaskState::SET_AS_NEW ||
      resultJsonMessage.getResultData().getDataItem(0).getSubtaskItem(0).getState() == DRCETaskRequest::TaskState::IN_PROGRESS ||
      resultJsonMessage.getResultData().getDataItem(0).getSubtaskItem(0).getState() == DRCETaskRequest::TaskState::FINISHED);
  ASSERT_EQ(resultJsonMessage.getResultData().getDataItem(0).getSubtaskItem(0).getSubtasksCount(), 2);

  ASSERT_TRUE(resultJsonMessage.getResultData().getDataItem(0).getSubtaskItem(1).getStderrStream().empty());
  ASSERT_EQ(resultJsonMessage.getResultData().getDataItem(0).getSubtaskItem(1).getErrorCode(), NO_ERROR);
  ASSERT_EQ(resultJsonMessage.getResultData().getDataItem(0).getSubtaskItem(1).getNodeName(), fObj.getNodeName());
  ASSERT_EQ(resultJsonMessage.getResultData().getDataItem(0).getSubtaskItem(1).getNodeHost(), fObj.getNodeHost());
  ASSERT_EQ(resultJsonMessage.getResultData().getDataItem(0).getSubtaskItem(1).getNodePort(), fObj.getNodePort());
  ASSERT_TRUE(resultJsonMessage.getResultData().getDataItem(0).getSubtaskItem(1).getState() == DRCETaskRequest::TaskState::SET_AS_NEW ||
      resultJsonMessage.getResultData().getDataItem(0).getSubtaskItem(1).getState() == DRCETaskRequest::TaskState::IN_PROGRESS ||
      resultJsonMessage.getResultData().getDataItem(0).getSubtaskItem(1).getState() == DRCETaskRequest::TaskState::FINISHED);
  ASSERT_EQ(resultJsonMessage.getResultData().getDataItem(0).getSubtaskItem(1).getSubtasksCount(), 2);

  for (unsigned int i=0;i<14;++i)
  {
    ASSERT_TRUE(isExistTaskData(startTaskId+i, nodeOptions));
  }
  cleanupTestData(nodeOptions);
  printSuccess("!!! Test Sync task and Mix subtasks");
}
//-----------------------------------------------------------------------------
void DRCECleanupTest::testAsyncTaskMixSubtasksCleanup(void)
{
  DRCENodeOptions nodeOptions = DRCETaskThreadingModeTest::getNodeOptions();
  cleanupTestData(nodeOptions);
  DRCEFunctionalObject fObj(nodeOptions.getNodeName(), nodeOptions.getHomeDir(), nodeOptions.getTasksDataDir(), nodeOptions.getTasksStatusDir());
  fObj.setNodeHost("NODE_HOST");
  fObj.setNodePort("NODE_PORT");
  fObj.setNodeName("NODE_NAME");

  unsigned int startTaskId = 600;

  std::string sourceJson = getInputJsonCleanup(startTaskId, SessionOptions::ThreadMode::tmAsync, "");

//  std::cout << "sourceJson: " << sourceJson << std::endl;

  std::string resultJson = fObj.Process(sourceJson);

  ASSERT_FALSE(fObj.isError()) << printError(fObj);
  ASSERT_EQ(fObj.getErrorCode(), NO_ERROR);
  ASSERT_TRUE(fObj.getErrorMsg().empty());
  printSuccess("DRCEFunctionalObject Process");

  std::cout << "result JSON: " << resultJson << std::endl;

  // check right result (unserialize and compare two objects)
  DRCEOutputJsonMessage resultJsonMessage(resultJson);

  ASSERT_FALSE(resultJsonMessage.isError()) << printError(resultJsonMessage);
  ASSERT_EQ(resultJsonMessage.getErrorCode(), NO_ERROR);
  ASSERT_TRUE(resultJsonMessage.getErrorMsg().empty());
  printSuccess("DRCEOutputJsonMessage unserialize");

  while (fObj.getAsyncTasksCount() || fObj.hasTask(startTaskId))
  {
    Poco::Thread::sleep(5000);
    std::cout << "wait end task" << std::endl;
  }
  Poco::Thread::sleep(2000);

  DRCEResultDataItem resultDataItem = DRCECleanupTest::extractResultData(startTaskId, fObj);

  ASSERT_TRUE(resultDataItem.getStderrStream().empty());
  ASSERT_EQ(resultDataItem.getErrorCode(), NO_ERROR);
  ASSERT_EQ(resultDataItem.getNodeName(), fObj.getNodeName());
  ASSERT_EQ(resultDataItem.getNodeHost(), fObj.getNodeHost());
  ASSERT_EQ(resultDataItem.getNodePort(), fObj.getNodePort());
  ASSERT_EQ(resultDataItem.getState(), DRCETaskRequest::TaskState::FINISHED);
  ASSERT_EQ(resultDataItem.getSubtasksCount(), 2);

  ASSERT_TRUE(resultDataItem.getSubtaskItem(0).getStderrStream().empty());
  ASSERT_EQ(resultDataItem.getSubtaskItem(0).getErrorCode(), NO_ERROR);
  ASSERT_EQ(resultDataItem.getSubtaskItem(0).getNodeName(), fObj.getNodeName());
  ASSERT_EQ(resultDataItem.getSubtaskItem(0).getNodeHost(), fObj.getNodeHost());
  ASSERT_EQ(resultDataItem.getSubtaskItem(0).getNodePort(), fObj.getNodePort());
  ASSERT_EQ(resultDataItem.getSubtaskItem(0).getState(), DRCETaskRequest::TaskState::FINISHED);
  ASSERT_EQ(resultDataItem.getSubtaskItem(0).getSubtasksCount(), 2);

  ASSERT_TRUE(resultDataItem.getSubtaskItem(1).getStderrStream().empty());
  ASSERT_EQ(resultDataItem.getSubtaskItem(1).getErrorCode(), NO_ERROR);
  ASSERT_EQ(resultDataItem.getSubtaskItem(1).getNodeName(), fObj.getNodeName());
  ASSERT_EQ(resultDataItem.getSubtaskItem(1).getNodeHost(), fObj.getNodeHost());
  ASSERT_EQ(resultDataItem.getSubtaskItem(1).getNodePort(), fObj.getNodePort());
  ASSERT_EQ(resultDataItem.getSubtaskItem(1).getState(), DRCETaskRequest::TaskState::FINISHED);
  ASSERT_EQ(resultDataItem.getSubtaskItem(1).getSubtasksCount(), 2);

  ASSERT_TRUE(isExistTaskData(startTaskId, nodeOptions));
  ASSERT_TRUE(isExistTaskData(startTaskId+2, nodeOptions));
  ASSERT_TRUE(isExistTaskData(startTaskId+6, nodeOptions));
  ASSERT_TRUE(isExistTaskData(startTaskId+14, nodeOptions));

  cleanupTestData(nodeOptions);
  printSuccess("!!! Test Async task and Mix subtasks Cleanup");
}
//-----------------------------------------------------------------------------
void DRCECleanupTest::testSyncTaskMixSubtasksCleanup(void)
{
  DRCENodeOptions nodeOptions = DRCETaskThreadingModeTest::getNodeOptions();
  cleanupTestData(nodeOptions);
  DRCEFunctionalObject fObj(nodeOptions.getNodeName(), nodeOptions.getHomeDir(), nodeOptions.getTasksDataDir(), nodeOptions.getTasksStatusDir());
  fObj.setNodeHost("NODE_HOST");
  fObj.setNodePort("NODE_PORT");
  fObj.setNodeName("NODE_NAME");

  unsigned int startTaskId = 600;

  std::string sourceJson = getInputJsonCleanup(startTaskId, SessionOptions::ThreadMode::tmSync, "");

//  std::cout << "sourceJson: " << sourceJson << std::endl;

  std::string resultJson = fObj.Process(sourceJson);

  ASSERT_FALSE(fObj.isError()) << printError(fObj);
  ASSERT_EQ(fObj.getErrorCode(), NO_ERROR);
  ASSERT_TRUE(fObj.getErrorMsg().empty());
  printSuccess("DRCEFunctionalObject Process");

//  std::cout << "result JSON: " << resultJson << std::endl;

  // check right result (unserialize and compare two objects)
  DRCEOutputJsonMessage resultJsonMessage(resultJson);

  ASSERT_FALSE(resultJsonMessage.isError()) << printError(resultJsonMessage);
  ASSERT_EQ(resultJsonMessage.getErrorCode(), NO_ERROR);
  ASSERT_TRUE(resultJsonMessage.getErrorMsg().empty());
  printSuccess("DRCEOutputJsonMessage unserialize");

  while (fObj.getAsyncTasksCount())
  {
    Poco::Thread::sleep(5000);
    std::cout << "wait end task" << std::endl;
  }
  Poco::Thread::sleep(2000);

  DRCEResultDataItem resultDataItem = DRCECleanupTest::extractResultData(startTaskId, fObj);

  ASSERT_TRUE(resultDataItem.getStderrStream().empty());
  ASSERT_EQ(resultDataItem.getErrorCode(), NO_ERROR);
  ASSERT_EQ(resultDataItem.getNodeName(), fObj.getNodeName());
  ASSERT_EQ(resultDataItem.getNodeHost(), fObj.getNodeHost());
  ASSERT_EQ(resultDataItem.getNodePort(), fObj.getNodePort());
  ASSERT_EQ(resultDataItem.getState(), DRCETaskRequest::TaskState::FINISHED);
  ASSERT_EQ(resultDataItem.getSubtasksCount(), 2);

  ASSERT_TRUE(resultDataItem.getSubtaskItem(0).getStderrStream().empty());
  ASSERT_EQ(resultDataItem.getSubtaskItem(0).getErrorCode(), NO_ERROR);
  ASSERT_EQ(resultDataItem.getSubtaskItem(0).getNodeName(), fObj.getNodeName());
  ASSERT_EQ(resultDataItem.getSubtaskItem(0).getNodeHost(), fObj.getNodeHost());
  ASSERT_EQ(resultDataItem.getSubtaskItem(0).getNodePort(), fObj.getNodePort());
  ASSERT_EQ(resultDataItem.getSubtaskItem(0).getState(), DRCETaskRequest::TaskState::FINISHED);
  ASSERT_EQ(resultDataItem.getSubtaskItem(0).getSubtasksCount(), 2);

  ASSERT_TRUE(resultDataItem.getSubtaskItem(1).getStderrStream().empty());
  ASSERT_EQ(resultDataItem.getSubtaskItem(1).getErrorCode(), NO_ERROR);
  ASSERT_EQ(resultDataItem.getSubtaskItem(1).getNodeName(), fObj.getNodeName());
  ASSERT_EQ(resultDataItem.getSubtaskItem(1).getNodeHost(), fObj.getNodeHost());
  ASSERT_EQ(resultDataItem.getSubtaskItem(1).getNodePort(), fObj.getNodePort());
  ASSERT_EQ(resultDataItem.getSubtaskItem(1).getState(), DRCETaskRequest::TaskState::FINISHED);
  ASSERT_EQ(resultDataItem.getSubtaskItem(1).getSubtasksCount(), 2);

  ASSERT_TRUE(isExistTaskData(startTaskId, nodeOptions));
  ASSERT_TRUE(isExistTaskData(startTaskId+2, nodeOptions));
  ASSERT_TRUE(isExistTaskData(startTaskId+6, nodeOptions));
  ASSERT_TRUE(isExistTaskData(startTaskId+14, nodeOptions));

  cleanupTestData(nodeOptions);
  printSuccess("!!! Test Sync task and Mix subtasks Cleanup");
}
//-----------------------------------------------------------------------------
void DRCECleanupTest::testRequestGetDataCleanup(void)
{
  DRCENodeOptions nodeOptions = DRCETaskThreadingModeTest::getNodeOptions();
  cleanupTestData(nodeOptions);
  DRCEFunctionalObject fObj(nodeOptions.getNodeName(), nodeOptions.getHomeDir(), nodeOptions.getTasksDataDir(), nodeOptions.getTasksStatusDir());
  fObj.setNodeHost("NODE_HOST");
  fObj.setNodePort("NODE_PORT");
  fObj.setNodeName("NODE_NAME");

  unsigned int startTaskId = 600;

  std::string sourceJson = getInputJsonCleanup(startTaskId, SessionOptions::ThreadMode::tmAsync, "");

//  std::cout << "sourceJson: " << sourceJson << std::endl;

  std::string resultJson = fObj.Process(sourceJson);

  ASSERT_FALSE(fObj.isError()) << printError(fObj);
  ASSERT_EQ(fObj.getErrorCode(), NO_ERROR);
  ASSERT_TRUE(fObj.getErrorMsg().empty());
  printSuccess("DRCEFunctionalObject Process");

//  std::cout << "result JSON: " << resultJson << std::endl;

  // check right result (unserialize and compare two objects)
  DRCEOutputJsonMessage resultJsonMessage(resultJson);

  ASSERT_FALSE(resultJsonMessage.isError()) << printError(resultJsonMessage);
  ASSERT_EQ(resultJsonMessage.getErrorCode(), NO_ERROR);
  ASSERT_TRUE(resultJsonMessage.getErrorMsg().empty());
  printSuccess("DRCEOutputJsonMessage unserialize");

  while (fObj.getAsyncTasksCount() || fObj.hasTask(startTaskId))
  {
    Poco::Thread::sleep(5000);
    std::cout << "wait end task" << std::endl;
  }
  Poco::Thread::sleep(5000);

  // make get data request
  std::string json;
  DRCETaskRequestGetData taskRequestGetData;
  taskRequestGetData.setFetchType(DRCETaskRequestGetData::FetchType::ftDeleteDataAfterFetch);
  taskRequestGetData.serialize(json);
  ASSERT_FALSE(taskRequestGetData.isError());
  printSuccess("DRCETaskRequestGetData serialize");

//  std::cout << "json: " << json << std::endl;

  // set request for execute task
  DRCEInputJsonMessage inputJsonMessage;
  inputJsonMessage.setRequestType(DRCETaskRequest::RequestType::rtGetTaskData);
  inputJsonMessage.setRequestData(json);
  inputJsonMessage.setRequestId(startTaskId);

  std::stringstream sourceJsonStream;
  sourceJsonStream << inputJsonMessage;

//  std::cout << "source JSON: " << sourceJson.str() << std::endl;

  ASSERT_FALSE(inputJsonMessage.isError()) << printError(inputJsonMessage);
  ASSERT_EQ(inputJsonMessage.getErrorCode(), NO_ERROR);
  ASSERT_TRUE(inputJsonMessage.getErrorMsg().empty());
  printSuccess("DRCEInputJsonMessage serialize");

  resultJson = fObj.Process(sourceJsonStream.str());

  ASSERT_FALSE(fObj.isError()) << printError(fObj);
  ASSERT_EQ(fObj.getErrorCode(), NO_ERROR);
  ASSERT_TRUE(fObj.getErrorMsg().empty());
  printSuccess("DRCEFunctionalObject Process");

//  std::cout << "result JSON: " << resultJson << std::endl;

  // check right result (unserialize and compare two objects)
  resultJsonMessage.unserialize(resultJson);

  ASSERT_FALSE(resultJsonMessage.isError()) << printError(resultJsonMessage);
  ASSERT_EQ(resultJsonMessage.getErrorCode(), NO_ERROR);
  ASSERT_TRUE(resultJsonMessage.getErrorMsg().empty());
  printSuccess("DRCEOutputJsonMessage unserialize");

  ASSERT_EQ(resultJsonMessage.getResultData().getItemsCount(), 1);
  ASSERT_TRUE(resultJsonMessage.getResultData().getDataItem(0).getStderrStream().empty());
  ASSERT_EQ(resultJsonMessage.getResultData().getDataItem(0).getErrorCode(), NO_ERROR);
  ASSERT_EQ(resultJsonMessage.getResultData().getDataItem(0).getNodeName(), fObj.getNodeName());
  ASSERT_EQ(resultJsonMessage.getResultData().getDataItem(0).getNodeHost(), fObj.getNodeHost());
  ASSERT_EQ(resultJsonMessage.getResultData().getDataItem(0).getNodePort(), fObj.getNodePort());
  ASSERT_TRUE(resultJsonMessage.getResultData().getDataItem(0).getState() == DRCETaskRequest::TaskState::FINISHED ||
      resultJsonMessage.getResultData().getDataItem(0).getState() == DRCETaskRequest::TaskState::IN_PROGRESS ||
      resultJsonMessage.getResultData().getDataItem(0).getState() == DRCETaskRequest::TaskState::SET_AS_NEW ||
      resultJsonMessage.getResultData().getDataItem(0).getState() == DRCETaskRequest::TaskState::QUEUED_TO_RUN);
  ASSERT_EQ(resultJsonMessage.getResultData().getDataItem(0).getSubtasksCount(), 2);

  ASSERT_TRUE(resultJsonMessage.getResultData().getDataItem(0).getSubtaskItem(0).getStderrStream().empty());
  ASSERT_EQ(resultJsonMessage.getResultData().getDataItem(0).getSubtaskItem(0).getErrorCode(), NO_ERROR);
  ASSERT_EQ(resultJsonMessage.getResultData().getDataItem(0).getSubtaskItem(0).getNodeName(), fObj.getNodeName());
  ASSERT_EQ(resultJsonMessage.getResultData().getDataItem(0).getSubtaskItem(0).getNodeHost(), fObj.getNodeHost());
  ASSERT_EQ(resultJsonMessage.getResultData().getDataItem(0).getSubtaskItem(0).getNodePort(), fObj.getNodePort());
  ASSERT_TRUE(resultJsonMessage.getResultData().getDataItem(0).getSubtaskItem(0).getState()==DRCETaskRequest::TaskState::FINISHED ||
      resultJsonMessage.getResultData().getDataItem(0).getSubtaskItem(0).getState()==DRCETaskRequest::TaskState::IN_PROGRESS ||
      resultJsonMessage.getResultData().getDataItem(0).getSubtaskItem(0).getState()==DRCETaskRequest::TaskState::SET_AS_NEW ||
      resultJsonMessage.getResultData().getDataItem(0).getSubtaskItem(0).getState()==DRCETaskRequest::TaskState::QUEUED_TO_RUN);
  ASSERT_EQ(resultJsonMessage.getResultData().getDataItem(0).getSubtaskItem(0).getSubtasksCount(), 2);

  ASSERT_TRUE(resultJsonMessage.getResultData().getDataItem(0).getSubtaskItem(1).getStderrStream().empty());
  ASSERT_EQ(resultJsonMessage.getResultData().getDataItem(0).getSubtaskItem(1).getErrorCode(), NO_ERROR);
  ASSERT_EQ(resultJsonMessage.getResultData().getDataItem(0).getSubtaskItem(1).getNodeName(), fObj.getNodeName());
  ASSERT_EQ(resultJsonMessage.getResultData().getDataItem(0).getSubtaskItem(1).getNodeHost(), fObj.getNodeHost());
  ASSERT_EQ(resultJsonMessage.getResultData().getDataItem(0).getSubtaskItem(1).getNodePort(), fObj.getNodePort());
  ASSERT_EQ(resultJsonMessage.getResultData().getDataItem(0).getSubtaskItem(1).getState(), DRCETaskRequest::TaskState::FINISHED ||
      resultJsonMessage.getResultData().getDataItem(0).getSubtaskItem(1).getState()==DRCETaskRequest::TaskState::IN_PROGRESS ||
      resultJsonMessage.getResultData().getDataItem(0).getSubtaskItem(1).getState()==DRCETaskRequest::TaskState::SET_AS_NEW ||
      resultJsonMessage.getResultData().getDataItem(0).getSubtaskItem(1).getState()==DRCETaskRequest::TaskState::QUEUED_TO_RUN);
  ASSERT_EQ(resultJsonMessage.getResultData().getDataItem(0).getSubtaskItem(1).getSubtasksCount(), 2);

  Poco::Thread::sleep(5000);
  for (unsigned int i=0;i<14;++i)
  {
    ASSERT_TRUE(isNotExistTaskData((startTaskId+i), nodeOptions)) << "Task ID: " << (startTaskId+i);
  }

  cleanupTestData(nodeOptions);
  printSuccess("!!! Test Request Get Data with Cleanup");
}
//-----------------------------------------------------------------------------
void DRCECleanupTest::testRequestDeleteDataCleanup(void)
{
  DRCENodeOptions nodeOptions = DRCETaskThreadingModeTest::getNodeOptions();
  cleanupTestData(nodeOptions);
  DRCEFunctionalObject fObj(nodeOptions.getNodeName(), nodeOptions.getHomeDir(), nodeOptions.getTasksDataDir(), nodeOptions.getTasksStatusDir());
  fObj.setNodeHost("NODE_HOST");
  fObj.setNodePort("NODE_PORT");
  fObj.setNodeName("NODE_NAME");

  unsigned int startTaskId = 700;

  std::string sourceJson = getInputJsonCleanup(startTaskId, SessionOptions::ThreadMode::tmAsync, "echo \"Hello\" && sleep 5 && ps ", 100000);

//  std::cout << "sourceJson: " << sourceJson << std::endl;

  std::string resultJson = fObj.Process(sourceJson);

  ASSERT_FALSE(fObj.isError()) << printError(fObj);
  ASSERT_EQ(fObj.getErrorCode(), NO_ERROR);
  ASSERT_TRUE(fObj.getErrorMsg().empty());
  printSuccess("DRCEFunctionalObject Process");

//  std::cout << "result JSON: " << resultJson << std::endl;

  // check right result (unserialize and compare two objects)
  DRCEOutputJsonMessage resultJsonMessage(resultJson);

  ASSERT_FALSE(resultJsonMessage.isError()) << printError(resultJsonMessage);
  ASSERT_EQ(resultJsonMessage.getErrorCode(), NO_ERROR);
  ASSERT_TRUE(resultJsonMessage.getErrorMsg().empty());
  printSuccess("DRCEOutputJsonMessage unserialize");

  while (fObj.getAsyncTasksCount() || fObj.hasTask(startTaskId))
  {
    Poco::Thread::sleep(10000);
    std::cout << "wait end task" << std::endl;
  }
  Poco::Thread::sleep(10000);

  // make get data request
  std::string json;
  DRCETaskRequestDeleteData taskRequestDeleteData;
  taskRequestDeleteData.serialize(json);
  ASSERT_FALSE(taskRequestDeleteData.isError());
  printSuccess("DRCETaskRequestDeleteData serialize");

//  std::cout << "json: " << json << std::endl;

  // set request for execute task
  DRCEInputJsonMessage inputJsonMessage;
  inputJsonMessage.setRequestType(DRCETaskRequest::RequestType::rtDeleteTaskData);
  inputJsonMessage.setRequestData(json);
  inputJsonMessage.setRequestId(startTaskId);

  std::stringstream sourceJsonStream;
  sourceJsonStream << inputJsonMessage;

//  std::cout << "source JSON: " << sourceJson.str() << std::endl;

  ASSERT_FALSE(inputJsonMessage.isError()) << printError(inputJsonMessage);
  ASSERT_EQ(inputJsonMessage.getErrorCode(), NO_ERROR) << printError(inputJsonMessage);
  ASSERT_TRUE(inputJsonMessage.getErrorMsg().empty()) << printError(inputJsonMessage);
  printSuccess("DRCEInputJsonMessage serialize");

  resultJson = fObj.Process(sourceJsonStream.str());

  ASSERT_FALSE(fObj.isError()) << printError(fObj);
  ASSERT_EQ(fObj.getErrorCode(), NO_ERROR);
  ASSERT_TRUE(fObj.getErrorMsg().empty());
  printSuccess("DRCEFunctionalObject Process");

  std::cout << "result JSON: " << resultJson << std::endl;

  // check right result (unserialize and compare two objects)
  resultJsonMessage.unserialize(resultJson);

  ASSERT_FALSE(resultJsonMessage.isError()) << printError(resultJsonMessage);
  ASSERT_EQ(resultJsonMessage.getErrorCode(), NO_ERROR) << printError(resultJsonMessage);
  ASSERT_TRUE(resultJsonMessage.getErrorMsg().empty()) << printError(resultJsonMessage);
  printSuccess("DRCEOutputJsonMessage unserialize");

  ASSERT_EQ(resultJsonMessage.getResultData().getItemsCount(), 1);
  ASSERT_TRUE(resultJsonMessage.getResultData().getDataItem(0).getStderrStream().empty());
  ASSERT_EQ(resultJsonMessage.getResultData().getDataItem(0).getErrorCode(), NO_ERROR) << resultJsonMessage.getResultData().getDataItem(0).getErrorCode();
  ASSERT_EQ(resultJsonMessage.getResultData().getDataItem(0).getNodeName(), fObj.getNodeName());
  ASSERT_EQ(resultJsonMessage.getResultData().getDataItem(0).getNodeHost(), fObj.getNodeHost());
  ASSERT_EQ(resultJsonMessage.getResultData().getDataItem(0).getNodePort(), fObj.getNodePort());
  ASSERT_EQ(resultJsonMessage.getResultData().getDataItem(0).getState(), DRCETaskRequest::TaskState::DELETED);
  ASSERT_EQ(resultJsonMessage.getResultData().getDataItem(0).getSubtasksCount(), 2);

  ASSERT_TRUE(resultJsonMessage.getResultData().getDataItem(0).getSubtaskItem(0).getStderrStream().empty());
  ASSERT_EQ(resultJsonMessage.getResultData().getDataItem(0).getSubtaskItem(0).getErrorCode(), NO_ERROR) << resultJsonMessage.getResultData().getDataItem(0).getSubtaskItem(0).getErrorCode();
  ASSERT_EQ(resultJsonMessage.getResultData().getDataItem(0).getSubtaskItem(0).getNodeName(), fObj.getNodeName());
  ASSERT_EQ(resultJsonMessage.getResultData().getDataItem(0).getSubtaskItem(0).getNodeHost(), fObj.getNodeHost());
  ASSERT_EQ(resultJsonMessage.getResultData().getDataItem(0).getSubtaskItem(0).getNodePort(), fObj.getNodePort());
  ASSERT_EQ(resultJsonMessage.getResultData().getDataItem(0).getSubtaskItem(0).getState(), DRCETaskRequest::TaskState::DELETED);
  ASSERT_EQ(resultJsonMessage.getResultData().getDataItem(0).getSubtaskItem(0).getSubtasksCount(), 2);

  ASSERT_TRUE(resultJsonMessage.getResultData().getDataItem(0).getSubtaskItem(1).getStderrStream().empty());
  ASSERT_EQ(resultJsonMessage.getResultData().getDataItem(0).getSubtaskItem(1).getErrorCode(), NO_ERROR) << resultJsonMessage.getResultData().getDataItem(0).getSubtaskItem(1).getErrorCode();
  ASSERT_EQ(resultJsonMessage.getResultData().getDataItem(0).getSubtaskItem(1).getNodeName(), fObj.getNodeName());
  ASSERT_EQ(resultJsonMessage.getResultData().getDataItem(0).getSubtaskItem(1).getNodeHost(), fObj.getNodeHost());
  ASSERT_EQ(resultJsonMessage.getResultData().getDataItem(0).getSubtaskItem(1).getNodePort(), fObj.getNodePort());
  ASSERT_EQ(resultJsonMessage.getResultData().getDataItem(0).getSubtaskItem(1).getState(), DRCETaskRequest::TaskState::DELETED);
  ASSERT_EQ(resultJsonMessage.getResultData().getDataItem(0).getSubtaskItem(1).getSubtasksCount(), 2);

  Poco::Thread::sleep(5000);
  for (unsigned int i=0;i<14;++i)
  {
    ASSERT_TRUE(isNotExistTaskData((startTaskId+i), nodeOptions)) << "Task ID: " << (startTaskId+i);
  }

  cleanupTestData(nodeOptions);
  printSuccess("!!! Test Request Delete Data with Cleanup");
}
//-----------------------------------------------------------------------------
void DRCECleanupTest::testRequestTerminateCleanup(void)
{
  DRCENodeOptions nodeOptions = DRCETaskThreadingModeTest::getNodeOptions();
  cleanupTestData(nodeOptions);
  DRCEFunctionalObject fObj(nodeOptions.getNodeName(), nodeOptions.getHomeDir(), nodeOptions.getTasksDataDir(), nodeOptions.getTasksStatusDir());
  fObj.setNodeHost("NODE_HOST");
  fObj.setNodePort("NODE_PORT");
  fObj.setNodeName("NODE_NAME");

  unsigned int startTaskId(600), terminateTaskId(603);
  const std::string commandLine = "echo \"scale=400; 4*a(1)\" | bc -l && sleep 10 && ps xau";

  std::string sourceJson = getInputJsonCleanup(startTaskId, SessionOptions::ThreadMode::tmAsync, SessionOptions::CleanupFlag::cfNotDelete, commandLine);

//  std::cout << "sourceJson: " << sourceJson << std::endl;

  std::string resultJson = fObj.Process(sourceJson);

  ASSERT_FALSE(fObj.isError()) << printError(fObj);
  ASSERT_EQ(fObj.getErrorCode(), NO_ERROR);
  ASSERT_TRUE(fObj.getErrorMsg().empty());
  printSuccess("DRCEFunctionalObject Process");

//  std::cout << "result JSON: " << resultJson << std::endl;

  // check right result (unserialize and compare two objects)
  DRCEOutputJsonMessage resultJsonMessage(resultJson);

  ASSERT_FALSE(resultJsonMessage.isError()) << printError(resultJsonMessage);
  ASSERT_EQ(resultJsonMessage.getErrorCode(), NO_ERROR);
  ASSERT_TRUE(resultJsonMessage.getErrorMsg().empty());
  printSuccess("DRCEOutputJsonMessage unserialize");

  // wait timeout and make terminate request
  int tryCount = 10;

  Poco::File statusFile(FileStream::getStatusFileName(nodeOptions.getTasksStatusDir(), terminateTaskId));
  while (!statusFile.exists() || tryCount-->0)
  {
    Poco::Thread::sleep(2000);
    std::cout << "Wait appear task (" << terminateTaskId << ") in async queue ..." << std::endl;
  }
  Poco::Thread::sleep(2000);

  // make get data request
  std::string json;
  DRCETaskRequestTerminate taskRequestTerminate;
  taskRequestTerminate.setAlgorithmType(DRCETaskRequestTerminate::AlgorithmType::atDefault);
  taskRequestTerminate.setDelayValue(1000);
  taskRequestTerminate.setRepeatValue(3);
  taskRequestTerminate.setSignalValue(9);
  taskRequestTerminate.setCleanupFlag(DRCETaskRequestTerminate::CleanupFlag::cfDelete);
  taskRequestTerminate.serialize(json);
  ASSERT_FALSE(taskRequestTerminate.isError());
  printSuccess("DRCETaskRequestTerminate serialize");

//  std::cout << "json: " << json << std::endl;

  // set request for execute task
  DRCEInputJsonMessage inputJsonMessage;
  inputJsonMessage.setRequestType(DRCETaskRequest::RequestType::rtTerminateTask);
  inputJsonMessage.setRequestData(json);
  inputJsonMessage.setRequestId(terminateTaskId);

  std::stringstream sourceJsonStream;
  sourceJsonStream << inputJsonMessage;

//  std::cout << "source JSON: " << sourceJson.str() << std::endl;

  ASSERT_FALSE(inputJsonMessage.isError()) << printError(inputJsonMessage);
  ASSERT_EQ(inputJsonMessage.getErrorCode(), NO_ERROR);
  ASSERT_TRUE(inputJsonMessage.getErrorMsg().empty());
  printSuccess("DRCEInputJsonMessage serialize");

  resultJson = fObj.Process(sourceJsonStream.str());

  ASSERT_FALSE(fObj.isError()) << printError(fObj);
  ASSERT_EQ(fObj.getErrorCode(), NO_ERROR);
  ASSERT_TRUE(fObj.getErrorMsg().empty());
  printSuccess("DRCEFunctionalObject Process");

  std::cout << "result JSON: " << resultJson << std::endl;

  // check right result (unserialize and compare two objects)
  resultJsonMessage.unserialize(resultJson);

  ASSERT_FALSE(resultJsonMessage.isError()) << printError(resultJsonMessage);
  ASSERT_EQ(resultJsonMessage.getErrorCode(), NO_ERROR);
  ASSERT_TRUE(resultJsonMessage.getErrorMsg().empty());
  printSuccess("DRCEOutputJsonMessage unserialize");

  while (fObj.getAsyncTasksCount() || fObj.hasTask(startTaskId))
  {
    Poco::Thread::sleep(5000);
    std::cout << "wait end task" << std::endl;
  }
  Poco::Thread::sleep(5000);

  ASSERT_EQ(resultJsonMessage.getResultData().getItemsCount(), 1);
  ASSERT_EQ(resultJsonMessage.getResultData().getDataItem(0).getRequestId(), 603);
  ASSERT_TRUE(resultJsonMessage.getResultData().getDataItem(0).getStderrStream().empty());
  ASSERT_EQ(resultJsonMessage.getResultData().getDataItem(0).getErrorCode(), NO_ERROR); //??
  ASSERT_EQ(resultJsonMessage.getResultData().getDataItem(0).getNodeName(), fObj.getNodeName());
  ASSERT_EQ(resultJsonMessage.getResultData().getDataItem(0).getNodeHost(), fObj.getNodeHost());
  ASSERT_EQ(resultJsonMessage.getResultData().getDataItem(0).getNodePort(), fObj.getNodePort());
  ASSERT_EQ(resultJsonMessage.getResultData().getDataItem(0).getState(), DRCETaskRequest::TaskState::TERMINATED);
  ASSERT_EQ(resultJsonMessage.getResultData().getDataItem(0).getSubtasksCount(), 2);

  ASSERT_TRUE(resultJsonMessage.getResultData().getDataItem(0).getSubtaskItem(0).getStderrStream().empty());
  ASSERT_EQ(resultJsonMessage.getResultData().getDataItem(0).getSubtaskItem(0).getErrorCode(), NO_ERROR);
  ASSERT_EQ(resultJsonMessage.getResultData().getDataItem(0).getSubtaskItem(0).getNodeName(), fObj.getNodeName());
  ASSERT_EQ(resultJsonMessage.getResultData().getDataItem(0).getSubtaskItem(0).getNodeHost(), fObj.getNodeHost());
  ASSERT_EQ(resultJsonMessage.getResultData().getDataItem(0).getSubtaskItem(0).getNodePort(), fObj.getNodePort());
  ASSERT_EQ(resultJsonMessage.getResultData().getDataItem(0).getSubtaskItem(0).getState(), DRCETaskRequest::TaskState::TERMINATED);
  ASSERT_EQ(resultJsonMessage.getResultData().getDataItem(0).getSubtaskItem(0).getSubtasksCount(), 0);

  ASSERT_TRUE(resultJsonMessage.getResultData().getDataItem(0).getSubtaskItem(1).getStderrStream().empty());
  ASSERT_EQ(resultJsonMessage.getResultData().getDataItem(0).getSubtaskItem(1).getErrorCode(), NO_ERROR);
  ASSERT_EQ(resultJsonMessage.getResultData().getDataItem(0).getSubtaskItem(1).getNodeName(), fObj.getNodeName());
  ASSERT_EQ(resultJsonMessage.getResultData().getDataItem(0).getSubtaskItem(1).getNodeHost(), fObj.getNodeHost());
  ASSERT_EQ(resultJsonMessage.getResultData().getDataItem(0).getSubtaskItem(1).getNodePort(), fObj.getNodePort());
  ASSERT_EQ(resultJsonMessage.getResultData().getDataItem(0).getSubtaskItem(1).getState(), DRCETaskRequest::TaskState::TERMINATED);
  ASSERT_EQ(resultJsonMessage.getResultData().getDataItem(0).getSubtaskItem(1).getSubtasksCount(), 0);

  Poco::Thread::sleep(10000);
  std::vector<unsigned int> cleanupTasks = {603, 607, 608};
  for (unsigned int i=0;i<cleanupTasks.size();++i)
  {
    ASSERT_TRUE(isNotExistTaskData(cleanupTasks[i], nodeOptions)) << "isNotExistTaskData  cleanupTasks[" << i << "]: " << cleanupTasks[i];
  }

  for (unsigned int i=0;i<14;++i)
  {
    unsigned int taskId = startTaskId+i;
    bool isNotExist = false;
    for (unsigned int i=0;i<cleanupTasks.size();++i)
    {
      if (cleanupTasks[i]==taskId)
      {
        isNotExist = true;
        break;
      }
    }
    if (!isNotExist)
    {
      ASSERT_TRUE(isExistTaskData(taskId, nodeOptions)) << " isExistTaskData taskId: " << taskId;
    }
  }

  cleanupTestData(nodeOptions);
  printSuccess("!!! Test Request Terminate with Cleanup");
}
//-----------------------------------------------------------------------------
void DRCECleanupTest::testRequestDeleteData(void)
{
  DRCENodeOptions nodeOptions = DRCETaskThreadingModeTest::getNodeOptions();
  cleanupTestData(nodeOptions);
  DRCEFunctionalObject fObj(nodeOptions.getNodeName(), nodeOptions.getHomeDir(), nodeOptions.getTasksDataDir(), nodeOptions.getTasksStatusDir());
  fObj.setNodeHost("NODE_HOST");
  fObj.setNodePort("NODE_PORT");
  fObj.setNodeName("NODE_NAME");

  unsigned int startTaskId = 600;

  DRCEInputJsonMessage inputJsonMessage = getInputJsonMessage(startTaskId, SessionOptions::ThreadMode::tmAsync, SessionOptions::CleanupFlag::cfNotDelete,
              "echo \"Hello\" && sleep 5 && ps xau",
              300);

  std::stringstream jsonStream;
  jsonStream << inputJsonMessage;

  ASSERT_FALSE(inputJsonMessage.isError()) << printError(inputJsonMessage);
  ASSERT_EQ(inputJsonMessage.getErrorCode(), NO_ERROR) << printError(inputJsonMessage);
  ASSERT_TRUE(inputJsonMessage.getErrorMsg().empty()) << printError(inputJsonMessage);

//  std::cout << "sourceJson: " << sourceJson << std::endl;

  std::string resultJson = fObj.Process(jsonStream.str());

  ASSERT_FALSE(fObj.isError()) << printError(fObj);
  ASSERT_EQ(fObj.getErrorCode(), NO_ERROR);
  ASSERT_TRUE(fObj.getErrorMsg().empty());
  printSuccess("DRCEFunctionalObject Process");

//  std::cout << "result JSON: " << resultJson << std::endl;

  // check right result (unserialize and compare two objects)
  DRCEOutputJsonMessage resultJsonMessage(resultJson);

  ASSERT_FALSE(resultJsonMessage.isError()) << printError(resultJsonMessage);
  ASSERT_EQ(resultJsonMessage.getErrorCode(), NO_ERROR);
  ASSERT_TRUE(resultJsonMessage.getErrorMsg().empty());
  printSuccess("DRCEOutputJsonMessage unserialize");

  while (fObj.getAsyncTasksCount() || fObj.hasTask(startTaskId))
  {
    Poco::Thread::sleep(5000);
    std::cout << "wait end task" << std::endl;
  }
  Poco::Thread::sleep(5000);

  // make get data request
  std::string json;
  DRCETaskRequestDeleteData taskRequestDeleteData;
  taskRequestDeleteData.serialize(json);
  ASSERT_FALSE(taskRequestDeleteData.isError());
  printSuccess("DRCETaskRequestDeleteData serialize");
  // set request for execute task
  inputJsonMessage.clear();
  inputJsonMessage.setRequestType(DRCETaskRequest::RequestType::rtDeleteTaskData);
  inputJsonMessage.setRequestData(json);
  inputJsonMessage.setRequestId(startTaskId);

  std::stringstream sourceJsonStream;
  sourceJsonStream << inputJsonMessage;

  //  std::cout << "source JSON: " << sourceJson.str() << std::endl;

  ASSERT_FALSE(inputJsonMessage.isError()) << printError(inputJsonMessage);
  ASSERT_EQ(inputJsonMessage.getErrorCode(), NO_ERROR) << printError(inputJsonMessage);
  ASSERT_TRUE(inputJsonMessage.getErrorMsg().empty()) << printError(inputJsonMessage);
  printSuccess("DRCEInputJsonMessage serialize");

  resultJson = fObj.Process(sourceJsonStream.str());

  ASSERT_FALSE(fObj.isError()) << printError(fObj);
  ASSERT_EQ(fObj.getErrorCode(), NO_ERROR);
  ASSERT_TRUE(fObj.getErrorMsg().empty());
  printSuccess("DRCEFunctionalObject Process");

//  std::cout << "result JSON: " << resultJson << std::endl;

  // check right result (unserialize and compare two objects)
  resultJsonMessage.unserialize(resultJson);

  ASSERT_FALSE(resultJsonMessage.isError()) << printError(resultJsonMessage);
  ASSERT_EQ(resultJsonMessage.getErrorCode(), NO_ERROR) << printError(resultJsonMessage);
  ASSERT_TRUE(resultJsonMessage.getErrorMsg().empty()) << printError(resultJsonMessage);
  printSuccess("DRCEOutputJsonMessage unserialize");
}
//-----------------------------------------------------------------------------
//-----------------------------------------------------------------------------
} // end namespace tests
} // end namespace drce 
} // end namespace HCE 
