#include <gtest/gtest.h>
#include <fstream>
#include <sstream>
#include <iomanip>
#include <utility>

#include "DRCEError.hpp"
#include "DRCESessionOptions.hpp"
#include "DRCEInputJsonMessage.hpp"
#include "DRCETaskRequestSetExecute.hpp"
#include "DRCEOutputJsonMessage.hpp"
#include "DRCEFunctionalObject.hpp"
#include "DRCEResultDataSerializator.hpp"
#include "DRCEResultData.hpp"
#include "DRCENotificationExecutor.hpp"

#include "DRCEProtocolsTest.hpp"
#include "DRCEPrintStatus.hpp"
#include "DRCEFunctionalObjectTest.hpp"
#include "DRCETaskThreadingModeTest.hpp"

namespace HCE
{
namespace drce
{
namespace tests
{
//-----------------------------------------------------------------------------
void DRCEFunctionalObjectTest::testApplyEnviroments(void)
{
  std::vector<std::pair<std::string, std::string> > testEnv={{"env1", "val1"}, {"env2", "val2"}, {"env3", "val3"}};

  DRCENodeOptions nodeOptions = DRCETaskThreadingModeTest::getNodeOptions();
  DRCEFunctionalObject fObj(nodeOptions.getNodeName(), nodeOptions.getHomeDir(), nodeOptions.getTasksDataDir(), nodeOptions.getTasksStatusDir());

  std::stringstream sourceJson;
  DRCETaskRequestSetExecute sourceMessage;
  sourceMessage.setCommandLine("env | grep PATH");

  SessionOptions sessionOption;
  sessionOption.shellName = "bash -c";
  sessionOption.environments = testEnv;
  sourceMessage.setSessionOptions(sessionOption);

  sourceJson << sourceMessage;

//  std::cout << "sourceJson = " << sourceJson.str() << std::endl;

  ASSERT_FALSE(sourceMessage.isError()) << printError(sourceMessage);
  ASSERT_EQ(sourceMessage.getErrorCode(), NO_ERROR);
  ASSERT_TRUE(sourceMessage.getErrorMsg().empty());
  printSuccess("DRCETaskRequestSetExecute serialize");

  DRCEInputJsonMessage inputJsonMessage;
  inputJsonMessage.setRequestType(DRCETaskRequest::RequestType::rtSetTaskExecute);
  inputJsonMessage.setRequestData(sourceJson.str());
  inputJsonMessage.setRequestId(2345);

  sourceJson.str("");
  sourceJson << inputJsonMessage;

  ASSERT_FALSE(inputJsonMessage.isError()) << printError(inputJsonMessage);
  ASSERT_EQ(inputJsonMessage.getErrorCode(), NO_ERROR);
  ASSERT_TRUE(inputJsonMessage.getErrorMsg().empty());
  printSuccess("DRCEInputJsonMessage serialize");

  std::string resultJson = fObj.Process(sourceJson.str());

//  std::cout << "resultJson = " << resultJson << std::endl;

  ASSERT_FALSE(fObj.isError()) << printError(fObj);
  ASSERT_EQ(fObj.getErrorCode(), NO_ERROR);
  ASSERT_TRUE(fObj.getErrorMsg().empty());
  printSuccess("DRCEFunctionalObject Process use Enviroments");
}
//-----------------------------------------------------------------------------
//-----------------------------------------------------------------------------
void DRCEFunctionalObjectTest::testRequestLocalShellSortInput(void)
{
  const std::string json = "{\"session\":{\"type\":0,\"port\":0,\"user\":\"\",\"tmode\":1,\"time_max\":0,\"password\":\"\",\"shell\":\"\",\"environment\":[{\"LINES\":\"80\",\"HOSTNAME\":\"node22\"}]},\"command\":\"sort\",\"input\":\"def\\nabc\\n123\\n\",\"files\":[{\"name\":\"file1\",\"data\":\"test content of file1\\n\",\"action\":13},{\"name\":\"file2\",\"data\":\"test content of file2\\n\",\"action\":1}]}";

  DRCENodeOptions nodeOptions = DRCETaskThreadingModeTest::getNodeOptions();
  DRCEFunctionalObject fObj(nodeOptions.getNodeName(), nodeOptions.getHomeDir(), nodeOptions.getTasksDataDir(), nodeOptions.getTasksStatusDir());

  DRCETaskRequestSetExecute inputJsonMessage(json);

  ASSERT_FALSE(inputJsonMessage.isError()) << printError(inputJsonMessage);
  ASSERT_EQ(inputJsonMessage.getErrorCode(), NO_ERROR);
  ASSERT_TRUE(inputJsonMessage.getErrorMsg().empty());
  printSuccess("DRCETaskRequestSetExecute unserialize");

  ASSERT_EQ(inputJsonMessage.getSessionOptions().sessionType, static_cast<SessionOptions::SessionType>(0));
  ASSERT_TRUE(inputJsonMessage.getSessionOptions().userName.empty());
  ASSERT_TRUE(inputJsonMessage.getSessionOptions().userPassword.empty());
  ASSERT_TRUE(inputJsonMessage.getSessionOptions().shellName.empty());
  ASSERT_EQ(inputJsonMessage.getSessionOptions().port, 0);
  ASSERT_EQ(inputJsonMessage.getSessionOptions().environments.size(), 2);
  for (size_t i=0;i<inputJsonMessage.getSessionOptions().environments.size();++i)
  {
    if (inputJsonMessage.getSessionOptions().environments[i].first=="LINES")
      ASSERT_EQ(inputJsonMessage.getSessionOptions().environments[i].second, "80");
    if (inputJsonMessage.getSessionOptions().environments[i].first=="HOSTNAME")
      ASSERT_EQ(inputJsonMessage.getSessionOptions().environments[i].second, "node22");
  }
  printSuccess("DRCESessionOptions checking members");

  ASSERT_EQ(inputJsonMessage.getFilesCount(), 2);
  for (size_t i=0;i<inputJsonMessage.getFilesCount();++i)
  {
    if (inputJsonMessage.getFileItem(i).name=="file1")
    {
      ASSERT_EQ(inputJsonMessage.getFileItem(i).data, "test content of file1\n");
      ASSERT_EQ(inputJsonMessage.getFileItem(i).actionType, 13);
    }
    if (inputJsonMessage.getFileItem(i).name=="file2")
    {
      ASSERT_EQ(inputJsonMessage.getFileItem(i).data, "test content of file2\n");
      ASSERT_EQ(inputJsonMessage.getFileItem(i).actionType, 1);
    }
  }
  printSuccess("DRCEFilesItemArray checking members");

  ASSERT_EQ(inputJsonMessage.getCommandLine(), "sort");
  ASSERT_EQ(inputJsonMessage.getInputStream(), "def\nabc\n123\n");
  printSuccess("Test Request for local Shell Input");
}
//-----------------------------------------------------------------------------
void DRCEFunctionalObjectTest::testRequestLocalShellSortOutput(void)
{
  const std::string json = "[{\"error_code\": 0,\"error_message\": \"Success\",\"exit_status\": 5,\"fields\": [],\"files\":[{\"action\":13,\"data\":\"test content of file1\\n\",\"name\":\"file1\"}],\"host\":\"\",\"id\":2345,\"node\":\"TEST_NODE\",\"pid\":2595,\"port\":\"\",\"state\":42,\"stderror\":\"\",\"stdout\":\"MTIzCmFiYwpkZWYK\",\"time\":5,\"type\":0}]";

  DRCEOutputJsonMessage outputJsonMessage(json);

  ASSERT_FALSE(outputJsonMessage.isError()) << printError(outputJsonMessage);
  ASSERT_EQ(outputJsonMessage.getErrorCode(), NO_ERROR);
  ASSERT_TRUE(outputJsonMessage.getErrorMsg().empty());
  printSuccess("DRCEOutputJsonMessage unserialize");

  ASSERT_EQ(outputJsonMessage.getResultData().getItemsCount(), 1);
  ASSERT_EQ(outputJsonMessage.getResultData().getDataItem(0).getStdoutStream(), "123\nabc\ndef\n");
  ASSERT_EQ(outputJsonMessage.getResultData().getDataItem(0).getStderrStream(), "");
  ASSERT_EQ(outputJsonMessage.getResultData().getDataItem(0).getErrorCode(), 0);
  ASSERT_EQ(outputJsonMessage.getResultData().getDataItem(0).getErrorMessage(), "Success");
  ASSERT_EQ(outputJsonMessage.getResultData().getDataItem(0).getExitStatus(), 5);
  ASSERT_EQ(outputJsonMessage.getResultData().getDataItem(0).getNodeHost(), "");
  ASSERT_EQ(outputJsonMessage.getResultData().getDataItem(0).getNodeName(), "TEST_NODE");
  ASSERT_EQ(outputJsonMessage.getResultData().getDataItem(0).getNodePort(), "");
  ASSERT_EQ(outputJsonMessage.getResultData().getDataItem(0).getRequestId(), 2345);
  ASSERT_EQ(outputJsonMessage.getResultData().getDataItem(0).getRequestType(), DRCETaskRequest::RequestType::rtSetTaskExecute);
  ASSERT_EQ(outputJsonMessage.getResultData().getDataItem(0).getTime(), 5);
  ASSERT_EQ(outputJsonMessage.getResultData().getDataItem(0).getState(), 42);
  ASSERT_EQ(outputJsonMessage.getResultData().getDataItem(0).getFilesCount(), 1);
  ASSERT_EQ(outputJsonMessage.getResultData().getDataItem(0).getFileItem(0).name, "file1");
  ASSERT_EQ(outputJsonMessage.getResultData().getDataItem(0).getFileItem(0).data, "test content of file1\n");
  ASSERT_EQ(outputJsonMessage.getResultData().getDataItem(0).getFileItem(0).actionType, 13);
  printSuccess("Test Request for local Shell Output");
}
//-----------------------------------------------------------------------------
void DRCEFunctionalObjectTest::testRequestLocalShellSort(void)
{
  const std::string requestJson = "{\"session\":{\"type\":0,\"port\":0,\"user\":\"\",\"time_max\":0,\"tmode\":1,\"time_max\":0,\"password\":\"\",\"shell\":\"\",\"environment\":[{\"LINES\":\"80\",\"HOSTNAME\":\"node22\"}]},\"command\":\"sort\",\"input\":\"def\\nabc\\n123\\n\",\"files\":[{\"name\":\"file1\",\"data\":\"test content of file1\\n\",\"action\":13},{\"name\":\"file2\",\"data\":\"test content of file2\\n\",\"action\":1}]}";

  DRCENodeOptions nodeOptions = DRCETaskThreadingModeTest::getNodeOptions();
  DRCEFunctionalObject fObj(nodeOptions.getNodeName(), nodeOptions.getHomeDir(), nodeOptions.getTasksDataDir(), nodeOptions.getTasksStatusDir());
  fObj.setNodeHost("NODE_HOST");
  fObj.setNodePort("NODE_PORT");
  fObj.setNodeName("NODE_NAME");

//  std::cout << "json: " << requestJson << std::endl;

  DRCEInputJsonMessage inputJsonMessage;
  inputJsonMessage.setRequestType(DRCETaskRequest::RequestType::rtSetTaskExecute);
  inputJsonMessage.setRequestData(requestJson);
  inputJsonMessage.setRequestId(2345);

  std::stringstream sourceJson;
  sourceJson << inputJsonMessage;

  ASSERT_FALSE(inputJsonMessage.isError()) << printError(inputJsonMessage);
  ASSERT_EQ(inputJsonMessage.getErrorCode(), NO_ERROR);
  ASSERT_TRUE(inputJsonMessage.getErrorMsg().empty());
  printSuccess("DRCEInputJsonMessage serialize");

//  std::cout << "IN JSON: " << sourceJson.str() << std::endl;

  std::string resultJson = fObj.Process(sourceJson.str());

  ASSERT_FALSE(fObj.isError()) << printError(fObj);
  ASSERT_EQ(fObj.getErrorCode(), NO_ERROR);
  ASSERT_TRUE(fObj.getErrorMsg().empty());
  printSuccess("DRCEFunctionalObject Process");

//  std::cout << "OUT JSON: " << resultJson << std::endl;

  DRCEOutputJsonMessage resultJsonMessage(resultJson);

  ASSERT_FALSE(resultJsonMessage.isError()) << printError(resultJsonMessage);
  ASSERT_EQ(resultJsonMessage.getErrorCode(), NO_ERROR);
  ASSERT_TRUE(resultJsonMessage.getErrorMsg().empty());
  printSuccess("DRCEOutputJsonMessage unserialize");

  ASSERT_EQ(resultJsonMessage.getResultData().getItemsCount(), 1);
  ASSERT_EQ(resultJsonMessage.getResultData().getDataItem(0).getStdoutStream(), "123\nabc\ndef\n");
  ASSERT_EQ(resultJsonMessage.getResultData().getDataItem(0).getStderrStream(), "");
  ASSERT_EQ(resultJsonMessage.getResultData().getDataItem(0).getErrorCode(), 0);
  ASSERT_EQ(resultJsonMessage.getResultData().getDataItem(0).getErrorMessage(), "");
  ASSERT_EQ(resultJsonMessage.getResultData().getDataItem(0).getExitStatus(), 0);
  ASSERT_EQ(resultJsonMessage.getResultData().getDataItem(0).getNodeHost(), "NODE_HOST");
  ASSERT_EQ(resultJsonMessage.getResultData().getDataItem(0).getNodeName(), "NODE_NAME");
  ASSERT_EQ(resultJsonMessage.getResultData().getDataItem(0).getNodePort(), "NODE_PORT");
  ASSERT_EQ(resultJsonMessage.getResultData().getDataItem(0).getRequestId(), 2345);
  ASSERT_EQ(resultJsonMessage.getResultData().getDataItem(0).getRequestType(), DRCETaskRequest::RequestType::rtSetTaskExecute);
  ASSERT_EQ(resultJsonMessage.getResultData().getDataItem(0).getFilesCount(), 1);
  ASSERT_EQ(resultJsonMessage.getResultData().getDataItem(0).getFileItem(0).name, "file1");
  ASSERT_EQ(resultJsonMessage.getResultData().getDataItem(0).getFileItem(0).data, "test content of file1\n");
  ASSERT_EQ(resultJsonMessage.getResultData().getDataItem(0).getFileItem(0).actionType, 13);
  printSuccess("Test Request for local Shell (sort)");
}
//-----------------------------------------------------------------------------
//-----------------------------------------------------------------------------
void DRCEFunctionalObjectTest::testRequestLocalShellPsInput(void)
{
  const std::string json = "{\"session\":{\"type\":0,\"port\":0,\"user\":\"\",\"password\":\"\",\"time_max\":0,\"tmode\":1,\"shell\":\"\",\"environment\":[{\"LINES\":\"80\",\"HOSTNAME\":\"node22\"}]},\"command\":\"ps > _ps_results.txt\",\"input\":\"\",\"files\":[{\"name\":\"_ps_results.txt\",\"data\":\"\",\"action\":12}]}";

  DRCENodeOptions nodeOptions = DRCETaskThreadingModeTest::getNodeOptions();
  DRCEFunctionalObject fObj(nodeOptions.getNodeName(), nodeOptions.getHomeDir(), nodeOptions.getTasksDataDir(), nodeOptions.getTasksStatusDir());

  DRCETaskRequestSetExecute taskRequestExecute(json);

  ASSERT_FALSE(taskRequestExecute.isError()) << printError(taskRequestExecute);
  ASSERT_EQ(taskRequestExecute.getErrorCode(), NO_ERROR);
  ASSERT_TRUE(taskRequestExecute.getErrorMsg().empty());
  printSuccess("DRCETaskRequestSetExecute unserialize");

  ASSERT_EQ(taskRequestExecute.getSessionOptions().sessionType, static_cast<SessionOptions::SessionType>(0));
  ASSERT_TRUE(taskRequestExecute.getSessionOptions().userName.empty());
  ASSERT_TRUE(taskRequestExecute.getSessionOptions().userPassword.empty());
  ASSERT_TRUE(taskRequestExecute.getSessionOptions().shellName.empty());
  ASSERT_EQ(taskRequestExecute.getSessionOptions().port, 0);
  ASSERT_EQ(taskRequestExecute.getSessionOptions().environments.size(), 2);

  for (size_t i=0;i<taskRequestExecute.getSessionOptions().environments.size();++i)
  {
    if (taskRequestExecute.getSessionOptions().environments[i].first=="LINES")
      ASSERT_EQ(taskRequestExecute.getSessionOptions().environments[i].second, "80");
    if (taskRequestExecute.getSessionOptions().environments[i].first=="HOSTNAME")
      ASSERT_EQ(taskRequestExecute.getSessionOptions().environments[i].second, "node22");
  }
  printSuccess("SessionOptions checking members");

  ASSERT_EQ(taskRequestExecute.getFilesCount(), 1);
  ASSERT_EQ(taskRequestExecute.getFileItem(0).name, "_ps_results.txt");
  ASSERT_TRUE(taskRequestExecute.getFileItem(0).data.empty());
  ASSERT_EQ(taskRequestExecute.getFileItem(0).actionType, 12);
  printSuccess("DRCEFilesItemList checking members");

  ASSERT_EQ(taskRequestExecute.getCommandLine(), "ps > _ps_results.txt");
  ASSERT_TRUE(taskRequestExecute.getInputStream().empty());
  printSuccess("Test Request for local Shell Input");
}
//-----------------------------------------------------------------------------
void DRCEFunctionalObjectTest::testRequestLocalShellPsOutput(void)
{
  const std::string json = "[{\"error_code\":1,\"error_message\":\"none\",\"exit_status\":127,\"fields\":[],\"files\":[{\"action\":12,\"data\":\"File content\",\"name\":\"_ps_results.txt\"}],\"host\":\"NODE_HOST\",\"id\":2345,\"node\":\"NODE_NAME\",\"pid\":4208,\"port\":\"NODE_PORT\",\"state\":42,\"stderror\":\"stderr_content\",\"stdout\":\"stdout_content\",\"time\":13,\"type\":1}]";

  DRCEOutputJsonMessage outputJsonMessage(json);

  ASSERT_FALSE(outputJsonMessage.isError()) << printError(outputJsonMessage);
  ASSERT_EQ(outputJsonMessage.getErrorCode(), NO_ERROR);
  ASSERT_TRUE(outputJsonMessage.getErrorMsg().empty());
  printSuccess("DRCEOutputJsonMessage unserialize");

  ASSERT_EQ(outputJsonMessage.getResultData().getItemsCount(), 1);
  ASSERT_EQ(outputJsonMessage.getResultData().getDataItem(0).getStdoutStream(), "");
  ASSERT_EQ(outputJsonMessage.getResultData().getDataItem(0).getStderrStream(), "");
  ASSERT_EQ(outputJsonMessage.getResultData().getDataItem(0).getErrorCode(), 1);
  ASSERT_EQ(outputJsonMessage.getResultData().getDataItem(0).getErrorMessage(), "none");
  ASSERT_EQ(outputJsonMessage.getResultData().getDataItem(0).getExitStatus(), 127);
  ASSERT_EQ(outputJsonMessage.getResultData().getDataItem(0).getNodeHost(), "NODE_HOST");
  ASSERT_EQ(outputJsonMessage.getResultData().getDataItem(0).getNodeName(), "NODE_NAME");
  ASSERT_EQ(outputJsonMessage.getResultData().getDataItem(0).getNodePort(), "NODE_PORT");
  ASSERT_EQ(outputJsonMessage.getResultData().getDataItem(0).getRequestId(), 2345);
  ASSERT_EQ(outputJsonMessage.getResultData().getDataItem(0).getRequestType(), DRCETaskRequest::RequestType::rtCheckTaskState);
  ASSERT_EQ(outputJsonMessage.getResultData().getDataItem(0).getTime(), 13);
  ASSERT_EQ(outputJsonMessage.getResultData().getDataItem(0).getState(), 42);
  ASSERT_EQ(outputJsonMessage.getResultData().getDataItem(0).getFilesCount(), 1);
  ASSERT_EQ(outputJsonMessage.getResultData().getDataItem(0).getFileItem(0).name, "_ps_results.txt");
  ASSERT_EQ(outputJsonMessage.getResultData().getDataItem(0).getFileItem(0).data, "File content");
  ASSERT_EQ(outputJsonMessage.getResultData().getDataItem(0).getFileItem(0).actionType, 12);
  printSuccess("Test Request for local Shell Output");
}
//-----------------------------------------------------------------------------
void DRCEFunctionalObjectTest::testRequestLocalShellPs(void)
{
  const std::string requestJson = "{\"session\":{\"type\":0,\"port\":0,\"user\":\"\",\"time_max\":0,\"tmode\":1,\"password\":\"\",\"shell\":\"\",\"environment\":[{\"LINES\":\"80\",\"HOSTNAME\":\"node22\"}]},\"command\":\"ps > _ps_results.txt\",\"input\":\"\",\"files\":[{\"name\":\"_ps_results.txt\",\"data\":\"\",\"action\":12}]}";

  DRCENodeOptions nodeOptions = DRCETaskThreadingModeTest::getNodeOptions();
  DRCEFunctionalObject fObj(nodeOptions.getNodeName(), nodeOptions.getHomeDir(), nodeOptions.getTasksDataDir(), nodeOptions.getTasksStatusDir());
  fObj.setNodeHost("NODE_HOST");
  fObj.setNodePort("NODE_PORT");
  fObj.setNodeName("NODE_NAME");

//  std::cout << "json: " << requestJson << std::endl;

  DRCEInputJsonMessage inputJsonMessage;
  inputJsonMessage.setRequestType(DRCETaskRequest::RequestType::rtSetTaskExecute);
  inputJsonMessage.setRequestData(requestJson);
  inputJsonMessage.setRequestId(2345);

  std::stringstream sourceJson;
  sourceJson << inputJsonMessage;

  ASSERT_FALSE(inputJsonMessage.isError()) << printError(inputJsonMessage);
  ASSERT_EQ(inputJsonMessage.getErrorCode(), NO_ERROR);
  ASSERT_TRUE(inputJsonMessage.getErrorMsg().empty());
  printSuccess("DRCEInputJsonMessage serialize");

//  std::cout << "IN JSON: " << sourceJson.str() << std::endl;

  std::string resultJson = fObj.Process(sourceJson.str());

  ASSERT_FALSE(fObj.isError()) << printError(fObj);
  ASSERT_EQ(fObj.getErrorCode(), NO_ERROR);
  ASSERT_TRUE(fObj.getErrorMsg().empty());
  printSuccess("DRCEFunctionalObject Process");

//  std::cout << "OUT JSON: " << resultJson << std::endl;

  DRCEOutputJsonMessage resultJsonMessage(resultJson);

  ASSERT_FALSE(resultJsonMessage.isError()) << printError(resultJsonMessage);
  ASSERT_EQ(resultJsonMessage.getErrorCode(), NO_ERROR);
  ASSERT_TRUE(resultJsonMessage.getErrorMsg().empty());
  printSuccess("DRCEOutputJsonMessage unserialize");

  ASSERT_EQ(resultJsonMessage.getResultData().getItemsCount(), 1);
  ASSERT_EQ(resultJsonMessage.getResultData().getDataItem(0).getStdoutStream(), "");
  ASSERT_EQ(resultJsonMessage.getResultData().getDataItem(0).getStderrStream(), "");
  ASSERT_EQ(resultJsonMessage.getResultData().getDataItem(0).getErrorCode(), 0);
  ASSERT_EQ(resultJsonMessage.getResultData().getDataItem(0).getErrorMessage(), "");
  ASSERT_EQ(resultJsonMessage.getResultData().getDataItem(0).getExitStatus(), 0);
  ASSERT_EQ(resultJsonMessage.getResultData().getDataItem(0).getNodeHost(), "NODE_HOST");
  ASSERT_EQ(resultJsonMessage.getResultData().getDataItem(0).getNodeName(), "NODE_NAME");
  ASSERT_EQ(resultJsonMessage.getResultData().getDataItem(0).getNodePort(), "NODE_PORT");
  ASSERT_EQ(resultJsonMessage.getResultData().getDataItem(0).getRequestId(), 2345);
  ASSERT_EQ(resultJsonMessage.getResultData().getDataItem(0).getRequestType(), DRCETaskRequest::RequestType::rtSetTaskExecute);
  ASSERT_EQ(resultJsonMessage.getResultData().getDataItem(0).getFilesCount(), 1);
  ASSERT_EQ(resultJsonMessage.getResultData().getDataItem(0).getFileItem(0).name, "_ps_results.txt");
  ASSERT_FALSE(resultJsonMessage.getResultData().getDataItem(0).getFileItem(0).data.empty());
  ASSERT_EQ(resultJsonMessage.getResultData().getDataItem(0).getFileItem(0).actionType, 12);
  printSuccess("Test Request for local Shell (ps)");  // ps -e
}
//-----------------------------------------------------------------------------
//-----------------------------------------------------------------------------
void DRCEFunctionalObjectTest::testRequestLocalShellPhpVersion(void)
{
  const std::string requestJson = "{\"session\":{\"type\":0,\"port\":0,\"user\":\"\",\"time_max\":0,\"tmode\":1,\"password\":\"\",\"shell\":\"\",\"environment\":[{\"LINES\":\"80\",\"HOSTNAME\":\"node22\"}]},\"command\":\"php -v > php_version.txt\",\"input\":\"\",\"files\":[{\"name\":\"php_version.txt\",\"data\":\"\",\"action\":12}]}";

  DRCENodeOptions nodeOptions = DRCETaskThreadingModeTest::getNodeOptions();
  DRCEFunctionalObject fObj(nodeOptions.getNodeName(), nodeOptions.getHomeDir(), nodeOptions.getTasksDataDir(), nodeOptions.getTasksStatusDir());
  fObj.setNodeHost("NODE_HOST");
  fObj.setNodePort("NODE_PORT");
  fObj.setNodeName("NODE_NAME");

  DRCEInputJsonMessage inputJsonMessage;
  inputJsonMessage.setRequestType(DRCETaskRequest::RequestType::rtSetTaskExecute);
  inputJsonMessage.setRequestData(requestJson);
  inputJsonMessage.setRequestId(2345);

  std::stringstream sourceJson;
  sourceJson << inputJsonMessage;

  ASSERT_FALSE(inputJsonMessage.isError()) << printError(inputJsonMessage);
  ASSERT_EQ(inputJsonMessage.getErrorCode(), NO_ERROR);
  ASSERT_TRUE(inputJsonMessage.getErrorMsg().empty());
  printSuccess("DRCEInputJsonMessage serialize");

  std::string resultJson = fObj.Process(sourceJson.str());

  ASSERT_FALSE(fObj.isError()) << printError(fObj);
  ASSERT_EQ(fObj.getErrorCode(), NO_ERROR);
  ASSERT_TRUE(fObj.getErrorMsg().empty());
  printSuccess("DRCEFunctionalObject Process");

//  std::cout << "JSON: " << resultJson << std::endl;
//  std::cout << "JSON: " << responseJson << std::endl;

  // check right result (unserialize and compare two objects)
  DRCEOutputJsonMessage resultJsonMessage(resultJson);

  ASSERT_FALSE(resultJsonMessage.isError()) << printError(resultJsonMessage);
  ASSERT_EQ(resultJsonMessage.getErrorCode(), NO_ERROR);
  ASSERT_TRUE(resultJsonMessage.getErrorMsg().empty());
  printSuccess("DRCEOutputJsonMessage unserialize");

  ASSERT_EQ(resultJsonMessage.getResultData().getItemsCount(), 1);
  ASSERT_EQ(resultJsonMessage.getResultData().getDataItem(0).getNodeName(), "NODE_NAME");
  ASSERT_EQ(resultJsonMessage.getResultData().getDataItem(0).getNodePort(), "NODE_PORT");
  ASSERT_EQ(resultJsonMessage.getResultData().getDataItem(0).getNodeHost(), "NODE_HOST");
  ASSERT_TRUE(resultJsonMessage.getResultData().getDataItem(0).getStdoutStream().empty());
  ASSERT_TRUE(resultJsonMessage.getResultData().getDataItem(0).getStderrStream().empty());
  ASSERT_EQ(resultJsonMessage.getResultData().getDataItem(0).getFilesCount(), 1);

  ASSERT_EQ(resultJsonMessage.getResultData().getDataItem(0).getFileItem(0).actionType, 12);
  ASSERT_EQ(resultJsonMessage.getResultData().getDataItem(0).getFileItem(0).name, "php_version.txt");
//  std::cout << "DATA: " << resultData.getDataItem(0).getFileItem(0).data << std::endl;

  printSuccess("Test Request for local Shell (php -v)");
}
//-----------------------------------------------------------------------------
void DRCEFunctionalObjectTest::testRequestLocalShellPhpScript(void)
{
  const std::string requestJson = "{\"session\":{\"type\":0,\"port\":0,\"user\":\"\",\"time_max\":0,\"tmode\":1,\"password\":\"\",\"shell\":\"\",\"environment\":[{\"LINES\":\"80\",\"HOSTNAME\":\"node22\"}]},\"command\":\"php test.php > result.txt\",\"input\":\"\",\"files\":[{\"name\":\"test.php\",\"data\":\"<?php\\necho 'Hello, world!';\\n?>\\n\",\"action\":9}, {\"name\":\"result.txt\",\"data\":\"\",\"action\":12}]}";

  DRCENodeOptions nodeOptions = DRCETaskThreadingModeTest::getNodeOptions();
  DRCEFunctionalObject fObj(nodeOptions.getNodeName(), nodeOptions.getHomeDir(), nodeOptions.getTasksDataDir(), nodeOptions.getTasksStatusDir());
  fObj.setNodeHost("NODE_HOST");
  fObj.setNodePort("NODE_PORT");
  fObj.setNodeName("NODE_NAME");
//  std::cout << "json: " << requestJson << std::endl;

  DRCEInputJsonMessage inputJsonMessage;
  inputJsonMessage.setRequestType(DRCETaskRequest::RequestType::rtSetTaskExecute);
  inputJsonMessage.setRequestData(requestJson);
  inputJsonMessage.setRequestId(2345);

  std::stringstream sourceJson;
  sourceJson << inputJsonMessage;

  ASSERT_FALSE(inputJsonMessage.isError()) << printError(inputJsonMessage);
  ASSERT_EQ(inputJsonMessage.getErrorCode(), NO_ERROR);
  ASSERT_TRUE(inputJsonMessage.getErrorMsg().empty());
  printSuccess("DRCEInputJsonMessage serialize");

//  std::cout << "IN JSON: " << sourceJson.str() << std::endl;

  std::string resultJson = fObj.Process(sourceJson.str());

  ASSERT_FALSE(fObj.isError()) << printError(fObj);
  ASSERT_EQ(fObj.getErrorCode(), NO_ERROR);
  ASSERT_TRUE(fObj.getErrorMsg().empty());
  printSuccess("DRCEFunctionalObject Process");

//  std::cout << "OUT JSON: " << resultJson << std::endl;

  // check right result (unserialize and compare two objects)
  DRCEOutputJsonMessage resultJsonMessage(resultJson);

  ASSERT_FALSE(resultJsonMessage.isError()) << printError(resultJsonMessage);
  ASSERT_EQ(resultJsonMessage.getErrorCode(), NO_ERROR);
  ASSERT_TRUE(resultJsonMessage.getErrorMsg().empty());
  printSuccess("DRCEOutputJsonMessage unserialize");

  ASSERT_EQ(resultJsonMessage.getResultData().getItemsCount(), 1);
  ASSERT_EQ(resultJsonMessage.getResultData().getDataItem(0).getNodeName(), "NODE_NAME");
  ASSERT_EQ(resultJsonMessage.getResultData().getDataItem(0).getNodePort(), "NODE_PORT");
  ASSERT_EQ(resultJsonMessage.getResultData().getDataItem(0).getNodeHost(), "NODE_HOST");
  ASSERT_TRUE(resultJsonMessage.getResultData().getDataItem(0).getStdoutStream().empty());
  ASSERT_TRUE(resultJsonMessage.getResultData().getDataItem(0).getStderrStream().empty());
  ASSERT_EQ(resultJsonMessage.getResultData().getDataItem(0).getFilesCount(), 1);

  ASSERT_EQ(resultJsonMessage.getResultData().getDataItem(0).getFileItem(0).actionType, 12);
  ASSERT_EQ(resultJsonMessage.getResultData().getDataItem(0).getFileItem(0).name, "result.txt");
  ASSERT_EQ(resultJsonMessage.getResultData().getDataItem(0).getFileItem(0).data, "Hello, world!");

  printSuccess("Test Request for local Shell (php test.php > result.txt)");
}
//-----------------------------------------------------------------------------
void DRCEFunctionalObjectTest::testRequestLocalShellPerlVersion(void)
{
  const std::string requestJson = "{\"session\":{\"type\":0,\"port\":0,\"user\":\"\",\"time_max\":0,\"tmode\":1,\"password\":\"\",\"shell\":\"\",\"environment\":[]},\"command\":\"perl -v > php_version.txt\",\"input\":\"\",\"files\":[{\"name\":\"php_version.txt\",\"data\":\"\",\"action\":12}]}";

  DRCENodeOptions nodeOptions = DRCETaskThreadingModeTest::getNodeOptions();
  DRCEFunctionalObject fObj(nodeOptions.getNodeName(), nodeOptions.getHomeDir(), nodeOptions.getTasksDataDir(), nodeOptions.getTasksStatusDir());
  fObj.setNodeHost("NODE_HOST");
  fObj.setNodePort("NODE_PORT");
  fObj.setNodeName("NODE_NAME");

  DRCEInputJsonMessage inputJsonMessage;
  inputJsonMessage.setRequestType(DRCETaskRequest::RequestType::rtSetTaskExecute);
  inputJsonMessage.setRequestData(requestJson);
  inputJsonMessage.setRequestId(2345);

  std::stringstream sourceJson;
  sourceJson << inputJsonMessage;

  ASSERT_FALSE(inputJsonMessage.isError()) << printError(inputJsonMessage);
  ASSERT_EQ(inputJsonMessage.getErrorCode(), NO_ERROR);
  ASSERT_TRUE(inputJsonMessage.getErrorMsg().empty());
  printSuccess("DRCEInputJsonMessage serialize");

  std::string resultJson = fObj.Process(sourceJson.str());

  ASSERT_FALSE(fObj.isError()) << printError(fObj);
  ASSERT_EQ(fObj.getErrorCode(), NO_ERROR);
  ASSERT_TRUE(fObj.getErrorMsg().empty());
  printSuccess("DRCEFunctionalObject Process");

//  std::cout << "JSON: " << resultJson << std::endl;
//  std::cout << "JSON: " << responseJson << std::endl;

  // check right result (unserialize and compare two objects)
  DRCEOutputJsonMessage resultJsonMessage(resultJson);

  ASSERT_FALSE(resultJsonMessage.isError()) << printError(resultJsonMessage);
  ASSERT_EQ(resultJsonMessage.getErrorCode(), NO_ERROR);
  ASSERT_TRUE(resultJsonMessage.getErrorMsg().empty());
  printSuccess("DRCEOutputJsonMessage unserialize");

  ASSERT_EQ(resultJsonMessage.getResultData().getItemsCount(), 1);
  ASSERT_EQ(resultJsonMessage.getResultData().getDataItem(0).getNodeName(), "NODE_NAME");
  ASSERT_EQ(resultJsonMessage.getResultData().getDataItem(0).getNodePort(), "NODE_PORT");
  ASSERT_EQ(resultJsonMessage.getResultData().getDataItem(0).getNodeHost(), "NODE_HOST");
  ASSERT_TRUE(resultJsonMessage.getResultData().getDataItem(0).getStdoutStream().empty());
  ASSERT_TRUE(resultJsonMessage.getResultData().getDataItem(0).getStderrStream().empty());
  ASSERT_EQ(resultJsonMessage.getResultData().getDataItem(0).getFilesCount(), 1);

  ASSERT_EQ(resultJsonMessage.getResultData().getDataItem(0).getFileItem(0).actionType, 12);
  ASSERT_EQ(resultJsonMessage.getResultData().getDataItem(0).getFileItem(0).name, "php_version.txt");
  ASSERT_FALSE(resultJsonMessage.getResultData().getDataItem(0).getFileItem(0).data.empty());

  printSuccess("Test Request for local Shell (perl -v > php_version.txt)");
}
//-----------------------------------------------------------------------------
void DRCEFunctionalObjectTest::testRequestLocalShellPerlScript(void)
{
  const std::string requestJson = "{\"session\":{\"type\":0,\"port\":0,\"user\":\"\",\"time_max\":0,\"tmode\":1,\"password\":\"\",\"shell\":\"\",\"environment\":[]},\"command\":\"perl test.pl > result.txt\",\"input\":\"\",\"files\":[{\"name\":\"test.pl\",\"data\":\"#!/usr/bin/perl\\nprint \\\"Hello, world!\\\";\",\"action\":9}, {\"name\":\"result.txt\",\"data\":\"\",\"action\":12}]}";

  DRCENodeOptions nodeOptions = DRCETaskThreadingModeTest::getNodeOptions();
  DRCEFunctionalObject fObj(nodeOptions.getNodeName(), nodeOptions.getHomeDir(), nodeOptions.getTasksDataDir(), nodeOptions.getTasksStatusDir());
  fObj.setNodeHost("NODE_HOST");
  fObj.setNodePort("NODE_PORT");
  fObj.setNodeName("NODE_NAME");

  DRCEInputJsonMessage inputJsonMessage;
  inputJsonMessage.setRequestType(DRCETaskRequest::RequestType::rtSetTaskExecute);
  inputJsonMessage.setRequestData(requestJson);
  inputJsonMessage.setRequestId(2345);

  std::stringstream sourceJson;
  sourceJson << inputJsonMessage;

  ASSERT_FALSE(inputJsonMessage.isError()) << printError(inputJsonMessage);
  ASSERT_EQ(inputJsonMessage.getErrorCode(), NO_ERROR);
  ASSERT_TRUE(inputJsonMessage.getErrorMsg().empty());
  printSuccess("DRCEInputJsonMessage serialize");

  std::string resultJson = fObj.Process(sourceJson.str());

  ASSERT_FALSE(fObj.isError()) << printError(fObj);
  ASSERT_EQ(fObj.getErrorCode(), NO_ERROR);
  ASSERT_TRUE(fObj.getErrorMsg().empty());
  printSuccess("DRCEFunctionalObject Process");

//  std::cout << "JSON: " << resultJson << std::endl;

  // check right result (unserialize and compare two objects)
  DRCEOutputJsonMessage resultJsonMessage(resultJson);

  ASSERT_FALSE(resultJsonMessage.isError()) << printError(resultJsonMessage);
  ASSERT_EQ(resultJsonMessage.getErrorCode(), NO_ERROR);
  ASSERT_TRUE(resultJsonMessage.getErrorMsg().empty());
  printSuccess("DRCEOutputJsonMessage unserialize");

  ASSERT_EQ(resultJsonMessage.getResultData().getItemsCount(), 1);
  ASSERT_EQ(resultJsonMessage.getResultData().getDataItem(0).getNodeName(), "NODE_NAME");
  ASSERT_EQ(resultJsonMessage.getResultData().getDataItem(0).getNodePort(), "NODE_PORT");
  ASSERT_EQ(resultJsonMessage.getResultData().getDataItem(0).getNodeHost(), "NODE_HOST");
  ASSERT_TRUE(resultJsonMessage.getResultData().getDataItem(0).getStdoutStream().empty());
  ASSERT_TRUE(resultJsonMessage.getResultData().getDataItem(0).getStderrStream().empty());
  ASSERT_EQ(resultJsonMessage.getResultData().getDataItem(0).getFilesCount(), 1);

  ASSERT_EQ(resultJsonMessage.getResultData().getDataItem(0).getFileItem(0).actionType, 12);
  ASSERT_EQ(resultJsonMessage.getResultData().getDataItem(0).getFileItem(0).name, "result.txt");
  ASSERT_EQ(resultJsonMessage.getResultData().getDataItem(0).getFileItem(0).data, "Hello, world!");

  printSuccess("Test Request for local Shell (perl test.pl > result.txt)");
}
//-----------------------------------------------------------------------------
//-----------------------------------------------------------------------------
} // end namespace tests
} // end namespace drce
} // end namespace HCE
