#include <gtest/gtest.h>
#include <fstream>
#include <utility>

#include "DRCEError.hpp"
#include "DRCESessionOptions.hpp"
#include "DRCEResourceLimits.hpp"
#include "DRCEInputJsonMessage.hpp"
#include "DRCETaskRequestSetExecute.hpp"
#include "DRCEOutputJsonMessage.hpp"
#include "DRCEFunctionalObject.hpp"
#include "DRCEResultDataSerializator.hpp"
#include "DRCEResultData.hpp"
#include "DRCENotificationExecutor.hpp"

#include "DRCEProtocolsTest.hpp"
#include "DRCEPrintStatus.hpp"

namespace HCE
{
namespace drce
{
namespace tests
{
//-----------------------------------------------------------------------------
void DRCEProtocolsTest::testResultDataSerializator(void)
{
  const size_t SZ = 3;
  FileItem files[SZ] = {{"file1", "data1", 2},{"file2", "data2", 4},{"file3", "data3", 8}};


  DRCEResultData resultData;
  DRCEResultDataItem resultDataItem;
  resultDataItem.setStderrStream("none");
  resultDataItem.setStdoutStream("Test message");
  for (size_t i=0;i<SZ;++i)
    resultDataItem.addFileItem(files[i].name, files[i].data, files[i].actionType);

  resultData.addDataItem(std::forward<DRCEResultDataItem>(resultDataItem));
  DRCEResultDataSerializator resultSerializator(resultData);

  std::string json;
  bool ret = resultSerializator.serialize(json);
  ASSERT_TRUE(ret) << printError(resultSerializator);

//  std::cout << "json: " << json << std::endl;

  resultData.clear();
  size_t itemsCount = resultData.getItemsCount();
  for (size_t i=0;i<itemsCount;++i)
  {
    ASSERT_TRUE(resultData.getDataItem(i).getStderrStream().empty());
    ASSERT_TRUE(resultData.getDataItem(i).getStdoutStream().empty());
    ASSERT_EQ(resultData.getDataItem(i).getFilesCount(), 0);
  }

  ret = resultSerializator.unserialize(json);
  ASSERT_TRUE(ret) << printError(resultSerializator);

  itemsCount = resultData.getItemsCount();
  ASSERT_EQ(itemsCount, 1);
  ASSERT_EQ(resultData.getDataItem(0).getStderrStream(), "none");
  ASSERT_EQ(resultData.getDataItem(0).getStdoutStream(), "Test message");
  const size_t filesCount = resultData.getDataItem(0).getFilesCount();
  ASSERT_EQ(filesCount, SZ);

  for (size_t i=0;i<SZ;++i)
  {
    ASSERT_EQ(resultData.getDataItem(0).getFileItem(i).name, files[i].name);
    ASSERT_EQ(resultData.getDataItem(0).getFileItem(i).data, files[i].data);
    ASSERT_EQ(resultData.getDataItem(0).getFileItem(i).actionType, files[i].actionType);
  }
  printSuccess("Test ResultDataSerializator");
}
//-----------------------------------------------------------------------------
void DRCEProtocolsTest::testInputJsonMessage(void)
{
  std::stringstream sourceJson, restoreJson;

  DRCETaskRequestSetExecute taskRequestSetExecute;
  SessionOptions sessionOptions;
  sessionOptions.sessionType = SessionOptions::SessionType::stSsh;
  sessionOptions.userName = "user1";
  sessionOptions.userPassword = "pass1";
  sessionOptions.shellName = "Shell";
  sessionOptions.homeDir = ".";
  sessionOptions.port = 22;
  sessionOptions.timeout = 43;
  sessionOptions.tmode = (SessionOptions::ThreadMode)5;
  sessionOptions.environments.push_back(std::make_pair("PATH", "../"));
  sessionOptions.environments.push_back(std::make_pair("LD_PATH", "./"));

  taskRequestSetExecute.setSessionOptions(std::forward<SessionOptions>(sessionOptions));
  taskRequestSetExecute.setCommandLine("cmd");
  taskRequestSetExecute.setInputStream("Stream");
  taskRequestSetExecute.addFileItem("file1", "data1", FileItem::DELETE_BEFORE_EXEC);
  taskRequestSetExecute.addFileItem("file2", "data2", FileItem::READ_AFTER_EXEC_AND_RETURN_RESULT);

  DRCEInputJsonMessage sourceMessage, restoreMessage;
/*
  sourceJson << taskRequestSetExecute;
  sourceMessage.setRequestType(DRCETaskRequest::RequestType::rtSetTaskExecute);
  sourceMessage.setRequestData(sourceJson.str());
*/
  sourceMessage.setTaskRequest(new DRCETaskRequestSetExecute(taskRequestSetExecute));
  sourceMessage.setRequestId(789);

  sourceJson.str("");
  sourceJson << sourceMessage;

//  sourceJson.str("{\"type\":0,\"data\":\"{\\\"session\\\":{\\\"type\\\":0,\\\"port\\\":0,\\\"user\\\":\\\"\\\",\\\"password\\\":\\\"\\\",\\\"shell\\\":\\\"\\\",\\\"environment\\\":[],\\\"home_dir\\\":\\\"\\\",\\\"timeout\\\":0,\\\"tmode\\\":1,\\\"time_max\\\":60000},\\\"command\\\":\\\"ls -la\\\",\\\"input\\\":\\\"\\\",\\\"files\\\":[],\\\"limits\\\":{\\\"proc_max\\\":0,\\\"threads_max\\\":0,\\\"cpu\\\":0,\\\"vram_free\\\":0,\\\"vram_free_min\\\":0,\\\"rram_free\\\":0,\\\"rram_free_min\\\":0,\\\"disk_free\\\":0,\\\"disk_free_min\\\":0}}\",\"id\":12345}");

  ASSERT_FALSE(sourceMessage.isError()) << printError(sourceMessage);
  ASSERT_EQ(sourceMessage.getErrorCode(), NO_ERROR);
  ASSERT_TRUE(sourceMessage.getErrorMsg().empty());
  printSuccess("DRCEInputJsonMessage serialize");

  sourceJson >> restoreMessage;  // unserialize

  ASSERT_FALSE(restoreMessage.isError()) << printError(restoreMessage);
  ASSERT_EQ(restoreMessage.getErrorCode(), NO_ERROR);
  ASSERT_TRUE(restoreMessage.getErrorMsg().empty());
  printSuccess("DRCEInputJsonMessage unserialize");

  restoreJson << restoreMessage;  // serialize
  ASSERT_FALSE(restoreMessage.isError()) << printError(restoreMessage);
//  std::cout << "json: " << sourceJson.str() << std::endl;
//  std::cout << "json: " << restoreJson.str() << std::endl;

  ASSERT_EQ(sourceMessage.getRequestType(), restoreMessage.getRequestType());
  ASSERT_EQ(sourceMessage.getRequestData(), restoreMessage.getRequestData());
  ASSERT_EQ(sourceMessage.getRequestId(), restoreMessage.getRequestId());
  ASSERT_EQ(sourceJson.str(), restoreJson.str());
  printSuccess("Test InputJsonMessage");
}
//-----------------------------------------------------------------------------
void DRCEProtocolsTest::testTaskExecuteRequest(void)
{
  std::stringstream sourceJson, restoreJson;
  DRCETaskRequestSetExecute sourceMessage, restoreMessage;

  SessionOptions sessionOptions;
  sessionOptions.sessionType = SessionOptions::SessionType::stSsh;
  sessionOptions.userName = "user1";
  sessionOptions.userPassword = "pass1";
  sessionOptions.shellName = "Shell";
  sessionOptions.homeDir = ".";
  sessionOptions.port = 22;
  sessionOptions.timeout = 43;
  sessionOptions.tmode = (SessionOptions::ThreadMode)5;
  sessionOptions.environments.push_back(std::make_pair("PATH", "../"));
  sessionOptions.environments.push_back(std::make_pair("LD_PATH", "./"));

  sourceMessage.setSessionOptions(std::forward<SessionOptions>(sessionOptions));
  sourceMessage.setCommandLine("cmd");
  sourceMessage.setInputStream("Stream");
  sourceMessage.addFileItem("file1", "data1", FileItem::DELETE_BEFORE_EXEC);
  sourceMessage.addFileItem("file2", "data2", FileItem::READ_AFTER_EXEC_AND_RETURN_RESULT);

  sourceJson << sourceMessage;

  ASSERT_FALSE(sourceMessage.isError()) << printError(sourceMessage);
  ASSERT_EQ(sourceMessage.getErrorCode(), NO_ERROR);
  ASSERT_TRUE(sourceMessage.getErrorMsg().empty());
  printSuccess("DRCETaskRequestSetExecute serialize");

  sourceJson >> restoreMessage;  // unserialize

  ASSERT_FALSE(restoreMessage.isError()) << printError(restoreMessage);
  ASSERT_EQ(restoreMessage.getErrorCode(), NO_ERROR);
  ASSERT_TRUE(restoreMessage.getErrorMsg().empty());
  printSuccess("DRCETaskRequestSetExecute unserialize");

  restoreJson << restoreMessage;  // serialize
  ASSERT_FALSE(restoreMessage.isError()) << printError(restoreMessage);
//  std::cout << "json: " << sourceJson.str() << std::endl;
//  std::cout << "json: " << restoreJson.str() << std::endl;

  ASSERT_EQ(sourceMessage.getSessionOptions().sessionType, restoreMessage.getSessionOptions().sessionType);
  ASSERT_EQ(sourceMessage.getSessionOptions().port, restoreMessage.getSessionOptions().port);
  ASSERT_EQ(sourceMessage.getSessionOptions().userName, restoreMessage.getSessionOptions().userName);
  ASSERT_EQ(sourceMessage.getSessionOptions().userPassword, restoreMessage.getSessionOptions().userPassword);
  ASSERT_EQ(sourceMessage.getSessionOptions().shellName, restoreMessage.getSessionOptions().shellName);
  ASSERT_EQ(sourceMessage.getSessionOptions().homeDir, restoreMessage.getSessionOptions().homeDir);
  ASSERT_EQ(sourceMessage.getSessionOptions().timeout, restoreMessage.getSessionOptions().timeout);
  ASSERT_EQ(sourceMessage.getSessionOptions().tmode, restoreMessage.getSessionOptions().tmode);
  ASSERT_EQ(sourceMessage.getSessionOptions().environments.size(), restoreMessage.getSessionOptions().environments.size());

  ASSERT_EQ(sourceMessage.getCommandLine(), restoreMessage.getCommandLine());
  ASSERT_EQ(sourceMessage.getInputStream(), restoreMessage.getInputStream());
  ASSERT_EQ(sourceMessage.getFilesCount(), restoreMessage.getFilesCount());
  const size_t filesCount = sourceMessage.getFilesCount();
  for (size_t i=0;i<filesCount;++i)
  {
    ASSERT_EQ(sourceMessage.getFileItem(i).name, restoreMessage.getFileItem(i).name);
    ASSERT_EQ(sourceMessage.getFileItem(i).data, restoreMessage.getFileItem(i).data);
    ASSERT_EQ(sourceMessage.getFileItem(i).actionType, restoreMessage.getFileItem(i).actionType);
  }
  ASSERT_EQ(sourceJson.str(), restoreJson.str());
  printSuccess("Test DRCETaskRequestSetExecute");
}
//-----------------------------------------------------------------------------
void DRCEProtocolsTest::testOutputJsonMessage(void)
{
  std::stringstream sourceJson, restoreJson;
  DRCEOutputJsonMessage sourceMessage, restoreMessage;

  DRCEResultData resultData;
  DRCEResultDataItem resultDataItem;
  resultDataItem.setErrorCode(ERROR_FUNCTIONAL_OBJECT);
  resultDataItem.setErrorMessage("Error message");
  resultDataItem.setTime(43);
  resultDataItem.setRequestId(12345);
  resultDataItem.setRequestType(DRCETaskRequest::RequestType::rtTerminateTask);
  resultDataItem.setNodeHost("node_host");
  resultDataItem.setNodePort("node_port");
  resultDataItem.setState(45);
  resultDataItem.setPid(10101);
  resultDataItem.setStdoutStream("stdout stream");
  resultDataItem.setStderrStream("stderr stream");
  resultDataItem.addFileItem("file1", "data1", FileItem::DELETE_BEFORE_EXEC);
  resultDataItem.addFileItem("file2", "data2", FileItem::READ_AFTER_EXEC_AND_RETURN_RESULT);
  resultData.addDataItem(std::forward<DRCEResultDataItem>(resultDataItem));
  sourceMessage.setResultData(resultData);

  std::string json;
  bool ret = sourceMessage.serialize(json);
  ASSERT_TRUE(ret) << printError(sourceMessage);
  ASSERT_FALSE(sourceMessage.isError()) << printError(sourceMessage);
  ASSERT_TRUE(sourceMessage.getErrorMsg().empty());
  ASSERT_EQ(sourceMessage.getErrorCode(), NO_ERROR);
  printSuccess("DRCEOutputJsonMessage serialize");

  sourceJson.str(json);
  sourceJson >> restoreMessage;  // unserialize

  ASSERT_FALSE(restoreMessage.isError()) << printError(restoreMessage);
  ASSERT_TRUE(restoreMessage.getErrorMsg().empty());
  ASSERT_EQ(restoreMessage.getErrorCode(), NO_ERROR);
  printSuccess("DRCEOutputJsonMessage unserialize");

  restoreJson << restoreMessage;  // serialize
  ASSERT_FALSE(restoreMessage.isError()) << printError(restoreMessage);

//  std::cout << "json: " << sourceJson.str() << std::endl;
//  std::cout << "json: " << restoreJson.str() << std::endl;

  ASSERT_EQ(sourceMessage.getResultData().getItemsCount(), 1);
  ASSERT_EQ(sourceMessage.getResultData().getItemsCount(), restoreMessage.getResultData().getItemsCount());
  ASSERT_EQ(sourceMessage.getResultData().getDataItem(0).getErrorCode(), restoreMessage.getResultData().getDataItem(0).getErrorCode());
  ASSERT_EQ(sourceMessage.getResultData().getDataItem(0).getErrorMessage(), restoreMessage.getResultData().getDataItem(0).getErrorMessage());
  ASSERT_EQ(sourceMessage.getResultData().getDataItem(0).getTime(), restoreMessage.getResultData().getDataItem(0).getTime());
  ASSERT_EQ(sourceMessage.getResultData().getDataItem(0).getRequestId(), restoreMessage.getResultData().getDataItem(0).getRequestId());
  ASSERT_EQ(sourceMessage.getResultData().getDataItem(0).getRequestType(), restoreMessage.getResultData().getDataItem(0).getRequestType());
  ASSERT_EQ(sourceMessage.getResultData().getDataItem(0).getNodeHost(), restoreMessage.getResultData().getDataItem(0).getNodeHost());
  ASSERT_EQ(sourceMessage.getResultData().getDataItem(0).getNodePort(), restoreMessage.getResultData().getDataItem(0).getNodePort());
  ASSERT_EQ(sourceMessage.getResultData().getDataItem(0).getState(), restoreMessage.getResultData().getDataItem(0).getState());
  ASSERT_EQ(sourceMessage.getResultData().getDataItem(0).getStdoutStream(), restoreMessage.getResultData().getDataItem(0).getStdoutStream());
  ASSERT_EQ(sourceMessage.getResultData().getDataItem(0).getStderrStream(), restoreMessage.getResultData().getDataItem(0).getStderrStream());

  ASSERT_EQ(sourceMessage.getResultData().getDataItem(0).getFilesCount(), 2);
  ASSERT_EQ(sourceMessage.getResultData().getDataItem(0).getFilesCount(), restoreMessage.getResultData().getDataItem(0).getFilesCount());
  ASSERT_EQ(sourceMessage.getResultData().getDataItem(0).getFileItem(0).name, restoreMessage.getResultData().getDataItem(0).getFileItem(0).name);
  ASSERT_EQ(sourceMessage.getResultData().getDataItem(0).getFileItem(0).data, restoreMessage.getResultData().getDataItem(0).getFileItem(0).data);
  ASSERT_EQ(sourceMessage.getResultData().getDataItem(0).getFileItem(0).actionType, restoreMessage.getResultData().getDataItem(0).getFileItem(0).actionType);
  ASSERT_EQ(sourceMessage.getResultData().getDataItem(0).getFileItem(1).name, restoreMessage.getResultData().getDataItem(0).getFileItem(1).name);
  ASSERT_EQ(sourceMessage.getResultData().getDataItem(0).getFileItem(1).data, restoreMessage.getResultData().getDataItem(0).getFileItem(1).data);
  ASSERT_EQ(sourceMessage.getResultData().getDataItem(0).getFileItem(1).actionType, restoreMessage.getResultData().getDataItem(0).getFileItem(1).actionType);

  ASSERT_EQ(sourceJson.str(), restoreJson.str());
  printSuccess("Test OutputJsonMessage");
}
//-----------------------------------------------------------------------------
void DRCEProtocolsTest::testSessionOptions(void)
{
  DRCETaskRequestSetExecute sourceMessage, restoreMessage;

  std::stringstream sourceJson;
  SessionOptions sourceOptions, restoreOptions;
  sourceOptions.sessionType = SessionOptions::SessionType::stSsh;
  sourceOptions.port = 22;
  sourceOptions.userName = "user";
  sourceOptions.userPassword = "123";
  sourceOptions.shellName = "Shell";
  sourceOptions.homeDir = ".";
  sourceOptions.timeout = 43;
  sourceOptions.tmode = (SessionOptions::ThreadMode)5;
  sourceOptions.timeMax = 500;
  sourceOptions.environments.push_back(std::make_pair("PATH", "./"));
  sourceOptions.environments.push_back(std::make_pair("LD_PATH", "../"));

  sourceMessage.setSessionOptions(std::forward<SessionOptions>(sourceOptions));

  sourceJson << sourceMessage;  // serialize

  ASSERT_FALSE(sourceMessage.isError()) << printError(sourceMessage);
  ASSERT_EQ(sourceMessage.getErrorCode(), NO_ERROR);
  ASSERT_TRUE(sourceMessage.getErrorMsg().empty());
  printSuccess("DRCETaskRequestSetExecute serialize");

  sourceJson >> restoreMessage;  // unserialize

  ASSERT_FALSE(restoreMessage.isError()) << printError(restoreMessage);
  ASSERT_EQ(restoreMessage.getErrorCode(), NO_ERROR);
  ASSERT_TRUE(restoreMessage.getErrorMsg().empty());
  printSuccess("DRCETaskRequestSetExecute unserialize");

  ASSERT_EQ(restoreMessage.getSessionOptions().sessionType, SessionOptions::SessionType::stSsh);
  ASSERT_EQ(restoreMessage.getSessionOptions().port, 22);
  ASSERT_EQ(restoreMessage.getSessionOptions().userName, "user");
  ASSERT_EQ(restoreMessage.getSessionOptions().userPassword, "123");
  ASSERT_EQ(restoreMessage.getSessionOptions().shellName, "Shell");
  ASSERT_EQ(restoreMessage.getSessionOptions().homeDir, ".");
  ASSERT_EQ(restoreMessage.getSessionOptions().timeout, 43);
  ASSERT_EQ(restoreMessage.getSessionOptions().tmode, (SessionOptions::ThreadMode)5);
  ASSERT_EQ(restoreMessage.getSessionOptions().timeMax, 500);
  ASSERT_EQ(restoreMessage.getSessionOptions().environments.size(), 2);
  ASSERT_EQ(restoreMessage.getSessionOptions().environments[0].first, "PATH");
  ASSERT_EQ(restoreMessage.getSessionOptions().environments[0].second, "./");
  ASSERT_EQ(restoreMessage.getSessionOptions().environments[1].first, "LD_PATH");
  ASSERT_EQ(restoreMessage.getSessionOptions().environments[1].second, "../");

  printSuccess("Test SessionOptions");
}
//-----------------------------------------------------------------------------
void DRCEProtocolsTest::testResourceLimits(void)
{
  DRCETaskRequestSetExecute sourceMessage, restoreMessage;

  std::stringstream sourceJson;
  ResourceLimits sourceLimits, restoreLimits;
  sourceLimits.procMax = 4;
  sourceLimits.threadsMax = 3;
  sourceLimits.cpuMax = 2;
  sourceLimits.vramFree = 125000;
  sourceLimits.vramFreeMin = 75;
  sourceLimits.rramFree = 512000;
  sourceLimits.rramFreeMin = 25;
  sourceLimits.diskFree = 1024000;
  sourceLimits.diskFreeMin = 50;
  sourceMessage.setResourceLimits(std::forward<ResourceLimits>(sourceLimits));

  sourceJson << sourceMessage;  // serialize

  ASSERT_FALSE(sourceMessage.isError()) << printError(sourceMessage);
  ASSERT_EQ(sourceMessage.getErrorCode(), NO_ERROR);
  ASSERT_TRUE(sourceMessage.getErrorMsg().empty());
  printSuccess("DRCETaskRequestSetExecute serialize");

  sourceJson >> restoreMessage;  // unserialize

  ASSERT_FALSE(restoreMessage.isError()) << printError(restoreMessage);
  ASSERT_EQ(restoreMessage.getErrorCode(), NO_ERROR);
  ASSERT_TRUE(restoreMessage.getErrorMsg().empty());
  printSuccess("DRCETaskRequestSetExecute unserialize");

  ASSERT_EQ(restoreMessage.getResourceLimits().procMax, 4);
  ASSERT_EQ(restoreMessage.getResourceLimits().threadsMax, 3);
  ASSERT_EQ(restoreMessage.getResourceLimits().cpuMax, 2);
  ASSERT_EQ(restoreMessage.getResourceLimits().vramFree, 125000);
  ASSERT_EQ(restoreMessage.getResourceLimits().vramFreeMin, 75);
  ASSERT_EQ(restoreMessage.getResourceLimits().rramFree, 512000);
  ASSERT_EQ(restoreMessage.getResourceLimits().rramFreeMin, 25);
  ASSERT_EQ(restoreMessage.getResourceLimits().diskFree, 1024000);
  ASSERT_EQ(restoreMessage.getResourceLimits().diskFreeMin, 50);

  printSuccess("Test ResourceLimits");}
//-----------------------------------------------------------------------------
void DRCEProtocolsTest::testFilesItemList(void)
{
  DRCETaskRequestSetExecute sourceMessage, restoreMessage;

  std::stringstream sourceJson;
  DRCEFilesList sourceFilesList, restoreFilesList;

  sourceMessage.addFileItem("name1", "data1", FileItem::CREATE_BEFORE_EXEC_AND_WRITE_DATA);
  sourceMessage.addFileItem("name2", "data2",
      FileItem::CREATE_BEFORE_EXEC_AND_WRITE_DATA | FileItem::DELETE_BEFORE_EXEC);
  sourceMessage.addFileItem("name3", "data3",
        FileItem::CREATE_BEFORE_EXEC_AND_WRITE_DATA | FileItem::DELETE_BEFORE_EXEC |
        FileItem::READ_AFTER_EXEC_AND_RETURN_RESULT);
  sourceMessage.addFileItem("name4", "data4",
        FileItem::CREATE_BEFORE_EXEC_AND_WRITE_DATA | FileItem::DELETE_BEFORE_EXEC |
        FileItem::READ_AFTER_EXEC_AND_RETURN_RESULT | FileItem::DELETE_AFTER_EXEC);

  sourceJson << sourceMessage;  // serialize

  ASSERT_FALSE(sourceMessage.isError()) << printError(sourceMessage);
  ASSERT_EQ(sourceMessage.getErrorCode(), NO_ERROR);
  ASSERT_TRUE(sourceMessage.getErrorMsg().empty());
  printSuccess("DRCETaskRequestSetExecute serialize");

  sourceJson >> restoreMessage;  // unserialize

  ASSERT_FALSE(restoreMessage.isError()) << printError(restoreMessage);
  ASSERT_EQ(restoreMessage.getErrorCode(), NO_ERROR);
  ASSERT_TRUE(restoreMessage.getErrorMsg().empty());
  printSuccess("DRCETaskRequestSetExecute unserialize");

  ASSERT_EQ(restoreMessage.getFilesCount(), 4);
  ASSERT_EQ(restoreMessage.getFileItem(0).name, "name1");
  ASSERT_EQ(restoreMessage.getFileItem(0).data, "data1");
  ASSERT_EQ(restoreMessage.getFileItem(0).actionType, FileItem::CREATE_BEFORE_EXEC_AND_WRITE_DATA);

  ASSERT_EQ(restoreMessage.getFileItem(1).name, "name2");
  ASSERT_EQ(restoreMessage.getFileItem(1).data, "data2");
  ASSERT_EQ(restoreMessage.getFileItem(1).actionType, (FileItem::CREATE_BEFORE_EXEC_AND_WRITE_DATA | FileItem::DELETE_BEFORE_EXEC));

  ASSERT_EQ(restoreMessage.getFileItem(2).name, "name3");
  ASSERT_EQ(restoreMessage.getFileItem(2).data, "data3");
  ASSERT_EQ(restoreMessage.getFileItem(2).actionType, (FileItem::CREATE_BEFORE_EXEC_AND_WRITE_DATA |
      FileItem::DELETE_BEFORE_EXEC | FileItem::READ_AFTER_EXEC_AND_RETURN_RESULT));

  ASSERT_EQ(restoreMessage.getFileItem(3).name, "name4");
  ASSERT_EQ(restoreMessage.getFileItem(3).data, "data4");
  ASSERT_EQ(restoreMessage.getFileItem(3).actionType, (FileItem::CREATE_BEFORE_EXEC_AND_WRITE_DATA |
      FileItem::DELETE_BEFORE_EXEC | FileItem::READ_AFTER_EXEC_AND_RETURN_RESULT | FileItem::DELETE_AFTER_EXEC));

  printSuccess("Test FilesItemList");
}
//-----------------------------------------------------------------------------
//-----------------------------------------------------------------------------
} // end namespace tests
} // end namespace drce
} // end namespace HCE
