#include <gtest/gtest.h>
#include <sstream>
#include <fstream>
#include <vector>
#include <limits>
#include <Poco/File.h>

#include "DRCENotificationExecutor.hpp"
#include "DRCEStressTest.hpp"
#include "DRCEPrintStatus.hpp"
#include "DRCECleanupTest.hpp"
#include "DRCEFileStream.hpp"
#include "DRCECommonTask.hpp"

namespace HCE
{
namespace drce
{
namespace tests
{
//-----------------------------------------------------------------------------
void DRCEStressTest::testMultipleRequests(DRCEFunctionalObject& fObj, SessionOptions::SessionType sessionType, unsigned int requestCount)
{
  fObj.setMaxThreadCount(requestCount);
  fObj.setResourceMonitorTimePeriod(10000);
  fObj.setLogStream(std::cout);

  std::vector<unsigned int> tasks;
  std::vector<std::string> jsons;

  unsigned int taskId(0), caseNumber(0);
  for (;taskId<requestCount;)
  {
    jsons.push_back(std::forward<std::string>(DRCEStressTest::getJsonRequestSetExecuteAsyncLongTask(sessionType, ++taskId)));
    tasks.push_back(taskId);

    jsons.push_back(std::forward<std::string>(DRCEStressTest::getJsonRequestSetExecuteAsyncExpitedTask(sessionType, ++taskId)));
    jsons.push_back(std::forward<std::string>(DRCEStressTest::getJsonRequestSetExecuteAsyncShortTask(sessionType, ++taskId)));
    jsons.push_back(std::forward<std::string>(DRCEStressTest::getJsonRequestSetExecuteSyncShortTask(sessionType, ++taskId)));

    jsons.push_back(std::forward<std::string>(DRCEStressTest::generateJsonRequest(caseNumber++, tasks.back())));
    if (caseNumber > 7)
      caseNumber = 0;
  }

  for (size_t i=0;i<tasks.size();++i)
  {
    jsons.push_back(std::forward<std::string>(DRCEStressTest::generateJsonRequest(caseNumber++, tasks.back())));
    if (caseNumber > 7)
      caseNumber = 0;
  }

  for (size_t i=0;i<jsons.size();++i)
  {
    DRCEStressTest::executeRequest(fObj, jsons[i]);
  }
  while(fObj.getAsyncTasksCount())
  {
    Poco::Thread::sleep(1000);
  }
  DRCECleanupTest::cleanupTestData(fObj);
  printSuccess ("Test Multiple Requests");
}
//-----------------------------------------------------------------------------
void DRCEStressTest::executeRequest(DRCEFunctionalObject& fObj, const std::string& json)
{
  std::string resultJson = fObj.Process(json);

  ASSERT_FALSE(fObj.isError()) << printError(fObj);
  ASSERT_EQ(fObj.getErrorCode(), NO_ERROR);
  ASSERT_TRUE(fObj.getErrorMsg().empty());

  DRCEOutputJsonMessage resultJsonMessage(resultJson);
  ASSERT_FALSE(resultJsonMessage.isError()) << printError(resultJsonMessage);
  ASSERT_EQ(resultJsonMessage.getErrorCode(), NO_ERROR);
  ASSERT_TRUE(resultJsonMessage.getErrorMsg().empty());
}
//-----------------------------------------------------------------------------
std::string DRCEStressTest::generateJsonRequest(unsigned int caseNumber, unsigned int requestId)
{
  std::string json;
  switch(caseNumber)
  {
    case 0: json = DRCEStressTest::getJsonRequestCheckState(DRCETaskRequestCheckState::CheckType::ctUninitialized, requestId);
      break;
    case 1: json = DRCEStressTest::getJsonRequestCheckState(DRCETaskRequestCheckState::CheckType::ctSimple, requestId);
      break;
    case 2: json = DRCEStressTest::getJsonRequestCheckState(DRCETaskRequestCheckState::CheckType::ctExtended, requestId);
      break;
    case 3: json = DRCEStressTest::getJsonRequestGetData(DRCETaskRequestGetData::FetchType::ftUninitialized, requestId);
      break;
    case 4: json = DRCEStressTest::getJsonRequestGetData(DRCETaskRequestGetData::FetchType::ftDeleteDataAfterFetch, requestId);
      break;
    case 5: json = DRCEStressTest::getJsonRequestGetData(DRCETaskRequestGetData::FetchType::ftNotDeleteDataAfterFetch, requestId);
      break;
    case 6: json = DRCEStressTest::getJsonRequestTerminate(DRCETaskRequestTerminate::AlgorithmType::atDefault, requestId);
      break;
    case 7: json = DRCEStressTest::getJsonRequestTerminate(DRCETaskRequestTerminate::AlgorithmType::atCustom, requestId);
      break;
    default:;
  }
  return json;
}
//-----------------------------------------------------------------------------
//-----------------------------------------------------------------------------
std::string DRCEStressTest::getJsonRequestSetExecute(DRCETaskRequestSetExecute& requestSetExecute, unsigned int requestId)
{
  std::stringstream sourceJson;
  sourceJson << requestSetExecute;

//  std::cout << "\nIN JSON: " << sourceJson.str() << std::endl;

  DRCEInputJsonMessage inputJsonMessage;
  inputJsonMessage.setRequestType(DRCETaskRequest::RequestType::rtSetTaskExecute);
  inputJsonMessage.setRequestData(sourceJson.str());
  inputJsonMessage.setRequestId(requestId);

  sourceJson.str("");
  sourceJson << inputJsonMessage;

  return sourceJson.str();
}
//-----------------------------------------------------------------------------
std::string DRCEStressTest::getJsonRequestSetExecuteAsyncExpitedTask(SessionOptions::SessionType sessionType,unsigned int requestId)
{
  const std::string command = "echo \"scale=4000; 4*a(1)\" | bc -l && sleep 60 && ps xau";

  DRCETaskRequestSetExecute requestSetExecute;

  SessionOptions sessionOptions;
  sessionOptions.sessionType = sessionType;
  sessionOptions.timeMax = 10000;  // msec
  sessionOptions.tmode = SessionOptions::ThreadMode::tmAsync;
  requestSetExecute.setSessionOptions(std::forward<SessionOptions>(sessionOptions));
  requestSetExecute.setCommandLine(command);

  return DRCEStressTest::getJsonRequestSetExecute(requestSetExecute, requestId);
}
//-----------------------------------------------------------------------------
std::string DRCEStressTest::getJsonRequestSetExecuteAsyncLongTask(SessionOptions::SessionType sessionType, unsigned int requestId)
{
  const std::string command = "echo \"scale=4000; 4*a(1)\" | bc -l && ps xau";

  DRCETaskRequestSetExecute requestSetExecute;

  SessionOptions sessionOptions;
  sessionOptions.sessionType = sessionType;
  sessionOptions.timeMax = 100000;  // msec
  sessionOptions.tmode = SessionOptions::ThreadMode::tmAsync;
  requestSetExecute.setSessionOptions(std::forward<SessionOptions>(sessionOptions));
  requestSetExecute.setCommandLine(command);

  return DRCEStressTest::getJsonRequestSetExecute(requestSetExecute, requestId);
}
//-----------------------------------------------------------------------------
std::string DRCEStressTest::getJsonRequestSetExecuteAsyncShortTask(SessionOptions::SessionType sessionType, unsigned int requestId)
{
  const std::string command = "sort";
  const std::string inputStream = "def\nabc\n123\n";

  DRCETaskRequestSetExecute requestSetExecute;

  SessionOptions sessionOptions;
  sessionOptions.sessionType = sessionType;
  sessionOptions.timeMax = 100000;  // msec
  sessionOptions.tmode = SessionOptions::ThreadMode::tmAsync;
  requestSetExecute.setSessionOptions(std::forward<SessionOptions>(sessionOptions));

  ResourceLimits resourceLimits;
  resourceLimits.procMax = 500;
  resourceLimits.rramFreeMin = 100000;
  requestSetExecute.setResourceLimits(resourceLimits);

  requestSetExecute.setCommandLine(command);
  requestSetExecute.setInputStream(inputStream);

  return DRCEStressTest::getJsonRequestSetExecute(requestSetExecute, requestId);
}
//-----------------------------------------------------------------------------
std::string DRCEStressTest::getJsonRequestSetExecuteSyncShortTask(SessionOptions::SessionType sessionType, unsigned int requestId)
{
  const std::string command = "sort";
  const std::string inputStream = "def\nabc\n123\n";

  DRCETaskRequestSetExecute requestSetExecute;

  SessionOptions sessionOptions;
  sessionOptions.sessionType = sessionType;
  sessionOptions.timeMax = 100000;  // msec
  sessionOptions.tmode = SessionOptions::ThreadMode::tmSync;
  requestSetExecute.setSessionOptions(std::forward<SessionOptions>(sessionOptions));
  requestSetExecute.setCommandLine(command);
  requestSetExecute.setInputStream(inputStream);

  return DRCEStressTest::getJsonRequestSetExecute(requestSetExecute, requestId);
}
//-----------------------------------------------------------------------------
//-----------------------------------------------------------------------------
std::string DRCEStressTest::getJsonRequestCheckState(DRCETaskRequestCheckState::CheckType checkType, unsigned int requestId)
{
  DRCETaskRequestCheckState requestCheckState;
  requestCheckState.setCheckType(checkType);

  std::stringstream sourceJson;
  sourceJson << requestCheckState;

  DRCEInputJsonMessage inputJsonMessage;
  inputJsonMessage.setRequestType(DRCETaskRequest::RequestType::rtCheckTaskState);
  inputJsonMessage.setRequestData(sourceJson.str());
  inputJsonMessage.setRequestId(requestId);

  sourceJson.str("");
  sourceJson << inputJsonMessage;

  return sourceJson.str();
}
//-----------------------------------------------------------------------------
std::string DRCEStressTest::getJsonRequestGetData(DRCETaskRequestGetData::FetchType fetchType, unsigned int requestId)
{
  DRCETaskRequestGetData requestGetData;
  requestGetData.setFetchType(fetchType);

  std::stringstream sourceJson;
  sourceJson << requestGetData;

  DRCEInputJsonMessage inputJsonMessage;
  inputJsonMessage.setRequestType(DRCETaskRequest::RequestType::rtGetTaskData);
  inputJsonMessage.setRequestData(sourceJson.str());
  inputJsonMessage.setRequestId(requestId);

  sourceJson.str("");
  sourceJson << inputJsonMessage;

  return sourceJson.str();
}
//-----------------------------------------------------------------------------
std::string DRCEStressTest::getJsonRequestTerminate(DRCETaskRequestTerminate::AlgorithmType algorithmType, unsigned int requestId)
{
  DRCETaskRequestTerminate requestTerminate;
  requestTerminate.setTaskId(requestId);
  requestTerminate.setAlgorithmType(algorithmType);
  requestTerminate.setDelayValue(1000);
  requestTerminate.setRepeatValue(3);
  requestTerminate.setSignalValue(9);

  std::stringstream sourceJson;
  sourceJson << requestTerminate;

  // set request for execute task
  DRCEInputJsonMessage inputJsonMessage;
  inputJsonMessage.setRequestType(DRCETaskRequest::RequestType::rtTerminateTask);
  inputJsonMessage.setRequestData(sourceJson.str());
  inputJsonMessage.setRequestId(requestId);

  sourceJson.str("");
  sourceJson << inputJsonMessage;

  return sourceJson.str();
}
//-----------------------------------------------------------------------------
std::string DRCEStressTest::getJsonRequestResourceLimits(SessionOptions::SessionType sessionType,
    SessionOptions::ThreadMode threadMode, unsigned int requestId, const ResourceLimits& resourceLimits, const std::string& commandLine)
{
  const std::string command = (commandLine.empty())?("echo \"scale=2000; 4*a(1)\" | bc -l && ps xau"):commandLine;

  DRCETaskRequestSetExecute requestSetExecute;

  SessionOptions sessionOptions;
  sessionOptions.sessionType = sessionType;
  sessionOptions.timeMax = 1000000;  // msec
  sessionOptions.tmode = threadMode;
  requestSetExecute.setSessionOptions(std::forward<SessionOptions>(sessionOptions));
  requestSetExecute.setCommandLine(command);
  requestSetExecute.setResourceLimits(resourceLimits);

  return DRCEStressTest::getJsonRequestSetExecute(requestSetExecute, requestId);
}
//-----------------------------------------------------------------------------
//-----------------------------------------------------------------------------
unsigned int DRCEStressTest::getRequestTime(DRCEFunctionalObject& fObj, unsigned int requestId)
{
  unsigned int elapsedTime = 0;
  std::string resultJson = fObj.Process(DRCEStressTest::getJsonRequestGetData(DRCETaskRequestGetData::FetchType::ftNotDeleteDataAfterFetch, requestId));

  DRCEOutputJsonMessage resultJsonMessage(resultJson);
  if (resultJsonMessage.getResultData().getItemsCount()==1)
    elapsedTime = resultJsonMessage.getResultData().getDataItem(0).getTime();

  return elapsedTime;
}
//-----------------------------------------------------------------------------
void DRCEStressTest::testRequestsTimeDelay(DRCEFunctionalObject& fObj, SessionOptions::SessionType sessionType, unsigned int requestCount, unsigned int& taskId,
    std::map<unsigned int, unsigned int>& delayTimes, std::map<unsigned int, unsigned int>& times)
{
  delayTimes.clear();
  times.clear();
  fObj.setMaxThreadCount(requestCount);
  fObj.setResourceMonitorTimePeriod(10000);
//  fObj.setLogStream(std::cout);

  for (;taskId<requestCount;)
  {
    Poco::Timestamp now;
    std::string resultJson = fObj.Process(DRCEStressTest::getJsonRequestSetExecuteAsyncLongTask(sessionType, ++taskId));

    ASSERT_FALSE(fObj.isError()) << printError(fObj);
    ASSERT_EQ(fObj.getErrorCode(), NO_ERROR);
    ASSERT_TRUE(fObj.getErrorMsg().empty());

    DRCEOutputJsonMessage resultJsonMessage(resultJson);
    ASSERT_FALSE(resultJsonMessage.isError()) << printError(resultJsonMessage);
    ASSERT_EQ(resultJsonMessage.getErrorCode(), NO_ERROR);
    ASSERT_TRUE(resultJsonMessage.getErrorMsg().empty());

    Poco::Timestamp::TimeDiff diff = now.elapsed();
    delayTimes[taskId] = diff;

//    ASSERT_EQ(resultJsonMessage.getResultData().getItemsCount(), 1);
//    delayTimes[taskId] = resultJsonMessage.getResultData().getDataItem(0).getTime();
  }

  while(fObj.getAsyncTasksCount())
  {
    Poco::Thread::sleep(5000);
  }

  for (size_t i=0;i<requestCount;++i)
  {
    times[i+1] = DRCEStressTest::getRequestTime(fObj, i+1);
  }
  DRCECleanupTest::cleanupTestData(fObj);
}
//-----------------------------------------------------------------------------
void DRCEStressTest::testRequestsTimeDelay(DRCEFunctionalObject& fObj, SessionOptions::SessionType sessionType, unsigned int requestCount, unsigned int repeat)
{
  const unsigned int SZ = repeat;
  unsigned int requestCounts[SZ];
  for (unsigned int i=0;i<SZ;++i)
  {
    requestCounts[i] = static_cast<double>(requestCount/(i+1));
  }
//  unsigned int requestCounts[SZ] = {requestCount, requestCount/2, requestCount/3, requestCount/4, requestCount/5};
  std::map<unsigned int, unsigned int> delayTimes[SZ];
  std::map<unsigned int, unsigned int> times[SZ];
  unsigned int taskId(0);

  for (unsigned int i=0;i<SZ;++i)
  {
    DRCEStressTest::testRequestsTimeDelay(fObj, sessionType, requestCounts[i], taskId, delayTimes[i], times[i]);
  }

  while(fObj.getAsyncTasksCount())
  {
    Poco::Thread::sleep(1000);
  }

  std::ofstream ofs("timeDelay.log", std::fstream::trunc);
  ofs << "time delay (in microseconds):\n";
  for (unsigned int i=0;i<SZ;++i)
  {
    for (size_t k=0;k<requestCount;++k)
      ofs << delayTimes[i][k+1] << "\t";
    ofs << std::endl;
  }
  ofs << "\nfull time (in msec):\n";
  for (unsigned int i=0;i<SZ;++i)
  {
    for (size_t k=0;k<requestCount;++k)
      ofs << times[i][k+1] << "\t";
    ofs << std::endl;
  }
  ofs.close();

  DRCECleanupTest::cleanupTestData(fObj);
  printSuccess ("Test Requests Time Delay");
}
//-----------------------------------------------------------------------------
//-----------------------------------------------------------------------------
void DRCEStressTest::testResourceLimits(DRCEFunctionalObject& fObj,
    SessionOptions::SessionType sessionType,
    SessionOptions::ThreadMode threadMode,
    const ResourceLimits& resourceLimits,
    unsigned int& taskId,
    bool exceededLimits,
    const std::string& commandLine)
{
  std::string resultJson = fObj.Process(DRCEStressTest::getJsonRequestResourceLimits(sessionType, threadMode, ++taskId, resourceLimits, commandLine));

//  std::cout << "resultJson: " << resultJson << std::endl;

  ASSERT_FALSE(fObj.isError()) << printError(fObj);
  ASSERT_EQ(fObj.getErrorCode(), NO_ERROR);
  ASSERT_TRUE(fObj.getErrorMsg().empty());

  DRCEOutputJsonMessage resultJsonMessage(resultJson);
  ASSERT_FALSE(resultJsonMessage.isError()) << printError(resultJsonMessage);
  ASSERT_EQ(resultJsonMessage.getErrorCode(), NO_ERROR) << printError(resultJsonMessage);
  ASSERT_TRUE(resultJsonMessage.getErrorMsg().empty()) << printError(resultJsonMessage);
  ASSERT_EQ(resultJsonMessage.getResultData().getItemsCount(), 1);


//  std::cout << ">>>>> taskId: " << taskId << " code: " << resultJsonMessage.getResultData().getDataItem(0).getErrorCode()
//      << " message: " << resultJsonMessage.getResultData().getDataItem(0).getErrorMessage()
//      << std::endl;

  if (exceededLimits==true)
    ASSERT_TRUE(resultJsonMessage.getResultData().getDataItem(0).getErrorCode()==ERROR_EXCEEDED_RESOURCE_LIMITS ||
        resultJsonMessage.getResultData().getDataItem(0).getErrorCode()==NO_ERROR);
  else
  {
    if (resultJsonMessage.getResultData().getDataItem(0).getErrorCode() != NO_ERROR)
    {
      std::cout << ">>>>> taskId: " << taskId << " code: " << resultJsonMessage.getResultData().getDataItem(0).getErrorCode()
            << " message: " << resultJsonMessage.getResultData().getDataItem(0).getErrorMessage()
            << std::endl;
    }
  }

  DRCECleanupTest::cleanupTestData(fObj);
}
//-----------------------------------------------------------------------------
void DRCEStressTest::testResourceLimits(DRCEFunctionalObject& fObj, SessionOptions::SessionType sessionType, SessionOptions::ThreadMode threadMode)
{
  fObj.setResourceMonitorTimePeriod(1000);

  unsigned int taskId(0);

  ResourceLimits resourceLimits;
  CpuUsageExtractor cpuUsageExtractor;

  resourceLimits.cpuMax = cpuUsageExtractor.getCpuUsage() * 2;
  DRCEStressTest::testResourceLimits(fObj, sessionType, threadMode, resourceLimits, taskId, false);
  resourceLimits.cpuMax = cpuUsageExtractor.getCpuUsage() / 3;
  DRCEStressTest::testResourceLimits(fObj, sessionType, threadMode, resourceLimits, taskId, true);
  printSuccess("Test Cpu usage");

  resourceLimits.cpuMax = 0;
  resourceLimits.iowaitMax = cpuUsageExtractor.getIOWait() * 2;
  DRCEStressTest::testResourceLimits(fObj, sessionType, threadMode, resourceLimits, taskId, false);
  resourceLimits.iowaitMax = cpuUsageExtractor.getIOWait() / 3;
  DRCEStressTest::testResourceLimits(fObj, sessionType, threadMode, resourceLimits, taskId, true);
  printSuccess("Test getting IOWAIT");

  resourceLimits.clear();
  DiskSpaceExtractor diskSpaceExtractor;
  resourceLimits.diskFreeMin = diskSpaceExtractor.getDiskFree()-1024*1024;
  DRCEStressTest::testResourceLimits(fObj, sessionType, threadMode, resourceLimits, taskId, false);
  resourceLimits.diskFreeMin = diskSpaceExtractor.getDiskFree()+1024*1024;
  DRCEStressTest::testResourceLimits(fObj, sessionType, threadMode, resourceLimits, taskId, true);
  printSuccess("Test Free Disk");

  resourceLimits.clear();
  resourceLimits.diskFree = diskSpaceExtractor.getDiskFreePercent()-1;
  DRCEStressTest::testResourceLimits(fObj, sessionType, threadMode, resourceLimits, taskId, false);
  resourceLimits.diskFree = diskSpaceExtractor.getDiskFreePercent()+1;
  DRCEStressTest::testResourceLimits(fObj, sessionType, threadMode, resourceLimits, taskId, true);
  printSuccess("Test Free Disk Percent");

  resourceLimits.clear();
  MemoryExtractor memoryExtractor;
  resourceLimits.rramFreeMin = memoryExtractor.getRramFree()-10*1024*1024;
  DRCEStressTest::testResourceLimits(fObj, sessionType, threadMode, resourceLimits, taskId, false);
  resourceLimits.rramFreeMin = memoryExtractor.getRramFree()+10*1024*1024;
  DRCEStressTest::testResourceLimits(fObj, sessionType, threadMode, resourceLimits, taskId, true);
  printSuccess("Test RRAM Free Min");

  resourceLimits.clear();
  resourceLimits.rramFree = memoryExtractor.getRramFreePercent()-30;
  DRCEStressTest::testResourceLimits(fObj, sessionType, threadMode, resourceLimits, taskId, false);
  resourceLimits.rramFree = memoryExtractor.getRramFreePercent()+30;
  DRCEStressTest::testResourceLimits(fObj, sessionType, threadMode, resourceLimits, taskId, true);
  printSuccess("Test RRAM Free Percent");

  resourceLimits.clear();
  resourceLimits.vramFreeMin = memoryExtractor.getVramFree()-10*1024*1024;
  DRCEStressTest::testResourceLimits(fObj, sessionType, threadMode, resourceLimits, taskId, false);
  resourceLimits.vramFreeMin = memoryExtractor.getVramFree()+10*1024*1024;
  DRCEStressTest::testResourceLimits(fObj, sessionType, threadMode, resourceLimits, taskId, true);
  printSuccess("Test VRAM Free Min");

  resourceLimits.clear();
  resourceLimits.vramFree = memoryExtractor.getVramFreePercent()-30;
  DRCEStressTest::testResourceLimits(fObj, sessionType, threadMode, resourceLimits, taskId, false);
  resourceLimits.vramFree = memoryExtractor.getVramFreePercent()+30;
  DRCEStressTest::testResourceLimits(fObj, sessionType, threadMode, resourceLimits, taskId, true);
  printSuccess("Test VRAM Free Percent");

  resourceLimits.clear();
  resourceLimits.procMax = 1000;
  DRCEStressTest::testResourceLimits(fObj, sessionType, threadMode, resourceLimits, taskId, false);
  resourceLimits.procMax = 10;
  DRCEStressTest::testResourceLimits(fObj, sessionType, threadMode, resourceLimits, taskId, true);
  printSuccess("Test Proc Max");

  resourceLimits.clear();
  resourceLimits.threadsMax = 1000;
  DRCEStressTest::testResourceLimits(fObj, sessionType, threadMode, resourceLimits, taskId, false);
  resourceLimits.threadsMax = 10;
  DRCEStressTest::testResourceLimits(fObj, sessionType, threadMode, resourceLimits, taskId, true);
  printSuccess("Test Threads Max");

  while(fObj.getAsyncTasksCount())
  {
    Poco::Thread::sleep(10000);
  }

  printSuccess ("Test Resource Limits");
}
//-----------------------------------------------------------------------------
void DRCEStressTest::testResourceUsageLimits(DRCEFunctionalObject& fObj, SessionOptions::SessionType sessionType, SessionOptions::ThreadMode threadMode)
{
  const std::string command = "echo Hello && sleep 10 && ps auxf && sleep 10 && ls -l";
  fObj.setResourceMonitorTimePeriod(10000);
  unsigned int taskId(0);

  ResourceLimits resourceLimits;
  CpuUsageExtractor cpuUsageExtractor;
  MemoryExtractor memoryExtractor;
  DiskSpaceExtractor diskSpaceExtractor;

  resourceLimits.usageLimits.iowaitMax = 10;
  DRCEStressTest::testResourceLimits(fObj, sessionType, threadMode, resourceLimits, taskId, false, command);
  resourceLimits.usageLimits.iowaitMax = 0.001;
  DRCEStressTest::testResourceLimits(fObj, sessionType, threadMode, resourceLimits, taskId, true, command);
  printSuccess("Test usage IOWAIT limit");

  resourceLimits.clear();
  resourceLimits.usageLimits.rramMax = memoryExtractor.getRramUsedPercent()-10;
  DRCEStressTest::testResourceLimits(fObj, sessionType, threadMode, resourceLimits, taskId, false, command);
  resourceLimits.usageLimits.rramMax = 0.0001;
  DRCEStressTest::testResourceLimits(fObj, sessionType, threadMode, resourceLimits, taskId, true, command);
  printSuccess("Test usage RRAM limit");

  resourceLimits.clear();
  resourceLimits.usageLimits.vramMax = memoryExtractor.getVramUsedPercent()-10;
  DRCEStressTest::testResourceLimits(fObj, sessionType, threadMode, resourceLimits, taskId, false, command);
  resourceLimits.usageLimits.vramMax = 0.0001;
  DRCEStressTest::testResourceLimits(fObj, sessionType, threadMode, resourceLimits, taskId, true, command);
  printSuccess("Test usage VRAM limit");

  resourceLimits.clear();
  resourceLimits.usageLimits.diskMax = diskSpaceExtractor.getDiskUsedPercent()-10;
  DRCEStressTest::testResourceLimits(fObj, sessionType, threadMode, resourceLimits, taskId, false, command);
  resourceLimits.usageLimits.diskMax = 0.0001;
  DRCEStressTest::testResourceLimits(fObj, sessionType, threadMode, resourceLimits, taskId, true, command);
  printSuccess("Test usage HDD limit percent");

  resourceLimits.clear();
  resourceLimits.usageLimits.diskMaxBytes = diskSpaceExtractor.getDiskUsed()-1024*1024;
  DRCEStressTest::testResourceLimits(fObj, sessionType, threadMode, resourceLimits, taskId, false, command);
  resourceLimits.usageLimits.diskMaxBytes = 1;
  DRCEStressTest::testResourceLimits(fObj, sessionType, threadMode, resourceLimits, taskId, true, command);
  printSuccess("Test usage HDD limit bytes");

  while(fObj.getAsyncTasksCount())
  {
    Poco::Thread::sleep(10000);
  }
  printSuccess ("Test Resource Usage Limits");
}
//-----------------------------------------------------------------------------
//-----------------------------------------------------------------------------
} // end namespace tests
} // end namespace drce 
} // end namespace HCE 
