#include <gtest/gtest.h>
#include <fstream>
#include <utility>
#include <vector>
#include <Poco/File.h>

#include "DRCEError.hpp"
#include "DRCESessionOptions.hpp"
#include "DRCEInputJsonMessage.hpp"
#include "DRCEOutputJsonMessage.hpp"
#include "DRCEFunctionalObject.hpp"
#include "DRCEResultDataSerializator.hpp"
#include "DRCEResultData.hpp"
#include "DRCEPrintStatus.hpp"
#include "DRCEFileStream.hpp"
#include "DRCETaskRequestGetData.hpp"
#include "DRCETaskRequestTerminate.hpp"
#include "DRCETaskRequestCheckState.hpp"
#include "DRCETaskRequestDeleteData.hpp"
#include "DRCETaskThreadingModeTest.hpp"
#include "DRCEStressTest.hpp"
#include "DRCECleanupTest.hpp"
#include "DRCETasksQueue.hpp"
#include "DRCEFileExtractor.hpp"
#include "CustomMessage.hpp"
#include "DRCEMessageConst.hpp"
#include "DRCEListAllTasks.hpp"
#include "DRCEResourceLimits.hpp"

namespace HCE
{
namespace drce
{
namespace tests
{
//-----------------------------------------------------------------------------
static size_t count = 0;
struct TestFunctor
{
  std::string operator()(const std::string& data)
  {
    ++count;
    return data+"\n";
  }
};
static TestFunctor testFunctor;
//-----------------------------------------------------------------------------
//-----------------------------------------------------------------------------
DRCENodeOptions DRCETaskThreadingModeTest::getNodeOptions(const std::string& homeDir)
{
  DRCENodeOptions nodeOptions("", "", "", "");
  nodeOptions.setNodeHost("NODE_HOST");
  nodeOptions.setNodePort("NODE_PORT");
  nodeOptions.setNodeName("NODE_NAME");
  if (homeDir.empty())
  {
    char* dir = get_current_dir_name();
    nodeOptions.setHomeDir(dir);
    free(dir);
    dir = nullptr;
  }
  nodeOptions.setTasksDataDir(nodeOptions.getHomeDir()+"/data");
  nodeOptions.setTasksStatusDir(nodeOptions.getHomeDir()+"/status");
  mkdir(nodeOptions.getTasksDataDir().c_str(), S_IRWXU | S_IRWXG | S_IRWXO);
  mkdir(nodeOptions.getTasksStatusDir().c_str(), S_IRWXU | S_IRWXG | S_IRWXO);
  return nodeOptions;
}
//-----------------------------------------------------------------------------
bool DRCETaskThreadingModeTest::isExistTask(DRCEFunctionalObject& fObj, unsigned int requestId)
{
  const std::string fileName = FileStream::getStatusFileName(fObj.getTasksStatusDir(), requestId);
  return Poco::File(fileName).exists();
}
//-----------------------------------------------------------------------------
//-----------------------------------------------------------------------------
void DRCETaskThreadingModeTest::testExecuteSyncRequest(DRCEFunctionalObject& fObj, unsigned int requestId)
{
  const std::string requestJson = "{\"session\":{\"type\":0,\"tmode\":1,\"cleanup\":0,\"time_max\":5000,\"port\":0,\"user\":\"\",\"password\":\"\",\"shell\":\"\",\"environment\":[{\"LINES\":\"80\",\"HOSTNAME\":\"node22\"}]},\"limits\":{\"proc_max\":400},\"command\":\"sort\",\"input\":\"def\\nabc\\n123\\n\",\"files\":[{\"name\":\"file1\",\"data\":\"test content of file1\\n\",\"action\":13},{\"name\":\"file2\",\"data\":\"test content of file2\\n\",\"action\":1}]}";
//  const std::string requestJson = "{\"session\":{\"type\":0,\"cleanup\":1,\"tmode\":1,\"time_max\":10000,\"port\":0,\"user\":\"\",\"password\":\"\",\"shell\":\"\",\"environment\":[{}]},\"command\":\"echo \\\"scale=4000; 4*a(1)\\\" | bc -l && ps xau && ls -la\",\"input\":\"\",\"files\":[]}";

  // set request for execute task
  DRCEInputJsonMessage inputJsonMessage;
  inputJsonMessage.setRequestType(DRCETaskRequest::RequestType::rtSetTaskExecute);
  inputJsonMessage.setRequestData(requestJson);
  inputJsonMessage.setRequestId(requestId);

  std::stringstream sourceJson;
//  sourceJson << inputJsonMessage;
  std::string json;
  inputJsonMessage.serialize(json);
  sourceJson.str(json);

//  std::cout << "source JSON: " << sourceJson.str() << std::endl;

  ASSERT_FALSE(inputJsonMessage.isError()) << printError(inputJsonMessage);
  ASSERT_EQ(inputJsonMessage.getErrorCode(), NO_ERROR);
  ASSERT_TRUE(inputJsonMessage.getErrorMsg().empty());
  printSuccess("DRCEInputJsonMessage serialize");

  fObj.setNodeHost("NODE_HOST");
  fObj.setNodePort("NODE_PORT");
  fObj.setNodeName("NODE_NAME");
  fObj.setTasksQueueDumpPeriod(2000);

  std::string resultJson = fObj.Process(sourceJson.str());

  ASSERT_FALSE(fObj.isError()) << printError(fObj);
  ASSERT_EQ(fObj.getErrorCode(), NO_ERROR);
  ASSERT_TRUE(fObj.getErrorMsg().empty());
  printSuccess("DRCEFunctionalObject Process");

//  std::cout << "result JSON: " << resultJson << std::endl;

  // check right result (unserialize and compare two objects)
  DRCEOutputJsonMessage resultJsonMessage(resultJson);

  ASSERT_FALSE(resultJsonMessage.isError()) << printError(resultJsonMessage);
  ASSERT_EQ(resultJsonMessage.getErrorCode(), NO_ERROR);
  ASSERT_TRUE(resultJsonMessage.getErrorMsg().empty());
  printSuccess("DRCEOutputJsonMessage unserialize");

  ASSERT_EQ(resultJsonMessage.getResultData().getItemsCount(), 1);
  ASSERT_FALSE(resultJsonMessage.getResultData().getDataItem(0).getStdoutStream().empty());
  ASSERT_TRUE(resultJsonMessage.getResultData().getDataItem(0).getStderrStream().empty());
  ASSERT_EQ(resultJsonMessage.getResultData().getDataItem(0).getNodeName(), fObj.getNodeName());
  ASSERT_EQ(resultJsonMessage.getResultData().getDataItem(0).getExitStatus(), 0);
  ASSERT_EQ(resultJsonMessage.getResultData().getDataItem(0).getFilesCount(), 1);
  ASSERT_EQ(resultJsonMessage.getResultData().getDataItem(0).getFileItem(0).name, "file1");
  ASSERT_EQ(resultJsonMessage.getResultData().getDataItem(0).getFileItem(0).data, "test content of file1\n");
  ASSERT_EQ(resultJsonMessage.getResultData().getDataItem(0).getFileItem(0).actionType, 13);

  DRCECleanupTest::cleanupTestData(fObj);
  printSuccess("!!! Execute sync request");
}
//-----------------------------------------------------------------------------
void DRCETaskThreadingModeTest::testSetTaskToExecuteRequest(DRCEFunctionalObject& fObj, unsigned int requestId, const std::string& requestData)
{
  std::string requestJson;
  if (requestData.empty())
    requestJson = "{\"session\":{\"type\":0,\"tmode\":2,\"time_max\":40000,\"port\":0,\"user\":\"\",\"password\":\"\",\"shell\":\"\",\"environment\":[{\"LINES\":\"80\",\"HOSTNAME\":\"node22\"}]},\"command\":\"sort\",\"input\":\"def\\nabc\\n123\\n\",\"files\":[{\"name\":\"file1\",\"data\":\"test content of file1\\n\",\"action\":13},{\"name\":\"file2\",\"data\":\"test content of file2\\n\",\"action\":1}]}";
  else
    requestJson = requestData;

//  std::cout << "json: " << requestJson << std::endl;

  // set request for execute task
  DRCEInputJsonMessage inputJsonMessage;
  inputJsonMessage.setRequestType(DRCETaskRequest::RequestType::rtSetTaskExecute);
  inputJsonMessage.setRequestData(requestJson);
  inputJsonMessage.setRequestId(requestId);

  std::stringstream sourceJson;
//  sourceJson << inputJsonMessage;
  std::string json;
  inputJsonMessage.serialize(json);
  sourceJson.str(json);

//  std::cout << "source JSON: " << sourceJson.str() << std::endl;

  ASSERT_FALSE(inputJsonMessage.isError()) << printError(inputJsonMessage);
  ASSERT_EQ(inputJsonMessage.getErrorCode(), NO_ERROR);
  ASSERT_TRUE(inputJsonMessage.getErrorMsg().empty());
  printSuccess("DRCEInputJsonMessage serialize");

  std::string resultJson = fObj.Process(sourceJson.str());

  ASSERT_FALSE(fObj.isError()) << printError(fObj);
  ASSERT_EQ(fObj.getErrorCode(), NO_ERROR);
  ASSERT_TRUE(fObj.getErrorMsg().empty());
  printSuccess("DRCEFunctionalObject Process");

  std::cout << "result JSON: " << resultJson << std::endl;

  // check right result (unserialize and compare two objects)
  DRCEOutputJsonMessage resultJsonMessage(resultJson);

  ASSERT_FALSE(resultJsonMessage.isError()) << printError(resultJsonMessage);
  ASSERT_EQ(resultJsonMessage.getErrorCode(), NO_ERROR);
  ASSERT_TRUE(resultJsonMessage.getErrorMsg().empty());
  printSuccess("DRCEOutputJsonMessage unserialize");

//  std::cout << "ItemsCount = " << resultJsonMessage.getResultData().getItemsCount() << std::endl;
//  for (size_t i=0;i<resultJsonMessage.getResultData().getItemsCount();++i)
//  {
//    std::cout << "stderr: " << resultJsonMessage.getResultData().getDataItem(i).getStderrStream() << std::endl;
//    std::cout << "stdout: " << resultJsonMessage.getResultData().getDataItem(i).getStdoutStream() << std::endl;
//  }

//  DRCECleanupTest::cleanupTestData(fObj);
  printSuccess("!!! Set task to async execute request");
}
//-----------------------------------------------------------------------------
void DRCETaskThreadingModeTest::testCheckTaskRequest(DRCEFunctionalObject& fObj, unsigned int requestId, const std::string& requestData)
{
  std::string requestJson;
  if (requestData.empty())
    requestJson = "{\"session\":{\"type\":0,\"tmode\":1,\"time_max\":40000,\"port\":0,\"user\":\"\",\"password\":\"\",\"shell\":\"\",\"environment\":[]},\"command\":\"sort\",\"input\":\"def\\nabc\\n123\\n\",\"files\":[]}";
  else
    requestJson = requestData;

  // set request for execute task
  DRCEInputJsonMessage inputJsonMessage;
  inputJsonMessage.setRequestType(DRCETaskRequest::RequestType::rtSetTaskExecute);
  inputJsonMessage.setRequestData(requestJson);
  inputJsonMessage.setRequestId(requestId);

   std::stringstream sourceJson;
   sourceJson << inputJsonMessage;

//  std::cout << "source JSON: " << sourceJson.str() << std::endl;

  ASSERT_FALSE(inputJsonMessage.isError()) << printError(inputJsonMessage);
  ASSERT_EQ(inputJsonMessage.getErrorCode(), NO_ERROR);
  ASSERT_TRUE(inputJsonMessage.getErrorMsg().empty());
  printSuccess("DRCEInputJsonMessage serialize");

  std::string resultJson = fObj.Process(sourceJson.str());

  std::cout << "SetTaskExecute resultJson: " << resultJson << std::endl;

  ASSERT_FALSE(fObj.isError()) << printError(fObj);
  ASSERT_EQ(fObj.getErrorCode(), NO_ERROR);
  ASSERT_TRUE(fObj.getErrorMsg().empty());
  printSuccess("DRCEFunctionalObject Process");

  Poco::Thread::sleep(5);
/*
  while (fObj.getAsyncTasksCount() || fObj.hasTask(requestId))
  {
    std::cout << "isExistTask on HDD: " << std::boolalpha << DRCETaskThreadingModeTest::isExistTask(fObj, requestId) << std::endl;

    Poco::Thread::sleep(10000);
    std::cout << "wait the end ...." << std::endl;
  }
  Poco::Thread::sleep(5000);
*/
  // Check result use Simple flag
  std::string json;
  DRCETaskRequestCheckState taskRequestCheckState;
  taskRequestCheckState.setCheckType(DRCETaskRequestCheckState::CheckType::ctSimple);
  taskRequestCheckState.serialize(json);
  ASSERT_FALSE(taskRequestCheckState.isError());
  printSuccess("DRCETaskRequestCheckState serialize");

  std::cout << "json: " << json << std::endl;

  // set request for execute task
  inputJsonMessage.clear();
  inputJsonMessage.setRequestType(DRCETaskRequest::RequestType::rtCheckTaskState);
  inputJsonMessage.setRequestData(json);
  inputJsonMessage.setRequestId(requestId);

  sourceJson.str("");
  sourceJson << inputJsonMessage;

//  std::cout << "source JSON: " << sourceJson.str() << std::endl;

  ASSERT_FALSE(inputJsonMessage.isError()) << printError(inputJsonMessage);
  ASSERT_EQ(inputJsonMessage.getErrorCode(), NO_ERROR);
  ASSERT_TRUE(inputJsonMessage.getErrorMsg().empty());
  printSuccess("DRCEInputJsonMessage serialize");

  resultJson = fObj.Process(sourceJson.str());

  ASSERT_FALSE(fObj.isError()) << printError(fObj);
  ASSERT_EQ(fObj.getErrorCode(), NO_ERROR);
  ASSERT_TRUE(fObj.getErrorMsg().empty());
  printSuccess("DRCEFunctionalObject Process");

  std::cout << "CheckTaskState result JSON: " << resultJson << std::endl;

  // check right result (unserialize and compare two objects)
  DRCEOutputJsonMessage resultJsonMessage(resultJson);

  ASSERT_FALSE(resultJsonMessage.isError()) << printError(resultJsonMessage);
  ASSERT_EQ(resultJsonMessage.getErrorCode(), NO_ERROR);
  ASSERT_TRUE(resultJsonMessage.getErrorMsg().empty());
  printSuccess("DRCEOutputJsonMessage unserialize");

  ASSERT_EQ(resultJsonMessage.getResultData().getItemsCount(), 1);
  ASSERT_TRUE(resultJsonMessage.getResultData().getDataItem(0).getStdoutStream().empty());
  ASSERT_TRUE(resultJsonMessage.getResultData().getDataItem(0).getStderrStream().empty());
  ASSERT_EQ(resultJsonMessage.getResultData().getDataItem(0).getNodeName(), fObj.getNodeName());
  ASSERT_EQ(resultJsonMessage.getResultData().getDataItem(0).getNodeHost(), fObj.getNodeHost());
  ASSERT_EQ(resultJsonMessage.getResultData().getDataItem(0).getNodePort(), fObj.getNodePort());

  printSuccess("Check state task use Simple flag");

  // Check result alredy ended task
  taskRequestCheckState.clear();
  taskRequestCheckState.setCheckType(DRCETaskRequestCheckState::CheckType::ctExtended);
  taskRequestCheckState.serialize(json);
  ASSERT_FALSE(taskRequestCheckState.isError());
  printSuccess("DRCETaskRequestCheckState serialize");

//  std::cout << "json: " << json << std::endl;
  //**********************
  while (fObj.getAsyncTasksCount() || fObj.hasTask(requestId))
  {
    std::cout << "isExistTask on HDD: " << std::boolalpha << DRCETaskThreadingModeTest::isExistTask(fObj, requestId) << std::endl;

    Poco::Thread::sleep(10000);
    std::cout << "wait the end ...." << std::endl;
  }
  //***********************
/*
  // set request for execute task
  inputJsonMessage.clear();
  inputJsonMessage.setRequestType(DRCETaskRequest::RequestType::rtCheckTaskState);
  inputJsonMessage.setRequestData(json);
  inputJsonMessage.setRequestId(requestId);

  sourceJson.str("");
  sourceJson << inputJsonMessage;

//  std::cout << "source JSON: " << sourceJson.str() << std::endl;

  ASSERT_FALSE(inputJsonMessage.isError()) << printError(inputJsonMessage);
  ASSERT_EQ(inputJsonMessage.getErrorCode(), NO_ERROR);
  ASSERT_TRUE(inputJsonMessage.getErrorMsg().empty());
  printSuccess("DRCEInputJsonMessage serialize");

  resultJson = fObj.Process(sourceJson.str());

  ASSERT_FALSE(fObj.isError()) << printError(fObj);
  ASSERT_EQ(fObj.getErrorCode(), NO_ERROR);
  ASSERT_TRUE(fObj.getErrorMsg().empty());
  printSuccess("DRCEFunctionalObject Process");

//  std::cout << "result JSON: " << resultJson << std::endl;

  // check right result (unserialize and compare two objects)
  resultJsonMessage.unserialize(resultJson);

  ASSERT_FALSE(resultJsonMessage.isError()) << printError(resultJsonMessage);
  ASSERT_EQ(resultJsonMessage.getErrorCode(), NO_ERROR);
  ASSERT_TRUE(resultJsonMessage.getErrorMsg().empty());
  printSuccess("DRCEOutputJsonMessage unserialize");

  ASSERT_EQ(resultJsonMessage.getResultData().getItemsCount(), 1);
  ASSERT_EQ(resultJsonMessage.getResultData().getDataItem(0).getStdoutStream(), "123\nabc\ndef\n");
  ASSERT_TRUE(resultJsonMessage.getResultData().getDataItem(0).getStderrStream().empty());
  ASSERT_EQ(resultJsonMessage.getResultData().getDataItem(0).getExitStatus(), 0);
  ASSERT_EQ(resultJsonMessage.getResultData().getDataItem(0).getNodeName(), fObj.getNodeName());
  ASSERT_EQ(resultJsonMessage.getResultData().getDataItem(0).getNodeHost(), fObj.getNodeHost());
  ASSERT_EQ(resultJsonMessage.getResultData().getDataItem(0).getNodePort(), fObj.getNodePort());
*/
  DRCECleanupTest::cleanupTestData(fObj);
  printSuccess("!!! Check state task request after execute");
}
//-----------------------------------------------------------------------------
void DRCETaskThreadingModeTest::testGetTaskDataRequest(DRCEFunctionalObject& fObj, unsigned int requestId)
{
  const std::string requestJson = "{\"session\":{\"type\":0,\"tmode\":2,\"time_max\":40000,\"port\":0,\"user\":\"\",\"password\":\"\",\"shell\":\"\",\"environment\":[]},\"command\":\"sort\",\"input\":\"def\\nabc\\n123\\n\",\"files\":[]}";

  // set request for execute task
  DRCEInputJsonMessage inputJsonMessage;
  inputJsonMessage.setRequestType(DRCETaskRequest::RequestType::rtSetTaskExecute);
  inputJsonMessage.setRequestData(requestJson);
  inputJsonMessage.setRequestId(requestId);

   std::stringstream sourceJson;
   sourceJson << inputJsonMessage;

//  std::cout << "source JSON: " << sourceJson.str() << std::endl;

  ASSERT_FALSE(inputJsonMessage.isError()) << printError(inputJsonMessage);
  ASSERT_EQ(inputJsonMessage.getErrorCode(), NO_ERROR);
  ASSERT_TRUE(inputJsonMessage.getErrorMsg().empty());
  printSuccess("DRCEInputJsonMessage serialize");

  std::string resultJson = fObj.Process(sourceJson.str());

  ASSERT_FALSE(fObj.isError()) << printError(fObj);
  ASSERT_EQ(fObj.getErrorCode(), NO_ERROR);
  ASSERT_TRUE(fObj.getErrorMsg().empty());
  printSuccess("DRCEFunctionalObject Process");

//  std::cout << "resultJson: " << resultJson << std::endl;

  while (fObj.getAsyncTasksCount() || fObj.hasTask(requestId))
  {
    Poco::Thread::sleep(10000);
    std::cout << "wait the end ...." << std::endl;
  }
  Poco::Thread::sleep(5000);

  std::string json;
  DRCETaskRequestGetData taskRequestGetData;
  taskRequestGetData.setFetchType(DRCETaskRequestGetData::FetchType::ftNotDeleteDataAfterFetch);
  taskRequestGetData.serialize(json);
  ASSERT_FALSE(taskRequestGetData.isError());
  printSuccess("DRCETaskRequestGetData serialize");

//  std::cout << "json: " << json << std::endl;

  // set request for execute task
  inputJsonMessage.clear();
  inputJsonMessage.setRequestType(DRCETaskRequest::RequestType::rtGetTaskData);
  inputJsonMessage.setRequestData(json);
  inputJsonMessage.setRequestId(requestId);

  sourceJson.str("");
  sourceJson << inputJsonMessage;

//  std::cout << "source JSON: " << sourceJson.str() << std::endl;

  ASSERT_FALSE(inputJsonMessage.isError()) << printError(inputJsonMessage);
  ASSERT_EQ(inputJsonMessage.getErrorCode(), NO_ERROR);
  ASSERT_TRUE(inputJsonMessage.getErrorMsg().empty());
  printSuccess("DRCEInputJsonMessage serialize");

  resultJson = fObj.Process(sourceJson.str());

  ASSERT_FALSE(fObj.isError()) << printError(fObj);
  ASSERT_EQ(fObj.getErrorCode(), NO_ERROR);
  ASSERT_TRUE(fObj.getErrorMsg().empty());
  printSuccess("DRCEFunctionalObject Process");

//  std::cout << "result JSON: " << resultJson << std::endl;

  // check right result (unserialize and compare two objects)
  DRCEOutputJsonMessage resultJsonMessage(resultJson);

  ASSERT_FALSE(resultJsonMessage.isError()) << printError(resultJsonMessage);
  ASSERT_EQ(resultJsonMessage.getErrorCode(), NO_ERROR);
  ASSERT_TRUE(resultJsonMessage.getErrorMsg().empty());
  printSuccess("DRCEOutputJsonMessage unserialize");

  ASSERT_EQ(resultJsonMessage.getResultData().getItemsCount(), 1);
  ASSERT_EQ(resultJsonMessage.getResultData().getDataItem(0).getStdoutStream(), "123\nabc\ndef\n");
  ASSERT_TRUE(resultJsonMessage.getResultData().getDataItem(0).getStderrStream().empty());
  ASSERT_EQ(resultJsonMessage.getResultData().getDataItem(0).getExitStatus(), 0);
  ASSERT_EQ(resultJsonMessage.getResultData().getDataItem(0).getNodeName(), fObj.getNodeName());
  ASSERT_EQ(resultJsonMessage.getResultData().getDataItem(0).getNodeHost(), fObj.getNodeHost());
  ASSERT_EQ(resultJsonMessage.getResultData().getDataItem(0).getNodePort(), fObj.getNodePort());

  DRCECleanupTest::cleanupTestData(fObj);
  printSuccess("!!! Get data task request after execute");
}
//-----------------------------------------------------------------------------
void DRCETaskThreadingModeTest::testTerminateTaskRequest(DRCEFunctionalObject& fObj, unsigned int requestId,
    DRCETaskRequestTerminate::AlgorithmType algorithmType, DRCETaskRequestTerminate::CleanupFlag cleanupFlag)
{
//  std::string requestJson = "{\"session\":{\"type\":0,\"tmode\":2,\"time_max\":80000,\"port\":0,\"user\":\"\",\"password\":\"\",\"shell\":\"\",\"environment\":[{}]},\"command\":\"echo \\\"scale=400; 4*a(1)\\\" | bc -l && sleep 10 && ps xau\",\"input\":\"\",\"files\":[{\"name\":\"file1\",\"data\":\"test content of file1\\n\",\"action\":13},{\"name\":\"file2\",\"data\":\"test content of file2\\n\",\"action\":1}]}";
  std::string requestJson = "{\"session\":{\"type\":0,\"cleanup\":0,\"tmode\":2,\"time_max\":80000,\"port\":0,\"user\":\"\",\"password\":\"\",\"shell\":\"\",\"environment\":[{}]},\"command\":\"echo \\\"scale=40; 4*a(1)\\\" | bc -l && /home/alexander/workspace/sleeper/Debug/sleeper 10 && echo Hello && /home/alexander/workspace/sleeper/Debug/sleeper 10\",\"input\":\"\",\"files\":[]}";

  // set request for execute task
  DRCEInputJsonMessage inputJsonMessage;
  inputJsonMessage.setRequestType(DRCETaskRequest::RequestType::rtSetTaskExecute);
  inputJsonMessage.setRequestData(requestJson);
  inputJsonMessage.setRequestId(requestId);

  std::stringstream sourceJson;
  sourceJson << inputJsonMessage;

//  std::cout << "source JSON: " << sourceJson.str() << std::endl;

  ASSERT_FALSE(inputJsonMessage.isError()) << printError(inputJsonMessage);
  ASSERT_EQ(inputJsonMessage.getErrorCode(), NO_ERROR);
  ASSERT_TRUE(inputJsonMessage.getErrorMsg().empty());
  printSuccess("DRCEInputJsonMessage serialize");

  std::string resultJson = fObj.Process(sourceJson.str());

//  std::cout << "result JSON: " << resultJson << std::endl;

  ASSERT_FALSE(fObj.isError()) << printError(fObj);
  ASSERT_EQ(fObj.getErrorCode(), NO_ERROR);
  ASSERT_TRUE(fObj.getErrorMsg().empty());
  printSuccess("DRCEFunctionalObject Process");

  Poco::Thread::sleep(5000);

  std::string json;
  DRCETaskRequestTerminate taskRequestTerminate;
  taskRequestTerminate.setTaskId(requestId);
  taskRequestTerminate.setAlgorithmType(algorithmType);
  taskRequestTerminate.setDelayValue(1000);
  taskRequestTerminate.setRepeatValue(3);
  taskRequestTerminate.setSignalValue(9);
  taskRequestTerminate.setCleanupFlag(cleanupFlag);
  taskRequestTerminate.serialize(json);
  ASSERT_FALSE(taskRequestTerminate.isError());
  printSuccess("DRCETaskRequestTerminate serialize");

//  std::cout << "json: " << json << std::endl;

  // set request for execute task
  inputJsonMessage.clear();
  inputJsonMessage.setRequestType(DRCETaskRequest::RequestType::rtTerminateTask);
  inputJsonMessage.setRequestData(json);
  inputJsonMessage.setRequestId(requestId);

  sourceJson.str("");
  sourceJson << inputJsonMessage;

//  std::cout << "source JSON: " << sourceJson.str() << std::endl;

  ASSERT_FALSE(inputJsonMessage.isError()) << printError(inputJsonMessage);
  ASSERT_EQ(inputJsonMessage.getErrorCode(), NO_ERROR);
  ASSERT_TRUE(inputJsonMessage.getErrorMsg().empty());
  printSuccess("DRCEInputJsonMessage serialize");

  resultJson = fObj.Process(sourceJson.str());

  ASSERT_FALSE(fObj.isError()) << printError(fObj);
  ASSERT_EQ(fObj.getErrorCode(), NO_ERROR);
  ASSERT_TRUE(fObj.getErrorMsg().empty());
  printSuccess("DRCEFunctionalObject Process");

  std::cout << "result JSON: " << resultJson << std::endl;

  // check right result (unserialize and compare two objects)
  DRCEOutputJsonMessage resultJsonMessage(resultJson);

  ASSERT_FALSE(resultJsonMessage.isError()) << printError(resultJsonMessage);
  ASSERT_EQ(resultJsonMessage.getErrorCode(), NO_ERROR);
  ASSERT_TRUE(resultJsonMessage.getErrorMsg().empty());
  printSuccess("DRCEOutputJsonMessage unserialize");

  ASSERT_EQ(resultJsonMessage.getResultData().getItemsCount(), 1);
  ASSERT_TRUE(resultJsonMessage.getResultData().getDataItem(0).getStdoutStream().empty());
  ASSERT_TRUE(resultJsonMessage.getResultData().getDataItem(0).getStderrStream().empty());
  ASSERT_EQ(resultJsonMessage.getResultData().getDataItem(0).getExitStatus(), (128+taskRequestTerminate.getSignalValue()));
  ASSERT_EQ(resultJsonMessage.getResultData().getDataItem(0).getNodeName(), fObj.getNodeName());
  ASSERT_EQ(resultJsonMessage.getResultData().getDataItem(0).getNodeHost(), fObj.getNodeHost());
  ASSERT_EQ(resultJsonMessage.getResultData().getDataItem(0).getNodePort(), fObj.getNodePort());
  ASSERT_EQ(resultJsonMessage.getResultData().getDataItem(0).getState(), DRCETaskRequest::TaskState::TERMINATED);

  while (fObj.getAsyncTasksCount())
  {
    Poco::Thread::sleep(5000);
    std::cout << "wait the end ...." << std::endl;
  }

  DRCECleanupTest::cleanupTestData(fObj);
  printSuccess("!!! Terminate task request");
}
//-----------------------------------------------------------------------------
void DRCETaskThreadingModeTest::testDeleteTaskDataRequest(DRCEFunctionalObject& fObj, unsigned int requestId)
{
  const std::string requestJson = "{\"session\":{\"type\":0,\"tmode\":2,\"time_max\":40000,\"port\":0,\"user\":\"\",\"password\":\"\",\"shell\":\"\",\"environment\":[]},\"command\":\"sort\",\"input\":\"def\\nabc\\n123\\n\",\"files\":[]}";

  // set request for execute task
  DRCEInputJsonMessage inputJsonMessage;
  inputJsonMessage.setRequestType(DRCETaskRequest::RequestType::rtSetTaskExecute);
  inputJsonMessage.setRequestData(requestJson);
  inputJsonMessage.setRequestId(requestId);

   std::stringstream sourceJson;
   sourceJson << inputJsonMessage;

//  std::cout << "source JSON: " << sourceJson.str() << std::endl;

  ASSERT_FALSE(inputJsonMessage.isError()) << printError(inputJsonMessage);
  ASSERT_EQ(inputJsonMessage.getErrorCode(), NO_ERROR);
  ASSERT_TRUE(inputJsonMessage.getErrorMsg().empty());
  printSuccess("DRCEInputJsonMessage serialize");

  std::string resultJson = fObj.Process(sourceJson.str());

  ASSERT_FALSE(fObj.isError()) << printError(fObj);
  ASSERT_EQ(fObj.getErrorCode(), NO_ERROR);
  ASSERT_TRUE(fObj.getErrorMsg().empty());
  printSuccess("DRCEFunctionalObject Process");

  while (fObj.getAsyncTasksCount())
  {
    Poco::Thread::sleep(5000);
    std::cout << "wait the end ...." << std::endl;
  }
  Poco::Thread::sleep(2000);

  std::string json;
  DRCETaskRequestDeleteData taskRequestDeleteData;
  taskRequestDeleteData.serialize(json);
  ASSERT_FALSE(taskRequestDeleteData.isError());
  printSuccess("DRCETaskRequestDeleteData serialize");

//  std::cout << "json: " << json << std::endl;

  // set request for execute task
  inputJsonMessage.clear();
  inputJsonMessage.setRequestType(DRCETaskRequest::RequestType::rtDeleteTaskData);
  inputJsonMessage.setRequestData(json);
  inputJsonMessage.setRequestId(requestId);

  sourceJson.str("");
  sourceJson << inputJsonMessage;

//  std::cout << "source JSON: " << sourceJson.str() << std::endl;

  ASSERT_FALSE(inputJsonMessage.isError()) << printError(inputJsonMessage);
  ASSERT_EQ(inputJsonMessage.getErrorCode(), NO_ERROR);
  ASSERT_TRUE(inputJsonMessage.getErrorMsg().empty());
  printSuccess("DRCEInputJsonMessage serialize");

  resultJson = fObj.Process(sourceJson.str());

  ASSERT_FALSE(fObj.isError()) << printError(fObj);
  ASSERT_EQ(fObj.getErrorCode(), NO_ERROR);
  ASSERT_TRUE(fObj.getErrorMsg().empty());
  printSuccess("DRCEFunctionalObject Process");

//  std::cout << "result JSON: " << resultJson << std::endl;

  // check right result (unserialize and compare two objects)
  DRCEOutputJsonMessage resultJsonMessage(resultJson);

  ASSERT_FALSE(resultJsonMessage.isError()) << printError(resultJsonMessage);
  ASSERT_EQ(resultJsonMessage.getErrorCode(), NO_ERROR);
  ASSERT_TRUE(resultJsonMessage.getErrorMsg().empty());
  printSuccess("DRCEOutputJsonMessage unserialize");

  if (!resultJsonMessage.getResultData().getDataItem(0).getErrorMessage().empty())
    std::cout << "getErrorMessage: " << resultJsonMessage.getResultData().getDataItem(0).getErrorMessage() << std::endl;

  ASSERT_EQ(resultJsonMessage.getResultData().getItemsCount(), 1);
  ASSERT_EQ(resultJsonMessage.getResultData().getDataItem(0).getErrorCode(), NO_ERROR);
  ASSERT_TRUE(resultJsonMessage.getResultData().getDataItem(0).getErrorMessage().empty());
  ASSERT_TRUE(resultJsonMessage.getResultData().getDataItem(0).getStdoutStream().empty());
  ASSERT_TRUE(resultJsonMessage.getResultData().getDataItem(0).getStderrStream().empty());
  ASSERT_EQ(resultJsonMessage.getResultData().getDataItem(0).getState(), static_cast<unsigned int>(HCE::drce::DRCETaskRequest::TaskState::DELETED))
  << "State: " << resultJsonMessage.getResultData().getDataItem(0).getState();
  ASSERT_EQ(resultJsonMessage.getResultData().getDataItem(0).getRequestId(), requestId);
  ASSERT_EQ(resultJsonMessage.getResultData().getDataItem(0).getRequestType(), HCE::drce::DRCETaskRequest::RequestType::rtDeleteTaskData);
  ASSERT_EQ(resultJsonMessage.getResultData().getDataItem(0).getFilesCount(), 0);
  ASSERT_EQ(resultJsonMessage.getResultData().getDataItem(0).getExitStatus(), 0);
  ASSERT_EQ(resultJsonMessage.getResultData().getDataItem(0).getNodeName(), fObj.getNodeName());
  ASSERT_EQ(resultJsonMessage.getResultData().getDataItem(0).getNodeHost(), fObj.getNodeHost());
  ASSERT_EQ(resultJsonMessage.getResultData().getDataItem(0).getNodePort(), fObj.getNodePort());

//  std::cout << "log: " << fObj.logMsg() << std::endl;
  DRCECleanupTest::cleanupTestData(fObj);
  printSuccess("!!! Delete task data request");
}
//-----------------------------------------------------------------------------
//-----------------------------------------------------------------------------
void DRCETaskThreadingModeTest::testExecuteNotification(DRCEFunctionalObject& fObj, unsigned int requestId)
{
  fObj.setTasksQueueDumpPeriod(1000);
  fObj.setResourceMonitorTimePeriod(1000);
  fObj.setNodeHost("NODE_HOST");
  fObj.setNodePort("NODE_PORT");
  fObj.setNodeName("NODE_NAME");
  fObj.setNotificationFunctor(testFunctor);
//  DRCETaskThreadingModeTest::testSetTaskToExecuteRequest(fObj, requestId);

  const std::string requestJson = "{\"session\":{\"type\":0,\"tmode\":2,\"time_max\":90000,\"port\":0,\"user\":\"\",\"password\":\"\",\"shell\":\"\",\"environment\":[{\"LINES\":\"80\",\"HOSTNAME\":\"node22\"}]},\"command\":\"echo \\\"scale=4000; 4*a(1)\\\" | bc -l && sleep 5 && ps xau\",\"input\":\"\",\"files\":[]}";

  const unsigned int repeat = 1;
  fObj.setMaxThreadCount(repeat);
  for (size_t i=0;i<repeat;++i)
  {
    std::cout << ">>>>> " << i << ") test enter !!!!!\n";
    DRCETaskThreadingModeTest::testSetTaskToExecuteRequest(fObj, (requestId+i), requestJson);
//    DRCETaskThreadingModeTest::testExecuteSyncRequest(fObj, (requestId+i));
    std::cout << ">>>>> " << i << ") test leave !!!!!\n";

    std::cout << "fObj.getAsyncTasksCount(): " << fObj.getAsyncTasksCount() << std::endl;
    Poco::Thread::sleep(500);
  }
//  Poco::Thread::sleep(2000);
  std::stringstream logMsg;
  while (fObj.getAsyncTasksCount())
  {
    Poco::Thread::sleep(5000);
//    logMsg << fObj.logMsg();
//    std::cout << "LOG: " << fObj.logMsg() << std::endl;
  }
  Poco::Thread::sleep(2000);
//  std::cout << "log: " << logMsg.str() << std::endl;

//  ASSERT_EQ(count, (2*repeat));
//  ASSERT_FALSE(logMsg.str().empty());

  DRCECleanupTest::cleanupTestData(fObj);
  printSuccess("!!! Test Execute Notification Functor");
}
//-----------------------------------------------------------------------------
//-----------------------------------------------------------------------------
void DRCETaskThreadingModeTest::testTerminateExpiredTaskSync(DRCEFunctionalObject& fObj, unsigned int requestId)
{
  std::string requestJson = "{\"session\":{\"type\":0,\"tmode\":1,\"time_max\":8000,\"port\":0,\"user\":\"\",\"password\":\"\",\"shell\":\"\",\"environment\":[{}]},\"command\":\"echo \\\"Hello\\\" && sleep 10 && ls -la && sleep 10 && ps auxf\",\"input\":\"\",\"files\":[]}";

  fObj.setNodeHost("NODE_HOST");
  fObj.setNodePort("NODE_PORT");
  fObj.setNodeName("NODE_NAME");
  fObj.setNotificationFunctor(testFunctor);

  DRCETaskThreadingModeTest::testSetTaskToExecuteRequest(fObj, requestId, requestJson);

  std::stringstream logMsg;
  while (fObj.getAsyncTasksCount())
  {
    Poco::Thread::sleep(1000);
//    logMsg << fObj.logMsg();
  }
//  std::cout << "log: " << logMsg.str() << std::endl;
//  ASSERT_FALSE(logMsg.str().empty());

  DRCECleanupTest::cleanupTestData(fObj);
  printSuccess("!!! Test Terminate Expired Sync Task");
}
//-----------------------------------------------------------------------------
void DRCETaskThreadingModeTest::testTerminateExpiredTaskAsync(DRCEFunctionalObject& fObj, unsigned int requestId)
{
  std::string requestJson = "{\"session\":{\"type\":0,\"tmode\":2,\"cleanup\":1,\"time_max\":8000,\"port\":0,\"user\":\"\",\"password\":\"\",\"shell\":\"\",\"environment\":[{}]},\"command\":\"echo \\\"Hello\\\" && sleep 20 && ls -la && sleep 10 && ps auxf\",\"input\":\"\",\"files\":[]}";

  fObj.setNodeHost("NODE_HOST");
  fObj.setNodePort("NODE_PORT");
  fObj.setNodeName("NODE_NAME");
  fObj.setNotificationFunctor(testFunctor);

  for (size_t i=0;i<3;++i)
    DRCETaskThreadingModeTest::testSetTaskToExecuteRequest(fObj, requestId+i, requestJson);

  std::stringstream logMsg;
  while (fObj.getAsyncTasksCount())
  {
    Poco::Thread::sleep(1000);
//    logMsg << fObj.logMsg();
  }
//  std::cout << "log: " << logMsg.str() << std::endl;
//  ASSERT_FALSE(logMsg.str().empty());

  DRCECleanupTest::cleanupTestData(fObj);
  printSuccess("!!! Test Terminate Expired Async Task");
}
//-----------------------------------------------------------------------------
//-----------------------------------------------------------------------------
void DRCETaskThreadingModeTest::testTaskRequestTimeLimits(DRCEFunctionalObject& fObj, unsigned int requestId, SessionOptions::ThreadMode threadMode, unsigned int timeMax)
{
  DRCETaskRequestSetExecute taskRequestSetExecute;

  SessionOptions sessionOptions;
  sessionOptions.sessionType = SessionOptions::SessionType::stHostShell;
  sessionOptions.homeDir = fObj.getHomeDir();
  sessionOptions.tmode = threadMode;
  sessionOptions.timeMax = timeMax;
  sessionOptions.cleanup = SessionOptions::CleanupFlag::cfNotDelete;

  taskRequestSetExecute.setSessionOptions(std::forward<SessionOptions>(sessionOptions));
  taskRequestSetExecute.setCommandLine("sort");
  taskRequestSetExecute.setInputStream("def\nabc\n123\n");

  std::stringstream sourceJson;
  sourceJson << taskRequestSetExecute;

  ASSERT_FALSE(taskRequestSetExecute.isError()) << printError(taskRequestSetExecute);
  ASSERT_EQ(taskRequestSetExecute.getErrorCode(), NO_ERROR);
  ASSERT_TRUE(taskRequestSetExecute.getErrorMsg().empty());
  printSuccess("DRCETaskRequestSetExecute serialize");

  DRCEInputJsonMessage inputJsonMessage;
  inputJsonMessage.setRequestType(DRCETaskRequest::RequestType::rtSetTaskExecute);
  inputJsonMessage.setRequestData(sourceJson.str());
  inputJsonMessage.setRequestId(requestId);

  sourceJson.str("");
  sourceJson << inputJsonMessage;

  ASSERT_FALSE(inputJsonMessage.isError()) << printError(inputJsonMessage);
  ASSERT_EQ(inputJsonMessage.getErrorCode(), NO_ERROR);
  ASSERT_TRUE(inputJsonMessage.getErrorMsg().empty());
  printSuccess("DRCEInputJsonMessage serialize");

//  std::cout << "IN JSON: " << sourceJson.str() << std::endl;

  std::string resultJson = fObj.Process(sourceJson.str());

  ASSERT_FALSE(fObj.isError()) << printError(fObj);
  ASSERT_EQ(fObj.getErrorCode(), NO_ERROR);
  ASSERT_TRUE(fObj.getErrorMsg().empty());
  printSuccess("DRCEFunctionalObject Process");

//  std::cout << "OUT JSON: " << resultJson << std::endl;

  DRCEOutputJsonMessage resultJsonMessage(resultJson);

  ASSERT_FALSE(resultJsonMessage.isError()) << printError(resultJsonMessage);
  ASSERT_EQ(resultJsonMessage.getErrorCode(), NO_ERROR);
  ASSERT_TRUE(resultJsonMessage.getErrorMsg().empty());
  printSuccess("DRCEOutputJsonMessage unserialize");

  ASSERT_EQ(resultJsonMessage.getResultData().getItemsCount(), 1);
  if (threadMode==SessionOptions::ThreadMode::tmSync)
  {
    ASSERT_EQ(resultJsonMessage.getResultData().getDataItem(0).getStdoutStream(), "123\nabc\ndef\n");
    ASSERT_EQ(resultJsonMessage.getResultData().getDataItem(0).getState(), DRCETaskRequest::TaskState::FINISHED);
  }
  else
  {
    if (resultJsonMessage.getResultData().getDataItem(0).getState() == DRCETaskRequest::TaskState::FINISHED)
      ASSERT_EQ(resultJsonMessage.getResultData().getDataItem(0).getStdoutStream(), "123\nabc\ndef\n");
    else
    {
      ASSERT_EQ(resultJsonMessage.getResultData().getDataItem(0).getStdoutStream(), "");
      ASSERT_TRUE(resultJsonMessage.getResultData().getDataItem(0).getState()==DRCETaskRequest::TaskState::SET_AS_NEW ||
          resultJsonMessage.getResultData().getDataItem(0).getState()==DRCETaskRequest::TaskState::IN_PROGRESS);
    }
  }

  ASSERT_EQ(resultJsonMessage.getResultData().getDataItem(0).getStderrStream(), "");
  ASSERT_EQ(resultJsonMessage.getResultData().getDataItem(0).getErrorCode(), 0);
  ASSERT_EQ(resultJsonMessage.getResultData().getDataItem(0).getErrorMessage(), "");
  ASSERT_EQ(resultJsonMessage.getResultData().getDataItem(0).getExitStatus(), 0);
  ASSERT_EQ(resultJsonMessage.getResultData().getDataItem(0).getNodeHost(), fObj.getNodeHost());
  ASSERT_EQ(resultJsonMessage.getResultData().getDataItem(0).getNodeName(), fObj.getNodeName());
  ASSERT_EQ(resultJsonMessage.getResultData().getDataItem(0).getNodePort(), fObj.getNodePort());
  ASSERT_EQ(resultJsonMessage.getResultData().getDataItem(0).getRequestId(), requestId);
  ASSERT_EQ(resultJsonMessage.getResultData().getDataItem(0).getRequestType(), DRCETaskRequest::RequestType::rtSetTaskExecute);

  if (threadMode==SessionOptions::ThreadMode::tmAsync)
  {
    while (fObj.getAsyncTasksCount() || fObj.hasTask(requestId))
    {
      Poco::Thread::sleep(5000);
      std::cout << "wait the end ...." << std::endl;
    }
    Poco::Thread::sleep(2000);

    DRCEResultDataItem resultDataItem = DRCECleanupTest::extractResultData(requestId, fObj);

    ASSERT_EQ(resultDataItem.getStdoutStream(), "123\nabc\ndef\n");
  }
  DRCECleanupTest::cleanupTestData(fObj);
  printSuccess("Test Request Time Limits");
}
//-----------------------------------------------------------------------------
//-----------------------------------------------------------------------------
void DRCETaskThreadingModeTest::testParallelAsyncExecution(DRCEFunctionalObject& fObj, unsigned int requestId, unsigned int repeatCount)
{
  const std::string requestJson = "{\"session\":{\"type\":0,\"tmode\":2,\"cleanup\":1,\"time_max\":120000,\"port\":0,\"user\":\"\",\"password\":\"\",\"shell\":\"\",\"environment\":[]},\"command\":\"echo \\\"scale=4000; 4*a(1)\\\" | bc -l && sleep 60 && ps xau\",\"input\":\"\",\"files\":[]}";

  fObj.setMaxThreadCount(repeatCount);
  fObj.setResourceMonitorTimePeriod(10000);
//  fObj.setLogStream(std::cout);

  Poco::Timestamp now;

  for (unsigned int i=0;i<repeatCount;++i)
  {
//    std::cout << i << ") test enter\n";
    DRCETaskThreadingModeTest::testSetTaskToExecuteRequest(fObj, (i+requestId), requestJson);
//    std::cout << i << ") test leave\n===========\n";
//    std::cout << "fObj.getAsyncTasksCount(): " << fObj.getAsyncTasksCount() << std::endl;
  }

  while (fObj.getAsyncTasksCount())
  {
    Poco::Thread::sleep(5000);
//    std::cout << "wait the end ...." << std::endl;
  }
  Poco::Thread::sleep(2000);
  Poco::Timestamp::TimeDiff diff = now.elapsed();

  std::cout << "Count tasks = " << repeatCount << std::endl;
  std::cout << "All time: " << (diff/1000) << " msec." << std::endl;
  DRCECleanupTest::cleanupTestData(fObj);
}
//-----------------------------------------------------------------------------
void DRCETaskThreadingModeTest::testMaxThreadLimits(DRCEFunctionalObject& fObj, unsigned int requestId, unsigned int maxAllowedThreadsCount, unsigned int repeatCount)
{
//  const std::string requestJson = "{\"session\":{\"type\":0,\"tmode\":2,\"time_max\":40000,\"port\":0,\"user\":\"\",\"password\":\"\",\"shell\":\"\",\"environment\":[]},\"command\":\"sort\",\"input\":\"def\\nabc\\n123\\n\",\"files\":[]}";
  const std::string requestJson = "{\"session\":{\"type\":0,\"tmode\":2,\"cleanup\":1,\"time_max\":40000,\"port\":0,\"user\":\"\",\"password\":\"\",\"shell\":\"\",\"environment\":[]},\"command\":\"sleep 20\",\"input\":\"\",\"files\":[]}";

//  fObj.setNotificationFunctor(testFunctor);

  fObj.setMaxThreadCount(maxAllowedThreadsCount);
  Poco::Timestamp now;

  for (unsigned int i=0;i<repeatCount;++i)
  {
    std::cout << i << ") test enter\n";
    DRCETaskThreadingModeTest::testSetTaskToExecuteRequest(fObj, (i+requestId), requestJson);
    std::cout << i << ") test leave\n===========\n";
    std::cout << "fObj.getAsyncTasksCount(): " << fObj.getAsyncTasksCount() << std::endl;
  }

  while (fObj.getAsyncTasksCount())
  {
    Poco::Thread::sleep(5000);
    std::cout << "wait the end ...." << std::endl;
  }
  Poco::Thread::sleep(2000);
//  std::cout << "log: " << fObj.logMsg() << std::endl;

  Poco::Timestamp::TimeDiff diff = now.elapsed();

  std::cout << "All threads = " << repeatCount << " each can sleep 20 sec."<< std::endl;
  std::cout << "All time: " << (diff/1000) << " msec." << std::endl;
  DRCECleanupTest::cleanupTestData(fObj);
}
//-----------------------------------------------------------------------------
//-----------------------------------------------------------------------------
void DRCETaskThreadingModeTest::testRequestLocalShellAsyncCleanUp(void)
{
  DRCENodeOptions nodeOptions = DRCETaskThreadingModeTest::getNodeOptions();
  DRCEFunctionalObject fObj(nodeOptions.getNodeName(), nodeOptions.getHomeDir(), nodeOptions.getTasksDataDir(), nodeOptions.getTasksStatusDir());

//  const std::string requestJson = "{\"session\":{\"type\":0,\"tmode\":1,\"time_max\":0,\"cleanup\":1,\"port\":0,\"user\":\"\",\"password\":\"\",\"shell\":\"\",\"environment\":[{\"LINES\":\"80\",\"HOSTNAME\":\"node22\"}]},\"command\":\"sort\",\"input\":\"def\\nabc\\n123\\n\",\"files\":[{\"name\":\"file1\",\"data\":\"test content of file1\\n\",\"action\":13},{\"name\":\"file2\",\"data\":\"test content of file2\\n\",\"action\":1}]}";
  const std::string requestJson = "{\"session\":{\"type\":0,\"tmode\":2,\"time_max\":40000,\"cleanup\":1,\"port\":0,\"user\":\"\",\"password\":\"\",\"shell\":\"\",\"environment\":[{\"LINES\":\"80\",\"HOSTNAME\":\"node22\"}]},\"command\":\"sort\",\"input\":\"def\\nabc\\n123\\n\",\"files\":[{\"name\":\"file1\",\"data\":\"test content of file1\\n\",\"action\":13},{\"name\":\"file2\",\"data\":\"test content of file2\\n\",\"action\":1}]}";
//  const std::string requestJson = "{\"session\":{\"type\":0,\"tmode\":2,\"time_max\":40000,\"port\":0,\"user\":\"\",\"password\":\"\",\"shell\":\"\",\"environment\":[{\"LINES\":\"80\",\"HOSTNAME\":\"node22\"}]},\"command\":\"sort\",\"input\":\"def\\nabc\\n123\\n\",\"files\":[{\"name\":\"file1\",\"data\":\"test content of file1\\n\",\"action\":13},{\"name\":\"file2\",\"data\":\"test content of file2\\n\",\"action\":1}]}";
  DRCETaskThreadingModeTest::testSetTaskToExecuteRequest(fObj, 567, requestJson);

  while (fObj.getAsyncTasksCount())
  {
    Poco::Thread::sleep(2000);
    std::cout << "wait the end ...." << std::endl;
  }
  Poco::Thread::sleep(5000);

  bool res = DRCETaskThreadingModeTest::isExistTask(fObj, 567);
  ASSERT_EQ(res, false);

  DRCECleanupTest::cleanupTestData(fObj);
  printSuccess("!!! Test request local shell async cleanup");
}
//-----------------------------------------------------------------------------
void DRCETaskThreadingModeTest::testTerminateTaskRequestCleanUp(void)
{
  DRCENodeOptions nodeOptions = DRCETaskThreadingModeTest::getNodeOptions();
  DRCEFunctionalObject fObj(nodeOptions.getNodeName(), nodeOptions.getHomeDir(), nodeOptions.getTasksDataDir(), nodeOptions.getTasksStatusDir());
  fObj.setNodeHost("NODE_HOST");
  fObj.setNodePort("NODE_PORT");
  fObj.setNodeName("NODE_NAME");
  fObj.setTasksQueueDumpPeriod(1000);
  const unsigned int taskId = 567;

  DRCETaskThreadingModeTest::testTerminateTaskRequest(fObj, taskId, DRCETaskRequestTerminate::AlgorithmType::atDefault, DRCETaskRequestTerminate::CleanupFlag::cfNotDelete);
  bool res = DRCETaskThreadingModeTest::isExistTask(fObj, taskId);
  ASSERT_EQ(res, true);

  DRCETaskThreadingModeTest::testTerminateTaskRequest(fObj, taskId, DRCETaskRequestTerminate::AlgorithmType::atDefault, DRCETaskRequestTerminate::CleanupFlag::cfDelete);
  int waitRepeat = 10;
  while (res != false && --waitRepeat>0)
  {
    Poco::Thread::sleep(5000);
    res = DRCETaskThreadingModeTest::isExistTask(fObj, taskId);
    std::cout << "task " << taskId << " awhile exist...\n";
  }
  ASSERT_EQ(res, false);

  DRCECleanupTest::cleanupTestData(fObj);
  printSuccess("!!! Test Terminate task request with cleanup");
}
//-----------------------------------------------------------------------------
//-----------------------------------------------------------------------------
void DRCETaskThreadingModeTest::testRequestLocalShellAsync(void)
{
  DRCENodeOptions nodeOptions = DRCETaskThreadingModeTest::getNodeOptions();
  DRCEFunctionalObject fObj(nodeOptions.getNodeName(), nodeOptions.getHomeDir(), nodeOptions.getTasksDataDir(), nodeOptions.getTasksStatusDir());
  fObj.setNodeHost("NODE_HOST");
  fObj.setNodePort("NODE_PORT");
  fObj.setNodeName("NODE_NAME");

  fObj.setNotificationFunctor(testFunctor);
  fObj.setTasksQueueDumpPeriod(1000);

  unsigned int taskId = 567;

  DRCETaskThreadingModeTest treadingModetest;
  treadingModetest.setNotificationFunctor(treadingModetest);
  fObj.setNotificationFunctor(treadingModetest);

  DRCETaskThreadingModeTest::testExecuteSyncRequest(fObj, taskId);

//  std::string requestJson = "{\"session\":{\"type\":0,\"cleanup\":0,\"tmode\":2,\"time_max\":200,\"port\":0,\"user\":\"\",\"password\":\"\",\"shell\":\"\",\"environment\":[{}]},\"command\":\"/home/alexander/workspace/sleeper/Debug/sleeper 10\",\"input\":\"\",\"files\":[]}";

//  const std::string requestJson = "{\"session\":{\"type\":0,\"cleanup\":0,\"tmode\":2,\"time_max\":10000,\"port\":0,\"user\":\"\",\"password\":\"\",\"shell\":\"\",\"environment\":[{\"LINES\":\"120\",\"PAGES\":\"52\"}]},\"command\":\"env\",\"input\":\"\",\"files\":[]}";
//  const std::string enviroments = "{\"DRCE_NODE\":\"1\",\"DRCE_NODE_COUNT\":\"4\"}";
//  fObj.setEnvironments(enviroments);

  DRCETaskThreadingModeTest::testSetTaskToExecuteRequest(fObj, taskId);

  while (fObj.getAsyncTasksCount())
  {
    Poco::Thread::sleep(5000);
    std::cout << "wait the end ...." << std::endl;
  }
  Poco::Thread::sleep(5000);
//  std::cout << "log: " << fObj.logMsg() << std::endl;

  DRCETaskThreadingModeTest::testGetTaskDataRequest(fObj, taskId);

  DRCETaskThreadingModeTest::testCheckTaskRequest(fObj, taskId);

  DRCETaskThreadingModeTest::testExecuteNotification(fObj, taskId);

  DRCETaskThreadingModeTest::testTerminateTaskRequest(fObj, taskId, DRCETaskRequestTerminate::AlgorithmType::atDefault, DRCETaskRequestTerminate::CleanupFlag::cfDelete);
  DRCETaskThreadingModeTest::testTerminateTaskRequest(fObj, taskId, DRCETaskRequestTerminate::AlgorithmType::atCustom);

  DRCETaskThreadingModeTest::testDeleteTaskDataRequest(fObj, taskId);

  DRCETaskThreadingModeTest::testTerminateExpiredTaskSync(fObj, taskId);
  DRCETaskThreadingModeTest::testTerminateExpiredTaskAsync(fObj, taskId);

  DRCETaskThreadingModeTest::testParallelAsyncExecution(fObj, taskId, 1);
  DRCETaskThreadingModeTest::testMaxThreadLimits(fObj, taskId, 10, 4);

  DRCETaskThreadingModeTest::testTaskRequestTimeLimits(fObj, taskId, SessionOptions::ThreadMode::tmSync, 40000);
  DRCETaskThreadingModeTest::testTaskRequestTimeLimits(fObj, taskId, SessionOptions::ThreadMode::tmAsync, 40000);
  DRCETaskThreadingModeTest::testTaskRequestTimeLimits(fObj, taskId, SessionOptions::ThreadMode::tmSync, 0);
  DRCETaskThreadingModeTest::testTaskRequestTimeLimits(fObj, taskId, SessionOptions::ThreadMode::tmAsync, 0);

  DRCETaskThreadingModeTest::testTaskRequestExpiredTime(fObj, taskId, SessionOptions::ThreadMode::tmSync, SessionOptions::CleanupFlag::cfNotDelete);
  DRCETaskThreadingModeTest::testTaskRequestExpiredTime(fObj, taskId, SessionOptions::ThreadMode::tmAsync, SessionOptions::CleanupFlag::cfNotDelete);
  DRCETaskThreadingModeTest::testTaskRequestExpiredTime(fObj, taskId, SessionOptions::ThreadMode::tmSync, SessionOptions::CleanupFlag::cfDelete);
  DRCETaskThreadingModeTest::testTaskRequestExpiredTime(fObj, taskId, SessionOptions::ThreadMode::tmAsync, SessionOptions::CleanupFlag::cfDelete);
}
//-----------------------------------------------------------------------------
void DRCETaskThreadingModeTest::testStressMultipleRequestsShellHost(void)
{
  DRCENodeOptions nodeOptions = DRCETaskThreadingModeTest::getNodeOptions();
  DRCECleanupTest::cleanupTestData(nodeOptions);
  DRCEFunctionalObject fObj(nodeOptions.getNodeName(), nodeOptions.getHomeDir(), nodeOptions.getTasksDataDir(), nodeOptions.getTasksStatusDir());

  unsigned int requestCount = 10;
  DRCEStressTest::testMultipleRequests(fObj, SessionOptions::SessionType::stHostShell, requestCount);
  DRCEStressTest::testRequestsTimeDelay(fObj, SessionOptions::SessionType::stHostShell, requestCount, requestCount);
  DRCEStressTest::testResourceLimits(fObj, SessionOptions::SessionType::stHostShell, SessionOptions::ThreadMode::tmAsync);
  DRCEStressTest::testResourceLimits(fObj, SessionOptions::SessionType::stHostShell, SessionOptions::ThreadMode::tmSync);
  DRCEStressTest::testResourceUsageLimits(fObj, SessionOptions::SessionType::stHostShell, SessionOptions::ThreadMode::tmAsync);
  DRCEStressTest::testResourceUsageLimits(fObj, SessionOptions::SessionType::stHostShell, SessionOptions::ThreadMode::tmSync);

  DRCECleanupTest::cleanupTestData(nodeOptions);
}
//-----------------------------------------------------------------------------
void DRCETaskThreadingModeTest::testResourceMonitor(void)
{
  DRCENodeOptions nodeOptions = DRCETaskThreadingModeTest::getNodeOptions();
  DRCECleanupTest::cleanupTestData(nodeOptions);
  DRCEFunctionalObject fObj(nodeOptions.getNodeName(), nodeOptions.getHomeDir(), nodeOptions.getTasksDataDir(), nodeOptions.getTasksStatusDir());
  fObj.setNodeHost("NODE_HOST");
  fObj.setNodePort("NODE_PORT");
  fObj.setNodeName("NODE_NAME");
  fObj.setResourceMonitorTimePeriod(1000);
//  fObj.setNotificationFunctor(testFunctor);

  const std::string syncRequestJson = "{\"session\":{\"type\":0,\"tmode\":1,\"time_max\":40000,\"port\":0,\"user\":\"\",\"password\":\"\",\"shell\":\"\",\"environment\":[{\"LINES\":\"80\",\"HOSTNAME\":\"node22\"}]},\"limits\":{\"proc_max\":400},\"command\":\"sort\",\"input\":\"def\\nabc\\n123\\n\",\"files\":[{\"name\":\"file1\",\"data\":\"test content of file1\\n\",\"action\":5},{\"name\":\"file2\",\"data\":\"test content of file2\\n\",\"action\":1}]}";
  const std::string shortRequestJson = "{\"session\":{\"type\":0,\"tmode\":2,\"time_max\":40000,\"port\":0,\"user\":\"\",\"password\":\"\",\"shell\":\"\",\"environment\":[{\"LINES\":\"80\",\"HOSTNAME\":\"node22\"}]},\"command\":\"sort\",\"input\":\"def\\nabc\\n123\\n\",\"files\":[{\"name\":\"file1\",\"data\":\"test content of file1\\n\",\"action\":5},{\"name\":\"file2\",\"data\":\"test content of file2\\n\",\"action\":1}]}";
  const std::string expiredRequestJson = "{\"session\":{\"type\":0,\"tmode\":2,\"time_max\":5000,\"port\":0,\"user\":\"\",\"password\":\"\",\"shell\":\"\",\"environment\":[{}]},\"command\":\"sleep 20\",\"input\":\"\",\"files\":[{\"name\":\"file1\",\"data\":\"test content of file1\\n\",\"action\":5},{\"name\":\"file2\",\"data\":\"test content of file2\\n\",\"action\":1}]}";
  const std::string longRequestJson = "{\"session\":{\"type\":0,\"tmode\":2,\"cleanup\":1,\"time_max\":120000,\"port\":0,\"user\":\"\",\"password\":\"\",\"shell\":\"\",\"environment\":[]},\"command\":\"echo \\\"scale=4000; 4*a(1)\\\" | bc -l && sleep 60 && ps xau\",\"input\":\"\",\"files\":[]}";
  const std::string bcJson = "{\"session\":{\"type\":0,\"tmode\":2,\"cleanup\":1,\"time_max\":120000,\"port\":0,\"user\":\"\",\"password\":\"\",\"shell\":\"\",\"environment\":[]},\"command\":\"echo \\\"scale=4000; 4*a(1)\\\" | bc -l && echo \\\"scale=400; 4*a(1)\\\" | bc -l && sleep 10\",\"input\":\"\",\"files\":[]}";

  const size_t requestCount = 10;

  const unsigned int offsetTaskId = 500;
  std::vector<std::string> jsonList = /*{bcJson, bcJson, bcJson, bcJson};*/{syncRequestJson, shortRequestJson, expiredRequestJson, longRequestJson};
  std::vector<std::pair<unsigned int, std::string> > tasks;
  unsigned int count = 0;
  while(count < requestCount)
  {
    for (size_t k=0;k<jsonList.size();++k)
    {
      tasks.push_back(std::pair<unsigned int, std::string>(++count+offsetTaskId, jsonList[k]));
      if (count >= requestCount)
        break;
    }
  }

  for (size_t i=0;i<requestCount;++i)
  {
    DRCETaskThreadingModeTest::testSetTaskToExecuteRequest(fObj, tasks[i].first, tasks[i].second);
  }

  do
  {
    Poco::Thread::sleep(5000);

    StatVariables vars = fObj.getStatVariables();
    std::cout << std::string(80, '=') << std::endl;

    std::cout << "             timeAsyncTask: " << vars.timeAsyncTask.min << "\t" << vars.timeAsyncTask.avg << "\t" << vars.timeAsyncTask.max << std::endl;
    std::cout << "              timeSyncTask: " << vars.timeSyncTask.min << "\t" << vars.timeSyncTask.avg << "\t" << vars.timeSyncTask.max << std::endl;
    std::cout << "  sizeInputBufferSyncTasks: " << vars.sizeInputBufferSyncTasks.min << "\t" << vars.sizeInputBufferSyncTasks.avg << "\t" << vars.sizeInputBufferSyncTasks.max << std::endl;
    std::cout << " sizeOutputBufferSyncTasks: " << vars.sizeOutputBufferSyncTasks.min << "\t" << vars.sizeOutputBufferSyncTasks.avg << "\t" << vars.sizeOutputBufferSyncTasks.max << std::endl;
    std::cout << " sizeInputBufferAsyncTasks: " << vars.sizeInputBufferAsyncTasks.min << "\t" << vars.sizeInputBufferAsyncTasks.avg << "\t" << vars.sizeInputBufferAsyncTasks.max << std::endl;
    std::cout << "sizeOutputBufferAsyncTasks: " << vars.sizeOutputBufferAsyncTasks.min << "\t" << vars.sizeOutputBufferAsyncTasks.avg << "\t" << vars.sizeOutputBufferAsyncTasks.max << std::endl;
    std::cout << "        cpuUsageAsyncTasks: " << vars.cpuUsageAsyncTasks.min << "\t" << vars.cpuUsageAsyncTasks.avg << "\t" << vars.cpuUsageAsyncTasks.max << std::endl;
    std::cout << "     memoryUsageAsyncTasks: " << vars.memoryUsageAsyncTasks.min << "\t" << vars.memoryUsageAsyncTasks.avg << "\t" << vars.memoryUsageAsyncTasks.max << std::endl;
    std::cout << "   countSyncTasksForMinute: " << vars.countSyncTasksForMinute.min << "\t" << vars.countSyncTasksForMinute.avg << "\t" << vars.countSyncTasksForMinute.max << std::endl;
    std::cout << "  countAsyncTasksForMinute: " << vars.countAsyncTasksForMinute.min << "\t" << vars.countAsyncTasksForMinute.avg << "\t" << vars.countAsyncTasksForMinute.max << std::endl;

    std::cout << "            countSyncTasks: " << vars.countSyncTasks << std::endl;
    std::cout << "           countAsyncTasks: " << vars.countAsyncTasks << std::endl;
    std::cout << "        countSyncTasksFail: " << vars.countSyncTasksFail << std::endl;
    std::cout << "       countAsyncTasksFail: " << vars.countAsyncTasksFail << std::endl;

    std::cout << std::string(80, '=') << std::endl;

  }while (fObj.getAsyncTasksCount());

  DRCECleanupTest::cleanupTestData(nodeOptions);
}
//-----------------------------------------------------------------------------
void DRCETaskThreadingModeTest::testCurrentTasksQueue(void)
{
  DRCENodeOptions nodeOptions = DRCETaskThreadingModeTest::getNodeOptions();
  DRCEFunctionalObject fObj(nodeOptions.getNodeName(), nodeOptions.getHomeDir(), nodeOptions.getTasksDataDir(), nodeOptions.getTasksStatusDir());
  fObj.setNodeHost("NODE_HOST");
  fObj.setNodePort("NODE_PORT");
  fObj.setNodeName("NODE_NAME");
  const std::string requestJson = "{\"session\":{\"type\":0,\"tmode\":2,\"cleanup\":0,\"time_max\":120000,\"port\":0,\"user\":\"\",\"password\":\"\",\"shell\":\"\",\"environment\":[]},\"command\":\"echo \\\"scale=40; 4*a(1)\\\" | bc -l && sleep 10 && ps xau\",\"input\":\"\",\"files\":[]}";
//  const std::string requestJson = "{\"session\":{\"type\":0,\"tmode\":2,\"time_max\":40000,\"port\":0,\"user\":\"\",\"password\":\"\",\"shell\":\"\",\"environment\":[]},\"command\":\"sort\",\"input\":\"def\\nabc\\n123\\n\",\"files\":[]}";


  const unsigned int requestCount = 5;
  const unsigned int startTaskId = 500;

  for (unsigned int i=0;i<requestCount;++i)
  {
    // set request for execute task
    DRCEInputJsonMessage inputJsonMessage;
    inputJsonMessage.setRequestType(DRCETaskRequest::RequestType::rtSetTaskExecute);
    inputJsonMessage.setRequestData(requestJson);
    inputJsonMessage.setRequestId(startTaskId+i);

    std::stringstream sourceJson;
    sourceJson << inputJsonMessage;

    ASSERT_FALSE(inputJsonMessage.isError()) << printError(inputJsonMessage);
    ASSERT_EQ(inputJsonMessage.getErrorCode(), NO_ERROR);
    ASSERT_TRUE(inputJsonMessage.getErrorMsg().empty());

    std::string resultJson = fObj.Process(sourceJson.str());

    ASSERT_FALSE(fObj.isError()) << printError(fObj);
    ASSERT_EQ(fObj.getErrorCode(), NO_ERROR);
    ASSERT_TRUE(fObj.getErrorMsg().empty());

    Poco::Thread::sleep(100);
//    std::cout << "currentTasksQueue: " << fObj.getCurrentTasksQueue() << std::endl << std::endl;
  }

  std::string currentTasksQueue = fObj.getCurrentTasksQueue();
  DRCETasksQueue tasksQueue;
  tasksQueue.unserialize(currentTasksQueue);

  ASSERT_FALSE(tasksQueue.isError()) << printError(tasksQueue);

  ASSERT_EQ(tasksQueue.getItemsCount(), requestCount);

  for (unsigned int i=0;i<tasksQueue.getItemsCount();++i)
  {
    ASSERT_EQ(tasksQueue.getItem(i).id, (startTaskId+i));
    ASSERT_TRUE(tasksQueue.getItem(i).sdate > 0);
    ASSERT_TRUE(tasksQueue.getItem(i).state==DRCETaskRequest::TaskState::SET_AS_NEW ||
        tasksQueue.getItem(i).state==DRCETaskRequest::TaskState::IN_PROGRESS ||
        tasksQueue.getItem(i).state==DRCETaskRequest::TaskState::FINISHED||
        tasksQueue.getItem(i).state==DRCETaskRequest::TaskState::CRASHED);
    ASSERT_EQ(tasksQueue.getItem(i).tmode, SessionOptions::ThreadMode::tmAsync);
  }

  while (fObj.getAsyncTasksCount() || fObj.hasTask(startTaskId))
  {
    Poco::Thread::sleep(5000);
//    std::cout << "currentTasksQueue: " << fObj.getCurrentTasksQueue() << std::endl << std::endl;
  }
  Poco::Thread::sleep(2000);

  currentTasksQueue = fObj.getCurrentTasksQueue();
  tasksQueue.unserialize(currentTasksQueue);

  ASSERT_FALSE(tasksQueue.isError()) << printError(tasksQueue);
  ASSERT_EQ(tasksQueue.getItemsCount(), 0);

  DRCECleanupTest::cleanupTestData(nodeOptions);
  printSuccess("!!! Test Getting Current Tasks Queue");
}
//-----------------------------------------------------------------------------
void DRCETaskThreadingModeTest::testListAllTasks(void)
{
  DRCENodeOptions nodeOptions = DRCETaskThreadingModeTest::getNodeOptions();
  DRCEFunctionalObject fObj(nodeOptions.getNodeName(), nodeOptions.getHomeDir(), nodeOptions.getTasksDataDir(), nodeOptions.getTasksStatusDir());
  fObj.setNodeHost("NODE_HOST");
  fObj.setNodePort("NODE_PORT");
  fObj.setNodeName("NODE_NAME");

  const std::string requestJson = "{\"session\":{\"type\":0,\"cleanup\":0,\"tmode\":2,\"time_max\":40000,\"port\":0,\"user\":\"\",\"password\":\"\",\"shell\":\"\",\"environment\":[]},\"command\":\"sort\",\"input\":\"def\\nabc\\n123\\n\",\"files\":[]}";

  const unsigned int requestCount = 5;
  const unsigned int startTaskId = 700;

  for (unsigned int i=0;i<requestCount;++i)
  {
    // set request for execute task
    DRCEInputJsonMessage inputJsonMessage;
    inputJsonMessage.setRequestType(DRCETaskRequest::RequestType::rtSetTaskExecute);
    inputJsonMessage.setRequestData(requestJson);
    inputJsonMessage.setRequestId(startTaskId+i);

    std::stringstream sourceJson;
    sourceJson << inputJsonMessage;

    ASSERT_FALSE(inputJsonMessage.isError()) << printError(inputJsonMessage);
    ASSERT_EQ(inputJsonMessage.getErrorCode(), NO_ERROR);
    ASSERT_TRUE(inputJsonMessage.getErrorMsg().empty());

    std::string resultJson = fObj.Process(sourceJson.str());

    ASSERT_FALSE(fObj.isError()) << printError(fObj);
    ASSERT_EQ(fObj.getErrorCode(), NO_ERROR);
    ASSERT_TRUE(fObj.getErrorMsg().empty());
  }

  while (fObj.getAsyncTasksCount())
  {
    Poco::Thread::sleep(2000);
  }
  Poco::Thread::sleep(2000);
  std::string listAllTasks = fObj.getListInfoAllTasks();

//  std::cout << "listAllTasks: " << listAllTasks << std::endl << std::endl;

  DRCEListAllTasks tasks;
  tasks.unserialize(listAllTasks);

  ASSERT_FALSE(tasks.isError()) << printError(tasks);
  ASSERT_EQ(tasks.getErrorCode(), NO_ERROR) << printError(tasks);
  ASSERT_TRUE(tasks.getErrorMsg().empty()) << printError(tasks);

  ASSERT_TRUE(tasks.getItemsCount() >= requestCount) << "ItemsCount: " << tasks.getItemsCount();

  std::vector<unsigned int> ids;
  for (unsigned int i=0;i<requestCount;++i)
    ids.push_back(startTaskId+i);

  unsigned int found = 0;
  for (unsigned int i=0;i<tasks.getItemsCount();++i)
  {
    for (unsigned int k=0;k<ids.size();++k)
    {
      if (tasks.getItem(i).taskId == ids[k])
      {
        ASSERT_FALSE(tasks.getItem(i).infoData.empty());
        ASSERT_TRUE(tasks.getItem(i).timeModified > 0);
        ++found;
        break;
      }
    }
  }
  ASSERT_EQ(found, requestCount) << "Count found of tasks: " << found;

  for (unsigned int i=0;i<ids.size();++i)
  {
    bool res = DRCETaskThreadingModeTest::isExistTask(fObj, ids[i]);
    ASSERT_EQ(res, true) << "Task data not found" << ids[i];
  }

  DRCECleanupTest::cleanupTestData(nodeOptions);
  printSuccess("!!! Test Getting Current Tasks Queue");
}
//-----------------------------------------------------------------------------
//-----------------------------------------------------------------------------
void DRCETaskThreadingModeTest::testTaskRequestExpiredTime(DRCEFunctionalObject& fObj,
    unsigned int requestId,
    SessionOptions::ThreadMode threadMode,
    SessionOptions::CleanupFlag cleanupFlag)
{
  const std::string commandLine = "echo \"Hello\" && sleep 10 && ps xau && sleep 10 && ls -la";
//  const std::string commandLine = "\"scale=4000; 4*a(1)\" | bc -l && ps xau && ls -la";
  std::string sourceJson = DRCECleanupTest::getInputJson(requestId, threadMode, cleanupFlag, commandLine, 5000);

//  std::cout << "sourceJson: " << sourceJson << std::endl;

  std::string resultJson = fObj.Process(sourceJson);

  ASSERT_FALSE(fObj.isError()) << printError(fObj);
  ASSERT_EQ(fObj.getErrorCode(), NO_ERROR);
  ASSERT_TRUE(fObj.getErrorMsg().empty());
  printSuccess("DRCEFunctionalObject Process");

//  std::cout << "result JSON: " << resultJson << std::endl;

  // check right result (unserialize and compare two objects)
  DRCEOutputJsonMessage resultJsonMessage(resultJson);

  ASSERT_FALSE(resultJsonMessage.isError()) << printError(resultJsonMessage);
  ASSERT_EQ(resultJsonMessage.getErrorCode(), NO_ERROR);
  ASSERT_TRUE(resultJsonMessage.getErrorMsg().empty());
  printSuccess("DRCEOutputJsonMessage unserialize");

  if (threadMode==SessionOptions::ThreadMode::tmSync)
  {
    ASSERT_EQ(resultJsonMessage.getResultData().getItemsCount(), 1);
    ASSERT_EQ(resultJsonMessage.getResultData().getDataItem(0).getRequestId(), requestId);
    ASSERT_EQ(resultJsonMessage.getResultData().getDataItem(0).getErrorCode(), ERROR_TERMINATE_EXPIRED_TASK);
    ASSERT_EQ(resultJsonMessage.getResultData().getDataItem(0).getNodeName(), "NODE_NAME");
    ASSERT_EQ(resultJsonMessage.getResultData().getDataItem(0).getNodeHost(), "NODE_HOST");
    ASSERT_EQ(resultJsonMessage.getResultData().getDataItem(0).getNodePort(), "NODE_PORT");
    ASSERT_EQ(resultJsonMessage.getResultData().getDataItem(0).getState(), static_cast<unsigned int>(DRCETaskRequest::TaskState::TERMINATED_BY_TTL));
    ASSERT_EQ(resultJsonMessage.getResultData().getDataItem(0).getSubtasksCount(), 2);

    ASSERT_FALSE(resultJsonMessage.getResultData().getDataItem(0).getSubtaskItem(0).getErrorMessage().empty());
    ASSERT_EQ(resultJsonMessage.getResultData().getDataItem(0).getSubtaskItem(0).getErrorCode(), ERROR_TERMINATE_SUBTASK_AFTER_EXPIRED_MAIN_TASK);
    ASSERT_EQ(resultJsonMessage.getResultData().getDataItem(0).getSubtaskItem(0).getNodeName(), "NODE_NAME");
    ASSERT_EQ(resultJsonMessage.getResultData().getDataItem(0).getSubtaskItem(0).getNodeHost(), "NODE_HOST");
    ASSERT_EQ(resultJsonMessage.getResultData().getDataItem(0).getSubtaskItem(0).getNodePort(), "NODE_PORT");
    ASSERT_EQ(resultJsonMessage.getResultData().getDataItem(0).getSubtaskItem(0).getState(), static_cast<unsigned int>(DRCETaskRequest::TaskState::TERMINATED_BY_TTL));
    ASSERT_EQ(resultJsonMessage.getResultData().getDataItem(0).getSubtaskItem(0).getSubtasksCount(), 2);

    ASSERT_FALSE(resultJsonMessage.getResultData().getDataItem(0).getSubtaskItem(1).getErrorMessage().empty());
    ASSERT_EQ(resultJsonMessage.getResultData().getDataItem(0).getSubtaskItem(1).getErrorCode(), ERROR_TERMINATE_SUBTASK_AFTER_EXPIRED_MAIN_TASK);
    ASSERT_EQ(resultJsonMessage.getResultData().getDataItem(0).getSubtaskItem(1).getNodeName(), "NODE_NAME");
    ASSERT_EQ(resultJsonMessage.getResultData().getDataItem(0).getSubtaskItem(1).getNodeHost(), "NODE_HOST");
    ASSERT_EQ(resultJsonMessage.getResultData().getDataItem(0).getSubtaskItem(1).getNodePort(), "NODE_PORT");
    ASSERT_EQ(resultJsonMessage.getResultData().getDataItem(0).getSubtaskItem(1).getState(), static_cast<unsigned int>(DRCETaskRequest::TaskState::TERMINATED_BY_TTL));
    ASSERT_EQ(resultJsonMessage.getResultData().getDataItem(0).getSubtaskItem(1).getSubtasksCount(), 2);

    printSuccess("!!! Test Task Request Expired Time for Sync mode");
  }
  else
  {
    while (fObj.getAsyncTasksCount() || fObj.hasTask(requestId))
    {
      Poco::Thread::sleep(5000);
    }

    Poco::File dataFile(FileStream::getDataFileName(fObj.getTasksDataDir(), requestId));

    ASSERT_EQ((cleanupFlag==SessionOptions::CleanupFlag::cfDelete), !dataFile.exists());

    if (dataFile.exists())
    {
      DRCEResultDataItem resultDataItem = DRCECleanupTest::extractResultData(requestId, fObj);

      ASSERT_EQ(resultDataItem.getRequestId(), requestId);
      ASSERT_EQ(resultDataItem.getErrorCode(), ERROR_TERMINATE_EXPIRED_TASK) << "ErrorCode: " << resultDataItem.getErrorCode();
      ASSERT_FALSE(resultDataItem.getErrorMessage().empty());
      ASSERT_EQ(resultDataItem.getNodeName(), fObj.getNodeName());
      ASSERT_EQ(resultDataItem.getNodeHost(), fObj.getNodeHost());
      ASSERT_EQ(resultDataItem.getNodePort(), fObj.getNodePort());
      ASSERT_EQ(resultDataItem.getState(), static_cast<unsigned int>(DRCETaskRequest::TaskState::TERMINATED_BY_TTL));
      ASSERT_EQ(resultDataItem.getSubtasksCount(), 2);

      ASSERT_TRUE(resultDataItem.getSubtaskItem(0).getErrorCode()==ERROR_TERMINATE_SUBTASK_AFTER_EXPIRED_MAIN_TASK ||
          resultDataItem.getSubtaskItem(0).getErrorCode()==NO_ERROR) << "ErrorCode: " << resultDataItem.getSubtaskItem(0).getErrorCode();

      if (resultDataItem.getSubtaskItem(0).getErrorCode()==ERROR_TERMINATE_SUBTASK_AFTER_EXPIRED_MAIN_TASK)
        ASSERT_FALSE(resultDataItem.getSubtaskItem(0).getErrorMessage().empty());
      else
        ASSERT_TRUE(resultDataItem.getSubtaskItem(0).getErrorMessage().empty());

      ASSERT_EQ(resultDataItem.getSubtaskItem(0).getNodeName(), fObj.getNodeName());
      ASSERT_EQ(resultDataItem.getSubtaskItem(0).getNodeHost(), fObj.getNodeHost());
      ASSERT_EQ(resultDataItem.getSubtaskItem(0).getNodePort(), fObj.getNodePort());
      ASSERT_TRUE(resultDataItem.getSubtaskItem(0).getState()==static_cast<unsigned int>(DRCETaskRequest::TaskState::TERMINATED_BY_TTL))
      << "State: " << resultDataItem.getSubtaskItem(0).getState();

      ASSERT_EQ(resultDataItem.getSubtaskItem(0).getSubtasksCount(), 2);

      ASSERT_TRUE(resultDataItem.getSubtaskItem(1).getErrorCode()==ERROR_TERMINATE_SUBTASK_AFTER_EXPIRED_MAIN_TASK ||
          resultDataItem.getSubtaskItem(1).getErrorCode()==NO_ERROR) << "ErrorCode: " << resultDataItem.getSubtaskItem(1).getErrorCode();

      if (resultDataItem.getSubtaskItem(1).getErrorCode()==ERROR_TERMINATE_SUBTASK_AFTER_EXPIRED_MAIN_TASK)
        ASSERT_FALSE(resultDataItem.getSubtaskItem(1).getErrorMessage().empty());
      else
        ASSERT_TRUE(resultDataItem.getSubtaskItem(1).getErrorMessage().empty());

      ASSERT_EQ(resultDataItem.getSubtaskItem(1).getNodeName(), fObj.getNodeName());
      ASSERT_EQ(resultDataItem.getSubtaskItem(1).getNodeHost(), fObj.getNodeHost());
      ASSERT_EQ(resultDataItem.getSubtaskItem(1).getNodePort(), fObj.getNodePort());
      ASSERT_EQ(resultDataItem.getSubtaskItem(1).getState(), static_cast<unsigned int>(DRCETaskRequest::TaskState::TERMINATED_BY_TTL))
      << "State: " << resultDataItem.getSubtaskItem(1).getState();

      ASSERT_EQ(resultDataItem.getSubtaskItem(1).getSubtasksCount(), 2);
    }
    printSuccess("!!! Test Task Request Expired Time for Async mode");
  }
  DRCECleanupTest::cleanupTestData(fObj);
}
//-----------------------------------------------------------------------------
//-----------------------------------------------------------------------------
} // namespace tests
} // namespace drce
} // namespace HCE
