#include <ApplicationPattern.h>
int ApplicationPattern::exitFlag = 0;
ApplicationPattern::ApplicationPattern(char **argv, const char *configFile, unsigned int logDirection, const char *ident, const char *title)
{
  this->argv = argv;
  this->configFile = strdup(configFile);
  if(!title)title = ident;
  log = new(std::nothrow) SysLog(title, logDirection | OPT_ISO8601);
  if(!log)
  {
    return;
  }
  exitFlag = 0;
  sig.Attach(SIGQUIT, signalFunc);
  sig.Attach(SIGINT, signalFunc);
  sig.Attach(SIGTERM, signalFunc);
  sig.Attach(SIGPIPE, signalFunc);
  sig.Attach(SIGUSR1, signalFunc);
  statisticHandler = NULL;
  statusHandler = NULL;
  adminClient = NULL;
  logHandler = NULL;
  this->ident = strdup(ident);
}

ApplicationPattern::~ApplicationPattern()
{
  delete log;
  free(configFile);
  if(adminClient)
  {
    adminClient->stop();
    adminClient->join();
    delete adminClient;
  }
  if(statisticHandler)
  {
    delete statisticHandler;
  }
  if(statusHandler)
  {
    delete statusHandler;
  }
  if(logHandler)
  {
    delete logHandler;
  }
  free(ident);
}

bool ApplicationPattern::init(const char *dumpFile)
{
  Config config(configFile);
  if(!config.isConfigExist())
  {
    log->Put(E_CONFIG, LEV_ERR, "Config file don't exists");
    return false;
  }
  statisticHandler = new(std::nothrow) AStorageHandler();
  if(!statisticHandler)
  {
    log->Put(E_MEMORY, LEV_EMERG, "Can't allocate memory");
    return false;
  }
  statusHandler = new(std::nothrow) StatusHandler(statisticHandler);
  if(!statusHandler)
  {
    log->Put(E_MEMORY, LEV_EMERG, "Can't allocate memory");
    return false;
  }
  statisticHandler->setValue("ApplicationState", 1ll);
  statisticHandler->setValue("StartDate", (long long)time(NULL));
  statisticHandler->setValue("Revision", (long long)REVISION);
  int maxLogLevel = 0;
  config.getAsInt(ident, "log_level", maxLogLevel, 3);
  log->setMaxLevel(maxLogLevel);
  logHandler = new(std::nothrow) LogHandler(log);
  if(!logHandler)
  {
    log->Put(E_MEMORY, LEV_EMERG, "Can't allocate memory");
    return false;
  }
  handlerList.pushDown(statisticHandler);
  handlerList.pushDown(statusHandler);
  handlerList.pushDown(logHandler);
  int adminPort = 0;
  config.getAsInt(ident, "admin_port", adminPort, 10000);
  char *adminHost = NULL;
  config.getAsString(ident, "admin_host", adminHost, "0.0.0.0");
  adminClient = new(std::nothrow) AdminClient(adminHost, adminPort, &handlerList);
  delete []adminHost;
  if(!adminClient)
  {
    log->Put(E_MEMORY, LEV_EMERG, "Can't allocate memory");
    return false;
  }
  if(adminClient->getLastErr())
  {
    log->Put(E_NETWORK, LEV_ERR, "AdminClient: %s", strerror(adminClient->getLastErr()));
    delete adminClient;
    adminClient = NULL;
    return false;
  }
  else
  {
    adminClient->start();
  }
  char *user = NULL;
  config.getAsString(ident, "user", user, "hce");
  if(!setUser(user))
  {
    delete []user;
    return false;
  }
  delete []user;
  char *path = NULL;
  config.getAsString(ident, "log_path", path,  "/var/log/hce/");
  log->setLogPath(path);
  delete []path;
  return true;
}

bool ApplicationPattern::setUser(const char *user)
{
  char *buf = new(std::nothrow) char[sysconf(_SC_GETPW_R_SIZE_MAX)];
  if(!buf)
  {
    return false;
  }
  struct passwd pwd;
  struct passwd *ppwd = NULL;
  memset(&pwd, 0, sizeof(pwd));
  int ret = getpwnam_r(user, &pwd, buf, size_t(sysconf(_SC_GETPW_R_SIZE_MAX)), &ppwd);
  if(ret == 0 && ppwd)
  {
    if(setgid(pwd.pw_gid) == 0)
    {
      if(setuid(pwd.pw_uid) == 0)
      {
        log->Put(E_OK, LEV_INFO, "Change user %s succefull", user);
        delete []buf;
        return true;
      }
      else
      {
        log->Put(E_SETUID, LEV_ERR, "Can't set uid to %d", pwd.pw_gid);
      }
    }
    else
    {
      log->Put(E_SETGID, LEV_ERR, "Can't set gid to %d", pwd.pw_gid);
    }
  }
  else
  {
    log->Put(E_GETUIDGID, LEV_ERR, "Can't get uid & gid for user: %s", user);
  }
  delete []buf;
  return false;
}


void ApplicationPattern::signalFunc(int sigNum)
{
  if(sigNum != SIGPIPE && sigNum != SIGUSR1)
  {
    exitFlag = 1;
  }
}

