#ifndef LOGGER_H
#define LOGGER_H

#include <Poco/Util/AbstractConfiguration.h>
#include <Poco/PatternFormatter.h>
#include <Poco/FileChannel.h>
#include <Poco/FormattingChannel.h>
#include <Poco/AutoPtr.h>
#include <Poco/Logger.h>
#include <Poco/ScopedLock.h>
#include <iostream>
#include <sstream>
#include <Poco/SharedPtr.h>

#include "exceptions/ChannelNotExistException.hpp"

#define DEFAULT_LOG_SUFFIX ".log.txt"

#define DEFAULT_LOG_ALWAYS_USE_DEF_LOG true
#define DEFAULT_LOG_LEVEL "information"

#define DEFAULT_LOG_NAME "Application"
#define DEFAULT_LOG_PATTERN_FORMATTER "%Y-%m-%d %H:%M:%S.%c %N[%P] :%p:%t %U:%u"
#define DEFAULT_LOG_ERROR "Default log is not exist"
#define MESSAGE_WRITTEN_INTO_UNREGISTER_LOG "message written into unregister log"

#define CHANNELS "CHANNELS"
#define PATTERN_FORMATTER "PatternFormatter"
#define BASE_NAME "application.baseName"
#define ALWAYS_USE_DEFAULT_LOG "LOG.alwaysUseDefaultLog"
#define LOG_LEVEL "LOG.log_level"

#define DEFAULT_PURGE_AGE "1 months"
#define DEFAULT_ARCHIVE "timestamp"
#define DEFAULT_ROTATION "daily"

#define LOGGER_ON_DEFAULT_LEVEL 7
#define LOGGER_OFF_LEVEL 0

/*
 * Application - default log which writes to the file "application_name.log.txt"
 * If log channel not set in configuration file :
 * if flag `alwaysUseDefaultLog` equals true - writes to the "application_name.log.txt"
 * else if flag `alwaysUseDefaultLog` not equals true - writes to the std::cout
 */

/*
 * HCE::Logger it's a static class , which may be used without creating HCE::Logger object
 */

namespace HCE {
    class Logger {
    public:
        //!< initialize loggers map
        static void initializeLoggers ();
    private:
        static void initializeDefaultLogger ();
        //!< check default logger, if logger exist - ok
        //!< else - throw exception and shutdown program.
        static void checkDefaultLogger ();
        //!< default channel = "default logger name from configuration file" + .log.txt
        static std::string getDefaultChannel ();
    private:
        static std::string createMsgIfThatChannelNotExist (const std::string& channelName);
    public:
        //!< use the default logger if specified logger does not exist (if "alwaysUseDefaultLog" equals true)
        static bool alwaysUseDefaultLog;
        //!< writing msg to setted channel
        static void writeLog (const Poco::Message& msg,const std::string& channel = DEFAULT_LOG_NAME);
        //!< if channel not exists and default channel not set - log message writes to unregister log
        static void writeUnregisterLog (const Poco::Message& msg,const std::string& channel);
    public:
        static void setLevelByChannelName (const std::string& name,const std::string& level);
        static void setLevelByChannelName (const std::string& name,const int& level);
    public:
        static void setLevelForAllChannel (const std::string& level);
        static void setLevelForAllChannel (const int& level);
    public:
        static void off ();
        static void on (const int& level = LOGGER_ON_DEFAULT_LEVEL);
        static void on (const std::string& level);
    };
}

#endif
