#include <MysqlLite.h>
MysqlLite::MysqlLite(const char *host, const char *user, const char *passwd,
                     const char *db, unsigned int port, const char *socket, bool connectNow)
{
  connected = false;
  if(connectNow)
  {
    sql = mysql_init(NULL);
    my_bool reconnect = true;
    mysql_options(sql, MYSQL_OPT_RECONNECT, (const char*)&reconnect);
    if (mysql_real_connect(sql, host, user, passwd, db, port, socket, 0))
    {
      connected = true;
      mysql_options(sql, MYSQL_OPT_RECONNECT, (const char*)&reconnect);
    }
  }
  else
  {
    sql = NULL;
  }
  if(host)this->host = strdup(host);
  if(user)this->user = strdup(user);
  if(passwd)this->passwd = strdup(passwd);
  if(db)this->db = strdup(db);
  this->port = port;
  if(socket)this->socket = strdup(socket);
  else this->socket = NULL;
}

MysqlLite::~MysqlLite()
{
  if(sql)mysql_close(sql);
  if(host)free(host);
  if(user)free(user);
  if(passwd)free(passwd);
  if(db)free(db);
  if(socket)free(socket);
  mysql_library_end();
}

void MysqlLite::setTimeout(unsigned int readTimeout, unsigned int writeTimeout)
{
  if(sql)
  {
     mysql_options(sql, MYSQL_OPT_READ_TIMEOUT, (const char*)&readTimeout);
     mysql_options(sql, MYSQL_OPT_WRITE_TIMEOUT, (const char*)&writeTimeout);
  }
}

bool MysqlLite::ping()
{
  return sql != NULL ? (mysql_ping(sql) == 0) : false;
}

void MysqlLite::disconnect()
{
  if(sql && connected)
  {
    mysql_close(sql);
    sql = NULL;
  }
}

bool MysqlLite::reconnect()
{
  if(sql)mysql_close(sql);
  sql = mysql_init(NULL);
  if (mysql_real_connect(sql, host, user, passwd, db, port, socket, 0))
  {
    connected = true;
    my_bool reconnect = true;
    mysql_options(sql, MYSQL_OPT_RECONNECT, (const char*)&reconnect);
  }
  else
  {
    connected = false;
  } 
  return connected;
}

bool MysqlLite::setCharacterSet(const char *character)
{
  if(sql && mysql_set_character_set(sql, character) == 0)
  {
    return true;
  }
  return false;
}
bool MysqlLite::query(const char *pattern, ...)
{
  char *data = NULL;
  int n, size = 100;
  va_list ap;
  if (!sql || (data = (char *) malloc(size)) == NULL)
  {
    return false;
  }
  while (1)
  {
    va_start(ap, pattern);
    n = vsnprintf(data, size, pattern, ap);
    va_end(ap);
    if (n > -1 && n < size)
    {
      break;
    }
    if (n > -1)
      size = n + 1;
    else
      size *= 2;
    if ((data = (char *) realloc(data, size)) == NULL)
    {
      return false;
    }
  }
  if (mysql_query(sql, data) == 0)
  {
    free(data);
    return true;
  }
  else
  {
    free(data);
    return false;
  }
}

MYSQL_RES *MysqlLite::store()
{
  return mysql_store_result(sql);
}

MYSQL_RES *MysqlLite::use()
{
  return mysql_use_result(sql);
}

MYSQL_ROW MysqlLite::fetch(MYSQL_RES * res)
{
  return mysql_fetch_row(res);
}

bool MysqlLite::is_connect()
{
  return connected;
}

my_ulonglong MysqlLite::insertId()
{
  return mysql_insert_id(sql);
}

const char *MysqlLite::errorStr()
{
  return mysql_error(sql);
}

const char *MysqlLite::errStr()
{
  return mysql_error(this->sql);
}

unsigned int MysqlLite::errNo()
{
  return mysql_errno(this->sql);
}

char *MysqlLite::escapeStr(const char *str)
{
  char *result = NULL;
  if (str)
  {
    unsigned long len = strlen(str);
    if( (result = (char *) malloc((len << 1) + 1)) == NULL )
    {
      return NULL;
    }
    len = mysql_real_escape_string(this->sql, result, str, len);
    result[len] = '\0';
    result = (char *) realloc(result, len + 1);
  }
  return result;
}

my_ulonglong MysqlLite::affectedRows()
{
  return mysql_affected_rows(this->sql);
}

my_ulonglong MysqlLite::numRows(MYSQL_RES * res)
{
  return mysql_num_rows(res);
}

PreparedQuery::PreparedQuery(const char *query, MysqlLite *sql)
{
  bindParam = NULL;
  stmt = mysql_stmt_init(sql->sql);
  if(stmt)
  {
    if(!mysql_stmt_prepare(stmt, query, strlen(query)))
    {
      paramCount= mysql_stmt_param_count(stmt);
      if(paramCount)
      {
        bindParam = new MYSQL_BIND[paramCount];
        bzero(bindParam, sizeof(MYSQL_BIND) * paramCount);
      }
    }
  }
}

PreparedQuery::~PreparedQuery()
{
  if(bindParam)delete []bindParam; 
  mysql_stmt_close(stmt);
}

bool PreparedQuery::execute()
{
  return mysql_stmt_execute(stmt)==0;
}

MYSQL_BIND* PreparedQuery::operator[](int index)
{
  if(index < paramCount)
  {
    return &(bindParam[index]);
  }
  return NULL;
}

bool PreparedQuery::bind()
{
  return mysql_stmt_bind_param(stmt, bindParam) == 0;
}

const char* PreparedQuery::errorStr()
{
  return mysql_stmt_error(stmt);
}

