/**
 * @file ServerApplication.hpp
 *
 * ServerApplication - base class for application
 *
 * @author Eduard <developers.hce@gmail.com>
 * @link http://hierarchical-cluster-engine.com/
 * @copyright Copyright &copy; 2013 IOIX Ukraine
 * @license http://hierarchical-cluster-engine.com/license/
 * @package HCE project node API
 * @since 0.1
 */

#ifndef HCE_SERVER_APPLICATION
#define HCE_SERVER_APPLICATION

#include <Poco/Util/HelpFormatter.h>
#include <Poco/Util/ServerApplication.h>
#include <Poco/Util/OptionSet.h>
#include <Poco/SharedPtr.h>
#include <Poco/AutoPtr.h>
#include <Poco/Exception.h>
#include <Poco/Bugcheck.h>

#include <string>
#include <vector>
#include <iostream>
#include <map>
#include <unistd.h>
#include <stdio.h>
#include <sstream>
#include <algorithm>

#include "Logger.hpp"
#include "DBSession.hpp"

#include "exceptions/AlreadyInitializedException.hpp"
#include "exceptions/NotSetDatabaseParameters.hpp"
#include "exceptions/NotSetDatabaseType.hpp"

//!< HCE::ServerApplication provides global variable 'HCE::database::Session' which initializes in function 'HCE::ServerApplication::initializeDataBaseSession()'

//!< HCE::ServerApplication provides global variable 'HCE::_config' which initializes in function 'HCE::ServerApplication::initializeConfigurationFile ()'
//!< we need to use variable 'HCE::_config' after calling function initialize ()

//!< HCE::ServerApplication provides global static class 'HCE::Logger' which initializes by calling function 'HCE::Logger::initializeLoggers()' from 'HCE::ServerApplication::initializeLoggers ()'

namespace HCE {
//!<
//!< base class for application's class.
//!< method `main (const std::vectro<string>& argv)` should be overridden in inherited class
//!< if you allows that exception can leave from method main , that function uninitialized never be called
//!< life cycle : constructor -> initialize -> main -> uninitialize -> destructor
class ServerApplication : public Poco::Util::ServerApplication {
public:
    ServerApplication();
    virtual ~ServerApplication() throw();
protected:
    //must be overridden
    void initialize (Poco::Util::Application& self);
    void reinitialize (Poco::Util::Application& self);
    void uninitialize ();
    void defineOptions (Poco::Util::OptionSet& options);

    //!< this function initializes global variable HCE::_config by filename which sets as value in pair --config-file='filename'
    void initializeConfigurationFile ();

    //!< this function return true if in configuration file exist section for database connection
    bool isDataBaseSectionExists();
    //!< return database connection type (by default it's MySQL)
    std::string getDBType () const;

    //!< host=*;user=*;password=*;db=*;compress=*;auto-reconnect=* - mysql
    //!< db=* - sqlite
    std::string initializeMySQLConnectionString() const;
    std::string initializeSQLiteConnectionString() const;

    //!< this functions must be caused both and if in configure file exists section for database connection
    void initializeDataBaseSession ();
    void uninitializeDataBaseSession ();

    //initialize loggers
    void initializeLoggers ();

    //!< this function allows us to understand that we will not use default configuration file
    //!< first parameter not set because we are not using option name
    virtual void handleConfigFile (const std::string&,const std::string& value);
    //!< this function shows us help information (all options and so on)
    virtual void handleHelp (const std::string& name,const std::string& value);
    //!< this function shows us version information (all options and so on)
    virtual void handleVersion (const std::string& name,const std::string& value);    
protected:
    //!< if we use default configuration file this variable is setted as false
    bool _configFileRequest;
    std::string _configFile;
    //!< program would not be started if we use key --help or -h .
    bool _showHelpInfo;
    //!< this variable does not allow initialize application twise
    //!< and verifies that our part of application (config,logger,database) correctly installed
    bool _initializedHCEServerApplication;
    std::string dataBaseType;
protected:
    //error messages, instead of macros definitions
    const std::string APPLICATION_NOT_INITIALIZED_EXCEPTION = "Class HCE::ServerApplication incorrectly initialized. Maybe problem with configuration,logger or database initializing.";
    const std::string MESSAGE_ALREADY_INITIALIZED_EXCEPTION = "Object already initialized.";
    const std::string NOT_SET_DB_TYPE_ERROR = "Not set database type in configuration type.";
    const std::string ADDITION_INFORM_STRING_NOT_FOUND = "config param or Not found config file";
};

}

#define HCE_SERVER_MAIN(App) \
		int main(int argc, char** argv) { \
			const std::string UNKNOWN_EXCEPTION = "Unknown exception"; \
			int rc = Poco::Util::Application::EXIT_OK; \
			try { \
				App app; \
				rc =  app.run(argc, argv); \
			} \
			catch (const Poco::Exception& exc) { \
				std::cerr << exc.displayText() << std::endl; \
				rc = exc.code(); \
			} \
			catch (const std::exception& exc) { \
				std::cerr << exc.what() << std::endl; \
				rc = Poco::Util::Application::EXIT_SOFTWARE; \
			} \
			catch (...) { \
				std::cerr << UNKNOWN_EXCEPTION << std::endl; \
				rc = Poco::Util::Application::EXIT_SOFTWARE; \
			} \
			return rc; \
		} \

#endif
