/**
 * @file sample.cpp
 *
 * Example Application main module. Example of how use HCE::Process class.
 *
 * @author Alexander Vybornyh <alexander.hce.cluster@gmail.com>
 * @link http://hierarchical-cluster-engine.com/
 * @copyright Copyright &copy; 2013 IOIX Ukraine
 * @license http://hierarchical-cluster-engine.com/license/
 * @package HCE project node API
 * @since 0.1
 */

#include <iostream>
#include <sstream>

#include "Process.hpp"
#include "ProcExec.hpp"
#include "CommandExecuter.hpp"

void printTitle(const std::string& title)
{
	std::cout << std::string(title.length(), '=') << std::endl << title << std::endl << std::string(title.length(), '=') << std::endl;
}

void sampleExecuteCommandSort(void)
{
	printTitle("Execute command Sort");

  std::string command = "sort";
  HCE::Process::Args args;
  std::string initialDirectory = "/tmp"; 
  std::stringstream inMsg("def\nabc\n123\n");
  std::stringstream outMsg, errMsg;
  HCE::Process::Env env = {{"key1", "value1"}};

  try
  {
    HCE::ProcessHandle ph = HCE::Process::launch(command, args, initialDirectory, inMsg, outMsg, errMsg, env);

    std::cout << "id: " << ph.id() << std::endl;
    std::cout << "wait: " << ph.wait() << std::endl;
    std::cout << "cout: " << outMsg.str() << std::endl;
    std::cout << "cerr: " << errMsg.str() << std::endl;

  }
  catch(std::exception& e)
  {
    std::cerr << "Exception: " << e.what() << std::endl;
  }
}

void sampleExecuteCommandReverseSort(void)
{
	printTitle("Execute command Reverse Sort");

  std::string command = "sort";
  HCE::Process::Args args = {"-r"};
  std::stringstream inMsg("def\nabc\n123\n");
  std::stringstream outMsg, errMsg;

  try
  {
    HCE::ProcessHandle ph = HCE::Process::launch(command, args, inMsg, outMsg, errMsg);

    std::cout << "id: " << ph.id() << std::endl;
    std::cout << "wait: " << ph.wait() << std::endl;
    std::cout << "cout: " << outMsg.str() << std::endl;
    std::cout << "cerr: " << errMsg.str() << std::endl;

  }
  catch(std::exception& e)
  {
    std::cerr << "Exception: " << e.what() << std::endl;
  }
}

void sampleRequestTerminate(void)
{
	printTitle("Request Terminate");

	std::string command = "sleep";
  HCE::Process::Args args = {"5"};
	std::stringstream inMsg, outMsg, errMsg;
	try
  {
    HCE::ProcessHandle ph = HCE::Process::launch(command, args, inMsg, outMsg, errMsg);

//		HCE::Process::requestTermination(ph);
		HCE::Process::requestKill(ph);

    std::cout << "id: " << ph.id() << std::endl;
    std::cout << "wait: " << ph.wait() << std::endl;
    std::cout << "cout: " << outMsg.str() << std::endl;
    std::cout << "cerr: " << errMsg.str() << std::endl;
  }
  catch(std::exception& e)
  {
    std::cerr << "Exception: " << e.what() << std::endl;
  }
}

void sampleGetCurrentProcessID(void)
{
	printTitle("Get Current Process ID");

	std::string command = "ps";
  HCE::Process::Args args = {"auxf"};
	try
  {
    HCE::ProcessHandle ph = HCE::Process::launch(command, args);

		HCE::Process::requestTermination(ph);

    std::cout << "Current process ID: " << HCE::Process::id() << std::endl;
		std::cout << "wait: " << ph.wait() << std::endl;
  }
  catch(std::exception& e)
  {
    std::cerr << "Exception: " << e.what() << std::endl;
  }
}

void sampleProcExec(void)
{
	printTitle("Use ProcExec");

	std::string command = "sort";
  HCE::Process::Args args = {"-r"};
  std::stringstream inMsg("def\nabc\n123\n");

	std::stringstream outMsg, errMsg;
	HCE::ProcExec procExec(command, args);
	std::cout << "Exec: " << std::boolalpha << procExec.exec(inMsg, outMsg, errMsg) << std::endl;
	std::cout << "isError: " << std::boolalpha << procExec.isError() << std::endl;
	std::cout << "Out: " << outMsg.str() << std::endl;
	std::cout << "Err: " << errMsg.str() << std::endl;
}	

void sampleCommandExecutor(void)
{
	printTitle("Use CommandExecutor");

	std::string command = "sort -r";
	std::string shell = "bash -c";
  std::stringstream inMsg("def\nabc\n123\n");

	HCE::Command cmd(command);
	cmd.setShell(shell);
	cmd.setInputStream(inMsg);
	HCE::CommandExecuter commandExecuter;
	HCE::CommandResultData commandResultData = commandExecuter.execute(cmd);

	std::cout << "isError: " << std::boolalpha << commandResultData.getIsError() << std::endl;
	std::cout << "ExitStatus: " << commandResultData.getExitStatus() << std::endl;
	std::cout << "Out: " << commandResultData.getOutStream().str() << std::endl;
	std::cout << "Err: " << commandResultData.getErrStream().str() << std::endl;
}

int main(void)
{
	sampleExecuteCommandSort();
	sampleExecuteCommandReverseSort();
	sampleRequestTerminate();
	sampleGetCurrentProcessID();
	sampleProcExec();
	sampleCommandExecutor();

  return 0;
}

