/**
 * @file JsonObjExtractor.hpp
 *
 * Definition of JsonObjExtractor class.
 * It class exctract nodes from given Poco::JSON::Object::Ptr object
 * by path, also can check nodes type. Also provides static method
 * which converts node's path from "std::vector<Poco::Dynamic::Var>" type 
 * to the it's string representation.
 *
 * @author Andrey Lapshov <developers.hce@gmail.com>
 * @link http://hierarchical-cluster-engine.com/
 * @copyright Copyright &copy; 2013 IOIX Ukraine
 * @license http://hierarchical-cluster-engine.com/license/
 * @package HCE project node API
 * @since 0.1
 */

#ifndef JSON_JBJ_EXTRACTOR_HPP
#define JSON_JBJ_EXTRACTOR_HPP

#include <Poco/Checksum.h>
#include <Poco/Dynamic/Var.h>
#include <Poco/Dynamic/Struct.h>
#include <Poco/Environment.h>
#include <Poco/JSON/Object.h>
#include <Poco/JSON/Array.h>
#include <map>
#include <vector>

#include "JsonObjectExeption.hpp"

namespace HCE
{
  namespace utils
  {
    using namespace HCE::exception;
    class JsonObjExtractor
    {
      protected:
        Poco::JSON::Object::Ptr localObj;
        Poco::JSON::Array::Ptr localArray;
        Poco::Dynamic::Var localVar;
        std::string errString;
        std::map<unsigned int, Poco::Dynamic::Var> globalMap;
      protected:
        inline void commonParsePrepareObj(Poco::JSON::Object::Ptr obj, std::vector<Poco::Dynamic::Var> &varPath);
        void commonParseObj(Poco::JSON::Object::Ptr obj, std::vector<Poco::Dynamic::Var> &varPath) throw (JsonObjectExeption);
        inline void mapExtract(std::vector<Poco::Dynamic::Var> &varPath);
        inline void commonParseSelectType(Poco::JSON::Object::Ptr obj, std::vector<Poco::Dynamic::Var> &varPath);
        void generateMapByObj(Poco::JSON::Object::Ptr obj, std::string &name);
        void generateMapByArray(Poco::JSON::Array::Ptr array, std::string &name);
        void clearLocalFields();
      public:
        JsonObjExtractor();
        inline Poco::Dynamic::Var getAsVar(Poco::JSON::Object::Ptr obj, std::vector<Poco::Dynamic::Var> &varPath);
        inline Poco::JSON::Object::Ptr getAsObject(Poco::JSON::Object::Ptr obj, std::vector<Poco::Dynamic::Var> &varPath);
        inline Poco::JSON::Array::Ptr getAsArray(Poco::JSON::Object::Ptr obj, std::vector<Poco::Dynamic::Var> &varPath);
        
        inline bool isVar(Poco::JSON::Object::Ptr obj, std::vector<Poco::Dynamic::Var> &varPath);
        inline bool isObject(Poco::JSON::Object::Ptr obj, std::vector<Poco::Dynamic::Var> &varPath);
        inline bool isArray(Poco::JSON::Object::Ptr obj, std::vector<Poco::Dynamic::Var> &varPath);
        bool isKeyPresent(Poco::JSON::Object::Ptr obj, std::vector<Poco::Dynamic::Var> &varPath);
        
        void generateMap(Poco::JSON::Object::Ptr obj, std::string &name);
        
        static char objSeparator;
        static char arraySeparatorBegin;
        static char arraySeparatorEnd;
        static std::string toJsonPath(std::vector<Poco::Dynamic::Var> &varPath) throw (JsonObjectExeption);
        
        ~JsonObjExtractor();
    };
    
    void JsonObjExtractor::commonParsePrepareObj(Poco::JSON::Object::Ptr obj, std::vector<Poco::Dynamic::Var> &varPath)
    {
      errString.clear();
      localObj = obj;
      localArray = nullptr;
      localVar.empty();
      try
      {
        commonParseObj(obj, varPath);
      }
      catch(HCE::exception::JsonObjectExeption &excp)
      {
        clearLocalFields();
        errString = excp.what();
      }
    }
    
    void JsonObjExtractor::mapExtract(std::vector<Poco::Dynamic::Var> &varPath)
    {
      Poco::Checksum crcCalc(Poco::Checksum::TYPE_CRC32);
      crcCalc.update(toJsonPath(varPath));
      std::map<unsigned int, Poco::Dynamic::Var>::iterator it;
      if((it = globalMap.find(crcCalc.checksum())) != globalMap.end())
      {
        if(!((*it).second.isEmpty()))
        {
          if((*it).second.type() == typeid(Poco::JSON::Object::Ptr))
          {
            localObj = (*it).second.extract<Poco::JSON::Object::Ptr>();
          }
          else if((*it).second.type() == typeid(Poco::JSON::Array::Ptr))
          {
            localArray = (*it).second.extract<Poco::JSON::Array::Ptr>();
          }
          else
          {
            localVar = (*it).second;
          }
        }
      }      
    }
    
    void JsonObjExtractor::commonParseSelectType(Poco::JSON::Object::Ptr obj, std::vector<Poco::Dynamic::Var> &varPath)
    {
      if(obj.isNull())
      {
        mapExtract(varPath);
      }
      else
      {
        commonParsePrepareObj(obj, varPath);
      }
    }
    
    Poco::Dynamic::Var JsonObjExtractor::getAsVar(Poco::JSON::Object::Ptr obj, std::vector<Poco::Dynamic::Var> &varPath)
    {
      commonParseSelectType(obj, varPath);
      return localVar;
    }
    
    Poco::JSON::Object::Ptr JsonObjExtractor::getAsObject(Poco::JSON::Object::Ptr obj, std::vector<Poco::Dynamic::Var> &varPath)
    {
      commonParseSelectType(obj, varPath);
      return localObj;
    }
    
    Poco::JSON::Array::Ptr JsonObjExtractor::getAsArray(Poco::JSON::Object::Ptr obj, std::vector<Poco::Dynamic::Var> &varPath)
    {
      commonParseSelectType(obj, varPath);
      return localArray;
    }
    
    bool JsonObjExtractor::isVar(Poco::JSON::Object::Ptr obj, std::vector<Poco::Dynamic::Var> &varPath)
    {
      commonParseSelectType(obj, varPath);
      return (localVar.isEmpty() ? false : true);
    }
    
    bool JsonObjExtractor::isObject(Poco::JSON::Object::Ptr obj, std::vector<Poco::Dynamic::Var> &varPath)
    {
      commonParseSelectType(obj, varPath);
      return (localObj.isNull() ? false : true);
    }
    
    bool JsonObjExtractor::isArray(Poco::JSON::Object::Ptr obj, std::vector<Poco::Dynamic::Var> &varPath)
    {
      commonParseSelectType(obj, varPath);
      return (localArray.isNull() ? false : true);
    }
    
  }
}

#endif