/**
 * @file BaseDB.hpp
 *
 * BaseDB - Base driver class.
 *
 * @author Alexey <developers.hce@gmail.com>
 * @link http://hierarchical-cluster-engine.com/
 * @copyright Copyright &copy; 2013 IOIX Ukraine
 * @license http://hierarchical-cluster-engine.com/license/
 * @package HCE project node API
 * @since 0.1
 */


#ifndef BASE_DB_HPP
#define BASE_DB_HPP


//!<
//!< Poco headers
//!<
#include <Poco/SharedPtr.h>

//!<
//!< std headers
//!<
#include <iostream>
#include <string>
#include <list>


//!<
//!< app headers
//!<
#include "Errors.hpp"


namespace HCE
{

typedef int STATUS;

/**
 * A base Search Engine driver's class
 * @brief Your Driver's class must be inherited from that base class
 */
class BaseDB
{
public:
	BaseDB() : _error(ACTION_SUCCESS) {}
	virtual ~BaseDB() {}
public:
	typedef Poco::SharedPtr<HCE::BaseDB> BaseDBPtr;
public:
    /**
     * @brief set the database name
     * @param dbName the database name
     */
	virtual inline void setDatabaseName(const std::string& dbName) { _dbName = dbName; }
    /**
     * @brief set the driver URL
     * @param dbUrl the driver URL
     */
	virtual inline void setDriverUrl(const std::string& dbUrl) { _bootstrapUrls.push_back(dbUrl); }
public:
	/**
	 * A pure virtual member.
     * @brief put document with given ID to Search Engine
	 * @param key the document's ID
	 * @param value the document itself
	 * @brief method must set _error to operation's result state:
	 * ACTION_SUCCESS - operation was completed successfully
	 * ACTION_GENERAL_FAILURE - operation was failed
	 */
	virtual void put(const std::string& key, const std::string& value) = 0;
    /**
     * A pure virtual member.
     * @brief get document with given ID from Search Engine
     * @param key the document's ID
     * @brief method must set _error to operation's result state:
     * DOCUMENT_PRESENT - operation was completed successfully (document was present)
     * DOCUMENT_NOT_PRESENT - operation was failed (document wasn't present)
     * @return document's content
     */
	virtual std::string get(const std::string& key) = 0;
    /**
     * A pure virtual member.
     * @brief get all documents with given ID from Search Engine
     * @param key the document's ID
     * @brief method must set _error to operation's result state:
     * DOCUMENT_PRESENT - operation was completed successfully (document was present)
     * DOCUMENT_NOT_PRESENT - operation was failed (document wasn't present)
     * @return document's content
     */
	virtual std::vector<std::string> get_all() = 0;
    /**
     * A pure virtual member.
     * @brief delete document with given ID from Search Engine
     * @param key the document's ID
     * @brief method must set _error to operation's result state:
     * ACTION_SUCCESS - operation was completed successfully
     * ACTION_GENERAL_FAILURE - operation was failed
     * @return document's content
     */
	virtual void del(const std::string& key) = 0;
    /**
     * A pure virtual member.
     * @brief get status last performed operation. Must be called after method to get operation's status
     * @return operation's status
     */
    virtual int getLastError() const = 0;//{ return _error; }
public:
    inline void setWriteUnique(const bool writeUnique) {_writeUnique=writeUnique;}
    inline void setWritePolicy(const bool writePolicy) {_writePolicy=writePolicy;}
    inline const bool getWriteUnique() const {return _writeUnique;}
    inline const bool getWritePolicy() const {return _writePolicy;}
protected:
	std::string _dbName;
	std::list<std::string> _bootstrapUrls;
	STATUS _error;
    bool _writeUnique;
    bool _writePolicy;
}; //!< class BaseDB


} //!< namespace HCE


#endif //!< BASE_DB_HPP
