/**
 * @file ORMFunctionalObject.hpp
 *
 * ORMFunctionalObject class
 *
 * @author Alexey <developers.hce@gmail.com>
 * @link http://hierarchical-cluster-engine.com/
 * @copyright Copyright &copy; 2013 IOIX Ukraine
 * @license http://hierarchical-cluster-engine.com/license/
 * @package HCE project node API
 * @since 0.1
 */


#ifndef ORM_FUNCTIONAL_OBJECT_HPP
#define ORM_FUNCTIONAL_OBJECT_HPP


//!<
//!< app headers
//!<
#include "FakeDB.hpp"
#include "BaseDB.hpp"
#include "Response.hpp"
#include "DataBaseDriverConfig.hpp"
#include "Logger.hpp"

#include <functional>


namespace HCE
{

//!<
//!< ORMFunctionalObject class
//!<
class ORMFunctionalObject
{
public:
    explicit ORMFunctionalObject(const DataBaseDriverConfig& db_config);
    virtual ~ORMFunctionalObject() throw();
public:
    //!<
    //!< process handler
    //!< ORMFunctionalObject takes input message in format Request struct serialized as string JSON
    //!< and gives output in format Response struct serialized as string JSON
    //!<
    std::string processJSON(const std::string& input_message);
    //!<
    //!< process handler
    //!< ORMFunctionalObject takes input message in format Request struct serialized as string JSON
    //!< and gives output in format Documents struct
    //!<
    //Documents processObj(const std::string& input_message);
    Poco::SharedPtr<Response> processObj(const Request&);
    //!<
    //!< getErrorCode() and getErrorMsg() return
    //!< ORMFunctionalObject state in numerical and literal
    //!< formats appropriately
    //!<
public:
    //!<
    //!< return error code as numerical value set by the Storage Engine driver
    //!<
    inline unsigned int getErrorCode() const { return _errorCode; }
    //!<
    //!< return error code as string value set by the Storage Engine driver
    //!<
    inline std::string getErrorMsg() const { return _errorMsg; }
private:
    //!<
    //!< fill array of the methods. Index is type of action (e.g. PUT=0 is "putData" method)
    //!< That array of the methods serve for replace the "switch" code block by the direct accessing the appropriate method by the index
    //!<
    void init();
private:
    void createDriverByName();
    void setDriverUrl();
    void setDatabaseName();
    void setWritePolicy();
    void setWriteUnique();
private:
    void put( const std::string & key, const std::string & value ) {
        this->_dbDriver->put(key,value);
        this->_status = this->_dbDriver->getLastError();
    }
    std::string get ( const std::string & key ) {
        std::string value(this->_dbDriver->get(key));
        this->_status = this->_dbDriver->getLastError();
        return value;
    }
    std::vector<std::string> get_all () {
    	std::vector<std::string> value (this->_dbDriver->get_all());
    	this->_status = this->_dbDriver->getLastError();
    	return value;
    }
    void del (const std::string& key) {
        this->_dbDriver->del(key);
        this->_status = _dbDriver->getLastError();
    }
private:
    //!<
    //!< that block of the methods in accordance with supported message's types, described in
    //!< hce_doc/03/Key-Value-DB_Functional_object_protocol.docx
    //!<
    Poco::SharedPtr<Response> putData(const std::string&);
    Poco::SharedPtr<Response> getData(const std::string&);
    Poco::SharedPtr<Response> delData(const std::string&);
    Poco::SharedPtr<Response> checkData(const std::string&);
    Poco::SharedPtr<Response> getAllKeys(const std::string&);
    Poco::SharedPtr<Response> adminData(const std::string&);
    //!<
    //!< return error code as numerical value set by the Storage Engine driver
    //!<
    inline int getLastError() { return _dbDriver->getLastError(); }
private:
//public:
    //template<typename T>
    //STATUS loadResourceFromJSON(const std::string&, T&);
    //!<
    //!< unserialization data from the "documents" section of the json message
    //!<
    STATUS loadResourcesFromJSON(const std::string&, std::map<std::string, std::string>&);
    STATUS loadDocIdsFromJSON(const std::string&, std::vector<std::string>&);
    //!<
    //!< fill array of the methods. Index is type of action (e.g. PUT=0 is "putData" method)
    //!< That array of the methods serve for replace the "switch" code block by the direct accessing the appropriate method by the index
    //!< must be called during initialization time and before first using the class's object
    //!<
    void init_command();
    const std::string getDBEngine() const { return this->_dbConfig.getDriverName(); }
private:
    unsigned int _errorCode;
    std::string _errorMsg;
    STATUS _status;
    //typedef Poco::SharedPtr<Response> (HCE::ORMFunctionalObject::*FP)(const std::string&);
    typedef std::function<Poco::SharedPtr<HCE::Response>(ORMFunctionalObject*,const std::string&)> FP;
    std::vector<FP> command;
private:
    Poco::SharedPtr<HCE::BaseDB> _dbDriver;
    HCE::DataBaseDriverConfig _dbConfig;
    Poco::DynamicFactory<HCE::BaseDB> _factory;
};

inline void ORMFunctionalObject::createDriverByName() {
	this->_status = ACTION_GENERAL_FAILURE;
	this->_dbDriver = this->_factory.createInstance( this->_dbConfig.getDriverName() );
	this->_status = ACTION_SUCCESS;
}

inline void ORMFunctionalObject::setDriverUrl() {
	this->_status = ACTION_GENERAL_FAILURE;
	if (this->_dbDriver) {
		this->_dbDriver->setDriverUrl( this->_dbConfig.getDataBaseUrl() );
		this->_status = ACTION_SUCCESS;
	}
}

inline void ORMFunctionalObject::setDatabaseName() {
    this->_status = ACTION_GENERAL_FAILURE;
    if (this->_dbDriver) {
        this->_dbDriver->setDatabaseName( this->_dbConfig.getDataBaseName() );
        this->_status = ACTION_SUCCESS;
    }
}


inline void ORMFunctionalObject::setWritePolicy() {
    this->_status = ACTION_GENERAL_FAILURE;
    if (this->_dbDriver) {
        this->_dbDriver->setWritePolicy(this->_dbConfig.getWritePolicy());
        this->_status = ACTION_SUCCESS;
    }
}


inline void ORMFunctionalObject::setWriteUnique() {
    this->_status = ACTION_GENERAL_FAILURE;
    if (this->_dbDriver) {
        this->_dbDriver->setWriteUnique(this->_dbConfig.getWriteUnique());
        this->_status = ACTION_SUCCESS;
    }
}


}

#endif
