/**
 * @file sample.cpp
 *
 * Example Application main module. Example of how use LoggerStream, MessagesCollection and CustomMessage classes.
 *
 * @author Alexander Vybornyh <alexander.hce.cluster@gmail.com>
 * @link http://hierarchical-cluster-engine.com/
 * @copyright Copyright &copy; 2013 IOIX Ukraine
 * @license http://hierarchical-cluster-engine.com/license/
 * @package HCE project node API
 * @since 0.1
 */

#include <iostream>

#include "LoggerStream.hpp"
#include "CustomMessage.hpp"
#include "RealLoggable.hpp"

void sampleLoggerStream(void)
{
  HCE::LoggerStream logger;
  HCE::tests::RealLoggable realLoggable; // some user implemented class (child of interface class Loggable)

  /* default mode write to buffer */
  logger << "test 1";
  std::cout << "MSG1: " << logger.logMsg() << std::endl;

  logger << "TEST 2" << std::endl;
  std::cout << "MSG2: " << logger.logMsg() << std::endl;

  /* below mode write to stream */
  logger.setLogStream(std::cout);   // turn on mode write to stream
  logger.log() << "TEST CONSOLE" << std::endl;

  logger.resetLogStream();          // turn off mode write to stream
  logger.log() << "THIS NEVER BEEN TO STREAM" << std::endl; // message will be write to buffer

  /* again mode write to buffer */
  logger.log() << "test 3" << std::endl;
  std::cout << "MSG3: " << logger.logMsg() << std::endl; // output buffer content (include "THIS NEVER BEEN TO STREAM")

  /* below used HCE::endl manipulator */
  logger << "test 4" << HCE::endl;
  std::cout << "MSG4: " << logger.logMsg() << std::endl;

  logger.setLoggable(realLoggable); // bind to class describing need behavior
  logger << "test 5" << std::endl << "test 6" << HCE::endl;
  std::cout << "MSG 5 and 6: " << logger.logMsg() << std::endl;

  logger.resetLoggable();   // unbind any classes describing need behavior
  logger << "test 7";
  std::cout << "MSG7: " << logger.logMsg() << std::endl;
}

void sampleCustomMessage(void)
{
  // list all messages need to functional objects
  std::vector<std::pair<unsigned int, std::string> > messages =
  {
      {0, "TEST START"},
      {1, "TEST DATA"},
      {2, "TEST MSG"},
      {3, "TEST STOP"}
  };

  HCE::MessagesCollection messagesCollection(messages); // initialization collection
  HCE::CustomMessage msg(messagesCollection);

  /* output messages with default delimiter tokens */
  std::cout << msg(0) << std::endl;
  std::cout << msg(1, 42) << std::endl;
  std::cout << msg(2, "USER MESSAGE") << std::endl;
  std::cout << msg(3) << std::endl;

  /* set custom delimiter tokens */
  HCE::CustomMessage::MsgConst msgConst = { '*', ';', '-'};
  msg.setMsgConst(msgConst);

  /* output messages with custom delimiter tokens */
  std::cout << msg(0) << std::endl;
  std::cout << msg(1, 42) << std::endl;
  std::cout << msg(2, "USER MESSAGE") << std::endl;
  std::cout << msg(3) << std::endl;
}

int main(int argc, char** argv)
{
  sampleLoggerStream();
  sampleCustomMessage();
  return 0;
}
