/**
 * @file LoggerStream.hpp
 *
 * Implementation class LoggerStream of a MessageStream tools in HCE cluster node.
 *
 * @author Alexander Vybornyh <alexander.hce.cluster@gmail.com>
 * @link http://hierarchical-cluster-engine.com/
 * @copyright Copyright &copy; 2013 IOIX Ukraine
 * @license http://hierarchical-cluster-engine.com/license/
 * @package HCE project node API
 * @since 0.1
 */

#ifndef LOGGER_STREAM_HPP
#define LOGGER_STREAM_HPP

#include <functional>
#include <iostream>
#include <sstream>
#include <typeinfo>

namespace HCE
{
//-----------------------------------------------------------------------------
typedef std::function<std::string(const std::string&)> Loggable;
//-----------------------------------------------------------------------------
class LoggerStream : public std::ostream
{
public:
  enum Priority
  {
    PRIO_NONE         = 0, // A none error. In the application turn off any log.
    PRIO_FATAL        = 1, // A fatal error. The application will most likely terminate. This is the highest priority.
    PRIO_CRITICAL     = 2, // A critical error. The application might not be able to continue running successfully.
    PRIO_ERROR        = 3, // An error. An operation did not complete successfully, but the application as a whole is not affected.
    PRIO_WARNING      = 4, // A warning. An operation completed with an unexpected result.
    PRIO_NOTICE       = 5, // A notice, which is an information with just a higher priority.
    PRIO_INFORMATION  = 6, // An informational message, usually denoting the successful completion of an operation.
    PRIO_DEBUG        = 7, // A debugging message.
    PRIO_TRACE        = 8  // A tracing message. This is the lowest priority.
  };
public:
  explicit LoggerStream(const std::string& source_="");
  virtual ~LoggerStream(void);

  void setPrefix(const std::string& prefix_) {prefix=prefix_;}
  std::string getPrefix(void) const {return prefix;}

  void setSuffix(const std::string& suffix_) {suffix=suffix_;}
  std::string getSuffix(void) const {return suffix;}

  unsigned int getLogPriority(void) const {return logPriority;}

  void setSource(const std::string& source_) {source=source_;}
  std::string getSource(void) const {return source;}

  void resetLogStream(void);
  void setLogStream(std::ostream& os_);
  std::string logMsg(bool isReset=true);
  std::ostream& log(unsigned int logPriority);
  std::ostream& log(Priority priority);

  void setLoggable(Loggable loggable_);
  void resetLoggable(void);
  std::ostream& put(const std::string&);
  LoggerStream& endl(void);
  LoggerStream& flush(void);

  void saveSafe(void);
private:
  std::stringbuf logBuf;
  std::basic_streambuf<char>* pBuf;

  Loggable loggable;

  std::string prefix;
  std::string suffix;
  unsigned int logPriority;
  std::string source;

  typedef std::ostream inherited;
};
//-----------------------------------------------------------------------------
//-----------------------------------------------------------------------------
std::ostream& endl(std::ostream& os);
std::ostream& flush(std::ostream& os);
LoggerStream& operator<<(LoggerStream& os, const std::string& rhs);
LoggerStream& operator<<(LoggerStream& os, const char* rhs);
//-----------------------------------------------------------------------------
//-----------------------------------------------------------------------------
} // end namespace HCE 

#endif // LOGGER_STREAM_HPP
