/**
 * @file ReduceTaskCore.hpp
 *
 * The file defines a set of abstract types which forms a
 * reduce task core.
 *
 * @author Igor Savonik <developers.hce@gmail.com>
 * @link http://hierarchical-cluster-engine.com/
 * @copyright Copyright &copy; 2013 IOIX Ukraine
 * @license http://hierarchical-cluster-engine.com/license/
 * @package HCE Reducing Task
 * @since 0.1
 */

#ifndef REDUCE_TASK_CORE_H
#define REDUCE_TASK_CORE_H

#include <Poco/SharedPtr.h>
#include <string>

namespace HCE{
namespace reduce{
namespace core{

/**
 * The IterableIf must support every class which acts as data storage
 * with ability to explore data
*/
class IterableIf{
public:
  virtual ~IterableIf(){}
  virtual bool hasNext() = 0;
  virtual void reset() = 0;
  virtual void getNext(std::string &key, std::string &value) = 0;
};


/**
 * The StorableIf must support every class which acts as data storage
 * with ability storage data
 */
class StorableIf{
public:
  virtual ~StorableIf(){}
  virtual void addKeyValue(const std::string &key, const std::string &value) = 0;
};


/**
 * The ReduceResultIf acts as storage for reduce job results. For user acts
 * as IterableIf to explore results and for ReduceTaskIf acts as storage to
 * store reduce job results
 */
class ReduceResultIf: public virtual IterableIf, public virtual StorableIf{
public:
  virtual ~ReduceResultIf(){}
};


/**
 * The ReduceTaskIf must support every class which consists some logic to process
 * reducing task
 */
class ReduceTaskIf: public virtual StorableIf{
public:
  virtual ~ReduceTaskIf(){}
  virtual void fillReduceResult(Poco::SharedPtr<ReduceResultIf> reduceResult) = 0;
};


/**
 * ReduceJonIf is simple wrapper interface to easy pass around data
 * related to reduce job
 */
class ReduceJobIf{
public:
  virtual ~ReduceJobIf(){}
  virtual Poco::SharedPtr<IterableIf> getDataStorage() const = 0;
  virtual Poco::SharedPtr<ReduceResultIf> getReduceResult() const = 0;
  virtual Poco::SharedPtr<ReduceTaskIf> getReduceTask() const = 0;
};


/**
 * ScheduleStrategyIf is common interface for implementation
 * of various running strategies of execution reduce jobs
 */
class ScheduleStrategyIf
{
public:
  virtual ~ScheduleStrategyIf(){}
  virtual void executeJob(const Poco::SharedPtr<ReduceJobIf> &reduceJob) = 0;
};

}
}
}

#endif
