#include <gtest/gtest.h>
#include "DRCETaskReducersManager.hpp"
#include "DRCEResultData.hpp"
#include "DRCEResultDataSerializator.hpp"

using HCE::drce::reduce_task::DRCETaskReducersManager;
using HCE::reduce::NotFoundByKeyException;
using HCE::reduce::WrongJSONStructureException;
using HCE::drce::DRCEResultData;
using HCE::drce::DRCEResultDataItem;
using HCE::drce::DRCEResultDataSerializator;

static void buildDRCEResultData( DRCEResultData& drceResultData ) {
	drceResultData.clear();

	DRCEResultDataItem item1;
	item1.setStdoutStream( "standard stream 1" );
	item1.setStderrStream( "standard error 1" );
	item1.setNodeName( "node_1" );
	item1.setTime( ::time( 0 ) );
	item1.setExitStatus( 0 );
	item1.addFileItem( "file1.txt", "test data 1", 4 );
	item1.addFileItem( "file2.txt", "test data 2", 1 );
	drceResultData.addDataItem( item1 );

	DRCEResultDataItem item2;
	item2.setStdoutStream( "standard stream 2" );
	item2.setStderrStream( "standard error 2" );
	item2.setNodeName( "node_2" );
	item2.setTime( ::time( 0 ) );
	item2.setExitStatus( 0 );
	item1.addFileItem( "file3.txt", "test data 3", 4 );
	item1.addFileItem( "file4.txt", "test data 4", 1 );		
	
	drceResultData.addDataItem( item2 );
}

static std::string buildDRCEResultDataJson() {
	DRCEResultData drceResultData;
	buildDRCEResultData( drceResultData );
	
	DRCEResultDataSerializator serializator( drceResultData );

	std::string json;
		
	serializator.serialize( json );
	
	return json;
} 
class DRCETaskReducersManagerTest: public ::testing::Test {
public:
	virtual void SetUp() {
		reduceTaskId = 1;
	}
	
	virtual void Destroy() {
	}
	
	DRCETaskReducersManager drceTaskReducersMgr;
	unsigned long long reduceTaskId; 
};

TEST_F( DRCETaskReducersManagerTest, ReduceTaskNotExist ) {
	EXPECT_FALSE( drceTaskReducersMgr.isReducerExist( reduceTaskId ) );
}

TEST_F( DRCETaskReducersManagerTest, CreateReducer ) {
	EXPECT_FALSE( drceTaskReducersMgr.isReducerExist( reduceTaskId ) );
	drceTaskReducersMgr.createReducer( reduceTaskId );
	EXPECT_TRUE( drceTaskReducersMgr.isReducerExist( reduceTaskId ) );
}

TEST_F( DRCETaskReducersManagerTest, DeleteReducer ) {
	drceTaskReducersMgr.createReducer( reduceTaskId );
	EXPECT_TRUE( drceTaskReducersMgr.isReducerExist( reduceTaskId ) );
	drceTaskReducersMgr.deleteReducerBy( reduceTaskId );
}

TEST_F( DRCETaskReducersManagerTest, DeleteNotExistReducer ) {
	EXPECT_FALSE( drceTaskReducersMgr.isReducerExist( reduceTaskId ) );
	EXPECT_THROW( drceTaskReducersMgr.deleteReducerBy( reduceTaskId ), NotFoundByKeyException );
}

TEST_F( DRCETaskReducersManagerTest, AddNonJsonData ) {
	drceTaskReducersMgr.createReducer( reduceTaskId );
	EXPECT_THROW( drceTaskReducersMgr.addDataInReducer( reduceTaskId, "balala..."), WrongJSONStructureException );
}

TEST_F( DRCETaskReducersManagerTest, AddEmptyResultData ) {
	drceTaskReducersMgr.createReducer( reduceTaskId );
	drceTaskReducersMgr.addDataInReducer( reduceTaskId, "[]");
	EXPECT_EQ( drceTaskReducersMgr.runReduceTaskForTaskId( reduceTaskId ), "[]");
}

TEST_F( DRCETaskReducersManagerTest, ReduceEmptyResult ) {
	drceTaskReducersMgr.createReducer( reduceTaskId );
	std::string json = drceTaskReducersMgr.runReduceTaskForTaskId( reduceTaskId );
	EXPECT_EQ( json, "[]");
}

TEST_F( DRCETaskReducersManagerTest, ReduceNonEmptyResult ) {
	std::string json = buildDRCEResultDataJson();
	
	drceTaskReducersMgr.createReducer( reduceTaskId );
	drceTaskReducersMgr.addDataInReducer( reduceTaskId, json );
	drceTaskReducersMgr.addDataInReducer( reduceTaskId, json );
	drceTaskReducersMgr.runReduceTaskForTaskId( reduceTaskId );
	std::string json1 = drceTaskReducersMgr.runReduceTaskForTaskId( reduceTaskId );
	EXPECT_EQ( json.substr( 1, json.size() - 2 ) + "," + json.substr( 1, json.size() - 2 ), json1.substr( 1, json1.size() - 2 ));
}