/*
 * ReducingHandlerIntergationTest.cpp
 *
 *  Created on: Jun 5, 2013
 *      Author: igor
 */
#include <gtest/gtest.h>
#include "ReducingHandler.hpp"
#include "ReducingPartsCountersStorage.hpp"
#include "ReduceAdditionsStorage.hpp"
#include "SphinxTaskReducersManager.hpp"
#include "ReducingOutputMessageBuilder.hpp"
#include "ConsoleSupportNotifier.hpp"
#include "ReducingExceptionsTranslator.hpp"
#include "TaskReducersManagerFactory.hpp"
#include "TestUtils.hpp"
#include "SphinxResultCompareUtils.hpp"

using namespace HCE::types;
using namespace HCE::reduce;
using namespace HCE::reduce_types;
using namespace HCE::sphinx;
using namespace HCE::sphinx::reduce_task;
using namespace Poco;
using namespace std;

class ReducingHandlerIntergationTest: public ::testing::Test{
protected:
  virtual void SetUp(){
    itemId = 2442352714LLU;
    ttl = 7;

    reducingPartsCountersStorage.assign(new ReducingPartsCountersStorage());
    reduceAdditionsStorage.assign(new ReduceAdditionsStorage());
    taskReducersManagerFactory.assign(new TaskReducersManagerFactory());
    reducingExceptionsTranslator.assign(new ReducingExceptionsTranslator());
    reducingOutputMessageBuilder.assign(new ReducingOutputMessageBuilder(reducingExceptionsTranslator));
    supportNotifier.assign(new ConsoleSupportNotifier());

    reducingHandler.assign(new ReducingHandler(reducingPartsCountersStorage,
                                                  reduceAdditionsStorage,
                                                  taskReducersManagerFactory,
                                                  reducingOutputMessageBuilder,
                                                  supportNotifier));
    buildSphinxResults();
  }

  void buildSphinxResults()
  {
    WordInfo word1 = WordInfo("word1", 10, 3);
    WordInfo word2 = WordInfo("word2", 1, 1);

    SphinxMatchInfo matchInfo1 = buildSimpleMatchInfo(7LLU, 111LLU);
    SphinxMatchInfo matchInfo2 = buildSimpleMatchInfo(9LLU, 222LLU);
    SphinxMatchInfo matchInfo3 = buildSimpleMatchInfo(15LLU, 674387LLU);
    SphinxMatchInfo matchInfo4 = buildSimpleMatchInfo(9LLU, 674387LLU);

    AttrInfo attrInfo1 = AttrInfo("first", "value");

    matchInfo1.addAttrInfo(attrInfo1);
    matchInfo2.addAttrInfo(attrInfo1);
    matchInfo3.addAttrInfo(attrInfo1);
    matchInfo4.addAttrInfo(attrInfo1);

    sphinxRequestInfo1.setMaxResultsNumber(5);
    sphinxRequestInfo1.setNodeName("nodeName1");
    sphinxRequestInfo1.setQuery("query");
    sphinxRequestInfo1.setTotal(10);
    sphinxRequestInfo1.setTotalFound(150);
    sphinxRequestInfo1.setTimeMsec(200);
    sphinxRequestInfo1.addWordInfo(word1);
    sphinxRequestInfo1.addWordInfo(word2);

    sphinxRequestInfo2.setMaxResultsNumber(5);
    sphinxRequestInfo2.setNodeName("nodeName2");
    sphinxRequestInfo2.addWordInfo(word1);
    sphinxRequestInfo2.addWordInfo(word2);

    sphinxResultData1.addMatchInfo(matchInfo1);
    sphinxResultData1.addMatchInfo(matchInfo2);

    sphinxResultData2.addMatchInfo(matchInfo3);
    sphinxResultData2.addMatchInfo(matchInfo4);

    expectedSphinxResultData12.addMatchInfo(matchInfo3);
    expectedSphinxResultData12.addMatchInfo(matchInfo1);
    expectedSphinxResultData12.addMatchInfo(matchInfo4);
    //expectedSphinxResultData12.addMatchInfo(matchInfo3);
    expectedSphinxResultData12.addRequestInfo(sphinxRequestInfo1);
    expectedSphinxResultData12.addRequestInfo(sphinxRequestInfo2);

    sphinxRequestInfo1.setMaxResultsNumber(2);
    sphinxRequestInfo1.setOrderBy(DESC_SORT);

    sphinxRequestInfo2.setMaxResultsNumber(2);
    sphinxRequestInfo2.setOrderBy(DESC_SORT);

    //because of using map, order by id different from input order
    expectedSphinxResultData12DescSort.addMatchInfo(matchInfo3);
    expectedSphinxResultData12DescSort.addMatchInfo(matchInfo4);
    //expectedSphinxResultData12DescSort.addMatchInfo(matchInfo3);
    expectedSphinxResultData12DescSort.addRequestInfo(sphinxRequestInfo1);
    expectedSphinxResultData12DescSort.addRequestInfo(sphinxRequestInfo2);
  }

  unsigned long long itemId;
  int ttl;
  SharedPtr<ReducingPartsCountersStorage> reducingPartsCountersStorage;
  SharedPtr<ReduceAdditionsStorage> reduceAdditionsStorage;
  SharedPtr<TaskReducersManagerFactory> taskReducersManagerFactory;
  SharedPtr<ReducingExceptionsTranslator>reducingExceptionsTranslator;
  SharedPtr<ReducingOutputMessageBuilder> reducingOutputMessageBuilder;
  SharedPtr<SupportNotifierIf> supportNotifier;

  SharedPtr<ReducingHandler>reducingHandler;

  SharedPtr<ReducingOutputMessage> reducingOutputMessage;
  SharedPtr<ReducingOutputMessage> etalonOutputMessage;
  SharedPtr<ReducingInputMessage> reducingInputMessage;
  SharedPtr<ReducingInputMessage> reducingInputMessage2;

  SphinxRequestInfo sphinxRequestInfo1;
  SphinxRequestInfo sphinxRequestInfo2;
  SphinxResultData sphinxResultData1;
  SphinxResultData sphinxResultData2;
  SphinxResultData expectedSphinxResultData12;
  SphinxResultData expectedSphinxResultData12DescSort;
};

TEST_F(ReducingHandlerIntergationTest, reducingTaskForInvalidKey)
{
  string errMsg = "ReduceAdditionsStorage - not found key " + std::to_string(itemId);
  NotFoundByKeyException notFoundByKeyException(errMsg);

  etalonOutputMessage = buildEtalonFrom(notFoundByKeyException);

  reducingOutputMessage = reducingHandler->makeReducing(itemId);
  ASSERT_TRUE(isEqualReducingOutputMessage(*reducingOutputMessage.get(),
                                           *etalonOutputMessage.get()));
}


TEST_F(ReducingHandlerIntergationTest, supressThrowsWhileDeleteReducingTaskByInvalidKey)
{
  ASSERT_NO_THROW(reducingHandler->deleteReducingTaskBy(itemId));
}


TEST_F(ReducingHandlerIntergationTest, throwExceptionWhenEncounterInvalidKeyForPartsNumberCounters)
{
  ASSERT_THROW(reducingHandler->getAccumulateReducingPartsNumberBy(itemId),
               NotFoundByKeyException);
}


TEST_F(ReducingHandlerIntergationTest, reduceTaskWithOnePartsNumber)
{
  sphinxRequestInfo1.setMaxResultsNumber(5);
  sphinxRequestInfo1.setOrderBy(NULL_SORT);
  sphinxRequestInfo2.setMaxResultsNumber(5);
  sphinxRequestInfo2.setOrderBy(NULL_SORT);

  sphinxResultData1.addRequestInfo(sphinxRequestInfo1);
  sphinxResultData2.addRequestInfo(sphinxRequestInfo2);

  string json_input = buildInputJSONFromSphinxResultData(sphinxResultData1);
  string json_output = buildOutputJSONFromSphinxResultData(sphinxResultData1);
  reducingInputMessage = buildReducingInputMessage(MessageType::mtSphinx, ttl, json_input);
  etalonOutputMessage = buildEtalonFrom(MessageType::mtSphinx, reducingInputMessage->getTTL(), json_output);

  reducingHandler->accumulateReducingData(itemId, reducingInputMessage);
  reducingOutputMessage = reducingHandler->makeReducing(itemId);
  ASSERT_TRUE(isEqualReducingOutputMessage(*reducingOutputMessage.get(),
                                           *etalonOutputMessage.get()));
}


TEST_F(ReducingHandlerIntergationTest, reduceTaskWithTwoPartsNumber)
{
  sphinxRequestInfo1.setMaxResultsNumber(5);
  sphinxRequestInfo1.setOrderBy(NULL_SORT);
  sphinxRequestInfo2.setMaxResultsNumber(5);
  sphinxRequestInfo2.setOrderBy(NULL_SORT);

  sphinxResultData1.addRequestInfo(sphinxRequestInfo1);
  sphinxResultData2.addRequestInfo(sphinxRequestInfo2);

  string expectedJSON = buildOutputJSONFromSphinxResultData(expectedSphinxResultData12);
  string json1 = buildInputJSONFromSphinxResultData(sphinxResultData1);
  string json2 = buildInputJSONFromSphinxResultData(sphinxResultData2);
  reducingInputMessage = buildReducingInputMessage(MessageType::mtSphinx, ttl, json1);
  reducingInputMessage2 = buildReducingInputMessage(MessageType::mtSphinx, ttl, json2);
  etalonOutputMessage = buildEtalonFrom(MessageType::mtSphinx, reducingInputMessage->getTTL(), expectedJSON);

  reducingHandler->accumulateReducingData(itemId, reducingInputMessage);
  reducingHandler->accumulateReducingData(itemId, reducingInputMessage2);
  reducingOutputMessage = reducingHandler->makeReducing(itemId);

  ASSERT_TRUE(isEqualReducingOutputMessage(*reducingOutputMessage.get(),
                                           *etalonOutputMessage.get()));
}


TEST_F(ReducingHandlerIntergationTest, reduceTaskWithTwoPartsNumberWithDescSort)
{
  sphinxResultData1.addRequestInfo(sphinxRequestInfo1);
  sphinxResultData2.addRequestInfo(sphinxRequestInfo2);

  string expectedJSON = buildOutputJSONFromSphinxResultData(expectedSphinxResultData12DescSort);
  string json1 = buildInputJSONFromSphinxResultData(sphinxResultData1);
  string json2 = buildInputJSONFromSphinxResultData(sphinxResultData2);
  reducingInputMessage = buildReducingInputMessage(MessageType::mtSphinx, ttl, json1);
  reducingInputMessage2 = buildReducingInputMessage(MessageType::mtSphinx, ttl, json2);
  etalonOutputMessage = buildEtalonFrom(MessageType::mtSphinx, reducingInputMessage->getTTL(), expectedJSON);

  reducingHandler->accumulateReducingData(itemId, reducingInputMessage);
  reducingHandler->accumulateReducingData(itemId, reducingInputMessage2);
  reducingOutputMessage = reducingHandler->makeReducing(itemId);

  ASSERT_TRUE(isEqualReducingOutputMessage(*reducingOutputMessage.get(),
                                           *etalonOutputMessage.get()));
}

