#include <gtest/gtest.h>
#include "HandlerMocks.hpp"
#include "ReducingHandler.hpp"
#include "ReducingInputMessage.hpp"
#include "ReducingOutputMessage.hpp"
#include "Exceptions.hpp"
#include "SphinxReduceJobBuilder.hpp"

using namespace HCE::reduce_types;
using namespace HCE::sphinx::reduce_task;

using ::testing::_;
using ::testing::Eq;
using ::testing::Return;
using ::testing::NiceMock;
using ::testing::Throw;
using ::testing::ReturnPointee;
using ::testing::Matcher;
using ::testing::InSequence;


class ReducingHandlerTest: public ::testing::Test{
protected:
  virtual void SetUp(){
    reducingPartsCountersStorage.assign(new ReducingPartsCountersStorageMock());
    reduceAdditionsStorage.assign(new ReduceAdditionsStorageMock());
    taskReducersManagerFactory.assign(new TaskReducersManagerFactoryMock());
    taskReducersManager.assign(new TaskReducersManagerMock());
    reducingOutputMessageBuilder.assign(new ReducingOutputMessageBuilderMock());
    supportNotifier.assign(new SupportNotifierMock());
    reducingPartsCountersStorageNice.assign(new NiceMock<ReducingPartsCountersStorageMock>());

    reducingHandler.assign(new ReducingHandler(reducingPartsCountersStorage,
                                               reduceAdditionsStorage,
                                               taskReducersManagerFactory,
                                               reducingOutputMessageBuilder,
                                               supportNotifier));
    itemId = 2;
    HCE::types::MessageType reduceType = HCE::types::MessageType::mtMysql; //need use const - must failure
    int ttl = 7;
    string body = "some body";

    reducingInputMessage.assign(new ReducingInputMessage(reduceType, ttl, body));
  }

  unsigned long long itemId;
  SharedPtr<ReducingPartsCountersStorageMock> reducingPartsCountersStorage;
  SharedPtr<ReduceAdditionsStorageMock> reduceAdditionsStorage;
  SharedPtr<TaskReducersManagerMock>taskReducersManager;
  SharedPtr<TaskReducersManagerFactoryMock> taskReducersManagerFactory;
  SharedPtr<ReducingOutputMessageBuilderMock> reducingOutputMessageBuilder;
  SharedPtr<SupportNotifierMock> supportNotifier;
  SharedPtr<ReducingInputMessage> reducingInputMessage;
  SharedPtr<NiceMock<ReducingPartsCountersStorageMock> >reducingPartsCountersStorageNice;
  SharedPtr<ReducingOutputMessage> reducingOutputMessage;
  SharedPtr<ReducingHandler>reducingHandler;
};


TEST_F(ReducingHandlerTest, createInitStructuresForAccumulating)
{
  NotFoundByKeyException notFoundByKeyException("");
  SharedPtr<TaskReducersManagerMock>taskReducersManagerN(new NiceMock<TaskReducersManagerMock>());

  InSequence s;

  EXPECT_CALL(*reducingPartsCountersStorage.get(), findBy(Eq(itemId)))
              .WillRepeatedly(Return(false));
  //init new structures
  EXPECT_CALL(*reduceAdditionsStorage.get(), findBy(Matcher<HCE::types::MessageType>(_)))
              .WillRepeatedly(Throw(notFoundByKeyException));
  EXPECT_CALL(*taskReducersManagerFactory.get(), build(_))
             .WillOnce(ReturnPointee(&taskReducersManagerN));
  EXPECT_CALL(*reduceAdditionsStorage.get(), addWith(Matcher<HCE::types::MessageType>(_), _))
           .Times(1);
  EXPECT_CALL(*reduceAdditionsStorage.get(), findBy(Matcher<HCE::types::MessageType>(_)))
              .WillRepeatedly(ReturnPointee(&taskReducersManagerN));
  EXPECT_CALL(*reduceAdditionsStorage.get(), addWith(Matcher<unsigned long long>(_), _))
            .Times(1);
  EXPECT_CALL(*reducingPartsCountersStorage.get(), addReducingPartsCounterWithKey(Eq(itemId)))
              .Times(1);
  EXPECT_CALL(*reduceAdditionsStorage.get(), findBy(Matcher<HCE::types::MessageType>(_)))
              .WillRepeatedly(ReturnPointee(&taskReducersManagerN));
  //usual flow
  EXPECT_CALL(*reducingPartsCountersStorage.get(), incrementBy(Eq(itemId)))
              .Times(1);
  EXPECT_CALL(*reduceAdditionsStorage.get(), findBy(Matcher<unsigned long long>(Eq(itemId))))
                .WillRepeatedly(ReturnPointee(&taskReducersManagerN));

  reducingHandler->accumulateReducingData(itemId, reducingInputMessage);
}


TEST_F(ReducingHandlerTest, translateJSONStructureThrowsToSupportNotifier)
{
  WrongJSONStructureException wrongJSONStructureException("");

  EXPECT_CALL(*reducingPartsCountersStorage.get(), findBy(_))
                .WillOnce(Return(true));

  EXPECT_CALL(*reducingPartsCountersStorage.get(), incrementBy(_))
              .Times(1);

  EXPECT_CALL(*reduceAdditionsStorage.get(), findBy(Matcher<unsigned long long>(Eq(itemId))))
                .WillRepeatedly(ReturnPointee(&taskReducersManager));

  EXPECT_CALL(*taskReducersManager.get(), addDataInReducer(_, _))
              .WillOnce(Throw(wrongJSONStructureException));

  EXPECT_CALL(*supportNotifier.get(), notifyWrongJSONStructure(_, _))
              .Times(1);

  reducingHandler->accumulateReducingData(itemId, reducingInputMessage);
}


TEST_F(ReducingHandlerTest, translateManagerFactoryNotFoundByKeyExceptionToSupportNotifier)
{
  NotFoundByKeyException notFoundByKeyException("");

  EXPECT_CALL(*reducingPartsCountersStorage.get(), findBy(_))
                .WillOnce(Return(false));

  EXPECT_CALL(*reduceAdditionsStorage.get(), findBy(Matcher<HCE::types::MessageType>(_)))
              .WillOnce(Throw(notFoundByKeyException));

  EXPECT_CALL(*taskReducersManagerFactory.get(), build(_))
             .WillOnce(Throw(notFoundByKeyException));

  EXPECT_CALL(*supportNotifier.get(), notifyNotFoundItemException(_))
              .Times(1);

  reducingHandler->accumulateReducingData(itemId, reducingInputMessage);
}


TEST_F(ReducingHandlerTest, makeReducing)
{
  string reducingJSON;
  types::MessageType msgType = types::MessageType::mtSphinx;

  EXPECT_CALL(*reduceAdditionsStorage.get(), findBy(Matcher<unsigned long long>(Eq(itemId))))
             .WillRepeatedly(ReturnPointee(&taskReducersManager));
  EXPECT_CALL(*taskReducersManager.get(), getType())
             .WillOnce(Return(msgType));

  EXPECT_CALL(*reduceAdditionsStorage.get(), findBy(Matcher<unsigned long long>(Eq(itemId))))
             .WillRepeatedly(ReturnPointee(&taskReducersManager));

  EXPECT_CALL(*taskReducersManager.get(), runReduceTaskForTaskId(Eq(itemId)))
             .WillOnce(Return(reducingJSON));
  EXPECT_CALL(*reducingOutputMessageBuilder.get(), build(_, _, Matcher<const string&>(Eq(reducingJSON))))
             .WillOnce(ReturnPointee(&reducingOutputMessage));

  reducingHandler->makeReducing(itemId);
}


TEST_F(ReducingHandlerTest, translateNotFoundSphinxTaskReducerExceptionToSupportNotifier)
{
  NotFoundByKeyException notFoundSphinxTaskReducerException("");
  types::MessageType msgType = types::MessageType::mtSphinx;

  EXPECT_CALL(*reduceAdditionsStorage.get(), findBy(Matcher<unsigned long long>(Eq(itemId))))
             .WillRepeatedly(ReturnPointee(&taskReducersManager));

  EXPECT_CALL(*taskReducersManager.get(), getType())
             .WillOnce(Return(msgType));

  EXPECT_CALL(*reduceAdditionsStorage.get(), findBy(Matcher<unsigned long long>(Eq(itemId))))
             .WillRepeatedly(ReturnPointee(&taskReducersManager));

  EXPECT_CALL(*taskReducersManager.get(), runReduceTaskForTaskId(Eq(itemId)))
             .WillOnce(Throw(notFoundSphinxTaskReducerException));
  EXPECT_CALL(*supportNotifier.get(), notifyNotFoundItemException(_))
             .Times(1);
  EXPECT_CALL(*reducingOutputMessageBuilder.get(),
               build(Matcher<const NotFoundByKeyException&>(_)))
             .WillOnce(ReturnPointee(&reducingOutputMessage));

  reducingHandler->makeReducing(itemId);
}


TEST_F(ReducingHandlerTest, cleanUpResourcesAfterDeleteItemId)
{
  EXPECT_CALL(*reducingPartsCountersStorage.get(), deleteBy(Eq(itemId)))
              .Times(1);
  EXPECT_CALL(*reduceAdditionsStorage.get(), findBy(Matcher<unsigned long long>(Eq(itemId))))
             .WillRepeatedly(ReturnPointee(&taskReducersManager));
  EXPECT_CALL(*taskReducersManager.get(), deleteReducerBy(Eq(itemId)))
             .Times(1);
  EXPECT_CALL(*reduceAdditionsStorage.get(), deleteBy(Matcher<unsigned long long>(Eq(itemId))))
             .Times(1);

  reducingHandler->deleteReducingTaskBy(itemId);
}


TEST_F(ReducingHandlerTest, translateNotFoundSphinxTaskReducerExceptionToSupportNotifierDuringDelete)
{
  NotFoundByKeyException notFoundSphinxTaskReducerException("");

  EXPECT_CALL(*reducingPartsCountersStorage.get(), deleteBy(Eq(itemId)))
              .WillOnce(Throw(notFoundSphinxTaskReducerException));
  EXPECT_CALL(*supportNotifier.get(), notifyNotFoundItemException(_))
              .Times(1);

  reducingHandler->deleteReducingTaskBy(itemId);
}


TEST_F(ReducingHandlerTest, getAccumulateReducingParts)
{
  EXPECT_CALL(*reducingPartsCountersStorage.get(), getAccumulateReducingParts(Eq(itemId)))
              .WillOnce(Return(1));

  reducingHandler->getAccumulateReducingPartsNumberBy(itemId);
}


TEST_F(ReducingHandlerTest, throwExceptionWhilegetAccumulateReducingPartsbyInvalidKey)
{
  NotFoundByKeyException notFoundSphinxTaskReducerException("");

  EXPECT_CALL(*reducingPartsCountersStorage.get(), getAccumulateReducingParts(Eq(itemId)))
              .WillOnce(Throw(notFoundSphinxTaskReducerException));

  ASSERT_THROW(reducingHandler->getAccumulateReducingPartsNumberBy(itemId),
               NotFoundByKeyException);
}
