/*
 * SphinxResultBuilderTest.cpp
 *
 *  Created on: May 14, 2013
 *      Author: igor
 */
#include <gtest/gtest.h>
#include "SphinxResultBuilder.hpp"
#include "SphinxReduceResult.hpp"
#include "SphinxResultCompareUtils.hpp"
#include "Exceptions.hpp"

using namespace HCE::sphinx;
using namespace HCE::sphinx::reduce_task;
using namespace HCE::reduce;
using namespace std;
using namespace Poco;

TEST(SphinxResultBuilder, buildFromEmptyData)
{
  SharedPtr<SphinxReduceDataStorage>sphinxReduceDataStorage(new SphinxReduceDataStorage());
  SharedPtr<SphinxReduceResult>sphinxReduceResult(new SphinxReduceResult);
  vector<SphinxRequestInfo> requestInfo;

  SphinxResultBuilder sphinxResultBuilder;
  SharedPtr<SphinxResultData> sphinxResult = sphinxResultBuilder.build(
                              sphinxReduceDataStorage, sphinxReduceResult, requestInfo);

  ASSERT_EQ(sphinxResult->getMatchInfoCount(), 0U);
  ASSERT_EQ(sphinxResult->getRequestInfoCount(), 0U);
}


class SphinxResultBuilderTest: public ::testing::Test{
  protected:
    virtual void SetUp(){
      nodeName = "node1";
      query = "My query";
      totalRes = 10;
      totalFound = 2;
      time = 1020;

      WordInfo word1 = WordInfo("word1", 10, 3);
      matchInfo1 = buildSimpleMatchInfo(7LLU, 111LLU);
      matchInfo2 = buildSimpleMatchInfo(9LLU, 222LLU);
      matchInfo3 = buildSimpleMatchInfo(15LLU, 444LLU);
      matchInfo4 = buildSimpleMatchInfo(3LLU, 99LLU);
      AttrInfo attrInfo("first", "value");

      matchInfo1.addAttrInfo(attrInfo);
      matchInfo2.addAttrInfo(attrInfo);
      matchInfo3.addAttrInfo(attrInfo);
      matchInfo4.addAttrInfo(attrInfo);

      sphinxRequestInfo.setQuery(query);
      sphinxRequestInfo.setTotal(totalRes);
      sphinxRequestInfo.setTotalFound(totalFound);
      sphinxRequestInfo.setTimeMsec(time);
      sphinxRequestInfo.addWordInfo(word1);

      sphinxResultData.assign(new SphinxResultData());
      sphinxResultData->addMatchInfo(matchInfo1);
      sphinxResultData->addMatchInfo(matchInfo2);
      sphinxResultData->addMatchInfo(matchInfo3);
      sphinxResultData->addMatchInfo(matchInfo4);
      sphinxResultData->addRequestInfo(sphinxRequestInfo);

      expectedSelectResultData.addMatchInfo(matchInfo1);
      expectedSelectResultData.addMatchInfo(matchInfo4);
      expectedSelectResultData.addRequestInfo(sphinxRequestInfo);

      expectedCopyResultData = *sphinxResultData.get();

      sphinxReduceDataStorage.assign(new SphinxReduceDataStorage());
      sphinxReduceResult.assign(new SphinxReduceResult());
    }

    string nodeName;
    string query;
    int totalRes;
    int totalFound;
    int time;
    SharedPtr<SphinxResultData> sphinxResultData;
    SphinxResultData expectedSelectResultData;
    SphinxResultData expectedCopyResultData;
    SphinxMatchInfo matchInfo1;
    SphinxMatchInfo matchInfo2;
    SphinxMatchInfo matchInfo3;
    SphinxMatchInfo matchInfo4;
    SphinxRequestInfo sphinxRequestInfo;

    vector<SphinxRequestInfo> requestInfo;
    SharedPtr<SphinxReduceDataStorage>sphinxReduceDataStorage;
    SharedPtr<SphinxReduceResult>sphinxReduceResult;
};


TEST_F(SphinxResultBuilderTest, build)
{
  sphinxReduceDataStorage->addData(matchInfo1);
  sphinxReduceDataStorage->addData(matchInfo2);
  sphinxReduceDataStorage->addData(matchInfo3);
  sphinxReduceDataStorage->addData(matchInfo4);

  sphinxReduceResult->addKeyValue(to_string(matchInfo1.getDocId()), matchInfo1.getWeight());
  sphinxReduceResult->addKeyValue(to_string(matchInfo2.getDocId()), matchInfo2.getWeight());
  sphinxReduceResult->addKeyValue(to_string(matchInfo3.getDocId()), matchInfo3.getWeight());
  sphinxReduceResult->addKeyValue(to_string(matchInfo4.getDocId()), matchInfo4.getWeight());

  requestInfo.push_back(sphinxRequestInfo);

  SphinxResultBuilder sphinxResultBuilder;
  SharedPtr<SphinxResultData> sphinxResult = sphinxResultBuilder.build(
                              sphinxReduceDataStorage, sphinxReduceResult, requestInfo);

  ASSERT_TRUE(isEqualSphinxResultData(*sphinxResult.get(), *sphinxResultData.get()));
}


TEST_F(SphinxResultBuilderTest, throwExceptionWhileResourceIsntPresentInDataStorage)
{
  sphinxReduceDataStorage->addData(matchInfo1);
  sphinxReduceDataStorage->addData(matchInfo2);

  sphinxReduceResult->addKeyValue(to_string(matchInfo1.getDocId()), matchInfo1.getWeight());
  sphinxReduceResult->addKeyValue("fake key", "fake weight");

  SphinxResultBuilder sphinxResultBuilder;
  ASSERT_THROW(sphinxResultBuilder.build(sphinxReduceDataStorage, sphinxReduceResult, requestInfo),
               NotFoundByKeyException);
}


TEST_F(SphinxResultBuilderTest, selectFirstDataElementFromEquals)
{
  SphinxMatchInfo matchInfo1Equal = buildSimpleMatchInfo(7LLU, 111LLU);

  sphinxReduceDataStorage->addData(matchInfo1);
  sphinxReduceDataStorage->addData(matchInfo1Equal);

  sphinxReduceResult->addKeyValue(to_string(matchInfo1.getDocId()), matchInfo1.getWeight());

  SphinxResultBuilder sphinxResultBuilder;
  SharedPtr<SphinxResultData> sphinxResult = sphinxResultBuilder.build(
                              sphinxReduceDataStorage, sphinxReduceResult, requestInfo);

  ASSERT_EQ(sphinxResult->getMatchInfoItem(0).getAttributes().size(), 1U);
}


TEST_F(SphinxResultBuilderTest, selectiveBuildBaseOnKeysList_emptyList)
{
  vector<pair<string, unsigned long long> > weightKeyPairs;

  SphinxResultBuilder sphinxResultBuilder;
  SharedPtr<SphinxResultData> sphinxResult  = sphinxResultBuilder.buildFromUsingKeys(
                                              sphinxResultData, weightKeyPairs);

  ASSERT_EQ(sphinxResult->getMatchInfoCount(), 0U);
  ASSERT_EQ(sphinxResult->getRequestInfoCount(), 1U);
}


TEST_F(SphinxResultBuilderTest, selectiveBuildBaseOnKeysList)
{
  vector<pair<string, unsigned long long> > weightKeyPairs;
  weightKeyPairs.push_back(make_pair("111", 7LLU));
  weightKeyPairs.push_back(make_pair("099", 3LLU));

  SphinxResultBuilder sphinxResultBuilder;
  SharedPtr<SphinxResultData> sphinxResult  = sphinxResultBuilder.buildFromUsingKeys(
                                              sphinxResultData, weightKeyPairs);

  ASSERT_TRUE(isEqualSphinxResultData(*sphinxResult.get(), expectedSelectResultData));
}


TEST_F(SphinxResultBuilderTest, selectiveBuildBaseOnKeysList_copy)
{
  vector<pair<string, unsigned long long> > weightKeyPairs;
  weightKeyPairs.push_back(make_pair("111", 7LLU));
  weightKeyPairs.push_back(make_pair("222", 9LLU));
  weightKeyPairs.push_back(make_pair("444", 15LLU));
  weightKeyPairs.push_back(make_pair("099", 3LLU));

  SphinxResultBuilder sphinxResultBuilder;
  SharedPtr<SphinxResultData> sphinxResult  = sphinxResultBuilder.buildFromUsingKeys(
                                              sphinxResultData, weightKeyPairs);

  ASSERT_TRUE(isEqualSphinxResultData(*sphinxResult.get(), expectedCopyResultData));
}
