/*
 * SphinxTaskReducersManagerIntegrationTest.cpp
 *
 *  Created on: May 21, 2013
 *      Author: igor
 */

#include <gtest/gtest.h>
#include "SphinxTaskReducersManager.hpp"
#include "SphinxResultCompareUtils.hpp"
#include "SphinxTaskReducersStorage.hpp"
#include "SphinxTaskReducerFactory.hpp"
#include "SphinxResultJSONConvertor.hpp"

using namespace HCE::sphinx::reduce_task;
using namespace HCE::sphinx;
using namespace HCE::reduce;
using namespace Poco;


class SphinxTaskReducersManagerIntegrationTest: public ::testing::Test{
  protected:
    virtual void SetUp(){
      sphinxTaskReducersStorage.assign(new SphinxTaskReducersStorage());
      sphinxTaskReducerFactory.assign(new SphinxTaskReducerFactory());
      sphinxResultJSONConvertor.assign(new SphinxResultJSONConvertor());

      WordInfo word1 = WordInfo("word1", 10, 3);
      WordInfo word2 = WordInfo("word2", 1, 1);

      matchInfo1 = buildSimpleMatchInfo(7LLU, 111LLU);
      matchInfo2 = buildSimpleMatchInfo(9LLU, 222LLU);
      matchInfo3 = buildSimpleMatchInfo(15LLU, 674387LLU);
      matchInfo4 = buildSimpleMatchInfo(9LLU, 674387LLU);

      AttrInfo attrInfo1 = AttrInfo("first", "value");

      matchInfo1.addAttrInfo(attrInfo1);
      matchInfo2.addAttrInfo(attrInfo1);
      matchInfo3.addAttrInfo(attrInfo1);
      matchInfo4.addAttrInfo(attrInfo1);

      sphinxRequestInfo1.setNodeName("nodeName1");
      sphinxRequestInfo1.setMaxResultsNumber(5);
      sphinxRequestInfo1.setQuery(query1);
      sphinxRequestInfo1.setTotal(totalRes1);
      sphinxRequestInfo1.setTotalFound(totalFound1);
      sphinxRequestInfo1.setTimeMsec(time1);
      sphinxRequestInfo1.addWordInfo(word1);
      sphinxRequestInfo1.addWordInfo(word2);

      sphinxRequestInfo2.setNodeName("nodeName2");
      sphinxRequestInfo2.setMaxResultsNumber(5);
      sphinxRequestInfo2.addWordInfo(word1);
      sphinxRequestInfo2.addWordInfo(word2);

      sphinxResultData1.addMatchInfo(matchInfo1);
      sphinxResultData1.addMatchInfo(matchInfo2);
      sphinxResultData1.addRequestInfo(sphinxRequestInfo1);

      sphinxResultData2.addMatchInfo(matchInfo3);
      sphinxResultData2.addMatchInfo(matchInfo4);
      sphinxResultData2.addRequestInfo(sphinxRequestInfo2);
    }

    virtual void TearDown(){
      sphinxTaskReducersStorage.assign(new SphinxTaskReducersStorage());
      sphinxTaskReducerFactory.assign(new SphinxTaskReducerFactory());
      sphinxResultJSONConvertor.assign(new SphinxResultJSONConvertor());
    }

    void buildExpectedSphinxResult(SphinxResultData& expected){
      expected.clear();
      expected.addMatchInfo(matchInfo3);
      expected.addMatchInfo(matchInfo1);
      expected.addMatchInfo(matchInfo4);
      //expected.addMatchInfo(matchInfo3);
      expected.addRequestInfo(sphinxRequestInfo1);
      expected.addRequestInfo(sphinxRequestInfo2);
    }

    SharedPtr<SphinxTaskReducersStorage>sphinxTaskReducersStorage;
    SharedPtr<SphinxTaskReducerFactory>sphinxTaskReducerFactory;
    SharedPtr<SphinxResultJSONConvertor>sphinxResultJSONConvertor;

    static string query1;
    static int totalRes1;
    static int totalFound1;
    static int time1;

    SphinxRequestInfo sphinxRequestInfo1;
    SphinxRequestInfo sphinxRequestInfo2;
    SphinxMatchInfo matchInfo1;
    SphinxMatchInfo matchInfo2;
    SphinxMatchInfo matchInfo3;
    SphinxMatchInfo matchInfo4;
    SphinxResultData sphinxResultData1;
    SphinxResultData sphinxResultData2;
};
string SphinxTaskReducersManagerIntegrationTest::query1;
int SphinxTaskReducersManagerIntegrationTest::totalRes1;
int SphinxTaskReducersManagerIntegrationTest::totalFound1;
int SphinxTaskReducersManagerIntegrationTest::time1;


TEST_F(SphinxTaskReducersManagerIntegrationTest, checkExistenceASphinxReducer)
{
  unsigned long long taskId = 2442352714LLU;
  SphinxTaskReducersManager sphinxTaskReducersManager(sphinxTaskReducersStorage,
                            sphinxTaskReducerFactory, sphinxResultJSONConvertor);

  sphinxTaskReducersManager.createReducer(taskId);

  ASSERT_TRUE(sphinxTaskReducersManager.isReducerExist(taskId) == true);
}


TEST_F(SphinxTaskReducersManagerIntegrationTest, reduceEmptySphinxResultResponses)
{
  SphinxResultData emptySphinxResult;
  string json = buildOutputJSONFromSphinxResultData(emptySphinxResult);
  string json_input = buildInputJSONFromSphinxResultData(emptySphinxResult);
  int taskId = 2442352714LLU;

  SphinxTaskReducersManager sphinxTaskReducersManager(sphinxTaskReducersStorage,
                            sphinxTaskReducerFactory, sphinxResultJSONConvertor);

  sphinxTaskReducersManager.createReducer(taskId);
  sphinxTaskReducersManager.addDataInReducer(taskId, json_input);
  string reducedJSON = sphinxTaskReducersManager.runReduceTaskForTaskId(taskId);

  ASSERT_EQ(json, reducedJSON);
}


TEST_F(SphinxTaskReducersManagerIntegrationTest, throwExceptionWhileAddDataWithInvalidTaskId)
{
  SphinxResultData emptySphinxResult;
  string validJSON = buildInputJSONFromSphinxResultData(emptySphinxResult);
  int invalidTaskId = 1;

  SphinxTaskReducersManager sphinxTaskReducersManager(sphinxTaskReducersStorage,
                             sphinxTaskReducerFactory, sphinxResultJSONConvertor);

  ASSERT_THROW(sphinxTaskReducersManager.addDataInReducer(invalidTaskId, validJSON),
               NotFoundByKeyException);
}


TEST_F(SphinxTaskReducersManagerIntegrationTest, throwExceptionWhileParseWrongJSONFormatString)
{
  string invalidJSON = "data";
  int taskId = 1;

  SphinxTaskReducersManager sphinxTaskReducersManager(sphinxTaskReducersStorage,
                             sphinxTaskReducerFactory, sphinxResultJSONConvertor);

  sphinxTaskReducersManager.createReducer(taskId);

  ASSERT_THROW(sphinxTaskReducersManager.addDataInReducer(taskId, invalidJSON),
               WrongJSONStructureException);
}


TEST_F(SphinxTaskReducersManagerIntegrationTest, reduceOneSphinxResult)
{
  string json = buildOutputJSONFromSphinxResultData(sphinxResultData1);
  string json_input = buildInputJSONFromSphinxResultData(sphinxResultData1);
  int taskId = 7;

  SphinxTaskReducersManager sphinxTaskReducersManager(sphinxTaskReducersStorage,
                            sphinxTaskReducerFactory, sphinxResultJSONConvertor);

  sphinxTaskReducersManager.createReducer(taskId);
  sphinxTaskReducersManager.addDataInReducer(taskId, json_input);
  string reducedJSON = sphinxTaskReducersManager.runReduceTaskForTaskId(taskId);

  ASSERT_EQ(json, reducedJSON);
}


TEST_F(SphinxTaskReducersManagerIntegrationTest, reduceTwoSphinxResults)
{
  SphinxResultData reducedSphinxResult;
  buildExpectedSphinxResult(reducedSphinxResult);

  string expectedJSON = buildOutputJSONFromSphinxResultData(reducedSphinxResult);
  string json1 = buildInputJSONFromSphinxResultData(sphinxResultData1);
  string json2 = buildInputJSONFromSphinxResultData(sphinxResultData2);
  int taskId = 7;

  SphinxTaskReducersManager sphinxTaskReducersManager(sphinxTaskReducersStorage,
                            sphinxTaskReducerFactory, sphinxResultJSONConvertor);

  sphinxTaskReducersManager.createReducer(taskId);
  sphinxTaskReducersManager.addDataInReducer(taskId, json1);
  sphinxTaskReducersManager.addDataInReducer(taskId, json2);

  string reducedJSON = sphinxTaskReducersManager.runReduceTaskForTaskId(taskId);

  ASSERT_EQ(expectedJSON, reducedJSON);
}

