/**
 * @file CldPlugin.cpp
 *
 * CldPlugin (Google Chromium Compact Language Detector) class implementation
 *
 * @author Alexey <developers.hce@gmail.com>
 * @link http://hierarchical-cluster-engine.com/
 * @copyright Copyright &copy; 2013 IOIX Ukraine
 * @license http://hierarchical-cluster-engine.com/license/
 * @package HCE project node API
 * @since 0.1
 */

#include <cld/ext_lang_enc.h>
#include <iostream>
#include <cld/encodings/proto/encodings.pb.h>

#include "CldPlugin.hpp"
#include "CWords.hpp"

namespace HCE
{

namespace component
{


CldPlugin::CldPlugin()
{
	//!<
	//!< adjust Language table
	//!< first of all, make all values from cld enum Language as LM_NONE
	//!<
	for ( unsigned i=0; i<MAX_NUM_LANG; ++i)
	{
        toLang[ static_cast<Language>(i) ] = LM_NONE;
	}

	//!<
	//!< next, adjust support language values to appropriate values by the numeric
	//!< ( used to calculate language mask )
	//!<
	toLang[ static_cast<Language>(0) ] = LM_ENGLISH;
	toLang[ static_cast<Language>(7) ] = LM_ENGLISH;
	toLang[ static_cast<Language>(31) ] = LM_ENGLISH;
	toLang[ static_cast<Language>(32) ] = LM_ENGLISH;
	toLang[ static_cast<Language>(40) ] = LM_ENGLISH;
	toLang[ static_cast<Language>(8) ] = LM_JAPANESE;
    toLang[ static_cast<Language>(13) ] = LM_RUSSIAN;
    toLang[ static_cast<Language>(27) ] = LM_RUSSIAN;

    _adjLang[ static_cast<Language>(13) ] = ::RUSSIAN;
    _adjLang[ static_cast<Language>(27) ] = ::RUSSIAN;
    _adjLang[ static_cast<Language>(29) ] = ::RUSSIAN;
    _adjLang[ static_cast<Language>(34) ] = ::RUSSIAN;
    _adjLang[ static_cast<Language>(36) ] = ::RUSSIAN;
    _adjLang[ static_cast<Language>(37) ] = ::RUSSIAN;
    _adjLang[ static_cast<Language>(47) ] = ::RUSSIAN;
    _adjLang[ static_cast<Language>(50) ] = ::RUSSIAN;

    //!<
	//!< TODO: add languages here as was done before to adjust languages table to your wishes
	//!<
	toLang[ static_cast<Language>(CATALAN) ] = LM_ENGLISH;

	//!<
	//!< adjust support language values to appropriate values by the string
	//!< ( used to select appropriate stemmer/pos-tagger )
	//!<
	toLangByString["en"] = "en";
	toLangByString["ms"] = "en";
	toLangByString["gl"] = "en";
	toLangByString["it"] = "en";
	toLangByString["fil"] = "en";
	toLangByString["sv"] = "en";
	toLangByString["ja"] = "ja";
	toLangByString["zh"] = "ja";
	toLangByString["zh-TW"] = "ja";
    //!<
    //!< Russian table : be bg mk sr uk ur
    //!<
	toLangByString["ru"] = "ru";
    toLangByString["be"] = "ru";
    toLangByString["bg"] = "ru";
    toLangByString["mk"] = "ru";
    toLangByString["sr"] = "ru";
    toLangByString["uk"] = "ru";
    toLangByString["ur"] = "ru";
    //!<
	//!< TODO: add languages here as was done before to adjust languages table to your wishes
	//!<
	toLangByString["ca"] = "en";

	//!<
	//!< fill export table
	//!<
    ProcessInfo processInfo;
    SET_PROCESS_INFO_P(processInfo, LDET, CHROMIUM_LDET, LM_ALL, &TPlugin::process, this)
    _exported.push_back( processInfo );
}


CldPlugin::~CldPlugin() {}


template <class T>
inline
const Language detectLanguage(const T& token)
{
	const char* src = token.c_str();
    bool is_plain_text = true;
    bool do_allow_extended_languages = true;
    bool do_pick_summary_language = false;
    bool do_remove_weak_matches = false;
    bool is_reliable;
    const char* tld_hint = NULL;
    int encoding_hint = UNKNOWN_ENCODING;
    Language language_hint = UNKNOWN_LANGUAGE;

    double normalized_score3[3];
    Language language3[3];
    int percent3[3];
    int text_bytes;

    Language lang;
    lang = CompactLangDet::DetectLanguage(0,
                                          src, strlen(src),
                                          is_plain_text,
                                          do_allow_extended_languages,
                                          do_pick_summary_language,
                                          do_remove_weak_matches,
                                          tld_hint,
                                          encoding_hint,
                                          language_hint,
                                          language3,
                                          percent3,
                                          normalized_score3,
                                          &text_bytes,
                                          &is_reliable);
  return lang;
}


void CldPlugin::process(LCoreData& lCData)
{
    //!<
    //!< detect preffered document language
    //!<
    _basis_lang_mask = detectLanguage(lCData._content);
#ifdef _DEBUG_
    std::cout << "basic language: " << _basis_lang_mask<<std::endl;
#endif
    std::map<WORD_CONTENT_OFFSET, SSTRING>::const_iterator it;
    for ( it=lCData._mapTokens.begin(); it!=lCData._mapTokens.end(); ++it )
    {
        //!<
        //!< get language by value (e.g. 0, 7, 13, etc.)
        //!<
        Language lang = detectLanguage( it->second );
        //!<
        //!< adjust language
        //!<
        adjustLanguage(lang);
        //!<
        //!< get language by string (e.g. "ru", "en", etc.)
        //!<
        //SSTRING iso639_1 = LanguageCode(lang);
        //std::string lstr = toLangByString[iso639_1];
        //lCData._mapLanguages[ it->first ] = (lstr=="") ? "en" : lstr;

        if(lang==UNKNOWN_LANGUAGE) {
            lCData._mapWordType[it->first ] = WORD_TYPE::DELIMITER;
            lang = langAdditionDetection(it->second, lCData._mapWordType[it->first ]);
        } else {
            lCData._mapWordType[it->first ] = WORD_TYPE::WORD;
        }
                //!<
                //!< SPIKE: insert "en" iso639_1 language if not appropriate ("")
                //!<
        lCData._mapLanguages[ it->first ] = correctLanguage(lang);
        lCData._mapLanguagesMasks[ it->first ] |= toLang[lang] | _basis_lang_mask<<1;
        //!<
        //!< fill map of language's vectors
        //!<
        lCData._mapVectorsOfLanguages[ static_cast<Language>( toLang[lang] ) ].push_back( it->first );


    }
#ifdef _DEBUG_
    std::cout << std::endl << "CLD plugin:" << std::endl << std::endl;
    //!<
    //!< display the map of the Languages
    //!<
    std::map<WORD_CONTENT_OFFSET, WORD_LENGTH>::const_iterator it1;
    std::cout << "token \t language" << std::endl;
    std::cout << "----- \t ----" << std::endl;
    for( it1=lCData._mapOffsets.begin(); it1!=lCData._mapOffsets.end(); ++it1)
    {
        std::cout << lCData._mapTokens[it1->first] << '\t'
                << lCData._mapLanguages[it1->first] << std::endl;
    }
    std::cout << "Tokens count: " << lCData._mapOffsets.size() << std::endl << std::endl;
#endif
}

/**
 * Addition language and word type detection
 */

Language CldPlugin::langAdditionDetection(const std::string &str, WORD_TYPE &wType)
{
  Language ret = UNKNOWN_LANGUAGE;
  for(unsigned int i = 0; i < str.size(); i++)
  {
    if(str[i] >= '0' && str[i] <= '9')
    {
      ret = ::ENGLISH;
      wType = (i == 0) ? WORD_TYPE::NUMBER : WORD_TYPE::WORD;
      break;
    }
  }  
  return ret;
}


} /* namespace component */

} /* namespace HCE */
