/**
 * @file Refine.cpp
 *
 * Refine class implementation
 *
 * @author Alexey <developers.hce@gmail.com>
 * @link http://hierarchical-cluster-engine.com/
 * @copyright Copyright &copy; 2013 IOIX Ukraine
 * @license http://hierarchical-cluster-engine.com/license/
 * @package HCE project node API
 * @since 0.1
 */

//!< std headers
#include <iostream>
#include <cld/compact_lang_det.h>

#include "Refine.hpp"
#include "OutDataRefine.hpp"
#include "RWords.hpp"
#include "CWords.hpp"
#include "LCoreData.hpp"
#include "HCETimer.hpp"


namespace HCE
{

namespace component
{


Refine::Refine(ComponentType inType) : _lCore()
{
	//!<
	//!< instance of the smth
	//!<
	ProcessInfo processInfo;

	/**
	 * Define content processing schema
	 * If input message hasn't provide it's own content processing schema
	 * Refine component apply default one:
	 *
	 * 1. Reduce tags from raw content
	 * 2. Split raw content on the tokens
	 * 3. Detect language's mask for each token in splitted content
	 * 4. Normalize Japanese tokens
	 * 5. Normalize European Languages tokens(Russian, English, etc.)
	 * 6. Part of speech of tokens
	 * 7. CRC64 of the normalized token's form
	 */


	//!<
	//!< tagger pos reduce
	//!<
	SET_PROCESS_INFO_L(processInfo, TAGS, TAG_REDUCE, LM_ALL )
	_processInfoTable.push_back( processInfo );

	//!<
	//!< split content into the tokens
	//!< Set type of the split content on the tokens
	//!< Available tokenizers:
	//!<  1. ICU
	//!<  2. Boost (methods: split and tokenizer)
	//!<  3. MeCab
	//!<
	//SET_PROCESS_INFO_L(processInfo, SPLIT, ICU_BOUNDARIES, LM_ALL)
    //_processInfoTable.push_back( processInfo );
	//!< or
	//SET_PROCESS_INFO_L(processInfo, SPLIT, BOOST_SPLIT, LM_ALL)
    //_processInfoTable.push_back( processInfo );
	//!< or
	SET_PROCESS_INFO_L(processInfo, SPLIT, BOOST_TOKENIZER, LM_ALL)
    _processInfoTable.push_back( processInfo );
	//!< or
	//SET_PROCESS_INFO_L(processInfo, NORM, MECAB_NORM, LM_ALL )
	//_processInfoTable.push_back( processInfo );

	//!<
	//!< detect language for each token
	//!<
	SET_PROCESS_INFO_L(processInfo, LDET, CHROMIUM_LDET, LM_ALL )
	_processInfoTable.push_back( processInfo );

	//!<
	//!< perform normalize for Japanese tokens
	//!<
    //SET_PROCESS_INFO_L(processInfo, NORM, MECAB_NORM, LM_JAPANESE )
    //SET_PROCESS_INFO_L(processInfo, NORM, MECAB_NORM, LM_ALL )
    //_processInfoTable.push_back( processInfo );

	//!<
	//!< perform normalize for other languages
	//!<
	SET_PROCESS_INFO_L(processInfo, NORM, SNOWBALL_NORM, LM_ALL )
	_processInfoTable.push_back( processInfo );

	//!<
	//!< Part Of Speech
	//!<
	SET_PROCESS_INFO_L(processInfo, POS, TURGLEM_POS, LM_ALL )
	_processInfoTable.push_back( processInfo );

	//!<
	//!< CRC64
	//!<
	SET_PROCESS_INFO_L(processInfo, HCRC, FAKE_HCRC, LM_ALL )
	_processInfoTable.push_back( processInfo );
}


Refine::~Refine() {}


Poco::SharedPtr<DataBase> Refine::process(const Poco::SharedPtr<DataBase> inData)
{
	//!<
	//!< timer statistic
	//!<
#ifdef _DEBUG_
	HCE::HCETimer hceTimer("Parsing content time: ");
#endif

	//!<
	//!<
	//!<
    LCoreData lCData( inData.cast<InDataRefine>()->getContent() );
    lCData._normalizationId = inData.cast<InDataRefine>()->getNormalizationId();
    lCData._delimiters = inData.cast<InDataRefine>()->getDelimiters();
    lCData._granularityLevel= inData.cast<InDataRefine>()->getGranularityLevel();
    std::vector<ProcessInfo> ProcessInfoTable = \
			( inData.cast<InDataRefine>()->getProcessInfoTable().size() ) ? \
					inData.cast<InDataRefine>()->getProcessInfoTable() :\
					this->_processInfoTable;

	//!<
	//!< main processing loop
	//!<
	//BOOST_FOREACH( ProcessInfo& processInfo, ProcessInfoTable )
	std::vector<ProcessInfo>::iterator PI_iter;
	for ( PI_iter = ProcessInfoTable.begin(); PI_iter != ProcessInfoTable.end(); ++PI_iter )
	{
#ifdef _DEBUG_
         std::cout << "------------------------" << std::endl;
#endif
        lCData._processInfo = *PI_iter; //processInfo;
		_lCore.selectMethod(lCData);
		/*processInfo*/PI_iter->_status = lCData._processInfo._status;
	}
	//!<
	//!< size of the map
	//!<
	//_log_ << "Tokens count" << lCData._mapOffsets.size() << _log_end_;

	//!<
	//!< fill OutDataRefine
	//!<
	Poco::SharedPtr<HCE::DataBase> outDataRefine(new HCE::OutDataRefine(HCE::CT_REFINE) );
	std::vector < CWords > cWords;
	std::vector < RWords > rWords;
	std::map<WORD_CONTENT_OFFSET, WORD_LENGTH>::const_iterator it;
#ifdef _DEBUG_
    std::cout << "token \t norm \t offset \t length \t pos \t hCrc64" << std::endl;
    std::cout << "----- \t ---- \t ------ \t ------ \t --- \t ------" << std::endl;
#endif

    //!<
	//!< for each token extracted from content
    //!<
	for ( it=lCData._mapOffsets.begin(); it!=lCData._mapOffsets.end(); ++it )
	{
		//!<
		//!< cword's instance
		//!<
		CWords cword;
		/**
		 * That fields must be inserted
		 *
		 *  unsigned char black;			//!< refine
			unsigned short simClass; 		//!< refine 	two  bytes morphology ( MorphChangeGrad )
			unsigned int hCrc;				//!< refine CRC32 word ( CRC word for highlight on CDR )
			unsigned int offset; 			//!< refine
		    unsigned int sentenceNumber; 	//!< refine 	(deprecated) number word's sentence, start from begin
		    unsigned char lingIntegrity; 	//!< refine	valuable of the word in the content ( val/unval content )
		    unsigned int initWordLen;		//!< refine
		    std::string normWord;			//!< refine
		    POSMaskBitset<POS_NUM> _posMask;
		 */
		//!<
		//!< set word blacklist
		//!<
		cword.setBlack( lCData._mapBlack[it->first] );

		//!<
		//!< set word morphology
		//!<
		cword.setSimClass( lCData._mapMorphology[it->first].getMask() );

		//!<
		//!< set word CRC for highlighting
		//!<
		//cword.setHCrc( lCData._mapHCRC64[ it->first ] );

		//!<
		//!< set word offset
		//!<
		cword.setOffset( it->first );

		//!<
		//!< set word's sentence number
		//!<
		cword.setSentenceNumber( lCData._mapSentences[ it->first ] );

		//!<
		//!< set word's linguistic integrity
		//!<
		cword.setLingIntegrity( lCData._mapLingIntegrity[ it->first ] );

		//!<
		//!< set init word length
		//!<
		cword.setInitWordLen( it->second );

        //!<
        //!< set original word form
        //!<
        cword.setOriginWord( lCData._mapTokens[it->first] );

        //!<
        //!< set normalized word form
        //!<
        cword.setNormWord( lCData._mapNormalizedForms[it->first] );

        //!<
        //!< set Part-Of-Speech word's mask
        //!<
        //cword.setPosMask ( lCData._mapPosMasks[ it->first ].getMask() );
        cword.setPosMask ( lCData._mapPosMasks[ it->first ]);

        //!<
        //!< set word's type
        //!<
        cword.setWordType(lCData._mapWordType[ it->first ]);

		//!<
		//!< insert cword to vector
		//!<
		cWords.push_back( cword );

		//!<
		//!< rword's instance
		//!<
		RWords rword;
		/**
		 * That fields must be inserted
		 *
		std::string _word;
		unsigned long long _crc64;
		POSMaskBitset<POS_NUM> _posMask;
		MorphChangeGradBitset<MCG_NUM> _morphChangeGrad;
		 */
		//!<
		//!< set word blacklist
		//!<
		rword.setNormWord ( lCData._mapNormalizedForms[it->first] );
		rword.setCrc64 ( lCData._mapHCRC64[ it->first ] );
		//rword.setPosMask ( lCData._mapPosMasks[ it->first ].getMask() );
        rword.setPosMask ( lCData._mapPosMasks[ it->first ] );
		rword.setMorphChangeGradMask ( lCData._mapMorphology[ it->first ].getMask() );
		rWords.push_back(rword);

#ifdef _DEBUG_
        std::cout << lCData._mapTokens[it->first] << '\t'
        		<< lCData._mapNormalizedForms[it->first] << '\t'
        		<< it->first << '\t'
				<< lCData._mapOffsets[it->first] << '\t'
                //<< lCData._mapPosMasks[it->first].getMask() << '\t'
                << lCData._mapPosMasks[it->first]<< '\t'
				<< lCData._mapHCRC64[it->first] << std::endl;
#endif
	}
	unsigned int langMask = 0;
	unsigned char err = 0;
	outDataRefine.cast<OutDataRefine>()->setCWords(cWords);
	outDataRefine.cast<OutDataRefine>()->setRWords(rWords);
	outDataRefine.cast<OutDataRefine>()->setLangMask(langMask);
	outDataRefine.cast<OutDataRefine>()->setLocalErrorCode(err);

	return outDataRefine;
}


} /* namespace component */

} /* namespace HCE */
