/**
 * @file main.cpp
 *
 * Test Application main module. Tests of JsonMessageCover class.
 *
 * @author Alexander Vybornyh <alexander.hce.cluster@gmail.com>
 * @link http://hierarchical-cluster-engine.com/
 * @copyright Copyright &copy; 2013 IOIX Ukraine
 * @license http://hierarchical-cluster-engine.com/license/
 * @package HCE project node API
 * @since 0.1
 */

#include <gtest/gtest.h>
#include <sstream>

#include "JsonMessageCover.hpp"

const std::string data = "Hello world!";
HCE::types::MessageType type = HCE::types::MessageType::mtSphinx;
unsigned int ttl = 234;	

const std::string validDecodeJson = "{\"data\":\"Hello world!\",\"ttl\":234,\"type\":1}";
const std::string validEncodeJson = "{\"data\":\"SGVsbG8gd29ybGQh\",\"ttl\":234,\"type\":1}";

const std::string invalidDecodeJson = "{\"data\":\"Hello world!\",\"ttl\":234,\"typ";
const std::string invalidEncodeJson = "{\"data\":\"SGVsbG8gd29ybGQh\",\"ttl\":234,\"typ";


TEST(JsonMessageCover, serializeJsonMessageCover)
{
	std::string json;
	HCE::JsonMessageCover jsonMessageCover;
	jsonMessageCover.setType(type);
	jsonMessageCover.setData(data);
	jsonMessageCover.setTTL(ttl);	

	ASSERT_TRUE(jsonMessageCover.serialize(json));
	ASSERT_FALSE(jsonMessageCover.isError());
	ASSERT_EQ(jsonMessageCover.getErrorCode(), HCE::JsonMessageCover::ErrorCode::NO_ERROR);
	ASSERT_TRUE(jsonMessageCover.getErrorMsg().empty());
	ASSERT_FALSE(json.empty());
}
 
TEST(JsonMessageCover, unserializeJsonMessageCover)
{
	std::string json;

#if JSON_USE_BASE64 == 1
	json = validEncodeJson;
#else
	json = validDecodeJson;
#endif

	HCE::JsonMessageCover jsonMessageCover;
	ASSERT_TRUE(jsonMessageCover.unserialize(json));
	ASSERT_FALSE(jsonMessageCover.isError());
	ASSERT_EQ(jsonMessageCover.getErrorCode(), HCE::JsonMessageCover::ErrorCode::NO_ERROR);
	ASSERT_TRUE(jsonMessageCover.getErrorMsg().empty());

	ASSERT_EQ(jsonMessageCover.getType(), HCE::types::MessageType::mtSphinx);
	ASSERT_EQ(jsonMessageCover.getData(), data);
	ASSERT_EQ(jsonMessageCover.getTTL(), ttl);	
}

TEST(JsonMessageCover, useOutputStream)
{
	HCE::JsonMessageCover jsonMessageCover;
	jsonMessageCover.setType(type);
	jsonMessageCover.setData(data);
	jsonMessageCover.setTTL(ttl);

	std::stringstream json;
	json << jsonMessageCover;

	ASSERT_FALSE(jsonMessageCover.isError());
	ASSERT_EQ(jsonMessageCover.getErrorCode(), HCE::JsonMessageCover::ErrorCode::NO_ERROR);
	ASSERT_TRUE(jsonMessageCover.getErrorMsg().empty());
	ASSERT_FALSE(json.str().empty());
}

TEST(JsonMessageCover, useInputStream)
{
	std::stringstream json;

#if JSON_USE_BASE64 == 1
	json << validEncodeJson;
#else
	json << validDecodeJson;
#endif

	HCE::JsonMessageCover jsonMessageCover;
	json >> jsonMessageCover;

	ASSERT_FALSE(jsonMessageCover.isError());
	ASSERT_EQ(jsonMessageCover.getErrorCode(), HCE::JsonMessageCover::ErrorCode::NO_ERROR);
	ASSERT_TRUE(jsonMessageCover.getErrorMsg().empty());

	ASSERT_EQ(jsonMessageCover.getType(), HCE::types::MessageType::mtSphinx);
	ASSERT_EQ(jsonMessageCover.getData(), data);
	ASSERT_EQ(jsonMessageCover.getTTL(), ttl);	
}

TEST(JsonMessageCover, errorMessageJsonMessageCover)
{
	std::stringstream json;

#if JSON_USE_BASE64 == 1
	json << invalidEncodeJson;
#else
	json << invalidDecodeJson;
#endif

	HCE::JsonMessageCover jsonMessageCover;
	json >> jsonMessageCover;

	ASSERT_TRUE(jsonMessageCover.isError());
	ASSERT_EQ(jsonMessageCover.getErrorCode(), HCE::JsonMessageCover::ErrorCode::ERROR_UNSERIALIZE);
	ASSERT_FALSE(jsonMessageCover.getErrorMsg().empty());
}

int main (int argc, char** argv)
{
  testing::InitGoogleTest(&argc, argv);
  return RUN_ALL_TESTS();
}

