#include <iostream>
#include <sstream>
#include <fstream>
#include <assert.h>
#include <string.h>
#include <vector>

#include "SphinxInputJsonMessage.hpp"
#include "SphinxOutputJsonMessage.hpp"
#include "SphinxSearcher.hpp"
#include "SphinxError.hpp"
#include "SphinxFunctionalObject.hpp"
#include "SphinxResultDataSerializator.hpp"
#include "SphinxInputJsonMessageSearch.hpp"

/*
     {"data":"{\"parameters\":[{\"queryId\":\"777\"},{\"JsonType\":\"7\"}],\"q\":\"test\"}","type":0}
*/

void testInputJsonMessage(void)
{
  const std::string inputJson = "{\"data\":\"{\\\"parameters\\\":[{\\\"queryId\\\":\\\"777\\\"},{\\\"JsonType\\\":\\\"7\\\"}],\\\"q\\\":\\\"test\\\"}\",\"type\":0}";

  // block sample of unserialize
  HCE::sphinx::SphinxInputJsonMessage inputMsg(inputJson);

  std::cout << "ErrorMsg = " << inputMsg.getErrorMsg() << std::endl;
  std::cout << "ErrorCode = " << inputMsg.getErrorCode() << std::endl;
  assert(!inputMsg.isError());

  std::cout << "Input json type = " << (unsigned int)(inputMsg.getType()) << std::endl;
  std::cout << "Json data count = " << inputMsg.getData().size() << std::endl;

  assert(inputMsg.getType()==HCE::sphinx::SphinxInputJsonMessage::MessageType::mtSearch);
  assert(!inputMsg.getData().empty());

  HCE::sphinx::SphinxInputJsonMessageSearch messageSearch(inputMsg.getData());
  assert(!messageSearch.isError());

  std::cout << "Query string: '" << messageSearch.getQueryString() << "'" << std::endl;
  std::cout << "Parameters count = " << messageSearch.getQueryParameters().size() << std::endl;

  assert(messageSearch.getQueryString()=="test");
  assert(messageSearch.getQueryParameters().size()==2);

  for (size_t i=0;i<messageSearch.getQueryParameters().size();++i)
  {
    if (messageSearch.getQueryParameters()[i].first=="queryId")
      assert(messageSearch.getQueryParameters()[i].second=="777");

    if (messageSearch.getQueryParameters()[i].first=="JsonType")
      assert(messageSearch.getQueryParameters()[i].second=="7");
    std::cout << "Parameter name: '" << messageSearch.getQueryParameters()[i].first << "' value: '" << messageSearch.getQueryParameters()[i].second  << "'" << std::endl;
  }

  // block sample of serialize
  std::string json;
  HCE::sphinx::SphinxInputJsonMessageSearch inputSearch;
  inputSearch.setQueryString("test");
  inputSearch.addQueryParameters("queryId", "777");
  inputSearch.addQueryParameters("JsonType", "7");
  inputSearch.serialize(json);

  inputMsg.setType(HCE::sphinx::SphinxInputJsonMessage::MessageType::mtSearch);
  inputMsg.setData(json);

  std::stringstream str;
  str << inputMsg;

  std::cout << " input json: '" << inputJson << "'\n";
  std::cout << "output json: '" << str.str() << "'\n";
  assert(str.str()==inputJson);
  std::cout << "testInputJsonMessage - OK" << std::endl;
}
//-----------------------------------------------------------------------------
/*
     {"data":"Some content of data","error_code":2000,"error_message":"Some error","time":23}
*/
// cppcheck-suppress unusedFunction
void testOutputJsonMessage(void)
{
  const std::string inputJson = "{\"data\":\"U29tZSBjb250ZW50IG9mIGRhdGE=\",\"error_code\":2000,\"error_message\":\"Some error\",\"time\":23}";

  // block sample of unserialize
  HCE::sphinx::SphinxOutputJsonMessage outputMessage;

  assert(outputMessage.unserialize(inputJson));
  assert(!outputMessage.isError());

  std::cout << "Error code = " << outputMessage.getErrorCode() << std::endl;
  std::cout << "Error message: '" << outputMessage.getErrorMessage() << "'" << std::endl;
  std::cout << "Json data: '" << outputMessage.getData() << "'" << std::endl;
  std::cout << "Time = " << outputMessage.getTime() << std::endl;

  assert(outputMessage.getErrorCode()==2000);
  assert(outputMessage.getErrorMessage()=="Some error");

  std::cout << "outputMessage.getData(): '" << outputMessage.getData() << "'" << std::endl;

  assert(outputMessage.getData()=="Some content of data");
  assert(outputMessage.getTime()==23);

  // block sample of serialize
  outputMessage.clear(); // clear object for repeat use
  outputMessage.setErrorCode(2000);
  outputMessage.setErrorMessage("Some error");
  outputMessage.setData("Some content of data");
  outputMessage.setTime(23);

  std::string outputJson;
  assert(outputMessage.serialize(outputJson));
  assert(!outputMessage.isError());

  std::cout << " input json: '" << inputJson << "'\n";
  std::cout << "output json: '" << outputJson << "'\n";
  assert(inputJson==outputJson);
  std::cout << "testOutputJsonMessage - OK" << std::endl;
}
//-----------------------------------------------------------------------------
//-----------------------------------------------------------------------------
// cppcheck-suppress unusedFunction
void testDefaultSearch(void)
{
  const std::string nodeName = "test_node";
  const std::string homeDir = "/home/alexander/.local/share/hce";
  const std::string indexName = "i004";

  HCE::sphinx::SphinxFunctionalObject fObj(nodeName, homeDir, indexName, HCE::sphinx::sphinx_search_const::searchdStart, HCE::sphinx::sphinx_search_const::searchdStop);

  std::cout << fObj.logMsg() << std::endl;
  fObj.setLogStream(std::cout);

  std::cout << "LOAD CONFIG => " << std::boolalpha << fObj.isLoadedConfig() << "\tIS ERROR => " << std::boolalpha << fObj.isError()
      << "\nERROR MSG: " << fObj.getErrorMsg() << "\tERROR CODE: " << fObj.getErrorCode() << std::endl;

  const std::string query = "apple";

  std::string json;
  HCE::sphinx::SphinxInputJsonMessageSearch msgSearch;
//  std::vector<std::string> attrs = {"1"};
//  msgSearch.addFilter((HCE::sphinx::SphinxFilter::FilterType)0, "mmedia", attrs, (HCE::sphinx::SphinxFilter::ExcludeType)0);

  msgSearch.setQueryString(query);
  msgSearch.addQueryParameters("queryId", "123");
  msgSearch.addQueryParameters("JsonType", "15");
  msgSearch.addQueryParameters("timeout", "5");
  msgSearch.addQueryParameters("max_results", "1000");
  msgSearch.addQueryParameters("offset", "0");
  msgSearch.addQueryParameters("limit", "0");
  msgSearch.addQueryParameters("cutoff", "0");
  msgSearch.addQueryParameters("sort_by", "mmedia");
  msgSearch.addQueryParameters("order_by", "0");
  msgSearch.addQueryParameters("ttl", "3");

  msgSearch.addExternalFields("uid");
  msgSearch.addExternalFields("sid");
  msgSearch.addExternalFields("lang");

  msgSearch.addOrderParameters("algorithm", "0");
  msgSearch.addOrderParameters("order_by", "2");
  msgSearch.addOrderFields("uid");
  msgSearch.addOrderFields("lang");
  msgSearch.addOrderFields("sid");

  assert(msgSearch.serialize(json));

  HCE::sphinx::SphinxInputJsonMessage handler;
  handler.setType(HCE::sphinx::SphinxInputJsonMessage::MessageType::mtSearch);
  handler.setData(json);

  std::stringstream str;
  str << handler;

  std::string ret = fObj.Process(str.str());  // run functional object

  if (!fObj.getErrorMsg().empty())
    std::cout << "fObj.getErrorMsg()  = " << fObj.getErrorMsg() << std::endl;
  assert(fObj.getErrorMsg().empty());

  HCE::sphinx::SphinxResultData Result;
  HCE::sphinx::SphinxResultDataSerializator ResultSerializator(Result);

  HCE::sphinx::SphinxOutputJsonMessage outputMessage(ret);
  assert(!outputMessage.isError());
  assert(ResultSerializator.unserialize(outputMessage.getData()));

  // sample of output search result into stream
  std::cout << "Result:\n" << Result << std::endl;

  // sample possibility getting of data out SphinxResultData class
  const size_t matchInfoCount = Result.getMatchInfoCount();
  for (size_t i=0;i<matchInfoCount;++i)
  {
    unsigned long long docId = Result.getMatchInfoItem(i).getDocId();
    std::string weight = Result.getMatchInfoItem(i).getWeight();

    std::cout << "docId: " << docId << " weight: " << weight << std::endl;

    const size_t attributesCount = Result.getMatchInfoItem(i).getAttributesCount();
    for (size_t j=0;j<attributesCount;++j)
    {
      std::cout << Result.getMatchInfoItem(i).getAttrItem(j).name << " = " << Result.getMatchInfoItem(i).getAttrItem(j).value << std::endl;
    }
  }

  const size_t requestInfoCount = Result.getRequestInfoCount();
  for (size_t i=0;i<requestInfoCount;++i)
  {
    std::cout << "nodeName: "         << Result.getRequestInfoItem(i).getNodeName() << std::endl;
    std::cout << "query: "            << Result.getRequestInfoItem(i).getQuery() << std::endl;
    std::cout << "queryId: "          << Result.getRequestInfoItem(i).getQueryId() << std::endl;
    std::cout << "maxResultsNumber: " << Result.getRequestInfoItem(i).getMaxResultsNumber() << std::endl;
    std::cout << "orderBy: "          << Result.getRequestInfoItem(i).getOrderBy() << std::endl;
    std::cout << "total: "            << Result.getRequestInfoItem(i).getTotal() << std::endl;
    std::cout << "totalFound: "       << Result.getRequestInfoItem(i).getTotalFound() << std::endl;
    std::cout << "timeMsec: "         << Result.getRequestInfoItem(i).getTimeMsec() << std::endl;

    const size_t wordsCount = Result.getRequestInfoItem(i).getWordsCount();
    for (size_t j=0;j<wordsCount;++j)
    {
      std::cout << "word: " << Result.getRequestInfoItem(i).getWordItem(j).word
          << ", hits: " << Result.getRequestInfoItem(i).getWordItem(j).hits
          << ", docs: " << Result.getRequestInfoItem(i).getWordItem(j).docs << std::endl;
    }
  }
}
//-----------------------------------------------------------------------------
//-----------------------------------------------------------------------------
// cppcheck-suppress unusedFunction
void testWriteToLogFunctionalObject(void)
{
  const std::string nodeName = "test_node";
  const std::string dataDir = "/var/lib/hce_cluster";
  const std::string defaultIndex = "default_index";
  HCE::sphinx::SphinxFunctionalObject fObj(nodeName, dataDir, defaultIndex);
  fObj.log(HCE::LoggerStream::Priority::PRIO_INFORMATION) << "Create object...\n";
  std::cout << "Log test:\n";
  std::cout << "--------------------------\n" << fObj.logMsg(false) << std::endl;
  fObj.log(HCE::LoggerStream::Priority::PRIO_INFORMATION) << "Write some text...\n";
  std::cout << "--------------------------\n" << fObj.logMsg() << std::endl;

  // make output stream for save log
  std::ofstream ofs("./out.log", std::fstream::trunc);
  // init functional object from stream
  fObj.setLogStream(ofs);
  fObj.log(HCE::LoggerStream::Priority::PRIO_INFORMATION) << "first line to cout...\n";
  fObj.log(HCE::LoggerStream::Priority::PRIO_INFORMATION) << "second line to cout...\n";
  fObj.log(HCE::LoggerStream::Priority::PRIO_INFORMATION) << "three line to cout...\n";
  ofs.close();
  // next don't wrote
  fObj.log(HCE::LoggerStream::Priority::PRIO_INFORMATION) << "Added some text...\n";
  // log buffer is empty
  std::cout << "--------------------------\n" << fObj.logMsg() << std::endl;
}
//-----------------------------------------------------------------------------
//-----------------------------------------------------------------------------


