/*
 * SphinxResultDataSerializator.cpp
 *
 *  Created on: Oct 30, 2013
 *      Author: alexander
 */

#include <Poco/JSON/Object.h>
#include <Poco/JSON/Array.h>
#include <Poco/JSON/JSON.h>
#include <Poco/JSON/Stringifier.h>
#include <Poco/JSON/Parser.h>
#include <Poco/JSON/JSONException.h>
#include <Poco/Dynamic/Var.h>
#include <sstream>
#include <iomanip>
#include <sys/time.h>

#include "SphinxInputJsonMessage.hpp"
#include "EncodeDecodeBase64.hpp"
#include "SphinxError.hpp"

#include "SphinxResultDataSerializator.hpp"

namespace HCE
{
namespace sphinx
{
//-----------------------------------------------------------------------------
std::string SphinxWeightPacker::pack(const std::string& data, unsigned int minNumberFieldsPacking)
{
  std::ostringstream ostr;

  if (minNumberFieldsPacking<=fieldsCount)
  {
    const size_t count = data.length()/fieldWidth;

    if (!fieldsCount)
      ++fieldsCount;

    for (size_t i=0;i<fieldsCount;++i)
    {
      std::string hexString;
      if (i<count)
        hexString = data.substr(i*fieldWidth, fieldWidth);
      size_t pos = hexString.find_first_not_of('0');
      ostr << delimiter;
      if (pos!=std::string::npos)
        ostr << hexString.substr(pos);
    }
  }
  else
    ostr << data;

  return ostr.str();
}
//-----------------------------------------------------------------------------
std::string SphinxWeightPacker::unpack(const std::string& data)
{
  std::ostringstream ostr;
  size_t count = 0;
  for (size_t i=0;i<data.size();++i)
    if (data[i]==delimiter)
      ++count;

  if (count)
  {
    std::istringstream istr(data);
    std::string line;
    bool begin = false;
    while(getline(istr, line, delimiter))
    {
      if (begin)
      {
        ostr << std::setfill('0') << std::setw(fieldWidth) << line;
        --count;
      }
      else
        begin = true;
    }
    if (count)
      ostr << std::setfill('0') << std::setw(fieldWidth) << "";
  }
  else
    ostr << data;

  return ostr.str();
}
//-----------------------------------------------------------------------------
//-----------------------------------------------------------------------------
SphinxResultDataSerializator::SphinxResultDataSerializator(SphinxResultData& resultData_)
: inherited(), resultData(resultData_), jsonType(FILL_MATCH_INFO | FILL_REQUEST_INFO | FILL_ATTRIBUTES | FILL_WORD_INFO)
{
}
//-----------------------------------------------------------------------------
SphinxResultDataSerializator::SphinxResultDataSerializator(SphinxResultData& resultData_, const std::string& json)
: inherited(), resultData(resultData_), jsonType(FILL_MATCH_INFO | FILL_REQUEST_INFO | FILL_ATTRIBUTES | FILL_WORD_INFO)
{
  unserialize(json);
}
//-----------------------------------------------------------------------------
SphinxResultDataSerializator::SphinxResultDataSerializator(SphinxResultData& resultData_, SphinxResultDataSerializator::JsonType jsonType_)
: inherited(), resultData(resultData_), jsonType(jsonType_)
{
}
//-----------------------------------------------------------------------------
SphinxResultDataSerializator::~SphinxResultDataSerializator(void)
{
}
//-----------------------------------------------------------------------------
bool SphinxResultDataSerializator::serialize(SphinxResultDataSerializator::JsonType jsonType, std::string& json, unsigned int fieldCount, unsigned int minNumberFieldsPacking)
{
  errorMsg.clear();
  errorCode = NO_ERROR;
  SphinxWeightPacker packer(fieldCount);
  try
  {
    Poco::JSON::Object::Ptr obj = new Poco::JSON::Object();

    // make json objects for matches info
    Poco::JSON::Array::Ptr pMatches = new Poco::JSON::Array();
    if ((jsonType & FILL_MATCH_INFO)!=0)
    {
      const size_t matchInfoCount = resultData.getMatchInfoCount();
      for (size_t i=0;i<matchInfoCount;++i)
      {
        Poco::JSON::Object::Ptr pMatchElem = new Poco::JSON::Object();

        std::string docId = std::to_string(resultData.getMatchInfoItem(i).getDocId());
        pMatchElem->set(search_result_json_const::docId, docId);
        pMatchElem->set(search_result_json_const::weight, packer.pack(resultData.getMatchInfoItem(i).getWeight(), minNumberFieldsPacking));

        Poco::JSON::Array::Ptr pAttributes = new Poco::JSON::Array();
        if ((jsonType & FILL_ATTRIBUTES)!=0)
        {
          const size_t attributesCount = resultData.getMatchInfoItem(i).getAttributesCount();
          for (size_t j=0;j<attributesCount;++j)
          {
            Poco::JSON::Object::Ptr pAttrElem = new Poco::JSON::Object();
            pAttrElem->set(resultData.getMatchInfoItem(i).getAttrItem(j).name, resultData.getMatchInfoItem(i).getAttrItem(j).value);
            pAttributes->add(pAttrElem);
          }
        }
        pMatchElem->set(search_result_json_const::attrArray, pAttributes);
        pMatches->add(pMatchElem);
      }
    }
    obj->set(search_result_json_const::matchInfo, pMatches);

    // make json objects for request info
    Poco::JSON::Array::Ptr pRequests = new Poco::JSON::Array();

    if ((jsonType & FILL_REQUEST_INFO)!=0)
    {
      const size_t requestInfoCount = resultData.getRequestInfoCount();
      for (size_t i=0;i<requestInfoCount;++i)
      {
        Poco::JSON::Object::Ptr pRequestElem = new Poco::JSON::Object();

        pRequestElem->set(search_result_json_const::nodeName,          resultData.getRequestInfoItem(i).getNodeName());
        pRequestElem->set(search_result_json_const::query,             resultData.getRequestInfoItem(i).getQuery());
        pRequestElem->set(search_result_json_const::queryId,           resultData.getRequestInfoItem(i).getQueryId());
        pRequestElem->set(search_result_json_const::maxResultsNumber,  resultData.getRequestInfoItem(i).getMaxResultsNumber());
        pRequestElem->set(search_result_json_const::orderBy,           resultData.getRequestInfoItem(i).getOrderBy());
        pRequestElem->set(search_result_json_const::total,             resultData.getRequestInfoItem(i).getTotal());
        pRequestElem->set(search_result_json_const::totalFound,        resultData.getRequestInfoItem(i).getTotalFound());
        pRequestElem->set(search_result_json_const::timeMsec,          resultData.getRequestInfoItem(i).getTimeMsec());

        Poco::JSON::Array::Ptr pWords = new Poco::JSON::Array();
        if ((jsonType & FILL_WORD_INFO)!=0)
        {
          const size_t wordsCount = resultData.getRequestInfoItem(i).getWordsCount();
          for (size_t j=0;j<wordsCount;++j)
          {
            Poco::JSON::Object::Ptr pWordElem = new Poco::JSON::Object();

            pWordElem->set(search_result_json_const::word, resultData.getRequestInfoItem(i).getWordItem(j).word);
            pWordElem->set(search_result_json_const::hits, resultData.getRequestInfoItem(i).getWordItem(j).hits);
            pWordElem->set(search_result_json_const::docs, resultData.getRequestInfoItem(i).getWordItem(j).docs);
            pWords->add(pWordElem);
          }
        }
        pRequestElem->set(search_result_json_const::wordArray, pWords);
        pRequests->add(pRequestElem);
      }
    }
    obj->set(search_result_json_const::requestInfo, pRequests);

    std::stringstream ostr;
    Poco::JSON::Stringifier::stringify(obj, ostr);
    json = ostr.str();
    _isError = false;
  }
  catch(std::exception& e)
  {
    _isError = true;
    errorMsg = e.what();
    errorCode = ERROR_SERIALIZE;
  }
  return !_isError;
}
//-----------------------------------------------------------------------------
bool SphinxResultDataSerializator::serialize(std::string& json, unsigned int fieldCount, unsigned int minNumberFieldsPacking)
{
  return serialize(jsonType, json, fieldCount, minNumberFieldsPacking);
}
//-----------------------------------------------------------------------------
bool SphinxResultDataSerializator::serialize(std::string& json)
{
  return serialize(json, 0, 0);
}
//-----------------------------------------------------------------------------
bool SphinxResultDataSerializator::unserialize(const std::string& json)
{
  resultData.clear();
  errorMsg.clear();
  errorCode = NO_ERROR;
  SphinxWeightPacker packer;
  try
  {
    Poco::JSON::Parser parser;
    Poco::Dynamic::Var res = parser.parse(json);
    Poco::JSON::Object::Ptr obj = res.extract<Poco::JSON::Object::Ptr>();
    Poco::Dynamic::Var tmp;

    // extract from json objects of matches info
    Poco::JSON::Array::Ptr pMatches = obj->getArray(search_result_json_const::matchInfo);
    if (!pMatches.isNull())
    {
      const size_t matchesCount = pMatches->size();
      for (size_t i=0;i<matchesCount;++i)
      {
        Poco::JSON::Object::Ptr pMatchElem = pMatches->getObject(i);
        if (!pMatchElem.isNull())
        {
          SphinxMatchInfo matchInfo;
          tmp = pMatchElem->get(search_result_json_const::docId);
          if (tmp.isString())
            matchInfo.setDocId(std::stoull(tmp.convert<std::string>()));

          tmp = pMatchElem->get(search_result_json_const::weight);
          if (tmp.isString())
            matchInfo.setWeightFromHexString(packer.unpack(tmp.convert<std::string>()));

          Poco::JSON::Array::Ptr pAttributes = pMatchElem->getArray(search_result_json_const::attrArray);
          if (!pAttributes.isNull())
          {
            const size_t attributesCount = pAttributes->size();
            for (size_t j=0;j<attributesCount;++j)
            {
              Poco::JSON::Object::Ptr pAttrElem = pAttributes->getObject(j);
              if (!pAttrElem.isNull())
              {
                std::vector<std::string> attrNames;
                pAttrElem->getNames(attrNames);
                const size_t attrNamesCount = attrNames.size();
                for (size_t k=0;k<attrNamesCount;++k)
                {
                  tmp = pAttrElem->get(attrNames[k]);
                  if (tmp.isString())
                    matchInfo.addAttrInfo(attrNames[k], tmp.convert<std::string>());
                }
              }
            }
          }
          resultData.addMatchInfo(std::forward<SphinxMatchInfo>(matchInfo));
        }
      }
    }
    // extract from json objects of request info
    Poco::JSON::Array::Ptr pRequests = obj->getArray(search_result_json_const::requestInfo);
    if (!pRequests.isNull())
    {
      const size_t requestsCount = pRequests->size();
      for (size_t i=0;i<requestsCount;++i)
      {
        SphinxRequestInfo requestInfo;
        Poco::JSON::Object::Ptr pRequestElem = pRequests->getObject(i);

        tmp = pRequestElem->get(search_result_json_const::nodeName);
        if (tmp.isString())
          requestInfo.setNodeName(tmp.convert<std::string>());

        tmp = pRequestElem->get(search_result_json_const::query);
        if (tmp.isString())
          requestInfo.setQuery(tmp.convert<std::string>());

        tmp = pRequestElem->get(search_result_json_const::queryId);
        requestInfo.setQueryId(convertVarToNumeric<unsigned int>(tmp, 0));

        tmp = pRequestElem->get(search_result_json_const::maxResultsNumber);
        requestInfo.setMaxResultsNumber(convertVarToNumeric<unsigned int>(tmp, input_json_message_const::defaultMaxResultsNumber));

        tmp = pRequestElem->get(search_result_json_const::orderBy);
        requestInfo.setOrderBy(convertVarToNumeric<unsigned int>(tmp, 0));

        tmp = pRequestElem->get(search_result_json_const::total);
        requestInfo.setTotal(convertVarToNumeric<unsigned int>(tmp, 0));

        tmp = pRequestElem->get(search_result_json_const::totalFound);
        requestInfo.setTotalFound(convertVarToNumeric<unsigned int>(tmp, 0));

        tmp = pRequestElem->get(search_result_json_const::timeMsec);
        requestInfo.setTimeMsec(convertVarToNumeric<unsigned int>(tmp, 0));

        Poco::JSON::Array::Ptr pWords = pRequestElem->getArray(search_result_json_const::wordArray);
        if (!pWords.isNull())
        {
          const size_t wordsCount = pWords->size();
          for (size_t j=0;j<wordsCount;++j)
          {
            Poco::JSON::Object::Ptr pWordElem = pWords->getObject(j);
            if (!pWordElem.isNull())
            {
              WordInfo wordInfo;
              tmp = pWordElem->get(search_result_json_const::word);
              if (tmp.isString())
                tmp.convert(wordInfo.word);

              tmp = pWordElem->get(search_result_json_const::hits);
              wordInfo.hits = convertVarToNumeric<unsigned int>(tmp, 0);

              tmp = pWordElem->get(search_result_json_const::docs);
              wordInfo.docs = convertVarToNumeric<unsigned int>(tmp, 0);

              requestInfo.addWordInfo(std::forward<WordInfo>(wordInfo));
            }
          }
        }
        resultData.addRequestInfo(std::forward<SphinxRequestInfo>(requestInfo));
      }
    }
    _isError = false;
  }
  catch(Poco::JSON::JSONException& e)
  {
    _isError = true;
    errorMsg = e.displayText();
    errorCode = ERROR_UNSERIALIZE;
  }
  catch(std::exception& e)
  {
    _isError = true;
    errorMsg = e.what();
    errorCode = ERROR_UNSERIALIZE;
  }
  return !_isError;
}
//-----------------------------------------------------------------------------
//-----------------------------------------------------------------------------
SphinxDefaultJSON::SphinxDefaultJSON(unsigned int matchesCount, unsigned int attributesCount, unsigned int maxResultNumber)
: json(""), errorMsg(""), errorCode(NO_ERROR), _isError(false)
{
  makeJSON(matchesCount, attributesCount, maxResultNumber);
}
//-----------------------------------------------------------------------------
bool SphinxDefaultJSON::fillResultData(SphinxResultData& resultData, unsigned int matchesCount, unsigned int attributesCount, unsigned int maxResultNumber)
{
  try
  {
    timeval now;
    gettimeofday(&now, 0);

    const size_t maxMatchesCount = now.tv_usec + matchesCount;

    for (size_t i=now.tv_usec;i<maxMatchesCount;++i)
    {
      SphinxMatchInfo matchInfo;
      matchInfo.setDocId(i);
      matchInfo.setWeightFromNumeric(i);

      for (size_t j=1;j<=attributesCount;++j)
        matchInfo.addAttrInfo("attr_"+std::to_string(j), "value_"+std::to_string(j));

      resultData.addMatchInfo(std::forward<SphinxMatchInfo>(matchInfo));
    }
    SphinxRequestInfo requestInfo;
    requestInfo.setNodeName("node");
    requestInfo.setQuery("query");
    requestInfo.setQueryId(matchesCount);
    requestInfo.setMaxResultsNumber(maxResultNumber);
    requestInfo.setOrderBy(attributesCount);
    requestInfo.setTotal(matchesCount);
    requestInfo.setTotalFound(matchesCount);
    requestInfo.addWordInfo("word1", 1 , 2);
    requestInfo.addWordInfo("word2", 2 , 3);
    requestInfo.addWordInfo("word3", 3 , 4);
    resultData.addRequestInfo(std::forward<SphinxRequestInfo>(requestInfo));

    _isError = false;
  }
  catch(std::exception& e)
  {
    _isError = true;
    errorMsg = e.what();
    errorCode = ERROR_MAKE_DEFAULT_JSON;
  }
  return !_isError;
}
//-----------------------------------------------------------------------------
bool SphinxDefaultJSON::makeJSON(unsigned int matchesCount, unsigned int attributesCount, unsigned int maxResultNumber)
{
  if (!maxResultNumber)
    maxResultNumber = input_json_message_const::defaultMaxResultsNumber;

  json.clear();
  errorMsg.clear();
  errorCode = NO_ERROR;
  _isError = false;
  SphinxResultData resultData;
  SphinxResultDataSerializator resultSerializator(resultData);

  bool success = true;
  if (!matchesCount)
    success = resultSerializator.serialize(NO_FILL, json);
  else
  {
    if (fillResultData(resultData, matchesCount, attributesCount, maxResultNumber))
      success = resultSerializator.serialize((FILL_MATCH_INFO | FILL_REQUEST_INFO | FILL_ATTRIBUTES | FILL_WORD_INFO), json);
  }

  if (!success)
  {
    _isError = resultSerializator.isError();
    errorMsg = resultSerializator.getErrorMsg();
    errorCode = resultSerializator.getErrorCode();
  }
  return !_isError;
}
//-----------------------------------------------------------------------------
std::ostream& operator<<(std::ostream& os, const SphinxDefaultJSON& defaultJson)
{
  return os << defaultJson.getJSON();
}
//-----------------------------------------------------------------------------
//-----------------------------------------------------------------------------
} // namespace sphinx 
} // namespace HCE 
