#include <string.h>
#include <stdio.h>
#include <sstream>
#include <iterator>
#include <time.h>
#include <sys/time.h>
#include <unistd.h>
#include <Poco/AutoPtr.h>

#include "SphinxSearcher.hpp"
#include "SphinxInputJsonMessage.hpp"
#include "SphinxOutputJsonMessage.hpp"
#include "SphinxError.hpp"
#include "ProcExec.hpp"
#include "SphinxAdminCommand.hpp"
#include "SphinxConfigCreator.hpp"
#include "SphinxWeightCalculator.hpp"
#include "SphinxInputJsonMessageSearch.hpp"
#include "SphinxMessageConst.hpp"

namespace HCE
{
namespace sphinx
{
//-------------------------------------------------------------------
SphinxSearcher::SphinxSearcher(SphinxFunctionalObject& fObj_, bool startSearchd_, bool stopSearchd_)
:SphinxSearchInterface(),
 client(nullptr), field_names(nullptr), field_weights(nullptr), fieldsCount(0), index_names(nullptr), index_weights(nullptr),
 indexesCount(0), fObj(fObj_), message(fObj.getCustomMessage()),
 serverHost(sphinx_search_const::defaultHost), serverPort(sphinx_search_const::defaultPort), indexFileName(""),
 _StopSearchd(stopSearchd_), _isConnected(false), _isStoppedSearchd(true), timeoutedRequests(0), defaultRequestTimeout(input_json_message_const::defaultTimeoutValue),
 allowedFields(), fullSchemaNames(), requestInfo(), resultData(), resultSerializator(resultData), resultMemoryManager(resultData)
{
	client = sphinx_create(SPH_TRUE);
	if (!client)
		errorMsg = message(message_const::FAILED_TO_CREATE_CLIENT);
}
//-------------------------------------------------------------------
SphinxSearcher::~SphinxSearcher(void)
{
  if (_StopSearchd)
    stopSearchd();
	sphinx_close(client);
	sphinx_destroy(client);	
	cleanup(field_names, field_weights, fieldsCount);
	cleanup(index_names, index_weights, indexesCount);
}
//-------------------------------------------------------------------
void SphinxSearcher::resizeResultData(size_t matchesCount, size_t attributesCount)
{
  resultMemoryManager.resize(matchesCount, attributesCount);
}
//-------------------------------------------------------------------
void SphinxSearcher::setAllowedFields(std::vector<std::string>& queryExtendedFields)
{
  allowedFields.clear();
  for (size_t i=0;i<queryExtendedFields.size();++i)
    allowedFields.insert(queryExtendedFields[i]);
}
//-------------------------------------------------------------------
bool SphinxSearcher::isAllowedFields(const std::string& extendedFieldName)
{
  return (allowedFields.find(extendedFieldName)!=allowedFields.end());
}
//-------------------------------------------------------------------
void SphinxSearcher::setFullSchemaNames(std::vector<std::string>& schemaNames)
{
  fullSchemaNames.clear();
  for (size_t i=0;i<schemaNames.size();++i)
    fullSchemaNames.insert(schemaNames[i]);
}
//-------------------------------------------------------------------
bool SphinxSearcher::isExistSchemaNames(const std::string& schemaName)
{
  return (fullSchemaNames.find(schemaName)!=fullSchemaNames.end());
}
//-------------------------------------------------------------------
bool SphinxSearcher::startSearchd(void)
{
  _isError = false;
  std::string configFile = fObj.getDataDir()+"/"+fObj.getIndexName()+"/"+sphinx_search_const::sphinxConfigName;

  fObj.log(LoggerStream::Priority::PRIO_INFORMATION) << message(message_const::START_SEARCHD_READ_CONFIG, configFile) << flush;

  try
  {
    stopSearchd();

    int count = sphinx_search_const::startStopSearchdLimit;
    fObj.log(LoggerStream::Priority::PRIO_INFORMATION) << message(message_const::WAIT_START_SEARCHD) << flush;

    while (!isActiveSearchd() && --count>0)
    {
      std::stringstream outMsg, errMsg;
      const std::string cmd("searchd");
      std::vector<std::string> args = {"--config", configFile};
      ProcExec procExec(cmd, args);
      procExec.exec(outMsg, errMsg);
      if (procExec.isError())
        throw Poco::Exception(message(message_const::START_SEARCHD_FAIL, errMsg.str()));

      if ((outMsg.str().find(sphinx_admin_command_const::procExecErrorMsg)!=std::string::npos) ||
          (outMsg.str().find(sphinx_admin_command_const::procExecFatalMsg)!=std::string::npos))
        throw Poco::Exception(message(message_const::START_SEARCHD_FAIL, outMsg.str()));

      if ((errMsg.str().find(sphinx_admin_command_const::procExecErrorMsg)!=std::string::npos) ||
          (errMsg.str().find(sphinx_admin_command_const::procExecFatalMsg)!=std::string::npos))
        throw Poco::Exception(message(message_const::START_SEARCHD_FAIL, errMsg.str()));

      sleep(1);
      fObj.log(LoggerStream::Priority::PRIO_INFORMATION) << ".";
      fObj.log(LoggerStream::Priority::PRIO_INFORMATION).flush();
      fObj.log(LoggerStream::Priority::PRIO_INFORMATION) << flush;
    }
    fObj.log(LoggerStream::Priority::PRIO_INFORMATION) << message(message_const::TRY_EASY_START) << flush;
    if (!easyStart())
    {
      errorMsg.clear();
      errorCode = NO_ERROR;
      sleep(1);
      if (!easyStart())
        throw Poco::Exception(errorMsg);
    }

    _isError = false;
    _isStoppedSearchd = false;
  }
  catch(Poco::Exception& e)
  {
    _isError = true;
    errorMsg = e.message();
    errorCode = ERROR_START_SEARCHD;

    if (!easyStart())
    {
      errorMsg.clear();
      errorCode = NO_ERROR;
      sleep(1);
      if (!easyStart())
        errorMsg.append("\n").append(e.message());
    }
    fObj.log(LoggerStream::Priority::PRIO_INFORMATION) << errorMsg << flush;
  }
  return !_isError;
}
//-------------------------------------------------------------------
bool SphinxSearcher::stopSearchd(void)
{
  _isError = false;
  std::string configFile = fObj.getDataDir()+"/"+fObj.getIndexName()+"/"+sphinx_search_const::sphinxConfigName;
  try
  {
    if (!close())
      fObj.resetError();

    if (!isStoppedSearchd())
    {
      std::stringstream outMsg, errMsg;
      const std::string cmd("searchd");
      std::vector<std::string> args = {"--stop", "--config", configFile};
      ProcExec procExec(cmd, args);
      if (!procExec.exec(outMsg, errMsg))
        throw Poco::Exception(message(message_const::STOP_SEARCHD, errMsg.str()));

      _isStoppedSearchd = true;
    }
    _isError = false;
  }
  catch(Poco::Exception& e)
  {
    _isError = true;
    errorMsg = e.message();
    errorCode = ERROR_STOP_SEARCHD;
  }
  return !_isError;
}
//-------------------------------------------------------------------
bool SphinxSearcher::isActiveSearchd()
{
  bool result = false;
  IndexStatusSearchd indexStatusSearchd(fObj);
  indexStatusSearchd.setIndexName(fObj.getIndexName());
  std::string outData;
  if (indexStatusSearchd.execute())
  {
    outData = indexStatusSearchd.getResultData();
    if (outData.find(sphinx_search_const::searchdGoodStatusFlag)!=std::string::npos)
      result = true;
  }
  // indexStatus.isActiveSearchd(); // not used !!! method search PID-file on HDD
  _isStoppedSearchd = !result;
  return result;
}
//-------------------------------------------------------------------
void SphinxSearcher::setMaxResultsNumber(unsigned int maxResultsNumber) throw (Poco::Exception)
{
  if (!client)
    throw Poco::Exception(message(message_const::INVALID_CLIENT_HANDLE), ERROR_FUNCTIONAL_OBJECT);

  if (!sphinx_set_limits(client, 0, maxResultsNumber, maxResultsNumber, 0))
    throw Poco::Exception(sphinx_error(client), ERROR_SET_LIMITS);
}
//-------------------------------------------------------------------
void SphinxSearcher::setLimits(unsigned int offset, unsigned int limit, unsigned int maxMatches, unsigned int cutoff) throw (Poco::Exception)
{
  if (!client)
    throw Poco::Exception(message(message_const::INVALID_CLIENT_HANDLE), ERROR_FUNCTIONAL_OBJECT);

  if (!sphinx_set_limits(client, offset, limit, maxMatches, cutoff))
    throw Poco::Exception(sphinx_error(client), ERROR_SET_LIMITS);
}
//-------------------------------------------------------------------
void SphinxSearcher::applyFilter(Poco::SharedPtr<SphinxFilter> pFilter) throw (Poco::Exception)
{
  if (!client)
    throw Poco::Exception(message(message_const::INVALID_CLIENT_HANDLE), ERROR_FUNCTIONAL_OBJECT);

  if (pFilter.isNull())
    throw Poco::Exception(message(message_const::INVALID_FILTER), ERROR_ADD_FILTER);

  bool exclude = (pFilter->getExcludeValue()==SphinxFilter::ExcludeType::etTrue);

  try
  {
    if (pFilter->getFilterType()==SphinxFilter::FilterType::ftSimple)
    {
      std::vector<long long> vFilters;
      for (size_t i=0;i<pFilter->getAttributeValues().size();++i)
        vFilters.push_back(std::stoll(pFilter->getAttributeValues()[i]));

      if (!addFilter(pFilter->getAttributeName(), vFilters, exclude))
        throw Poco::Exception(message(message_const::ADD_FILTER, pFilter->getAttributeName())+message(message_const::ERROR, sphinx_error(client)), ERROR_ADD_FILTER);
    }
    else if(pFilter->getFilterType()==SphinxFilter::FilterType::ftRangeInteger)
    {
      if (pFilter->getAttributeValues().size() < 2)
        throw Poco::Exception(message(message_const::INVALID_RANGE_OF_FILTER), ERROR_ADD_FILTER);

      if (!addFilterRange(pFilter->getAttributeName(), std::stoull(pFilter->getAttributeValues()[0]), std::stoull(pFilter->getAttributeValues()[1]), exclude))
        throw Poco::Exception(message(message_const::ADD_FILTER, pFilter->getAttributeName())+message(message_const::ERROR, sphinx_error(client)), ERROR_ADD_FILTER);
    }
    else if(pFilter->getFilterType()==SphinxFilter::FilterType::ftRangeFloat)
    {
      if (pFilter->getAttributeValues().size() < 2)
        throw Poco::Exception(message(message_const::INVALID_RANGE_OF_FILTER), ERROR_ADD_FILTER);

      if (!addFilterFloatRange(pFilter->getAttributeName(), std::stof(pFilter->getAttributeValues()[0]), std::stof(pFilter->getAttributeValues()[1]), exclude))
        throw Poco::Exception(message(message_const::ADD_FILTER, pFilter->getAttributeName())+message(message_const::ERROR, sphinx_error(client)), ERROR_ADD_FILTER);
    }
    else if(pFilter->getFilterType()==SphinxFilter::FilterType::ftSimpleMulti)
    {
      for (size_t i=0;i<pFilter->getAttributeValues().size();++i)
      {
        if (!addFilter(pFilter->getAttributeName(), std::stoull(pFilter->getAttributeValues()[i]), exclude))
          throw Poco::Exception(message(message_const::ADD_FILTER, pFilter->getAttributeName())+message(message_const::ERROR, sphinx_error(client)), ERROR_ADD_FILTER);
      }
    }
    else
      throw Poco::Exception(message(message_const::UNKNOWN_FILTER_TYPE, std::to_string(static_cast<unsigned int>(pFilter->getFilterType()))), ERROR_ADD_FILTER);
  }
  catch(std::exception& e)
  {
    throw Poco::Exception(e.what(), ERROR_ADD_FILTER);
  }
}
//-------------------------------------------------------------------
bool SphinxSearcher::easyStart(int matchMode, int sortMode, int rankingMode, const std::string& rankexpr)
{
  errorMsg.clear();
  errorCode = NO_ERROR;
  try
  {
    if (!client)
      throw Poco::Exception(message(message_const::INVALID_CLIENT_HANDLE), ERROR_FUNCTIONAL_OBJECT);

    if (serverHost.empty())
      throw Poco::Exception(message(message_const::SERVER_HOST_IS_EMPTY), ERROR_FUNCTIONAL_OBJECT);

    if (serverPort==0)
      throw Poco::Exception(message(message_const::SERVER_PORT_IS_NULL), ERROR_FUNCTIONAL_OBJECT);

    if (!sphinx_set_server(client, serverHost.c_str(), serverPort))
      throw Poco::Exception(sphinx_error(client), ERROR_SET_SERVER);

    fObj.log(LoggerStream::Priority::PRIO_INFORMATION) << message(message_const::SET_SERVER_SUCCESS)
        << message(message_const::HOST, serverHost)
        << message(message_const::PORT, serverPort) << flush;

    if (setMatchMode(matchMode))
    {
      if (setRankingMode(rankingMode, rankexpr))
      {
        if (setSortMode(sortMode))
        {
          if (!sphinx_open(client))
            throw Poco::Exception(message(message_const::OPEN_CONNECTION_ERROR, sphinx_error(client)), ERROR_OPEN_CONNECTION);
          else
            _isConnected = true;
        }
      }
    }

    fObj.log(LoggerStream::Priority::PRIO_INFORMATION) << message(message_const::EASY_START_SUCCESS)
        << message(message_const::HOST, serverHost)
        << message(message_const::PORT, serverPort) << flush;

    _isError = false;
  }
  catch(Poco::Exception& e)
  {
    _isError = true;
    errorMsg = e.message();
    errorCode = e.code();
    fObj.log(LoggerStream::Priority::PRIO_INFORMATION) << message(message_const::EASY_START_FALLEN)
        << message(message_const::HOST, serverHost)
        << message(message_const::PORT, serverPort)
        << message(message_const::ERROR, errorMsg) << flush;
  }
  return !_isError;
}
//-------------------------------------------------------------------
bool SphinxSearcher::open(void)
{
  errorMsg.clear();
  errorCode = NO_ERROR;
	try
	{
	  if (!client)
	    throw Poco::Exception(message(message_const::INVALID_CLIENT_HANDLE), ERROR_FUNCTIONAL_OBJECT);

	  if (!sphinx_set_server(client, serverHost.c_str(), serverPort))
	    throw Poco::Exception(sphinx_error(client), ERROR_SET_SERVER);

	  fObj.log(LoggerStream::Priority::PRIO_INFORMATION) << message(message_const::SET_SERVER_SUCCESS)
        << message(message_const::HOST, serverHost)
        << message(message_const::PORT, serverPort) << flush;

	  if (!sphinx_open(client))
	    throw Poco::Exception(sphinx_error(client), ERROR_OPEN_CONNECTION);

	  _isConnected = true;
	  _isError = false;
	}
	catch(Poco::Exception& e)
	{
	  _isError = true;
	  errorMsg = e.displayText();
	  errorCode = e.code();
	}
	return !_isError;
}
//-------------------------------------------------------------------
bool SphinxSearcher::close(void)
{
  errorMsg.clear();
  errorCode = NO_ERROR;
  try
  {
    if (!client)
      throw Poco::Exception(message(message_const::INVALID_CLIENT_HANDLE), ERROR_FUNCTIONAL_OBJECT);

    if (!sphinx_close(client))
      throw Poco::Exception(sphinx_error(client), ERROR_CLOSE_CONNECTION);

     _isConnected = false;
    _isError = false;
  }
  catch(Poco::Exception& e)
  {
    _isError = true;
    errorMsg = e.displayText();
    errorCode = e.code();
  }
  return !_isError;
}
//-------------------------------------------------------------------
void SphinxSearcher::cleanup(char**& names, int*& weights, size_t& count)
{
	if (names)
	{
		for (size_t i=0;i<count;++i)
			if (names[i])
				delete [] names[i];
		delete [] names;
		names = NULL;
	}

	if (weights)
	{
		delete [] weights;
		weights = NULL;
	}
	count = 0;
	errorMsg.clear();
}
//-------------------------------------------------------------------
bool SphinxSearcher::setWeights(char**& names, int*& weights, size_t& count, std::vector <std::pair<std::string, int> >& Vec)
{
	bool result = false;
	if (client)
	{
		// free if need memory
		cleanup(names, weights, count);
		// allocate memory
		count = Vec.size();
		names = new char*[count];
		weights = new int[count];
		// copy data
		for (size_t i=0;i<count;++i)
		{
			names[i] = new char[Vec[i].first.length()+1];
			memset(names[i], 0, Vec[i].first.length()+1);
			memcpy(names[i], Vec[i].first.c_str(), Vec[i].first.length());		
			weights[i] = Vec[i].second;
		}
		result = true;
	}
	return result;
}
//-------------------------------------------------------------------
bool SphinxSearcher::setFieldWeights(std::vector <std::pair<std::string, int> >& Vec)
{
	bool result = false;

	if (setWeights(field_names, field_weights, fieldsCount, Vec))
		result = sphinx_set_field_weights(client, fieldsCount, (const char**)field_names, field_weights);

	if (!result)
		errorMsg = sphinx_error(client);

	return result;
}
//-------------------------------------------------------------------
bool SphinxSearcher::setIndexWeights(std::vector <std::pair<std::string, int> >& Vec)
{
	bool result = false;

	if (setWeights(index_names, index_weights, indexesCount, Vec))
		result = sphinx_set_index_weights(client, indexesCount, (const char**)index_names, index_weights);

	if (!result)
		errorMsg = sphinx_error(client);

	if (Vec.size() && indexFileName.empty())
		result = setIndexFileName(Vec.begin()->first);

	return result;
}
//-------------------------------------------------------------------
void SphinxSearcher::getWeights(char**& names, int*& weights, size_t& count, std::vector <std::pair<std::string, int> >& Vec)
{
	Vec.clear();
	for (size_t i=0;i<count;++i)
	{
		if (names[i]!=nullptr)
			Vec.push_back(std::make_pair <std::string, int> (static_cast<std::string>(names[i]), static_cast<int>(weights[i])));
	}
}
//-------------------------------------------------------------------
void SphinxSearcher::getFieldWeights(std::vector <std::pair<std::string, int> >& Vec)
{
	getWeights(field_names, field_weights, fieldsCount, Vec);
}
//-------------------------------------------------------------------
void SphinxSearcher::getIndexWeights(std::vector <std::pair<std::string, int> >& Vec)
{
	getWeights(index_names, index_weights, indexesCount, Vec);
}
//-------------------------------------------------------------------
std::vector <std::pair<std::string, int> >::iterator SphinxSearcher::find(std::vector <std::pair<std::string, int> >& Vec, const std::string& name)
{
	std::vector <std::pair<std::string, int> >::iterator iter;
	for (iter=Vec.begin();iter!=Vec.end();++iter)
	{
		if ((*iter).first == name)
			break;
	}
	return iter;
}
//-------------------------------------------------------------------
// cppcheck-suppress unusedFunction
bool SphinxSearcher::isExistField(const std::string& name)
{
	std::vector <std::pair<std::string, int> > Vec;
	this->getFieldWeights(Vec);
	return (this->find(Vec, name) != Vec.end());
}
//-------------------------------------------------------------------
// cppcheck-suppress unusedFunction
bool SphinxSearcher::isExistIndexFile(const std::string& name)
{
	std::vector <std::pair<std::string, int> > Vec;
	this->getIndexWeights(Vec);
	return (this->find(Vec, name) != Vec.end());
}
//-------------------------------------------------------------------
// cppcheck-suppress unusedFunction
int SphinxSearcher::getFieldWeight(const std::string& name)
{
	std::vector <std::pair<std::string, int> > Vec;
	this->getFieldWeights(Vec);
	std::vector <std::pair<std::string, int> >::iterator iter = this->find(Vec, name);
	return (iter!=Vec.end())?(*iter).second:0;
}
//-------------------------------------------------------------------
// cppcheck-suppress unusedFunction
int SphinxSearcher::getIndexWeight(const std::string& name)
{
	std::vector <std::pair<std::string, int> > Vec;
	this->getIndexWeights(Vec);
	std::vector <std::pair<std::string, int> >::iterator iter = this->find(Vec, name);
	return (iter!=Vec.end())?(*iter).second:0;
}
//-------------------------------------------------------------------
// cppcheck-suppress unusedFunction
bool SphinxSearcher::addField(const std::string& name, int weight)
{
	errorMsg.clear();
	bool result = false;

	std::vector <std::pair<std::string, int> > Vec;
	getFieldWeights(Vec);

	std::vector <std::pair<std::string, int> >::iterator iter = this->find(Vec, name);
	if (iter==Vec.end())
	{
		Vec.push_back(std::make_pair (name, weight));
		result = setFieldWeights(Vec);
	}
	else
		errorMsg = message(message_const::FIELD_ALREADY_EXIST, name);

	return result;
}
//-------------------------------------------------------------------
// cppcheck-suppress unusedFunction
bool SphinxSearcher::setField(const std::string& name, int weight)
{
	errorMsg.clear();
	bool result = false;

	std::vector <std::pair<std::string, int> > Vec;
	getFieldWeights(Vec);

	std::vector <std::pair<std::string, int> >::iterator iter = this->find(Vec, name);
	if (iter!=Vec.end())
	{
		(*iter).second = weight;
		result = setFieldWeights(Vec);
	}
	else
		errorMsg = message(message_const::FIELD_NOT_FOUND, name);

	return result;
}
//-------------------------------------------------------------------
// cppcheck-suppress unusedFunction
bool SphinxSearcher::delField(const std::string& name)
{
	errorMsg.clear();
	bool result = false;

	std::vector <std::pair<std::string, int> > Vec;
	getFieldWeights(Vec);

	std::vector <std::pair<std::string, int> >::iterator iter = this->find(Vec, name);
	if (iter!=Vec.end())
	{
		Vec.erase(iter);
		result = setFieldWeights(Vec);
	}
	else
		errorMsg = message(message_const::FIELD_NOT_FOUND, name);

	return result;
}
//-------------------------------------------------------------------
// cppcheck-suppress unusedFunction
bool SphinxSearcher::addIndexFile(const std::string& name, int weight)
{
	errorMsg.clear();
	bool result = false;

	std::vector <std::pair<std::string, int> > Vec;
	getIndexWeights(Vec);

	std::vector <std::pair<std::string, int> >::iterator iter = this->find(Vec, name);
	if (iter==Vec.end())
	{
		Vec.push_back(std::make_pair (name, weight));
		result = setIndexWeights(Vec);
	}
	else
		errorMsg =message(message_const::INDEX_ALREADY_EXIST, name);

	return result;
}
//-------------------------------------------------------------------
// cppcheck-suppress unusedFunction
bool SphinxSearcher::setIndexFile(const std::string& name, int weight)
{
	errorMsg.clear();
	bool result = false;

	std::vector <std::pair<std::string, int> > Vec;
	getIndexWeights(Vec);

	std::vector <std::pair<std::string, int> >::iterator iter = this->find(Vec, name);
	if (iter!=Vec.end())
	{
		(*iter).second = weight;
		result = setIndexWeights(Vec);
	}
	else
		errorMsg = message(message_const::INDEX_NOT_FOUND, name);

	return result;
}
//-------------------------------------------------------------------
// cppcheck-suppress unusedFunction
bool SphinxSearcher::delIndexFile(const std::string& name)
{
	errorMsg.clear();
	bool result = false;

	std::vector <std::pair<std::string, int> > Vec;
	getIndexWeights(Vec);

	std::vector <std::pair<std::string, int> >::iterator iter = this->find(Vec, name);
	if (iter!=Vec.end())
	{
		Vec.erase(iter);
		result = setIndexWeights(Vec);
	}
	else
		errorMsg = message(message_const::INDEX_NOT_FOUND, name);

	return result;
}
//-------------------------------------------------------------------
bool SphinxSearcher::setMatchMode(int mode)
{
  bool result = false;
  try
  {
    if (!client)
      throw Poco::Exception(message(message_const::INVALID_CLIENT_HANDLE), ERROR_FUNCTIONAL_OBJECT);

    if (!sphinx_set_match_mode(client, mode))
      throw Poco::Exception(sphinx_error(client), ERROR_SET_MATCH_MODE);

    result = true;
	}
  catch(Poco::Exception& e)
  {
    errorMsg = e.displayText();
    errorCode = e.code();
  }
	return result;
}
//-------------------------------------------------------------------
bool SphinxSearcher::setSortMode(int mode, const std::string& sortby)
{
  bool result = false;
  try
  {
    if (!client)
      throw Poco::Exception(message(message_const::INVALID_CLIENT_HANDLE), ERROR_FUNCTIONAL_OBJECT);

    if (sortby.empty())
    {
      if (!sphinx_set_sort_mode(client, SPH_SORT_RELEVANCE, nullptr))
        throw Poco::Exception(sphinx_error(client), ERROR_SET_SORT_MODE);
    }
    else
    {
      if (!sphinx_set_sort_mode(client, mode, sortby.c_str()))
        throw Poco::Exception(sphinx_error(client), ERROR_SET_SORT_MODE);
    }
    result = true;
  }
  catch(Poco::Exception& e)
  {
    errorMsg = e.displayText();
    errorCode = e.code();
  }
  return result;
}
//-------------------------------------------------------------------
bool SphinxSearcher::setRankingMode(int ranker, const std::string& rankexpr)
{
  bool result = false;
  try
  {
    if (!client)
      throw Poco::Exception(message(message_const::INVALID_CLIENT_HANDLE), ERROR_FUNCTIONAL_OBJECT);

    if (rankexpr.empty())
    {
      if (!sphinx_set_ranking_mode(client, ranker, nullptr))
        throw Poco::Exception(sphinx_error(client), ERROR_SET_RANKING_MODE);
    }
    else
    {
      if (!sphinx_set_ranking_mode(client, ranker, rankexpr.c_str()))
        throw Poco::Exception(sphinx_error(client), ERROR_SET_RANKING_MODE);
    }
    result = true;
  }
  catch(Poco::Exception& e)
  {
    errorMsg = e.displayText();
    errorCode = e.code();
  }
  return result;
}
//-------------------------------------------------------------------
void SphinxSearcher::setMaxQueryTime(unsigned long maxQueryTime) throw (Poco::Exception)
{
  if (!client)
    throw Poco::Exception(message(message_const::INVALID_CLIENT_HANDLE), ERROR_FUNCTIONAL_OBJECT);

  if (!sphinx_set_max_query_time(client, maxQueryTime))
    throw Poco::Exception(sphinx_error(client), ERROR_SET_QUERY_TIME);
}
//-------------------------------------------------------------------
bool SphinxSearcher::setIndexFileName(const std::string& indexFileName_)
{
	bool result = false;
	for (size_t i=0;i<indexesCount;++i)
	{
		if (indexFileName_ == index_names[i])
		{
		  indexFileName = indexFileName_;
			result = true;
			break;
		}
	}
	if (!result)
	{
		if (indexesCount)
			errorMsg = message(message_const::INDEX_NAME, indexFileName)+message(message_const::NOT_FOUND_IN_CONFIG);
		else
			errorMsg = message(message_const::LIST_OF_INDEXES_IS_EMPTY);
	}
	return result;
}
//-------------------------------------------------------------------
bool SphinxSearcher::searchDocument(const std::string& query, SphinxResultDataMemoryManager& resultMemoryManager, SphinxRequestInfo& request)
{
	bool result = false;
	if (client)
	{
		sphinx_result* res = 0;
		int i, j, k, mva_len;
		unsigned int* mva;
		result = true;

		for (size_t z=0;z<indexesCount;++z)
		{
		  if (!indexFileName.empty() && indexFileName!=index_names[z])
		    continue;

      const char* indexName = index_names[z];
      res = sphinx_query(client, query.c_str(), indexName, nullptr);

      if (!res)
      {
        errorMsg = sphinx_error(client);
        fObj.log(LoggerStream::Priority::PRIO_INFORMATION) << message(message_const::SEARCHD_ERROR, errorMsg) << flush;
        result = false;
      }
      else
      {
        std::string warningMsg(sphinx_warning(client));
        if (!warningMsg.empty())
        {
          fObj.log(LoggerStream::Priority::PRIO_INFORMATION) << message(message_const::SEARCHD_WARNING, warningMsg) << flush;
          ++timeoutedRequests;
        }

        request.setQuery(query);
        request.setTotal(res->total);
        request.setTotalFound(res->total_found);
        request.setTimeMsec(res->time_msec);

        for (i=0; i<res->num_words; ++i)
          request.addWordInfo(WordInfo(res->words[i].word, res->words[i].hits, res->words[i].docs));

        for (i=0; i<res->num_matches; ++i)
        {
          resultMemoryManager.resetMatchInfo();
          const unsigned long long sphinxDocumentId = sphinx_get_id(res, i);
          resultMemoryManager.setMatchInfoDocId(sphinxDocumentId);
          const std::string sphinxWeight = std::to_string(sphinx_get_weight(res, i));
          resultMemoryManager.setMatchInfoWeight(sphinxWeight);
          for (j=0; j<res->num_attrs; ++j)
          {
            if (allowedFields.empty() || isAllowedFields(res->attr_names[j]))
            {
              std::stringstream valueStr;

              switch(res->attr_types[j])
              {
                case SPH_ATTR_MULTI64:
                case SPH_ATTR_MULTI:
                  mva = sphinx_get_mva(res, i, j);
                  mva_len = *mva++;
                  for (k=0; k<mva_len; ++k)
                    valueStr << (res->attr_types[j]==SPH_ATTR_MULTI ? mva[k] : (unsigned int)sphinx_get_mva64_value(mva, k));
                  break;
                case SPH_ATTR_FLOAT:	valueStr <<	sphinx_get_float(res, i, j);
                  break;
                case SPH_ATTR_STRING:	valueStr << sphinx_get_string(res, i, j);
                  break;
                default: valueStr << (sphinx_int64_t)sphinx_get_int(res, i, j);
                  break;
              }
              if (!resultMemoryManager.addMatchInfoAttr(AttrInfo(res->attr_names[j], valueStr.str())))
                break;
            }
          }
          resultMemoryManager.addMatchInfoAttr(AttrInfo(search_result_json_const::sphinxWeight, sphinxWeight)); // save weight as attribute
          resultMemoryManager.addMatchInfoAttr(AttrInfo(search_result_json_const::sphinxDocumentId, std::to_string(sphinxDocumentId))); // save document ID as attribute
          if (!resultMemoryManager.applyMatchInfo())
            break;
        }
      }
		}
	}
	if (!result)
	{
	  bool activeSearchd = isActiveSearchd();
	  if (!activeSearchd)
	  {
	    IndexStatusSearchd indexStatusSearchd(fObj);
	    indexStatusSearchd.setIndexName(fObj.getIndexName());
	    bool existsPidFile = indexStatusSearchd.isExistPidFile();

	    fObj.log(LoggerStream::Priority::PRIO_INFORMATION) << message(message_const::SEARCHD_STOPPED_PID_FILE_EXIST, ((existsPidFile)?"true":"false")) << flush;

	    if (existsPidFile)
	    {
	      errorCode = ERROR_SEARCHD_STOPED_PID_FILE_EXISTS;
	      errorMsg = message(message_const::SEARCHD_STOPPED_BUT_PID_FILE_EXIST, errorMsg);
	    }
	    else
	    {
	      errorCode = ERROR_NOT_ACTIVITY_SEARCHD;
	      errorMsg = message(message_const::SEARCHD_NOT_ACTIVITY, errorMsg);
	    }
	  }
	  else
	    errorCode = ERROR_SEARCH_REQUEST;
	}
	_isError = !result;
	return result;
}
//-------------------------------------------------------------------
bool SphinxSearcher::addFilter(const std::string& attrName, long long value, bool exclude)
{
	return sphinx_add_filter(client, attrName.c_str(), 1, &value, exclude);
}
//-------------------------------------------------------------------
bool SphinxSearcher::addFilter(const std::string& attrName, std::vector<long long>& values, bool exclude)
{
  bool result = false;
  long long* pVal = new long long[values.size()];
  if (pVal)
  {
    for (size_t i=0;i<values.size();++i)
      pVal[i] = values[i];
    result = sphinx_add_filter(client, attrName.c_str(), values.size(), pVal, exclude);
    delete [] pVal;
  }
  return result;
}
//-------------------------------------------------------------------
bool SphinxSearcher::addFilterRange(const std::string& attrName, long long umin, long long umax, bool exclude)
{
	return sphinx_add_filter_range(client, attrName.c_str(), umin, umax, exclude);
}
//-------------------------------------------------------------------
bool SphinxSearcher::addFilterFloatRange(const std::string& attrName, float fmin, float fmax, bool exclude)
{
	return sphinx_add_filter_float_range(client, attrName.c_str(), fmin, fmax, exclude);
}
//-------------------------------------------------------------------
void SphinxSearcher::resetFilters(void)
{
	sphinx_reset_filters(client);
}
//-------------------------------------------------------------------
std::string SphinxSearcher::Process(const std::string& json)
{
  _isError = false;
  errorMsg.clear();
  errorCode = NO_ERROR;
  std::string resultData;
  timeval tvStart, tvStop;
  gettimeofday(&tvStart, 0); //gettimeofday does not support TZ adjust on Linux.

  try
  {
    SphinxInputJsonMessage inputMessage(json);
    if (inputMessage.isError())
      throw Poco::Exception(inputMessage.getErrorMsg(), PARSE_ERROR);

    if (inputMessage.getType() == SphinxInputJsonMessage::MessageType::mtSearch)
    {
      resultData = makeSearchCommand(inputMessage.getData());
      if (resultData.empty())
      {
        SphinxDefaultJSON defaultJson;
        resultData = defaultJson.getJSON();
      }
    }
    else
      throw Poco::Exception(message(message_const::INCORRECT_JSON_TYPE), COMMAND_ERROR);
  }
  catch(Poco::Exception& e)
  {
    _isError = true;
    errorCode = e.code();
    errorMsg = e.displayText();

    SphinxDefaultJSON defaultJson;
    resultData = defaultJson.getJSON();
  }

  std::string resJson;
  gettimeofday(&tvStop, 0);

  SphinxOutputJsonMessage outputMessage;
  outputMessage.setErrorCode(errorCode);
  outputMessage.setErrorMessage(errorMsg);
  outputMessage.setData(resultData);
  outputMessage.setTime(getTimeInterval(tvStart, tvStop)/1000);
  bool success = outputMessage.serialize(resJson);
  if (!_isError)
    _isError = !success;

  return resJson;
}
//-------------------------------------------------------------------
std::string SphinxSearcher::makeSearchCommand(const std::string& json) throw (Poco::Exception, std::exception)
{
  resultMemoryManager.reset();
  requestInfo.clear();
  resultData.clearRequestInfo();

  _isError = false;
  errorMsg.clear();
  errorCode = NO_ERROR;

  SphinxInputJsonMessageSearch messageSearch(json);
  if (messageSearch.isError())
    throw Poco::Exception(messageSearch.getErrorMsg(), ERROR_FUNCTIONAL_OBJECT);

  unsigned int offsetResults = 0;
  unsigned int limitResults = 0;
  unsigned int cutoffResults = 0;
  unsigned int orderBy = 0;
  unsigned int maxResultsNumber = input_json_message_const::defaultMaxResultsNumber;
  unsigned long long timeoutValue = defaultRequestTimeout;

  setAllowedFields(messageSearch.getExternalFields());  // set list of external fields need get in search request
  std::vector<std::string>& orderFieldsNames = messageSearch.getOrderFields(); // set list of external fields use in calculate weight
  std::string weightAlgirithm(""); // type of calculate weight algorithm in string

  try
  {
    // read query parameters
    for (size_t k=0;k<messageSearch.getQueryParameters().size();++k)
    {
      try
      {
        if (insensitiveCompare(messageSearch.getQueryParameters()[k].first, input_json_message_const::queryId))
          requestInfo.setQueryId(std::stoul(messageSearch.getQueryParameters()[k].second, nullptr, 0));

        if (insensitiveCompare(messageSearch.getQueryParameters()[k].first, input_json_message_const::JsonType))
          resultSerializator.setJsonType(std::stoul(messageSearch.getQueryParameters()[k].second, nullptr, 0));

        std::string sortFieldName;
        if (insensitiveCompare(messageSearch.getQueryParameters()[k].first, input_json_message_const::sortFieldName))
        {
          sortFieldName = messageSearch.getQueryParameters()[k].second;
          if (!sortFieldName.empty())
            if (!isExistSchemaNames(sortFieldName))
              throw Poco::Exception(message(message_const::FIELD_NOT_FOUND_IN_CURRENT_SCHEMA, sortFieldName), ERROR_SEARCH_REQUEST);
        }

        if (insensitiveCompare(messageSearch.getQueryParameters()[k].first, input_json_message_const::sortOrderBy))
        {
          // set different sort modes
          unsigned int sortOrderBy = std::stoul(messageSearch.getQueryParameters()[k].second);
          if (!setSortMode(sortOrderBy, sortFieldName))
            throw Poco::Exception(errorMsg, errorCode);
        }

        if (insensitiveCompare(messageSearch.getQueryParameters()[k].first, input_json_message_const::offsetResultsNumber))
          offsetResults = std::stoul(messageSearch.getQueryParameters()[k].second);

        if (insensitiveCompare(messageSearch.getQueryParameters()[k].first, input_json_message_const::limitResultsNumber))
          limitResults = std::stoul(messageSearch.getQueryParameters()[k].second);

        if (insensitiveCompare(messageSearch.getQueryParameters()[k].first, input_json_message_const::cutoffResultNumber))
          cutoffResults = std::stoul(messageSearch.getQueryParameters()[k].second);

        if (insensitiveCompare(messageSearch.getQueryParameters()[k].first, input_json_message_const::maxResultsNumber))
          maxResultsNumber = std::stoul(messageSearch.getQueryParameters()[k].second);

        if (insensitiveCompare(messageSearch.getQueryParameters()[k].first, input_json_message_const::timeoutValueName))
           timeoutValue = std::stoul(messageSearch.getQueryParameters()[k].second);
      }
      catch(std::exception& e)
      {
        throw Poco::Exception("'"+messageSearch.getQueryParameters()[k].first+"' "+e.what(), ERROR_SEARCH_REQUEST);
      }
    }

    // read order parameters
    for (size_t k=0;k<messageSearch.getOrderParameters().size();++k)
    {
      try
      {
        if (insensitiveCompare(messageSearch.getOrderParameters()[k].first, input_json_message_const::weightAlgorithm))
          weightAlgirithm = messageSearch.getOrderParameters()[k].second;

        if (insensitiveCompare(messageSearch.getOrderParameters()[k].first, input_json_message_const::orderBy))
          orderBy = std::stoul(messageSearch.getOrderParameters()[k].second);
      }
      catch(std::exception& e)
      {
        throw Poco::Exception("'"+messageSearch.getOrderParameters()[k].first+"' "+e.what(), ERROR_SEARCH_REQUEST);
      }
    }
  }
  catch(Poco::Exception& e)
  {
    _isError = true;
    fObj.log(LoggerStream::Priority::PRIO_INFORMATION) << message(message_const::ERROR_MSG, e.message()) << message(message_const::ERROR_CODE, e.code()) << flush;
    throw Poco::Exception(message(message_const::GETTING_PARAMETERS_FAIL, e.message()), ERROR_SEARCH_REQUEST);
  }

  if (limitResults==0)
    limitResults = maxResultsNumber;
  // set different limits
  setLimits(offsetResults, limitResults, maxResultsNumber, cutoffResults);

  // cleanup filters
  resetFilters();
  // add filters
  for (size_t i=0;i<messageSearch.getFilters().getFilters().size();++i)
    applyFilter(messageSearch.getFilters().getFilters()[i]);
  // set max allowed query time
  setMaxQueryTime(timeoutValue);

  if (!searchDocument(messageSearch.getQueryString(), resultMemoryManager, requestInfo))
    throw Poco::Exception(getErrorMsg(), ERROR_SEARCH_REQUEST);

  requestInfo.setNodeName(fObj.getNodeName());
  requestInfo.setMaxResultsNumber(maxResultsNumber);
  requestInfo.setOrderBy(orderBy);
  resultData.addRequestInfo(std::forward<SphinxRequestInfo>(requestInfo));

  SphinxWeightCalculator calculator;
  CalculateStrategyFactory factory;
  calculator.setStrategy(factory.create(weightAlgirithm));
  calculator.calculate(orderFieldsNames, resultData);
  fObj.log(LoggerStream::Priority::PRIO_INFORMATION) << calculator.getLogString();

  std::string resultJson;
  if (!resultSerializator.serialize(resultJson, orderFieldsNames.size(), fObj.getMinNumberFieldsPacking()))
    throw Poco::Exception(resultSerializator.getErrorMsg(), ERROR_SEARCH_REQUEST);

  return resultJson;
}
//-------------------------------------------------------------------
SphinxOutputJsonMessage SphinxSearcher::Process(SphinxInputJsonMessage& inputJsonMessage)
{
  _isError = false;
  errorMsg.clear();
  errorCode = NO_ERROR;
  std::string resultData;
  timeval tvStart, tvStop;
  gettimeofday(&tvStart, 0);          //gettimeofday does not support TZ adjust on Linux.

  try
  {
    if (inputJsonMessage.getType() != SphinxInputJsonMessage::MessageType::mtSearch)
      throw Poco::Exception(message(message_const::INCORRECT_JSON_TYPE), COMMAND_ERROR);

    resultData = makeSearchCommand(inputJsonMessage.getData());
    if (resultData.empty())
    {
      SphinxDefaultJSON defaultJson;
      resultData = defaultJson.getJSON();
    }
  }
  catch(Poco::Exception& e)
  {
    _isError = true;
    errorCode = e.code();
    errorMsg = e.displayText();

    SphinxDefaultJSON defaultJson;
    resultData = defaultJson.getJSON();
  }
  gettimeofday(&tvStop, 0);

  SphinxOutputJsonMessage outputMessage;
  outputMessage.setErrorCode(errorCode);
  outputMessage.setErrorMessage(errorMsg);
  outputMessage.setData(resultData);
  outputMessage.setTime(getTimeInterval(tvStart, tvStop)/1000);

  return outputMessage;
}
//-------------------------------------------------------------------
//-------------------------------------------------------------------
} // end namespace sphinx
} // end namespace HCE


