/**
 * @file SphinxSearcher.hpp
 *
 * Implementation of a functional object is responsible for search use sphinx in HCE project node
 *
 * @author Alexander Vybornyh <alexander.hce.cluster@gmail.com>
 * @link http://hierarchical-cluster-engine.com/
 * @copyright Copyright &copy; 2013 IOIX Ukraine
 * @license http://hierarchical-cluster-engine.com/license/
 * @package HCE project node API
 * @since 0.1
 */

#ifndef SPHINX_SEARCHER_HPP
#define SPHINX_SEARCHER_HPP

#include <set>

#include "sphinxclient.h"

#include "SphinxFilters.hpp"
#include "SphinxSearchInterface.hpp"
#include "SphinxResultDataSerializator.hpp"
#include "SphinxResultDataMemoryManager.hpp"
#include "CustomMessage.hpp"

typedef struct st_sphinx_client sphinx_client;

namespace HCE
{
namespace sphinx
{
class SphinxFunctionalObject;
//-------------------------------------------------------------------
class SphinxSearcher : public SphinxSearchInterface
{
	sphinx_client * client;
	char** field_names;
	int* field_weights;
	size_t fieldsCount;

	char** index_names;
	int* index_weights;
	size_t indexesCount;
public:
	explicit SphinxSearcher(SphinxFunctionalObject& fObj, bool startSearchd=false, bool stopSearchd_=false);
	virtual ~SphinxSearcher(void);

	bool easyStart(int matchMode=SPH_MATCH_EXTENDED2, int sortMode=SPH_SORT_RELEVANCE, int rankingMode=SPH_RANK_DEFAULT, const std::string& rankexpr="");
	bool startSearchd(void);
	bool stopSearchd(void);
	bool isActiveSearchd(void);
	bool isNeedStopSearchd(void) const {return _StopSearchd;}
	bool isConnected(void) const {return _isConnected;}
	bool isStoppedSearchd(void) const {return _isStoppedSearchd;}

	std::string Process(const std::string& json); // make search (black box)
	SphinxOutputJsonMessage Process(SphinxInputJsonMessage& inputJsonMessage); // make search (black box)

	bool open(void);  // persistence connection
	bool close(void); // close connection

	void setServerHost(const std::string& serverHost_) {serverHost=serverHost_;}
	std::string getServerHost(void) const {return serverHost;}

	void setServerPort(int serverPort_) {serverPort=serverPort_;}
	int getServerPort(void) const {return serverPort;}


	bool setFieldWeights(std::vector <std::pair<std::string, int> >& Vec);
	bool setIndexWeights(std::vector <std::pair<std::string, int> >& Vec);

	void getFieldWeights(std::vector <std::pair<std::string, int> >& Vec);
	void getIndexWeights(std::vector <std::pair<std::string, int> >& Vec);

	bool addField(const std::string& name, int weight=0);
	bool setField(const std::string& name, int weight);
	bool delField(const std::string& name);

	bool addIndexFile(const std::string& name, int weight=0);
	bool setIndexFile(const std::string& name, int weight);
	bool delIndexFile(const std::string& name);

	bool isExistField(const std::string& name);
	bool isExistIndexFile(const std::string& name);

	int getFieldWeight(const std::string& name);
	int getIndexWeight(const std::string& name);

	bool setMatchMode(int mode);	/* Match mode can be:
																	 SPH_MATCH_ALL        = 0,
																	 SPH_MATCH_ANY        = 1,
																	 SPH_MATCH_PHRASE     = 2,
																	 SPH_MATCH_BOOLEAN    = 3,
																	 SPH_MATCH_EXTENDED   = 4,
																	 SPH_MATCH_FULLSCAN   = 5,
																	 SPH_MATCH_EXTENDED2  = 6 */

	bool setSortMode(int mode, const std::string& sortby="");   /* Sort mode can be:
 																																 SPH_SORT_RELEVANCE      = 0,
																														     SPH_SORT_ATTR_DESC      = 1,
																														     SPH_SORT_ATTR_ASC       = 2,
																														     SPH_SORT_TIME_SEGMENTS  = 3,
																														     SPH_SORT_EXTENDED       = 4,
																														     SPH_SORT_EXPR           = 5 */


	bool setRankingMode(int ranker, const std::string& rankexpr="");  /* Ranking mode can be:
                                                                    SPH_RANK_PROXIMITY_BM25 = 0,
                                                                    SPH_RANK_BM25           = 1,
                                                                    SPH_RANK_NONE           = 2,
                                                                    SPH_RANK_WORDCOUNT      = 3,
                                                                    SPH_RANK_PROXIMITY      = 4,
                                                                    SPH_RANK_MATCHANY       = 5,
                                                                    SPH_RANK_FIELDMASK      = 6,
                                                                    SPH_RANK_SPH04          = 7,
                                                                    SPH_RANK_EXPR           = 8,
                                                                    SPH_RANK_TOTAL          = 9,
                                                                    SPH_RANK_DEFAULT        = SPH_RANK_PROXIMITY_BM25 */

	size_t getFieldsCount(void) const {return fieldsCount;}
	size_t getIndexesCount(void) const {return indexesCount;}

	bool setIndexFileName(const std::string& indexFileName_);
	std::string getIndexFileName(void) const {return indexFileName;}

	void setTimeoutedRequests(unsigned long long timeoutedRequests_) {timeoutedRequests=timeoutedRequests_;}
	unsigned long long getTimeoutedRequests(void) const {return timeoutedRequests;}
	void resetTimeoutedRequests(void) {timeoutedRequests=0;}

	void setDefaultRequestTimeout(unsigned long defaultRequestTimeout_) {defaultRequestTimeout=defaultRequestTimeout_;}
	unsigned long getDefaultRequestTimeout(void) const {return defaultRequestTimeout;}

	void resizeResultData(size_t matchesCount, size_t attributesCount);
protected:
	bool addFilter(const std::string& attrName, long long value, bool exclude=false);
  bool addFilter(const std::string& attrName, std::vector<long long>& values, bool exclude=false);
	bool addFilterRange(const std::string& attrName, long long umin, long long umax, bool exclude=false);
	bool addFilterFloatRange(const std::string& attrName, float fmin, float fmax, bool exclude=false);
	void resetFilters(void);

	void setMaxResultsNumber(unsigned int maxResultsNumber) throw (Poco::Exception);    // set limits max result number (with offset==0 and limit==maxResultNumber)
	void setLimits(unsigned int offset, unsigned int limit, unsigned int maxMatches, unsigned int cutoff=0) throw (Poco::Exception);   // detail set limits before search
	void applyFilter(Poco::SharedPtr<SphinxFilter> pFilter) throw (Poco::Exception);  // add attributes filter before search
	void setMaxQueryTime(unsigned long maxQueryTime) throw (Poco::Exception); // set allowed query time before search
public:
	bool searchDocument(const std::string& query, SphinxResultDataMemoryManager& resultMemoryManager, SphinxRequestInfo& requestInfo);

protected:
	std::string makeSearchCommand(const std::string& json) throw (Poco::Exception, std::exception);

	SphinxFunctionalObject& fObj;
	CustomMessage& message;

	std::string serverHost;
	int serverPort;

	std::string indexFileName;

	bool _StopSearchd;
	bool _isConnected;
	bool _isStoppedSearchd;
	unsigned long long timeoutedRequests;
	unsigned long defaultRequestTimeout;

	void cleanup(char**& names, int*& weights, size_t& count);
	bool setWeights(char**& names, int*& weights, size_t& count, std::vector <std::pair<std::string, int> >& Vec);
	void getWeights(char**& names, int*& weights, size_t& count, std::vector <std::pair<std::string, int> >& Vec);
	static std::vector <std::pair<std::string, int> >::iterator find(std::vector <std::pair<std::string, int> >& Vec, const std::string& name);

	void setAllowedFields(std::vector<std::string>& queryExtendedFields);
	std::set<std::string>& getAlloedFields(void) {return allowedFields;}
	bool isAllowedFields(const std::string& extendedFieldName);

	void setFullSchemaNames(std::vector<std::string>& schemaNames);
	std::set<std::string>& getFullSchemaNames(void) {return fullSchemaNames;}
	bool isExistSchemaNames(const std::string& schemaName);

	std::set<std::string> allowedFields;
	std::set<std::string> fullSchemaNames;

	SphinxRequestInfo requestInfo;
	SphinxResultData resultData;
	SphinxResultDataSerializator resultSerializator;
	SphinxResultDataMemoryManager resultMemoryManager;

	friend class SphinxFunctionalObject;
};
//-------------------------------------------------------------------
//-------------------------------------------------------------------
} // end namespace sphinx
} // end namespace HCE

#endif // SPHINX_SEARCHER_HPP

