#include <gtest/gtest.h>
#include <iomanip>
#include <sstream>
#include <fstream>
#include <set>

#include <Poco/SharedPtr.h>
#include "SphinxTestPrinter.hpp"

#include "EncodeDecodeBase64.hpp"
#include "SphinxError.hpp"
#include "SphinxManageTest.hpp"
#include "SphinxInputJsonMessage.hpp"
#include "SphinxOutputJsonMessage.hpp"
#include "SphinxFunctionalObject.hpp"
#include "SphinxFunctionalObjectTest.hpp"
#include "SphinxPrintStatus.hpp"

namespace HCE
{
namespace sphinx
{
namespace tests
{
extern std::string homeDir;
extern Poco::SharedPtr<HCE::sphinx::SphinxFunctionalObject> pObj;

const std::string nodeName = "test_node";
const std::string defaultIndex = "i001";
const std::string indexName = "i003";
const std::string copyIndexName = "i004";
const std::string renameIndexName = "i005";
//-----------------------------------------------------------------------------
TEST(SphinxAdminCommand, notLoadConfigInConstructor)
{
  HCE::sphinx::SphinxFunctionalObject fObject(nodeName, 234, homeDir);
  ASSERT_FALSE(fObject.isLoadedConfig());
  ASSERT_FALSE(fObject.isActiveSearchd());
}
//-----------------------------------------------------------------------------
TEST(SphinxAdminCommand, loadConfigInConstructor)
{
  HCE::sphinx::SphinxFunctionalObject* pObj = new HCE::sphinx::SphinxFunctionalObject(nodeName, 234, homeDir);
  ASSERT_NE(pObj, nullptr);

  SphinxManageTest::testIndexCreate(*pObj, defaultIndex);
  ASSERT_TRUE(pObj->getErrorCode()==NO_ERROR || pObj->getErrorCode()==ERROR_CREATE_INDEX) << printError(*pObj);
  delete pObj;

  HCE::sphinx::SphinxFunctionalObject* pObject = new HCE::sphinx::SphinxFunctionalObject(nodeName, 234, homeDir, defaultIndex, sphinx_search_const::searchdStart, sphinx_search_const::searchdStop);
  ASSERT_NE(pObject, nullptr);
  pObject->setLogStream(std::cout);
  ASSERT_TRUE(pObject->isLoadedConfig());
  ASSERT_TRUE(pObject->isActiveSearchd());
  delete pObject;

  pObj = new HCE::sphinx::SphinxFunctionalObject("", 0, homeDir);
  ASSERT_NE(pObj, nullptr);
  SphinxManageTest::testIndexRemove(*pObj, defaultIndex);
  ASSERT_FALSE(pObj->isError()) << printError(*pObj);
  delete pObj;
}
//-----------------------------------------------------------------------------
//-----------------------------------------------------------------------------
class SphinxAdminCommandTest : public ::testing::Test
{
protected:
  static void SetUpTestCase(void)
  {
    pObj = Poco::SharedPtr<HCE::sphinx::SphinxFunctionalObject>(new HCE::sphinx::SphinxFunctionalObject(nodeName, 234, homeDir));
//    pObj->setLogStream(std::cout);
  }
  // cppcheck-suppress unusedFunction
  static void TearDownTestCase(void)
  {
    pObj = nullptr;
  }
};

std::vector<std::string> branches;
//-----------------------------------------------------------------------------
//-----------------------------------------------------------------------------
TEST_F(SphinxAdminCommandTest, isCreatedSphinxFunctionalObject)
{
  ASSERT_FALSE(pObj.isNull());
}
//-----------------------------------------------------------------------------
TEST_F(SphinxAdminCommandTest, isActiveSearchd)
{
  ASSERT_FALSE(pObj->isActiveSearchd());
}
//-----------------------------------------------------------------------------
TEST_F(SphinxAdminCommandTest, isNotLoadedConfig)
{
  ASSERT_FALSE(pObj->isLoadedConfig());
}
//-----------------------------------------------------------------------------
TEST_F(SphinxAdminCommandTest, testIndexCreateRemove)
{
  SphinxManageTest::testIndexCreate(*pObj, indexName);
  ASSERT_TRUE(pObj->getErrorCode()==NO_ERROR || pObj->getErrorCode()==ERROR_CREATE_INDEX) << printError(*pObj);
  SphinxManageTest::testIndexRemove(*pObj, indexName);
  ASSERT_FALSE(pObj->isError()) << printError(*pObj);
}
//-----------------------------------------------------------------------------
//-----------------------------------------------------------------------------
TEST_F(SphinxAdminCommandTest, testIndexCheckExist)
{
  SphinxManageTest::testIndexCreate(*pObj, indexName);
  ASSERT_TRUE(pObj->getErrorCode()==NO_ERROR || pObj->getErrorCode()==ERROR_CREATE_INDEX) << printError(*pObj);
  SphinxManageTest::testIndexCheck(*pObj, indexName, true);
  ASSERT_TRUE(pObj->getErrorCode()==NO_ERROR || pObj->getErrorCode()==ERROR_CHECK_INDEX) << printError(*pObj);
  SphinxManageTest::testIndexRemove(*pObj, indexName);
  ASSERT_FALSE(pObj->isError()) << printError(*pObj);
}
//-----------------------------------------------------------------------------
//-----------------------------------------------------------------------------
TEST_F(SphinxAdminCommandTest, testIndexCheckNotExist)
{
  SphinxManageTest::testIndexCheck(*pObj, "branch_0006", false);
  ASSERT_TRUE(pObj->getErrorCode()==NO_ERROR || pObj->getErrorCode()==ERROR_CHECK_INDEX) << printError(*pObj);
}
//-----------------------------------------------------------------------------
//-----------------------------------------------------------------------------
TEST_F(SphinxAdminCommandTest, testIndexStoreDataFile)
{
  SphinxManageTest::testIndexCreate(*pObj, indexName);
  ASSERT_TRUE(pObj->getErrorCode()==NO_ERROR || pObj->getErrorCode()==ERROR_CREATE_INDEX) << printError(*pObj);
  branches.clear();
  branches.push_back("branch_0005");
  branches.push_back("branch_0006");
  branches.push_back("branch_0007");
  for (size_t i=0;i<branches.size();++i)
  {
    SphinxManageTest::testIndexStoreDataFile(*pObj, indexName, branches[i], "");
    SphinxManageTest::testIndexStoreDataFile(*pObj, indexName, branches[i]);
  }
  ASSERT_FALSE(pObj->isError()) << printError(*pObj);
  SphinxManageTest::testIndexRemove(*pObj, indexName);
  ASSERT_FALSE(pObj->isError()) << printError(*pObj);
}
//-----------------------------------------------------------------------------
//-----------------------------------------------------------------------------
TEST_F(SphinxAdminCommandTest, testIndexStoreSchemaFile)
{
  SphinxManageTest::testIndexCreate(*pObj, indexName);
  ASSERT_TRUE(pObj->getErrorCode()==NO_ERROR || pObj->getErrorCode()==ERROR_CREATE_INDEX) << printError(*pObj);
  SphinxManageTest::testIndexStoreSchemaFile(*pObj, indexName, "");
  SphinxManageTest::testIndexStoreSchemaFile(*pObj, indexName);
  ASSERT_FALSE(pObj->isError()) << printError(*pObj);
  SphinxManageTest::testIndexRemove(*pObj, indexName);
  ASSERT_FALSE(pObj->isError()) << printError(*pObj);
}
//-----------------------------------------------------------------------------
//-----------------------------------------------------------------------------
TEST_F(SphinxAdminCommandTest, testIndexMaxDocId)
{
  SphinxManageTest::testIndexCreate(*pObj, indexName);
  ASSERT_TRUE(pObj->getErrorCode()==NO_ERROR || pObj->getErrorCode()==ERROR_CREATE_INDEX) << printError(*pObj);
  branches.clear();
  branches.push_back("branch_0005");
  for (size_t i=0;i<branches.size();++i)
    SphinxManageTest::testIndexStoreDataFile(*pObj, indexName, branches[i]);
  SphinxManageTest::testIndexStoreSchemaFile(*pObj, indexName);
  SphinxManageTest::testIndexRebuild(*pObj, indexName, branches);
  SphinxManageTest::testIndexMaxDocId(*pObj, indexName);
  ASSERT_FALSE(pObj->isError()) << printError(*pObj);
  SphinxManageTest::testIndexRemove(*pObj, indexName);
  ASSERT_FALSE(pObj->isError()) << printError(*pObj);
}
//-----------------------------------------------------------------------------
//-----------------------------------------------------------------------------
TEST_F(SphinxAdminCommandTest, testIndexMaxDocIdExtended)
{
  SphinxManageTest::testIndexCreate(*pObj, indexName);
  ASSERT_TRUE(pObj->getErrorCode()==NO_ERROR || pObj->getErrorCode()==ERROR_CREATE_INDEX) << printError(*pObj);
  branches.clear();
  branches.push_back("branch_0005");
  branches.push_back("branch_0006");
  for (size_t i=0;i<branches.size();++i)
    SphinxManageTest::testIndexStoreDataFile(*pObj, indexName, branches[i]);
  SphinxManageTest::testIndexStoreSchemaFile(*pObj, indexName);
  SphinxManageTest::testIndexRebuild(*pObj, indexName, branches);
  SphinxManageTest::testIndexMaxDocIdExtended(*pObj, indexName);
  ASSERT_FALSE(pObj->isError()) << printError(*pObj);
  SphinxManageTest::testIndexRemove(*pObj, indexName);
  ASSERT_FALSE(pObj->isError()) << printError(*pObj);
}
//-----------------------------------------------------------------------------
//-----------------------------------------------------------------------------
TEST_F(SphinxAdminCommandTest, testIndexRebuild)
{
  SphinxManageTest::testIndexCreate(*pObj, indexName);
  ASSERT_TRUE(pObj->getErrorCode()==NO_ERROR || pObj->getErrorCode()==ERROR_CREATE_INDEX) << printError(*pObj);
  SphinxManageTest::testIndexStoreSchemaFile(*pObj, indexName);
  branches.clear();
  branches.push_back("branch_0005");
  branches.push_back("branch_0006");
  branches.push_back("branch_0007");
  for (size_t i=0;i<branches.size();++i)
    SphinxManageTest::testIndexStoreDataFile(*pObj, indexName, branches[i]);
  branches.push_back("branch_0008");
  branches.push_back("branch_0009");
  SphinxManageTest::testIndexRebuild(*pObj, indexName, branches);

  branches.clear();
  branches.push_back("*");
  SphinxManageTest::testIndexRebuild(*pObj, indexName, branches);
  ASSERT_FALSE(pObj->isError()) << printError(*pObj);
  SphinxManageTest::testIndexRemove(*pObj, indexName);
  ASSERT_FALSE(pObj->isError()) << printError(*pObj);
}
//-----------------------------------------------------------------------------
//-----------------------------------------------------------------------------
TEST_F(SphinxAdminCommandTest, testIndexMergeStart)
{
  SphinxManageTest::testIndexCreate(*pObj, indexName);
  ASSERT_TRUE(pObj->getErrorCode()==NO_ERROR || pObj->getErrorCode()==ERROR_CREATE_INDEX) << printError(*pObj);
  SphinxManageTest::testIndexStoreSchemaFile(*pObj, indexName);
  branches.clear();
  branches.push_back("branch_0005");
  branches.push_back("branch_0006");
  branches.push_back("branch_0007");
  for (size_t i=0;i<branches.size();++i)
    SphinxManageTest::testIndexStoreDataFile(*pObj, indexName, branches[i]);
  SphinxManageTest::testIndexRebuild(*pObj, indexName, branches);
  SphinxManageTest::testIndexMerge(*pObj, indexName, branches);
  SphinxManageTest::testIndexStart(*pObj, indexName);
  ASSERT_FALSE(pObj->isError()) << printError(*pObj);
  SphinxManageTest::testIndexRemove(*pObj, indexName);
  ASSERT_FALSE(pObj->isError()) << printError(*pObj);
}
//-----------------------------------------------------------------------------
//-----------------------------------------------------------------------------
TEST_F(SphinxAdminCommandTest, testIndexDeleteDataFile)
{
  SphinxManageTest::testIndexCreate(*pObj, indexName);
  ASSERT_TRUE(pObj->getErrorCode()==NO_ERROR || pObj->getErrorCode()==ERROR_CREATE_INDEX) << printError(*pObj);
  branches.clear();
  branches.push_back("branch_0005");
  for (size_t i=0;i<branches.size();++i)
    SphinxManageTest::testIndexStoreDataFile(*pObj, indexName, branches[i]);
  SphinxManageTest::testIndexDeleteDataFile(*pObj, indexName, branches);
  ASSERT_FALSE(pObj->isError()) << printError(*pObj);
  SphinxManageTest::testIndexRemove(*pObj, indexName);
  ASSERT_FALSE(pObj->isError()) << printError(*pObj);
}
//-----------------------------------------------------------------------------
//-----------------------------------------------------------------------------
TEST_F(SphinxAdminCommandTest, testIndexRebuildMergeStart)
{
  SphinxManageTest::testIndexCreate(*pObj, indexName);
  ASSERT_TRUE(pObj->getErrorCode()==NO_ERROR || pObj->getErrorCode()==ERROR_CREATE_INDEX) << printError(*pObj);
  SphinxManageTest::testIndexStoreSchemaFile(*pObj, indexName);
  branches.clear();
  branches.push_back("branch_0005");
  branches.push_back("branch_0006");
  branches.push_back("branch_0007");
  for (size_t i=0;i<branches.size();++i)
    SphinxManageTest::testIndexStoreDataFile(*pObj, indexName, branches[i]);
  branches.clear();
  branches.push_back("*");
  SphinxManageTest::testIndexRebuild(*pObj, indexName, branches);
  SphinxManageTest::testIndexMerge(*pObj, indexName, branches);
  SphinxManageTest::testIndexStart(*pObj, indexName);
  ASSERT_FALSE(pObj->isError()) << printError(*pObj);
  SphinxManageTest::testIndexRemove(*pObj, indexName);
  ASSERT_FALSE(pObj->isError()) << printError(*pObj);
}
//-----------------------------------------------------------------------------
//-----------------------------------------------------------------------------
TEST_F(SphinxAdminCommandTest, testIndexMergeNotExist)
{
  SphinxManageTest::testIndexCreate(*pObj, indexName);
  ASSERT_TRUE(pObj->getErrorCode()==NO_ERROR || pObj->getErrorCode()==ERROR_CREATE_INDEX) << printError(*pObj);
  SphinxManageTest::testIndexStoreSchemaFile(*pObj, indexName);
  branches.clear();
  branches.push_back("branch_0005");
  branches.push_back("branch_0006");
  branches.push_back("branch_0007");
  for (size_t i=0;i<branches.size();++i)
    SphinxManageTest::testIndexStoreDataFile(*pObj, indexName, branches[i]);
  SphinxManageTest::testIndexRebuild(*pObj, indexName, branches);
  SphinxManageTest::testIndexMerge(*pObj, indexName, branches);
  branches.push_back("branch_0008");  // not exists
  branches.push_back("branch_0009");  // not exists
  SphinxManageTest::testIndexMerge(*pObj, indexName, branches);
  ASSERT_FALSE(pObj->isError()) << printError(*pObj);
  SphinxManageTest::testIndexRemove(*pObj, indexName);
  ASSERT_FALSE(pObj->isError()) << printError(*pObj);
}
//-----------------------------------------------------------------------------
//-----------------------------------------------------------------------------
TEST_F(SphinxAdminCommandTest, testIndexStartStopStart)
{
  SphinxManageTest::testIndexCreate(*pObj, indexName);
  ASSERT_TRUE(pObj->getErrorCode()==NO_ERROR || pObj->getErrorCode()==ERROR_CREATE_INDEX) << printError(*pObj);
  SphinxManageTest::testIndexStoreSchemaFile(*pObj, indexName);
  branches.clear();
  branches.push_back("branch_0005");
  for (size_t i=0;i<branches.size();++i)
    SphinxManageTest::testIndexStoreDataFile(*pObj, indexName, branches[i]);
  SphinxManageTest::testIndexRebuild(*pObj, indexName, branches);
  SphinxManageTest::testIndexMerge(*pObj, indexName, branches);
  SphinxManageTest::testIndexStart(*pObj, indexName);
  SphinxManageTest::testIndexStop(*pObj, indexName);
  SphinxManageTest::testIndexStart(*pObj, indexName);
  ASSERT_FALSE(pObj->isError()) << printError(*pObj);
  SphinxManageTest::testIndexRemove(*pObj, indexName);
  ASSERT_FALSE(pObj->isError()) << printError(*pObj);
}
//-----------------------------------------------------------------------------
TEST_F(SphinxAdminCommandTest, testIndexDeleteDataFileAndDeleteSchemaFile)
{
  SphinxManageTest::testIndexCreate(*pObj, indexName);
  ASSERT_TRUE(pObj->getErrorCode()==NO_ERROR || pObj->getErrorCode()==ERROR_CREATE_INDEX) << printError(*pObj);
  SphinxManageTest::testIndexStoreSchemaFile(*pObj, indexName);
  branches.clear();
  branches.push_back("branch_0005");
  branches.push_back("branch_0006");
  branches.push_back("branch_0007");
  for (size_t i=0;i<branches.size();++i)
    SphinxManageTest::testIndexStoreDataFile(*pObj, indexName, branches[i]);
  SphinxManageTest::testIndexDeleteDataFile(*pObj, indexName, branches);
  for (size_t i=0;i<branches.size();++i)
    SphinxManageTest::testIndexStoreDataFile(*pObj, indexName, branches[i]);
  branches.clear();
  branches.push_back("*");
  SphinxManageTest::testIndexDeleteDataFile(*pObj, indexName, branches);
  SphinxManageTest::testIndexDeleteSchemaFile(*pObj, indexName);
  ASSERT_FALSE(pObj->isError()) << printError(*pObj);
  SphinxManageTest::testIndexRemove(*pObj, indexName);
  ASSERT_FALSE(pObj->isError()) << printError(*pObj);
}
//-----------------------------------------------------------------------------
//-----------------------------------------------------------------------------
TEST_F(SphinxAdminCommandTest, testIndexAppendDataFile)
{
  SphinxManageTest::testIndexCreate(*pObj, indexName);
  ASSERT_TRUE(pObj->getErrorCode()==NO_ERROR || pObj->getErrorCode()==ERROR_CREATE_INDEX) << printError(*pObj);
  SphinxManageTest::testIndexStoreSchemaFile(*pObj, indexName);
  branches.clear();
  branches.push_back("branch_0005");
  branches.push_back("branch_0006");
  branches.push_back("branch_0007");
  for (size_t i=0;i<branches.size();++i)
    SphinxManageTest::testIndexStoreDataFile(*pObj, indexName, branches[i]);
  SphinxManageTest::testIndexRebuild(*pObj, indexName, branches);
  SphinxManageTest::testIndexMerge(*pObj, indexName, branches);
  SphinxManageTest::testIndexStart(*pObj, indexName);
  branches.clear();
  branches.push_back("branch_0005");
  branches.push_back("branch_0006");
  for (size_t i=0;i<branches.size();++i)
    SphinxManageTest::testIndexStoreDataFile(*pObj, indexName, branches[i]);
  SphinxManageTest::testIndexStoreSchemaFile(*pObj, indexName);
  for (size_t i=0;i<branches.size();++i)
    SphinxManageTest::testIndexAppendDataFile(*pObj, indexName, branches[i]);
  ASSERT_FALSE(pObj->isError()) << printError(*pObj);
  SphinxManageTest::testIndexRemove(*pObj, indexName);
  ASSERT_FALSE(pObj->isError()) << printError(*pObj);
}
//-----------------------------------------------------------------------------
//-----------------------------------------------------------------------------
TEST_F(SphinxAdminCommandTest, testIndexDeleteDoc)
{
  SphinxManageTest::testIndexCreate(*pObj, indexName);
  ASSERT_TRUE(pObj->getErrorCode()==NO_ERROR || pObj->getErrorCode()==ERROR_CREATE_INDEX) << printError(*pObj);
  SphinxManageTest::testIndexStoreSchemaFile(*pObj, indexName);
  branches.clear();
  branches.push_back("branch_0005");
  branches.push_back("branch_0006");
  branches.push_back("branch_0007");
  for (size_t i=0;i<branches.size();++i)
    SphinxManageTest::testIndexStoreDataFile(*pObj, indexName, branches[i]);
  SphinxManageTest::testIndexRebuild(*pObj, indexName, branches);
  SphinxManageTest::testIndexMerge(*pObj, indexName, branches);

  std::vector<unsigned long long> ids = { 123456789012, 123456789023, 123456789034, 123456789045, 123456789056, 123456789067, 123456789078, 123456789089, 123456789090, 1234567890101};
  SphinxManageTest::testIndexDeleteDoc(*pObj, indexName, ids);
  SphinxManageTest::testIndexDeleteDocNumber(*pObj, indexName);
  ASSERT_FALSE(pObj->isError()) << printError(*pObj);
  SphinxManageTest::testIndexRemove(*pObj, indexName);
  ASSERT_FALSE(pObj->isError()) << printError(*pObj);
}
//-----------------------------------------------------------------------------
//-----------------------------------------------------------------------------
TEST_F(SphinxAdminCommandTest, testIndexPackDocData)
{
  SphinxManageTest::testIndexCreate(*pObj, indexName);
  ASSERT_TRUE(pObj->getErrorCode()==NO_ERROR || pObj->getErrorCode()==ERROR_CREATE_INDEX) << printError(*pObj);
  SphinxManageTest::testIndexStoreSchemaFile(*pObj, indexName);
  branches.clear();
  branches.push_back("branch_0005");
  branches.push_back("branch_0006");
  branches.push_back("branch_0007");
  for (size_t i=0;i<branches.size();++i)
    SphinxManageTest::testIndexStoreDataFile(*pObj, indexName, branches[i]);

  SphinxManageTest::testIndexDeleteDocNumber(*pObj, indexName);
  SphinxManageTest::testIndexPackDocData(*pObj, indexName, branches);

  SphinxManageTest::testIndexRebuild(*pObj, indexName, branches);
  SphinxManageTest::testIndexMerge(*pObj, indexName, branches);
  branches.clear();
  branches.push_back("*");
  SphinxManageTest::testIndexPackDocData(*pObj, indexName, branches);
  SphinxManageTest::testIndexMerge(*pObj, indexName, branches);
  ASSERT_FALSE(pObj->isError()) << printError(*pObj);
  SphinxManageTest::testIndexRemove(*pObj, indexName);
  ASSERT_FALSE(pObj->isError()) << printError(*pObj);
}
//-----------------------------------------------------------------------------
//-----------------------------------------------------------------------------
TEST_F(SphinxAdminCommandTest, testIndexRemove)
{
  SphinxManageTest::testIndexCreate(*pObj, indexName);
  ASSERT_TRUE(pObj->getErrorCode()==NO_ERROR || pObj->getErrorCode()==ERROR_CREATE_INDEX) << printError(*pObj);
  SphinxManageTest::testIndexRemove(*pObj, indexName);
  SphinxManageTest::testIndexCheck(*pObj, indexName, false);
  ASSERT_TRUE(pObj->isError());
}
//-----------------------------------------------------------------------------
//-----------------------------------------------------------------------------
TEST_F(SphinxAdminCommandTest, testIndexCopy)
{
  // prepare for next copy index
  SphinxManageTest::testIndexCreate(*pObj, indexName);
  SphinxManageTest::testIndexCheck(*pObj, indexName, true);
  branches.clear();
  branches.push_back("branch_0005");
  branches.push_back("branch_0006");
  branches.push_back("branch_0007");
  for (size_t i=0;i<branches.size();++i)
    SphinxManageTest::testIndexStoreDataFile(*pObj, indexName, branches[i]);
  SphinxManageTest::testIndexStoreSchemaFile(*pObj, indexName);
  SphinxManageTest::testIndexRebuild(*pObj, indexName, branches);
  SphinxManageTest::testIndexMerge(*pObj, indexName, branches);
  // make copy index
  SphinxManageTest::testIndexStart(*pObj, indexName);
  SphinxManageTest::testIndexStop(*pObj, indexName);
  SphinxManageTest::testIndexStart(*pObj, indexName);
  SphinxManageTest::testIndexCopy(*pObj, indexName, copyIndexName);
  SphinxManageTest::testIndexCheck(*pObj, copyIndexName, true);
  ASSERT_FALSE(pObj->isError()) << printError(*pObj);
  SphinxManageTest::testIndexRemove(*pObj, indexName);
  ASSERT_FALSE(pObj->isError()) << printError(*pObj);
  SphinxManageTest::testIndexRemove(*pObj, copyIndexName);
  ASSERT_FALSE(pObj->isError()) << printError(*pObj);
}
//-----------------------------------------------------------------------------
//-----------------------------------------------------------------------------
TEST_F(SphinxAdminCommandTest, testIndexRename)
{
  // prepare for next copy index
  SphinxManageTest::testIndexCreate(*pObj, indexName);
  SphinxManageTest::testIndexCheck(*pObj, indexName, true);
  branches.clear();
  branches.push_back("branch_0005");
  branches.push_back("branch_0006");
  branches.push_back("branch_0007");
  for (size_t i=0;i<branches.size();++i)
    SphinxManageTest::testIndexStoreDataFile(*pObj, indexName, branches[i]);
  SphinxManageTest::testIndexStoreSchemaFile(*pObj, indexName);
  SphinxManageTest::testIndexRebuild(*pObj, indexName, branches);
  SphinxManageTest::testIndexMerge(*pObj, indexName, branches);
  SphinxManageTest::testIndexStart(*pObj, indexName);

  // make rename index
  SphinxManageTest::testIndexRename(*pObj, indexName, renameIndexName);
  ASSERT_FALSE(pObj->isError()) << printError(*pObj);
  SphinxManageTest::testIndexCheck(*pObj, indexName, false);
  pObj->resetError();
  SphinxManageTest::testIndexCheck(*pObj, renameIndexName, true);
  SphinxManageTest::testIndexRemove(*pObj, renameIndexName);
  ASSERT_FALSE(pObj->isError()) << printError(*pObj);
}
//-----------------------------------------------------------------------------
//-----------------------------------------------------------------------------
TEST_F(SphinxAdminCommandTest, testIndexSetConfigVarGetConfigVar)
{
  std::cout << "HOME DIR: " << pObj->getHomeDir() << std::endl;

  SphinxManageTest::testIndexCreate(*pObj, renameIndexName);
  SphinxManageTest::testIndexCheck(*pObj, renameIndexName, true);
  branches.clear();
  branches.push_back("branch_0005");
  branches.push_back("branch_0006");
  branches.push_back("branch_0007");
  for (size_t i=0;i<branches.size();++i)
    SphinxManageTest::testIndexStoreDataFile(*pObj, renameIndexName, branches[i]);
  SphinxManageTest::testIndexStoreSchemaFile(*pObj, renameIndexName);
  SphinxManageTest::testIndexRebuild(*pObj, renameIndexName, branches);
  SphinxManageTest::testIndexMerge(*pObj, renameIndexName, branches);

  SphinxManageTest::testIndexStart(*pObj, renameIndexName);
  // make remove index
  SphinxManageTest::testIndexRemove(*pObj, renameIndexName);
  SphinxManageTest::testIndexCheck(*pObj, renameIndexName, false);

  SphinxManageTest::testIndexCreate(*pObj, indexName);
  branches.clear();
  branches.push_back("branch_0005");
  for (size_t i=0;i<branches.size();++i)
    SphinxManageTest::testIndexStoreDataFile(*pObj, indexName, branches[i]);
  SphinxManageTest::testIndexStoreSchemaFile(*pObj, indexName);
  SphinxManageTest::testIndexRebuild(*pObj, indexName, branches);
  SphinxManageTest::testIndexMerge(*pObj, indexName, branches);

  // set config property value
  SphinxManageTest::testIndexSetConfigVar(*pObj, indexName, "", "searchd", "test_param", "test_value1");
  SphinxManageTest::testIndexSetConfigVar(*pObj, indexName, "", "index branch_0001", "test_param", "test_value2");
  SphinxManageTest::testIndexSetConfigVar(*pObj, indexName, "branch_0005", "searchd", "test_param", "test_value1");
  SphinxManageTest::testIndexSetConfigVar(*pObj, indexName, "branch_0005", "index branch_0001", "test_param", "test_value2");
  // get config property value
  std::string paramValue;
  SphinxManageTest::testIndexGetConfigVar(*pObj, indexName, "", "searchd", "test_param", paramValue);
  ASSERT_FALSE(pObj->isError()) << printError(*pObj);
  ASSERT_EQ(paramValue, "test_value1");
  SphinxManageTest::testIndexGetConfigVar(*pObj, indexName, "", "index branch_0001", "test_param", paramValue);
  ASSERT_FALSE(pObj->isError()) << printError(*pObj);
  ASSERT_EQ(paramValue, "test_value2");
  SphinxManageTest::testIndexGetConfigVar(*pObj, indexName, "branch_0005", "searchd", "test_param", paramValue);
  ASSERT_FALSE(pObj->isError()) << printError(*pObj);
  ASSERT_EQ(paramValue, "test_value1");
  SphinxManageTest::testIndexGetConfigVar(*pObj, indexName, "branch_0005", "index branch_0001", "test_param", paramValue);
  ASSERT_FALSE(pObj->isError()) << printError(*pObj);
  ASSERT_EQ(paramValue, "test_value2");
  // check validation of schema data and save in index if valid
  SphinxManageTest::testIndexDeleteSchemaFile(*pObj, indexName);

  SphinxManageTest::testIndexRemove(*pObj, indexName);
  SphinxManageTest::testIndexCreate(*pObj, indexName);
  branches.clear();
  branches.push_back("branch_0005");
  for (size_t i=0;i<branches.size();++i)
    SphinxManageTest::testIndexStoreDataFile(*pObj, indexName, branches[i]);
  SphinxManageTest::testIndexStoreSchemaFile(*pObj, indexName);
  SphinxManageTest::testIndexRebuild(*pObj, indexName, branches);
  SphinxManageTest::testIndexMerge(*pObj, indexName, branches);
  SphinxManageTest::testIndexStart(*pObj, indexName);
  ASSERT_FALSE(pObj->isError()) << printError(*pObj);
  SphinxManageTest::testIndexRemove(*pObj, indexName);
  ASSERT_FALSE(pObj->isError()) << printError(*pObj);
}
//-----------------------------------------------------------------------------
//-----------------------------------------------------------------------------
TEST_F(SphinxAdminCommandTest, testIndexCheckSchema)
{
  SphinxManageTest::testIndexCreate(*pObj, indexName);
  SphinxManageTest::testIndexStoreSchemaFile(*pObj, indexName);
  SphinxManageTest::testIndexCheckSchema(*pObj, indexName);
  ASSERT_FALSE(pObj->isError());
  SphinxManageTest::testIndexRemove(*pObj, indexName);
  ASSERT_FALSE(pObj->isError()) << printError(*pObj);
}
//-----------------------------------------------------------------------------
//-----------------------------------------------------------------------------
TEST_F(SphinxAdminCommandTest, testIndexStatusSearchd)
{
  SphinxManageTest::testIndexCreate(*pObj, indexName);
  SphinxManageTest::testIndexCheck(*pObj, indexName, true);
  branches.clear();
  branches.push_back("branch_0005");
  for (size_t i=0;i<branches.size();++i)
    SphinxManageTest::testIndexStoreDataFile(*pObj, indexName, branches[i]);
  SphinxManageTest::testIndexStoreSchemaFile(*pObj, indexName);
  SphinxManageTest::testIndexRebuild(*pObj, indexName, branches);
  SphinxManageTest::testIndexMerge(*pObj, indexName, branches);
  // make copy index
  SphinxManageTest::testIndexStart(*pObj, indexName);
  SphinxManageTest::testIndexStatusSearchd(*pObj, indexName);
  ASSERT_FALSE(pObj->isError()) << printError(*pObj);
  SphinxManageTest::testIndexRemove(*pObj, indexName);
  ASSERT_FALSE(pObj->isError()) << printError(*pObj);
}
//-----------------------------------------------------------------------------
//-----------------------------------------------------------------------------
TEST_F(SphinxAdminCommandTest, testIndexStatus)
{
  SphinxManageTest::testIndexCreate(*pObj, indexName);
  SphinxManageTest::testIndexCheck(*pObj, indexName, true);
  branches.clear();
  branches.push_back("branch_0005");
  for (size_t i=0;i<branches.size();++i)
    SphinxManageTest::testIndexStoreDataFile(*pObj, indexName, branches[i]);
  SphinxManageTest::testIndexStoreSchemaFile(*pObj, indexName);
  SphinxManageTest::testIndexRebuild(*pObj, indexName, branches);
  SphinxManageTest::testIndexMerge(*pObj, indexName, branches);
  // make copy index
  SphinxManageTest::testIndexStart(*pObj, indexName);
  SphinxManageTest::testIndexStatus(*pObj, indexName);
  ASSERT_FALSE(pObj->isError()) << printError(*pObj);
  SphinxManageTest::testIndexRemove(*pObj, indexName);
  ASSERT_FALSE(pObj->isError()) << printError(*pObj);
}
//-----------------------------------------------------------------------------
//-----------------------------------------------------------------------------
TEST_F(SphinxAdminCommandTest, testIndexDataList)
{
  SphinxManageTest::testIndexCreate(*pObj, indexName);
  SphinxManageTest::testIndexCheck(*pObj, indexName, true);
  branches.clear();
  branches.push_back("branch_0005");
  for (size_t i=0;i<branches.size();++i)
    SphinxManageTest::testIndexStoreDataFile(*pObj, indexName, branches[i]);
  SphinxManageTest::testIndexStoreSchemaFile(*pObj, indexName);
  SphinxManageTest::testIndexRebuild(*pObj, indexName, branches);
  SphinxManageTest::testIndexMerge(*pObj, indexName, branches);
  // make copy index
  SphinxManageTest::testIndexStart(*pObj, indexName);
  SphinxManageTest::testIndexDataList(*pObj, indexName);
  ASSERT_FALSE(pObj->isError()) << printError(*pObj);
  SphinxManageTest::testIndexRemove(*pObj, indexName);
  ASSERT_FALSE(pObj->isError()) << printError(*pObj);
}
//-----------------------------------------------------------------------------
//-----------------------------------------------------------------------------
TEST_F(SphinxAdminCommandTest, testIndexBranchesList)
{
  SphinxManageTest::testIndexCreate(*pObj, indexName);
  SphinxManageTest::testIndexCheck(*pObj, indexName, true);
  branches.clear();
  branches.push_back("branch_0005");
  for (size_t i=0;i<branches.size();++i)
    SphinxManageTest::testIndexStoreDataFile(*pObj, indexName, branches[i]);
  SphinxManageTest::testIndexStoreSchemaFile(*pObj, indexName);
  SphinxManageTest::testIndexRebuild(*pObj, indexName, branches);
  SphinxManageTest::testIndexMerge(*pObj, indexName, branches);
  // make copy index
  SphinxManageTest::testIndexStart(*pObj, indexName);
  SphinxManageTest::testIndexBranchesList(*pObj, indexName);
  ASSERT_FALSE(pObj->isError()) << printError(*pObj);
  SphinxManageTest::testIndexRemove(*pObj, indexName);
  ASSERT_FALSE(pObj->isError()) << printError(*pObj);
}
//-----------------------------------------------------------------------------
//-----------------------------------------------------------------------------
TEST_F(SphinxAdminCommandTest, testIndexBranchesInfo)
{
  SphinxManageTest::testIndexCreate(*pObj, indexName);
  SphinxManageTest::testIndexCheck(*pObj, indexName, true);
  branches.clear();
  branches.push_back("branch_0005");
  for (size_t i=0;i<branches.size();++i)
    SphinxManageTest::testIndexStoreDataFile(*pObj, indexName, branches[i]);
  SphinxManageTest::testIndexStoreSchemaFile(*pObj, indexName);
  SphinxManageTest::testIndexRebuild(*pObj, indexName, branches);
  SphinxManageTest::testIndexMerge(*pObj, indexName, branches);
  // make copy index
  SphinxManageTest::testIndexStart(*pObj, indexName);
  SphinxManageTest::testIndexBranchesInfo(*pObj, indexName, branches);
  ASSERT_FALSE(pObj->isError()) << printError(*pObj);
  SphinxManageTest::testIndexRemove(*pObj, indexName);
  ASSERT_FALSE(pObj->isError()) << printError(*pObj);
}
//-----------------------------------------------------------------------------
//-----------------------------------------------------------------------------
TEST_F(SphinxAdminCommandTest, testIndexBranchesStatus)
{
  SphinxManageTest::testIndexCreate(*pObj, indexName);
  SphinxManageTest::testIndexCheck(*pObj, indexName, true);
  branches.clear();
  branches.push_back("branch_0005");
  for (size_t i=0;i<branches.size();++i)
    SphinxManageTest::testIndexStoreDataFile(*pObj, indexName, branches[i]);
  SphinxManageTest::testIndexStoreSchemaFile(*pObj, indexName);
  SphinxManageTest::testIndexRebuild(*pObj, indexName, branches);
  SphinxManageTest::testIndexMerge(*pObj, indexName, branches);
  SphinxManageTest::testIndexStart(*pObj, indexName);
  SphinxManageTest::testIndexBranchesStatus(*pObj, indexName, branches);
  ASSERT_FALSE(pObj->isError()) << printError(*pObj);
  SphinxManageTest::testIndexRemove(*pObj, indexName);
  ASSERT_FALSE(pObj->isError()) << printError(*pObj);
}
//-----------------------------------------------------------------------------
//-----------------------------------------------------------------------------
TEST_F(SphinxAdminCommandTest, testIndexConnectDisconnect)
{
  SphinxManageTest::testIndexCreate(*pObj, indexName);
  SphinxManageTest::testIndexCheck(*pObj, indexName, true);
  branches.clear();
  branches.push_back("branch_0005");
  for (size_t i=0;i<branches.size();++i)
    SphinxManageTest::testIndexStoreDataFile(*pObj, indexName, branches[i]);
  SphinxManageTest::testIndexStoreSchemaFile(*pObj, indexName);
  SphinxManageTest::testIndexRebuild(*pObj, indexName, branches);
  SphinxManageTest::testIndexMerge(*pObj, indexName, branches);
  SphinxManageTest::testIndexStart(*pObj, indexName);
  SphinxManageTest::testIndexDisconnect(*pObj);
  SphinxManageTest::testIndexConnect(*pObj, indexName);
  SphinxManageTest::testIndexRemove(*pObj, indexName);
  ASSERT_FALSE(pObj->isError()) << printError(*pObj);
  SphinxManageTest::testIndexCheck(*pObj, indexName, false);
  ASSERT_TRUE(pObj->isError());
}
//-----------------------------------------------------------------------------
//-----------------------------------------------------------------------------
TEST_F(SphinxAdminCommandTest, fullAdminCommandsTest)
{
  SphinxFunctionalObjectTest::fullAdminCommandsTest(*pObj);
  ASSERT_FALSE(pObj->isError()) << printError(*pObj);
}
//-----------------------------------------------------------------------------
//-----------------------------------------------------------------------------
} // end namespace tests
} // end namespace sphinx
} // end namespace HCE



