#include <Poco/SharedPtr.h>
#include <Poco/ScopedLock.h>
#include <Poco/ThreadPool.h>
#include <fstream>
#include <assert.h>
#include <iomanip>
#include <map>

#include "SphinxMultiIndexTest.hpp"
#include "SphinxManageTest.hpp"
#include "SphinxSearchTest.hpp"
#include "SphinxIndexTest.hpp"
#include "SphinxAdminCommand.hpp"
#include "SphinxInputJsonMessage.hpp"
#include "SphinxOutputJsonMessage.hpp"
#include "SphinxConfigCreator.hpp"
#include "SphinxInputJsonMessageSearch.hpp"
#include "SphinxPrintStatus.hpp"

namespace HCE
{
namespace sphinx
{
namespace tests
{
//-----------------------------------------------------------------------------
SearchTask::SearchTask(const std::string& indexName_, std::vector<std::string>& queries_, bool start, bool stop)
: inherited("search_"+indexName_), indexName(indexName_), queries(queries_),
  fObj("node_"+indexName_, sphinx_search_const::defaultDataDir, indexName_, start, stop), fps(0)
{
  if (!start)
    fObj.easyStart();
}
//-----------------------------------------------------------------------------
// cppcheck-suppress unusedFunction
void SearchTask::runTask(void)
{
  fps = 0;
  if (queries.size())
  {
    float localTime = 0;
    float progress = 0;
    const float progressStep = 100/queries.size();

    for (size_t i=0;i<queries.size();++i)
    {
      localTime += SphinxMultiIndexTest::testSearchIndex(fObj, indexName, queries[i]);
      progress += progressStep;
      setProgress(progress);
    }

    localTime /= 1000;
    localTime /= queries.size();
    fps = 1/localTime;
  }
}
//-----------------------------------------------------------------------------
//-----------------------------------------------------------------------------
void SearchProgressHandler::onProgress(Poco::TaskProgressNotification* pNf)
{
  std::cout << pNf->task()->name() << " progress: " << pNf->progress() << std::endl;
  pNf->release();
}
//-----------------------------------------------------------------------------
void SearchProgressHandler::onFinished(Poco::TaskFinishedNotification* pNf)
{
  std::cout << pNf->task()->name() << " finished." << std::endl;

  Poco::Mutex::ScopedLock lock(mutex);
  fps += static_cast<SearchTask*>(pNf->task())->getFps();

  pNf->release();
}
//-----------------------------------------------------------------------------
float SearchProgressHandler::getFps(void)
{
  Poco::Mutex::ScopedLock lock(mutex);
  return fps;
}
//-----------------------------------------------------------------------------
//-----------------------------------------------------------------------------
void SphinxMultiIndexTest::printSuccess(const std::string& msg, std::ostream& os)
{
  os << std::setw(40) << msg << " - OK\n";
}
//-----------------------------------------------------------------------------
void SphinxMultiIndexTest::printWarning(const std::string& msg, std::ostream& os)
{
  os << std::setw(40) << msg << " - WARNING\n";
}
//-----------------------------------------------------------------------------
//-----------------------------------------------------------------------------
void SphinxMultiIndexTest::testCreateIndex(SphinxFunctionalObject& fObj, const std::string& indexName, const std::string& dataContent)
{
  SphinxManageTest::testIndexCreate(fObj, indexName);
  SphinxManageTest::testIndexCheck(fObj, indexName, true);

  std::vector<std::string> branches;
  branches.push_back("branch_"+indexName);

  if (!dataContent.empty())
  {
    for (size_t i=0;i<branches.size();++i)
      SphinxManageTest::testIndexStoreDataFile(fObj, indexName, branches[i], dataContent);// add data content
    SphinxMultiIndexTest::printSuccess("Using data file content", fObj.log(HCE::LoggerStream::PRIO_INFORMATION));
  }
  else
  {
    for (size_t i=0;i<branches.size();++i)
      SphinxManageTest::testIndexStoreDataFile(fObj, indexName, branches[i]);
    SphinxMultiIndexTest::printWarning("Using internal test data content", fObj.log(HCE::LoggerStream::PRIO_INFORMATION));
  }

  SphinxManageTest::testIndexStoreSchemaFile(fObj, indexName);
  SphinxManageTest::testIndexRebuild(fObj, indexName, branches);
  SphinxManageTest::testIndexMerge(fObj, indexName, branches);
}
//-----------------------------------------------------------------------------
unsigned int SphinxMultiIndexTest::testSearchIndex(SphinxFunctionalObject& fObj, const std::string& indexName, const std::string& query)
{
  std::string json;
  SphinxInputJsonMessageSearch msgSearch;
  msgSearch.setQueryString(query);
 // msgSearch.setMaxResultsNumber(5);
  msgSearch.addQueryParameters("queryId", "777");
  msgSearch.addQueryParameters("JsonType", "7");
  msgSearch.addQueryParameters("offset", "0");
  msgSearch.addQueryParameters("limit", "0");
  msgSearch.addQueryParameters("cutoff", "0");
  assert(msgSearch.serialize(json));
  printSuccess("Serialize SphinxInputJsonMessageSearch", fObj.log(HCE::LoggerStream::PRIO_INFORMATION));

  SphinxInputJsonMessage handler;
  handler.setType(SphinxInputJsonMessage::MessageType::mtSearch);
  handler.setData(json);

  std::stringstream str;
  str << handler;

  std::string ret = fObj.Process(str.str());

//  std::cout << "ret: " << ret << std::endl;
  if (!fObj.getErrorMsg().empty())
    std::cout << "fObj.getErrorMsg()  = " << fObj.getErrorMsg() << std::endl;

  if (!fObj.getErrorMsg().empty())
    fObj.log(HCE::LoggerStream::PRIO_INFORMATION) << "fObj.getErrorMsg()  = " << fObj.getErrorMsg() << std::endl;
  assert(fObj.getErrorMsg().empty());

  SphinxOutputJsonMessage outputMessage(ret);
  assert(!outputMessage.isError());

  fObj.log(HCE::LoggerStream::PRIO_INFORMATION) << "Time = " << outputMessage.getTime() << std::endl;
  printSuccess("Search Process Query", fObj.log(HCE::LoggerStream::PRIO_INFORMATION));
  return outputMessage.getTime();
}
//-----------------------------------------------------------------------------
void SphinxMultiIndexTest::makeIndexNames(size_t count, std::vector<std::string>& names)
{
  names.clear();
  for (size_t i=1;i<=count;++i)
  {
    std::stringstream indexName;
    indexName << "i";
    indexName.width(3);
    indexName.fill('0');
    indexName << i;
    names.push_back(indexName.str());
  }
}
//-----------------------------------------------------------------------------
void SphinxMultiIndexTest::testCreateIndexes(const std::string& homeDir, size_t count, const std::string& dataFileName, std::vector<std::string>& names)
{
  std::string dataContent;
  std::ifstream ifs(dataFileName.c_str(), std::fstream::binary);
  if (ifs.is_open())
  {
    std::streambuf* pbuf = ifs.rdbuf();
    std::streamsize size = pbuf->pubseekoff(0,ifs.end);
    pbuf->pubseekoff(0,ifs.beg);       // rewind
    dataContent.resize(size);
    pbuf->sgetn (const_cast<char*>(dataContent.c_str()), size);
    ifs.close();
    printSuccess("Load data file for test '"+dataFileName+"'");
  }

  SphinxFunctionalObject fObj("test_node", homeDir);
//  fObj.loadConfiguration(sphinx_search_const::defaultIndexName);
//  fObj.setLogStream(std::cout);
//  if (!fObj.isLoadedConfig())
//    std::cout << "load config: '" << fObj.getErrorMsg() << "' errorCode = " << fObj.getErrorCode() << std::endl;
//  assert(fObj.isLoadedConfig());

  SphinxMultiIndexTest::makeIndexNames(count, names);

  for (size_t i=0;i<names.size();++i)
  {
    SphinxMultiIndexTest::testCreateIndex(fObj, names[i], dataContent);
  }
}
//-----------------------------------------------------------------------------
float SphinxMultiIndexTest::testSearchIndexes(std::vector<std::string>& names, std::vector<std::string> queries)
{
  assert(names.size());
  Poco::ThreadPool threadPool(1, names.size());
  Poco::TaskManager tm(threadPool);
  SearchProgressHandler pm;
  tm.addObserver(Poco::Observer<SearchProgressHandler, Poco::TaskProgressNotification>(pm, &SearchProgressHandler::onProgress));
  tm.addObserver(Poco::Observer<SearchProgressHandler, Poco::TaskFinishedNotification>(pm, &SearchProgressHandler::onFinished));

  for (size_t i=0;i<names.size();++i)
  {
    tm.start(new SearchTask(names[i], queries));
  }
  tm.joinAll();

  return pm.getFps();
}
//-----------------------------------------------------------------------------
void SphinxMultiIndexTest::setIndexProperty(const std::string& homeDir, const std::string& indexName, const IndexProperty& property) throw (Poco::Exception)
{
  std::string propFile = homeDir+"/"+sphinx_search_const::defaultDataDir+"/"+indexName+"/"+sphinx_search_const::sphinxPropertyFile;
  Poco::AutoPtr<SphinxConfigCreator> pConf = new SphinxConfigCreator(propFile);

  std::vector<std::string> keys;
  pConf->keys(keys);

  for (size_t i=0;i<keys.size();++i)
  {
    if (keys[i].find(sphinx_search_const::indexNameField)!=std::string::npos)
    {
      std::string name = keys[i]+"."+sphinx_search_const::pathNameField;
      pConf->setString(name, property.path);
    }
  }
  pConf->setString(sphinx_search_const::searchdListen, property.listen);

  pConf->save(propFile);
  pConf->setPropertyFileConfiguration(propFile);

  std::string sphinxConfig = sphinx_search_const::defaultDataDir+"/"+indexName+"/"+sphinx_search_const::sphinxConfigName;
  std::ofstream ofs(sphinxConfig.c_str(), std::fstream::trunc);
  ofs << *pConf;
  ofs.close();
}
//-----------------------------------------------------------------------------
void SphinxMultiIndexTest::getIndexProperty(const std::string& homeDir, const std::string& indexName, IndexProperty& property) throw (Poco::Exception)
{
  std::string propFile = homeDir+"/"+sphinx_search_const::defaultDataDir+"/"+indexName+"/"+sphinx_search_const::sphinxPropertyFile;
  Poco::AutoPtr<SphinxConfigCreator> pConf = new SphinxConfigCreator(propFile);

  std::vector<std::string> keys;
  pConf->keys(keys);

  for (size_t i=0;i<keys.size();++i)
  {
    if (keys[i].find(sphinx_search_const::indexNameField)!=std::string::npos)
    {
      std::string name = keys[i]+"."+sphinx_search_const::pathNameField;
      property.path = pConf->getString(name);
      break;
    }
  }
  property.listen = pConf->getString(sphinx_search_const::searchdListen);
}
//-----------------------------------------------------------------------------
float SphinxMultiIndexTest::testSearchIndexes(const std::string& homeDir, std::vector<std::string>& names, std::vector<std::string> queries, const std::string& indexName)
{
  assert(names.size());

  Poco::ThreadPool threadPool(1, names.size());
  Poco::TaskManager tm(threadPool);
  SearchProgressHandler pm;
  tm.addObserver(Poco::Observer<SearchProgressHandler, Poco::TaskProgressNotification>(pm, &SearchProgressHandler::onProgress));
  tm.addObserver(Poco::Observer<SearchProgressHandler, Poco::TaskFinishedNotification>(pm, &SearchProgressHandler::onFinished));

  try
  {
    IndexProperty property;
    SphinxMultiIndexTest::getIndexProperty(homeDir, indexName, property); // get value for applicable to continue next

    // correct configuration files
    std::map<std::string, IndexProperty> Map;
    for (size_t i=0;i<names.size();++i)
    {
      if (indexName==names[i])
        continue;
      IndexProperty localProperty;
      SphinxMultiIndexTest::getIndexProperty(homeDir, names[i], localProperty);
      Map[names[i]] = localProperty;
      SphinxMultiIndexTest::setIndexProperty(homeDir, names[i], property);
    }

    // next make search
    tm.start(new SearchTask(indexName, queries, true, false));
    for (size_t i=0;i<names.size();++i)
    {
      if (indexName==names[i])
        continue;
      tm.start(new SearchTask(names[i], queries, false, false));
    }
    tm.joinAll();

    SphinxFunctionalObject fObj("node_"+indexName, homeDir, indexName, false, true);

    // restore configuration files
    for (std::map<std::string, IndexProperty>::iterator iter=Map.begin();iter!=Map.end();++iter)
    {
//      std::cout << "[" << (*iter).first << "] = " << (*iter).second.path << "\t" << (*iter).second.listen << std::endl;
      SphinxMultiIndexTest::setIndexProperty(homeDir, (*iter).first, (*iter).second);
    }
  }
  catch(Poco::Exception& e)
  {
    std::cout << "Error ("<< e.code() << "): " << e.message() << std::endl;
  }
  return pm.getFps();
}
//-----------------------------------------------------------------------------
void SphinxMultiIndexTest::testRemoveIndexes(const std::string& homeDir, std::vector<std::string>& names)
{
  SphinxFunctionalObject fObj("test_node", homeDir);//, sphinx_search_const::defaultIndexName);
  for (size_t i=0;i<names.size();++i)
  {
    SphinxManageTest::testIndexRemove(fObj, names[i]);
  }
  names.clear();
}
//-----------------------------------------------------------------------------
//-----------------------------------------------------------------------------
} // namespace tests
} // namespace sphinx 
} // namespace HCE 
